--[[
			Advanced Tools for Mission Editor - ATME
			Author : Sunski34
			
			Testers : Snowsniper/Sunski34

			Oriented Object lua core module
			
			V.2.0.1 - Beta Version - 06/2021 - Known issue
				-> getParking ... Regarder bug confirmé dans DCS, remonté support ED.


			DO NOT MODIFY ANYTHING - For any idea or bug, please send a message on the forum
			
			Notes :
			-----
			ATME function initObjectsInMission is based on existing MIST function to initialize mission context
			
			All comments in that file are in french because of ATMECore software debugging. If you want 
			informations about ATME, classes, functions, events or handlers, please refer to english documentation.
]]--


local versionATME = "V2.0.1"


----------------------------------------------------------------------------------------------------------
-- DEFINITION DES VALEURS CONSTANTES & VARIABLES GLOBALES
----------------------------------------------------------------------------------------------------------

-- Données locales à ATME y compris descripteurs de classe créés ultérieurement
local ATMEDatas = {
	exitATME = false,					-- Flag de sortie d'ATME, en cas d'erreur. cf ATMEFct.errorOutputCore
	debugSwitch = false,				-- Activation ou non du mode affichage debug pour le module Core
	error = nil,						-- Erreur : code en cas de retour nil sur une fonction.
	files = {},							-- si le fichier MissionScripting.lua de DCS ..\Scripts a été modifié,
										-- tableau de gestion des fichiers
	language = "FR",					-- Langue utilisée par défaut
	isRunning = false,					-- isRunning est à false jusqu'au lancement par ATME.run après chargement des modules
	theatreName = "",
	logDisplay = false,					-- Log niveau 1 affiché ou non à l'écran
	currentModuleName = "",				-- Nom du module courant : Mis à jour lors des appels aux events ou handlers d'un module
	localMembers = {},
	allGroupsInMission = {},			-- Fixe les groupes définis dans la mission
	allAIUnitsInMission = {},			-- Fixe les unités AI définis dans la mission
	allStaticsInMission = {},			-- Fixe les objets définis dans la mission
	allCargosInMission = {},			-- Fixe les cargos définis dans la mission
	allStaticAirbasesInMission = {},	-- Fixe et pas de spawn sur les airfields et FARP
	unitsOutOfControl = {},				-- Liste des unités d'où est sorti un joueur
	playersControl = {}, 				-- Liste des joueurs sortants d'une unité
	players = {},						-- Liste des joueurs actifs
	flagDeleteUnit = {},				-- Liste flags de destruction pour les unités
	playerNamesByGroupId = {},			-- Liste des joueurs par leur groupId (A REFAIRE)
	reservedUnitNames = {},				-- Liste indexée par nom des unités actives ou embarquées. Ces noms sont réservés
	reservedGroupNames = {},			-- Liste indexée par nom des groupes actifs ou embarqués. Ces noms sont réservés
	reservedAsPlayers = {},				-- Liste indexée par nom des unités joueurs définies dans la mission. Ces noms sont réservés
	reservedAsPlayerGroups = {},		-- Liste indexée par nom des groupes avec au moins un joueur défini dans la mission. Ces noms sont réservés
	nbPlayers = 0,
	groups = {},						-- Groupes actifs
	units = {},							-- Unités actives
	spawnedGroups = {},					-- Table de gestion des id XXX max de groupes avec suffixe "- #XXX", une entrée pour un nom de groupe donné
	spawnedUnits = {},					-- Table de gestion des id XXX max de unités avec suffixe "- #XXX", une entrée pour un nom d'unité donné
	spawnedStatics = {},				-- Table de gestion des id XXX max d'objets statiques avec suffixe "- #XXX", une entrée pour un nom d'object statique donné
	spawnedCargos = {},					-- Table de gestion des id XXX max des cargos avec suffixe "- #XXX", une entrée pour un nom de cargos donné
	repeatSpawnGroupId = 1,            	-- Nombre total de groupes spawnés en répétition automatique : permet de gérer l'ID repeat des groupes et unités
	statics = {},						-- Objets statiques actifs
	cargos = {},						-- Cargos actifs
	soundsOn = {						-- Nécessaire à la gestion des diffusions de sons
		coalitions = {},
		countries = {},
		groups = {},
	},
	airbasesById = {},					-- Liste des airbases du théatre indexés par ID (y compris FARP et ship)
	airbases = {
		["SHIP"] = {},
		["HELIPAD"] = {},
		["AIRBASE"] = {},
	},									-- Liste des airbases (y compris FARP et ship)
	modules = {},						-- Liste nommée des modules (utilisée pour ATME & les classes disposant de la variable modules)
	modulesLocalMembers = {},			-- Liste des localMembers des modules, indexés par le nom du module.
	modulesNames = {},					-- Liste indexée stockant les noms des modules de la mission
	refModules = {},					-- Table référencant les noms de fichier temporaires des modules dans la mission
	readyToBoardGroups = {},			-- Liste des groupes prêts à embarquer
	reservedBoardingGroup = {},			-- Liste des groupes réservés par un transport et en attente/en cours d'embarquement
	respawnedGroups = {},				-- Liste des noms de groupes à respawn auto ayant été détruits
	spawnedGroupsWaypoints = {},		-- Liste pour mémoriser les WP des groupes spawnés.
	groupsOnBoard = {},					-- Liste des groupes embarqués, en cas de destruction du transport
	groupsToRespawn = {}, 				-- Liste des groupes à respawner au prochain cycle timer
	staticsToBirth = {},				-- Liste des objets statiques spawnés et non encore créés suite à S_EVENT_BIRTH
	cargosToBirth = {},					-- Liste des cargos spawnés et non encore créés suite à S_EVENT_BIRTH
	spawnGroupDatasRepeat = {},			-- Liste des données pour les groupes datas à multiple spawn
	idDCSRadioTransmissions = 0,		-- Compteur de création des radios transmissions
	markId = 0,							-- Identifiant pour les marqueurs de la carte F10
	marks = {},							-- Liste des marqueurs
	totalCharge = 0,					-- Valeur pour le calcul de charge dans le main timer pour lisser (traitements sécables)
	radioFrequencies = {},				-- Liste des radio-fréquences indexées par fréquence et modulation
	unitsReadyToCreate = {},			-- Unités ayant déclenché un S_EVENT_BIRTH pendant un addGroup. Parfois le groupe DCS n'est
										-- alors pas encore initialisé. Aussi, elles sont ensuite intégrées à ATME dès la fin d'exécution
										-- de addgroup.
	groupsReadyToBirth = {},			-- Pour les groupes aircraft, le S_EVENT_BIRTH est déclenché bien après le addGroup. Cette table
										-- permet la mémorisation des informations qui sera utilisée lors du S_EVENT_BIRTH
	unitsReadyToBirth = {},				-- Pour les groupes aircrafts, les unités associées sont mémorisées pour éviter les doublons
	newCreatedGroups = {},				-- Liste de gestion des groupes créés et/ou spawnés entre deux cycles timer. Ces listes sont
										-- mises à jour dans la fonction lors de la création d'un groupe pour ATMEDatas.newCreatedGroups. 
										-- Cette table est aussi utilisée pour le spawn avec une key isSpawned mise à jour.
	newCreatedCargos = {},				-- Idem pour les cargos
	newCreatedStatics = {},				-- Idem pour les statics
	newPreparedGroups = {},				-- Liste des groupes préparés, nécessaire lors de l'initialisation
	newPreparedCargos = {},				-- Idem pour les cargos
	newPreparedStatics = {},			-- idem pour les statics
	unitsToDisable = {},
	groupsToDisable = {},
	staticsToDisable = {},
	cargosToDisable = {},
	startOffset = 0,					-- Délai d'attente au démarrage avant initialisation
	startMissionOk = 0,					-- Vaut 0 avant initialisation, 1 lorsque tout est ok et avant onStartMissionHandler, et 2 après.
	weapons = {},						-- Table des armes tirées non impactées ou désactivées
	weaponsToDelete = {},				-- Table des armes avant impactée (S_EVENT_HIT) pour délai avant destruction ATME -> Cause S_EVENT_BDA
	sceneryObjects = {},				-- Table des objets de scène créés
}

-- Tables des fonctions locales à ATME core

local ATMEAIIndFct = {}
local ATMEFct = {}
local ATMETimers = {}

local ATMEConst = {
	smokeColors = {
		["GREEN"] = trigger.smokeColor.Green,
		["RED"] = trigger.smokeColor.Red,
		["WHITE"] = trigger.smokeColor.White,
		["ORANGE"] = trigger.smokeColor.Orange,
		["BLUE"] = trigger.smokeColor.Blue,
		["RANDOM"] = -1,
	},
	flareColors = {
		["GREEN"] = trigger.flareColor.Green,
		["RED"] = trigger.flareColor.Red,
		["WHITE"] = trigger.flareColor.White,
		["YELLOW"] = trigger.flareColor.Yellow,
		["RANDOM"] = -1,
	},
	flareColorsIndex = {
		"GREEN",
		"RED",
		"WHITE",
		"YELLOW"
	},
	smokeColorsIndex = {
		"GREEN",
		"RED",
		"WHITE",
		"ORANGE",
		"BLUE",
	},
	surfaceType = {
		["ALL"] = true,
		["LAND"] = true,
		["WATER"] = true,
	},
	bigSmokeType = {
		["SMALL"] = 5,
		["MEDIUM"] = 6,
		["LARGE"] = 7,
		["HUGE"] = 8,
	},
	groupCategories = {
		["plane"] = Group.Category.AIRPLANE,
		["helicopter"] = Group.Category.HELICOPTER,
		["vehicle"] = Group.Category.GROUND,
		["ship"] = Group.Category.SHIP,
	},
	skill = {
		["RANDOM"] = "Random",
		["AVERAGE"] = "Average", 
		["GOOD"] = "Good",
		["HIGH"] = "High", 
		["EXCELLENT"] = "Excellent",
	},
	JTACStatus = {		-- Status pour test JTAC dans C_Task_getTaskDatas
		["plane"] = 2,
		["helicopter"] = 2,
		["vehicle"] = 1,
		["ship"] = 0,
	},
	categoryNames = {
		[Group.Category.AIRPLANE] = "AIRPLANE",
		[Group.Category.HELICOPTER] = "HELICOPTER",
		[Group.Category.GROUND] = "GROUND",
		[Group.Category.SHIP] = "SHIP",
	},
	categoryNameFilters = {
		["AIRPLANE"] = "plane",
		["HELICOPTER"] = "helicopter",
		["GROUND"] = "vehicle",
		["SHIP"] = "ship",
	},
	canFly = {
		["AIRPLANE"] = true,
		["HELICOPTER"] = true,
		["GROUND"] = false,
		["SHIP"] = false,
	},
	altRouteSet = {
		["plane"] = false,
		["helicopter"] = false,
		["vehicle"] = true,
		["ship"] = true,
	},
	fctRadius = {
		["AIRPLANE"] = function (_action, _velocity, _inAir)
			if _inAir == true then
				local v2 = _velocity.x * _velocity.x + _velocity.z * _velocity.z
				local factor = 1 + v2 / 19600	-- explications : hausse au carré de v en ratio 140m/s qui donne environ 500km/h
				
				if action == "Fly Over Point" then
					return 200 * factor
				else
					return 1500 * factor
				end
			else
				return 50
			end
		end,
		["HELICOPTER"] = function (_action, _velocity, _inAir)
			if _inAir == true then
				local v2 = _velocity.x * _velocity.x + _velocity.z * _velocity.z
				local factor = 1 + v2 / 3600	-- explications : hausse au carré de v en ratio 60m/s qui donne environ 216km/h
				
				if action == "Fly Over Point" then
					return 100 * factor
				else
					return 250 * factor
				end
			else
				return 50
			end
		end,
		["GROUND"] = function (_action, _velocity, _inAir)
			if action == "On Road" then
				return 20
			else
				return 40
			end
		end,
		["SHIP"] = function (_action, _velocity, _inAir)
			return 100
		end,
	},
	coalitionNames = {
		[coalition.side.RED] = "RED",
		[coalition.side.BLUE] = "BLUE",
		[coalition.side.NEUTRAL] = "NEUTRAL",
		[3] = "NEUTRAL",							-- Bug DCS si l'aéroport est à la fois bleu et rouge en début de mission
	},
	coalitionIds = {
		["RED"] = coalition.side.RED,
		["BLUE"] = coalition.side.BLUE,
		["NEUTRAL"] = coalition.side.NEUTRAL,
	},
	coalitionIdsBlueRed = {
		["RED"] = coalition.side.RED,
		["BLUE"] = coalition.side.BLUE,
	},
	coalitionEnemy = {
		["BLUE"] = "RED",
		["RED"] = "BLUE",
	},
	personnelCarrierAttributes = {
		["UH-1H"] = {maxUnits = 8, minDistance = 25,},
		["Mi-8MT"] = {maxUnits = 16, minDistance = 28,},
		["SA342M"] = {maxUnits = 2, minDistance = 22,},
		["SA342L"] = {maxUnits = 2, minDistance = 22,},
		["SA342Mistral"] = {maxUnits = 2, minDistance = 22,},
		["AAV7"] = {maxUnits = 25, minDistance = 20,},
		["M-113"] = {maxUnits = 11, minDistance = 20,},
		["LAV-25"] = {maxUnits = 6, minDistance = 20,},
		["M1126 Stryker ICV"] = {maxUnits = 9, minDistance = 20,},
		["M-2 Bradley"] = {maxUnits = 6, minDistance = 20,},
		["BTR-80"] = {maxUnits = 9, minDistance = 20,},
		["BTR_D"] = {maxUnits = 12, minDistance = 20,},
		["MTLB"] = {maxUnits = 10, minDistance = 20,},
		["BMD-1"] = {maxUnits = 4, minDistance = 20,},
		["BMP-1"] = {maxUnits = 8, minDistance = 20,},
		["BMP-2"] = {maxUnits = 7, minDistance = 20,},
		["BMP-3"] = {maxUnits = 7, minDistance = 20,},
		["UAZ-469"] = {maxUnits = 6, minDistance = 20,},
		["Tigr_233036"] = {maxUnits = 6, minDistance = 20,},
		["GAZ-3307"] = {maxUnits = 16, minDistance = 20,},
		["GAZ-3308"] = {maxUnits = 16, minDistance = 20,},
		["Ural-4320T"] = {maxUnits = 20, minDistance = 20,},
		["Ural-4320-31"] = {maxUnits = 20, minDistance = 20,},
		["M 818"] = {maxUnits = 20, minDistance = 20,},
		["KAMAZ Truck"] = {maxUnits = 20, minDistance = 20,},
		["Bedford_MWD"] = {maxUnits = 10, minDistance = 20,},
		["Land_Rover_109_S3"] = {maxUnits = 10, minDistance = 20,},
		["CCKW_353"] = {maxUnits = 20, minDistance = 20,},
		["LAZ Bus"] = {maxUnits = 30, minDistance = 20,},
		["LiAZ Bus"] = {maxUnits = 30, minDistance = 20,},
		["Willys_MB"] = {maxUnits = 5, minDistance = 20,},
		["Sd_Kfz_7"] = {maxUnits = 7, minDistance = 20,},
		["Horch_901_typ_40_kfz_21"] = {maxUnits = 4, minDistance = 20,},
		["Sd_Kfz_2"] = {maxUnits = 2, minDistance = 20,},
		["Kubelwagen_82"] = {maxUnits = 4, minDistance = 20,},
		["M4_Tractor"] = {maxUnits = 6, minDistance = 20,},
		["Land_Rover_101_FC"] = {maxUnits = 10, minDistance = 20,},
		["Blitz_36-6700A"] = {maxUnits = 12, minDistance = 20,},
		["VAZ Car"] = {maxUnits = 4, minDistance = 20,},
		["Trolley bus"] = {maxUnits = 30, minDistance = 20,},
		["IKARUS Bus"] = {maxUnits = 30, minDistance = 20,},
		["M2A1_halftrack"] = {maxUnits = 8, minDistance = 20,},
		["TPZ"] = {maxUnits = 10, minDistance = 20,},
	},
	staticCategories = {
		-- Bug DCS sur certains objets dont le nom est toujours 'static'
		-- LTAvehicles/Effects ne fonctionnent pas en spawn.
		-- Catégorie Heliports considérée comme Airbase donc non traitée ici.
		
		["SHIP"]= {
			category = "Ships",
		},
		["AIRPLANE"]= {
			category = "Planes",
		},
		["HELICOPTER"]= {
			category = "Helicopters",
		},
		["WAREHOUSE"]= {
			category = "Warehouses",
		},
		["GROUND"]= {
			category = "Unarmed",
		},
		["FORTIFICATION"]= {
			category = "Fortifications",
		},
		["ADEQUIPMENT"]= {
			category = "ADEquipment",
		},
		["PERSONNEL"]= {
			category = "Personnel",
		},
		["ANIMAL"] = {
			category = "Animal",
		},
	},
	backStaticObjectCategories = {
		["Ships"] = "SHIP",
		["Planes"] = "AIRPLANE",
		["Helicopters"] = "HELICOPTER",
		["Unarmed"] = "GROUND",
		["Warehouses"] = "WAREHOUSE",
		["Fortifications"] = "FORTIFICATION",
		["ADEquipment"] = "ADEQUIPMENT",
		["Personnel"] = "PERSONNEL",
		["Animal"] = "ANIMAL",
	},
	supportedLanguages = {
		["FR"] = "French",
		["EN"] = "English",
		["RUS"] = "Russian",
		["GER"] = "German",
	},
	specialAircraftTypes = {
		["E-3A"] = "awacs",
		["E-2C"] = "awacs",
		["KC-135"] = "tankers",
		["KC135MPRS"] = "tankers",
		["KC130"] = "tankers",
		["A-10A"] = "A10",
		["A-10C"] = "A10",
		["A-10C_2"] = "A10",
	},
	blueCallsigns = {
		["aircrafts"] = {
			["ENFIELD"] = {
				name = "Enfield",
				id = 1,
				type = "standard",
			},
			["SPRINGFIELD"] = {
				name = "SpringField",
				id = 2,
				type = "standard",
			},
			["UZI"] = {
				name = "Uzi",
				id = 3,
				type = "standard",
			},
			["COLT"] = {
				name = "Colt",
				id = 4,
				type = "standard",
			},
			["DODGE"] = {
				name = "Dodge",
				id = 5,
				type = "standard",
			},
			["FORD"] = {
				name = "Ford",
				id = 6,
				type = "standard",
			},
			["CHEVY"] = {
				name = "Chevy",
				id = 7,
				type = "standard",
			},
			["PONTIAC"] = {
				name = "Pontiac",
				id = 8,
				type = "standard",
			},
			["HAWG"] = {
				name = "Hawg",
				id = 9,
				type = "A10",
			},
			["BOAR"] = {
				name = "Boar",
				id = 10,
				type = "A10",
			},
			["PIG"] = {
				name = "Pig",
				id = 11,
				type = "A10",
			},
			["TUSK"] = {
				name = "Tusk",
				id = 12,
				type = "A10",
			},
			["TEXACO"] = {
				name = "Texaco",
				id = 1,
				type = "tankers",
			},
			["ARCO"] = {
				name = "Arco",
				id = 2,
				type = "tankers",
			},
			["SHELL"] = {
				name = "Shell",
				id = 3,
				type = "tankers",
			},
			["OVERLORD"] = {
				name = "Overlord",
				id = 1,
				type = "awacs",
			},
			["MAGIC"] = {
				name = "Magic",
				id = 2,
			},
			["WIZARD"] = {
				name = "Wizard",
				id = 3,
				type = "awacs",
			},
			["FOCUS"] = {
				name = "Focus",
				id = 4,
				type = "awacs",
			},
			["DARKSTAR"] = {
				name = "Darkstar",
				id = 5,
				type = "awacs",
			},
		},
		["jtacs"] = {
			["AXEMAN"] = {
				name = "Axeman",
				id = 1,
			},
			["DARKNIGHT"] = {
				name = "Darknight",
				id = 2,
			},
			["WARRIOR"] = {
				name = "Warrior",
				id = 3,
			},
			["POINTER"] = {
				name = "Pointer",
				id = 4,
			},
			["EYEBALL"] = {
				name = "Eyeball",
				id = 5,
			},
			["MOONBEAM"] = {
				name = "Moonbeam",
				id = 6,
			},
			["WHIPLASH"] = {
				name = "Whiplash",
				id = 7,
			},
			["FINGER"] = {
				name = "Finger",
				id = 8,
			},
			["PINPOINT"] = {
				name = "Pinpoint",
				id = 9,
			},
			["FERRET"] = {
				name = "Ferret",
				id = 10,
			},
			["SHABA"] = {
				name = "Shaba",
				id = 11,
			},
			["PLAYBOY"] = {
				name = "Playboy",
				id = 12,
			},
			["HAMMER"] = {
				name = "Hammer",
				id = 13,
			},
			["JAGUAR"] = {
				name = "Jaguar",
				id = 14,
			},
			["DEATHSTAR"] = {
				name = "Deathstar",
				id = 15,
			},
			["ANVIL"] = {
				name = "Anvil",
				id = 16,
			},
			["FIREFLY"] = {
				name = "Firefly",
				id = 17,
			},
			["MANTIS"] = {
				name = "Mantis",
				id = 18,
			},
			["BADGER"] = {
				name = "Badger",
				id = 19,
			},
		},
	},
	moduleHandlers = {
		["onCreateUnitHandler"] = true,
		["onDeleteUnitHandler"] = true,
		["onBaseCapturedUnitHandler"] = true,
		["onCreateGroupHandler"] = true,
		["onSpawnGroupHandler"] = true,
		["onDeleteGroupHandler"] = true,
		["onCreateStaticHandler"] = true,
		["onDeleteStaticHandler"] = true,
		["onSpawnStaticHandler"] = true,
		["onCreateCargoHandler"] = true,
		["onDeleteCargoHandler"] = true,
		["onSpawnCargoHandler"] = true,
		["onTimerHandler"] = true,
		["onStartMissionHandler"] = true,
		["onEndMissionHandler"] = true,
	},
	directionValue = {					-- Table de gestion des directions
		[-1] = "RIGHT_TO_LEFT",
		[1] = "LEFT_TO_RIGHT",
	},
	weaponFlags = {
		["NO_WEAPON"] = 0,
		["LGB"] = 2,
		["TVGB"] = 4,
		["SNSGB"] = 8,
		["HE_BOMB"] = 16,
		["PENETRATOR"] = 32,
		["NAPALM_BOMB"] = 64,
		["FAE_BOMB"] = 128,
		["CLUSTER_BOMB"] = 256,
		["DISPENCER"] = 512,
		["CANDLE_BOMB"] = 1024,
		["PARACHUTE_BOMB"] = 2147483648,
		["LIGHT_ROCKET"] = 2048,
		["MARKER_ROCKET"] = 4096,
		["CANDLE_ROCKET"] = 8192,
		["HEAVY_ROCKET"] = 16384,
		["ANTI_RADAR_MISSILE"] = 32768,
		["ANTI_SHIP_MISSILE"] = 65536,
		["ANTI_TANK_MISSILE"] = 131072,
		["FIRE_AND_FORGET_ASM"] = 262144,
		["LASER_ASM"] = 524288,
		["TELE_ASM"] = 1048576,
		["CRUISE_MISSILE"] = 2097152,
		["ANTI_RADAR_MISSILE_2"] = 1073741824,
		["LR_AAM"] = 16777216,
		["SR_AAM"] = 4194304,
		["MR_AAM"] = 8388608,
		["IR_AAM"] = 33554432,
		["SAR_AAM"] = 67108864,
		["AR_AAM"] = 134217728,
		["GUN_POD"] = 268435456,
		["BUILT_IN_CANNON"] = 536870912,
		["ANY_WEAPON"] = 3221225470,
		["MARKER_WEAPON"] = 13312,
		["ARM_WEAPON"] = 3221212158,
	},
	weaponExpend = {
		["QUARTER"] = AI.Task.WeaponExpend.QUARTER,
		["TWO"] = AI.Task.WeaponExpend.TWO,
		["ONE"] = AI.Task.WeaponExpend.ONE,
		["FOUR"] = AI.Task.WeaponExpend.FOUR,
		["HALF"] = AI.Task.WeaponExpend.HALF,
		["ALL"] = AI.Task.WeaponExpend.ALL,
	},
	orbitPattern = {
		["CIRCLE"] = AI.Task.OrbitPattern.CIRCLE,
		["RACE_TRACK"] = AI.Task.OrbitPattern.RACE_TRACK,
	},
	designationAI = {
		["NO"] = AI.Task.Designation.NO,
		["WP"] = AI.Task.Designation.WP,
		["IR_POINTER"] = AI.Task.Designation.IR_POINTER,
		["LASER"] = AI.Task.Designation.LASER,
		["AUTO"] = AI.Task.Designation.AUTO,
	},
	aircraftTypeJTAC = {
		["RQ-1A Predator"] = true,
		["MQ-9 Reaper"] = true,
	},
	offsetTracking = {
		["C_Unit"] = 1,
		["C_Weapon"] = 0.2,
	},
	targetsAttributes = {
		["PLANE_CARRIERS"] = "plane_carrier",
		["AA_FLAK"] = "AA_flak",
		["AA_MISSILES"] = "AA_missile",
		["CRUISE_MISSILES"] = "Cruise missiles",
		["ANTI_SHIP_MISSILES"] = "Anti-Ship missiles",
		["MISSILES"] = "Missiles",
		["AIR"] = "Air",
		["AIRPLANES"] = "Planes",
		["UAVS"] = "UAVs",
		["HELICOPTERS"] = "Helicopters",
		["BATTLE_AIRPLANES"] = "Battle airplanes",
		["ATTACK_HELICOPTERS"] = "Attack helicopters",
		["FIGHTERS"] = "Fighters",
		["INTERCEPTORS"]= "Interceptors",
		["BOMBERS"] = "Bombers",
		["STRATEGIC_BOMBERS"] = "Strategic Bombers",
		["MULTIROLE_FIGHTERS"] = "Multirole fighters",
		["TRANSPORT_AIRPLANES"] = "Transports",
		["TRANSPORT_HELICOPTERS"] = "Transport helicopters",
		["AWACS"] = "AWACS",
		["TANKERS"] = "Tankers",
		["AUX"] = "Aux",
		["CARS"] = "Cars",
		["TRUCKS"] = "Trucks",
		["INFANTRY"] = "Infantry",
		["TANKS"] = "Tanks",
		["ARTILLERY"] = "Artillery",
		["MLRS"] = "MLRS",
		["IFV"] = "IFV",
		["APC"] = "APC",
		["FORTIFICATIONS"] = "Fortifications",
		["ARMED_VEHICLES"] = "Armed vehicles",
		["STATIC_AAA"] = "Static AAA",
		["MOBILE_AAA"]= "Mobile AAA",
		["SAM_SR"] = "SAM SR",
		["SAM_TR"] = "SAM TR",
		["SAM_LL"] = "SAM LL",
		["SAM_CC"] = "SAM CC",
		["SAM_AUX"] = "SAM AUX",
		["SR_SAM"] = "SR SAM",
		["MR_SAM"] = "MR SAM",
		["LR_SAM"] = "LR SAM",
		["SAM_ELEMENTS"] = "SAM Elements",
		["IR_GUIDED_SAM"] = "IR Guided SAM",
		["SAM"] = "SAM",
		["SAM_RELATED"] = "SAM related",
		["AAA"] = "AAA",
		["EWR"] = "EWR",
		["AIR_DEFENCE_VEHICLE"] = "Air Defence vehicle",
		["MANPADS"] = "MANPADS",
		["MANPADS_AUX"] = "MANPADS AUX",
		["UNARMED_VEHICLES"] = "Unarmed vehicles",
		["ARMED_GROUND_UNITS"] = "Armed ground units",
		["ARMED_AIR_DEFENCE"] = "Armed Air Defence",
		["AIR_DEFENCE"] = "Air defence",
		["AIRCRAFT_CARRIERS"] = "Aircraft Carriers",
		["CRUISERS"] = "Cruisers",
		["DESTROYERS"] = "Destroyers",
		["FRIGATES"] = "Frigates",
		["CORVETTES"] = "Corvettes",
		["SUBMARINES"] = "Submarines",
		["HEAVY_ARMED_SHIPS"] = "Heavy armed ships",
		["LIGHT_ARMED_SHIPS"] = "Light armed ships",
		["ARMED_SHIPS"] = "Armed ships",
		["UNARMED_SHIPS"] = "Unarmed ships",
		["GROUND_VEHICLES"] = "Ground vehicles",
		["SHIPS"] = "Ships",
		["BUILDINGS"] = "Buildings",
		["HEAVY_ARMORED_UNITS"] = "HeavyArmoredUnits",
		["LIGHT_ARMORED_UNITS"] = "LightArmoredUnits",
		["ATGM"] = "ATGM",
		["OLD_TANKS"] = "Old Tanks",
		["MODERN_TANKS"] = "Modern Tanks",
		["ROCKET_ATTACK_VALID_AIR_DEFENCE"] = "Rocket Attack Valid AirDefence",
		["ALL"] = "All",
		["INFANTRY_CARRIERS"] = "Infantry carriers",
		["VEHICLES"] = "Vehicles",
		["GROUND_UNITS"] = "Ground Units",
		["GROUND_UNITS_NON_AIR_DEFENCE"] = "Ground Units Non Airdefence",
		["ARMORED_VEHICLES"] = " Armored vehicles",
		["ANTI_AIR_ARMED_VEHICLES"] = "AntiAir Armed Vehicles",
		["AIRBASES"] = "Airfields",
		["HELIPADS"] = "Helipad",
		["GRASS_AIRFIELDS"] = "Grass Airfields",
		["POINT"] = "Point",
		["NON_ARMORED_UNITS"] = "NonArmoredUnits",
		["NON_AND_LIGHT_ARMORED_UNITS"] = "NonAndLightArmoredUnits",
		["HUMAN_VEHICLE"] = "Human_vehicle",
		["RADAR_BAND1_FOR_ARM"] = "RADAR_BAND1_FOR_ARM",
		["RADAR_BAND2_FOR_ARM"] = "RADAR_BAND2_FOR_ARM",
		["PRONE"] = "Prone",
		["DETECTION_BY_AWACS"] = "DetectionByAWACS",
		["DATALINK"] = "Datalink",
		["CUSTOM_AIM_POINT"] = "CustomAimPoint",
		["INDIRECT_FIRE"] = "Indirect fire",
		["REFUELABLE"] = "Refuelable",
		["WEAPON"] = "Weapon",
	},
	backTargetsAttributes = {
		["plane_carrier"] = {
			value = "PLANE_CARRIERS",
			priority = 0,
		},
		["AA_flak"] = {
			value = "AA_FLAK",
			priority = 0,
		},
		["AA_missile"] = {
			value = "AA_MISSILES",
			priority = 0,
		},
		["Cruise missiles"] = {
			value = "CRUISE_MISSILES",
			priority = 0,
		},
		["Anti-Ship missiles"] = {
			value = "ANTI_SHIP_MISSILES",
			priority = 0,
		},
		["Missiles"] = {
			value = "MISSILES",
			priority = 0,
			optionTarget = true,
		},
		["Air"] = {
			value = "AIR",
			priority = 0,
			airTarget = true,
			target = true,
			optionTarget = true,
		},
		["Planes"] = {
			value = "AIRPLANES",
			priority = 0,
			airTarget = true,
			optionTarget = true,
		},
		["UAVs"] = {
			value = "UAVS",
			priority = 0,
		},
		["Helicopters"] = {
			value = "HELICOPTERS",
			priority = 0,
			airTarget = true,
			target = true,
			optionTarget = true,
		},
		["Battle airplanes"] = {
			value = "BATTLE_AIRPLANES",
			priority = 2,
		},
		["Attack helicopters"] = {
			value = "ATTACK_HELICOPTERS",
			priority = 2,
		},
		["Fighters"] = {
			value = "FIGHTERS",
			priority = 0,
			airTarget = true,
			optionTarget = true,
		},
		["Interceptors"] = {
			value = "INTERCEPTORS",
			priority = 0,
		},
		["Bombers"] = {
			value = "BOMBERS",
			priority = 0,
			airTarget = true,
			optionTarget = true,
		},
		["Strategic Bombers"] = {
			value = "STRATEGIC_BOMBERS",
			priority = 0,
		},
		["Multirole fighters"] = {
			value = "MULTIROLE_FIGHTERS",
			priority = 0,
			airTarget = true,
			optionTarget = true,
		},
		["Transports"] = {
			value = "TRANSPORT_AIRPLANES",
			priority = 0,
		},
		["Transport helicopters"] = {
			value = "TRANSPORT_HELICOPTERS",
			priority = 0,
		},
		["AWACS"] = {
			value = "AWACS",
			priority = 0,
		},
		["Tankers"] = {
			value = "TANKERS",
			priority = 0,
		},
		["Aux"] = {
			value = "AUX",
			priority = 0,
		},
		["Cars"] = {
			value = "CARS",
			priority = 0,
		},
		["Trucks"] = {
			value = "TRUCKS",
			priority = 0,
		},
		["Infantry"] = {
			value = "INFANTRY",
			priority = 0,
			target = true,
			optionTarget = true,
		},
		["Tanks"] = {
			value = "TANKS",
			priority = 0,
			target = true,
			optionTarget = true,
		},
		["Artillery"] = {
			value = "ARTILLERY",
			priority = 0,
			target = true,
			optionTarget = true,
		},
		["MLRS"] = {
			value = "MLRS",
			priority = 0,
		},
		["IFV"] = {
			value = "IFV",
			priority = 0,
			target = true,
			optionTarget = true,
		},
		["APC"] = {
			value = "APC",
			priority = 0,
			target = true,
			optionTarget = true,
		},
		["Fortifications"] = {
			value = "FORTIFICATIONS",
			priority = 0,
			target = true,
		},
		["Armed vehicles"] = {
			value = "ARMED_VEHICLES",
			priority = 0,
		},
		["Static AAA"] = {
			value = "STATIC_AAA",
			priority = 0,
		},
		["Mobile AAA"] = {
			value = "MOBILE_AAA",
			priority = 0,
		},
		["SAM SR"] = {
			value = "SAM_SR",
			priority = 5,
		},
		["SAM TR"] = {
			value = "SAM_TR",
			priority = 9,
		},
		["SAM LL"] = {
			value = "SAM_LL",
			priority = 1,
		},
		["SAM CC"] = {
			value = "SAM_CC",
			priority = 0,
		},
		["SAM AUX"] = {
			value = "SAM_AUX",
			priority = 0,
		},
		["SR SAM"] = {
			value = "SR_SAM",
			priority = 6,
			target = true,
			optionTarget = true,
		},
		["MR SAM"] = {
			value = "MR_SAM",
			priority = 7,
			target = true,
			optionTarget = true,
		},
		["LR SAM"] = {
			value = "LR_SAM",
			priority = 8,
			target = true,
			optionTarget = true,
		},
		["SAM Elements"] = {
			value = "SAM_ELEMENTS",
			priority = 0,
		},
		["IR Guided SAM"] = {
			value = "IR_GUIDED_SAM",
			priority = 0,
		},
		["SAM"] = {
			value = "SAM",
			priority = 0,
			target = true,
			optionTarget = true,
		},
		["SAM related"] = {
			value = "SAM_RELATED",
			priority = 0,
		},
		["AAA"] = {
			value = "AAA",
			priority = 4,
			target = true,
			optionTarget = true,
		},
		["EWR"] = {
			value = "EWR",
			priority = 5,
		},
		["Air Defence vehicle"] = {
			value = "AIR_DEFENCE_VEHICLE",
			priority = 0,
		},
		["MANPADS"] = {
			value = "MANPADS",
			priority = 0,
		},
		["MANPADS AUX"] = {
			value = "MANPADS_AUX",
			priority = 0,
		},
		["Unarmed vehicles"] = {
			value = "UNARMED_VEHICLES",
			priority = 1,
		},
		["Armed ground units"] = {
			value = "ARMED_GROUND_UNITS",
			priority = 0,
		},
		["Armed Air Defence"] = {
			value = "ARMED_AIR_DEFENCE",
			priority = 0,
		},
		["Air defence"] = {
			value = "AIR_DEFENCE",
			priority = 0,
			target = true,
			optionTarget = true,
		},
		["Aircraft Carriers"] = {
			value = "AIRCRAFT_CARRIERS",
			priority = 0,
			optionTarget = true,
		},
		["Cruisers"] = {
			value = "CRUISERS",
			priority = 0,
			optionTarget = true,
		},
		["Destroyers"] = {
			value = "DESTROYERS",
			priority = 0,
			optionTarget = true,
		},
		["Frigates"] = {
			value = "FRIGATES",
			priority = 0,
			optionTarget = true,
		},
		["Corvettes"] = {
			value = "CORVETTES",
			priority = 0,
			optionTarget = true,
		},
		["Submarines"] = {
			value = "SUBMARINES",
			priority = 0,
			optionTarget = true,
		},
		["Heavy armed ships"] = {
			value = "HEAVY_ARMED_SHIPS",
			priority = 0,
			optionTarget = true,
		},
		["Light armed ships"] = {
			value = "LIGHT_ARMED_SHIPS",
			priority = 0,
			target = true,
			optionTarget = true,
		},
		["Armed ships"] = {
			value = "ARMED_SHIPS",
			priority = 0,
			optionTarget = true,
		},
		["Unarmed ships"] = {
			value = "UNARMED_SHIPS",
			priority = 0,
			target = true,
			optionTarget = true,
		},
		["Ground vehicles"] = {
			value = "GROUND_VEHICLES",
			priority = 0,
		},
		["Ships"] = {
			value = "SHIPS",
			priority = 0,
			target = true,
			optionTarget = true,
		},
		["Buildings"] = {
			value = "BUILDINGS",
			priority = 0,
		},
		["HeavyArmoredUnits"] = {
			value = "HEAVY_ARMORED_UNITS",
			priority = 0,
		},
		["LightArmoredUnits"] = {
			value = "LIGHT_ARMORED_UNITS",
			priority = 0,
		},
		["ATGM"] = {
			value = "ATGM",
			priority = 0,
		},
		["Old Tanks"] = {
			value = "OLD_TANKS",
			priority = 0,
		},
		["Modern Tanks"] = {
			value = "MODERN_TANKS",
			priority = 0,
		},
		["Rocket Attack Valid AirDefence"] = {
			value = "ROCKET_ATTACK_VALID_AIR_DEFENCE",
			priority = 0,
		},
		["All"] = {
			value = "ALL",
			priority = 0,
			target = true,
			optionTarget = true,
		},
		["Infantry carriers"] = {
			value = "INFANTRY_CARRIERS",
			priority = 0,
		},
		["Vehicles"] = {
			value = "VEHICLES",
			priority = 0,
		},
		["Ground Units"] = {
			value = "GROUND_UNITS",
			priority = 0,
			target = true,
		},
		["Ground Units Non Airdefence"] = {
			value = "GROUND_UNITS_NON_AIR_DEFENCE",
			priority = 0,
		},
		["Armored vehicles"] = {
			value = "ARMORED_VEHICLES",
			priority = 3,
		},
		["AntiAir Armed Vehicles"] = {
			value = "ANTI_AIR_ARMED_VEHICLES",
			priority = 0,
		},
		["Airfields"] = {
			value = "AIRBASES",
			priority = 0,
		},
		["Helipad"] = {
			value = "HELIPADS",
			priority = 0,
		},
		["Grass Airfields"] = {
			value = "GRASS_AIRFIELDS",
			priority = 0,
		},
		["Point"] = {
			value = "POINT",
			priority = 0,
		},
		["NonArmoredUnits"] = {
			value = "NON_ARMORED_UNITS",
			priority = 0,
		},
		["NonAndLightArmoredUnits"] = {
			value = "NON_AND_LIGHT_ARMORED_UNITS",
			priority = 0,
		},
		["Human_vehicle"] = {
			value = "HUMAN_VEHICLE",
			priority = 0,
		},
		["RADAR_BAND1_FOR_ARM"] = {
			value = "RADAR_BAND1_FOR_ARM",
			priority = 0,
		},
		["RADAR_BAND2_FOR_ARM"] = {
			value = "RADAR_BAND2_FOR_ARM",
			priority = 0,
		},
		["Prone"] = {
			value = "PRONE",
			priority = 0,
		},
		["DetectionByAWACS"] = {
			value = "DETECTION_BY_AWACS",
			priority = 0,
		},
		["Datalink"] = {
			value = "DATALINK",
			priority = 0,
		},
		["CustomAimPoint"] = {
			value = "CUSTOM_AIM_POINT",
			priority = 0,
		},
		["Indirect fire"] = {
			value = "INDIRECT_FIRE",
			priority = 0,
		},
		["Refuelable"] = {
			value = "REFUELABLE",
			priority = 0,
		},
		["Weapon"] = {
			value = "WEAPON",
			priority = 0,
		},
	},
	findCategories = {
		["AIRPLANE"] = true,
		["HELICOPTER"] = true,
		["GROUND"] = true,
		["SHIP"] = true,
		["HELIPAD"] = true,
		["AIRBASE"] = true,
	},
	WPActions = {
		["plane"] = {
			["TURNING_POINT"] = "Turning Point",
			["FLY_OVER_POINT"] = "Fly Over Point",
		},
		["helicopter"] = {
			["TURNING_POINT"] = "Turning Point",
			["FLY_OVER_POINT"] = "Fly Over Point",
		},
		["vehicle"] = {
			["VEE"] = "Vee",
			["ECHELON_RIGHT"] = "EchelonR",
			["OFF_ROAD"] = "Off Road",
			["RANK"] = "Rank",
			["ECHELON_LEFT"]  = "EchelonL",
			["ON_ROAD"] = "On Road",
			["CONE"] = "Cone",
			["DIAMOND"] = "Diamond",
		},
		["ship"] = {
			["TURNING_POINT"] = "Turning Point",
		},
	},
	mainTask = {
		["NOTHING"] = "Nothing",
		["GROUND_ATTACK"] = "Ground Attack",
		["AFAC"] = "AFAC",
		["ANTI_SHIP_STRIKE"] = "Antiship Strike",
		["AWACS"] = "AWACS",
		["CAP"] = "CAP",
		["CAS"] = "CAS",
		["ESCORT"] = "Escort",
		["FIGHTER_SWEEP"] = "Fighter Sweep",
		["INTERCEPT"] = "Intercept",
		["PIN_POINT_STRIKE"] = "Pinpoint Strike",
		["RECONNAISSANCE"] = "Reconnaissance",
		["REFUELING"] = "Refueling",
		["RUNWAY_ATTACK"] = "Runway Attack",
		["SEAD"] = "SEAD",
		["TRANSPORT"] = "Transport",
	},
	taskAI = {
		["REFUELING"] = { -- Main task
			id = "Refueling",
			filters = {
				type = {
					["plane"] = true,
				},
			},
			args = {
			},
		},
		["FOLLOW"] = { -- Main task
			id = "Follow",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
				},
			},
			args = {
				["group"] = {
					parameter = {
						name = "groupId",
						fct = "C_Group_getIDWithWP",
						linkedParams = {
							"lastWptIndexFlag",
						},
					},
					multi = false,
					type = "C_Group",
					mandatory = true,
				},
				["pos"] = {
					parameter = {
						name = "pos",
						fct = "C_Vector_getPos",
					},
					multi = false,
					type = "C_Vector3D",
					mandatory = true,
				},
				["lastWaypointID"] = {
					parameter = {
						name = "lastWptIndex",
						fct = "C_getWaypointNumberWithFlag",
						linkedParams = {
							"lastWptIndexFlag",
						},
					},
					multi = false,
					type = "number",
					min = 0,
					mandatory = false,
				},
			},
		},
		["FOLLOW_BIG_FORMATION"] = { -- Main task
			id = "FollowBigFormation",
			filters = {
				type = {
					["plane"] = true,
				},
			},
			args = {
				["group"] = {
					parameter = {
						name = "groupId",
						fct = "C_Group_getIDWithWP",
						linkedParams = {
							"lastWptIndexFlag",
						},
					},
					multi = false,
					type = "C_Group",
					mandatory = true,
				},
				["pos"] = {
					parameter = {
						name = "pos",
						fct = "C_Vector_getPos",
					},
					multi = false,
					type = "C_Vector3D",
					mandatory = true,
				},
				["lastWaypointID"] = {
					parameter = {
						name = "lastWptIndex",
						fct = "C_getWaypointNumberWithFlag",
						linkedParams = {
							"lastWptIndexFlag",
						},
					},
					multi = false,
					type = "number",
					min = 0,
					mandatory = false,
				},
			},
		},
		["ESCORT"] = { -- Main task
			id = "Escort",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
				},
			},
			args = {
				["group"] = {
					parameter = {
						name = "groupId",
						fct = "C_Group_getIDWithWP",
						linkedParams = {
							"lastWptIndexFlag",
						},
					},
					multi = false,
					type = "C_Group",
					mandatory = true,
				},
				["pos"] = {
					parameter = {
						name = "pos",
						fct = "C_Vector_getPos",
					},
					multi = false,
					type = "C_Vector3D",
					mandatory = true,
				},
				["maxDistance"] = {
					parameter = {
						name = "engagementDistMax",
					},
					multi = false,
					type = "number",
					min = 0,
					mandatory = true,
				},
				["targetsType"] = {
					parameter = {
						name = "targetTypes",
						fct = "C_getAirTargetsType",
					},
					type = "string",
					multi = true,
					mandatory = true,
				},
				["lastWaypointID"] = {
					parameter = {
						name = "lastWptIndex",
						fct = "C_getWaypointNumberWithFlag",
						linkedParams = {
							"lastWptIndexFlag",
						},
					},
					multi = false,
					type = "number",
					min = 0,
					mandatory = false,
				},
			},
		},
		["LAND"] = { -- Main task
			id = "Land",
			filters = {
				type = {
					["helicopter"] = true,
				},
			},
			args = {
				["point"] = {
					parameter = {
						name = "point",
						fct = "C_getPoint2D",
					},
					multi = false,
					type = "table",
					mandatory = true,
				},
				["duration"] = {
					parameter = {
						name = "duration",
						fct = "C_getNumberWithFlag",
						linkedParams = {
							"durationFlag",
						},
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = false,
				},
			},
		},
		["FIRE_AT_POINT"] = { -- Main task
			id = "FireAtPoint",
			filters = {
				type = {
					["vehicle"] = true,
					["ship"] = true,
				},
			},
			const = {
				["radius"] = 100,
			},
			fctValidate = "C_TaskValidate_Altitude",
			args = {
				["circle"] = {
					parameter = {
						name = "point",
						fct = "C_Circle_getCenterRadius",
						linkedParams = {
							"radius",
						},
					},
					multi = false,
					type = "C_Circle",
					mandatory = true,
				},
				["weaponType"] = {
					parameter = {
						name = "weaponType",
						fct = "C_getWeaponFlag",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["qty"] = {
					parameter = {
						name = "expendQty",
						fct = "C_getNumberWithFlag",
						linkedParams = {
							"expendQtyEnabled",
						},
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = false,
				},
				["altitude"] = {
					parameter = {
						name = "altitude",
					},
					multi = false,
					type = "number",
					min = 0,
					mandatory = false,
				},
				["altitudeType"] = {
					parameter = {
						name = "alt_type",
						fct = "C_getAltitudeType",
					},
					multi = false,
					type = "number",
					min = 0,
					mandatory = false,
				},
			},
		},
		["BOMBING"] = { -- Main task
			id = "Bombing",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
				},
			},
			args = {
				["point"] = {
					parameter = {
						name = "point",
						fct = "C_getPoint2D",
					},
					multi = false,
					type = "table",
					mandatory = true,
				},
				["qty"] = {
					parameter = {
						name = "attackQty",
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = true,
				},

				["weaponType"] = {
					parameter = {
						name = "weaponType",
						fct = "C_getWeaponFlag",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["expend"] = {
					parameter = {
						name = "expend",
						fct = "C_getExpend",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["direction"] = {
					parameter = {
						name = "direction",
					},
					multi = false,
					type = "number",
					mandatory = false,
				},
				["altitude"] = {
					parameter = {
						name = "altitude",
						fct = "C_getNumberWithFlag",
						linkedParams = {
							"altitudeEnabled",
						},
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = false,
				},
				["groupAttack"] = {
					parameter = {
						name = "groupAttack",
					},
					multi = false,
					type = "boolean",
					mandatory = false,
				},
			},
		},
		["BOMBING_RUNWAY"] = { -- Main task
			id = "BombingRunway",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
				},
			},
			args = {
				["airbase"] = {
					parameter = {
						name = "runwayId",
						fct = "C_getAirbase",
					},
					multi = false,
					type = "C_Airbase",
					mandatory = true,
				},
				["weaponType"] = {
					parameter = {
						name = "weaponType",
						fct = "C_getWeaponFlag",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["expend"] = {
					parameter = {
						name = "expend",
						fct = "C_getExpend",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["qty"] = {
					parameter = {
						name = "attackQty",
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = false,
				},
				["direction"] = {
					parameter = {
						name = "direction",
					},
					multi = false,
					type = "number",
					mandatory = false,
				},
				["groupAttack"] = {
					parameter = {
						name = "groupAttack",
					},
					multi = false,
					type = "boolean",
					mandatory = false,
				},
			},
		},
		["CARPET_BOMBING"] = { -- Main task
			id = "CarpetBombing",
			filters = {
				type = {
					["plane"] = true,
				},
			},
			const = {
				["attackType"] = "Carpet",
			},
			args = {
				["point"] = {
					parameter = {
						name = "point",
						fct = "C_getPoint2D",
					},
					multi = false,
					type = "table",
					mandatory = true,
				},
				["qty"] = {
					parameter = {
						name = "attackQty",
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = true,
				},
				["carpetLength"] = {
					parameter = {
						name = "carpetLength",
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = true,
				},
				["weaponType"] = {
					parameter = {
						name = "weaponType",
						fct = "C_getWeaponFlag",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["expend"] = {
					parameter = {
						name = "expend",
						fct = "C_getExpend",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["altitude"] = {
					parameter = {
						name = "altitude",
						fct = "C_getNumberWithFlag",
						linkedParams = {
							"altitudeEnabled",
						},
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = false,
				},
				["groupAttack"] = {
					parameter = {
						name = "groupAttack",
					},
					multi = false,
					type = "boolean",
					mandatory = false,
				},
			},
		},
		["ATTACK_MAP_OBJECT"] = { -- Main task
			id = "AttackMapObject",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
				},
			},
			args = {
				["point"] = {
					parameter = {
						name = "point",
						fct = "C_getPoint2D",
					},
					multi = false,
					type = "table",
					mandatory = true,
				},
				["weaponType"] = {
					parameter = {
						name = "weaponType",
						fct = "C_getWeaponFlag",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["expend"] = {
					parameter = {
						name = "expend",
						fct = "C_getExpend",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["qty"] = {
					parameter = {
						name = "attackQty",
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = false,
				},
				["direction"] = {
					parameter = {
						name = "direction",
					},
					multi = false,
					type = "number",
					mandatory = false,
				},
				["groupAttack"] = {
					parameter = {
						name = "groupAttack",
					},
					multi = false,
					type = "boolean",
					mandatory = false,
				},
			},
		},
		["ATTACK_GROUP"] = { -- Main task
			id = "AttackGroup",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
				},
			},
			args = {
				["group"] = {
					parameter = {
						name = "groupId",
						fct = "C_Group_getID",
					},
					multi = false,
					type = "C_Group",
					mandatory = true,
				},
				["weaponType"] = {
					parameter = {
						name = "weaponType",
						fct = "C_getWeaponFlag",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["expend"] = {
					parameter = {
						name = "expend",
						fct = "C_getExpend",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["qty"] = {
					parameter = {
						name = "attackQty",
						fct = "C_getNumberWithFlag",
						linkedParams = {
							"attackQtyLimit",
						},
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = false,
				},
				["direction"] = {
					parameter = {
						name = "direction",
						fct = "C_getNumberWithFlag",
						linkedParams = {
							"directionEnabled",
						},
					},
					multi = false,
					type = "number",
					mandatory = false,
				},
				["altitude"] = {
					parameter = {
						name = "altitude",
						fct = "C_getNumberWithFlag",
						linkedParams = {
							"altitudeEnabled",
						},
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = false,
				},
			},
		},
		["ATTACK_UNIT"] = { -- Main task
			id = "AttackUnit",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
				},
			},
			args = {
				["unit"] = {
					parameter = {
						name = "unitId",
						fct = "C_Unit_getID",
					},
					multi = false,
					type = "C_Unit",
					mandatory = true,
				},
				["weaponType"] = {
					parameter = {
						name = "weaponType",
						fct = "C_getWeaponFlag",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["expend"] = {
					parameter = {
						name = "expend",
						fct = "C_getExpend",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["qty"] = {
					parameter = {
						name = "attackQty",
						fct = "C_getNumberWithFlag",
						linkedParams = {
							"attackQtyLimit",
						},
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = false,
				},
				["direction"] = {
					parameter = {
						name = "direction",
						fct = "C_getNumberWithFlag",
						linkedParams = {
							"directionEnabled",
						},
					},
					multi = false,
					type = "number",
					mandatory = false,
				},
				["groupAttack"] = {
					parameter = {
						name = "groupAttack",
					},
					multi = false,
					type = "boolean",
					mandatory = false,
				},
			},
		},
		["FAC_ATTACK_GROUP"] = { -- Main task
			id = "FAC_AttackGroup",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
					["vehicle"] = true,
				},
			},
			args = {
				["frequency"] = {
					parameter = {
						name = "frequency",
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = true,
				},
				["modulation"] = {
					parameter = {
						name = "modulation",
						fct = "C_getModulation",
					},
					multi = false,
					type = "string",
					mandatory = true,
				},
				["callsign"] = {
					parameter = {
						name = "callname",
						fct = "C_getCallname",
						linkedParams = {
							"number",
						},
					},
					multi = false,
					type = "string",
					mandatory = true,
				},
				["group"] = {
					parameter = {
						name = "groupId",
						fct = "C_Group_getID",
					},
					multi = false,
					type = "C_Group",
					mandatory = true,
				},
				["weaponType"] = {
					parameter = {
						name = "weaponType",
						fct = "C_getWeaponFlag",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["designation"] = {
					parameter = {
						name = "designation",
						fct = "C_getDesignation",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["datalink"] = {
					parameter = {
						name = "datalink",
					},
					multi = false,
					type = "boolean",
					mandatory = false,
				},
			},
		},
		["ORBIT"] = { -- Main task
			id = "Orbit",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
				},
			},
			args = {
				["pattern"] = {
					parameter = {
						name = "pattern",
						fct = "C_getOrbitPattern",
					},
					multi = false,
					type = "string",
					mandatory = true,
				},
				["point"] = {
					parameter = {
						name = "point",
						fct = "C_getPoint2D",
					},
					multi = false,
					type = "table",
					mandatory = false,
				},
				["point2"] = {
					parameter = {
						name = "point2",
						fct = "C_getPoint2D",
					},
					multi = false,
					type = "table",
					mandatory = false,
				},
				["speed"] = {
					parameter = {
						name = "speed",
					},
					multi = false,
					min = 0,
					type = "number",
					mandatory = false,
				},
				["altitude"] = {
					parameter = {
						name = "altitude",
					},
					multi = false,
					min = 0,
					type = "number",
					mandatory = false,
				},
			},
		},
		["AWACS"] = { -- Enroute task
			id = "AWACS",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
				},
			},
			args = {
			},
		},
		["TANKER"] = { -- Enroute task
			id = "Tanker",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
				},
			},
			args = {
			},
		},
		["ENGAGE_GROUP"] = { -- Enroute task
			id = "EngageGroup",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
				},
			},
			args = {
				["group"] = {
					parameter = {
						name = "groupId",
						fct = "C_Group_getID",
					},
					multi = false,
					type = "C_Group",
					mandatory = true,
				},
				["priority"] = {
					parameter = {
						name = "priority",
					},
					multi = false,
					type = "number",
					min = 0,
					mandatory = false,
				},
				["weaponType"] = {
					parameter = {
						name = "weaponType",
						fct = "C_getWeaponFlag",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["expend"] = {
					parameter = {
						name = "expend",
						fct = "C_getExpend",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["qty"] = {
					parameter = {
						name = "attackQty",
						fct = "C_getNumberWithFlag",
						linkedParams = {
							"attackQtyLimit",
						},
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = false,
				},
				["direction"] = {
					parameter = {
						name = "direction",
					},
					multi = false,
					type = "number",
					mandatory = false,
				},
			},
		},
		["ENGAGE_TARGETS"] = { -- Enroute task
			id = "EngageTargets",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
				},
			},
			args = {
				["targetsType"] = {
					parameter = {
						name = "targetTypes",
						fct = "C_getTargetsType",
					},
					type = "string",
					multi = true,
					mandatory = true,
				},
				["priority"] = {
					parameter = {
						name = "priority",
					},
					multi = false,
					type = "number",
					min = 0,
					mandatory = false,
				},
				["maxDistance"] = {
					parameter = {
						name = "maxDist",
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = false,
				},
			},
		},
		["ENGAGE_TARGETS_IN_CIRCLE"] = { -- Enroute task
			id = "EngageTargetsInZone",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
				},
			},
			args = {
				["circle"] = {
					parameter = {
						name = "point",
						fct = "C_Circle_getCenterRadius",
						linkedParams = {
							"zoneRadius",
						},
					},
					multi = false,
					type = "C_Circle",
					mandatory = true,
				},
				["targetsType"] = {
					parameter = {
						name = "targetTypes",
						fct = "C_getTargetsType",
					},
					type = "string",
					multi = true,
					mandatory = true,
				},
				["priority"] = {
					parameter = {
						name = "priority",
					},
					multi = false,
					type = "number",
					min = 0,
					mandatory = false,
				},
			},
		},
		["ENGAGE_UNIT"] = { -- Enroute task
			id = "EngageUnit",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
				},
			},
			args = {
				["unit"] = {
					parameter = {
						name = "unitId",
						fct = "C_Unit_getID",
					},
					multi = false,
					type = "C_Unit",
					mandatory = true,
				},
				["priority"] = {
					parameter = {
						name = "priority",
					},
					multi = false,
					type = "number",
					min = 0,
					mandatory = false,
				},
				["weaponType"] = {
					parameter = {
						name = "weaponType",
						fct = "C_getWeaponFlag",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["expend"] = {
					parameter = {
						name = "expend",
						fct = "C_getExpend",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["qty"] = {
					parameter = {
						name = "attackQty",
						fct = "C_getNumberWithFlag",
						linkedParams = {
							"attackQtyLimit",
						},
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = false,
				},
				["direction"] = {
					parameter = {
						name = "direction",
						fct = "C_getNumberWithFlag",
						linkedParams = {
							"directionEnabled",
						},
					},
					multi = false,
					type = "number",
					mandatory = false,
				},
				["groupAttack"] = {
					parameter = {
						name = "groupAttack",
					},
					multi = false,
					type = "boolean",
					mandatory = false,
				},
			},
		},
		["FAC_ENGAGE_GROUP"] = { -- Enroute task
			id = "FAC_EngageGroup",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
					["vehicle"] = true,
				},
			},
			args = {
				["frequency"] = {
					parameter = {
						name = "frequency",
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = true,
				},
				["modulation"] = {
					parameter = {
						name = "modulation",
						fct = "C_getModulation",
					},
					multi = false,
					type = "string",
					mandatory = true,
				},
				["callsign"] = {
					parameter = {
						name = "callname",
						fct = "C_getCallname",
						linkedParams = {
							"number",
						},
					},
					multi = false,
					type = "string",
					mandatory = true,
				},
				["group"] = {
					parameter = {
						name = "groupId",
						fct = "C_Group_getID",
					},
					multi = false,
					type = "C_Group",
					mandatory = true,
				},
				["priority"] = {
					parameter = {
						name = "priority",
					},
					multi = false,
					type = "number",
					min = 0,
					mandatory = false,
				},
				["weaponType"] = {
					parameter = {
						name = "weaponType",
						fct = "C_getWeaponFlag",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["designation"] = {
					parameter = {
						name = "designation",
						fct = "C_getDesignation",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["datalink"] = {
					parameter = {
						name = "datalink",
					},
					multi = false,
					type = "boolean",
					mandatory = false,
				},
			},
		},
	},
	optionAI = {
		["plane"] = {
			["ROE"] = {
				id = AI.Option.Air.id.ROE,
				type = "string",
				enum = {
					["WEAPON_FREE"] = {
						value = AI.Option.Air.val.ROE.WEAPON_FREE,
						params = {},
					},
					["OPEN_FIRE_WEAPON_FREE"] = {
						value = AI.Option.Air.val.ROE.OPEN_FIRE_WEAPON_FREE,
						params = {},
					},
					["OPEN_FIRE"] = {
						value = AI.Option.Air.val.ROE.OPEN_FIRE,
						params = {},
					},
					["RETURN_FIRE"] = {
						value = AI.Option.Air.val.ROE.RETURN_FIRE,
						params = {},
					},
					["WEAPON_HOLD"] = {
						value = AI.Option.Air.val.ROE.WEAPON_HOLD,
						params = {},
					},
				},
			},
			["FORMATION"] = {
				id = AI.Option.Air.id.FORMATION,
				type = "string",
				enum = {
					["LINE_ABREAST"] = {
						value = 65536,
						params = {
							{
								enum = {
									["CLOSE"] = 1, 
									["OPEN"] = 2,
									["GROUP_CLOSE"] = 3,
								},
							},
						},
					},
					["TRAIL"] = 
					{
						value = 131072,
						params = {
							{
								enum = {
									["CLOSE"] = 1, 
									["OPEN"] = 2,
									["GROUP_CLOSE"] = 3,
								},
							},
						},
					}, 
					["WEDGE"] = {
						value = 196608,
						params = {
							{
								enum = {
									["CLOSE"] = 1, 
									["OPEN"] = 2,
									["GROUP_CLOSE"] = 3,
								},
							},
						},
					}, 
					["ECHELON_RIGHT"] = {
						value = 262144,
						params = {
							{
								enum = {
									["CLOSE"] = 1, 
									["OPEN"] = 2,
									["GROUP_CLOSE"] = 3,
								},
							},
						},
					}, 
					["ECHELON_LEFT"] = {
						value = 327680,
						params = {
							{
								enum = {
									["CLOSE"] = 1, 
									["OPEN"] = 2,
									["GROUP_CLOSE"] = 3,
								},
							},
						},
					}, 
					["FINGER_FOUR"] = {
						value = 393216,
						params = {
							{
								enum = {
									["CLOSE"] = 1, 
									["OPEN"] = 2,
									["GROUP_CLOSE"] = 3,
								},
							},
						},
					}, 
					["SPREAD_FOUR"] = {
						value = 458752,
						params = {
							{
								enum = {
									["CLOSE"] = 1, 
									["OPEN"] = 2,
									["GROUP_CLOSE"] = 3,
								},
							},
						},
					}, 
					["WW2_BOMBER_ELEMENT"] = {
						value = 786432,
						params = {
							{
								enum = {
									["CLOSE"] = 1, 
									["OPEN"] = 2,
								},
							},
						},
					},
					["WW2_BOMBER_ELEMENT_HEIGHT"] = {
						value = 851968,
						params = {},
					},
					["WW2_FIGHTER_VIC"] = {
						value = 917504,
						params = {
							{
								enum = {
									["CLOSE"] = 1, 
									["OPEN"] = 2,
								},
							},
						},
					},
				},
			},
			["REACTION_ON_THREAT"] = {
				id = AI.Option.Air.id.REACTION_ON_THREAT,
				type = "string",
				enum = {
					["NO_REACTION"] = {
						value = AI.Option.Air.val.REACTION_ON_THREAT.NO_REACTION,
						params = {},
					},
					["PASSIVE_DEFENCE"] = {
						value = AI.Option.Air.val.REACTION_ON_THREAT.PASSIVE_DEFENCE,
						params = {},
					},
					["EVADE_FIRE"] = {
						value = AI.Option.Air.val.REACTION_ON_THREAT.EVADE_FIRE,
						params = {},
					},
					["BYPASS_AND_ESCAPE"] = {
						value = AI.Option.Air.val.REACTION_ON_THREAT.BYPASS_AND_ESCAPE,
						params = {},
					},
					["ALLOW_ABORT_MISSION"] = {
						value = AI.Option.Air.val.REACTION_ON_THREAT.ALLOW_ABORT_MISSION,
						params = {},
					},
				},
			},
			["RADAR_USING"] = {
				id = AI.Option.Air.id.RADAR_USING,
				type = "string",
				enum = {
					["NEVER"] = {
						value = AI.Option.Air.val.RADAR_USING.NEVER,
						params = {},
					},
					["FOR_ATTACK_ONLY"] = {
						value = AI.Option.Air.val.RADAR_USING.FOR_ATTACK_ONLY,
						params = {},
					},
					["FOR_SEARCH_IF_REQUIRED"] = {
						value = AI.Option.Air.val.RADAR_USING.FOR_SEARCH_IF_REQUIRED,
						params = {},
					},
					["FOR_CONTINUOUS_SEARCH"] = {
						value = AI.Option.Air.val.RADAR_USING.FOR_CONTINUOUS_SEARCH,
						params = {},
					},
				},
			},
			["FLARE_USING"] = {
				type = "string",
				enum = {
					["NEVER"] = {
						value = AI.Option.Air.val.FLARE_USING.NEVER,
						params = {},
					},
					["AGAINST_FIRED_MISSILE"] = {
						value = AI.Option.Air.val.FLARE_USING.AGAINST_FIRED_MISSILE,
						params = {},
					},
					["FLYING_IN_SAM_WEZ"] = {
						value = AI.Option.Air.val.FLARE_USING.WHEN_FLYING_IN_SAM_WEZ,
						params = {},
					},
					["FLYING_NEAR_ENEMIES"] = {
						value = AI.Option.Air.val.FLARE_USING.WHEN_FLYING_NEAR_ENEMIES,
						params = {},
					},
				},
			},
			["ECM_USING"] = {
				id = AI.Option.Air.id.ECM_USING,
				type = "string",
				enum = {
					["NEVER"] = {
						value = AI.Option.Air.val.ECM_USING.NEVER_USE,
						params = {},
					},
					["ONLY_LOCKED_BY_RADAR"] = {
						value = AI.Option.Air.val.ECM_USING.USE_IF_ONLY_LOCK_BY_RADAR,
						params = {},
					},
					["DETECTED_LOCK_BY_RADAR"] = {
						value = AI.Option.Air.val.ECM_USING.USE_IF_DETECTED_LOCK_BY_RADAR,
						params = {},
					},
					["ALWAYS"] = {
						value = AI.Option.Air.val.ECM_USING.ALWAYS_USE,
						params = {},
					},
				},
			},
			["MISSILE_ATTACK"] = {
				id = AI.Option.Air.id.MISSILE_ATTACK,
				type = "string",
				enum = {
					["MAX_RANGE"] = {
						value = AI.Option.Air.val.MISSILE_ATTACK.MAX_RANGE,
						params = {},
					},
					["NEZ_RANGE"] = {
						value = AI.Option.Air.val.MISSILE_ATTACK.NEZ_RANGE,
						params = {},
					},
					["HALF_WAY_RMAX_NEZ"] = {
						value = AI.Option.Air.val.MISSILE_ATTACK.HALF_WAY_RMAX_NEZ,
						params = {},
					},
					["TARGET_THREATEST"] = {
						value = AI.Option.Air.val.MISSILE_ATTACK.TARGET_THREAT_EST,
						params = {},
					},
					["RANDOM_RANGE"] = {
						value = AI.Option.Air.val.MISSILE_ATTACK.RANDOM_RANGE,
						params = {},
					},
				},
			},
			["SILENCE"] = {
				id = AI.Option.Air.id.SILENCE,
				type = "boolean",
			},
			["RTB_BINGO"] = {
				id = AI.Option.Air.id.RTB_ON_BINGO,
				type = "boolean",
			},
			["PROHIBIT_AA"] = {
				id = AI.Option.Air.id.PROHIBIT_AA,
				type = "boolean",
			},
			["PROHIBIT_AG"] = {
				id = AI.Option.Air.id.PROHIBIT_AG,
				type = "boolean",
			},
			["PROHIBIT_AFTER_BURNER"] = {
				id = AI.Option.Air.id.PROHIBIT_AB,
				type = "boolean",
			},
			["PROHIBIT_JETTISON"] = {
				id = AI.Option.Air.id.PROHIBIT_JETT,
				type = "boolean",
			},
			["PROHIBIT_WP_REPORT"] = {
				id = AI.Option.Air.id.PROHIBIT_WP_PASS_REPORT,
				type = "boolean",
			},
			["RTB_OUT_OF_AMMO"] = {
				id = AI.Option.Air.id.RTB_ON_OUT_OF_AMMO,
				type = "string",
				fct = "C_getWeaponFlag",
				enum = {},
			},
			["RADIO_USAGE_KILL"] = {
				id = AI.Option.Air.id.OPTION_RADIO_USAGE_KILL,
				type = "string",
				fct = "C_getOptionTargetsType",
				enum = {},
			},
			["RADIO_USAGE_CONTACT"] = {
				id = AI.Option.Air.id.OPTION_RADIO_USAGE_CONTACT,
				type = "string",
				fct = "C_getOptionTargetsType",
				enum = {},
			},
			["RADIO_USAGE_ENGAGE"] = {
				id = AI.Option.Air.id.OPTION_RADIO_USAGE_ENGAGE,
				type = "string",
				fct = "C_getOptionTargetsType",
				enum = {},
			},
		},
		["helicopter"] = {
			["ROE"] = {
				id = AI.Option.Air.id.ROE,
				type = "string",
				enum = {
					["WEAPON_FREE"] = {
						value = AI.Option.Air.val.ROE.WEAPON_FREE,
						params = {},
					},
					["OPEN_FIRE_WEAPON_FREE"] = {
						value = AI.Option.Air.val.ROE.OPEN_FIRE_WEAPON_FREE,
						params = {},
					},
					["OPEN_FIRE"] = {
						value = AI.Option.Air.val.ROE.OPEN_FIRE,
						params = {},
					},
					["RETURN_FIRE"] = {
						value = AI.Option.Air.val.ROE.RETURN_FIRE,
						params = {},
					},
					["WEAPON_HOLD"] = {
						value = AI.Option.Air.val.ROE.WEAPON_HOLD,
						params = {},
					},
				},
			},
			["FORMATION"] = {
				id = AI.Option.Air.id.FORMATION,
				type = "string",
				enum = {
					["COLUMN"] = {
						value = 720896,
						params = {},
					},
					["WEDGE"] = {
						value = 524288,
						params = {},
					},
					["FRONT"] = {
						value = 655360,
						params = {
							{
								enum = {
									["LEFT"] = 256, 
									["RIGHT"] = 0,
								},
							},
							{
								enum = {
									["300"] = 1, 
									["600"] = 32 
								},
							},
						},
					},
					["ECHELON"] = {
						value = 589824,
						params = {
							{
								enum = {
									["LEFT"] = 256, 
									["RIGHT"] = 0,
								},
							},
							{
								enum = {
									["70"] = 1, 
									["300"] = 2, 
									["600"] = 3, 
								},
							},
						},
					},
				},
			},
			["REACTION_ON_THREAT"] = {
				id = AI.Option.Air.id.REACTION_ON_THREAT,
				type = "string",
				enum = {
					["NO_REACTION"] = {
						value = AI.Option.Air.val.REACTION_ON_THREAT.NO_REACTION,
						params = {},
					},
					["PASSIVE_DEFENCE"] = {
						value = AI.Option.Air.val.REACTION_ON_THREAT.PASSIVE_DEFENCE,
						params = {},
					},
					["EVADE_FIRE"] = {
						value = AI.Option.Air.val.REACTION_ON_THREAT.EVADE_FIRE,
						params = {},
					},
					["BYPASS_AND_ESCAPE"] = {
						value = AI.Option.Air.val.REACTION_ON_THREAT.BYPASS_AND_ESCAPE,
						params = {},
					},
					["ALLOW_ABORT_MISSION"] = {
						value = AI.Option.Air.val.REACTION_ON_THREAT.ALLOW_ABORT_MISSION,
						params = {},
					},
				},
			},
			["RADAR_USING"] = {
				id = AI.Option.Air.id.RADAR_USING,
				type = "string",
				enum = {
					["NEVER"] = {
						value = AI.Option.Air.val.RADAR_USING.NEVER,
						params = {},
					},
					["FOR_ATTACK_ONLY"] = {
						value = AI.Option.Air.val.RADAR_USING.FOR_ATTACK_ONLY,
						params = {},
					},
					["FOR_SEARCH_IF_REQUIRED"] = {
						value = AI.Option.Air.val.RADAR_USING.FOR_SEARCH_IF_REQUIRED,
						params = {},
					},
					["FOR_CONTINUOUS_SEARCH"] = {
						value = AI.Option.Air.val.RADAR_USING.FOR_CONTINUOUS_SEARCH,
						params = {},
					},
				},
			},
			["FLARE_USING"] = {
				type = "string",
				enum = {
					["NEVER"] = {
						value = AI.Option.Air.val.FLARE_USING.NEVER,
						params = {},
					},
					["AGAINST_FIRED_MISSILE"] = {
						value = AI.Option.Air.val.FLARE_USING.AGAINST_FIRED_MISSILE,
						params = {},
					},
					["FLYING_IN_SAM_WEZ"] = {
						value = AI.Option.Air.val.FLARE_USING.WHEN_FLYING_IN_SAM_WEZ,
						params = {},
					},
					["FLYING_NEAR_ENEMIES"] = {
						value = AI.Option.Air.val.FLARE_USING.WHEN_FLYING_NEAR_ENEMIES,
						params = {},
					},
				},
			},
			["ECM_USING"] = {
				id = AI.Option.Air.id.ECM_USING,
				type = "string",
				enum = {
					["NEVER"] = {
						value = AI.Option.Air.val.ECM_USING.NEVER_USE,
						params = {},
					},
					["ONLY_LOCKED_BY_RADAR"] = {
						value = AI.Option.Air.val.ECM_USING.USE_IF_ONLY_LOCK_BY_RADAR,
						params = {},
					},
					["DETECTED_LOCK_BY_RADAR"] = {
						value = AI.Option.Air.val.ECM_USING.USE_IF_DETECTED_LOCK_BY_RADAR,
						params = {},
					},
					["ALWAYS"] = {
						value = AI.Option.Air.val.ECM_USING.ALWAYS_USE,
						params = {},
					},
				},
			},
			["MISSILE_ATTACK"] = {
				id = AI.Option.Air.id.MISSILE_ATTACK,
				type = "string",
				enum = {
					["MAX_RANGE"] = {
						value = AI.Option.Air.val.MISSILE_ATTACK.MAX_RANGE,
						params = {},
					},
					["NEZ_RANGE"] = {
						value = AI.Option.Air.val.MISSILE_ATTACK.NEZ_RANGE,
						params = {},
					},
					["HALF_WAY_RMAX_NEZ"] = {
						value = AI.Option.Air.val.MISSILE_ATTACK.HALF_WAY_RMAX_NEZ,
						params = {},
					},
					["TARGET_THREATEST"] = {
						value = AI.Option.Air.val.MISSILE_ATTACK.TARGET_THREAT_EST,
						params = {},
					},
					["RANDOM_RANGE"] = {
						value = AI.Option.Air.val.MISSILE_ATTACK.RANDOM_RANGE,
						params = {},
					},
				},
			},
			["SILENCE"] = {
				id = AI.Option.Air.id.SILENCE,
				type = "boolean",
			},
			["RTB_BINGO"] = {
				id = AI.Option.Air.id.RTB_ON_BINGO,
				type = "boolean",
			},
			["PROHIBIT_AA"] = {
				id = AI.Option.Air.id.PROHIBIT_AA,
				type = "boolean",
			},
			["PROHIBIT_AG"] = {
				id = AI.Option.Air.id.PROHIBIT_AG,
				type = "boolean",
			},
			["PROHIBIT_JETTISON"] = {
				id = AI.Option.Air.id.PROHIBIT_JETT,
				type = "boolean",
			},
			["PROHIBIT_WP_REPORT"] = {
				id = AI.Option.Air.id.PROHIBIT_WP_PASS_REPORT,
				type = "boolean",
			},
			["RTB_OUT_OF_AMMO"] = {
				id = AI.Option.Air.id.RTB_ON_OUT_OF_AMMO,
				type = "string",
				fct = "C_getWeaponFlag",
				enum = {},
			},
		},
		["vehicle"] = {
			["ROE"] = {
				id = AI.Option.Ground.id.ROE,
				type = "string",
				enum = {
					["OPEN_FIRE"] = {
						value = AI.Option.Ground.val.ROE.OPEN_FIRE,
						params = {},
					},
					["RETURN_FIRE"] = {
						value = AI.Option.Ground.val.ROE.RETURN_FIRE,
						params = {},
					},
					["WEAPON_HOLD"] = {
						value = AI.Option.Ground.val.ROE.WEAPON_HOLD,
						params = {},
					},
				},
			},
			["ALARM_STATE"] = {
				id = AI.Option.Ground.id.ALARM_STATE,
				type = "string",
				enum = {
					["AUTO"] = {
						value = AI.Option.Ground.val.ALARM_STATE.AUTO,
						params = {},
					},
					["RED"] = {
						value = AI.Option.Ground.val.ALARM_STATE.RED,
						params = {},
					},
					["GREEN"] = {
						value = AI.Option.Ground.val.ALARM_STATE.GREEN,
						params = {},
					},
				},
			},
			["DISPERSE_ON_ATTACK"] = {
				id = AI.Option.Ground.id.DISPERSE_ON_ATTACK,
				type = "number",
				min = 0,
				max = 120,
			},
			["SILENCE"] = {
				id = AI.Option.Ground.id.SILENCE,
				type = "boolean",
			},
			["RANGE_RESTRICTION"] = {
				id = AI.Option.Ground.id.AC_ENGAGEMENT_RANGE_RESTRICTION,
				type = "number",
				min = 0,
				max = 100,
			},
			["ENGAGE_AIR_WEAPONS"] = {
				id = AI.Option.Ground.id.ENGAGE_AIR_WEAPONS,
				type = "boolean",
			},
		},
		["ship"] = {
			["ROE"] = {
				id = AI.Option.Naval.id.ROE,
				type = "string",
				enum = {
					["OPEN_FIRE"] = {
						value = AI.Option.Naval.val.ROE.OPEN_FIRE,
						params = {},
					},
					["RETURN_FIRE"] = {
						value = AI.Option.Naval.val.ROE.RETURN_FIRE,
						params = {},
					},
					["WEAPON_HOLD"] = {
						value = AI.Option.Naval.val.ROE.WEAPON_HOLD,
						params = {},
					},
				},
			},
			["SILENCE"] = {
				id = AI.Option.Naval.id.SILENCE,
				type = "boolean",
			},
		},
	},
	typesOfSensor = {
		["VISUAL"] = Controller.Detection.VISUAL,
		["OPTIC"] = Controller.Detection.OPTIC,
		["RADAR"] = Controller.Detection.RADAR,
		["IRST"] = Controller.Detection.IRST,
		["RWR"] = Controller.Detection.RWR,
		["DLINK"] = Controller.Detection.DLINK,
		["ALL"] = true,
	},
	cardinal = {
		["N"] = "NORTH",
		["S"] = "SOUTH",
		["E"] = "EAST",
		["W"] = "WEST",
	},
	weaponAttributes = {
		--categories = {
			"SHELL",
			"MISSILE",
			"ROCKET",
			"BOMB",
		--guidanceType = {
			"INS",
			"IR",
			"RADAR_ACTIVE",
			"RADAR_SEMI_ACTIVE",
			"RADAR_PASSIVE",
			"TV",
			"LASER",
			"TELE",
		-- missileCategories = {
			"AAM",
			"SAM",
			"BM",
			"ANTI_SHIP",
			"CRUISE",
			"OTHER",    
		--warheadTypes = {
			"AP",
			"HE",
			"SHAPED_CHARGE",
	},
	stackTypeExtract = {
		["HEAD"] = {
			load = "first",
			direction = "next",
			raz = "previous",
		},
		["TAIL"] = {
			load = "last",
			direction = "previous",
			raz = "next",
		},
	},
	polygonMaxVertices = 40,

	maxTimeSmoke = 300,
	stackTimerMaxTotalCharge = 120,		-- Valeurs de tuning pour optimiser la gestion du timer sur les traitements sécables (stacktimer)
	modulationValues = {
		["AM"] = radio.modulation.AM,
		["FM"] = radio.modulation.FM, 
	},
	indZoneConfig ={
		["C_Circle"] = {
			fctIsInside = "C_Circle_isInside",
			fctGetRandomPosition = "C_Circle_getRandomPosition",
			fctEvalSearchVolume = "C_Circle_searchVolume",
		},
		["C_Rectangle"] = {
			fctIsInside = "C_Rectangle_isInside",
			fctGetRandomPosition = "C_Rectangle_getRandomPosition",
			fctEvalSearchVolume = "C_Rectangle_searchVolume",
		},
		["C_Sphere"] = {
			fctIsInside = "C_Sphere_isInside",
			fctGetRandomPosition = "C_Sphere_getRandomPosition",
			fctEvalSearchVolume = "C_Sphere_searchVolume",
		},
		["C_Cylinder"] = {
			fctIsInside = "C_Cylinder_isInside",
			fctGetRandomPosition = "C_Cylinder_getRandomPosition",
			fctEvalSearchVolume = "C_Cylinder_searchVolume",
		},
		["C_Box"] = {
			fctIsInside = "C_Box_isInside",
			fctGetRandomPosition = "C_Box_getRandomPosition",
			fctEvalSearchVolume = "C_Box_searchVolume",
		},
		["C_PickupZone"] = {
			fctIsInside = "C_Circle_isInside",
			fctGetRandomPosition = "C_Circle_getRandomPosition",
			fctEvalSearchVolume = "C_Circle_searchVolume",
		},
		["C_Polygon"] = {
			fctIsInside = "C_Polygon_isInside",
			fctGetRandomPosition = "C_Polygon_getRandomPosition",
			fctEvalSearchVolume = "C_Polygon_searchVolume",
		},
	},
	randomSign = {
		[1] = -1,
		[2] = 1,
	},
	basicRouteDelay = 2,
	tokenTypes = {
		["RANDOM"] = {
			type = "TOKEN_TYPE_RANDOM",
		},
		["TRUE"] = {
			type = "TOKEN_INTEGER",
			value = 1,
		},
		["FALSE"] = {
			type = "TOKEN_INTEGER",
			value = 0,
		},
		["FLARE"] = {
			type = "TOKEN_SIGNAL_TYPE",
		},
		["SMOKE"] = {
			type = "TOKEN_SIGNAL_TYPE",
		},
		["WAYPOINT"] = {
			type = "TOKEN_TYPE_WAYPOINT",
		},
		["TIME"] = {
			type = "TOKEN_TYPE_TIME",
		},
		["MAX_TURNS"] = {
			type = "TOKEN_TYPE_MAX_TURNS",
		},
		["FLAG"] = {
			type = "TOKEN_TYPE_FLAG",
		},
		["FUEL_RATIO"] = {
			type = "TOKEN_TYPE_FUEL_RATIO",
		},
		["TACAN_AA"] = {
			type = "TOKEN_TYPE_TACAN_AA",
		},
		["TACAN"] = {
			type = "TOKEN_TYPE_TACAN",
		},
		["ICLS"] = {
			type = "TOKEN_TYPE_ICLS",
		}
	},
	relativeStatus = {
		"ABSOLUTE",
		"RELATIVE",
	},
	findClasses = { 
		["AIRBASES"] = { -- AIRBASES inopérant cause bug DCS pour recherche avec world.searchObject. Traitement spécifique
			findSubCategories = true,
		},
		["GROUPS"] = {
			findSubCategories = true,
			category = {Object.Category.UNIT},
			search_fct = function(_object, _datas)
				ATMEFct.search_UNIT(_object, _datas)
			end,
			get_objects = function()
				return ATMEDatas.units
			end,
			fct = function(_datas, _group, _unitName)
				local groupName = _group:getName()
				if _datas.objects[groupName] == nil then -- Cause recherche multi-zones sur un ATME.C_Zones
					if _datas.nbUnitsGroup[groupName] == nil then
						_datas.allGroupIn[groupName] = {}
						_datas.nbUnitsGroup[groupName] = 0
					end

					table.insert(_datas.allGroupIn[groupName], _unitName)
					_datas.nbUnitsGroup[groupName] = _datas.nbUnitsGroup[groupName] + 1

					if _datas.nbUnitsGroup[groupName] == _group:getNbUnits() then
						local record
						local status
						if _datas.localMembers.callbacks[_datas.moduleName] ~= nil then
							local datas = {
								zone = _datas.zone,
								object = _group,
								unitsName = _datas.allGroupIn[groupName],
							}
							status, record = xpcall(function() return _datas.localMembers.callbacks[_datas.moduleName](_datas.findContext, datas) end, function() ATMEFct.errUserHandler('Filter callback') end)
						end
						
						_datas.allGroupIn[groupName] = nil
						_datas.nbUnitsGroup[groupName] = nil
						if record == nil or (type(record) == "boolean" and record == true) then
							_datas.objects[groupName] = _group
							table.insert(_datas.objectsIndex, groupName)
						end
					end
				end
			end,
		},
		["UNITS"] = {
			findSubCategories = true,
			category = {Object.Category.UNIT},
			search_fct = function(_object, _datas)
				ATMEFct.search_UNIT(_object, _datas)
			end,
			get_objects = function()
				return ATMEDatas.units
			end,
			fct = function(_datas, _group, _unitName)
				if _datas.objects[_unitName] == nil then -- Cause recherche multi-zones sur un ATME.C_Zones
					if ATMEDatas.units[_unitName] ~= nil then
						local record
						local status
						if _datas.localMembers.callbacks[_datas.moduleName] ~= nil then
							local datas = {
								zone = _datas.zone,
								object = ATMEDatas.units[_unitName],
							}
							status, record = xpcall(function() return _datas.localMembers.callbacks[_datas.moduleName](_datas.findContext, datas) end, function() ATMEFct.errUserHandler('Filter callback') end)
						end
						
						if record == nil or (type(record) == "boolean" and record == true) then
							_datas.objects[_unitName] = ATMEDatas.units[_unitName]
							table.insert(_datas.objectsIndex, _unitName)
						end
					end
				end
			end,
		},
		["PLAYERS"] = {
			findSubCategories = true,
			category = {Object.Category.UNIT},
			search_fct = function(_object, _datas)
				ATMEFct.search_PLAYER(_object, _datas)
			end,
			get_objects = function()
				return ATMEDatas.players
			end,
			fct = function(_datas, _group, _unitName)
				if _datas.objects[_unitName] == nil then -- Cause recherche multi-zones sur un ATME.C_Zones
					if ATMEDatas.players[_unitName] ~= nil then
						local record
						local status
						if _datas.localMembers.callbacks[_datas.moduleName] ~= nil then
							local datas = {
								zone = _datas.zone,
								object = ATMEDatas.players[_unitName],
							}
							status, record = xpcall(function() return _datas.localMembers.callbacks[_datas.moduleName](_datas.findContext, datas) end, function() ATMEFct.errUserHandler('Filter callback') end)
						end
						
						if record == nil or (type(record) == "boolean" and record == true) then
							_datas.objects[_unitName] = ATMEDatas.players[_unitName]
							table.insert(_datas.objectsIndex, _unitName)
						end
					end
				end
			end,
		},
		["WEAPONS"] = {
			findSubCategories = true,
			category = {
				Object.Category.WEAPON,
			},
			search_fct = function(_object, _datas)
				ATMEFct.search_WEAPON(_object, _datas)
			end,
			get_objects = function()
				return ATMEDatas.weapons
			end,
			fct = function(_datas, _group, _objectName)
				if _datas.objects[_objectName] == nil then -- Cause recherche multi-zones sur un ATME.C_Zones
					local record
					local status
					if _datas.localMembers.callbacks[_datas.moduleName] ~= nil then
						local datas = {
							zone = _datas.zone,
							object = ATMEDatas.weapons[_objectName],
						}
						status, record = xpcall(function() return _datas.localMembers.callbacks[_datas.moduleName](_datas.findContext, datas) end, function() ATMEFct.errUserHandler('Filter callback') end)
					end
				
					if record == nil or (type(record) == "boolean" and record == true) then
						_datas.objects[_objectName] = ATMEDatas.weapons[_objectName]
						table.insert(_datas.objectsIndex, _objectName)
					end
				end
			end,
		},
		["STATICS"] = {
			findSubCategories = true,
			category = {
				Object.Category.STATIC,
			},
			search_fct = function(_object, _datas)
				ATMEFct.search_STATIC(_object, _datas)
			end,
			get_objects = function()
				return ATMEDatas.statics
			end,
			fct = function(_datas, _group, _objectName)
				if _datas.objects[_objectName] == nil then -- Cause recherche multi-zones sur un ATME.C_Zones
					if ATMEDatas.statics[_objectName] ~= nil then
						local record
						local status
						if _datas.localMembers.callbacks[_datas.moduleName] ~= nil then
							local datas = {
								zone = _datas.zone,
								object = ATMEDatas.statics[_objectName],
							}
							status, record = xpcall(function() return _datas.localMembers.callbacks[_datas.moduleName](_datas.findContext, datas) end, function() ATMEFct.errUserHandler('Filter callback') end)
						end
						
						
						if record == nil or (type(record) == "boolean" and record == true) then
							_datas.objects[_objectName] = ATMEDatas.statics[_objectName]
							table.insert(_datas.objectsIndex, _objectName)
						end
					end
				end
			end,
		},
		["CARGOS"] = {
			findSubCategories = false,
			category = {
				Object.Category.CARGO,
			},
			search_fct = function(_object, _datas)
				ATMEFct.search_CARGO(_object, _datas)
			end,
			get_objects = function()
				return ATMEDatas.cargos
			end,
			fct = function(_datas, _group, _objectName)
				if _datas.objects[_objectName] == nil then -- Cause recherche multi-zones sur un ATME.C_Zones
					if ATMEDatas.cargos[_objectName] ~= nil then
						local record
						local status
						if _datas.localMembers.callbacks[_datas.moduleName] ~= nil then
							local datas = {
								zone = _datas.zone,
								object = ATMEDatas.cargos[_objectName],
							}
							status, record = xpcall(function() return _datas.localMembers.callbacks[_datas.moduleName](_datas.findContext, datas) end, function() ATMEFct.errUserHandler('Filter callback') end)
						end
						
						if record == nil or (type(record) == "boolean" and record == true) then
							_datas.objects[_objectName] = ATMEDatas.cargos[_objectName]
							table.insert(_datas.objectsIndex, _objectName)
						end
					end
				end
			end,
		},
		["SCENERY_OBJECTS"] = {
			findSubCategories = false,
			category = {
				Object.Category.SCENERY,
			},
			search_fct = function(_object, _datas)
				ATMEFct.search_SCENERY(_object, _datas)
			end,
			get_objects = function()
				return ATMEDatas.sceneryObjects
			end,
			fct = function(_datas, _group, _objectName)
				if _datas.objects[_objectName] == nil then -- Cause recherche multi-zones sur un ATME.C_Zones
					local record
					local status
					if _datas.localMembers.callbacks[_datas.moduleName] ~= nil then
						local datas = {
							zone = _datas.zone,
							object = ATMEDatas.sceneryObjects[_objectName],
						}
						status, record = xpcall(function() return _datas.localMembers.callbacks[_datas.moduleName](_datas.findContext, datas) end, function() ATMEFct.errUserHandler('Filter callback') end)
					end
					
					if record == nil or (type(record) == "boolean" and record == true) then
						_datas.objects[_objectName] = ATMEDatas.sceneryObjects[_objectName]
						table.insert(_datas.objectsIndex, _objectName)
					end
				end
			end,
		},
	},
	referenceClasses = {
		["C_Unit"] = true,
		["C_Static"] = true,
		["C_Cargo"] = true,
		["C_Weapon"] = true,
		["C_SceneryObject"] = true,
	},
	tacanMode = {
		["X"] = 0,
		["Y"] = 1,
	},
	tacanFrequencyOffset = {
		["Xmit"] = {
			["X"] = 1024,
			["Y"] = 1024,
		},
		["Recv"] = {
			["X"] = 961,
			["Y"] = 1087,
		},
	},
	beaconTypes = {
		["TOKEN_TYPE_TACAN_AA"] = "TACAN",
		["TOKEN_TYPE_TACAN"] = "TACAN",
		["TOKEN_TYPE_ICLS"] = "ICLS",
	},
	coreEventTypes = {
		["TAKE_OFF"] = function(_localMembers)
			return {}
		end,
		["LAND"] = function(_localMembers)
			return {}
		end,
		["ENGINE"] = function(_localMembers)
			return {
				action = _localMembers.datas.action,
			}
		end,
		["TRANSPORT"] = function(_localMembers)
			return {
				action = _localMembers.datas.action,
				index = _localMembers.datas.index,
				onBoardGroupName = _localMembers.datas.onBoardGroupName,
				nbUnitsOnBoard = _localMembers.datas.nbUnitsOnBoard,
				nbDeadUnits = _localMembers.datas.nbDeadUnits,
				linkedGroup = _localMembers.datas.linkedGroup,
			}
		end,
		["SPOT"] = function(_localMembers)
			return {
				action = _localMembers.datas.action,
				target = _localMembers.datas.target,
				code = _localMembers.datas.code,
				vector = _localMembers.datas.vector,
			}
		end,
		["SENSOR"] = function(_localMembers)
			return {
				action = _localMembers.datas.action,
				target = _localMembers.datas.target,
			}
		end,
		["WEAPON"] = function(_localMembers)
			return {
				action = _localMembers.datas.action,
				target = _localMembers.datas.target,
				weapon = _localMembers.datas.weapon,
			}
		end,	
		["REFUELING"] = function(_localMembers)
			return {
				action = _localMembers.datas.action,
			}
		end,
		["MARK"] = function(_localMembers)
			return {
				action = _localMembers.datas.action,
				mark = _localMembers.datas.mark,
				linkedGroup = _localMembers.datas.linkedGroup,
				coalitionName = _localMembers.datas.coalitionName,
			}
		end,
	},
}

-- Ajout des données nécessaires aux ATMEConst déjà définies

do
	ATMEConst.errorLabelReference = " or table (x,y)/(x,y,z)"
	local first = true
	for _id, _ in pairs(ATMEConst.referenceClasses) do
		if first == true then
			ATMEConst.errorLabelReference = "ATME." .. _id .. ATMEConst.errorLabelReference
		else
			ATMEConst.errorLabelReference = "ATME." .. _id .. ", " .. ATMEConst.errorLabelReference
		end
		
		first = false
	end
	
	for _id, _ in pairs(ATMEConst.smokeColors) do
		if _id ~= "RANDOM" then
			ATMEConst.tokenTypes[_id] = {
				type = "TOKEN_COLOR",
			}
		end
	end
	
	for _id, _ in pairs(ATMEConst.flareColors) do
		if _id ~= "RANDOM" then
			ATMEConst.tokenTypes[_id] = {
				type = "TOKEN_COLOR",
			}
		end
	end
	
	for _id, _ in pairs(ATMEConst.blueCallsigns.aircrafts) do
		ATMEConst.tokenTypes[_id] = {
			type = "TOKEN_AIRCRAFT_CALLSIGN_NAME",
		}
	end
	
	for _id, _ in pairs(ATMEConst.blueCallsigns.jtacs) do
		ATMEConst.tokenTypes[_id] = {
			type = "TOKEN_JTAC_CALLSIGN_NAME",
		}
	end
end

-- Définitions complémentaires

local POINT_BAD = 0
local POINT_2D = 1
local POINT_3D = 2

local customEventHandler = {}			-- Déclaration du handler d'événements DCS


----------------------------------------------------------------------------------------------------------
-- DEFINITION DES FONCTIONS UTILES DIRECTEMENT A ATME
----------------------------------------------------------------------------------------------------------

ATMEFct.logDisplayText = function(_debugMode, _label, _level)
	if _debugMode == true and _level == 1 and ATMEDatas.logDisplay == true then
		trigger.action.outText(_label, 5)
	elseif _debugMode == false and _level > 2 then
		return
	end
	
	env.info(_label, false)
end

ATMEFct.warningDisplayText = function(_debugMode, _label)
	if _debugMode == true and ATMEDatas.logDisplay == true then
		trigger.action.outText(_label, 5)
	end
	
	env.warning(_label, false)
end

ATMEFct.errorOutputCore = function(_module, _fctDefinition, _text, _needObjectPrecision)
	-- errorOutputCore : fonction de base de gestion des erreurs, engendre un arrêt de ATME (plus de réarmement des timers)
	
	if ATMEDatas.exitATME == false then
		ATMEFct.setError("FATAL_ERROR")
		
		local err = debug.traceback()
		local id, text, stack = ATMEFct.analyzeTraceback(err)
		local fctDefinition = ""
	
		local label = "[ATME"
		if id ~= nil and ATMEDatas.refModules[id] ~= nil then
			label = label .. "-" .. ATMEDatas.refModules[id] .. "] Error in ATME module lua file"
		elseif id ~= nil then
			label = label .. "] Error in DO SCRIPT FILE in Mission Editor (not an ATME module lua file)"
		else
			label = label .. "] Error in DO SCRIPT in Mission Editor"
		end
		
		if type(_fctDefinition) == "string" and _fctDefinition:len() > 0 then
			fctDefinition = _fctDefinition .. "\n"
		end
		
		if text ~= nil then
			label = label .. "\n" .. text .. fctDefinition .. "\n   -> " .. _text
		else
			label = label .. "\n" .. fctDefinition .. "\n   -> " .. _text
		end
		
		if _needObjectPrecision == true then
			label = label .. " (or missing ':' before function name)"
		end
		
		ATMEDatas.exitATME = true
		
		label = label .. "\nError code : " .. ATMEDatas.error.code .. "\n\nATME STOPPED"
		env.error(label .. "\n\nstack traceback:\n" .. stack, true)
	end
end

ATMEFct.warningOutputCore = function(_module, _fctDefinition, _text, _needObjectPrecision)
	-- warningOutputCore : fonction de base de gestion des erreurs non fatales, sans arrêt ATME.
	
	if ATMEDatas.exitATME == false then
		if ATMEDatas.error == nil then
			ATMEFct.setError("UNKNOWN_WARNING")
		end
		
		local err = debug.traceback()
		local id, text, stack = ATMEFct.analyzeTraceback(err)
		local fctDefinition = ""
	
		local label = "[ATME"
		if id ~= nil and ATMEDatas.refModules[id] ~= nil then
			label = label .. "-" .. ATMEDatas.refModules[id] .. "] Warning in ATME module lua file"
		elseif id ~= nil then
			label = label .. "] Warning in DO SCRIPT FILE in Mission Editor (not an ATME module lua file)"
		else
			label = label .. "] Warning in DO SCRIPT in Mission Editor"
		end
		
		if type(_fctDefinition) == "string" and _fctDefinition:len() > 0  then
			fctDefinition = _fctDefinition .. "\n"
		end
		
		if text ~= nil then
			label = label .. "\n" .. text .. _fctDefinition .. "\n   -> " .. _text
		else
			label = label .. "\n" .. _fctDefinition .. "\n   -> " .. _text
		end
		
		if _needObjectPrecision == true then
			label = label .. " (or missing ':' before function name)"
		end
		
		ATMEFct.warningDisplayText(ATMEConst.debugSwitch, label)
	end
end

ATMEFct.outputCore = function(_module, _text, _level)
	-- outputCore : fonction de base des messages de debug en fonction du niveau et de l'activation ou non du mode debug
	if ATMEDatas.exitATME == false and (_debugMode == true or _level < 3) then
		local label = "[ATME"
		local debugMode = ATMEConst.debugSwitch
		
		if _module ~= nil then
			label = label .. "-" .. _module:getName()
			debugMode = _module:isDebugOn()
		end
		
		label = label .. "] " .. _text
		
		if _level > 0 then
			ATMEFct.logDisplayText(debugMode, label, _level)
		else
			ATMEFct.warningDisplayText(debugMode, label)
		end
	end
end

ATMEFct.outputVarCore = function(_module, _title, _var, _level)
	if ATMEDatas.exitATME == false and (_debugMode == true or _level < 3) then
		local debugMode = ATMEConst.debugSwitch
		local tablesRedundancy = {}
		local startLabel = "[ATME"
		
		if _module ~= nil then
			startLabel = startLabel .. "-" .. _module:getName()
			debugMode = _module:isDebugOn()
		end
		
		startLabel = startLabel .. "] "
		
		local function logVar(_v, _title)
			local label = _title
			
			if type(_v) == "boolean" then
				local v = "True"
				if _v == false then
					v = "False"
				end
				label = label .. " : " .. v
			elseif _v ~= nil and type(_v) ~= "function" then
				label = label ..  " : " .. _v
			end
			
			ATMEFct.logDisplayText(debugMode, label, _level)
		end
		
		local function logObject(_v, _title)
			ATMEFct.logDisplayText(debugMode, _title .. _v:getClassName()  .. " class", _level)
		end
		
		local function logPoint(_point, _indent)
			local lat, lon, alt = coord.LOtoLL(ATMEFct.convertToPoint3D(_point))
					
			local dirlon, dlon, mlon, slon = ATMEFct.convertToDMS(lon)
			local dirlat, dlat, mlat, slat = ATMEFct.convertToDMS(lat)
			
			local label 
			if  _point.z == nil then
				label = _indent .. "  Found keys x,y point 2D at lat : "
			else
				label = _indent .. "  Found keys x,y,z point 3D at lat : "
			end
			
			if dirlat < 0 then
				label = label .. "S"
			else
				label = label .. "N"
			end
			label = label .. string.format("%02d°%02d'%02d\"", dlat, mlat, slat) .. " - lon : "
			
			if dirlon < 0 then
				label = label .. "W"
			else
				label = label .. "E"
			end
			
			ATMEFct.logDisplayText(debugMode, label .. string.format("%03d°%02d'%02d\"", dlon, mlon, slon) .. " -> alt : " .. math.floor(alt) .. "m (" .. math.floor(ATME.convertMtoFts(alt)) .. " feet)\n", _level)
		end
		
		local function logTable(_t, _indent)
			if _indent == nil then
				_indent = "  "
			else
				_indent = _indent .. "  "
			end
			
			local found = false
			for _k, _v in pairs(_t) do
				found = true
				
				if type(_v) == "table" then
					if tablesRedundancy[_v] ~= nil then
						ATMEFct.logDisplayText(debugMode, _indent .. "table '" .. _k .. "' redundancy of " .. tablesRedundancy[_v] .. "\n", _level)
					elseif ATMEFct.isATMEObject(_v) == false then
						ATMEFct.logDisplayText(debugMode, _indent .. "table '" .. _k .. "' listing\n", _level)
						tablesRedundancy[_v] = _k
						if logTable(_v, _indent) == false then
							ATMEFct.logDisplayText(debugMode, _indent .. "  Empty table", _level)
						elseif type(_v.x) == "number" and type(_v.y) == "number" and (_v.z == nil or type(_v.z) == "number") then
							logPoint(_v, _indent)
						end
					else
						logObject(_v, _indent .. "key '" .. _k .. "' - Object instance of ATME.")
					end
				else
					logVar(_v, _indent .. "key '" .. _k .. "' - Type " .. type(_v))
				end
			end
			
			return found
		end
		
		if type(_var) == "table" then
			if ATMEFct.isATMEObject(_var) == false then
				startLabel = startLabel .. "Start Table '" .. _title .. "' :\n"
				ATMEFct.logDisplayText(debugMode, startLabel, _level)
				tablesRedundancy[_var] = _title
				if logTable(_var) == false then
					ATMEFct.logDisplayText(debugMode, "  Empty table", _level)
				elseif type(_var.x) == "number" and type(_var.y) == "number" and (_var.z == nil or type(_var.z) == "number") then
					logPoint(_var, "")
				end
			else
				logObject(_var, startLabel .. "Object '" .. _title .. "' - Instance of ATME.")
			end
		else
			logVar(_var, startLabel .. "Var '" .. _title .. "' - Type " .. type(_var))
		end
	end
end

ATMEFct.setError = function(_code)
	ATMEDatas.error = {
		code = _code,
	}
end

ATMEFct.copyTable = function(_datas)
	local err = false 
	local function _copy(datas)
		if err == false then
			if type(datas) ~= "table" or ATMEFct.isATMEObject(datas) == true then
				return datas
			end
			
			if getmetatable(_datas) == nil then
				local newTable = {}
				for _index, _value in pairs(datas) do
					newTable[_index] = _copy(_value)
				end
					
				return newTable
			else
				err = true
			end
		end
	end
	
	local t = _copy(_datas)
	if err == false then
		return t
	end
end

ATMEFct.convertToDMS = function(_degrees)
	local dir = 1
	if _degrees < 0 then
		dir = -1
	end
	
	local d = math.abs(_degrees)
	local dfloor = math.floor(d)
	local m = (d - dfloor) * 60
	local mfloor = math.floor(m)
	local s = (m - mfloor) * 60
	local sfloor = math.floor(s)
	
	return dir, dfloor, mfloor, sfloor, math.floor((s - sfloor) * 100)
end

ATMEFct.convertToDMd = function(_degrees, _highPrecision)
	local dir = 1
	if _degrees < 0 then
		dir = -1
	end
	
	local factor = 10000
	if _highPrecision == false then
		factor = 100
	end

	local dabs = math.abs(_degrees)
	local dfloor = math.floor(dabs)
	local m = (dabs - dfloor) * 60
	local mfloor = math.floor(m)
	
	return dir, dfloor, mfloor, math.floor((m - mfloor) * factor)
end

ATMEFct.search_UNIT = function(_unit, _datas)
	local unitName = _unit:getName()
	
	if ATMEDatas.units[unitName] ~= nil then
		local localMembers = ATMEDatas.localMembers[ATMEDatas.units[unitName]:getIndex()]
		
		if _unit:isExist() == true then
			local id = _unit:getCoalition()
			
			if _datas.coalitionName == nil or ATMEConst.coalitionNames[id] == _datas.coalitionName then
				if ATMEFct.C_FindContext_Object_isInZone(ATMEDatas.units[unitName], _datas) == true then
						
					ATMEFct.C_FindContext_setToList(_datas, localMembers.group, unitName)
				end
			end
		end
	end
end

ATMEFct.search_PLAYER = function(_unit, _datas)
	local unitName = _unit:getName()
	
	if ATMEDatas.players[unitName] ~= nil then
		local localMembers = ATMEDatas.localMembers[ATMEDatas.players[unitName]:getIndex()]
		
		if _unit:isExist() == true then
			local id = _unit:getCoalition()

			if _datas.coalitionName == nil or ATMEConst.coalitionNames[id] == _datas.coalitionName then
				if ATMEFct.C_FindContext_Object_isInZone(ATMEDatas.players[unitName], _datas) == true then
						
					ATMEFct.C_FindContext_setToList(_datas, localMembers.group, unitName)
				end
			end
		end
	end
end

ATMEFct.search_STATIC = function(_object, _datas)
	local objectName = _object:getName()

	if ATMEDatas.statics[objectName] ~= nil then
		local localMembers = ATMEDatas.localMembers[ATMEDatas.statics[objectName]:getIndex()]
		
		if _object:isExist() == true then
			local id = _object:getCoalition()
			
			if _datas.coalitionName == nil or ATMEConst.coalitionNames[id] == _datas.coalitionName then
				if ATMEFct.C_FindContext_Object_isInZone(ATMEDatas.statics[objectName], _datas) == true then
					ATMEFct.C_FindContext_setToList(_datas, nil, objectName)
				end
			end
		end
	end
end

ATMEFct.search_CARGO = function(_object, _datas)
	local objectName = _object:getName()

	if ATMEDatas.cargos[objectName] ~= nil then
		local localMembers = ATMEDatas.localMembers[ATMEDatas.cargos[objectName]:getIndex()]
		
		if _object:isExist() == true then
			local id = _object:getCoalition()
			
			if _datas.coalitionName == nil or ATMEConst.coalitionNames[id] == _datas.coalitionName then
				if ATMEFct.C_FindContext_Object_isInZone(ATMEDatas.cargos[objectName], _datas) == true then
					ATMEFct.C_FindContext_setToList(_datas, nil, objectName)
				end
			end
		end
	end
end

ATMEFct.search_SCENERY = function(_object, _datas)
	local objectName = _object:getName()

	if ATMEDatas.sceneryObjects[objectName] == nil then
		ATMEFct.C_SceneryObject_create(_object)
	end
							
	if ATMEFct.C_FindContext_Object_isInZone(ATMEDatas.sceneryObjects[objectName], _datas) == true then
		ATMEFct.C_FindContext_setToList(_datas, nil, objectName)
	end
end

ATMEFct.search_WEAPON = function(_object, _datas)
	local objectName = _object:getName()
	
	if ATMEFct.C_FindContext_Object_isInZone(ATMEDatas.weapons[objectName], _datas) == true then
		ATMEFct.C_FindContext_setToList(_datas, nil, objectName)
	end
end

ATMEFct.fctZoneTask = {
	[1] = function(_param, _item)
		_param.groupId = ATMEAIIndFct.C_Group_getID(ATMEDatas.groups[_item.targetGroupName])[1]
	end,
}

ATMEFct.onZoneTasks = {	-- Tâche principale pilotable par ATME pour traiter des zones
						-- La fonction d'initialisation des paramètres est indiquée
	["FAC_AttackGroup"] = ATMEFct.fctZoneTask[1],
	["AttackGroup"] = ATMEFct.fctZoneTask[1],
}

ATMEFct.whichSide = function(_A, _B, _C)
	local vectAB = ATMEFct.vectors["3D"].createFromReference(nil, nil, _A, _B)
	local vectAC = ATMEFct.vectors["3D"].createFromReference(nil, nil, _A, _C)
	
	local det = vectAB:getX() * vectAC:getZ() - vectAB:getZ() * vectAC:getX()
	
	-- det < 0, C est à gauche de la droite AB
	-- det > 0,  C est à droite de la droite AB
	-- det = 0, C est dans sur la droite AB
	
	if det > 0 then
		return 1
	elseif det < 0 then
		return -1
	else
		return 0
	end
end	

ATMEFct.whichSideLine2D = function(_line2D, _C)
	local vector = _line2D:getVector()
	local origin = _line2D:getOrigin()
	
	local vectAC = ATMEFct.vectors["2D"].createFromReference(nil, nil, origin, _C)
	
	local det = vector:getX() * vectAC:getZ() - vector:getZ() * vectAC:getX()
	
	-- det < 0, C est à gauche de la droite orientée line2D
	-- det > 0,  C est à droite de la droite orientée line2D
	-- det = 0, C est dans sur la droite orientée line2D
	
	if det > 0 then
		return 1
	elseif det < 0 then
		return -1
	else
		return 0
	end
end		


ATMEFct.isAzimuthInLimits = function(_referencePoint, _azimuth, _targetPoint)
	local vector = ATMEFct.vectors["3D"].createFromReference(nil, nil, _referencePoint, _targetPoint)
	local radius = 1000
	
	if vector:getHModule() > 10000 then
		radius = 0.1 * vector:getHModule()
	end
	
	local h = math.sqrt(vector:getHModule()^2 + radius^2)
	local deltaAzimuth = math.deg(math.asin(radius/h))
	local vectorHeading = vector:getAzimuth()

	local minAzimuth = vectorHeading - deltaAzimuth
	local maxAzimuth = vectorHeading + deltaAzimuth

	if maxAzimuth >= 360 then
		maxAzimuth = maxAzimuth - 360
	end
	
	if minAzimuth < 0 then
		minAzimuth = minAzimuth + 360
	end

	return _azimuth >= minAzimuth and _azimuth <= maxAzimuth
end

ATMEFct.decompString = function(_sequence)
	local tokens = {}
	
	-- Le '-' devient un séparateur dans la décomposition d'une chaine en tags suite à modif DCS des copies (' #%03d' remplacé par '-%d')
	-- Ainsi "Naval-1" a bien 2 tags "Naval" et "1"
	for _word in _sequence:gmatch("[^%s%-]+") do
		table.insert(tokens, _word)
    end 
   
	if #tokens > 0 then
		return tokens
	end
end

ATMEFct.decompStringSpaceOnly = function(_sequence)
	local tokens = {}
	
	for _word in _sequence:gmatch("[^%s]+") do
		table.insert(tokens, _word)
    end 
   
	if #tokens > 0 then
		return tokens
	end
end

ATMEFct.compareTags = function(_tags, _tagsToCompare)
	local function findTags(_sequence)
		local index = 1
		local cursor = 1
		local tags = {}
		local err = false
		local lg = _sequence:len()

		while cursor <= lg and err == false do
			local d,f = _sequence:find("[|]", cursor)
			if d ~= nil then
				local s = _sequence:sub(cursor, d - 1)
				local tagsTemp = ATMEFct.decompString(s)

				if tagsTemp ~= nil and #tagsTemp <= 1 then
					if #tagsTemp == 1 then
						tags[index] = tagsTemp[1]
					end
				end
				
				index = index + 1
				cursor = d + 1 
			else
				local s = _sequence:sub(cursor, lg) 
				local tagsTemp = ATMEFct.decompString(s)

				if tagsTemp ~= nil and #tagsTemp == 1 then
					tags[index] = tagsTemp[1]
					index = index + 1
					cursor = lg + 1 
				else
					err = true
				end  
			end
		end

		return err, tags
	end
	
	local err, tags = findTags(_tagsToCompare)

	if err == false then
		local ok = true
		for _i, _tag in ipairs(_tags) do
			if tags[_i] ~= nil and _tag ~= tags[_i] then
				ok = false
			end
		end

		return ok
	end
end

ATMEFct.analyzeString = function(_sequence)
	local pBegin = {}
	local pEnd = {}
	local pString = {}
	local index = 0
	
	local cursor = 1
	local lg =  _sequence:len()
      

	local function defineToken(_token)
		local sign = 1
		
		local function returnRelativeStatus(_token)
			local c = string.sub(_token, 1, 1)
			local relative = 1

			if c == "+" then
				relative = 2
			elseif c == "-" then
				relative = 2
				sign = -1
			end

			return relative
		end
		
		local function tokenType(_token, _item)
			local token = string.upper(_token)
			
			if ATMEConst.tokenTypes[token] ~= nil then
				_item.type = ATMEConst.tokenTypes[token].type
				_item.status = "ABSOLUTE"
        
				if ATMEConst.tokenTypes[token].value ~= nil then
				  _item.value = ATMEConst.tokenTypes[token].value
				else
				  _item.value = token
				end
        
				return true
			end
			
			return false
		end    

		local function tokenEval(_token, _item)
			local str
			local percent = 1
			
			local pos = returnRelativeStatus(_token)
			_item.status = ATMEConst.relativeStatus[pos]

			str = string.match(_token,"^%d%d:%d%d:%d%d$",pos)

			if str ~= nil then
				local hd = tonumber(string.sub(str,1,2))
				local md = tonumber(string.sub(str,4,5))
				local sd = tonumber(string.sub(str,7,8))

				if hd >= 0 and hd <= 23 and md >= 0 and md <= 59 and sd >= 0 and sd <= 59 then
					_item.type = "TOKEN_INTEGER"
					_item.value = (hd * 3600 + md * 60 + sd) * sign
					return true
				end
			end

			local lastChar = string.sub(_token, -1)

			local tacanChannel = false
			if lastChar == "%" then
				str = string.sub(_token, 1, _token:len() - 1)
				percent = 0.01
			elseif ATMEConst.tacanMode[lastChar] ~= nil then
				tacanChannel = true
				str = string.sub(_token, 1, _token:len() - 1)
			else
				str = _token
			end

			local n = tonumber(str)

			if n ~= nil then
				if tacanChannel == false then
					if n == math.floor(n) then
						_item.type = "TOKEN_INTEGER"
					else
						_item.type = "TOKEN_FLOAT"
					end
			
					_item.value = n * percent
					return true
				else
					local channel = math.floor(n)
					
					if channel > 0 and channel < 64 then
						_item.value = channel .. lastChar
						_item.type = "TOKEN_TACAN_CHANNEL"
						return true
					end
				end
			end

			return false
		end        

		local fcts = {
			[1] = tokenType,
			[2] = tokenEval,
		}

		local item = {}
		local indexfct = 1
		while indexfct <= #fcts and fcts[indexfct](_token, item) == false do
			indexfct = indexfct + 1
		end

		if indexfct > #fcts then
			item.value = _token
			item.type = "TOKEN_ERROR"
			item.status = "ABSOLUTE"
		end

		return item
	end
	
	local function setValue(_start, _end)
		local separator
		if _sequence:sub(_start,_start) == "'" then
			separator = "'"
		elseif _sequence:sub(_start,_start) == '"' then
			separator = '"'
		end
		
		if separator ~= nil then
			local ds,fs = string.find(_sequence, separator, _start + 1)
			if ds ~= nil then
				local d,f = string.find(_sequence, "%s+", fs)
				if fs == lg or d ~= nil then
					pBegin[index] = _start
					pEnd[index] = ds
					pString[index] = true

					if d ~= nil then
						return f
					else
						return fs
					end
				end
			end
		end
		
		pBegin[index] = _start
		pEnd[index] = _end
		pString[index] = false   
	end

	
	while cursor <= lg do
		local d,f = string.find(_sequence, "%s+", cursor)

		if d ~= nil then
			if d == cursor then
				cursor = f + 1
			else
				index = index + 1
				local fs = setValue(cursor, d - 1)
				
				if fs ~= nil then
					f = fs
				end
					
				cursor = f + 1
			end
		else
			index = index + 1
			setValue(cursor, lg)
			cursor = lg + 1
		end
	end

	local tokens = {}
	
	for i = 1,index do
		if pString[i] == false then
			tokens[i] = defineToken(_sequence:sub(pBegin[i], pEnd[i]))
		else
			local value = ""
			
			if pEnd[i] - pBegin[i] > 1 then
				value = _sequence:sub(pBegin[i] + 1, pEnd[i] - 1)
			end
			tokens[i] = {
				value = value,
				type = "TOKEN_STRING",
				status = "ABSOLUTE",
			}
		end
		tokens[i].pStart = pBegin[i]
		tokens[i].pEnd = pEnd[i]
	end

	return tokens
end

ATMEFct.getTimeValue = function(_rules)
	local tokenError = false
	local timeStart
	local isRelative = false
	
	if type(_rules) == "string" then
		local tokens = ATMEFct.analyzeString(_rules)
		
		if #tokens >= 1 and #tokens <= 3 then
			if #tokens >= 2 and tokens[1].type == "TOKEN_TYPE_RANDOM" then
				if tokens[2].type == "TOKEN_INTEGER" then
					local ok = false
					local minRandom = 0
					local maxRandom = 0
					
					if tokens[2].value >= 0 then
						maxRandom = tokens[2].value
						
						if #tokens == 3 then
							minRandom = tokens[2].value
							
							if tokens[3].type == "TOKEN_INTEGER" then
								if tokens[3].value >= 0 then
									maxRandom = tokens[3].value
									if tokens[3].status == tokens[2].status then
										ok = true
									else
										tokenError = true
										timeStart = "Parameters #2 & #3 must have same status (+ for all or nothing for all) in string format rules"
									end
								else
									tokenError = true
									timeStart = "Parameter #3 must be >= 0 in string format rules"
								end
							else
								tokenError = true
								timeStart = "Parameter #3 must be number or time value in string format rules"
							end
						else
							ok = true
						end
					else
						tokenError = true
						timeStart = "Parameter #2 must be >= 0 in string format rules"
					end
					
					if ok == true then
						if minRandom >= maxRandom  then
							tokenError = true
							timeStart = "Parameter #2 must be <= Parameter #3 in string format rules"
						else
							if tokens[2].status == "RELATIVE" then
								isRelative = true
							end
						
							timeStart = math.random(minRandom, maxRandom)
						end
					end
				else
					tokenError = true
					timeStart = "Parameter #2 must be number or time value in string format rules"
				end
			elseif tokens[1].type == "TOKEN_INTEGER" then
				if tokens[1].value >= 0 then
					if tokens[1].status == "RELATIVE" then
						isRelative = true
					end
					
					timeStart = tokens[1].value
				else
					tokenError = true
					timeStart = "Parameter #1 must be >= 0 in string format rules"
				end
			else
				tokenError = true
				timeStart = "Parameter #1 must be number or time value or 'RANDOM' in string format rules"
			end
		else
			tokenError = true
			timeStart = "Bad string format rules"
		end
	else
		isRelative = true
		if _rules > 0 then
			timeStart = _rules
		else
			tokenError = true
			timeStart = "rules must be > 0"
		end
	end

	return tokenError, timeStart, isRelative
end

ATMEFct.getRandomDatas = function(_datas)
	local randomSet = false
	local minValue = 0
	local maxValue
	
	if type(_datas) == "string" then
		local tokens = ATMEFct.analyzeString(_datas)
		local ok = false

		if #tokens >= 1 and #tokens <= 3 and (tokens[1].type == "TOKEN_TYPE_RANDOM" or tokens[1].type == "TOKEN_INTEGER") then
			if tokens[1].type == "TOKEN_TYPE_RANDOM" then
				randomSet = true
				if #tokens >= 2 and tokens[2].type == "TOKEN_INTEGER" then
					minValue = tokens[2].value
					
					if #tokens == 3 and tokens[3].type == "TOKEN_INTEGER" then
						maxValue = tokens[3].value
						ok = true
					elseif #tokens == 2 then
						maxValue = minValue
						minValue = 1
						ok = true
					else
						return "Parameter #3 must be number in string format "
					end
				else
					return "Parameter #2 must be number in string format "
				end
				
				if ok == true then
					if minValue >= maxValue then
						return "Parameter #2 must be < Parameter #3 in string format "
					end
				end
			else
				minValue = tokens[1].value
				ok = true
			end
		else
			return "Parameter #1 must be number or 'RANDOM' in string format "
		end
		
		if ok == false then
			return "Bad string format "
		end
	else
		minValue = math.floor(_datas)
	end
	
	return nil, randomSet, minValue, maxValue
end

ATMEFct.searchMainChunk = function(_err)
	local line = string.match(_err,'[^\n]* in main chunk')

	return string.sub(string.match(line,'[%w%s]*%.lua"%]'), 1, -7)
end

ATMEFct.analyzeTraceback = function (_err)
	local line = _err
	local id
	local startBack = 1
	local found = false
	local refName
	local last = false
	local numLine

	local _, pos = string.find(_err, "\n") -- La première ligne est ignorée
	pos = pos + 1

	while last == false and found == false do
		local _, f = string.find(_err, "\n", pos)
		id = nil
		
		if f ~= nil then
			line = string.sub(_err, pos, f - 1)
			local subline = string.match(line,'[%w%s]*%.lua"%]')
			if subline ~= nil then
				id = string.sub(subline, 1, -7)
			end
			
			startBack = pos
			pos = f + 1
		else -- Dernière ligne
			local d, f = string.find(_err,'[^\n]*$')
			line = string.sub(_err, d, f)
			local subline = string.match(line,'[%w%s]*%.lua"%]')
			if subline ~= nil then
				id = string.sub(subline, 1, -7)
			end
			
			startBack = d
			last = true
		end
		
		if refName == nil then
			refName = id
		elseif refName ~= id then 
			found = true
		end
	end
	
	numLine = string.match(line, ":%d+:")
	
	local text
	if numLine ~= nil then
		numLine = string.sub(numLine, 2, numLine:len() - 1)
		text = "At line " .. numLine .. " - "
	end
	
	return id, text, string.sub(_err, startBack, _err:len())
end

ATMEFct.getErrorPrefix = function(_object)
	ATMEDatas.error = nil
	if _object ~= nil then
		return "Class ATME." .. _object:getClassName().. "\nFunction : "
	else
		return "Function ATME."
	end
end

ATMEFct.errUserHandler = function(_handlerName, _isMenuHandler)
	if _isMenuHandler == nil or _isMenuHandler == false then
		ATMEFct.errorOutputCore(nil, "User function assigned to '" .. _handlerName .. "'", "fatal error or try to access a nil value", false)
	else
		ATMEFct.errorOutputCore(nil, "User function assigned to menu '" .. _handlerName .. "'", "fatal error or try to access a nil value", false)
	end
end

ATMEFct.getRandomPlayer = function(_coalitionId)
	local playerUnitsName = {}
	
	if _coalitionId == nil then
		for _id = 1, 2 do
			local players = coalition.getPlayers(_id)
			
			for _i = 1, #players do
				if ATMEDatas.players[players[_i]:getName()] ~= nil then
					local localMembers = ATMEDatas.localMembers[ATMEDatas.players[players[_i]:getName()]:getIndex()]
					
					if localMembers.objectDCS:isExist() == true then
						table.insert(playerUnitsName, players[_i]:getName())
					end
				end
			end
		end
	else
		local players = coalition.getPlayers(_coalitionId)
		
		for _i = 1, #players do
			if ATMEDatas.players[players[i]:getName()] ~= nil then
				local localMembers = ATMEDatas.localMembers[ATMEDatas.players[players[_i]:getName()]:getIndex()]
				
				if localMembers.objectDCS:isExist() == true then
					table.insert(playerUnitsName, players[_i]:getName())
				end
			end
		end
	end
	
	return ATMEDatas.players[playerUnitsName[math.random(1, #playerUnitsName)]]
end

ATMEFct.foreachObject = function(_coalitionName, _typeObject, _callback, _datas)
	for _, _object in pairs(ATMEDatas[_typeObject]) do
		local localMembers = ATMEDatas.localMembers[_object:getIndex()]
		
		if localMembers.objectDCS:isExist() == true then
			local id = localMembers.objectDCS:getCoalition()
			if  _coalitionName == nil or ATMEConst.coalitionNames[id] == _coalitionName then
				if _callback(ATMEDatas[_typeObject][_object:getName()], _datas) == true then
					return
				end
			end
		end
	end
	
	return
end

ATMEFct.getDCSZone = function(_name)
	return trigger.misc.getZone(_name)
end

-- isATMEAObject : fonction interne ATME pour vérifier si une table est une instance
-- de classe ATME.

ATMEFct.isATMEObject = function(_object)
	local ok = false
	
	if type(_object) == "table" then
		xpcall(
			function() 
				local value = _object:getClassName()
		
				if type(value) == "string" then
					if type(ATMEDatas[value .. "_MT"].mt) == "table" then
						ok = true
					end
				end
			end, 
			function() 
				return false
			end)
	end
	
	return ok
end

-- isObjectClass : fonction interne ATME pour vérifier si une table est une instance
-- de classe ATME particulière.

ATMEFct.isObjectClass = function(_object, _class)
	local function testObjectClass(_object, _class)
		if _object:getClassName() == _class:getClassName() then
			return true
		else
			local object = _object:getParentClass()
			if object ~= nil then
				return testObjectClass(object, _class)
			end
		end
		
		return false
	end
	
	local function testObjectClassFromName(_object, _name)
		if _object:getClassName() == _name then
			return true
		else
			local object = _object:getParentClass()
			if object ~= nil then
				return testObjectClassFromName(object, _class)
			end
		end
		
		return false
	end

	if ATMEFct.isATMEObject(_object) == true then
		if ATMEFct.isATMEObject(_class) == true then
			return testObjectClass(_object, _class)
		elseif type(_class) == "string" then
			return testObjectClassFromName(_object, _class)
		end
	end
	
	return false
end



ATMEFct.isReference = function(_reference)
	if ATME.isPoint(_reference) ~= POINT_BAD then
		return true
	elseif ATMEFct.isATMEObject(_reference) == true then
		return ATMEConst.referenceClasses[_reference:getClassName()] ~= nil
	else
		return false
	end
end

ATMEFct.convertToPoint3D = function(_point)
	local point = {}
			
	point.x = _point.x
	if ATME.isPoint(_point) == POINT_2D then
		point.z = _point.y
		point.y = ATMEFct.getSurfaceAltitude(_point)
	else
		point.y = _point.y
		point.z = _point.z
	end
	
	return point
end

ATMEFct.getPointFromReference = function(_reference)
	if ATME.isPoint(_reference) ~= POINT_BAD then
		return ATMEFct.convertToPoint3D(_reference)
	else
		return _reference:getPosition()
	end
end

ATMEFct.coordLocalToDCS = function(_point, _pointB, _azimuth)
	-- coordLocalToDCS : Transforme les coordonnées locales d'un point relatives au repère défini par sa référence d'origine et son azimuth
	local angle = math.rad(_azimuth)
	
	local point
	
	if _point.z == nil then
		point = {
			x = _point.x,
			z = _point.y,
		}
	else
		point = _point
	end
	
	local pointDCS =  ATMEFct.convertToPoint3D({
										x = _pointB.x + point.x * math.cos(angle) - point.z * math.sin(angle),
										y = _pointB.z + point.x * math.sin(angle) + point.z * math.cos(angle),
									})

	if point.y ~= nil and point.y > pointDCS.y then
		pointDCS.y = point.y
	end
	
	return pointDCS
end

ATMEFct.coordLocalToDCSReference = function(_point, _originReference, _azimuth)
	local origin = ATMEFct.getPointFromReference(_originReference)
	
	return ATMEFct.coordLocalToDCS(_point, origin, _azimuth)
end

ATMEFct.hasAttributes = function(_object, _attributes)
	for _, _attr in ipairs(_attributes) do
		if _object:hasAttribute(ATMEConst.targetsAttributes[_attr]) == true or _attr == "ALL" then
			return true
		end
	end

	return false
end

ATMEFct.extractDatasFromStringOrTable = function(_attributes)
	local datas = {}
	
	if type(_attributes) == "string" then
		local id = 1
		for _item in _attributes:gmatch("([^%s]+)") do
			id = id + 1
			table.insert(datas, string.upper(_item))
		end
	elseif type(_attributes) == "table" then
		for _id, _item in ipairs(_attributes) do
			if type(_item) == "string" then
				table.insert(datas, string.upper(_item))
			else
				return true, "item #" .. _id .. " type must be string"
			end
		end
	else
		return true, "attributes type must be string or table"
	end
	
	return false, datas
end

ATMEFct.getSurfaceAltitude = function(_point)
	-- getSurfaceAltitude : Donne l'altitude du terrain à la verticale du point
	if  _point.z == nil then
		return land.getHeight({x = _point.x, y = _point.y})
	else
		return land.getHeight({x = _point.x, y = _point.z})
	end
end

ATMEFct.isSurfaceWater = function(_reference)
	local point = ATMEFct.getPointFromReference(_reference)
	local landType = land.getSurfaceType({
											x = point.x,
											y = point.z,
										})
										
	return landType == land.SurfaceType.WATER or landType == land.SurfaceType.SHALLOW_WATER
end

ATMEFct.getAzimuthMod = function(_azimuth, _mod)
	local azimuth = math.fmod(_azimuth, 360)

	if _mod == 360 then	
		if azimuth < 0 then 
			azimuth = azimuth + 360
		end
	else
		if azimuth < -180 then
			azimuth = azimuth + 360
		elseif azimuth > 180 then
			azimuth = azimuth - 360
		end
	end
	
	return azimuth
end

ATMEFct.testTakeOffFunctions = {
	["Turning Point"] = function()
		return false
	end,
	["TakeOffParkingHot"] = function()
		return true
	end,
	["TakeOffParking"] = function()
		return true
	end,
	["TakeOff"] = function()
		return true
	end,
	["TakeOffGroundHot"] = function()
		return true
	end,
	["TakeOffGround"] = function()
		return true
	end,
	["Land"] = function()
		return false
	end,
}

ATMEFct.isTakeOffWaypoint = function(_type)
	return ATMEFct.testTakeOffFunctions[_type]()
end

ATMEFct.isLandingWaypoint = function(_type)
	return _type == "Land"
end

ATMEFct.getWindVector = function(_reference, _withTurbulence)
	local point = ATMEFct.getPointFromReference(_reference)
			
	local alt = ATMEFct.getSurfaceAltitude(point)
	if point.y - alt < 20 then 
		point.y = alt + 20
	end
	
	if _withTurbulence == true then
		local wind = atmosphere.getWindWithTurbulence(point)
		return ATMEFct.vectors["3D"].createFromTable(nil, nil, wind) 
	else
		local wind = atmosphere.getWind(point)
		return ATMEFct.vectors["3D"].createFromTable(nil, nil, wind) 
	end
end

----------------------------------------------------------------------------------------------------------
-- DEFINITION DE LA TABLE ATME ET DES FONCTIONS GENERALES
--
-- Bien que reprenant la structure des classes définies plus loin, ATME n'est pas une classe en tant que telle.
-- ATME définit des fonctions de base et la liste des classes existantes dans ATME_Core. A ce titre, elle n'est
-- pas préfixée par C_
--
-- ATME_MT définit une métatable pour ATME (mt) et pour les extensions de modules afin d'avoir une variable globale
-- ATME.modules["nom du module"] pour les modules lancés. L'entrée "keys" interdit la création de clés nouvelles
-- dans la table ATME.
----------------------------------------------------------------------------------------------------------

ATMEDatas.ATME_MT = {
	mt = {},
	mtModules = {},

	keys = {
		["C_Vector3D"] = true,
		["C_Vector2D"] = true,
		["C_Line2D"] = true,
		["C_List"] = true,
		["C_Stack"] = true,
		["C_Airbase"] = true,
		["C_Zones"] = true,
		["C_Menu"] = true,
		["C_Alarm"] = true,
		["C_Module"] = true, 
		["C_Smoke"] = true,
		["C_Flare"] = true,
		["C_Event"] = true,
		["C_GroupCloningContext"] = true,
		["C_StaticCloningContext"] = true,
		["C_CargoCloningContext"] = true,
		["C_Group"] = true,
		["C_Unit"] = true,
		["C_Static"] = true,
		["C_Cargo"] = true,
		["C_Task"] = true,
		["C_Route"] = true,
		["C_Sound"] = true,
		["C_Circle"] = true,
		["C_Rectangle"] = true,
		["C_Polygon"] = true,
		["C_Zone"] = true,
		["C_Zone2D"] = true,
		["C_Zone3D"] = true,
		["C_MovableZone2D"] = true,
		["C_MovableZone3D"] = true,
		["C_FindContext"] = true,
		["C_Segment2D"] = true,
		["C_PickupZone"] = true,
		["C_Sphere"] = true,
		["C_Cylinder"] = true,
		["C_Box"] = true,
		["C_Weapon"] = true,
		["C_SceneryObject"] = true,
		["C_RadioChannel"] = true,
		["C_DataSharing"] = true,
		["C_Mark"] = true,
		["run"] = false,
		["loadDLL"] = true,
		["modules"] = false,
	},
}

ATME = {
	getLanguage = function()
		return ATMEDatas.language
	end,
	
	isPoint = function(_point)
		-- isPoint : Vérifie si la table passée en paramètre est un point et retourne POINT_BAD (0) si ce n'est pas un point :
		--           Si point 2D si présence de deux nombres _point.x et _point.y -> Retourne POINT_2D (1)
		--			 Si point 3D si présence de trois nombres _point.x, _point.y et _point.z/gModule -> Retourne POINT_3D (2)
		if type(_point) == "table" then
			if type(_point.x) == "number" and type(_point.y) == "number" then
				if type(_point.z) == "number" then
					return POINT_3D
				else
					return POINT_2D
				end
			end
		end
		
		return POINT_BAD
	end,
	
	isPoint2D = function(_point)
		-- isPoint2D : Vérifie si la table passée en paramètre est un point 2D, avec x et y renseignés
		--             Ces nombres sont relatifs au plan horizontal XZ de DCS, y étant en 2D la coordonnée sur Z.
		if type(_point) == "table" then
			if type(_point.x) == "number" and type(_point.y) == "number" then
				return true
			end
		end
		
		return false
	end,
	
	isPoint3D = function(_point)
		-- isPoint3D : Vérifie si la table passée en paramètre est un point 3D, avec x, y et z renseignés
		--             Ces nombres sont relatifs au repère XYZ de DCS
		if type(_point) == "table" then
			if type(_point.x) == "number" and type(_point.y) == "number" then
				if type(_point.z) == "number" then
					return true
				end
			end
		end
		
		return false
	end,
	
	getError = function()
		if ATMEDatas.error ~= nil then
			local errorCode = ATMEDatas.error.code
			ATMEDatas.error = nil
			
			return errorCode
		end
	end,
	
	isObject = function(_object, _class)
		local fctDefinition = ATMEFct.getErrorPrefix() .. "isObject(object, [class])"
		
		if _class ~= nil then
			if ATMEFct.isATMEObject(_class) == true then
				return ATMEFct.isObjectClass(_object, _class)
			elseif type(_class) == "string" then
				if ATMEDatas.ATME_MT.keys[_class] ~= nil and ATMEDatas.ATME_MT.keys[_class] == true then
					return ATMEFct.isObjectClass(_object, _class)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "class must be an ATME class name", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "class type must be string or ATME class", false)
			end
		else
			return ATMEFct.isATMEObject(_object)
		end
	end,
	
	getTheatre = function()
		-- getTheatre : Retourne la carte (map) associée à la mission
		-- 			    actuellement "NEVADA", "CAUCASUS", "NORMANDY" ou ""
		
		return ATMEDatas.theatreName
	end,
	
	getFlag = function(_id)
		-- getFlag : Retourne la valeur d'un flag. 
		-- 			 Pour être compatible avec le flags de l'éditeur de mission, il faut un id entier positif.
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "getFlag(id)"
		
		if type(_id) == "number" or type(_id) == "string" then
			local id
			if type(_id) == "number" then
				if _id > 0 then
					id = tostring(math.floor(_id))
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "id must be > 0 if number type", false)
					return
				end
			else
				id = _id
			end
			
			return trigger.misc.getUserFlag(id)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "id type must be number", false)
		end
	end,
	
	setFlag = function(_id, _value)
		-- setFlag : Modifie la valeur l'état d'un flag.
		--			 Pour être compatible avec le flags de l'éditeur de mission, il faut un id entier positif. Si le flag
		--			 est modifié ici et qu'il existe un trigger dans l'éditeur sur ce même flag pour la valeur renseignée,
		--           le trigger défini dans l'éditeur de mission sera déclenché
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "setFlag(id, value)"
		
		if (type(_id) == "number" or type(_id) == "string") and (type(_value) == "number" or type(_value) == "boolean") then
			local id
			if type(_id) == "number" then
				if _id > 0 then
					id = tostring(math.floor(_id))
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "id must be > 0 if number type", false)
					return
				end
			else
				id = _id
			end
			
			if (type(_value) == "number" and _value >= 0) or type(_value) == "boolean" then
				if type(_value) == "number" then
					_value = math.floor(_value)
				end
				
				trigger.action.setUserFlag(id, _value)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "value must be > 0 or true/false", false)
			end
		else
			if type(_id) ~= "number" and type(_id) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "id type must be number or string", false)
			elseif type(_value) ~= "number" and type(_value) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "value type must be number or boolean", false)
			end
		end
	end,
	
	getDCSZone = function(_name, _centerAltitude)
		-- getDCSZone : Retourne une zone définie dans l'éditeur de mission, nil si inexistante.
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "getDCSZone(name, [centerAltitude])"
		
		if type(_name) == "string" and (_centerAltitude == nil or type(_centerAltitude) == "boolean") then
			local centerAltitude = true
			
			if type(_centerAltitude) == "boolean" then
				centerAltitude = _centerAltitude
			end
			
			local zone = ATMEFct.getDCSZone(_name)
			
			if zone ~= nil then
				if centerAltitude == true then
					zone.point.y = ATMEFct.getSurfaceAltitude(zone.point)
				end
			end
			
			return zone
		else
			if type(_name) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
			elseif _centerAltitude ~= nil and type(_centerAltitude) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "centerAltitude type must be nil or boolean", false)
			end
		end
	end,
	
	getTime = function()
		-- getTime : Retourne le temps de la mission.
		
		return timer.getTime()
	end,
	
	getAzimuth360 = function(_azimuth)
		local fctDefinition = ATMEFct.getErrorPrefix() .. "getAzimuth360(azimuth)"
		
		if type(_azimuth) == "number" then
			return ATMEFct.getAzimuthMod(_azimuth, 360)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "azimuth type must be number", false)
		end
	end,
	
	getAzimuth180 = function(_azimuth)
		local fctDefinition = ATMEFct.getErrorPrefix() .. "getAzimuth180(azimuth)"
		
		if type(_azimuth) == "number" then
			return ATMEFct.getAzimuthMod(_azimuth, 180)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "azimuth type must be number", false)
		end
	end,
	
	copyTable = function(_tableToCopy)
		local fctDefinition = ATMEFct.getErrorPrefix() .. "copyTable(tableToCopy)"
		
		if type(_tableToCopy) == "table" then
			return ATMEFct.copyTable(_tableToCopy)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tableToCopy type must be table", false)
		end
	end,
	
	explode = function(_reference, _power)
		-- explode : Génère une explosion au point indiqué.
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "explode(reference, power)"

		if ATMEFct.isReference(_reference) == true and type(_power) == "number" then
			local power = math.floor(_power)
			if power > 0 then
				trigger.action.explosion(ATMEFct.getPointFromReference(_reference), power)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "power must be > 0", false)
			end
		else
			if ATMEFct.isReference(_reference) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
			elseif type(_power) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "power type must be number", false)
			end
		end
	end,
	
	displayForAll = function(_text, _duration, _clear)
		-- displayForAll : Affiche un message à tous les joueurs actifs avec un durée donnée
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "displayForAll(text, duration, [clear])"
		
		if type(_text) == "string" and type(_duration) == "number" and
				(_clear == nil or type(_clear) == "boolean") then
				
			local clear = false
			
			if type(_clear) == "boolean" then
				clear = _clear
			end
			
			if _duration >= 0 then
				trigger.action.outText(_text, _duration, clear)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "duration type must be >= 0", false)
			end
		else
			if type(_text) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "text type must be string", false)
			elseif type(_duration) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "duration type must be string", false)
			elseif _clear ~= nil and type(_clear) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "clear type must be nil or string", false)
			end
		end
	end,
	
	displayForCoalition = function(_coalitionName, _text, _duration, _clear)
		-- displayForCoalition : Affiche un message à tous les joueurs actifs d'une coalition avec une durée donnée
		--                       La coalition doit être "RED" ou "BLUE"
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "displayForCoalition(coalitionName, text, duration, [clear])"
		
		if type(_coalitionName) == "string" and type(_text) == "string" and type(_duration) == "number" and
				(_clear == nil or type(_clear) == "boolean") then
				
			local clear = false
			
			if type(_clear) == "boolean" then
				clear = _clear
			end
			
			if _duration >= 0 then
				local coalitionName = string.upper(_coalitionName)
				if ATMEConst.coalitionIdsBlueRed[coalitionName] ~= nil then
					trigger.action.outTextForCoalition(ATMEConst.coalitionIds[coalitionName], _text, _duration, clear)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName must be 'BLUE' or 'RED'", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "duration type must be >= 0", false)
			end
		else
			if type(_coalitionName) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName type must be string", false)
			elseif type(_text) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "text type must be string", false)
			elseif type(_duration) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "duration type must be string", false)
			elseif _clear ~= nil and type(_clear) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "clear type must be nil or string", false)
			end
		end
	end,
	
	displayForCountry = function(_countryName, _text, _duration, _clear)
		-- displayForCoalition : Affiche un message à tous les joueurs actifs d'une coalition avec une durée donnée
		--                       La nation est celle définie par DCS.
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "displayForCountry(countryName, text, duration, [clear])"
		
		if type(_countryName) == "string" and type(_text) == "string" and type(_duration) == "number" and
				(_clear == nil or type(_clear) == "boolean") then
			
			local clear = false
			
			if type(_clear) == "boolean" then
				clear = _clear
			end
			
			if _duration >= 0 then
				_countryName = string.upper(_countryName)
				if country.id[_countryName] ~= nil then
					trigger.action.outTextForCountry(country.id[_countryName], _text, _duration, clear)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "Country " .. _countryName .. " doesn't exist", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "duration type must be >= 0", false)
			end
		else
			if type(_countryName) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "countryName type must be string", false)
			elseif type(_text) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "text type must be string", false)
			elseif type(_duration) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "duration type must be string", false)
			elseif _clear ~= nil and type(_clear) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "clear type must be nil or string", false)
			end
		end
	end,
	
	logDisplay = function()
		-- logDisplay : Définit le niveau de débug. Tout module en mode debug est concerné. 

		ATMEDatas.logDisplay = true
	end,
	
	isSurfaceWater = function(_reference)
		-- isSurfaceWater : Vérifie si un point est au dessus de l'eau
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "isSurfaceWater(reference)"
		
		if ATMEFct.isReference(_reference) == true then
			return ATMEFct.isSurfaceWater(_reference)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
		end
	end,
	
	isSurfaceRunway = function(_reference)
		-- isSurfaceRunway : Vérifie si un point est au dessus d'une piste ou d'un taxiway
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "isSurfaceRunway(reference)"
		
		if ATMEFct.isReference(_reference) == true then
			local point = ATMEFct.getPointFromReference(_reference)
			
			local landType = land.getSurfaceType({
													x = point.x,
													y = point.z,
												})
			
			return landType == land.SurfaceType.RUNWAY
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
		end
	end,
	
	isSurfaceLand = function(_reference)
		-- isSurfaceNad : Vérifie si un point est au dessus de la terre hors piste ou taxiway
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "isSurfaceLand(reference)"
		
		if ATMEFct.isReference(_reference) == true then
			local point = ATMEFct.getPointFromReference(_reference)
			
			local landType = land.getSurfaceType({
													x = point.x,
													y = point.z,
												})
			
			return landType == land.SurfaceType.LAND
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
		end
	end,
	
	isSurfaceRoad = function(_reference)
		-- isSurfaceRoad : Vérifie si un point est au dessus de la terren hors piste ou taxiway
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "isSurfaceRoad(reference)"
		
		if ATMEFct.isReference(_reference) == true then
			local point = ATMEFct.getPointFromReference(_reference)
			
			local landType = land.getSurfaceType({
													x = point.x,
													y = point.z,
												})
			
			return landType == land.SurfaceType.ROAD
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
		end
	end,

	getSurfaceAltitude = function(_reference)
		-- getSurfaceAltitude : Donne l'altitude du terrain à la verticale du point
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "getSurfaceAltitude(reference)"
		
		if  ATMEFct.isReference(_reference) == true then
			return ATMEFct.getSurfaceAltitude(ATMEFct.getPointFromReference(_reference))
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
		end
	end,
	
	getPointAltitude = function(_point)
		-- getPointAltitude : Donne l'altitude du point. Si le point est 2D, ce sera
		--					  l'altitude du terrain
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "getPointAltitude(point)"
		
		local isPoint = ATME.isPoint(_point)
		if  isPoint ~= POINT_BAD then
			if isPoint == POINT_3D then
				return _point.y
			else
				return ATMEFct.getSurfaceAltitude(_point)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "point type must be table (x,y)/(x,y,z)", false)
		end
	end,
	
	getOnRoadPoint = function(_reference)
		local fctDefinition = ATMEFct.getErrorPrefix() .. "getOnRoadPoint(reference)"
		
		if ATMEFct.isReference(_reference) == true then
			local point = ATMEFct.getPointFromReference(_reference)
			
			local xTemp, yTemp = land.getClosestPointOnRoads("roads", point.x, point.z)
			
			return ATMEFct.convertToPoint3D({x = xTemp, y = yTemp})
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
		end
	end,
	
	convertToPoint3D = function(_point)
		-- Permet de convertir un point en point 3D
		-- afin de s'assurer que le point retourné est bien 3D.
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertToPoint3D(point)"
		
		if ATME.isPoint(_point) ~= POINT_BAD then
			return ATMEFct.convertToPoint3D(_point)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "point type must be table (x,y)/(x,y,z)", false)
		end
	end,
	
	areVisible = function(_refA, _refB)
		-- areVisible : Vérifie si deux références sont mutuellement visibles (ie non masqués) par le terrain
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "areVisible(refA, refB)"
		
		if ATMEFct.isReference(_refA) == true and ATMEFct.isReference(_refB) == true then
			local pA = ATMEFct.getPointFromReference(_refA)
			local pB = ATMEFct.getPointFromReference(_refB)
			
			-- Ajout d'un offset sur l'altitude pour éviter un pb si l'objet est sur le 
			-- sol avec une altitude dans le sol (cf doc DCS pour land.isVisible)
			pA.y = pA.y + 3
			pB.y = pB.y + 3
			return land.isVisible(pA, pB)
		else
			if ATMEFct.isReference(_refA) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "refA type must be" .. ATMEConst.errorLabelReference, false)
			elseif ATMEFct.isReference(_refB) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "refB type must be" .. ATMEConst.errorLabelReference, false)
			end
		end
	end,
	
	getTemperatureAndPressure = function(_reference)
		-- getTemperatureAndPressure : Retourne température et pression à hauter donnée.
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "getTemperatureAndPressure(reference)"
		
		if  ATMEFct.isReference(_reference) == true then
			local point = ATMEFct.getPointFromReference(_reference)
			
			return atmosphere.getTemperatureAndPressure(point) 
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
		end
	end,
	
	getWindVector = function(_reference, _withTurbulence)
		-- getWindVector : Retourne un instance de ATME.C_Vector3D correspond au vecteur vent
		--                 avec ou sans turbulences
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "getWindVector(reference, withTurbulence)"
		
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if  ATMEFct.isReference(_reference) == true and type(_withTurbulence) == "boolean" then
			return ATMEFct.getWindVector(_reference, _withTurbulence)
		else
			if ATMEFct.isReference(_reference) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
			elseif type(_withTurbulence) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "withTurbulence type must be boolean", false)
			end
		end
	end,
	
	getWindSpeed = function(_reference, _withTurbulence)
		-- getWindSpeed : Retourne la vitesse du vent calculée sur les 3 dimensions
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "getWindSpeed(reference, withTurbulence)"
		
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if ATMEFct.isReference(_reference) == true and type(_withTurbulence) == "boolean" then
			local wind = ATMEFct.getWindVector(_reference, _withTurbulence)

			return wind:getModule()
		else
			if ATMEFct.isReference(_reference) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
			elseif type(_withTurbulence) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "withTurbulence type must be boolean", false)
			end
		end
	end,
	
	getWindHSpeed = function(_reference, _withTurbulence)
		-- getWindSpeed : Retourne la vitesse horizontale du vent calculée sur les 2 dimensions x et z
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "getWindHSpeed(reference, withTurbulence)"
		
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if ATMEFct.isReference(_reference) == true and type(_withTurbulence) == "boolean" then
			local wind = ATMEFct.getWindVector(_reference, _withTurbulence)

			return wind:getHModule()
		else
			if ATMEFct.isReference(_reference) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
			elseif type(_withTurbulence) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "withTurbulence type must be boolean", false)
			end
		end
	end,
	
	getWindAzimuth = function(_reference, _withTurbulence)
		-- getWindAzimuth : Retourne l'azimuth du vent (attention pas son origine mais sa direction)
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "getWindAzimuth(reference, withTurbulence)"
		
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if ATMEFct.isReference(_reference) == true and type(_withTurbulence) == "boolean" then
			local wind = ATMEFct.getWindVector(_reference, _withTurbulence)

			return wind:getAzimuth()
		else
			if ATMEFct.isReference(_reference) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
			elseif type(_withTurbulence) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "withTurbulence type must be boolean", false)
			end
		end
	end,
	
	getTags = function(_text)
		local fctDefinition = ATMEFct.getErrorPrefix() .. "getTags(text)"
		
		if type(_text) == "string" then
			return ATMEFct.decompString(_text)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "text type must be string", false)
		end
	end,
		
	convertMtoNM = function(_distance)
		-- convertMtoNM : Convertit des mètres en miles nautiques
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertMtoNM(distance)"
		
		if type(_distance) == "number" then
			return _distance * 0.000539957
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "distance type must be number", false)
		end
	end,

	convertNMtoM = function(_distance)
		-- convertNMtoM : Convertit des miles nautiques en mètres
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertNMtoM(distance)"
		
		if type(_distance) == "number" then
			return _distance / 0.000539957
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "distance type must be number", false)
		end
	end,
	
	convertMtoFts = function(_distance)
		-- convertMtoFts : Convertit des mètres en pieds
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertMtoFts(distance)"
		
		if type(_distance) == "number" then
			return _distance * 3.28084
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "distance type must be number", false)
		end	
	end,
	
	convertFtstoM =  function(_distance)
		-- convertFtstoM : Convertit des pieds en mètres
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertFtstoM(distance)"
		
		if type(_distance) == "number" then
			return _distance / 3.28084
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "distance type must be number", false)
		end
	end,
	
	convertMpstoNds = function(_speed)
		-- convertMpstoNds : Convertit des m/s en noeuds
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertMpstoNds(speed)"
		
		if type(_speed) == "number" then
			return _speed * 1.9438
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "speed type must be number", false)
		end
	end,
	
	convertNdstoMps = function(_speed)
		-- convertNdstoMps : Convertit des noeuds en m/s
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertNdstoMps(speed)"
		
		if type(_speed) == "number" then
			return _speed / 1.9438
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "speed type must be number", false)
		end
	end,

	convertMpstoKph = function(_speed)
		-- convertMpstoKph : Convertit des m/s en km/h
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertMpstoKph(speed)"
		
		if type(_speed) == "number" then
			return _speed * 3.6
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "speed type must be number", false)
		end
	end,

	convertKphtoMps = function(_speed)
		-- convertKphtoMps : Convertit des km/h en m/s
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertKphtoMps(speed)"
		
		if type(_speed) == "number" then
			return _speed / 3.6
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "speed type must be number", false)
		end
	end,
	
	convertToDMS = function(_degrees)
		-- convertToDMS : Convertit des angles en degrés décimaux en degrés/minutes/secondes/centièmes
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertToDMS(degrees)"
		
		if type(_degrees) == "number" then
			return ATMEFct.convertToDMS(_degrees)
		else
			if type(_degrees) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "degrees type must be number", false)
			end
		end
	end,
	
	convertToDMd = function(_degrees)
		-- convertToDMd : Convertit des angles en degrés décimaux en degrés/minutes/décimales haute précision
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertToDMd(degrees)"
		
		if type(_degrees) == "number" then
			return ATMEFct.convertToDMd(_degrees, true)
		else
			if type(_degrees) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "degrees type must be number", false)
			end
		end
	end,
	
	LLToDMSString = function(_latitude, _longitude, _highPrecision)
		local fctDefinition = ATMEFct.getErrorPrefix() .. "LLToDMSString(latitude, longitude, [highPrecision])"
		
		if type(_latitude) == "number" and type(_longitude) == "number" and (_highPrecision == nil or type(_highPrecision) == "boolean") then
			local highPrecision = false
			if type(_highPrecision) == "boolean" then
				highPrecision = _highPrecision
			end
			
			local dirlat, dlat, mlat, slat, ddlat  = ATMEFct.convertToDMS(_latitude)
			local dirlon, dlon, mlon, slon, ddlon  = ATMEFct.convertToDMS(_longitude)
			
			local westEast = "N"
			if dirlat == -1 then
				westEast = "S"
			end
			
			local northSouth = "E"
			if dirlon == -1 then
				northSouth = "W"
			end
			
			local textlat
			local textlon

			if highPrecision == true then
				textlat = string.format("%s%02d°%02d'%02d\"%02d", westEast, dlat, mlat, slat, ddlat)
				textlon = string.format("%s%03d°%02d'%02d\"%02d", northSouth, dlon, mlon, slon, ddlon)
			else
				textlat = string.format("%s%02d°%02d'%02d\"", westEast, dlat, mlat, slat)
				textlon = string.format("%s%03d°%02d'%02d\"", northSouth, dlon, mlon, slon)
			end
			
			return textlat, textlon
		else
			if type(_latitude) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "latitude type must be number", false)
			elseif type(longitude) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "longitude type must be number", false)
			elseif _highPrecision ~= nil and type(_highPrecision) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "highPrecision type must be nil or boolean", false)
			end
		end
	end,
	
	LLToDMdString = function(_latitude, _longitude, _highPrecision)
		local fctDefinition = ATMEFct.getErrorPrefix() .. "LLToDMdString(latitude, longitude, [highPrecision])"
		
		if type(_latitude) == "number" and type(_longitude) == "number" and (_highPrecision == nil or type(_highPrecision) == "boolean") then
			local highPrecision = false
			if type(_highPrecision) == "boolean" then
				highPrecision = _highPrecision
			end

			local dirlat, dlat, mlat, slat  = ATMEFct.convertToDMd(_latitude, highPrecision)
			local dirlon, dlon, mlon, slon  = ATMEFct.convertToDMd(_longitude, highPrecision)

			local westEast = "N"
			if dirlat == -1 then
				westEast = "S"
			end
			
			local northSouth = "E"
			if dirlon == -1 then
				northSouth = "W"
			end
		
			local textlat
			local textlon
			if highPrecision == true then
				textlat = string.format("%s%02d°%02d.%04d", westEast, dlat, mlat, slat)
				textlon = string.format("%s%03d°%02d.%04d", northSouth, dlon, mlon, slon)
			else
				textlat = string.format("%s%02d°%02d.%02d", westEast, dlat, mlat, slat)
				textlon = string.format("%s%03d°%02d.%02d", northSouth, dlon, mlon, slon)	
			end
			
			return textlat, textlon
		else
			if type(_latitude) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "latitude type must be number", false)
			elseif type(longitude) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "longitude type must be number", false)
			elseif _highPrecision ~= nil and type(_highPrecision) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "highPrecision type must be nil or boolean", false)
			end
		end
	end,
	
	convertToHMS = function (_seconds)
		-- convertToHMS : Convertit un temps en secondes en heures/minutes/secondes/100ème
		-- En cas de temps négatif, retourne une valeur négative sur la première valeur significative,
		-- h, m ou s -> exemple : -60s donnera m = -1
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertToHMS(seconds)"
		
		if type(_seconds) == "number" then
			local seconds = _seconds
			local sign = 1
			if _seconds < 0 then
				seconds = -_seconds
				sign = -1
			end
	
			local h = math.floor(seconds/3600)
			local m =  math.floor((seconds - h * 3600)/60)
			local s = math.floor(seconds - h * 3600 - m * 60)
			local d = (seconds - math.floor(seconds)) * 100
			
			if h > 0 then
				h = h * sign
			elseif m > 0 then
				m = m * sign
			elseif s > 0 then
				s = s * sign
			else
				d = d * sign
			end
			
			return h, m, s, d
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "seconds type must be number", false)
		end
	end,

	convertPointToLL = function(_point)
		-- convertPointToLL : Convertit un point en coordonnées latitude/longitude/altitude
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertPointToLL(point)"
		
		if ATME.isPoint(_point) ~= POINT_BAD then
			return coord.LOtoLL(ATMEFct.convertToPoint3D(_point))
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "point type must be table (x,y)/(x,y,z)", false)
		end
	end,
	
	convertLLToPoint = function(_latitude, _longitude, _altitude)
		-- convertLLToPoint : Convertit un triplet latitude/longitude/altitude en un point 3D
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertLLToPoint(latitude, longitude, altitude)"
		
		if type(_latitude) == "number" and type(_longitude) == "number" and type(_altitude) == "number" then
			return coord.LLtoLO(_latitude, _longitude, _altitude)
		else
			if type(_latitude) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "latitude type must be number", false)
			elseif type(_longitude) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "longitude type must be number", false)
			elseif type(_altitude) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "altitude type must be number", false)
			end
		end
	end,
	
	convertLLToMGRS = function(_latitude, _longitude)
		-- convertLLToMGRS : Convertit une latitude/longitude en référence MGRS
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertLLToMGRS(latitude, longitude)"
		
		if type(_latitude) == "number" and type(_longitude) == "number" then
			return coord.LLtoMGRS(_latitude, _longitude)
		else
			if type(_latitude) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "latitude type must be number", false)
			elseif type(_longitude) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "longitude type must be number", false)
			end
		end
	end,
	
	convertMGRSToLL = function(_tableMGRS)
		-- convertMGRSToLL : Convertit une référence MGRS en latitude/longitude
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertMGRSToLL(tableMGRS)"
		
		if type(_tableMGRS) == "table" and type(table.UTMZone) == "string" and type(table.MGRSDigraph) == "string" and type(table.Easting) == "number" and type(table.Northing ) == "number" then
			return coord.MGRStoLL(_tableMGRS)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tableMGRS must have MGRS format", false)
		end
	end,
	
	getHScalarProductFromPoints = function(_A, _B, _C)
		-- HScalarProductFromPoints : Calcule le produit scalaire des vecteurs AB et AC, sur leurs coordonnées x,z
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "getHScalarProductFromPoints(A, B, C)"
		
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if ATME.isPoint(_A) ~= POINT_BAD and ATME.isPoint(_B) ~= POINT_BAD and ATME.isPoint(_C) ~= POINT_BAD then
			local vectAB = ATMEFct.vectors["3D"].createFromReference(nil, nil, _A, _B)
			local vectAC = ATMEFct.vectors["3D"].createFromReference(nil, nil, _A, _C)
			
			return vectAB:getX() * vectAC:getX() + vectAB:getZ() * vectAC:getZ()
		else
			if ATME.isPoint(_A) == POINT_BAD then
				ATMEFct.errorOutputCore(nil, fctDefinition, "A type must be table (x,y)/(x,y,z)", false)
			elseif ATME.isPoint(_B) == POINT_BAD then
				ATMEFct.errorOutputCore(nil, fctDefinition, "B type must be table (x,y)/(x,y,z)", false)
			elseif ATME.isPoint(_C) == POINT_BAD then
				ATMEFct.errorOutputCore(nil, fctDefinition, "C type must be table (x,y)/(x,y,z)", false)
			end
		end
	end,
	
	whichSide = function(_referenceA, _referenceB, _referenceC)
		-- whichSide : Vérifie si un point C est à droite ou à gauche d'un axe orienté AB :
		--               - 0 si le point est sur la droite
		--				 - 1 si le point est à droite
		--               - -1 si le point est à gauche
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "whichSide(referenceA, referenceB, referenceC)"
		
		if ATMEFct.isReference(_referenceA) == true and ATMEFct.isReference(_referenceB) == true and ATMEFct.isReference(_referenceC) == true then
			return ATMEFct.whichSide(ATMEFct.getPointFromReference(_referenceA), ATMEFct.getPointFromReference(_referenceB), ATMEFct.getPointFromReference(_referenceC))
		else
			if ATMEFct.isReference(_referenceA) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "referenceA type must be" .. ATMEConst.errorLabelReference, false)
			elseif ATMEFct.isReference(_referenceB) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "referenceB type must be" .. ATMEConst.errorLabelReference, false)
			elseif ATMEFct.isReference(_referenceC) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "referenceC type must be".. ATMEConst.errorLabelReference, false)
			end
		end
	end,
	
	coordLocalToDCS = function(_point, _originReference, _azimuth)
		-- coordLocalToDCS : Transforme les coordonnées locales d'un point relatives au repère défini par sa référence d'origine et son azimuth
		-- L'azimuth est en degrés. A noter : le repère est inversé x en vertical et repère xOy indirect (norme DCS)
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "coordLocalToDCS(point, originReference, azimuth)"

		if ATME.isPoint(_point) ~= POINT_BAD and ATMEFct.isReference(_originReference) == true and type(_azimuth) == "number" then
			return ATMEFct.coordLocalToDCSReference(_point, _originReference, _azimuth)
		else
			if ATMEFct.isPoint(_point) == POINT_BAD then
				ATMEFct.errorOutputCore(nil, fctDefinition, "point type must be table (x,y)/(x,y,z)", false)
			elseif ATMEFct.isReference(_originReference) == false  then
				ATMEFct.errorOutputCore(nil, fctDefinition, "originReference type must be" .. ATMEConst.errorLabelReference, false)
			elseif type(_azimuth) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "azimuth type must be number", false)
			end
		end
	end,
	
	coordDCSToLocal = function(_reference, _originReference, _azimuth)		
		-- localCoord : Transforme les coordonnées DCS d'une référence en coordonnées locales relatives au repère défini par sa référence 
		-- d'origine en coordonnées DCS et son azimuth
		-- L'azimuth est en degrés. A noter : le repère est inversé x en vertical et repère xOy indirect (norme DCS)
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "coordDCSToLocal(reference, originReference, azimuth)"
		
		if ATMEFct.isReference(_reference) == true and ATMEFct.isReference(_originReference) == true and type(_azimuth) == "number" then
			local angle = math.rad(_azimuth)
			local point = ATMEFct.getPointFromReference(_reference)
			local origin = ATMEFct.getPointFromReference(_originReference)
			
			return {
				x = (point.x - origin.x) * math.cos(angle) + (point.z - origin.z) * math.sin(angle),
				y = point.y,
				z = -(point.x - origin.x) * math.sin(angle) + (point.z - origin.z) * math.cos(angle),
			}
		else
			if ATMEFct.isReference(_reference) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
			elseif ATMEFct.isReference(_originReference) == false  then
				ATMEFct.errorOutputCore(nil, fctDefinition, "originReference type must be" .. ATMEConst.errorLabelReference, false)
			elseif type(_azimuth) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "azimuth type must be number", false)
			end
		end
	end,
	effectBigSmoke = function(_reference, _typeSmoke, _density)
		local fctDefinition = ATMEFct.getErrorPrefix() .. "effectBigSmoke(reference, typeSmoke, density)"
		
		if ATMEFct.isReference(_reference) == true and type(_typeSmoke) == "string" and type(_density) == "number" then
			if _density > 0 then
				local tokens = ATMEFct.decompStringSpaceOnly(_typeSmoke)
				local value = (1 - #tokens) * 4
				
				if #tokens >= 1 and #tokens <= 2 then
					if #tokens == 1 or tokens[2] == "FIRE" then
						if ATMEConst.bigSmokeType[tokens[1]] ~= nil then
							local point = ATMEFct.getPointFromReference(_reference)
							trigger.action.effectSmokeBig(point, ATMEConst.bigSmokeType[tokens[1]] + value, _density)
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "typeSmoke string bad format", false)
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "typeSmoke string bad format", false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "typeSmoke string bad format", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "density must be > 0", false)
			end
		else
			if ATMEFct.isReference(_reference) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
			elseif type(_typeSmoke) ~= "string"  then
				ATMEFct.errorOutputCore(nil, fctDefinition, "typeSmoke type must be string", false)
			elseif type(_azimuth) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "density type must be number", false)
			end
		end
	end,
}

setmetatable(ATME, ATMEDatas.ATME_MT.mt)

ATMEDatas.ATME_MT.mt.__metatable = "Private"
ATMEDatas.ATME_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.ATME_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "ATME : Bad key " .. k, "table locked", false)
	end
end


ATME.modules = {}  -- Déclaration de ATME.modules pour les variables globales des modules

setmetatable(ATME.modules, ATMEDatas.ATME_MT.mtModules)
ATMEDatas.ATME_MT.mtModules.__metatable = "Private"
ATMEDatas.ATME_MT.mtModules.__newindex = function(t, k, v)
	if ATMEDatas.modules[k] ~= nil then
		rawset(t, k, v)
	else  
		ATMEFct.errorOutputCore(nil,"ATME : ATME.modules Bad key " .. k, "table locked", false)
	end
end



----------------------------------------------------------------------------------------------------------
-- DEFINITION DES CLASSES D'OBJETS DEFINIS DANS ATME
--
-- Quelques points essentiels :
-- 			- Une classe d'objet est une table lua particulière, préfixée de "C_"
--			- Il existe une notion de classes virtuelles qui sont des regroupements de classes
--			- Une instance d'objet est un objet créé à partir d'une classe, c'est donc aussi une table qui peut
--			  être cependant étendue par le création de méthodes dans le constructeur local C_Class_create.
--			- Les fonctions liées à une classe et non à une instance sont définies dans la classe.
--			- Une classe d'objet a également une classe spéciale suffixée par _MT et locale à ATME_Core. Cette
--            classe définit : 
--				-> la metatable des instances (entrée "mt")
--				-> l'entrée "keys" des instances pour en interdire la création, lors de la création de l'instance. En effet,
--                 le constructeur local C_Class_create crée l'instance et lui associe des méthodes spécifique pour  
--				   les variables locales (closure). Ce sont ces méthodes, non gérées par la metaclasse qui sont listées
--				   dans l'entrée "keys" pour le contrôle des entrées de la table représentant l'instance.
--				-> le nom de la classe entrée "className"
--				-> la création du constructeur "public" de la classe (__call). Par défaut seule la fonction locale
--                 C_Class_create est créée ce qui interdit toute création d'instance dans les modules externes. C'est
--				   le cas pour les classes C_Group (Groupes au format ATME), C_Unit (Unités AI au format ATME) et 
--                 C_Static (objets statiques au format ATME sauf FARP et Airfields.
--				-> Certaines propriétés dont l'accès ne doit être que local et dont l'accès à la modification pourrait
--  			   engendrer des incohérences. Ces propriétés sont regroupées dans l'entrée "localMembers" et accessibles
--				   par les fonctions appropriées à partir du module ATME_Core uniquement (en lecture comme en écriture).
--				   L'instance contient un index spécifique accessible en lecture seule y compris à partir de modules
--				   externes, mais pour ces derniers cet index ne sert à rien.
----------------------------------------------------------------------------------------------------------


----------------------------------------------------------------------------------------------------------
-- DEFINITION DES CLASSES VIRTUELLES
----------------------------------------------------------------------------------------------------------

----------------------------------------------------------------------------------------------------------
--	ATME.Zone : 
--  Classe virtuelle pour
-- 		                   ATME.C_Zone2D
--		                   ATME.C_Zone3D
----------------------------------------------------------------------------------------------------------

ATMEFct.C_Zone_createNamed = function(_namedTable, _name, _fct, ...)
	if _namedTable[_name] ~= nil then
		ATMEFct.setError("ALREADY_CREATED")
		return true, "named zone already exists"
	else
		local d,f = string.find(_name, "|", 1)
		if d == nil then
			local args = {...}
			return false, _fct(_name, _namedTable, unpack(args))
		else
			ATMEFct.setError("BAD_FORMAT")
			return true, "name can't contain '|' character"
		end
	end
end

ATMEFct.C_Zone_randomPositionTest = function(_zone, _datas)
	local localMembersZone = ATMEDatas.localMembers[_zone:getIndex()]
	local datas = {
		zone = _zone,
		actualCenter = ATMEFct.C_Zone_getActualCenter(localMembersZone),
		actualAzimuth = ATMEFct.C_Zone_getActualAzimuth(localMembersZone),
	}

    local point = ATMEFct[ATMEConst.indZoneConfig[_zone:getClassName()].fctGetRandomPosition](datas)

	local landType = land.getSurfaceType({
						x = point.x,
						y = point.z})
						
	local isInWater = landType == land.SurfaceType.WATER or landType == land.SurfaceType.SHALLOW_WATER
	
	if _datas.surfaceFilter == "ALL" or (isInWater == false and _datas.surfaceFilter == "LAND") then
		if point.y >= _datas.altMin then
			if _datas.isFlat == true and _datas.surfaceFilter == "LAND" then
				local ok = true
				for _i = -1, 1 do
					for _j = -1, 1 do
						if _i ~= 0 or _j ~= 0 then
							local deltaX = _i * 10
							local deltaZ = _j * 10
							local d2 = deltaX * deltaX + deltaZ * deltaZ
							local h = ATMEFct.getSurfaceAltitude({x = point.x + deltaX, y = point.z + deltaZ}) - point.y

							-- Pente de 10% max soit h > d*10/100 -> h > d/10
							if h * h > d2 / 100 then
								ok = false
							end
						end
					end
				end
				
				if ok == true then
					return point
				end
			else
				return point
			end
		end
	elseif isInWater == true and _datas.surfaceFilter == "WATER" then
		return point
	end
end

ATMEFct.C_Zone_getRandomPosition = function(_datas)
	local nbTries = 0
	local aborted = false
	
	repeat
		repeat 
			_datas.point = ATMEFct.C_Zone_randomPositionTest(_datas.zone, _datas)
			nbTries = nbTries + 1
		until _datas.point ~= nil or nbTries % 10 == 0
		
		if _datas.localMembers.coroutines[_datas.moduleName] ~= nil and _datas.point == nil then
			aborted = _datas.localMembers.resultCallbacks[_datas.moduleName].aborted
			coroutine.yield()
			
		elseif _datas.localMembers.coroutines[_datas.moduleName] == nil and nbTries >= 100 then
			aborted = true
		end
	until _datas.point ~= nil or aborted == true
end

ATMEFct.C_Zone_abortResult = function(_localMembers)
	if _localMembers.resultCallbacks[ATMEDatas.currentModuleName] ~= nil then
		if _localMembers.coroutines[ATMEDatas.currentModuleName] ~= nil then
			if _localMembers.resultCallbacks[ATMEDatas.currentModuleName].idTimer ~= nil then
				timer.removeFunction(_localMembers.resultCallbacks[ATMEDatas.currentModuleName].idTimer)
				_localMembers.resultCallbacks[ATMEDatas.currentModuleName].idTimer = nil
			end
			
			_localMembers.resultCallbacks[ATMEDatas.currentModuleName].aborted = true
			while coroutine.status(_localMembers.coroutines[ATMEDatas.currentModuleName]) ~= "dead" do
				if coroutine.status(_localMembers.coroutines[ATMEDatas.currentModuleName]) == "suspended" then
					coroutine.resume(_localMembers.coroutines[ATMEDatas.currentModuleName])
				end
			end

			_localMembers.coroutines[ATMEDatas.currentModuleName] = nil
			_localMembers.resultCallbacks[ATMEDatas.currentModuleName].aborted = false
		end
	end
end

ATMEFct.C_Zone_searchResultFct = {
	["dead"] = function(_datas)
		if _datas.localMembers.resultCallbacks[_datas.moduleName] ~= nil then
			xpcall(function() _datas.localMembers.resultCallbacks[_datas.moduleName].fct(_datas.zone, _datas.point) end, function() ATMEFct.errUserHandler('Result callback') end)
		end

		_datas.localMembers.coroutines[_datas.moduleName] = nil
		return false
	end,
	["suspended"]= function(_datas)
		local ok = coroutine.resume(_datas.localMembers.coroutines[_datas.moduleName])
		return ok
	end,
	["running"] = function(_datas)
		return true
	end
}

ATMETimers.C_Zone_searchResultTimer = function(_datas)
	if ATMEDatas.exitATME == false then
		local ok = ATMEFct.C_Zone_searchResultFct[coroutine.status(_datas.localMembers.coroutines[_datas.moduleName])](_datas)

		if ok == true then
			return timer.getTime() + 0.01
		end
		
		_datas.localMembers.resultCallbacks[_datas.moduleName].idTimer = nil
	end
end

ATMEFct.C_Zone_getActualCenter = function(_localMembers)
	if _localMembers.boundUnit ~= nil then
		local localMembersUnit = ATMEDatas.localMembers[_localMembers.boundUnit.unit:getIndex()]
		local position
		
		if localMembersUnit.objectDCS:isExist() == true then
			position = localMembersUnit.objectDCS:getPosition()
		else
			position = localMembersUnit.lastPosition
		end
		
		local point = {
			x = _localMembers.boundUnit.distance * math.cos(_localMembers.boundUnit.azimuthRad),
			y = _localMembers.boundUnit.distance * math.sin(_localMembers.boundUnit.azimuthRad),
		}
		
		local vector = ATMEFct.vectors["3D"].createFromTable(nil, nil, position.x)
		local unitAngle = math.rad(vector:getAzimuth())
		
		if _localMembers.is3D == false then
			return {
				x = position.p.x + point.x * math.cos(unitAngle) - point.y * math.sin(unitAngle),
				y = 0,
				z = position.p.z + point.x * math.sin(unitAngle) + point.y * math.cos(unitAngle),
			}
		else
			return {
				x = position.p.x + point.x * math.cos(unitAngle) - point.y * math.sin(unitAngle),
				y = position.p.y + _localMembers.boundUnit.deltaAltitude,
				z = position.p.z + point.x * math.sin(unitAngle) + point.y * math.cos(unitAngle),
			}
		end
	end
	
	if _localMembers.center ~= nil then
		if _localMembers.is3D == false then
			return {
				x = _localMembers.center.x,
				y = 0,
				z = _localMembers.center.z,
			}
		else
			return {
				x = _localMembers.center.x,
				y = _localMembers.center.y,
				z = _localMembers.center.z,
			}
		end
	end
end

ATMEFct.C_Zone_getActualAzimuth = function(_localMembers)
	if _localMembers.boundUnit ~= nil and _localMembers.azimuth ~= nil then
		local localMembersUnit = ATMEDatas.localMembers[_localMembers.boundUnit.unit:getIndex()]
		
		local position
		if localMembersUnit.objectDCS:isExist() == true then
			position = localMembersUnit.objectDCS:getPosition().x
		else
			position = localMembersUnit.lastPosition.x
		end
		
		local vector = ATMEFct.vectors["3D"].createFromTable(nil, nil, position)
		return ATMEFct.getAzimuthMod(_localMembers.azimuth + vector:getAzimuth(), 360)
	else
		return _localMembers.azimuth
	end
end


ATMETimers.C_Zone_timer = function(_datas)
	if ATMEDatas.exitATME == false then
		if _datas.localMembers.tracking[_datas.name].localMembersObject.objectDCS:isExist() == true then
			ATMEFct.outputCore(nil, "Tracking sur zone ".. _datas.localMembers.tracking[_datas.name].name .. " - classe " .. _datas.localMembers.tracking[_datas.name].className , 2)
			if _datas.localMembers.tracking[_datas.name].className == "C_Weapon" or _datas.localMembers.tracking[_datas.name].localMembersObject.objectDCS:isActive() == true then
				local isInside = false
				
				local datas = {
					zone = _datas.zone,
					actualCenter = ATMEFct.C_Zone_getActualCenter(_datas.localMembers),
					actualAzimuth = ATMEFct.C_Zone_getActualAzimuth(_datas.localMembers),
				}

				if ATMEFct[ATMEConst.indZoneConfig[_datas.zone:getClassName()].fctIsInside](_datas.localMembers.tracking[_datas.name].object:getPosition(), datas) == true then
					isInside = true
				end

				if isInside == true then
					if _datas.localMembers.tracking[_datas.name].flag == nil or _datas.localMembers.tracking[_datas.name].flag  == false then
						_datas.localMembers.tracking[_datas.name].flag = true
						local datas = {
							isInside = true,
							object = _datas.localMembers.tracking[_datas.name].object,
						}
						
						for _moduleName, _callback in pairs(_datas.localMembers.tracking[_datas.name].callbacks) do
							ATMEDatas.currentModuleName = _moduleName
							
							ATMEFct.outputCore(nil,"Appel callback - Entrée zone pour unité " .. _datas.localMembers.tracking[_datas.name].name .. " pour le module " .. _moduleName, 2)
							xpcall(function() _callback(_datas.zone, datas) end, function() ATMEFct.errUserHandler('Tracking callback') end)
						end
					end
				else
					if _datas.localMembers.tracking[_datas.name].flag  == true then
						_datas.localMembers.tracking[_datas.name].flag = false
						local datas = {
							isInside = false,
							object = _datas.localMembers.tracking[_datas.name].object,
						}
						
						for _moduleName, _callback in pairs(_datas.localMembers.tracking[_datas.name].callbacks) do
							ATMEDatas.currentModuleName = _moduleName
							
							ATMEFct.outputCore(nil,"Appel callback - Sortie zone pour unité " .. _datas.localMembers.tracking[_datas.name].name .. " pour le module " .. _moduleName, 2)
							xpcall(function() _callback(_datas.zone, datas) end, function() ATMEFct.errUserHandler('Tracking callback') end)
						end
					end
				end
			end

			return timer.getTime() + ATMEConst.offsetTracking[_datas.localMembers.tracking[_datas.name].className]
		else
			_datas.localMembers.tracking[_datas.name] = nil
		end
	end
end

ATMEDatas.C_Zone_MT = {
	className = "C_Zone",
	
	mt = {},
}

ATME.C_Zone = {
	getClassName = function(self)
		return ATMEDatas.C_Zone_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Zone_MT.parentClass
	end,
	
	getByName = function(_name)
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Zone.getByName(name)", "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Zone.getByName(name)", "name type must be string", false)
		end
	end,
}

setmetatable(ATME.C_Zone, ATMEDatas.C_Zone_MT)

ATMEDatas.C_Zone_MT.mt.__metatable = "Private"
ATMEDatas.C_Zone_MT.mt.__index = ATME.C_Zone

ATMEDatas.C_Zone_MT.mt.__newindex = function(t, k, v)
	ATMEFct.errorOutputCore(nil, "Function ATME.C_Zone : Bad key " .. k, "table locked", false)
end

----------------------------------------------------------------------------------------------------------
--	ATME.Zone2D : 
--  Classe virtuelle pour
-- 		                   ATME.C_Polygon
--		                   ATME.C_MovableZone2D
--						   ATME.C_PickupZone
----------------------------------------------------------------------------------------------------------

ATMEDatas.C_Zone2D_MT = {
	className = "C_Zone2D",
	parentClass = ATME.C_Zone,
	
	mt = {},
}

ATME.C_Zone2D = {
	getClassName = function(self)
		return ATMEDatas.C_Zone2D_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Zone2D_MT.parentClass
	end,
	
	getByName = function(_name)
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Zone2D.getByName(name)", "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			if ATMEFct.isObjectClass(ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name], ATME.C_Zone2D) == true then
				return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name]
			end
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Zone2D.getByName(name)", "name type must be string", false)
		end
	end,
}

setmetatable(ATME.C_Zone2D, ATMEDatas.C_Zone2D_MT)

ATMEDatas.C_Zone2D_MT.mt.__metatable = "Private"
ATMEDatas.C_Zone2D_MT.mt.__index = ATME.C_Zone2D

ATMEDatas.C_Zone2D_MT.mt.__newindex = function(t, k, v)
	ATMEFct.errorOutputCore(nil, "Function ATME.C_Zone2D : Bad key " .. k, "table locked", false)
end

----------------------------------------------------------------------------------------------------------
--	ATME.Zone3D : 
--  Classe virtuelle pour
-- 		                   ATME.C_MovableZone3D
----------------------------------------------------------------------------------------------------------

ATMEDatas.C_Zone3D_MT = {
	className = "C_Zone3D",
	parentClass = ATME.C_Zone,
	
	mt = {},
}

ATME.C_Zone3D = {
	getClassName = function(self)
		return ATMEDatas.C_Zone3D_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Zone3D_MT.parentClass
	end,
	
	getByName = function(_name)
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil,  "Function ATME.C_Zone3D.getByName(name)", "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			if ATMEFct.isObjectClass(ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name], ATME.C_Zone3D) == true then
				return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name]
			end
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Zone3D.getByName(name)", "name type must be string", false)
		end
	end,
}

setmetatable(ATME.C_Zone3D, ATMEDatas.C_Zone3D_MT)

ATMEDatas.C_Zone3D_MT.mt.__metatable = "Private"
ATMEDatas.C_Zone3D_MT.mt.__index = ATME.C_Zone3D

ATMEDatas.C_Zone3D_MT.mt.__newindex = function(t, k, v)
	ATMEFct.errorOutputCore(nil, "Function ATME.C_Zone3D : Bad key " .. k, "table locked", false)
end

----------------------------------------------------------------------------------------------------------
--	ATME.MovableZone2D : Définit les zones 2D pouvant être rattachées à une unité si associée à un zones
--  Classe virtuelle pour
--		                   ATME.C_Circle
--						   ATME.C_Rectangle
----------------------------------------------------------------------------------------------------------

ATMEDatas.C_MovableZone2D_MT = {
	className = "C_MovableZone2D",
	parentClass = ATME.C_Zone2D,
	
	mt = {},
}

ATME.C_MovableZone2D = {
	getClassName = function(self)
		return ATMEDatas.C_MovableZone2D_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_MovableZone2D_MT.parentClass
	end,
	
	getByName = function(_name)
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, "Function ATME.C_MovableZone2D.getByName(name)", "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			if ATMEFct.isObjectClass(ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name], ATME.C_MovableZone2D) == true then
				return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name]
			end
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_MovableZone2D.getByName(name)", "name type must be string", false)
		end
	end,
}

setmetatable(ATME.C_MovableZone2D, ATMEDatas.C_MovableZone2D_MT)

ATMEDatas.C_MovableZone2D_MT.mt.__metatable = "Private"
ATMEDatas.C_MovableZone2D_MT.mt.__index = ATME.C_MovableZone2D

ATMEDatas.C_MovableZone2D_MT.mt.__newindex = function(t, k, v)
	ATMEFct.errorOutputCore(nil, "Function ATME.C_MovableZone2D : Bad key " .. k, "table locked", false)
end

----------------------------------------------------------------------------------------------------------
--	ATME.MovableZone3D : Définit les zones 3D pouvant être rattachées à une unité si associée à un zones
--  Classe virtuelle pour
--		                   ATME.C_Sphere
--						   ATME.C_Cylinder
--                         ATME.C_Box
----------------------------------------------------------------------------------------------------------

ATMEDatas.C_MovableZone3D_MT = {
	className = "C_MovableZone3D",
	parentClass = ATME.C_Zone3D,
	
	mt = {},
}

ATME.C_MovableZone3D = {
	getClassName = function(self)
		return ATMEDatas.C_MovableZone3D_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_MovableZone3D_MT.parentClass
	end,
	
	getByName = function(_name)
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, "Function ATME.C_MovableZone3D.getByName(name)", "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			if ATMEFct.isObjectClass(ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name], ATME.C_MovableZone3D) == true then
				return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name]
			end
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_MovableZone3D.getByName(name)", "name type must be string", false)
		end
	end,
}

setmetatable(ATME.C_MovableZone3D, ATMEDatas.C_MovableZone3D_MT)

ATMEDatas.C_MovableZone3D_MT.mt.__metatable = "Private"
ATMEDatas.C_MovableZone3D_MT.mt.__index = ATME.C_MovableZone3D

ATMEDatas.C_MovableZone3D_MT.mt.__newindex = function(t, k, v)
	ATMEFct.errorOutputCore(nil, "Function ATME.C_MovableZone3D : Bad key " .. k, "table locked", false)
end

----------------------------------------------------------------------------------------------------------
-- DEFINITION DES CLASSES STANDARD
----------------------------------------------------------------------------------------------------------

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_List : gère une liste indexée par des entiers. Constructeur de classe accessible aux modules 
-- externes. En cas de suppression, l'index des suivants ne change pas.
----------------------------------------------------------------------------------------------------------

ATMEDatas.C_List_MT = {
	className = "C_List",
	mt = {},
	
	keys = {
		["add"] = true,
		["remove"] = true,
		["get"] = true,
		["getFirst"] = true,
		["getNext"] = true,
		["getPrevious"] = true,
		["getRandom"] = true,
		["getLast"] = true,
		["pairs"] = true,
		["getCount"] = true,
		["removeAll"] = true,
	},
}

ATME.C_List = {
	getClassName = function(self)
		return ATMEDatas.C_List_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_List_MT.parentClass
	end,
}

ATMEFct.C_List_create = function()
	local maxIndex = 1
	local idsRef = {}
	local list = {}
	local newList = {}
	local firstValidId
	local lastValidId
	local scanUp, scanDown
		
	setmetatable(newList, ATMEDatas.C_List_MT.mt)
	
	function newList.add(self, _item)
		local fctDefinition = ATMEFct.getErrorPrefix(newList) .. "add(item)"
		
		if _item ~= nil then
			local index = maxIndex
			
			if #idsRef == 0 then
				firstValidId = index
			else
				list[lastValidId].next = index
			end
			
			list[index] = {
				item = _item,
				previous = lastValidId,
			}
			
			lastValidId = index
			table.insert(idsRef, index)
			
			maxIndex = maxIndex + 1
			
			return index
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "item type can't be nil", true)
		end
	end
	
	function newList.remove(self, _id)
		local fctDefinition = ATMEFct.getErrorPrefix(newList) .. "remove([id])"
		if _id == nil or type(_id) == "number" then
			local id = firstValidId
			
			if _id ~= nil then
				id = math.floor(_id)
			end
			
			if id > 0 then
				if list[id] ~= nil then
					if #idsRef == 1 then
						firstValidId = nil
						lastValidId = nil
						maxIndex = 1
						
						table.remove(idsRef, 1)
					else
						local previousId = list[id].previous
						local nextId = list[id].next

						if id == scanDown then
							scanDown = list[id].previous
						elseif id == scanUp then
							scanUp = list[id].next
						end
						
						if  previousId ~= nil then
							list[previousId].next = nextId
						else
							firstValidId = nextId
						end
						
						if nextId ~= nil then
							list[nextId].previous = previousId
						else
							lastValidId = previousId
						end
						
						local minId = 1
						local maxId = #idsRef
						local found
						
						if idsRef[minId] == id then
							found = minId
						elseif idsRef[maxId] == id then
							found = maxId
						end
						
						local delta = maxId - minId
						local currentId
						while found == nil do
							currentId = math.floor(delta/2) + minId
							
							if idsRef[currentId] == id then
								found = currentId
							else
								if idsRef[currentId] > id then
									maxId = currentId
								else
									minId = currentId
								end
								
								delta = maxId - minId
							end
						end
						
						table.remove(idsRef, found)
					end
					
					
					local item = list[id].item
					list[id] = nil
					
					return item
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "id must be >0", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "id type must be nil or number", true)
		end
	end
	
	function newList.get(self,_id)
		local fctDefinition = ATMEFct.getErrorPrefix(newList) .. "get(id)"
		if type(_id) == "number" then
			local id = math.floor(_id)
			if id > 0 then
				if list[id] ~= nil then
					scanDown = list[id].previous
					scanUp = list[id].next
			
					return list[id].item, id
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "id must be >0", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "id type must be number", true)
		end
	end
	
	function newList.getFirst(self)
		if firstValidId ~= nil then
			scanDown = nil
			scanUp = list[firstValidId].next
			return list[firstValidId].item, firstValidId
		end
	end
	
	function newList.getLast(self)
		if lastValidId ~= nil then
			scanDown = list[lastValidId].previous
			scanUp = nil
			return list[lastValidId].item, lastValidId
		end
	end
	
	function newList.getNext(self)
		if scanUp ~= nil and list[scanUp] ~= nil then
			local id = scanUp
			scanDown = list[id].previous
			scanUp = list[id].next
			
			return list[id].item, id
		else
			scanUp = nil
		end
	end
	
	function newList.getPrevious(self)
		if scanDown ~= nil and list[scanDown] ~= nil then
			local id = scanDown
			scanDown = list[id].previous
			scanUp = list[id].next
			
			return list[id].item, id
		else
			scanDown = nil
		end
	end
	
	function newList.getCount(self)
		return #idsRef
	end
	
	function newList.getRandom(self)
		if #idsRef > 0 then
			local rand = math.random(1, #idsRef)
			return list[idsRef[rand]].item, idsRef[rand]
		end
	end
	
    function newList.pairs(self)
		local i = firstValidId
		return function()
			
			if i ~= nil then
				local index = i
				i = list[index].next

				return index, list[index].item
			end
		end
    end
	
	function newList.removeAll(self)
		list = {}
		idsRef = {}
		maxIndex = 1
		firstValidId = nil
		lastValidId = nil
	end
	
	return newList
end

ATMEDatas.C_List_MT.__call = function(t)
	return ATMEFct.C_List_create()
end

setmetatable(ATME.C_List,ATMEDatas.C_List_MT)

ATMEDatas.C_List_MT.mt.__metatable = "Private"
ATMEDatas.C_List_MT.mt.__index = ATME.C_List
ATMEDatas.C_List_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_List_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else  
		ATMEFct.errorOutputCore(nil, "Function ATME.C_List : Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Stack : gère une pile. Constructeur de classe accessible aux modules 
-- externes. 
-- Elle permet une gestion des piles FIFO et LIFO
----------------------------------------------------------------------------------------------------------

ATMEDatas.C_Stack_MT = {
	className = "C_Stack",
	mt = {},
	
	keys = {
		["push"] = true,
		["extract"] = true,
		["pairs"] = true,
		["getCount"] = true,
		["removeAll"] = true,
		["getIndex"] = true,
	},
	
	add = function(self)
		local item = {
			nbItems= 0,
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
	end,
}

ATME.C_Stack = {
	getClassName = function(self)
		return ATMEDatas.C_Stack_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Stack_MT.parentClass
	end,
}

ATMEFct.C_Stack_extract = function(_localMembers, _typeExtract)
	if _localMembers[ATMEConst.stackTypeExtract[_typeExtract].load] ~= nil then
		local item = _localMembers[ATMEConst.stackTypeExtract[_typeExtract].load].item
		if _localMembers.nbItems ~= 1 then
			_localMembers[ATMEConst.stackTypeExtract[_typeExtract].load] = _localMembers[ATMEConst.stackTypeExtract[_typeExtract].load][ATMEConst.stackTypeExtract[_typeExtract].direction]
			_localMembers[ATMEConst.stackTypeExtract[_typeExtract].load][ATMEConst.stackTypeExtract[_typeExtract].raz] = nil
		else
			_localMembers.first = nil
			_localMembers.last = nil
		end
		
		_localMembers.nbItems = _localMembers.nbItems - 1
		
		return item
	end
end

ATMEFct.C_Stack_create = function()
	local newStack = {}

	setmetatable(newStack, ATMEDatas.C_Stack_MT.mt)
	local index = ATMEDatas.C_Stack_MT:add()
	local localMembers = ATMEDatas.localMembers[index]

	function newStack.getIndex(self)
		return index
	end
	
	function newStack.push(self, _item)
		local data = {
			item = _item,
		}

		if localMembers.first == nil then
			localMembers.first = data
			localMembers.last = data
		else
			localMembers.first.previous = data
			data.next = localMembers.first
			localMembers.first = data
		end

		localMembers.nbItems = localMembers.nbItems + 1
	end
	
	function newStack.extract(self, _typeExtract)
		local fctDefinition = ATMEFct.getErrorPrefix(newStack) .. "extract([type])"
		
		if _typeExtract == nil or type(_typeExtract) == "string" then
			local typeExtract
			if _typeExtract == nil then
				typeExtract = "TAIL"
			else
				typeExtract = string.upper(_typeExtract)
			end
			
			if ATMEConst.stackTypeExtract[typeExtract] ~= nil then		
				return ATMEFct.C_Stack_extract(localMembers, typeExtract)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "type must be 'HEAD' or 'TAIL'", false)
			end
		else
			if _typeExtract ~= nil and type(_typeExtract) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "type type must be nil or string", true)
			end
		end
	end
	
	function newStack.getCount(self)
		return localMembers.nbItems
	end
	
    function newStack.pairs(self, _typeRead)
		local data
		local i = 0
		
		local typeRead
		if _typeRead == nil or type(_typeRead) ~= "string" then
			typeRead = "TAIL"
		else
			typeRead = _typeRead:upper()
			if ATMEConst.stackTypeExtract[typeRead] == nil then
				typeRead = "TAIL"
			end
		end
		
		return function()
			if data == nil then
				data = localMembers[ATMEConst.stackTypeExtract[typeRead].load]
			else
				data = data[ATMEConst.stackTypeExtract[typeRead].direction]
			end
			
			i = i + 1
			
			if data ~= nil then
				return i, data.item
			end
		end
    end
	
	function newStack.removeAll(self)
		localMembers.first = nil
		localMembers.last = nil
		localMembers.nbItems = 0
	end

	return newStack
end

ATMEDatas.C_Stack_MT.__call = function(t)
	return ATMEFct.C_Stack_create()
end

setmetatable(ATME.C_Stack,ATMEDatas.C_Stack_MT)

ATMEDatas.C_Stack_MT.mt.__metatable = "Private"
ATMEDatas.C_Stack_MT.mt.__index = ATME.C_Stack
ATMEDatas.C_Stack_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Stack_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else  
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Stack : Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Variables globales utilisant les C_Index_List & C_Stack
----------------------------------------------------------------------------------------------------------
ATMEDatas.signalCapabilityUnits = ATMEFct.C_List_create()

----------------------------------------------------------------------------------------------------------
-- Fonctions générales
----------------------------------------------------------------------------------------------------------		


-- Fonction de disable pour les units
ATMEFct.C_Unit_disable = function(_unit, _reserved)
	local localMembers = ATMEDatas.localMembers[_unit:getIndex()]
	local localMembersGroup = ATMEDatas.localMembers[localMembers.group:getIndex()]
	localMembers.isEnabled = false
	
	ATMEFct.C_Unit_playerLeaves(_unit, localMembersGroup)
	localMembers.objectDCS:destroy()
	ATMEFct.C_Unit_delete(_unit, _reserved)
end			

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_DataSharing : Gère les partages de données entre joueur pour les sons, les marqueurs
-- et les affichages. Les niveaux hiérarchisés sont :
--		- all -> Echanges entre tous les joueurs de la mission
--		- coalition -> Echanges entre les joueurs d'une coalition "BLUE" ou "RED"
--		- nation -> Echanges entre les joueurs d'une nation DCS
--		- groupe -> Echanges entre les joueurs d'un groupe particulier
----------------------------------------------------------------------------------------------------------

ATMEFct.C_DataSharing_initPlaylist = function(_soundDatas)
	_soundDatas.currentReadIndex = 1
	
	ATMEFct.C_DataSharing_soundsOnOff[_soundDatas.localMembers.level](true, _soundDatas)

	_soundDatas.localMembers.idTimer = timer.scheduleFunction(ATMETimers.C_DataSharing_timer, _soundDatas, timer.getTime() + 1)
end

ATMEFct.C_DataSharing_soundsOnOff = {
	["groups"] = function(_flag, _soundDatas)
		ATMEDatas.soundsOn.groups[_soundDatas.localMembersGroup.objectDCSId] = _flag
	end,
	
	["coalitions"] = function(_flag, _soundDatas)
		ATMEDatas.soundsOn.coalitions[_soundDatas.localMembers.coalition] = _flag
	end,
	
	["countries"] = function(_flag, _soundDatas)
		ATMEDatas.soundsOn.countries[_soundDatas.localMembers.country] = _flag
	end,
	
	["all"] = function(_flag, _soundDatas)
		ATMEDatas.soundsOn.all = _flag
	end,
}

ATMEFct.C_DataSharing_testSoundsOn = {
	["groups"] = function(_soundDatas)
		return ATMEDatas.soundsOn.all == nil and 
					ATMEDatas.soundsOn.coalitions[_soundDatas.localMembers.coalition] == nil and 
					ATMEDatas.soundsOn.countries[_soundDatas.localMembers.country] == nil and 
					ATMEDatas.soundsOn.groups[_soundDatas.localMembersGroup.objectDCSId] == nil 
	end,
	
	["coalitions"] = function(_soundDatas)
		return ATMEDatas.soundsOn.all == nil and 
					ATMEDatas.soundsOn.coalitions[_soundDatas.localMembers.coalition] == nil
	end,
	
	["countries"] = function(_soundDatas)
		return ATMEDatas.soundsOn.all == nil and 
					ATMEDatas.soundsOn.coalitions[_soundDatas.localMembers.coalition] == nil and 
					ATMEDatas.soundsOn.countries[_soundDatas.localMembers.country] == nil
	end,
	
	["all"] = function(_soundDatas)
		return ATMEDatas.soundsOn.all == nil
	end,
}

ATMEFct.C_DataSharing_continue = {
	["groups"] = function(_soundDatas)
		if ATMEDatas.soundsOn.all == nil and 
					ATMEDatas.soundsOn.coalitions[_soundDatas.localMembers.coalition] == nil and 
					ATMEDatas.soundsOn.countries[_soundDatas.localMembers.country] == nil and 
					ATMEDatas.soundsOn.groups[_soundDatas.localMembersGroup.objectDCSId] ~= nil then
			
			trigger.action.outSoundForGroup(_soundDatas.localMembersGroup.objectDCSId, _soundDatas.localMembers.currentVoice .. _soundDatas.playlist[_soundDatas.currentReadIndex].fileName)

			if _soundDatas.playlist[_soundDatas.currentReadIndex].subtitle.duration > 0 then
				trigger.action.outTextForGroup(_soundDatas.localMembersGroup.objectDCSId, _soundDatas.playlist[_soundDatas.currentReadIndex].subtitle.text, _soundDatas.playlist[_soundDatas.currentReadIndex].subtitle.duration, _soundDatas.localMembers.clear)
			end
		else
			ATMEFct.C_DataSharing_soundsOnOff[_soundDatas.localMembers.level](nil, _soundDatas)
			timer.removeFunction(_soundDatas.localMembers.idTimer)
			_soundDatas.localMembers.idTimer = nil
			_soundDatas.localMembers.soundDatas = nil
		end
	end,
	
	["coalitions"] = function(_soundDatas)
		if ATMEDatas.soundsOn.all == nil and 
					ATMEDatas.soundsOn.coalitions[_soundDatas.localMembers.coalition] ~= nil then
			
			trigger.action.outSoundForCoalition(ATMEConst.coalitionIds[_soundDatas.localMembers.coalition], _soundDatas.localMembers.currentVoice .. _soundDatas.playlist[_soundDatas.currentReadIndex].fileName)

			if _soundDatas.playlist[_soundDatas.currentReadIndex].subtitle.duration > 0 then
				trigger.action.outTextForCoalition(ATMEConst.coalitionIds[_soundDatas.localMembers.coalition], _soundDatas.playlist[_soundDatas.currentReadIndex].subtitle.text, _soundDatas.playlist[_soundDatas.currentReadIndex].subtitle.duration, _soundDatas.localMembers.clear)
			end
		else
			timer.removeFunction(_soundDatas.localMembers.idTimer)
			ATMEFct.C_DataSharing_soundsOnOff[_soundDatas.localMembers.level](nil, _soundDatas)
			_soundDatas.localMembers.idTimer = nil
			_soundDatas.localMembers.soundDatas = nil
		end
	end,
	
	["countries"] = function(_soundDatas)
		if ATMEDatas.soundsOn.all == nil and 
					ATMEDatas.soundsOn.coalitions[_soundDatas.localMembers.coalition] == nil and 
					ATMEDatas.soundsOn.countries[_soundDatas.localMembers.country] ~= nil then
			
			trigger.action.outSoundForCountry(_soundDatas.localMembers.country, _soundDatas.localMembers.currentVoice .. _soundDatas.playlist[_soundDatas.currentReadIndex].fileName)

			if _soundDatas.playlist[_soundDatas.currentReadIndex].subtitle.duration > 0 then
				trigger.action.outTextForCountry(_soundDatas.localMembers.country, _soundDatas.playlist[_soundDatas.currentReadIndex].subtitle.text, _soundDatas.playlist[_soundDatas.currentReadIndex].subtitle.duration, _soundDatas.localMembers.clear)
			end
		else
			ATMEFct.C_DataSharing_soundsOnOff[_soundDatas.localMembers.level](nil, _soundDatas)
			timer.removeFunction(_soundDatas.localMembers.idTimer)
			_soundDatas.localMembers.idTimer = nil
			_soundDatas.localMembers.soundDatas = nil
		end
	end,
	
	["all"] = function(_soundDatas)
		if ATMEDatas.soundsOn.all ~= nil then 
			trigger.action.outSound(_soundDatas.localMembers.currentVoice .. _soundDatas.playlist[_soundDatas.currentReadIndex].fileName)

			if _soundDatas.playlist[_soundDatas.currentReadIndex].subtitle.duration > 0 then
				trigger.action.outText(_soundDatas.playlist[_soundDatas.currentReadIndex].subtitle.text, _soundDatas.playlist[_soundDatas.currentReadIndex].subtitle.duration, _soundDatas.localMembers.clear)
			end
		else
			ATMEFct.C_DataSharing_soundsOnOff[_soundDatas.localMembers.level](nil, _soundDatas)
			timer.removeFunction(_soundDatas.localMembers.idTimer)
			_soundDatas.localMembers.idTimer = nil
			_soundDatas.localMembers.soundDatas = nil
		end
	end,
}

ATMEFct.C_DataSharing_display = {
	["groups"] = function(_displayDatas)
		if _displayDatas.localMembersGroup.objectDCS:isExist() == true and #_displayDatas.localMembersGroup.objectDCS:getUnits() ~= 0 then
			trigger.action.outTextForGroup(_displayDatas.localMembersGroup.objectDCSId, _displayDatas.text, _displayDatas.duration, _displayDatas.localMembers.clear)
		end
	end,
	
	["coalitions"] = function(_displayDatas)
		trigger.action.outTextForCoalition(ATMEConst.coalitionIds[_displayDatas.localMembers.coalition], _displayDatas.text, _displayDatas.duration, _displayDatas.localMembers.clear)
	end,
	
	["countries"] = function(_displayDatas)
		trigger.action.outTextForCountry(_displayDatas.localMembers.country, _displayDatas.text, _displayDatas.duration, _displayDatas.localMembers.clear)
	end,
	
	["all"] = function(_displayDatas)
		trigger.action.outText(_displayDatas.text, _displayDatas.duration, _displayDatas.localMembers.clear)
	end,
}

ATMEFct.C_DataSharing_mark = {
	["groups"] = function(_markDatas)
		if _markDatas.localMembersGroup.objectDCS:isExist() == true and #_markDatas.localMembersGroup.objectDCS:getUnits() ~= 0 then
			trigger.action.markToGroup(_markDatas.id, _markDatas.text, _markDatas.point, _markDatas.localMembersGroup.objectDCSId, _markDatas.readOnly, _markDatas.message)
			return true
		else
			ATMEFct.setError("BAD_MARK")
			return false, "Group '" .. _markDatas.localMembersGroup.name .. "' doesn't exist"
		end
	end,
	
	["coalitions"] = function(_markDatas)
		trigger.action.markToCoalition(_markDatas.id, _markDatas.text, _markDatas.point, ATMEConst.coalitionIds[_displayDatas.localMembers.coalition], _markDatas.readOnly, _markDatas.message)
		return true
	end,
	
	["countries"] = function(_markDatas)
		-- n'existe pas DCS 2.7
		ATMEFct.setError("BAD_MARK")
		return false, "No mark for countries"
	end,
	
	["all"] = function(_markDatas)
		trigger.action.markToAll(_markDatas.id, _markDatas.text, _markDatas.point, _markDatas.readOnly, _markDatas.message)
		return true
	end,
}

ATMETimers.C_DataSharing_timer = function(_soundDatas)
	if ATMEDatas.exitATME == false then
		if _soundDatas.randomRead == true then 
			_soundDatas.currentReadIndex = math.random(1, #_soundDatas.playlist)
		end

		local tips = _soundDatas.playlist[_soundDatas.currentReadIndex].tipsBase
		
		ATMEFct.C_DataSharing_continue[_soundDatas.localMembers.level](_soundDatas)
		_soundDatas.currentReadIndex = _soundDatas.currentReadIndex + 1
		
		if _soundDatas.currentReadIndex > #_soundDatas.playlist then
			if _soundDatas.readOnce == true then
				ATMEFct.C_DataSharing_soundsOnOff[_soundDatas.localMembers.level](nil, _soundDatas)
				_soundDatas.localMembers.soundDatas = nil
				_soundDatas.localMembers.idTimer = nil
				return
			else
				_soundDatas.currentReadIndex = 1
			end
		end

		return timer.getTime() + tips
	end
end

ATMEFct.C_DataSharing_setChannel = {
	["ALL"] = function()
		return "all"
	end,
	["BLUE"] = function()
		return "coalitions", "BLUE"
	end,
	["RED"] = function()
		return "coalitions", "RED"
	end,
}

ATMEFct.C_DataSharing_createNamed = function(_namedTable, _name, _level, _channel, _country)
	if _namedTable[_name] ~= nil then
		ATMEFct.setError("ALREADY_CREATED")
		return true, "named Radio channel already exists"
	else
		local d,f = string.find(_name, "|", 1)
		if d == nil then
			return ATMEFct.C_DataSharing_create(_name, _namedTable, _level, _channel, _country)
		else
			ATMEFct.setError("BAD_FORMAT")
			return true, "name can't contain '|' character"
		end
	end
end

ATMEDatas.C_DataSharing_MT = {
	className = "C_DataSharing",
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["getUserPrivateDatas"] = true,
		["soundPlay"] = true,
		["soundStop"] = true,
		["clearDisplay"] = true,
		["setVoice"] = true,
		["getVoice"] = true,
		["display"] = true,
		["markAdd"] = true,
		["markRemove"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
	},
	
	add = function(self, _level, _channel, _country, _moduleName)
		local localMembersGroup
		local coalition
		local countryId = _country
		
		if type(_channel) == "string" then
			coalition = _channel
		elseif _channel ~= nil then
			localMembersGroup = _channel
			countryId = country.id[localMembersGroup.country]
			coalition = ATMEConst.coalitionNames[localMembersGroup.objectDCS:getCoalition()]
		end
		
		local item = {
			level = _level,
			localMembersGroup = localMembersGroup,
			coalition = coalition,
			country = countryId,
			currentVoice = "",
			clear = false,
			userPrivateDatas = {},
			moduleName = _moduleName,
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
	end,
}

ATME.C_DataSharing = {
	getClassName = function(self)
		return ATMEDatas.C_DataSharing_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_DataSharing_MT.parentClass
	end,
	
	createNamed = function(_name, _channel)
		local fctDefinition = "Function ATME.C_DataSharing.createNamed(name, channel)"
		
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" and (type(_channel) == "string" or ATMEFct.isObjectClass(_channel, ATME.C_Group) == true) then
			local level
			local channel
			local countryId

			if type(_channel) == "string" then
				channel = string.upper(_channel)
				
				if ATMEFct.C_DataSharing_setChannel[channel] ~= nil then
					level, channel = ATMEFct.C_DataSharing_setChannel[channel]()
				elseif country.id[channel] ~= nil then
					countryId = country.id[channel]
					channel = ATMEConst.coalitionNames[coalition.getCountryCoalition(countryId)]
					level = "countries"
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "Channel " .. _channel .. " doesn't exist", false)
					return
				end
			else
				level = "groups"
				channel = ATMEDatas.localMembers[_channel:getIndex()]
			end
			
			local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
			local err, newDataSharing = ATMEFct.C_DataSharing_createNamed(localMembersModule.dataSharings, _name, level, channel, countryId, ATMEDatas.currentModuleName)

			if err == true then
				ATMEFct.warningOutputCore(nil, fctDefinition, newDataSharing, false)
				return
			end

			return newDataSharing
		else
			if type(_name) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
			elseif type(_channel) ~= "string" and ATMEFct.isObjectClass(_channel, ATME.C_Group) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "channel type must be string or ATME.C_Group", false)
			end
		end
	end,
	
	getByName = function(_name)
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, "Function ATME.C_DataSharing.getByName(name)", "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].dataSharings[_name]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_DataSharing.getByName(name)", "name type must be string", false)
		end
	end,
}

ATMEFct.C_DataSharing_create = function(_name, _dataSharings, _level, _channel, _country, _moduleName)
	local newDataSharing = {}
	local name
	local tagsName

	setmetatable(newDataSharing, ATMEDatas.C_DataSharing_MT.mt)
	local index = ATMEDatas.C_DataSharing_MT:add(_level, _channel, _country, _moduleName)
	local localMembers = ATMEDatas.localMembers[index]
	
	if _name ~= nil and _dataSharings ~= nil then
		name = _name
		_dataSharings[name] = newDataSharing
		tagsName = ATMEFct.decompString(_name)
	end
	
	function newDataSharing.getIndex(self)
		return index
	end
	
	function newDataSharing.getName(self)
		return name
	end
	
	function newDataSharing.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newDataSharing) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newDataSharing.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newDataSharing.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newDataSharing) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newDataSharing.clearDisplay(self, _active)
		local fctDefinition = ATMEFct.getErrorPrefix(newDataSharing) .. "clearDisplay(active)"
		if type(_active) == "boolean" then
			localMembers.clear = _active
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "active type must be boolean", true)
		end
	end
	
	function newDataSharing.setVoice(self, _name)
		local fctDefinition = ATMEFct.getErrorPrefix(newDataSharing) .. "setVoice([name])"
		
		if _name == nil or type(_name) == "string" then
			if _name ~= nil then
				local d, f, temp = _name:find("^(%a[%a%d_]*)")
				
				if temp == _name then
					localMembers.currentVoice = _name
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "name bad format", false)
				end
			else
				localMembers.currentVoice = ""
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be nil or string", true)
		end
	end
	
	function newDataSharing.getVoice(self)
		if localMembers.currentVoice ~= "" then
			return localMembers.currentVoice
		end
	end
	
	function newDataSharing.soundPlay(self, _text, _readMode)
		local fctDefinition = ATMEFct.getErrorPrefix(newDataSharing) .. "soundPlay(text, [readMode])"
		
		if type(_text) == "string" and (_readMode == nil or type(_readMode) == "boolean") then
			if localMembers.soundDatas == nil then
				localMembers.soundDatas = {
					localMembersGroup = localMembers.localMembersGroup,
					localMembers = localMembers,
				}

				if ATMEFct.C_DataSharing_testSoundsOn[localMembers.level](localMembers.soundDatas) == true then
					local randomRead = false
					local readOnce = true
					
					if _readMode ~= nil then
						randomRead = _readMode
						readOnce = false
					end

					local err, playlist = ATMEFct.C_Sound_getPlaylist(_text, randomRead, nil, localMembers.moduleName)

					if err == true then
						ATMEFct.warningOutputCore(nil, fctDefinition, playlist, false)
						return false
					end

					localMembers.soundDatas.randomRead = randomRead
					localMembers.soundDatas.readOnce = readOnce
					localMembers.soundDatas.playlist = playlist
	
					ATMEFct.C_DataSharing_initPlaylist(localMembers.soundDatas)
					
					return true
				else
					ATMEFct.setError("ALREADY_PLAYING_SOUND")
					ATMEFct.warningOutputCore(nil, fctDefinition, "already playing sound", false)
					localMembers.soundDatas = nil
					
					return false
				end
			else
				ATMEFct.setError("ALREADY_PLAYING_SOUND")
				ATMEFct.warningOutputCore(nil, fctDefinition, "already playing sound", false)

				return false
			end
		else
			if type(_text) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "text type must be string", true)
			elseif _readMode ~= nil and type(_readMode) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "readMode type must be nil or boolean", true)
			end
		end
	end
	
	function newDataSharing.soundStop(self)
		if localMembers.idTimer ~= nil then
			ATMEFct.C_DataSharing_soundsOnOff[localMembers.level](nil, localMembers.soundDatas)
			localMembers.soundDatas = nil
			
			timer.removeFunction(localMembers.idTimer)
			localMembers.idTimer = nil
		end
	end
	
	function newDataSharing.display(self, _text, _duration)
		local fctDefinition = ATMEFct.getErrorPrefix(newDataSharing) .. "display(text, duration)"
		
		if type(_text) == "string" and type(_duration) == "number" then
			local duration = math.ceil(_duration)
			
			if duration > 0 then
				ATMEFct.C_DataSharing_display[localMembers.level]({
					localMembersGroup = localMembers.localMembersGroup,
					localMembers = localMembers,
					duration = duration,
					text = _text,
				})
				
				return true
			else
				ATMEFct.setError("BAD_DURATION")
				ATMEFct.warningOutputCore(nil, fctDefinition, "duration must be > 0", false)
	
				return false
			end
		else
			if type(_text) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "text type must be string", true)
			elseif type(_duration) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "duration type must be number", false)
			end
		end
	end
	
	function newDataSharing.markAdd(self, _text, _reference, _readOnly, _message)
		local fctDefinition = ATMEFct.getErrorPrefix(newDataSharing) .. "markAdd(text, reference, readOnly, [message])"
		
		if type(_text) == "string" and ATMEFct.isReference(_reference) == true and type(_readOnly) == "boolean" and
						(_message == nil or type(_message) == "string") then
			local point = ATMEFct.getPointFromReference(_reference)
			
			ATMEDatas.markId = ATMEDatas.markId + 1
			
			local ok, err =  ATMEFct.C_DataSharing_mark[localMembers.level]({
										localMembersGroup = localMembers.localMembersGroup,
										localMembers = localMembers,
										point = point,
										text = _text,
										readOnly = _readOnly,
										message = _message,
										id = ATMEDatas.markId,
			})
			
			if ok == true then
				return ATMEDatas.markId
			else
				ATMEFct.warningOutputCore(nil, fctDefinition, err, false)
			end
		else
			if type(_text) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "text type must be string", true)
			elseif ATMEFct.isReference(_reference) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
			elseif type(_readOnly) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "readOnly type must be boolean", false)
			elseif _message ~= nil and type(_message) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "message type must be nil or string", false)
			end
		end
	end
	
	function newDataSharing.markRemove(self, _id)
		local fctDefinition = ATMEFct.getErrorPrefix(newDataSharing) .. "markRemove(id)"
		
		if type(_id) == "number" then
			local id = math.floor(_id)
			
			trigger.action.removeMark(id)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "id type must be number", true)
		end
	end	
	
	function newDataSharing.getUserPrivateDatas(self, _moduleName)
		local fctDefinition = ATMEFct.getErrorPrefix(newDataSharing) .. "getUserPrivateDatas([moduleName])"
		local moduleName
		
		if _moduleName ~= nil then
			if type(_moduleName) == "string" then
				if localMembers.userPrivateDatas[_moduleName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
					return
				else
					moduleName = _moduleName
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
				return
			end
		else
			moduleName = ATMEDatas.currentModuleName
		end
		
		return localMembers.userPrivateDatas[moduleName]
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end

	return false, newDataSharing
end

ATMEDatas.C_DataSharing_MT.__call = function(t, _channel)
	local fctDefinition = "Constructor :\nATME.C_DataSharing(channel)"
	
	if ATMEDatas.startMissionOk < 2 then
		ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
		return
	end

	if type(_channel) == "string" or ATMEFct.isObjectClass(_channel, ATME.C_Group) == true then
		local level
		local channel	-- Groupe ATME ou coalition BLUE/RED
		local country
		
		if type(_channel) == "string" then
			channel = string.upper(_channel)
			
			if ATMEFct.C_DataSharing_setChannel[channel] ~= nil then
				level, channel = ATMEFct.C_DataSharing_setChannel[channel]()
			elseif country.id[channel] ~= nil then
				countryId = country.id[channel]
				channel = ATMEConst.coalitionNames[coalition.getCountryCoalition(countryId)]
				level = "countries"
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Channel " .. _channel .. " doesn't exist", false)
				return
			end
		else
			level = "groups"
			channel = ATMEDatas.localMembers[_channel:getIndex()]
		end

		local err, newDataSharing = ATMEFct.C_DataSharing_create(nil, nil, level, channel, country, ATMEDatas.currentModuleName)

		if err == true then
			ATMEFct.warningOutputCore(nil, fctDefinition, newDataSharing, false)
			return
		end

		return newDataSharing
	else
		ATMEFct.errorOutputCore(nil, fctDefinition, "channel type must be string or ATME.C_Group", false)
	end
end

setmetatable(ATME.C_DataSharing,ATMEDatas.C_DataSharing_MT)

ATMEDatas.C_DataSharing_MT.mt.__metatable = "Private"
ATMEDatas.C_DataSharing_MT.mt.__index = ATME.C_DataSharing

ATMEDatas.C_DataSharing_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_DataSharing_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_DataSharing : Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_RadioChannel : gère les fréquences radios pour les transmissions.
----------------------------------------------------------------------------------------------------------

ATMEFct.C_RadioChannel_getIdFrequency = function(_modulation, _frequency)
	return "FR_" .. string.format("%09d", _frequency) .. _modulation
end

ATMETimers.C_RadioChannel_transmissionTimer = function(_idFrequency)
	if ATMEDatas.exitATME == false then
		local currentTransmission = ATMEDatas.radioFrequencies[_idFrequency].current
		local toRemove = false

		if currentTransmission.randomRead == true then 
			currentTransmission.currentReadIndex = math.random(1, #currentTransmission.playlist)
		end
		
		if currentTransmission.unitName == nil then
			ATMEFct.C_RadioChannel_stopTransmission(currentTransmission)
			ATMEFct.C_RadioChannel_startTransmission(currentTransmission)
				
			currentTransmission.tipsBase = currentTransmission.playlist[currentTransmission.currentReadIndex].tipsBase
			
			if currentTransmission.tipsBase < 1.5 then
				currentTransmission.tipsBase = 1.5
			end
		else
			if ATMEDatas.units[currentTransmission.unitName] ~= nil then
				ATMEFct.C_Unit_stopTransmission(currentTransmission)
				ATMEFct.C_Unit_startTransmission(currentTransmission)
				currentTransmission.tipsBase = currentTransmission.playlist[currentTransmission.currentReadIndex].tipsBase
			else
				toRemove = true
			end
		end
		
		if currentTransmission.readOnce == true then
			currentTransmission.currentReadIndex = currentTransmission.currentReadIndex + 1
			if currentTransmission.currentReadIndex > #currentTransmission.playlist then
				toRemove = true
			end
		else
			if currentTransmission.randomRead == false then
				currentTransmission.currentReadIndex = currentTransmission.currentReadIndex + 1
				if currentTransmission.currentReadIndex > #currentTransmission.playlist then
					currentTransmission.currentReadIndex = 1
				end
			end
		end
		
		if toRemove == true then
			local stackLocalMembers = ATMEDatas.localMembers[ATMEDatas.radioFrequencies[_idFrequency].stack:getIndex()]
			ATMEDatas.radioFrequencies[_idFrequency].current = ATMEFct.C_Stack_extract(stackLocalMembers, "TAIL")
				
			if ATMEDatas.radioFrequencies[_idFrequency].current ~= nil then 
				return timer.getTime() + math.random(1,5)
			end

			ATMEDatas.radioFrequencies[_idFrequency].timerId = nil
		else
			-- On réarme pour la durée du son
			return timer.getTime() + currentTransmission.tipsBase
		end
	end
end

ATMEFct.C_RadioChannel_initTransmission = function(_extraDatas)
	-- C_RadioChannel_initTransmission : fonction d'initialisation des paramètres pour les transmissions radio
	-- 						             pour une mise en forme au format DCS

	local newRadioTransmission = {
		playlist = _extraDatas.playlist,
		unitName = _extraDatas.unitName,
		tipsBase = 0,
		readOnce = _extraDatas.readOnce,
		randomRead = _extraDatas.randomRead,
		replay = true,
		currentReadIndex = 1,
		started = false,
		localMembers = _extraDatas.localMembers,
	}

	if _extraDatas.unitName == nil then 			-- Pour la transmission à partir d'un point
		newRadioTransmission.point = _extraDatas.point
		newRadioTransmission.power = _extraDatas.power
		newRadioTransmission.textId = "Z" .. ATMEDatas.idDCSRadioTransmissions
		ATMEDatas.idDCSRadioTransmissions = ATMEDatas.idDCSRadioTransmissions + 1
	end

	if ATMEDatas.radioFrequencies[_extraDatas.localMembers.idFrequency].timerId == nil then
		ATMEDatas.radioFrequencies[_extraDatas.localMembers.idFrequency].current = newRadioTransmission
		ATMEDatas.radioFrequencies[_extraDatas.localMembers.idFrequency].timerId = timer.scheduleFunction(ATMETimers.C_RadioChannel_transmissionTimer, _extraDatas.localMembers.idFrequency, timer.getTime() + 1)
	else
		if _extraDatas.localMembers.highPriority == false then
			ATMEDatas.radioFrequencies[_extraDatas.localMembers.idFrequency].stack:push(newRadioTransmission)
		else
			if ATMEDatas.radioFrequencies[_extraDatas.localMembers.idFrequency].timerId ~= nil then
				timer.removeFunction(ATMEDatas.radioFrequencies[_extraDatas.localMembers.idFrequency].timerId)
				if ATMEDatas.radioFrequencies[_extraDatas.localMembers.idFrequency].current.unitName ~= nil then	
					ATMEFct.C_Unit_stopTransmission(ATMEDatas.radioFrequencies[_extraDatas.localMembers.idFrequency].current)
				else
					ATMEFct.C_RadioChannel_stopTransmission(ATMEDatas.radioFrequencies[_extraDatas.localMembers.idFrequency].current)
				end
				
				ATMEDatas.radioFrequencies[_extraDatas.localMembers.idFrequency].current = newRadioTransmission
				ATMEDatas.radioFrequencies[_extraDatas.localMembers.idFrequency].timerId = timer.scheduleFunction(ATMETimers.C_RadioChannel_transmissionTimer, _extraDatas.localMembers.idFrequency, timer.getTime() + 1)
			end
		end
	end
	
	return newRadioTransmission
end

ATMEFct.C_RadioChannel_startTransmission = function(_radioTransmission)
	-- ATTENTION Bug DCS dans la transmission à partir d'un point si le point est à la position exacte du joueur qui écoute.
	-- Dans ce cas, soit une partie de la com est bouffée soit toute la com, avec quelques rares exceptions ou ca fonctionne.
	
	if _radioTransmission.started == false and _radioTransmission.playlist[_radioTransmission.currentReadIndex].pause == false then
		trigger.action.radioTransmission("l10n/DEFAULT/" .. _radioTransmission.localMembers.currentVoice .. _radioTransmission.playlist[_radioTransmission.currentReadIndex].fileName,
											_radioTransmission.point , ATMEConst.modulationValues[_radioTransmission.localMembers.modulation],
											false, _radioTransmission.localMembers.frequency,
											_radioTransmission.power, _radioTransmission.textId)

		_radioTransmission.started = true
	end
end

ATMEFct.C_RadioChannel_stopTransmission = function(_radioTransmission)
	if _radioTransmission.started == true then
		trigger.action.stopRadioTransmission(_radioTransmission.textId)
		_radioTransmission.started = false
	end
end

ATMEFct.C_RadioChannel_createNamed = function(_namedTable, _name, _modulation, _frequency, _highPriority)
	if _namedTable[_name] ~= nil then
		ATMEFct.setError("ALREADY_CREATED")
		return true, "named Radio channel already exists"
	else
		local d,f = string.find(_name, "|", 1)
		if d == nil then
			return ATMEFct.C_RadioChannel_create(_name, _namedTable, _modulation, _frequency, _highPriority)
		else
			ATMEFct.setError("BAD_FORMAT")
			return true, "name can't contain '|' character"
		end
	end
end

ATMEDatas.C_RadioChannel_MT = {
	className = "C_RadioChannel",
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["getUserPrivateDatas"] = true,
		["setVoice"] = true,
		["getVoice"] = true,
		["sendFromUnit"] = true,
		["sendFromReference"] = true,
		["stop"] = true,
		["stopAll"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
	},
	
	add = function(self, _modulation, _frequency, _highPriority)
		local item = {
			frequency = _frequency,
			modulation = _modulation,
			idFrequency = ATMEFct.C_RadioChannel_getIdFrequency(_modulation, _frequency),
			userPrivateDatas = {},
			currentVoice = "",
			highPriority = _highPriority,
			moduleName = ATMEDatas.currentModuleName,
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
	end,
}

ATME.C_RadioChannel = {
	getClassName = function(self)
		return ATMEDatas.C_RadioChannel_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_RadioChannel_MT.parentClass
	end,
	
	createNamed = function(_name, _modulation, _frequency, _highPriority)
		local fctDefinition = "Function ATME.C_RadioChannel.createNamed(name, modulation, frequency, [highPriority])"
		
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" and type(_modulation) == "string" and type(_frequency) == "number" and (_highPriority == nil or type(_highPriority) == "boolean") then
			if ATMEConst.modulationValues[_modulation] ~= nil then
				local frequencyValue = math.floor(_frequency)
				if frequencyValue > 0 then
					local highPriority = false
					
					if _highPriority ~= nil then
						highPriority = _highPriority
					end
					local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
					local err, newRadioChannel = ATMEFct.C_RadioChannel_createNamed(localMembersModule.radioChannels, _name, _modulation, frequencyValue, highPriority)

					if err == true then
						ATMEFct.warningOutputCore(nil, fctDefinition, newRadioChannel, false)
						return
					end

					return newRadioChannel
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "frequency must be > 0", false)
				end	
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "modulation must be 'AM' or 'FM'", false)
			end
		else
			if type(_name) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
			elseif type(_modulation) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "modulation type must be string", false)
			elseif type(_frequency) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "frequency type must be number", false)
			elseif _highPriority ~= nil and type(_highPriority) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "highPriority type must be nil or boolean", false)
			end
		end
	end,
	
	getByName = function(_name)
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, "Function ATME.C_RadioChannel.getByName(name)", "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].radioChannels[_name]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_RadioChannel.getByName(name)", "name type must be string", false)
		end
	end,
}

ATMEFct.C_RadioChannel_create = function(_name, _radioChannels, _modulation, _frequency, _highPriority)
	local newRadioChannel = {}
	local name
	local tagsName

	setmetatable(newRadioChannel, ATMEDatas.C_RadioChannel_MT.mt)
	
	local index = ATMEDatas.C_RadioChannel_MT:add(_modulation, _frequency, _highPriority)
	local localMembers = ATMEDatas.localMembers[index]
	
	if _name ~= nil and _radioChannels ~= nil then
		name = _name
		_radioChannels[name] = newRadioChannel
		tagsName = ATMEFct.decompString(_name)
	end
	
	if ATMEDatas.radioFrequencies[localMembers.idFrequency] == nil then
		ATMEDatas.radioFrequencies[localMembers.idFrequency] = {
			stack = ATMEFct.C_Stack_create(),
		}
	end
	
	function newRadioChannel.getIndex(self)
		return index
	end
	
	function newRadioChannel.getName(self)
		return name
	end
	
	function newRadioChannel.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newRadioChannel) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newRadioChannel.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newRadioChannel.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newRadioChannel) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newRadioChannel.setVoice(self, _name)
		local fctDefinition = ATMEFct.getErrorPrefix(newRadioChannel) .. "setVoice([name])"
		
		if _name == nil or type(_name) == "string" then
			if _name ~= nil then
				local d, f, temp = _name:find("^(%a[%a%d_]*)")
				if temp == _name then
					localMembers.currentVoice = _name
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "name bad format", false)
				end
			else
				localMembers.currentVoice = ""
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be nil or string", true)
		end
	end
	
	function newRadioChannel.getVoice(self)
		if localMembers.currentVoice ~= "" then
			return localMembers.currentVoice
		end
	end
	
	function newRadioChannel.sendFromUnit(self, _text, _unit, _readMode)
		local fctDefinition = ATMEFct.getErrorPrefix(newRadioChannel) .. "sendFromUnit(text, unit, [readMode])"
		
		if type(_text) == "string" and ATMEFct.isObjectClass(_unit, ATME.C_Unit) == true and 
				(_readMode == nil or type(_readMode) == "boolean") then
			local randomRead = false
			local readOnce = true
			
			if type(_readMode) == "boolean" then
				randomRead = _readMode
				readOnce = false
			end
			
			local err, playlistNew = ATMEFct.C_Sound_getPlaylist(_text, randomRead, nil, localMembers.moduleName)
			
			if err == true then
				ATMEFct.warningOutputCore(nil, fctDefinition, playlistNew, false)
				return
			end
			
			local localMembersUnit = ATMEDatas.localMembers[_unit:getIndex()]
			local voice = localMembersUnit.currentVoice
			if voice == "" then
				voice = localMembers.currentVoice
			end
			
			ATMEFct.C_RadioChannel_initTransmission({
				unitName = _unit:getName(),
				playlist = playlistNew,
				readOnce = readOnce,
				randomRead = randomRead,
				localMembers = localMembers,
			})
		else
			if type(_text) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "text type must be string", true)
			elseif ATMEFct.isObjectClass(_unit, ATME.C_Unit) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "unit type must be ATME.C_Unit", false)
			elseif _readMode ~= nil and type(_readMode) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "readMode type must be nil or boolean", false)
			end
		end
	end
	
	function newRadioChannel.sendFromReference(self, _text, _reference, _power, _readMode)
		local fctDefinition = ATMEFct.getErrorPrefix(newRadioChannel) .. "sendFromReference(text, reference, power, readMode)"
		
		if type(_text) == "string" and ATMEFct.isReference(_reference) == true and type(_power) == "number" and 
												(_readMode == nil or type(_readMode) == "boolean") then
			local powerValue = math.floor(_power)
			
			if powerValue > 0 then
				local randomRead = false
				local readOnce = true
				
				if type(_readMode) == "boolean" then
					randomRead = _readMode
					readOnce = false
				end
			
				local err, playlistNew = ATMEFct.C_Sound_getPlaylist(_text, false, nil, localMembers.moduleName)

				if err == true then
					ATMEFct.warningOutputCore(nil, fctDefinition, playlistNew, false)
					return
				end

				ATMEFct.C_RadioChannel_initTransmission({
					playlist = playlistNew,
					readOnce = readOnce,
					randomRead = randomRead,
					point = ATMEFct.getPointFromReference(_reference),
					power = powerValue,
					localMembers = localMembers,
				})
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "power must be > 0", false)
			end
		else
			if type(_text) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "text type must be string", true)
			elseif ATMEFct.isReference(_reference) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
			elseif type(_power) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "power type must be number", false)
			elseif _readMode ~= nil and type(_readMode) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "readMode type must be nil or boolean", false)
			end
		end
	end

	function newRadioChannel.stop(self)
		if ATMEDatas.radioFrequencies[localMembers.idFrequency].timerId ~= nil then
			timer.removeFunction(ATMEDatas.radioFrequencies[localMembers.idFrequency].timerId)
			if ATMEDatas.radioFrequencies[localMembers.idFrequency].current.unitName ~= nil then	
				ATMEFct.C_Unit_stopTransmission(ATMEDatas.radioFrequencies[localMembers.idFrequency].current)
			else
				ATMEFct.C_RadioChannel_stopTransmission(ATMEDatas.radioFrequencies[localMembers.idFrequency].current)
			end
			
			local stackLocalMembers = ATMEDatas.localMembers[ATMEDatas.radioFrequencies[localMembers.idFrequency].stack:getIndex()]
			ATMEDatas.radioFrequencies[localMembers.idFrequency].current = ATMEFct.C_Stack_extract(stackLocalMembers, "TAIL")
				
			if ATMEDatas.radioFrequencies[localMembers.idFrequency].current ~= nil then 
				ATMEDatas.radioFrequencies[localMembers.idFrequency].timerId = timer.scheduleFunction(ATMETimers.C_RadioChannel_transmissionTimer, localMembers.idFrequency, timer.getTime() + math.random(1,5))
			else
				ATMEDatas.radioFrequencies[localMembers.idFrequency].timerId = nil
			end
		end
	end
	
	function newRadioChannel.stopAll(self)
		if ATMEDatas.radioFrequencies[localMembers.idFrequency].timerId ~= nil then
			timer.removeFunction(ATMEDatas.radioFrequencies[localMembers.idFrequency].timerId)
			if ATMEDatas.radioFrequencies[localMembers.idFrequency].current.unitName ~= nil then	
				ATMEFct.C_Unit_stopTransmission(ATMEDatas.radioFrequencies[localMembers.idFrequency].current)
			else
				ATMEFct.C_RadioChannel_stopTransmission(ATMEDatas.radioFrequencies[localMembers.idFrequency].current)
			end
			
			ATMEDatas.radioFrequencies[localMembers.idFrequency].timerId = nil
		end
		
		ATMEDatas.radioFrequencies[localMembers.idFrequency].stack:removeAll()
	end
	
	function newRadioChannel.getUserPrivateDatas(self, _moduleName)
		local fctDefinition = ATMEFct.getErrorPrefix(newRadioChannel) .. "getUserPrivateDatas([moduleName])"
		local moduleName
		
		if _moduleName ~= nil then
			if type(_moduleName) == "string" then
				if localMembers.userPrivateDatas[_moduleName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
					return
				else
					moduleName = _moduleName
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
				return
			end
		else
			moduleName = ATMEDatas.currentModuleName
		end
		
		return localMembers.userPrivateDatas[moduleName]
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return false, newRadioChannel
end

ATMEDatas.C_RadioChannel_MT.__call = function(t, _modulation, _frequency, _highPriority)
	local fctDefinition = "Constructor :\nATME.C_RadioChannel(modulation, frequency, [highPriority])"
	
	if ATMEDatas.startMissionOk < 2 then
		ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
		return
	end
	
	if type(_modulation) == "string" and type(_frequency) == "number" and (_highPriority == nil or type(_highPriority) == "boolean") then
		if ATMEConst.modulationValues[_modulation] ~= nil then
			local frequencyValue = math.floor(_frequency)
			if frequencyValue > 0 then
				local highPriority = false
					
				if _highPriority ~= nil then
					highPriority = _highPriority
				end
				
				local err, newRadioChannel =  ATMEFct.C_RadioChannel_create(nil, nil, _modulation, frequencyValue, highPriority)
				
				if err == true then
					ATMEFct.errorOutputCore(nil, fctDefinition, newRadioChannel, false)
					return
				end
					
				return newRadioChannel
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "frequency must be > 0", false)
			end	
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "modulation must be 'AM' or 'FM'", false)
		end
	else
		if type(_modulation) ~= "string" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "modulation type must be string", false)
		elseif type(_frequency) ~= "number" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "frequency type must be number", false)
		elseif _highPriority ~= nil and type(_highPriority) ~= "boolean" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "highPriority type must be nil or boolean", false)
		end
	end
end

setmetatable(ATME.C_RadioChannel,ATMEDatas.C_RadioChannel_MT)

ATMEDatas.C_RadioChannel_MT.mt.__metatable = "Private"
ATMEDatas.C_RadioChannel_MT.mt.__index = ATME.C_RadioChannel

ATMEDatas.C_RadioChannel_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_RadioChannel_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_RadioChannel : Bad key " .. k, "table locked", false)
	end
end


-- Fonctions génériques vectors

ATMEFct.vectors = {
	["2D"] = {
		createFromTable = function(_name, _namedTable, _table)
			return ATMEFct.C_Vector2D_create(_name, _namedTable, _table.x, _table.z)
		end,
		createFromVector3D = function(_name, _namedTable, _vector)
			return ATMEFct.C_Vector2D_create(_name, _namedTable, _vector:getX(), _vector:getZ())
		end,
		createFromVector2D = function(_name, _namedTable, _vector)
			return ATMEFct.C_Vector2D_create(_name, _namedTable, _vector:getX(), _vector:getZ())
		end,
		createFromReference = function(_name, _namedTable, _refA, _refB)
			local p1 = ATMEFct.getPointFromReference(_refA)
			local p2 = ATMEFct.getPointFromReference(_refB)

			return ATMEFct.C_Vector2D_create(_name, _namedTable, p2.x - p1.x, p2.z - p1.z)
		end,
	},
	["3D"] = {
		createFromTable = function(_name, _namedTable, _table)
			return ATMEFct.C_Vector3D_create(_name, _namedTable, _table.x, _table.y, _table.z)
		end,
		createFromVector3D = function(_name, _namedTable, _vector)
			return ATMEFct.C_Vector3D_create(_name, _namedTable, _vector:getX(), _vector:getY(), _vector:getZ())
		end,
		createFromVector2D = function(_name, _namedTable, _vector)
			return ATMEFct.C_Vector3D_create(_name, _namedTable, _vector:getX(), 0, _vector:getZ())
		end,
		createFromReference = function(_name, _namedTable, _refA, _refB)
			local p1 = ATMEFct.getPointFromReference(_refA)
			local p2 = ATMEFct.getPointFromReference(_refB)

			return ATMEFct.C_Vector3D_create(_name, _namedTable, p2.x - p1.x, p2.y - p1.y, p2.z - p1.z)
		end,
	},
}

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Vector3D : gère les vecteurs 3D. Constructeur de classe accessible aux modules externes.
-- Elle donne accès aux calculs de module (module du vecteur AB = distance AB), à l'angle formé avec le nord
-- par AB (Azimuth), à la transformation en vecteur unitaire (vecteur identique de module 1) et au produit
-- scalaire calculé pour les coordonnées horizontales (x et z).
----------------------------------------------------------------------------------------------------------

ATMEFct.C_Vector3D_createNamed = function(_namedTable, _name, ...)
	if _namedTable[_name] ~= nil then
		ATMEFct.setError("ALREADY_CREATED")
		return 2, "named vector3D already exists"
	else
		local d,f = string.find(_name, "|", 1)
		if d == nil then
			local args = {...}
			
			if #args == 0 then
				return 0, ATMEFct.C_Vector3D_create(_name, _namedTable, 0, 0, 0)
			elseif #args == 1 then
				if ATMEFct.isObjectClass(args[1], ATME.C_Vector3D) == true then  
					return 0, ATMEFct.vectors["3D"].createFromVector3D(_name, _namedTable, args[1])
				elseif ATMEFct.isObjectClass(args[1], ATME.C_Vector2D) == true then
					return 0, ATMEFct.vectors["3D"].createFromVector2D(_name, _namedTable, args[1])
				elseif type(args[1]) == "table" and type(args[1].x) == "number" and type(args[1].y) == "number" and type(args[1].z) == "number" then
					return 0, ATMEFct.C_Vector3D_create(_name, _namedTable, args[1].x, args[1].y, args[1].z)
				end
			elseif #args == 2 and ATMEFct.isReference(args[1]) == true and ATMEFct.isReference(args[2]) == true then
				return 0, ATMEFct.vectors["3D"].createFromReference(_name, _namedTable, args[1], args[2])
			elseif #args == 3 and type(args[1]) == "number" and type(args[2]) == "number" and type(args[3]) == "number" then
				return 0, ATMEFct.C_Vector3D_create(_name, _namedTable, args[1], args[2], args[3])
			else
				if #args == 3 then
					if type(args[1]) ~= "number" then
						return 1, "x type must be number"
					elseif type(args[2]) ~= "number" then
						return 1, "y type must be number"
					elseif type(args[3]) ~= "number" then
						return 1, "z type must be number"
					end
				elseif #args == 2 then
					if ATMEFct.isReference(args[1]) == false then
						return 1, "refA type must be string," .. ATMEConst.errorLabelReference
					elseif ATMEFct.isReference(args[2]) == false then
						return 1, "refB type must be string," .. ATMEConst.errorLabelReference
					end
				elseif #args == 1 and ATMEFct.isObjectClass(args[1], ATME.C_Vector3D) == false and ATMEFct.isObjectClass(args[1], ATME.C_Vector2D) == false and
					(type(args[1]) ~= "table" or (type(args[1]) == "table" and (type(args[1].x) ~= "number" or type(args[1].y) ~= "number" or type(args[1].z) ~= "number"))) then
					return 1, "vector type must be (x, y, z), ATME.C_Vector2D or ATME.C_Vector3D"
				else
					return 1, "must have 0 to 3 parameters"
				end
			end
		else
			ATMEFct.setError("BAD_FORMAT")
			return 2, "name can't contain '|' character"
		end
	end
end

ATMEDatas.C_Vector3D_MT = {
	className = "C_Vector3D",
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["getUserPrivateDatas"] = true,
		["getAzimuth"] = true, 
		["getHModule"] = true,
		["getModule"] = true,
		["get"] = true, 
		["getX"] = true,
		["getY"] = true,
		["getZ"] = true,
		["toUnitVector"] = true,
		["toVector2D"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
	},
	
	add = function(self)
		local item = {
			userPrivateDatas = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
	end,
}

ATME.C_Vector3D = {
	getClassName = function(self)
		return ATMEDatas.C_Vector3D_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Vector3D_MT.parentClass
	end,
	
	createNamed = function(_name, ...)
		local fctDefinition = "Function ATME.C_Vector3D.createNamed(name)\nATME.C_Vector3D.createNamed(name, vector)\nATME.C_Vector3D.createNamed(name, refA, refB)\nATME.C_Vector3D.createNamed(name, x, y, z)"
		
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
			local err, newVector3D = ATMEFct.C_Vector3D_createNamed(localMembersModule.vectors3D, _name, ...)				
			
			if err > 0 then
				if err == 1 then
					ATMEFct.errorOutputCore(nil, fctDefinition, newVector3D, false)
				else
					ATMEFct.warningOutputCore(nil, fctDefinition, newVector3D, false)
				end
				
				return
			end
			
			return newVector3D
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
		end
	end,
	
	getByName = function(_name)
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Vector3D.getByName(name)", "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].vectors3D[_name]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Vector3D.getByName(name)", "name type must be string", false)
		end
	end,
}

ATMEFct.C_Vector3D_create = function(_name, _vectors3D, _x, _y, _z)
	local x = _x
	local y = _y
	local z = _z
	
	local newVector = {}
	local localMembers
	local index
	
	local name
	local tagsName
	
	setmetatable(newVector, ATMEDatas.C_Vector3D_MT.mt)
	index = ATMEDatas.C_Vector3D_MT:add()
	localMembers = ATMEDatas.localMembers[index]
		
	if _name ~= nil and _vectors3D ~= nil then
		name = _name
		_vectors3D[name] = newVector
		tagsName = ATMEFct.decompString(_name)
	end
	
	local hModule = math.sqrt(x^2 + z^2)
	local gModule = math.sqrt(x^2 + y^2 + z^2)

	local angle = 0
	if x ~= 0 and z ~= 0 then
		angle = math.deg(math.atan2(z, x))
	end
	
	
	if angle < 0 then
		angle = angle + 360
	end
	
	function newVector.getIndex(self)
		return index
	end
	
	function newVector.getName(self)
		return name
	end
	
	function newVector.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newVector) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newVector.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newVector.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newVector) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newVector.getAzimuth(self)
		return angle
	end
	
	function newVector.getHModule(self)
		return hModule
	end
	
	function newVector.getModule(self)
		return gModule
	end
	
	function newVector.get(self)
		return x, y, z
	end
	
	function newVector.getX(self)
		return x
	end

	function newVector.getY(self)
		return y
	end

	function newVector.getZ(self)
		return z
	end
	
	function newVector.toUnitVector(self)
		local unitVector
		if gModule ~= 0 then
			unitVector = ATMEFct.C_Vector3D_create(nil, nil, x/gModule, y/gModule, z/gModule)
		else
			unitVector = ATMEFct.C_Vector3D_create(nil, nil, 0, 0, 0)
		end
		
		return unitVector
	end
	
	function newVector.toVector2D(self)
		return ATMEFct.C_Vector2D_create(nil, nil, x, z)
	end
	
	function newVector.getUserPrivateDatas(self, _moduleName)
		local fctDefinition = ATMEFct.getErrorPrefix(newVector) .. "getUserPrivateDatas([moduleName])"
		local moduleName
		
		if _moduleName ~= nil then
			if type(_moduleName) == "string" then
				if localMembers.userPrivateDatas[_moduleName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
					return
				else
					moduleName = _moduleName
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
				return
			end
		else
			moduleName = ATMEDatas.currentModuleName
		end
		
		return localMembers.userPrivateDatas[moduleName]
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return newVector
end

ATMEDatas.C_Vector3D_MT.__call = function(t, ...)
	local args = {...}
	
	local fctDefinition = "Constructor :\nATME.C_Vector3D()\nATME.C_Vector3D(vector)\nATME.C_Vector3D(refA, refB)\nATME.C_Vector3D(x, y, z)"
	
	if ATMEDatas.startMissionOk < 2 then
		ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
		return
	end
	
	local x = 0
	local y = 0
	local z = 0
	
	if #args == 0 then
		return ATMEFct.C_Vector3D_create(nil, nil, 0, 0, 0)
	elseif #args == 1 then
		if ATMEFct.isObjectClass(args[1], ATME.C_Vector3D) == true then  
			return ATMEFct.vectors["3D"].createFromVector3D(nil, nil, args[1])
		elseif ATMEFct.isObjectClass(args[1], ATME.C_Vector2D) == true then
			return ATMEFct.vectors["3D"].createFromVector2D(nil, nil, args[1])
		elseif type(args[1]) == "table" and type(args[1].x) == "number" and type(args[1].y) == "number" and type(args[1].z) == "number" then
			return ATMEFct.C_Vector3D_create(nil, nil, args[1].x, args[1].y, args[1].z)
		end
	elseif #args == 2 and ATMEFct.isReference(args[1]) == true and ATMEFct.isReference(args[2]) == true then
		return ATMEFct.vectors["3D"].createFromReference(nil, nil, args[1], args[2])
	elseif #args == 3 and type(args[1]) == "number" and type(args[2]) == "number" and type(args[3]) == "number" then
		return ATMEFct.C_Vector3D_create(nil, nil, args[1], args[2], args[3])
	else
		if #args == 3 then
			if type(args[1]) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "x type must be number", false)
			elseif type(args[2]) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "y type must be number", false)
			elseif type(args[3]) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "z type must be number", false)
			end
		elseif #args == 2 then
			if ATMEFct.isReference(args[1]) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "refA type must be".. ATMEConst.errorLabelReference, false)
			elseif ATMEFct.isReference(args[2]) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "refB type must be".. ATMEConst.errorLabelReference, false)
			end
		elseif #args == 1 and ATMEFct.isObjectClass(args[1], ATME.C_Vector3D) == false and ATMEFct.isObjectClass(args[1], ATME.C_Vector2D) == false and
			(type(args[1]) ~= "table" or (type(args[1]) == "table" and (type(args[1].x) ~= "number" or type(args[1].y) ~= "number" or type(args[1].z) ~= "number"))) then
			ATMEFct.errorOutputCore(nil, fctDefinition, "vector type must be (x, y, z), ATME.C_Vector2D or ATME.C_Vector3D", false)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "must have 0 to 3 parameters", false)
		end
	end
end

setmetatable(ATME.C_Vector3D,ATMEDatas.C_Vector3D_MT)

ATMEDatas.C_Vector3D_MT.mt.__metatable = "Private"
ATMEDatas.C_Vector3D_MT.mt.__index = ATME.C_Vector3D

ATMEDatas.C_Vector3D_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Vector3D_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Vector3D : Bad key " .. k, "table locked", false)
	end
end

ATMEDatas.C_Vector3D_MT.mt.__add = function(_v1, _v2)
	if type(_v1) == "number" or type(_v2) == "number" then
		ATMEFct.errorOutputCore(nil, "Vectors add v1 + v2", "v1 or v2 type can't be numbers", false)
		return
	end
	
	local v1x, v1y, v1z = _v1:get()
	local v2x, v2y, v2z
	
	if ATMEFct.isObjectClass(_v2, ATME.C_Vector3D) then
		v2x, v2y, v2z = _v2:get()
	elseif ATMEFct.isObjectClass(_v2, ATME.C_Vector2D) then
		v2x, v2z = _v2:get()
		v2y = 0
	end

	return ATMEFct.C_Vector2D_create(nil, nil, v1x + v2x, v1y + v2y, v1z + v2z)
end

ATMEDatas.C_Vector3D_MT.mt.__sub = function(_v1, _v2)
	if type(_v1) == "number" or type(_v2) == "number" then
		ATMEFct.errorOutputCore(nil, "Vectors add v1 - v2", "v1 or v2 type can't be numbers", false)
		return
	end
	
	local v1x, v1y, v1z = _v1:get()
	local v2x, v2y, v2z
	
	if ATMEFct.isObjectClass(_v2, ATME.C_Vector3D) then
		v2x, v2y, v2z = _v2:get()
	elseif ATMEFct.isObjectClass(_v2, ATME.C_Vector2D) then
		v2x, v2z = _v2:get()
		v2y = 0
	end

	return ATMEFct.C_Vector2D_create(nil, nil, v1x - v2x, v1y - v2y, v1z - v2z)
end

ATMEDatas.C_Vector3D_MT.mt.__mul = function(_v1, _v2)
	if type(_v1) == "number" then
		return ATMEFct.C_Vector3D_create(nil, nil, _v2:getX() * _v1, _v2:getY() * _v1, _v2:getZ() * _v1)
	elseif type(_v2) == "number" then
		return ATMEFct.C_Vector3D_create(nil, nil, _v1:getX() * _v2, _v1:getY() * _v2, _v1:getZ() * _v2)
	else
		local v1x, v1y, v1z = _v1:get()
		local v2x, v2y, v2z
	
		if ATMEFct.isObjectClass(_v2, ATME.C_Vector3D) then
			v2x, v2y, v2z = _v2:get()
		elseif ATMEFct.isObjectClass(_v2, ATME.C_Vector2D) then
			v2x, v2z = _v2:get()
			v2y = 0
		end
		
		return v1x * v2x + v1y * v2y + v1z * v2z
	end
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Vector2D : gère les vecteurs 2D, format DCS repère z0x. 
-- Constructeur de classe accessible aux modules externes.
--
-- NOTE : l'axe y est non significatif
----------------------------------------------------------------------------------------------------------

ATMEFct.C_Vector2D_createNamed = function(_namedTable, _name, ...)
	if _namedTable[_name] ~= nil then
		ATMEFct.setError("ALREADY_CREATED")
		return 2, "named vector2D already exists"
	else
		local d,f = string.find(_name, "|", 1)
		if d == nil then
			local args = {...}
			
			if #args == 0 then
				return ATMEFct.C_Vector2D_create(_name, _namedTable, 0, 0)
			elseif #args == 1 then
				if ATMEFct.isObjectClass(args[1], ATME.C_Vector3D) == true then  
					return 0, ATMEFct.vectors["2D"].createFromVector3D(_name, _namedTable, args[1])
				elseif ATMEFct.isObjectClass(args[1], ATME.C_Vector2D) == true then
					return 0, ATMEFct.vectors["2D"].createFromVector2D(_name, _namedTable, args[1])
				elseif type(args[1]) == "table" and type(args[1].x) == "number" and type(args[1].z) == "number" then
					return 0, ATMEFct.C_Vector2D_create(_name, _namedTable, args[1].x, args[1].z)
				end
			elseif #args == 2 then
				if ATMEFct.isReference(args[1]) == true and ATMEFct.isReference(args[2]) == true then
					return 0, ATMEFct.vectors["2D"].createFromReference(_name, _namedTable, args[1], args[2])
				elseif type(args[1]) == "number" and type(args[2]) == "number" then
					return 0, ATMEFct.C_Vector2D_create(_name, _namedTable, args[1], args[2])
				end
			else
				if #args == 2 then
					if ATMEFct.isReference(args[1]) == false and type(args[1]) ~= "number" then
						return 1, "x, refA type must be number," .. ATMEConst.errorLabelReference
					elseif ATMEFct.isReference(args[2]) == false and type(args[2]) ~= "number" then
						return 1, "z, refB type must be number," .. ATMEConst.errorLabelReference
					end
				elseif #args == 1 and ATMEFct.isObjectClass(args[1], ATME.C_Vector3D) == false and ATMEFct.isObjectClass(args[1], ATME.C_Vector2D) == false and
					(type(args[1]) ~= "table" or (type(args[1]) == "table" and (type(args[1].x) ~= "number" or type(args[1].y) ~= "number" or type(args[1].z) ~= "number"))) then
					return 1, "vector type must be (x, y, z), ATME.C_Vector2D or ATME.C_Vector3D"
				else
					return 1, "must have 0 to 2 parameters"
				end
			end
		else
			ATMEFct.setError("BAD_FORMAT")
			return 2, "name can't contain '|' character"
		end
	end
end

ATMEDatas.C_Vector2D_MT = {
	className = "C_Vector2D",
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["getUserPrivateDatas"] = true,
		["getAzimuth"] = true, 
		["getModule"] = true,
		["get"] = true, 
		["getX"] = true,
		["getZ"] = true,
		["toUnitVector"] = true,
		["toVector3D"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
	},
	
	add = function(self)
		local item = {
			userPrivateDatas = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
	end,
}

ATME.C_Vector2D = {
	getClassName = function(self)
		return ATMEDatas.C_Vector2D_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Vector2D_MT.parentClass
	end,
	
	createNamed = function(_name, ...)
		local fctDefinition = "Function ATME.C_Vector2D.createNamed(name)\nATME.C_Vector2D.createNamed(name, vector)\nATME.C_Vector2D.createNamed(name, refA, refB)\nATME.C_Vector2D.createNamed(name, x, z)"
		
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
			local err, newVector2D = ATMEFct.C_Vector2D_createNamed(localMembersModule.vectors2D, _name, ...)				
			
			if err > 0 then
				if err == 1 then
					ATMEFct.errorOutputCore(nil, fctDefinition, newVector2D, false)
				else
					ATMEFct.warningOutputCore(nil, fctDefinition, newVector2D, false)
				end
				
				return
			end
			
			return newVector2D
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
		end
	end,
	
	getByName = function(_name)
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Vector2D.getByName(name)", "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].vectors2D[_name]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Vector2D.getByName(name)", "name type must be string", false)
		end
	end,
}

ATMEFct.C_Vector2D_create = function(_name, _vectors2D, _x, _z)
	local x = _x
	local z = _z
	
	local newVector = {}
	local localMembers
	local index
	
	local name
	local tagsName
	
	setmetatable(newVector, ATMEDatas.C_Vector2D_MT.mt)
	index = ATMEDatas.C_Vector2D_MT:add()
	localMembers = ATMEDatas.localMembers[index]
		
	if _name ~= nil and _vectors2D ~= nil then
		name = _name
		_vectors2D[name] = newVector
		tagsName = ATMEFct.decompString(_name)
	end
	
	local gModule = math.sqrt(x^2 + z^2)

	local angle = 0
	if x ~= 0 and z ~= 0 then
		angle = math.deg(math.atan2(z, x))
	end
	
	
	if angle < 0 then
		angle = angle + 360
	end
	
	function newVector.getIndex(self)
		return index
	end
	
	function newVector.getName(self)
		return name
	end
	
	function newVector.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newVector) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newVector.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newVector.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newVector) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newVector.getAzimuth(self)
		return angle
	end
	
	function newVector.getModule(self)
		return gModule
	end
	
	function newVector.get(self)
		return x, z
	end
	
	function newVector.getX(self)
		return x
	end

	function newVector.getZ(self)
		return z
	end
	
	function newVector.toUnitVector(self)
		local unitVector
		if gModule ~= 0 then
			unitVector = ATMEFct.C_Vector2D_create(nil, nil, x/gModule, z/gModule)
		else
			unitVector = ATMEFct.C_Vector2D_create(nil, nil)
		end
		
		return unitVector
	end
	
	function newVector.toVector3D(self)
		return ATMEFct.C_Vector3D_create(nil, nil, x, 0, z)
	end
	
	function newVector.getUserPrivateDatas(self, _moduleName)
		local fctDefinition = ATMEFct.getErrorPrefix(newVector) .. "getUserPrivateDatas([moduleName])"
		local moduleName
		
		if _moduleName ~= nil then
			if type(_moduleName) == "string" then
				if localMembers.userPrivateDatas[_moduleName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
					return
				else
					moduleName = _moduleName
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
				return
			end
		else
			moduleName = ATMEDatas.currentModuleName
		end
		
		return localMembers.userPrivateDatas[moduleName]
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return newVector
end

ATMEDatas.C_Vector2D_MT.__call = function(t, ...)
	local args = {...}
	
	local fctDefinition = "Constructor :\nATME.C_Vector2D()\nATME.C_Vector2D(vector)\nATME.C_Vector2D(refA, refB)\nATME.C_Vector2D(x, z)"
	
	if ATMEDatas.startMissionOk < 2 then
		ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
		return
	end
	
	local x = 0
	local z = 0
	
	if #args == 0 then
		return ATMEFct.C_Vector2D_create(nil, nil, 0, 0)
	elseif #args == 1 then
		if ATMEFct.isObjectClass(args[1], ATME.C_Vector3D) == true then  
			return ATMEFct.vectors["2D"].createFromVector3D(nil, nil, args[1])
		elseif ATMEFct.isObjectClass(args[1], ATME.C_Vector2D) == true then
			return ATMEFct.vectors["2D"].createFromVector2D(nil, nil, args[1])
		elseif type(args[1]) == "table" and type(args[1].x) == "number" and type(args[1].z) == "number" then
			return ATMEFct.C_Vector3D_create(nil, nil, args[1].x, args[1].z)
		end
	elseif #args == 2 then
		if ATMEFct.isReference(args[1]) == true and ATMEFct.isReference(args[2]) == true then
			return ATMEFct.vectors["2D"].createFromReference(nil, nil, args[1], args[2])
		elseif type(args[1]) == "number" and type(args[2]) == "number" then
			return ATMEFct.C_Vector2D_create(nil, nil, args[1], args[2])
		end
	else
		if #args == 2 then
			if ATMEFct.isReference(args[1]) == false and type(args[1]) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "x or refA type must be number, ".. ATMEConst.errorLabelReference, false)
			elseif ATMEFct.isReference(args[2]) == false and type(args[2]) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "z or refB type must be number, ".. ATMEConst.errorLabelReference, false)
			end
		elseif #args == 1 and ATMEFct.isObjectClass(args[1], ATME.C_Vector3D) == false and ATMEFct.isObjectClass(args[1], ATME.C_Vector2D) == false and
			(type(args[1]) ~= "table" or (type(args[1]) == "table" and (type(args[1].x) ~= "number" or type(args[1].y) ~= "number" or type(args[1].z) ~= "number"))) then
			ATMEFct.errorOutputCore(nil, fctDefinition, "vector type must be (x, y, z), ATME.C_Vector2D or ATME.C_Vector3D", false)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "must have 0 to 3 parameters", false)
		end
	end
end

setmetatable(ATME.C_Vector2D,ATMEDatas.C_Vector2D_MT)

ATMEDatas.C_Vector2D_MT.mt.__metatable = "Private"
ATMEDatas.C_Vector2D_MT.mt.__index = ATME.C_Vector2D

ATMEDatas.C_Vector2D_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Vector2D_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Vector2D : Bad key " .. k, "table locked", false)
	end
end

ATMEDatas.C_Vector2D_MT.mt.__add = function(_v1, _v2)
	if type(_v1) == "number" or type(_v2) == "number" then
		ATMEFct.errorOutputCore(nil, "Vectors add v1 + v2", "v1 or v2 type can't be numbers", false)
		return
	end
	
	local v1x, v1z = _v1:get()
	local v2x, v2z
	
	if ATMEFct.isObjectClass(_v2, ATME.C_Vector3D) then
		v2x, _, v2z = _v2:get()
	elseif ATMEFct.isObjectClass(_v2, ATME.C_Vector2D) then
		v2x, v2z = _v2:get()
	end

	return ATMEFct.C_Vector2D_create(nil, nil, v1x + v2x, v1z + v2z)
end

ATMEDatas.C_Vector2D_MT.mt.__sub = function(_v1, _v2)
	if type(_v1) == "number" or type(_v2) == "number" then
		ATMEFct.errorOutputCore(nil, "Vectors add v1 - v2", "v1 or v2 type can't be numbers", false)
		return
	end
	
	local v1x, v1z = _v1:get()
	local v2x, v2z
	
	if ATMEFct.isObjectClass(_v2, ATME.C_Vector3D) then
		v2x, _, v2z = _v2:get()
	elseif ATMEFct.isObjectClass(_v2, ATME.C_Vector2D) then
		v2x, v2z = _v2:get()
	end

	return ATMEFct.C_Vector2D_create(nil, nil, v1x - v2x, v1z - v2z)
end

ATMEDatas.C_Vector2D_MT.mt.__mul = function(_v1, _v2)
	if type(_v1) == "number" then
		return ATMEFct.C_Vector2D_create(nil, nil, _v2:getX() * _v1, _v2:getZ() * _v1)
	elseif type(_v2) == "number" then
		return ATMEFct.C_Vector2D_create(nil, nil, _v1:getX() * _v2, _v1:getZ() * _v2)
	else
		local v1x, v1z = _v1:get()
		local v2x, v2z
		
		if ATMEFct.isObjectClass(_v2, ATME.C_Vector3D) then
			v2x, _, v2z = _v2:get()
		elseif ATMEFct.isObjectClass(_v2, ATME.C_Vector2D) then
			v2x, v2z = _v2:get()
		end
		
		return v1x * v2x + v1z * v2z
	end
end

-- Fonctions génériques lines

ATMEFct.lines = {
	["2D"] = {
		createFromLine = function(_name, _namedTable, _line)
			return ATMEFct.C_Line2D_create(_name, _namedTable, _line:getA(), _line:getB(), _line:getC(), _line:getOrigin())
		end,
		createFromVector = function(_name, _namedTable, _vector, _ref)
			local origin = ATMEFct.getPointFromReference(_ref)
			local a = _vector:getX()
			local b = -_vector:getZ()
			
			return ATMEFct.C_Line2D_create(_name, _namedTable, a, b, - b * origin.x - a * origin.z, origin)
		end,
		createFromReference = function(_name, _namedTable, _refA, _refB)
			local origin = ATMEFct.getPointFromReference(_refA)
			local p2 = ATMEFct.getPointFromReference(_refB)
			local a = p2.x - origin.x
			local b = origin.z - p2.z
			
			if p2.x ~= origin.x or p2.z ~= origin.z then
				return ATMEFct.C_Line2D_create(_name, _namedTable, a, b, - b * origin.x - a * origin.z, origin)
			end
		end,
	},
}

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Line2D : gère les droites horizontales (2D). Format DCS repère z0x
--
-- Une droite est formée d'un vecteur et d'un point d'origine, référence 0 de la droite pour cette classe.
--
-- Forme Az + Bx + C = 0
-- Vecteur directeur v(xv, yv, zv) => A = xv et B = -zv
-- Point d'origine (xref, yref, zref) => C = -A * zref - B * xref
--
-- Droite perpendiculaire déduite et passant au point d'origine :
-- Vecteur directeur vp(xvp, yvp, zvp) => xvp = B et zvp = A
-- Forme : -Bz + Ax + C = 0 et C = -A * zref + B * xref
--
-- NOTE : l'axe y est non significatif
----------------------------------------------------------------------------------------------------------

ATMEFct.C_Line2D_createNamed = function(_namedTable, _name, ...)
	if _namedTable[_name] ~= nil then
		ATMEFct.setError("ALREADY_CREATED")
		return 1, "named line2D already exists"
	else
		local d,f = string.find(_name, "|", 1)
		if d == nil then
			local a
			local b
			local c
			local origin
			local args = {...}
			
			if #args == 1 and ATMEFct.isObjectClass(args[1], ATME.C_Line2D) == true then
				return 0, ATMEFct.lines["2D"].createFromLine(_name, _namedTable, args[1])
			elseif	#args == 2 then
				if ATMEFct.isReference(args[1]) == true and ATMEFct.isReference(args[2]) == true then
					local line = ATMEFct.lines["2D"].createFromReference(_name, _namedTable, args[1], args[2])
					
					if line ~= nil then
						return 0, line
					else
						ATMEFct.setError("SAME_REF")
						return 2, "refA and refB must be differents"
					end
				elseif ATMEFct.isObjectClass(args[1], ATME.C_Vector3D) == true and ATMEFct.isReference(args[2]) == true then
					return 0, ATMEFct.lines["2D"].createFromVector(_name, _namedTable, args[1], args[2])
				elseif ATMEFct.isObjectClass(args[1], ATME.C_Vector2D) == true and ATMEFct.isReference(args[2]) == true then
					return 0, ATMEFct.lines["2D"].createFromVector(_name, _namedTable, args[1], args[2])
				end
			else
				if #args == 2 then
					if ATMEFct.isReference(args[1]) == false and ATMEFct.isObjectClass(args[1], ATME.C_Vector3D) == false then
						return 1, "vector or refA type must be ATME.C_Vector3D, string," .. ATMEConst.errorLabelReference
					elseif ATMEFct.isReference(args[2]) == false then
						return 1, "origin or refB type must be string,".. ATMEConst.errorLabelReference
					end
				elseif #args == 1 and ATMEFct.isObjectClass(args[1], ATME.C_Line2D) == false then
					return 1, "line type must be ATME.C_Line2D"
				else
					return 1, "must have 1 or 2 parameters"
				end
			end
		else
			ATMEFct.setError("BAD_FORMAT")
			return 2, "name can't contain '|' character"
		end
	end
end

ATMEDatas.C_Line2D_MT = {
	className = "C_Line2D",
	mt = {},
	
	keys = {
		["get"] = true,
		["getA"] = true,
		["getB"] = true,
		["getC"] = true,
		["getX"] = true,
		["getZ"] = true,
		["isNSAxis"] = true,
		["isWEAxis"] = true,
		["getOrigin"] = true,
		["getPerpendicular"] = true,
		["getSegment2D"] = true,
		["getBissectrix"] = true,
		["getPointFromOrigin"] = true,
		["getIntersection"] = true,
		["getPointHDistance"] = true,
		["getPointDistance"] = true,
		["getVector"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["getIndex"] = true,
		["getUserPrivateDatas"] = true,
	},
	
	add = function(self)
		local item = {
			userPrivateDatas = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
	end,
}

ATME.C_Line2D = {
	getClassName = function(self)
		return ATMEDatas.C_Line2D_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Line2D_MT.parentClass
	end,
	
	createNamed = function(_name, ...)
		local fctDefinition = "Function ATME.C_Line2D.createNamed(name, line)\nATME.C_Line2D.createNamed(name, vector, origin)\nATME.C_Line2D.createNamed(name, refA, refB)"
		
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
			local err, newLine2D = ATMEFct.C_Line2D_createNamed(localMembersModule.lines2D, _name, ...)				
			
			if err > 0 then
				if err == 1 then
					ATMEFct.errorOutputCore(nil, fctDefinition, newLine2D, false)
				else
					ATMEFct.warningOutputCore(nil, fctDefinition, newLine2D, false)
				end
				
				return
			end
			
			return newLine2D
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
		end
	end,

	getByName = function(_name)
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Line2D.getByName(name)", "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].lines2D[_name]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Line2D.getByName(name)", "name type must be string", false)
		end
	end,
}

ATMEFct.C_Line2D_special = function(_line2D, _line2D1, _intersect, _parallel)
	local a, b, c = _line2D:get()
	local a1, b1, c1 = _line2D1:get()
	local result
	
	local delta = a1 * b - b1 * a
			
	if delta ~= 0 then
		if a1 ~= 0 then
			if a ~= 0 then
				local xTemp = (a * c1 - a1 * c) / delta
				result = ATMEFct.convertToPoint3D({
									x = xTemp,
									y = (-b * xTemp - c) / a,
								})
			else
				local xTemp = -c / b
				result = ATMEFct.convertToPoint3D({
									x = xTemp,
									y = (-b1 * xTemp - c1) / a1,
								})
			end
		else
			local xTemp = -c1 / b1
			result = ATMEFct.convertToPoint3D({
								x = xTemp,
								y = (-b * xTemp - c) / a,
							})
		end
		
		if _intersect == true then
			-- Intersection
			return result
		end
		
		-- Bissectrices
		
		local d = math.sqrt(a * a + b * b)
		local d1 = math.sqrt(a1 * a1 + b1 * b1)
		
		local v1 = ATMEFct.C_Vector2D_create(nil, nil, d1 * a + d * a1, -d1 * b - d * b1)
		local bis1 = ATMEFct.lines["2D"].createFromVector(nil, nil, v1, result)
		
		local v2 = ATMEFct.C_Vector2D_create(nil, nil,  d1 * a - d * a1, -d1 * b + d * b1)
		local bis2 = ATMEFct.lines["2D"].createFromVector(nil, nil, v2, result)
		
		return bis1, bis2
	else
		if _parallel == true then	
			
			local deltac = a1 * c - c1 * a
			if math.abs(deltac) < 1e-3 then
				return _line2D
			end
		else
			-- Test si parallèle, pour éviter les erreurs d'arrondis sur tracking WP 
			-- => _parallel = false
			
			return _line2D
		end
	end
end

ATMEFct.C_Line2D_getPointFromOrigin = function(_a, _b, _origin, _offset)
	local n = math.sqrt(_offset^2/(_a^2 + _b^2))
	local point = {}
	
	if _offset < 0 then
		n = -n
	end

	point.x = _origin.x + n * _a
	point.z = _origin.z - n * _b
	point.y = ATMEFct.getSurfaceAltitude(point)
	
	return point
end

ATMEFct.C_Line2D_getSegment2D = function(_origin, _a, _b, _offset1, _offset2)
	local n1 = math.sqrt(_offset1^2/(_a^2 + _b^2))
	local points = {}
	
	if _offset1 < 0 then
		n1 = -n1
	end

	points[1] = {}
	points[1].x = _origin.x + n1 * _a
	points[1].z = _origin.z - n1 * _b
	points[1].y = ATMEFct.getSurfaceAltitude(points[1])
	
	local n2
	if _offset2 == nil then
		n2 = -n1
	else
		n2 = math.sqrt(_offset2^2/(_a^2 + _b^2))
		
		if _offset2 < 0 then
			n2 = -n2
		end
	end
	
	points[2] = {}
	points[2].x = _origin.x + n2 * _a
	points[2].z = _origin.z - n2 * _b
	points[2].y = ATMEFct.getSurfaceAltitude(points[2])
	
	return ATMEFct.C_Segment2D_create(nil, nil, points)
end

ATMEFct.C_Line2D_create = function(_name, _lines2D, _a, _b, _c, _origin)
	local a = _a
	local b = _b
	local c = _c
	local origin = _origin
	local newLine = {}
	
	setmetatable(newLine, ATMEDatas.C_Line2D_MT.mt)
	
	local index = ATMEDatas.C_Line2D_MT:add()
	local localMembers = ATMEDatas.localMembers[index]
	
	local name
	local tagsName

	if _name ~= nil and _lines2D ~= nil then
		name = _name
		_lines2D[name] = newLine

		tagsName = ATMEFct.decompString(_name)
	end

	local vector = ATMEFct.C_Vector2D_create(nil, nil, a, -b)
	
	function newLine.getIndex(self)
		return index
	end
	
	function newLine.getUserPrivateDatas(self, _moduleName)
		local fctDefinition = ATMEFct.getErrorPrefix(newLine) .. "getUserPrivateDatas([moduleName])"
		local moduleName
		
		if _moduleName ~= nil then
			if type(_moduleName) == "string" then
				if localMembers.userPrivateDatas[_moduleName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
					return
				else
					moduleName = _moduleName
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
				return
			end
		else
			moduleName = ATMEDatas.currentModuleName
		end
		
		return localMembers.userPrivateDatas[moduleName]
	end
	
	function newLine.getName(self)
		return name
	end
	
	function newLine.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newLine) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newLine.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newLine.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newLine) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newLine.get(self)
		return a, b, c
	end
	
	function newLine.getA(self)
		return a
	end

	function newLine.getB(self)
		return b
	end

	function newLine.getC(self)
		return c
	end
	
	function newLine.getZ(self, _x)
		if a ~= 0 then
			return (-b * _x - c) / a
		end
	end
	
	function newLine.getX(self, _z)
		if b ~= 0 then
			return (-a * _z - c) / b
		end
	end
	
	function newLine.isNSAxis(self)
		return b == 0
	end
	
	function newLine.isWEAxis(self)
		return a == 0
	end
	
	function newLine.getOrigin(self)
		return {
			x = origin.x,
			y = origin.y,
			z = origin.z,
		}
	end
	
	function newLine.getVector(self)
		return vector
	end
	
	function newLine.getPointFromOrigin(self, _offset)
		local fctDefinition = ATMEFct.getErrorPrefix(newLine) .. "getPointFromOrigin(offset)"
		
		if type(_offset) == "number" then
			return ATMEFct.C_Line2D_getPointFromOrigin(a, b, origin, _offset)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "offset type must be number", true)
		end
	end
	
	function newLine.getSegment2D(self, _offset1, _offset2)
		local fctDefinition = ATMEFct.getErrorPrefix(newLine) .. "getSegment2D(offset1, [offset2])"
		
		if type(_offset1) == "number" and (_offset2 == nil or type(_offset2) == "number") then
			if (_offset1 ~= 0 and _offset2 == nil) or _offset1 ~= _offset2 then
				return ATMEFct.C_Line2D_getSegment2D(origin, a, b, _offset1, _offset2)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "offset1 must be ~= 0 or ~= offset2", false)
			end
		else
			if type(_offset1) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "offset1 type must be number", true)
			elseif _offset2 ~= nil and  type(_offset2) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "offset2 type must be nil or number", false)
			end
		end
	end
		
	function newLine.getPerpendicular(self, _offset)
		local fctDefinition = ATMEFct.getErrorPrefix(newLine) .. "getPerpendicular([offset])"
		
		if _offset == nil or type(_offset) == "number" then
			local v = ATMEFct.C_Vector2D_create(nil, nil, b, a)
			
			local perpendicularLine
			
			if _offset == nil or _offset == 0 then
				perpendicularLine = ATMEFct.lines["2D"].createFromVector(nil, nil, v, origin)
			else
				perpendicularLine = ATMEFct.lines["2D"].createFromVector(nil, nil, v, ATMEFct.C_Line2D_getPointFromOrigin(a, b, origin, _offset))
			end
			
			return perpendicularLine
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "offset type must be nil or number", true)
		end
	end
	
	function newLine.getBissectrix(self, _line2D)
		local fctDefinition = ATMEFct.getErrorPrefix(newLine) .. "getBissectrix(line2D)"
		
		if ATMEFct.isObjectClass(_line2D, ATME.C_Line2D) == true then
			return ATMEFct.C_Line2D_special(self, _line2D, false, true)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "line2D type must be ATME.C_Line2D", true)
		end
	end
	
	function newLine.getPointHDistance(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newLine) .. "getPointHDistance(reference)"
		
		if ATMEFct.isReference(_reference) == true then
			local point = ATME.getPointFromReference(_reference)
			
			return math.abs(a * point.z + b * point.x + c) / math.sqrt(a * a + b * b)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
		end
	end
	
	function newLine.getPointDistance(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newLine) .. "getPointDistance(reference)"
		
		if ATMEFct.isReference(_reference) == true then
			local point = ATME.getPointFromReference(_reference)
			
			local hdist = math.abs(a * point.z + b * point.x + c) / math.sqrt(a * a + b * b)
			
			return math.sqrt(hdist * hdist + point.y * point.y)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
		end
	end
	
	function newLine.getIntersection(self, _line2D)
		local fctDefinition = ATMEFct.getErrorPrefix(newLine) .. "getIntersection(line2D)"
		
		if ATMEFct.isObjectClass(_line2D, ATME.C_Line2D) == true then
			return ATMEFct.C_Line2D_special(self, _line2D, true, true)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "line2D type must be ATME.C_Line2D", true)
		end
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return newLine
end

ATMEDatas.C_Line2D_MT.__call = function(t, ...)
	local args = {...}
	local fctDefinition = "Constructor :\nATME.C_Line2D(line)\nATME.C_Line2D(vector, origin)\nATME.C_Line(refA, refB)"
	
	if ATMEDatas.startMissionOk < 2 then
		ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
		return
	end
	
	if #args == 1 and ATMEFct.isObjectClass(args[1], ATME.C_Line2D) == true then
		return ATMEFct.lines["2D"].createFromLine(_namedTable, _name, args[1])
	elseif	#args == 2 then
		if ATMEFct.isReference(args[1]) == true and ATMEFct.isReference(args[2]) == true then
			local line = ATMEFct.lines["2D"].createFromReference(_namedTable, _name, args[1], args[2])
			
			if line ~= nil then
				return line
			else
				ATMEFct.setError("SAME_REF")
				ATMEFct.warningOutputCore(nil, fctDefinition, "refA and refB must be differents", false)
			end
		elseif ATMEFct.isObjectClass(args[1], ATME.C_Vector3D) == true and ATMEFct.isReference(args[2]) == true then
			return ATMEFct.lines["2D"].createFromVector(_namedTable, _name, args[1], args[2])
		elseif ATMEFct.isObjectClass(args[1], ATME.C_Vector2D) == true and ATMEFct.isReference(args[2]) == true then
			return ATMEFct.lines["2D"].createFromVector(_namedTable, _name, args[1], args[2])
		end
	else
		if #args == 2 then
			if ATMEFct.isReference(args[1]) == false and ATMEFct.isObjectClass(args[1], ATME.C_Vector3D) == false and
						ATMEFct.isObjectClass(args[1], ATME.C_Vector2D) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "vector or refA type must be ATME.C_Vector3D," .. ATMEConst.errorLabelReference, false)
			elseif ATMEFct.isReference(args[2]) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "origin or refB type must be".. ATMEConst.errorLabelReference, false)
			end
		elseif #args == 1 and ATMEFct.isObjectClass(args[1], ATME.C_Line2D) == false then
			ATMEFct.errorOutputCore(nil, fctDefinition, "line type must be ATME.C_Line2D", false)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "must have 1 or 2 parameters", false)
		end
	end
end

setmetatable(ATME.C_Line2D,ATMEDatas.C_Line2D_MT)

ATMEDatas.C_Line2D_MT.mt.__metatable = "Private"
ATMEDatas.C_Line2D_MT.mt.__index = ATME.C_Line2D

ATMEDatas.C_Line2D_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Line2D_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Line2D : Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Segment2D : gère des segments 2D ayant ou non des points communs. 
-- Format DCS repère z0x.
--
-- Un ensemble de segments se caractérise par un ensemble de points consécutifs.
--
-- NOTE : l'axe y est non significatif
----------------------------------------------------------------------------------------------------------

ATMEFct.C_Segment2D_createNamed = function(_namedTable, _name, _points)
	if _namedTable[_name] ~= nil then
		ATMEFct.setError("ALREADY_CREATED")
		return true, "named segment2D already exists"
	else
		local d,f = string.find(_name, "|", 1)
		if d == nil then
			return false, ATMEFct.C_Segment2D_create(_name, _namedTable, _points)
		else
			ATMEFct.setError("BAD_FORMAT")
			return true, "name can't contain '|' character"
		end
	end
end

ATMEFct.C_Segment2D_intersect = function(_datas, _pA, _pB)
	local p1 = _datas.p1
	local p2 = _datas.p2
	local vector1 = _datas.vector
	
	local vector2 = {
		x = _pB.x - _pA.x,
		y = 0,
		z = _pB.z - _pA.z,
	}

	local det = vector1:getZ() * vector2.x - vector1:getX() * vector2.z
	
	if det == 0 then
		return false
	end
	
	local t = ((p1.x - _pA.x) * vector2.z - (p1.z - _pA.z) * vector2.x) / det
	if t < 0 or t > 1 then
		return false
	end
	
	local u = ((p1.x - _pA.x) * vector1:getZ() - (p1.z - _pA.z) * vector1:getX()) / det
	if u < 0 or u > 1 then
		return false
	end
	
	local side = ATMEFct.whichSide(p1, p2, _pB)
	if side < 0 then
		side = -1
	else
		side = 1
	end
	
	return true, side
end

ATMETimers.C_Segment2D_timer = function(_datas)
	if ATMEDatas.exitATME == false then
		if _datas.tracking.localMembersObject.objectDCS:isExist() == true then
			if _datas.tracking.className == "C_Weapon" or _datas.tracking.localMembersObject.objectDCS:isActive() == true then
				local currentPos = _datas.tracking.object:getPosition()
				local intersect, side = ATMEFct.C_Segment2D_intersect(_datas.localMembers.datas, _datas.tracking.pos, currentPos)
				
				_datas.tracking.pos = currentPos

				if intersect == true then		
					local datas = {
						direction = ATMEConst.directionValue[side],
						object = _datas.tracking.object,
					}

					for _moduleName, _callback in pairs(_datas.tracking.callbacks) do
						ATMEDatas.currentModuleName = _moduleName
						
						ATMEFct.outputCore(nil,"Appel callback - Franchissement de segment par l'unité " .. _datas.tracking.name .. " pour le module " .. _moduleName, 2)
						xpcall(function() _callback(_datas.segment2D, datas) end, function() ATMEFct.errUserHandler('Tracking callback') end)
					end
				end
			end
			
			return timer.getTime() + ATMEConst.offsetTracking[_datas.tracking.className]
		else
			_datas.localMembers.tracking[_datas.tracking.name] = nil
		end
	end
end

ATMEDatas.C_Segment2D_MT = {
	className = "C_Segment2D",
	mt = {},
	
	keys = {
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["getIndex"] = true,
		["getUserPrivateDatas"] = true,
		["getDatas"] = true,
		["addTrackingCallback"] = true,
		["removeTrackingCallback"] = true,
	},
	
	add = function(self)
		local item = {
			datas = {},
			tracking = {},
			userPrivateDatas = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].tracking = nil
	end,
}

ATME.C_Segment2D = {
	getClassName = function(self)
		return ATMEDatas.C_Segment2D_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Segment2D_MT.parentClass
	end,
	
	createNamed = function(_name, _refA, _refB)
		local fctDefinition = "Function ATME.C_Segment2D.createNamed(name, refA, refB)"
		
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" and ATMEFct.isReference(_refA) == true and ATMEFct.isReference(_refB) == true then
			local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
			local points = {
				[1] = ATMEFct.getPointFromReference(_refA),
				[2] = ATMEFct.getPointFromReference(_refB),
			}

			local err, newSegment2D = ATMEFct.C_Segment2D_createNamed(localMembersModule.segments2D, _name, points)				
			
			if err == true then
				ATMEFct.warningOutputCore(nil, fctDefinition, newSegment2D, false)
				return
			end

			return newSegment2D
		else
			if type(_name) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
			elseif ATMEFct.isReference(_refA) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "refA type must be".. ATMEConst.errorLabelReference, false)
			elseif ATMEFct.isReference(_refB) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "refB type must be".. ATMEConst.errorLabelReference, false)
			end
		end
	end,
	
	getByName = function(_name)
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Segment2D.getByName(name)", "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].segments2D[_name]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Segment2D.getByName(name)", "name type must be string", false)
		end
	end,
	
	createSegments = function(_points)
		local fctDefinition = "Function : ATME.C_Segment2D.createSegments(points)\nATME.C_Segment2D.createSegments(dcsZoneName)"
		
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_points) == "table" or type(_points) == "string" then
			local pErr
			local points = {}
			if type(_points) == "table" then
				for _key, _value in ipairs(_points) do
					if pErr == nil then
						local point
						
						if ATME.isPoint(_value) ~= POINT_BAD then
							point = ATME.convertToPoint3D(_value)
						elseif type(_value) == "string" then
							local zone = ATMEFct.getDCSZone(_value)
							
							if zone ~= nil then
								point = zone.point
							else
								pErr = "Bad zone in item key " .. _key .. " : DCS Zone " .. _value .. " doesn't exist"
							end
						else
							pErr = "Bad type for item key " .. _key
						end
						
						if pErr == nil then
							table.insert(points, point)
						end
					end
				end
			else
				-- Le '-' devient un séparateur suite à modif DCS, report ici pour uniformisation (" #%02d" remplacé par "-%d")
				for _i = 1, 40 do
					local currentName = _points .. string.format("-%d", _i)
					local zone = ATMEFct.getDCSZone(currentName)
					
					if zone ~= nil then
						table.insert(points, zone.point)
					end
				end
			end
			
			if pErr ~= nil then
				ATMEFct.errorOutputCore(nil, fctDefinition, pErr, false)
			else
				if #points >= 2 then
					local segments = {}
					
					for _i = 2, #points do
						table.insert(segments, ATMEFct.C_Segment2D_create(nil, nil, {points[_i - 1], points[_i]}))
					end
					
					return segments
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "#points must be >= 2 and <= 40", false)
				end
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "points type must be table or string", false)
		end
	end,
}

ATMEFct.C_Segment2D_create = function(_name, _segments2D, _points)
	local newSegment = {}
	
	setmetatable(newSegment, ATMEDatas.C_Segment2D_MT.mt)
	
	local index = ATMEDatas.C_Segment2D_MT:add()
	local localMembers = ATMEDatas.localMembers[index]
	
	local name
	local tagsName
	
	if _name ~= nil and _segments2D ~= nil then
		name = _name
		_segments2D[name] = newSegment
		tagsName = ATMEFct.decompString(_name)
	end
	
	localMembers.datas.p1 = {
		x = _points[1].x,
		y = _points[1].y,
		z = _points[1].z,
	}
	localMembers.datas.p2 = {
		x = _points[2].x,
		y = _points[2].y,
		z = _points[2].z,
	}
	
	localMembers.datas.vector = ATMEFct.C_Vector3D_create(nil, nil, _points[2].x - _points[1].x, 0, _points[2].z - _points[1].z)
	
	function newSegment.getIndex(self)
		return index
	end
	
	function newSegment.getName(self)
		return name
	end
	
	function newSegment.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newSegment) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newSegment.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newSegment.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newSegment) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newSegment.getDatas(self)
		local fctDefinition = ATMEFct.getErrorPrefix(newSegment) .. "getDatas()"
		
		return localMembers.datas.p1, localMembers.datas.p2, localMembers.datas.vector
	end
	
	function newSegment.addTrackingCallback(self, _fct, _object)
		local fctDefinition = ATMEFct.getErrorPrefix(newSegment) .. "addTrackingCallback(fct, object)"

		if type(_fct) == "function" and (ATMEFct.isObjectClass(_object, ATME.C_Unit) == true or ATMEFct.isObjectClass(_object, ATME.C_Weapon) == true) then
			local className = _object:getClassName()
			local name = _object:getName()
			if localMembers.tracking[name] == nil then
				localMembers.tracking[name] = {
					className = className,
					name = name,
					object = _object,
					localMembersObject = ATMEDatas.localMembers[_object:getIndex()],
					callbacks = {},
				}
				
				localMembers.tracking[name].pos = _object:getPosition()
				
				local datas = {
					tracking = localMembers.tracking[name],
					localMembers = localMembers,
					segment2D = self,
				}
				
				timer.scheduleFunction(ATMETimers.C_Segment2D_timer, datas, timer.getTime() + ATMEConst.offsetTracking[className])
			end
			
			if localMembers.tracking[name].callbacks[ATMEDatas.currentModuleName] == nil then
				localMembers.tracking[name].callbacks[ATMEDatas.currentModuleName] = _fct
			end
		else
			if type(_fct) ~= "function" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
			elseif ATMEFct.isObjectClass(_object, ATME.C_Unit) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "object type must be ATME.C_Unit or ATME.C_Weapon", false)
			end
		end
	end
	
	function newSegment.removeTrackingCallback(self, _object)
		local fctDefinition = ATMEFct.getErrorPrefix(newSegment) .. "removeTrackingCallback(object)"
		
		if ATMEFct.isObjectClass(_object, ATME.C_Unit) == true or ATMEFct.isObjectClass(_object, ATME.C_Weapon) == true then
			if localMembers.tracking[_object:getName()] ~= nil then
				localMembers.tracking[_object:getName()] = nil
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "object type must be ATME.C_Unit or ATME.C_Weapon", true)
		end
	end
	
	function newSegment.getUserPrivateDatas(self, _moduleName)
		local fctDefinition = ATMEFct.getErrorPrefix(newSegment) .. "getUserPrivateDatas([moduleName])"
		local moduleName
		
		if _moduleName ~= nil then
			if type(_moduleName) == "string" then
				if localMembers.userPrivateDatas[_moduleName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
					return
				else
					moduleName = _moduleName
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
				return
			end
		else
			moduleName = ATMEDatas.currentModuleName
		end
		
		return localMembers.userPrivateDatas[moduleName]
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return newSegment
end

ATMEDatas.C_Segment2D_MT.__call = function(t, _refA, _refB)
	local fctDefinition = "Constructor :\nATME.C_Segment2D(refA, refB)"
	
	if ATMEDatas.startMissionOk < 2 then
		ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
		return
	end
	
	if ATMEFct.isReference(_refA) == true and ATMEFct.isReference(_refB) == true then	
		local points = {
			[1] = ATMEFct.getPointFromReference(_refA),
			[2] = ATMEFct.getPointFromReference(_refB),
		}

		return ATMEFct.C_Segment2D_create(nil, nil, points)
	else
		if ATMEFct.isReference(_refA) == false then
			ATMEFct.errorOutputCore(nil, fctDefinition, "refA type must be" .. ATMEConst.errorLabelReference, false)
		elseif ATMEFct.isReference(_refB) == false then
			ATMEFct.errorOutputCore(nil, fctDefinition, "refB type must be" .. ATMEConst.errorLabelReference, false)
		end
	end
end

setmetatable(ATME.C_Segment2D,ATMEDatas.C_Segment2D_MT)

ATMEDatas.C_Segment2D_MT.mt.__metatable = "Private"
ATMEDatas.C_Segment2D_MT.mt.__index = ATME.C_Segment2D

ATMEDatas.C_Segment2D_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Segment2D_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Segment2D : Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Classes Zones 2D et 3D
--      ATME.C_Circle : gère les cercles sur le plan zOx.
--		ATME.C_Polygon : gère les polygones non complexes dans le plan zOx
-- 		ATME.C_Rectangle : gère les rectangles dans le plan zOx
-- 		ATME.C_PickupZone : gère les zones de chargement d'unité qui sont des zones DCS particulières
--                          donc aussi des cercles 2D.
--		ATME.C_Sphere : gère les sphères en 3D
--		ATME.C_Cylinder : gère les cylindres verticaux en 3D (rotation autour de l'axe y de DCS)
--		ATME.C_Box : gère les boites en 3D
----------------------------------------------------------------------------------------------------------

ATMEFct.C_Circle_isInside = function(_point, _datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	
	local dx2 = (_point.x - _datas.actualCenter.x) * (_point.x - _datas.actualCenter.x)
	local dz2 = (_point.z - _datas.actualCenter.z) * (_point.z - _datas.actualCenter.z)
	local d = math.sqrt(dx2 + dz2) 

	return d >= localMembers.radiusMin and d <= localMembers.radiusMax
end

ATMEFct.C_Circle_getRandomPosition = function(_datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	local angle = 2 * math.pi * math.random()
	local radius = (localMembers.radiusMax - localMembers.radiusMin) * math.random() + localMembers.radiusMin
	
	local point = {
		x = _datas.actualCenter.x + radius * math.sin(angle),
		y = _datas.actualCenter.z + radius * math.cos(angle),
	}
	
	return {
		x = point.x,
		y = ATMEFct.getSurfaceAltitude(point),
		z = point.y,
	}
end

ATMEFct.C_Circle_searchVolume = function(_datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]

	return 	world.VolumeType.BOX,
			"C_FindContext_search",
			{
				min = {
					x = _datas.actualCenter.x - localMembers.radiusMax,
					y = 0,
					z = _datas.actualCenter.z - localMembers.radiusMax,
				},
				max = {
					x = _datas.actualCenter.x + localMembers.radiusMax,
					y = 200000,
					z = _datas.actualCenter.z + localMembers.radiusMax,
				},
			}
end

ATMEDatas.C_Circle_MT = {
	className = "C_Circle",
	parentClass = ATME.C_MovableZone2D,
	
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["getUserPrivateDatas"] = true,
		["isInside"] = true,
		["getRandomPosition"] = true,
		["addTrackingCallback"] = true,
		["removeTrackingCallback"] = true,
		["addResultCallback"] = true,
		["removeResultCallback"] = true,
		["abortResultCallback"] = true,
		["getCenter"] = true,
		["getRadius"] = true,
		["changeThickness"] = true,
		["getThickness"] = true,
		["bindUnit"] = true,
		["getBoundUnit"] = true,
	},
	
	add = function(self, _center, _radiusMax)
		local item = {
			center = _center,
			radiusMax = _radiusMax,
			radiusMin = 0,
			thickness = 0,
			randomRatio = math.pi * _radiusMax * _radiusMax,
			is3D = false,
			box = {
				min = {
					x = _center.x - _radiusMax,
					y = 0,
					z = _center.z - _radiusMax,
				},
				max = {
					x = _center.x + _radiusMax,
					y = 200000,
					z = _center.z + _radiusMax,
				},
			},
			tracking = {},
			userPrivateDatas = {},
			resultCallbacks = {},
			coroutines = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].tracking = nil
		ATMEDatas.localMembers[_id].box = nil
		ATMEDatas.localMembers[_id].resultCallbacks = nil
		ATMEDatas.localMembers[_id].coroutines = nil
	end,
}

ATME.C_Circle = {
	getClassName = function(self)
		return ATMEDatas.C_Circle_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Circle_MT.parentClass
	end,
	
	createNamed = function(_name, ...)
		local args = {...}
		local fctDefinition = "Function :\nATME.C_Circle.createNamed(name, DCSZoneName)\nATME.C_Circle.createNamed(name, reference, radius)"
		local point, radius
		
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
			return
		end
		
		if type(_name) == "string" then
			if type(args[1]) == "string" then
				local zone = ATMEFct.getDCSZone(args[1])
				
				if zone ~=  nil then
					point = zone.point
					radius = zone.radius
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone '" .. args[1] .. "' unknown", false)
					return
				end
			elseif ATMEFct.isReference(args[1]) == true then
				if type(args[2]) == "number" then
					radius = args[2]
					if radius > 0 then
						point = ATMEFct.getPointFromReference(args[1])
						point.y = 0
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "radius must be > 0", false)
						return
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "radius type must be number", false)
					return
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 type must be string," .. ATMEConst.errorLabelReference, false)
				return
			end
			
			local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
			local err, newCircle = ATMEFct.C_Zone_createNamed(localMembersModule.zones, _name, ATMEFct.C_Circle_create, point, radius)				
		
			if err == true then
				ATMEFct.warningOutputCore(nil, fctDefinition, newCircle, false)
				return
			end

			return newCircle
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
		end
	end,
	
	getByName = function(_name)
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Circle.getByName(name)", "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			if ATMEFct.isObjectClass(ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name], ATME.C_Circle) == true then
				return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name]
			end
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Circle.getByName(name)", "name type must be string", false)
		end
	end,
}

ATMEFct.C_Circle_create = function(_name, _zones, _center, _radiusMax)
	local newCircle = {}
	
	setmetatable(newCircle, ATMEDatas.C_Circle_MT.mt)
	local index = ATMEDatas.C_Circle_MT:add(_center, _radiusMax)
	local localMembers = ATMEDatas.localMembers[index]

	local name
	local tagsName

	if _name ~= nil and _zones ~= nil then
		name = _name
		_zones[name] = newCircle
		tagsName = ATMEFct.decompString(_name)
	end
	
	function newCircle.getIndex(self)
		return index
	end
	
	function newCircle.getName(self)
		return name
	end
	
	function newCircle.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newCircle) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newCircle.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newCircle.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newCircle) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newCircle.changeThickness(self, _thickness)
		local fctDefinition = ATMEFct.getErrorPrefix(newCircle) .. "changeThickness([thickness])"
		
		if _thickness == nil or type(_thickness) == "number" then
			local thickness = 0
			if type(_thickness) == "number" then
				thickness = _thickness
			end
			
			if thickness >= localMembers.radiusMax then
				thickness = 0
			elseif thickness < 0 then
				ATMEFct.errorOutputCore(nil, fctDefinition, "thickness must be >= 0", false)
				return
			end
			
			localMembers.thickness = thickness
			
			if thickness == 0 then
				localMembers.radiusMin = 0
			else
				localMembers.radiusMin = localMembers.radiusMax - thickness
			end
			
			localMembers.randomRatio = math.pi * (localMembers.radiusMax * localMembers.radiusMax - localMembers.radiusMin * localMembers.radiusMin)
		else
			if _thickness ~= nil and type(_thickness) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "thickness type must be nil or number", true)
			end
		end
	end
	
	function newCircle.getThickness(self)	
		if localMembers.thickness ~= 0 then
			return localMembers.thickness
		end
	end
	
	function newCircle.getUserPrivateDatas(self, _moduleName)
		local fctDefinition = ATMEFct.getErrorPrefix(newCircle) .. "getUserPrivateDatas([moduleName])"
		local moduleName
		
		if _moduleName ~= nil then
			if type(_moduleName) == "string" then
				if localMembers.userPrivateDatas[_moduleName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
					return
				else
					moduleName = _moduleName
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
				return
			end
		else
			moduleName = ATMEDatas.currentModuleName
		end
		
		return localMembers.userPrivateDatas[moduleName]
	end
	
	function newCircle.isInside(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newCircle) .. "isInside(reference)"
		
		if ATMEFct.isReference(_reference) == true then
			local point = ATMEFct.getPointFromReference(_reference)
			
			local datas = {
				zone = self,
				actualCenter = ATMEFct.C_Zone_getActualCenter(localMembers),
				actualAzimuth = ATMEFct.C_Zone_getActualAzimuth(localMembers),
			}
			
			return ATMEFct.C_Circle_isInside(point, datas)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
		end
		
		return false
	end
	
	function newCircle.addResultCallback(self, _fct)
		local fctDefinition = ATMEFct.getErrorPrefix(newCircle) .. "addResultCallback(fct)"

		if type(_fct) == "function" then
			localMembers.resultCallbacks[ATMEDatas.currentModuleName] = {
				fct = _fct,
				aborted = false,
			}
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
		end
	end
	
	function newCircle.removeResultCallback(self)
		ATMEFct.C_Zone_abortResult(localMembers)
		localMembers.resultCallbacks[ATMEDatas.currentModuleName] = nil
	end
	
	function newCircle.abortResultCallback(self)
		ATMEFct.C_Zone_abortResult(localMembers)
	end
	
	function newCircle.getRandomPosition(self, _surfaceFilter, _isFlat, _altMin)
		local fctDefinition = ATMEFct.getErrorPrefix(newCircle) .. "getRandomPosition([surfaceFilter], [flat], [altMin])"
	
		if (_surfaceFilter == nil or type(_surfaceFilter) == "string") and
					(_isFlat == nil or type(_isFlat) =="boolean") and (_altMin == nil or type(_altMin) =="number") then

			local surfaceFilter = "ALL"
			
			if _surfaceFilter ~= nil then
				surfaceFilter = string.upper(_surfaceFilter)
			end
			
			if ATMEConst.surfaceType[surfaceFilter] ~= nil then
				local datas = {
					zone = self,
					surfaceFilter = surfaceFilter,
					isFlat = false,
					altMin = 0,
					localMembers = localMembers,
					moduleName = ATMEDatas.currentModuleName,
				}
				
				if _isFlat ~= nil then
					datas.isFlat = _isFlat
				end
				
				if type(_altMin) == "number" then
					datas.altMin = _altMin
				end
				
				if localMembers.resultCallbacks[ATMEDatas.currentModuleName] == nil then
					ATMEFct.C_Zone_getRandomPosition(datas)
					return datas.point
				else
					if localMembers.coroutines[ATMEDatas.currentModuleName] == nil then
						ATMEFct.outputCore(nil, "newCircle.getRandomPosition appel asynchrone", 2)
						localMembers.coroutines[ATMEDatas.currentModuleName] = coroutine.create(ATMEFct.C_Zone_getRandomPosition)
						coroutine.resume(localMembers.coroutines[ATMEDatas.currentModuleName], datas)
						localMembers.resultCallbacks[ATMEDatas.currentModuleName].idTimer = timer.scheduleFunction(ATMETimers.C_Zone_searchResultTimer, datas, timer.getTime() + 0.001)
						return
					else
						ATMEFct.setError("ALREADY_STARTED")
						ATMEFct.warningOutputCore(nil, fctDefinition, "search already started", false)
						return
					end
				end
			end
			
			ATMEFct.errorOutputCore(nil, fctDefinition, "surfaceFilter must be 'ALL', 'WATER' or 'LAND'", false)
		else
			if _surfaceFilter ~= nil and type(_surfaceFilter) ~= "string" then 
				ATMEFct.errorOutputCore(nil, fctDefinition, "surfaceFilter type must be nil or string", true)
			elseif _isFlat ~= nil and type(_isFlat) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "flat type must be nil or boolean", false)
			elseif _altMin ~= nil and type(_altMin) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "altMin type must be nil or number", false)
			end
		end
	end
	
	function newCircle.getRadius(self)
		return localMembers.radiusMax
	end
	
	function newCircle.getCenter(self)
		return ATMEFct.C_Zone_getActualCenter(localMembers)
	end
	
	function newCircle.bindUnit(self, _unit, _azimuth, _distance, _deltaAltitude)
		local fctDefinition = ATMEFct.getErrorPrefix(newCircle) .. "bindUnit()\nbindUnit(unit, [azimuth, distance, [deltaAltitude]])"
		
		if _unit == nil or ATMEFct.isObjectClass(_unit, ATME.C_Unit) == true then
			if _unit == nil then
				localMembers.boundUnit = nil
			else
				local localMembersUnit = ATMEDatas.localMembers[_unit:getIndex()]
			
				if localMembersUnit.objectDCS:isExist() == true then
					localMembers.boundUnit = {
						unit = _unit,
						azimuthRad = 0,
						distance = 0,
						deltaAltitude = 0
					}

					if (_azimuth == nil and _distance == nil) or (type(_azimuth) == "number" and type(_distance) == "number") then
						local azimuth = 0
						local distance = 0
						
						if _distance ~= nil then
							distance = _distance
							azimuth = _azimuth
						end
						
						if distance >= 0 then
							if _deltaAltitude == nil or type(_deltaAltitude) == "number" then
								localMembers.boundUnit.azimuthRad = math.rad(ATMEFct.getAzimuthMod(azimuth, 360))
								localMembers.boundUnit.distance = math.floor(distance)
								
								if type(_deltaAltitude) == "number" then
									localMembers.boundUnit.deltaAltitude = _deltaAltitude
								end
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "deltaAltitude type must be nil or number", false)
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "distance must be >= 0", false)
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "azimuth and distance type must be both nil or number", false)
					end
				else
					localMembers.boundUnit = nil
					ATMEFct.setError("ALREADY_DISABLED")
					ATMEFct.warningOutputCore(nil, fctDefinition, "bound unit already disabled", false)
				end
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "unit type must be nil or ATME.C_Unit", true)
		end
	end
	
	function newCircle.getBoundUnit(self)
		if localMembers.boundUnit ~= nil then
			local localMembersUnit = ATMEDatas.localMembers[localMembers.boundUnit.unit:getIndex()]
			
			if localMembersUnit.objectDCS:isExist() == true then
				return localMembers.boundUnit.unit
			end
		end
	end
	
	function newCircle.addTrackingCallback(self, _fct, _object)
		local fctDefinition = ATMEFct.getErrorPrefix(newCircle) .. "addTrackingCallback(fct, object)"

		if type(_fct) == "function" and (ATMEFct.isObjectClass(_object, ATME.C_Unit) == true or ATMEFct.isObjectClass(_object, ATME.C_Weapon) == true) then
			local className = _object:getClassName()
			local name = _object:getName()
			if localMembers.tracking[name] == nil then
				localMembers.tracking[name] = {
					className = className,
					name = name,
					object = _object,
					localMembersObject = ATMEDatas.localMembers[_object:getIndex()],
					callbacks = {},
				}
			end
				
			localMembers.tracking[name].pos = _object:getPosition()
			
			localMembers.tracking[name].callbacks[ATMEDatas.currentModuleName] = _fct
			
			local datas = {
				name = name,
				localMembers = localMembers,
				zone = self,
			}
			
			timer.scheduleFunction(ATMETimers.C_Zone_timer, datas, timer.getTime() + ATMEConst.offsetTracking[className])
		else
			if type(_fct) ~= "function" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
			elseif ATMEFct.isObjectClass(_object, ATME.C_Unit) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "object type must be ATME.C_Unit or ATME.C_Weapon", false)
			end
		end
	end
	
	function newCircle.removeTrackingCallback(self, _object)
		local fctDefinition = ATMEFct.getErrorPrefix(newCircle) .. "removeTrackingCallback(object)"
		
		if ATMEFct.isObjectClass(_object, ATME.C_Unit) == true or ATMEFct.isObjectClass(_object, ATME.C_Weapon) == true then
			if localMembers.tracking[_object:getName()] ~= nil then
				localMembers.tracking[_object:getName()].callbacks[ATMEDatas.currentModuleName] = nil
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "object type must be ATME.C_Unit or ATME.C_Weapon", true)
		end
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return newCircle
end


ATMEDatas.C_Circle_MT.__call = function(t, ...)
	local args = {...}
	local fctDefinition = "Constructor :\nATME.C_Circle(DCSZoneName)\nATME.C_Circle(reference, radius)"
	local point, radius
	
	if ATMEDatas.startMissionOk < 2 then
		ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
		return
	end
	
	if type(args[1]) == "string" then
		local zone = ATMEFct.getDCSZone(args[1])
		
		if zone ~=  nil then	
			point = zone.point
			radius = zone.radius
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone '" .. args[1] .. "' unknown", false)
			return
		end
	elseif ATMEFct.isReference(args[1]) == true then
		if type(args[2]) == "number" then
			radius = args[2]
			if radius > 0 then
				point = ATMEFct.getPointFromReference(args[1])
				point.y = 0
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "radius must be > 0", false)
				return
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "radius type must be number", false)
			return
		end
	else
		ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #1 type must be string," .. ATMEConst.errorLabelReference, false)
		return
	end

	return ATMEFct.C_Circle_create(nil, nil, point, radius)
end

setmetatable(ATME.C_Circle, ATMEDatas.C_Circle_MT)

ATMEDatas.C_Circle_MT.mt.__metatable = "Private"
ATMEDatas.C_Circle_MT.mt.__index = ATME.C_Circle

ATMEDatas.C_Circle_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Circle_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Circle : Bad key " .. k, "table locked", false)
	end
end

ATMEFct.generateSignal = function(_datas)
	if _datas.timeStart == nil then
		_datas.timeStart = timer:getTime()
		
		if _datas.type == "SMOKE" then
			local color
			if ATMEConst.smokeColors[_datas.color] == -1 then
				color = ATMEConst.smokeColors[ATMEConst.smokeColorsIndex[math.random(1,#ATMEConst.smokeColorsIndex)]]
			else
				color = ATMEConst.smokeColors[_datas.color]
			end
			
			trigger.action.smoke(_datas.point, color)
			_datas.offsetTime = math.random(5,10) + ATMEConst.maxTimeSmoke
		else
			local color
			if ATMEConst.flareColors[_datas.color] == -1 then
				color = ATMEConst.flareColors[ATMEConst.flareColorsIndex[math.random(1, #ATMEConst.flareColorsIndex)]]
			else
				color = ATMEConst.flareColors[_datas.color]
			end
			
			trigger.action.signalFlare(_datas.point, color, math.random() * 2 * math.pi)
			_datas.offsetTime = math.random(5,20)
		end
	else
		if _datas.timeStart + _datas.offsetTime < timer:getTime() then
			if _datas.type == "SMOKE" then
				local color
				if ATMEConst.smokeColors[_datas.color] == -1 then
					color = ATMEConst.smokeColors[ATMEConst.smokeColorsIndex[math.random(1,#ATMEConst.smokeColorsIndex)]]
				else
					color = ATMEConst.smokeColors[_datas.color]
				end
				
				trigger.action.smoke(_datas.point, color)
				_datas.offsetTime = math.random(5,10) + ATMEConst.maxTimeSmoke
			else
				local color
				if ATMEConst.flareColors[_datas.color] == -1 then
					color = ATMEConst.flareColors[ATMEConst.flareColorsIndex[math.random(1, #ATMEConst.flareColorsIndex)]]
				else
					color = ATMEConst.flareColors[_datas.color]
				end
				
				trigger.action.signalFlare(_datas.point, color, math.random() * 2 * math.pi)
				_datas.offsetTime = math.random(5,20)
			end
			
			_datas.timeStart = timer.getTime()
		end
	end
end

ATMETimers.C_PickupZone_signalTimer = function(_localMembersPickupZone)
	if ATMEDatas.exitATME == false then
		if _localMembersPickupZone.datasSignal.currentDatas == nil then
			_localMembersPickupZone.datasSignal.currentDatas = ATMEDatas.signalCapabilityUnits:getFirst()
		else
			_localMembersPickupZone.datasSignal.currentDatas = ATMEDatas.signalCapabilityUnits:getNext()
			
			if _localMembersPickupZone.datasSignal.currentDatas == nil then
				_localMembersPickupZone.datasSignal.currentDatas = ATMEDatas.signalCapabilityUnits:getFirst()
			end
		end
		
		if _localMembersPickupZone.datasSignal.currentDatas ~= nil then
			local name = _localMembersPickupZone.datasSignal.currentDatas.localMembers.name
			if _localMembersPickupZone.datasSignal.currentDatas.localMembers.objectDCS:isExist() == true then
				local id = _localMembersPickupZone.datasSignal.currentDatas.localMembers.objectDCS:getCoalition()
				local point = _localMembersPickupZone.datasSignal.currentDatas.localMembers.objectDCS:getPoint()
				if ATMEConst.coalitionNames[id] == _localMembersPickupZone.coalitionName then
					if _localMembersPickupZone.datasSignal.signal.point ~= nil then
						if _localMembersPickupZone.datasSignal.currentDatas.localMembers.objectDCS:inAir() == true then 
							local v = ATMEFct.vectors["3D"].createFromReference(nil, nil, _localMembersPickupZone.datasSignal.signal.point, point)
							if _localMembersPickupZone.datasSignal.coalitionUnitsInZone[_localMembersPickupZone.datasSignal.currentDatas.localMembers.name] == nil and v:getHModule() <= _localMembersPickupZone.datasSignal.signal.radius then
								_localMembersPickupZone.datasSignal.coalitionUnitsInZone[_localMembersPickupZone.datasSignal.currentDatas.localMembers.name] = true
								local datas = {
									unit = _localMembersPickupZone.datasSignal.currentDatas.object,
									isInside = true,
								}
								
								for _moduleName, _callback in pairs(_localMembersPickupZone.callbacks) do
									ATMEDatas.currentModuleName = _moduleName

									ATMEFct.outputCore(nil,"Appel handler d'entrée de pickupzone signalée pour le module " .. _moduleName, 2)
									xpcall(function() _callback(_localMembersPickupZone.datasSignal.pickupZone, datas) end, function() ATMEFct.errUserHandler('pickupZoneSignalCallbackHandler') end)
								end
											
								_localMembersPickupZone.datasSignal.nbUnits = _localMembersPickupZone.datasSignal.nbUnits + 1
							elseif _localMembersPickupZone.datasSignal.coalitionUnitsInZone[_localMembersPickupZone.datasSignal.currentDatas.localMembers.name] ~= nil and v:getHModule() > _localMembersPickupZone.datasSignal.signal.radius then
								_localMembersPickupZone.datasSignal.coalitionUnitsInZone[_localMembersPickupZone.datasSignal.currentDatas.localMembers.name] = nil
								local datas = {
									unit = _localMembersPickupZone.datasSignal.currentDatas.object,
									isInside = false,
								}
								
								for _moduleName, _callback in pairs(_localMembersPickupZone.callbacks) do
									ATMEDatas.currentModuleName = _moduleName

									ATMEFct.outputCore(nil,"Appel handler de sortie de pickupzone signalée pour le module " .. _moduleName, 2)
									xpcall(function() _callback(_localMembersPickupZone.datasSignal.pickupZone, datas) end, function() ATMEFct.errUserHandler('pickupZoneSignalCallbackHandler') end)
								end
								
								_localMembersPickupZone.datasSignal.nbUnits = _localMembersPickupZone.datasSignal.nbUnits - 1
							end
						else
							if _localMembersPickupZone.datasSignal.coalitionUnitsInZone[_localMembersPickupZone.datasSignal.currentDatas.localMembers.name] ~= nil then
								_localMembersPickupZone.datasSignal.coalitionUnitsInZone[_localMembersPickupZone.datasSignal.currentDatas.localMembers.name] = nil
								_localMembersPickupZone.datasSignal.nbUnits = _localMembersPickupZone.datasSignal.nbUnits - 1
							end
						end
					end
				end
			else
				if _localMembersPickupZone.datasSignal.coalitionUnitsInZone[_localMembersPickupZone.datasSignal.currentDatas.localMembers.name] ~= nil then
					_localMembersPickupZone.datasSignal.nbUnits = _localMembersPickupZone.datasSignal.nbUnits - 1
					_localMembersPickupZone.datasSignal.coalitionUnitsInZone[_localMembersPickupZone.datasSignal.currentDatas.localMembers.name] = nil
				end
			end
		end

		if _localMembersPickupZone.datasSignal.signal ~= nil then
			if _localMembersPickupZone.datasSignal.nbUnits > 0 and _localMembersPickupZone.nbReadyToBoard > 0 then 
				ATMEFct.generateSignal(_localMembersPickupZone.datasSignal.signal)
			end
		end
		
		
		if _localMembersPickupZone.nbReadyToBoard == 0 and _localMembersPickupZone.datasSignal.nbUnits == 0 then
			_localMembersPickupZone.idTimer = nil
			return
		end
			
		return timer.getTime() + 0.5
	end
end

ATMETimers.C_PickupZone_readyToBoardTimer = function(_datas)
	if ATMEDatas.exitATME == false then
		if _datas.localMembersObject.objectDCS:isExist() == true and #_datas.localMembersObject.objectDCS:getUnits() ~= 0 and ATMEDatas.groups[_datas.localMembersObject.objectDCS:getName()] ~= nil then
			ATMEFct.outputCore(nil, "Tracking pickupZone sur groupe ".. _datas.name , 2)
			
			if _datas.object:isActivated() == true then
				if _datas.object:isReadyToBoard() == false then
					local distance = ATMEFct.C_Group_evalDistances(_datas.object, _datas.localMembersPickupZone.center, "MIN_H")

					if distance <= _datas.localMembersPickupZone.radiusMax then
						_datas.isInZone = true
					end
					
					local ok = false
					if _datas.isInZone == true then
						local firstUnit = ATMEFct.getFirstUnit(_datas.object)
						
						if firstUnit ~= nil then
							local vector = ATMEFct.vectors["3D"].createFromTable(nil, nil, ATMEDatas.localMembers[firstUnit:getIndex()].objectDCS:getVelocity())
							if _datas.nbTries == 3 or math.random() < _datas.randomMax or vector:getModule() == 0 then
								ok = true
							else
								_datas.nbTries = _datas.nbTries + 1
							end
						end
					elseif _datas.nbTries > 0 then
						ok = true
					end

					if ok == true then
						ATMEFct.C_Group_changeReadyToBoard(_datas.object, true, _datas.nextWaypoint)

						_datas.localMembersPickupZone.nbReadyToBoard = _datas.localMembersPickupZone.nbReadyToBoard + 1
						_datas.localMembersObject.idTimerPickupZone = nil
						
						if _datas.localMembersPickupZone.datasSignal ~= nil and _datas.localMembersPickupZone.idTimer == nil then
							_datas.localMembersPickupZone.idTimer = timer.scheduleFunction(ATMETimers.C_PickupZone_signalTimer, _datas.localMembersPickupZone, timer.getTime() + 1)
						end
						
						return
					end
				end
			end
			
			return timer:getTime() + 2
		else
			_datas.localMembersObject.idTimerPickupZone = nil
			_datas.localMembersObject.pickupZone = nil
		end
	end
end

ATMEDatas.C_PickupZone_MT = {
	className = "C_PickupZone",
	parentClass = ATME.C_Zone2D,
	
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["getUserPrivateDatas"] = true,
		["getRandomPosition"] = true,
		["getCenter"] = true,
		["getRadius"] = true,
		["getCoalitionName"]= true,
		["addTrackingCallback"] = true,
		["removeTrackingCallback"] = true,
		["addResultCallback"] = true,
		["removeResultCallback"] = true,
		["abortResultCallback"] = true,
		["isInside"] = true,
		["addSignalCallback"] = true,
		["removeSignalCallback"] = true,
	},
	
	add = function(self, _coalitionName, _center, _radiusMax)
		local item = {
			coalitionName = _coalitionName,
			center = _center,
			radiusMax = _radiusMax,
			radiusMin = 0,
			randomRatio = math.pi * _radiusMax * _radiusMax,
			is3D = false,
			box = {
				min = {
					x = _center.x - _radiusMax,
					y = 0,
					z = _center.z - _radiusMax,
				},
				max = {
					x = _center.x + _radiusMax,
					y = 200000,
					z = _center.z + _radiusMax,
				},
			},
			nbReadyToBoard = 0,
			userPrivateDatas = {},
			tracking = {},
			callbacks = {},
			resultCallbacks = {},
			coroutines = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].tracking = nil
		ATMEDatas.localMembers[_id].box = nil
		ATMEDatas.localMembers[_id].callbacks = nil
		ATMEDatas.localMembers[_id].resultCallbacks = nil
		ATMEDatas.localMembers[_id].coroutines = nil
	end,
}

ATME.C_PickupZone = {
	getClassName = function(self)
		return ATMEDatas.C_PickupZone_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_PickupZone_MT.parentClass
	end,
	
	createNamed = function(_name, ...)
		local args = {...}
		local fctDefinition = "Function :\nATME.C_PickupZone.createNamed(name, DCSZoneName, coalitionName)\nATME.C_PickupZone.createNamed(name, reference, radius, coalitionName)\nATME.C_PickupZone.createNamed(name, DCSZoneName, coalitionName, signalRadius, [signalType])\nATME.C_PickupZone.createNamed(name, reference, radius, coalitionName, signalRadius, [signalType])"
		local point, radius
		local argId = 2
		
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			if type(args[1]) == "string" then
				local zone = ATMEFct.getDCSZone(args[1])
				
				if zone ~= nil then
					point = zone.point
					radius = zone.radius
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone '" .. args[1] .. "' unknown", false)
					return
				end
			elseif ATMEFct.isReference(args[1]) == true then
				argId = 3
				if type(args[2]) == "number" then
					radius = args[2]
					if radius > 0 then
						point = ATMEFct.getPointFromReference(args[1])
						point.y = 0
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "radius must be > 0", false)
						return
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "radius type must be number", false)
					return
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 type must be string," .. ATMEConst.errorLabelReference, false)
				return
			end
			
			if type(args[argId]) == "string" then
				local coalitionName = string.upper(args[argId])
				if ATMEConst.coalitionIdsBlueRed[coalitionName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName must be 'BLUE' or 'RED'", false)
					return
				end
				
				if args[argId + 1] == nil or type(args[argId + 1]) == "number" then
					local signalRadius
					if args[argId + 1] ~= nil then
						signalRadius = math.floor(args[argId + 1])
						if signalRadius < 2000 then
							signalRadius = 2000
						end
					else
						signalRadius = 2000
					end
					
					local datasSignal = {
						radius = signalRadius,
					}

					if args[argId + 2] == nil or type(args[argId + 2]) == "string" then
						if args[argId + 2] ~= nil then
							local tokens = ATMEFct.analyzeString(args[argId + 2])
							if #tokens == 2 and tokens[1].type == "TOKEN_SIGNAL_TYPE" and (tokens[2].type == "TOKEN_COLOR" or tokens[2].type == "TOKEN_TYPE_RANDOM") then
								if tokens[1].value == "SMOKE" and ATMEConst.smokeColors[tokens[2].value] ~= nil then
									datasSignal.type = tokens[1].value
									datasSignal.color = tokens[2].value
								elseif tokens[1].value == "FLARE" and ATMEConst.flareColors[tokens[2].value] ~= nil then
									datasSignal.type = tokens[1].value
									datasSignal.color = tokens[2].value
								else
									ATMEFct.errorOutputCore(nil, fctDefinition, "signalType string bad format", false)
									return
								end
							end
						end

						local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
						local err, newPickupZone = ATMEFct.C_Zone_createNamed(localMembersModule.zones, _name, ATMEFct.C_PickupZone_create, coalitionName, point, radius, datasSignal)				

						if err == true then
							ATMEFct.warningOutputCore(nil, fctDefinition, newPickupZone, false)
							return
						end

						return newPickupZone
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "signalType type must be nil or string", false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "signalRadius type must be nil or number", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName type must be string", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
		end
	end,
	
	getByName = function(_name)
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, "Function ATME.C_PickupZone.getByName(name)", "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			if ATMEFct.isObjectClass(ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name], ATME.C_PickupZone) == true then
				return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name]
			end
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_PickupZone.getByName(name)", "name type must be string", false)
		end
	end,
}

ATMEFct.C_PickupZone_create = function(_name, _zones, _coalitionName, _center, _radiusMax, _datasSignal)
	local newPickupZone = {}

	setmetatable(newPickupZone, ATMEDatas.C_PickupZone_MT.mt)
	local index = ATMEDatas.C_PickupZone_MT:add(_coalitionName, _center, _radiusMax)
	local localMembers = ATMEDatas.localMembers[index]
							
	local name
	local tagsName

	if _name ~= nil and _zones ~= nil then
		name = _name
		_zones[name] = newPickupZone
		tagsName = ATMEFct.decompString(_name)
	end

	function newPickupZone.getIndex(self)
		return index
	end
	
	function newPickupZone.getName(self)
		return name
	end
	
	function newPickupZone.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newPickupZone) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newPickupZone.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newPickupZone.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newPickupZone) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newPickupZone.isInside(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newPickupZone) .. "isInside(reference)"
		
		if ATMEFct.isReference(_reference) == true then		
			local point = ATMEFct.getPointFromReference(_reference)
			
			local datas = {
				zone = self,
				actualCenter = localMembers.center,
			}
			
			return ATMEFct.C_Circle_isInside(point, datas)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
		end
		
		return false
	end
	
	function newPickupZone.getCoalitionName(self)
		return localMembers.coalitionName
	end
	
	function newPickupZone.addResultCallback(self, _fct)
		local fctDefinition = ATMEFct.getErrorPrefix(newPickupZone) .. "addResultCallback(fct)"

		if type(_fct) == "function" then
			localMembers.resultCallbacks[ATMEDatas.currentModuleName] = {
				fct = _fct,
				aborted = false,
			}
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
		end
	end
	
	function newPickupZone.removeResultCallback(self)
		ATMEFct.C_Zone_abortResult(localMembers)
		localMembers.resultCallbacks[ATMEDatas.currentModuleName] = nil
	end
	
	function newPickupZone.abortResultCallback(self)
		ATMEFct.C_Zone_abortResult(localMembers)
	end
	
	function newPickupZone.getRandomPosition(self, _surfaceFilter, _isFlat, _altMin)
		local fctDefinition = ATMEFct.getErrorPrefix(newPickupZone) .. "getRandomPosition([surfaceFilter], [flat], [altMin])"
	
		if (_surfaceFilter == nil or type(_surfaceFilter) == "string") and
					(_isFlat == nil or type(_isFlat) =="boolean") and (_altMin == nil or type(_altMin) =="number") then

			local surfaceFilter = "ALL"
			
			if _surfaceFilter ~= nil then
				surfaceFilter = string.upper(_surfaceFilter)
			end
			
			if ATMEConst.surfaceType[surfaceFilter] ~= nil then
				local datas = {
					zone = self,
					surfaceFilter = surfaceFilter,
					isFlat = false,
					altMin = 0,
					localMembers = localMembers,
					moduleName = ATMEDatas.currentModuleName,
				}
				
				if _isFlat ~= nil then
					datas.isFlat = _isFlat
				end
				
				if type(_altMin) == "number" then
					datas.altMin = _altMin
				end
				
				if localMembers.resultCallbacks[ATMEDatas.currentModuleName] == nil then
					ATMEFct.C_Zone_getRandomPosition(datas)
					return datas.point
				else
					if localMembers.coroutines[ATMEDatas.currentModuleName] == nil then
						ATMEFct.outputCore(nil, "newPickupZone.getRandomPosition appel asynchrone", 2)
						localMembers.coroutines[ATMEDatas.currentModuleName] = coroutine.create(ATMEFct.C_Zone_getRandomPosition)
						coroutine.resume(localMembers.coroutines[ATMEDatas.currentModuleName], datas)
						localMembers.resultCallbacks[ATMEDatas.currentModuleName].idTimer = timer.scheduleFunction(ATMETimers.C_Zone_searchResultTimer, datas, timer.getTime() + 0.001)
						return
					else
						ATMEFct.setError("ALREADY_STARTED")
						ATMEFct.warningOutputCore(nil, fctDefinition, "search already started", false)
						return
					end
				end
			end
			
			ATMEFct.errorOutputCore(nil, fctDefinition, "surfaceFilter must be 'ALL', 'WATER' or 'LAND'", false)
		else
			if _surfaceFilter ~= nil and type(_surfaceFilter) ~= "string" then 
				ATMEFct.errorOutputCore(nil, fctDefinition, "surfaceFilter type must be nil or string", true)
			elseif _isFlat ~= nil and type(_isFlat) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "flat type must be nil or boolean", false)
			elseif _altMin ~= nil and type(_altMin) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "altMin type must be nil or number", false)
			end
		end
	end
	
	function newPickupZone.getRadius(self)
		return localMembers.radiusMax
	end
	
	function newPickupZone.getCenter(self)
		return ATMEFct.C_Zone_getActualCenter(localMembers)
	end
	
	function newPickupZone.getUserPrivateDatas(self, _moduleName)
		local fctDefinition = ATMEFct.getErrorPrefix(newPickupZone) .. "getUserPrivateDatas([moduleName])"
		local moduleName
		
		if _moduleName ~= nil then
			if type(_moduleName) == "string" then
				if localMembers.userPrivateDatas[_moduleName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
					return
				else
					moduleName = _moduleName
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
				return
			end
		else
			moduleName = ATMEDatas.currentModuleName
		end
		
		return localMembers.userPrivateDatas[moduleName]
	end
	
	function newPickupZone.addSignalCallback(self, _fct)
		local fctDefinition = ATMEFct.getErrorPrefix(newPickupZone) .. "addSignalCallback(fct)"
		
		if type(_fct) == "function" then
			localMembers.callbacks[ATMEDatas.currentModuleName] = _fct
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
		end
	end
	
	function newPickupZone.removeSignalCallback(self)
		localMembers.callbacks[ATMEDatas.currentModuleName] = nil
	end
	
	function newPickupZone.addTrackingCallback(self, _fct, _object)
		local fctDefinition = ATMEFct.getErrorPrefix(newPickupZone) .. "addTrackingCallback(fct, object)"

		if type(_fct) == "function" and (ATMEFct.isObjectClass(_object, ATME.C_Unit) == true or ATMEFct.isObjectClass(_object, ATME.C_Weapon) == true) then
			local className = _object:getClassName()
			local name = _object:getName()
			if localMembers.tracking[name] == nil then
				localMembers.tracking[name] = {
					className = className,
					name = name,
					object = _object,
					localMembersObject = ATMEDatas.localMembers[_object:getIndex()],
					callbacks = {},
				}
			end
				
			localMembers.tracking[name].pos = _object:getPosition()
			
			localMembers.tracking[name].callbacks[ATMEDatas.currentModuleName] = _fct
			
			local datas = {
				name = name,
				localMembers = localMembers,
				zone = self,
			}
			
			timer.scheduleFunction(ATMETimers.C_Zone_timer, datas, timer.getTime() + ATMEConst.offsetTracking[className])
		else
			if type(_fct) ~= "function" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
			elseif ATMEFct.isObjectClass(_object, ATME.C_Unit) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "object type must be ATME.C_Unit or ATME.C_Weapon", false)
			end
		end
	end
	
	function newPickupZone.removeTrackingCallback(self, _object)
		local fctDefinition = ATMEFct.getErrorPrefix(newPickupZone) .. "removeTrackingCallback(object)"
		
		if ATMEFct.isObjectClass(_object, ATME.C_Unit) == true or ATMEFct.isObjectClass(_object, ATME.C_Weapon) == true then
			if localMembers.tracking[_object:getName()] ~= nil then
				localMembers.tracking[_object:getName()].callbacks[ATMEDatas.currentModuleName] = nil
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "object type must be ATME.C_Unit or ATME.C_Weapon", true)
		end
	end
	
	if _datasSignal ~= nil then
		localMembers.datasSignal = {
			coalitionUnitsInZone = {},
			signal = {
				type = _datasSignal.type,
				color = _datasSignal.color,
				radius = _datasSignal.radius,
			},
			nbUnits = 0,
			delay = 0,
			pickupZone = self,
		}
		
		if localMembers.datasSignal.signal.type == "SMOKE" then
			localMembers.datasSignal.signal.point = {
				x = _center.x + math.random(-20,20),
				y = _center.y,
				z = _center.z + math.random(-20,20),
			}
		elseif _datasSignal.type == "FLARE" then
			localMembers.datasSignal.signal.point = _center
		end
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end

	return newPickupZone
end


ATMEDatas.C_PickupZone_MT.__call = function(t, ...)
	local args = {...}
	local fctDefinition = "Constructor :\nATME.C_PickupZone(DCSZoneName, coalitionName)\nATME.C_PickupZone(reference, radius, coalitionName)\nATME.C_PickupZone(DCSZoneName, coalitionName, signalRadius, [signalType])\nATME.C_PickupZone(reference, radius, coalitionName, signalRadius, [signalType])"
	local point, radius
	local argId = 2
	
	if ATMEDatas.startMissionOk < 2 then
		ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
		return
	end
	
	if type(args[1]) == "string" then
		local zone = ATMEFct.getDCSZone(args[1])
		
		if zone ~= nil then
			point = zone.point
			radius = zone.radius
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone '" .. args[1] .. "' unknown", false)
			return
		end
	elseif ATMEFct.isReference(args[1]) == true then
		argId = 3
		if type(args[2]) == "number" then
			radius = args[2]
			if radius > 0 then
				point = ATMEFct.getPointFromReference(args[1])
				point.y = 0
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "radius must be > 0", false)
				return
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "radius type must be number", false)
			return
		end
	else
		ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #1 type must be string," .. ATMEConst.errorLabelReference, false)
		return
	end
	
	if type(args[argId]) == "string" then
		local coalitionName = string.upper(args[argId])
		if ATMEConst.coalitionIdsBlueRed[coalitionName] == nil then
			ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName must be 'BLUE' or 'RED'", false)
			return
		end
		
		if args[argId + 1] == nil or type(args[argId + 1]) == "number" then
			local signalRadius
			if args[argId + 1] ~= nil then
				signalRadius = math.floor(args[argId + 1])
				if signalRadius < 2000 then
					signalRadius = 2000
				end
			else
				signalRadius = 2000
			end
			
			local datasSignal = {
				radius = signalRadius,
			}

			if args[argId + 2] == nil or type(args[argId + 2]) == "string" then
				if args[argId + 2] ~= nil then
					local tokens = ATMEFct.analyzeString(args[argId + 2])
					if #tokens == 2 and tokens[1].type == "TOKEN_SIGNAL_TYPE" and (tokens[2].type == "TOKEN_COLOR" or tokens[2].type == "TOKEN_TYPE_RANDOM") then
						if tokens[1].value == "SMOKE" and ATMEConst.smokeColors[tokens[2].value] ~= nil then
							datasSignal.type = tokens[1].value
							datasSignal.color = tokens[2].value
						elseif tokens[1].value == "FLARE" and ATMEConst.flareColors[tokens[2].value] ~= nil then
							datasSignal.type = tokens[1].value
							datasSignal.color = tokens[2].value
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "signalType string bad format", false)
							return
						end
					end
				end
				
				return ATMEFct.C_PickupZone_create(nil, nil, coalitionName, point, radius, datasSignal)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "signalType type must be nil or string", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "signalRadius type must be nil or number", false)
		end
	else
		ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName type must be string", false)
	end
end

setmetatable(ATME.C_PickupZone, ATMEDatas.C_PickupZone_MT)

ATMEDatas.C_PickupZone_MT.mt.__metatable = "Private"
ATMEDatas.C_PickupZone_MT.mt.__index = ATME.C_PickupZone

ATMEDatas.C_PickupZone_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_PickupZone_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_PickupZone : Bad key " .. k, "table locked", false)
	end
end

ATMEFct.C_Polygon_isInTriangle = function(_points, _point)
	local dRef
	local isInside = true

	for _i = 1, 3 do
		local j = _i + 1
		if j == 4 then
			j = 1
		end

		local d = ATMEFct.whichSide(_points[_i], _points[j], _point)
		if d ~= 0 then
			if dRef == nil then
				dRef = d
			elseif dRef ~= d then
				isInside = false
			end
		end
	end

	return isInside
end

ATMEFct.C_Polygon_isInside = function(_point, _datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	local isInside = true
	local index = 1
	
	repeat
		isInside = ATMEFct.C_Polygon_isInTriangle(localMembers.triangles[index].points, _point)
		index = index + 1
	until isInside == true or index == #localMembers.triangles + 1

	return isInside
end

ATMEFct.C_Polygon_getRandomPosition = function(_datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	
	local idTriangle = math.random(1, #localMembers.triangles)
	
	local sectorMin = math.random(1, #localMembers.triangles[idTriangle].points)
	local sectorMax = sectorMin + 1
	
	if sectorMin == #localMembers.triangles[idTriangle].points then
		sectorMax = 1
	end
	
	local angleMin
	local angleMax
	
	if localMembers.triangles[idTriangle].direction == 1 then
		angleMin = localMembers.triangles[idTriangle].angleRad[sectorMin].angle
		angleMax = localMembers.triangles[idTriangle].angleRad[sectorMax].angle
	else
		angleMin = localMembers.triangles[idTriangle].angleRad[sectorMax].angle
		angleMax = localMembers.triangles[idTriangle].angleRad[sectorMin].angle
	end
	
	local angleRad = angleMin + math.random() * (angleMax - angleMin)
	
	local pt = {
		x = localMembers.triangles[idTriangle].barycentre.x + 10 * math.cos(angleRad),
		y = localMembers.triangles[idTriangle].barycentre.z + 10 * math.sin(angleRad),
	}
	
	local line = ATMEFct.lines["2D"].createFromReference(nil, nil, localMembers.triangles[idTriangle].barycentre, pt)
	
	local intersectPoint = line:getIntersection(localMembers.triangles[idTriangle].sides[sectorMin])
	
	local dx2 = (intersectPoint.x - localMembers.triangles[idTriangle].barycentre.x) * (intersectPoint.x - localMembers.triangles[idTriangle].barycentre.x)
	local dy2 = (intersectPoint.z - localMembers.triangles[idTriangle].barycentre.z) * (intersectPoint.z - localMembers.triangles[idTriangle].barycentre.z)
	local distance = math.sqrt(dx2 + dy2)
	
	local randomDistance = distance * math.random()
	
	return ATMEFct.C_Line2D_getPointFromOrigin(line:getA(), line:getB(), line:getOrigin(), randomDistance)
end

ATMEFct.C_Polygon_searchVolume = function(_datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	
	return world.VolumeType.BOX,
			"C_FindContext_search",
			localMembers.box
end

ATMEDatas.C_Polygon_MT = {
	className = "C_Polygon",
	parentClass = ATME.C_Zone2D,
	
	mt = {},
	
	keys = {
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["isInside"] = true,
		["getRandomPosition"] = true,
		["addTrackingCallback"] = true,
		["removeTrackingCallback"] = true,
		["addResultCallback"] = true,
		["removeResultCallback"] = true,
		["abortResultCallback"] = true,
		["getIndex"] = true,
		["getUserPrivateDatas"] = true,
	},
	
	add = function(self, _points, _triangles, _surface)
		local minPoint = {
			x = _points[1].x,
			y = 0,
			z = _points[1].z,
		}
			
		local maxPoint = {
			x = _points[1].x,
			y = 200000,
			z = _points[1].z,
		}
		
		for _, _point in ipairs(_points) do
			if minPoint.x > _point.x then
				minPoint.x = _point.x
			end
			if minPoint.z > _point.z then
				minPoint.z = _point.z
			end
			
			if maxPoint.x < _point.x then
				maxPoint.x = _point.x
			end
			if maxPoint.z < _point.z then
				maxPoint.z = _point.z
			end
		end
		
		local item = {
			points = _points,
			triangles = _triangles,
			randomRatio = _surface,
			is3D = false,
			box = {
				min = minPoint,
				max = maxPoint,
			},
			userPrivateDatas = {},
			tracking = {},
			resultCallbacks = {},
			coroutines = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].tracking = nil
		ATMEDatas.localMembers[_id].box = nil
		ATMEDatas.localMembers[_id].resultCallbacks = nil
		ATMEDatas.localMembers[_id].coroutines = nil
	end,
}

ATME.C_Polygon = {
	getClassName = function(self)
		return ATMEDatas.C_Polygon_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Polygon_MT.parentClass
	end,
	
	createNamed = function(_name, _vertices)
		local fctDefinition = "Function :\nATME.C_Polygon.createNamed(name, vertices)\nATME.C_Polygon.createNamed(name, DCSZoneName)"
		
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" and type(_vertices) == "table" or type(_vertices) == "string" then
			local pErr
			local points = {}
			if type(_vertices) == "table" then
				for _key, _value in ipairs(_vertices) do
					if pErr == nil then
						local point
						
						if ATME.isPoint(_value) ~= POINT_BAD then
							point = ATMEFct.convertToPoint3D(_value)
						elseif type(_value) == "string" then
							local zone = ATMEFct.getDCSZone(_value)
							
							if zone ~= nil then
								point = zone.point
							else
								pErr = "Bad zone in item key " .. _key .. " : DCS Zone " .. _value .. " doesn't exist"
							end
						else
							pErr = "Bad type for item key " .. _key
						end
						
						if pErr == nil then
							table.insert(points, point)
						end
					end
				end
			else
				-- Le '-' devient un séparateur suite à modif DCS, report ici pour uniformisation (" #%03d" remplacé par "-%d")
				for _i = 1, ATMEConst.polygonMaxVertices do
					local currentName = _vertices .. string.format("-%d", _i)
					local zone = ATMEFct.getDCSZone(currentName)
					
					if zone ~= nil then
						table.insert(points, zone.point)
					end
				end
			end
			
			if pErr ~= nil then
				ATMEFct.errorOutputCore(nil, fctDefinition, pErr, false)
			else
				if #points > 2 and #points <= ATMEConst.polygonMaxVertices then
				
					-- Vérification de polygone croisé
					local intersect = false
					
					for _i = 1, #points - 1 do
						for _j = _i + 1, #points - 1 do
							local nr = (points[_i].z - points[_j].z) * (points[_j + 1].x - points[_j].x) - (points[_i].x - points[_j].x) * (points[_j + 1].z - points[_j].z)
							local ns = (points[_i].z - points[_j].z) * (points[_i + 1].x - points[_i].x) - (points[_i].x - points[_j].x) * (points[_i + 1].z - points[_i].z)
							local d = (points[_i + 1].x - points[_i].x) * (points[_j + 1].z - points[_j].z) - (points[_i + 1].z - points[_i].z) * (points[_j + 1].x - points[_j].x)

							if d ~= 0 then
								local r = nr / d
								local s = ns / d
								if r > 0 and r < 1 and s > 0 and s < 1 then
									intersect = true
								end
							end
						end
					end


					if intersect == false then
						local pErr
						-- Vérification des distances entre points consécutifs
						for _i = 1, #points - 1 do
							for _j = _i + 1, #points do
								local dX2 = (points[_i].x - points[_j].x) * (points[_i].x - points[_j].x)
								local dY2 = (points[_i].z - points[_j].z) * (points[_i].z - points[_j].z)
								local dist2 = dX2 + dY2

								if dist2 < 1 then
									-- Points trop proches distance < 1
									
									pErr = {
										first = _i,
										last = _j,
									}
								end
							end
						end

						if pErr == nil then
							local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
							local err, newPolygon = ATMEFct.C_Zone_createNamed(localMembersModule.zones, _name, ATMEFct.C_Polygon_create, points)				
			
							if err == true then
								ATMEFct.warningOutputCore(nil, fctDefinition, newPolygon, false)
								return
							end

							return newPolygon
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "Bad polygon : Vertex " .. pErr.first .. " and " .. pErr.last .. " too near (distance < 1m)", false)
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Bad polygon : crossed vertices", false)
					end
				else
					if #points < 3 then
						ATMEFct.errorOutputCore(nil, fctDefinition, "Bad polygon : Less than 3 vertices", false)
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Bad polygon : More than limit of " .. ATMEConst.polygonMaxVertices .. " vertices", false)
					end
				end
			end
		else
			if type(_name) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "vertices type must be table or DCSZoneName type must be string", false)
			end
		end
	end,
	
	getByName = function(_name)
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Polygon.getByName(name)", "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			if ATMEFct.isObjectClass(ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name], ATME.C_Polygon) == true then
				return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name]
			end
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Polygon.getByName(name)", "name type must be string", false)
		end
	end,
}

ATMEFct.C_Polygon_create = function(_name, _zones, _points)
	local triangles = {}
	
	local function zMin(_points)
		local zMin = _points[1].z
		local index = 1
		for _i = 2, #_points do
			if _points[_i].z < zMin then
				index = _i
				zMin = _points[_i].z
			end
		end

		return index
	end
  
	local function nextVertexIndex(_points, _index)
		local newIndex = _index + 1
		local nbPoints = #_points

		if newIndex > nbPoints then
			newIndex = 1
		end

		return newIndex
	end

	local function previousVertexIndex(_points, _index)
		local newIndex = _index - 1
		local nbPoints = #_points

		if newIndex == 0 then
			newIndex = nbPoints
		end

		return newIndex
	end

	local function newPolygon(_polygon, _start, _end)
		local points = {}
		local i = _start

		while i ~= _end do
			table.insert(points, {
									x = _polygon[i].x,
									y = _polygon[i].y,
									z = _polygon[i].z,
									})

			i = nextVertexIndex(_polygon, i)
		end
		
		table.insert(points, {
								x = _polygon[i].x,
								y = _polygon[i].y,
								z = _polygon[i].z,
								})

		return points
	end

	local function inTriangle(_points, _M)
		local det = -1
		
		if ATMEFct.C_Polygon_isInTriangle(_points, _M) == true then
			det = math.abs((_points[3].z - _points[2].z) * (_M.x - _points[2].x) - 
								(_points[3].x - _points[2].x) * (_M.z - _points[2].z))
		end

		return det
	end
	
	local function findVertexInTriangle(_points, _iP1, _iP2, _iP3)
		local nbPoints = #_points
		local j
		local maxDet = 0

		for _i = 1, nbPoints do
			if _i ~= _iP1 and _i ~= _iP2 and _i ~= _iP3 then
				local pts = {
					[1] = _points[_iP1], 
					[2] = _points[_iP2], 
					[3] = _points[_iP3],
				}
				local det = inTriangle(pts, _points[_i])

				if det > maxDet then
					maxDet = det
					j = _i  
				end
			end
		end

		return j
	end
  
	local function analyze(_points)
		-- Fonction récursive de triangulation d'un polygone
		-- Pour un polygone en cours défini par ses point, cette triangulation 
		-- est basée sur :
		--    - La recherche du z minimal dans le plan zOx appelé Start
		--    - Si aucun autre point du polygon coupe le segment défini par précédent/suivant, 
		--      le premier triangle est crée avec ces 3 points. Le point Start est supprimé et 
		--      une nouvelle analyse est faite.
		--    - sinon le point le plus loin (pour optimiser utilisation du produit vectoriel)
		--      du segment défini par précédent/suivant est mémorisé (ici M) et le nouveau segment
		--      défini par Start et M induisent la création de deux polygones à partir du premier,
		--      une nouvelle analyse est alors lancée pour chaque polygone.
		--
		-- Dans tous les cas, l'analyse s'arrête s'il ne reste que 3 points et ce tringle est alors
		-- mémorisé. Il faut noter que les triangles plats ne sont pas mémorisés. Le triangle ABC est 
		-- un triangle plat si le produit vectoriel vaut 0 pour les vecteurs AB et AC.
		
		local startIndex = zMin(_points)
		local nextIndex = nextVertexIndex(_points, startIndex)
		local prevIndex = previousVertexIndex(_points, startIndex)

		local j = findVertexInTriangle(_points, startIndex, nextIndex, prevIndex)
		
		if j == nil then
			if ATMEFct.whichSide(_points[startIndex], _points[nextIndex], _points[prevIndex]) ~= 0 then
				table.insert(triangles, {
											points = {
												_points[startIndex], 
												_points[nextIndex], 
												_points[prevIndex],
											}
										})
			end
			
			local poly = newPolygon(_points, nextIndex, prevIndex)
			if #poly == 3 then
				if ATMEFct.whichSide(poly[1], poly[2], poly[3]) ~= 0 then
					table.insert(triangles, {
												points = poly,
											})
				end
			else
				analyze(poly)
			end
		else
			local poly1 = newPolygon(_points, startIndex, j)
			local poly2 = newPolygon(_points, j, startIndex)
			if #poly1 == 3 then
				if ATMEFct.whichSide(poly1[1], poly1[2], poly1[3]) ~= 0 then
					table.insert(triangles, {
												points = poly1,
											})
				end
			else
				analyze(poly1)
			end
			if #poly2 == 3 then
				if ATMEFct.whichSide(poly2[1], poly2[2], poly2[3]) ~= 0 then
					table.insert(triangles, {
												points = poly2,
											})
				end
			else
				analyze(poly2)
			end
		end
	end
  
 
	analyze(_points)
	
	
	-- Calcul des barycentres et des droites rayon qui relient le barycentre à chaque sommet.
	-- Mémorisation des côtés représentés par des droites 2D. La droite 2D i relie le sommet i au 
	-- sommet i + 1. Un triangle = 3 côtés et 3 rayons, donc pas besoin de recherche de direction.
	-- L'angle de chaque rayon est également mémorisé. Toutes ces valeurs seront utilisées pour déterminer
	-- de manière déterministe un point aléatoire dans un polygone.
	
	-- Calcul du barycentre et des côtés de chaque triangle.
	
	local surface = 0
	for _, _triangle in ipairs(triangles) do
		_triangle.sides = {}
		
		local sx = 0
		local sy = 0
		for _i = 1, #_triangle.points - 1 do
			sx = sx + _triangle.points[_i].x
			sy = sy + _triangle.points[_i].z
			
			_triangle.sides[_i] = ATMEFct.lines["2D"].createFromReference(nil, nil, _triangle.points[_i], _triangle.points[_i + 1])
		end
		
		sx = sx + _triangle.points[#_triangle.points].x
		sy = sy + _triangle.points[#_triangle.points].z
		
		_triangle.sides[#_triangle.points] = ATMEFct.lines["2D"].createFromReference(nil, nil, _triangle.points[#_triangle.points], _triangle.points[1])
		
		_triangle.barycentre = ATMEFct.convertToPoint3D({
								x = sx / #_triangle.points,
								y = sy / #_triangle.points,
							})
		
		local a = math.sqrt((_triangle.points[2].x - _triangle.points[1].x) ^ 2 + (_triangle.points[2].z - _triangle.points[1].z) ^ 2)
		local b = math.sqrt((_triangle.points[3].x - _triangle.points[2].x) ^ 2 + (_triangle.points[3].z - _triangle.points[2].z) ^ 2)
		local c = math.sqrt((_triangle.points[1].x - _triangle.points[3].x) ^ 2 + (_triangle.points[1].z - _triangle.points[3].z) ^ 2)
		local abc = (a + b + c) / 2
		
		surface = surface + math.sqrt(abc * (abc - a) * (abc - b) * (abc - c)) 
			
		-- Mémorisation des droites rayons : lient le barycentre de chaque triangle à ses sommets. 
		-- Evaluation de l'angle
		_triangle.angleRad = {}
		for _i = 1, #_triangle.points do
			local line = ATMEFct.lines["2D"].createFromReference(nil, nil, _triangle.barycentre,  _triangle.points[_i])
			_triangle.angleRad[_i] = {
				angle = math.atan2(-line:getB(), line:getA()),
				index = _i,
			}
		end
	end


	local newPolygon = {}
	
	setmetatable(newPolygon, ATMEDatas.C_Polygon_MT.mt)
	local index = ATMEDatas.C_Polygon_MT:add(_points, triangles, surface)
	local localMembers = ATMEDatas.localMembers[index]
	
	local name
	local tagsName

	if _name ~= nil and _zones ~= nil then
		name = _name
		_zones[name] = newPolygon
		tagsName = ATMEFct.decompString(_name)
	end

	function newPolygon.getIndex(self)
		return index
	end
	
	function newPolygon.getName(self)
		return name
	end
	
	function newPolygon.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newPolygon) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newPolygon.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newPolygon.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newPolygon) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newPolygon.getUserPrivateDatas(self, _moduleName)
		local fctDefinition = ATMEFct.getErrorPrefix(newPolygon) .. "getUserPrivateDatas([moduleName])"
		local moduleName
		
		if _moduleName ~= nil then
			if type(_moduleName) == "string" then
				if localMembers.userPrivateDatas[_moduleName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
					return
				else
					moduleName = _moduleName
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
				return
			end
		else
			moduleName = ATMEDatas.currentModuleName
		end
		
		return localMembers.userPrivateDatas[moduleName]
	end
	
	function newPolygon.isInside(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newPolygon) .. "isInside(reference)"
		local isInside = false
		
		if ATMEFct.isReference(_reference) == true then		
			local point = ATMEFct.getPointFromReference(_reference)
			
			local datas = {
				zone = self,
			}
				
			isInside = ATMEFct.C_Polygon_isInside(point, datas)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
		end
		
		return isInside
	end
	
	function newPolygon.addResultCallback(self, _fct)
		local fctDefinition = ATMEFct.getErrorPrefix(newPolygon) .. "addResultCallback(fct)"

		if type(_fct) == "function" then
			localMembers.resultCallbacks[ATMEDatas.currentModuleName] = {
				fct = _fct,
				aborted = false,
			}
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
		end
	end
	
	function newPolygon.removeResultCallback(self)
		ATMEFct.C_Zone_abortResult(localMembers)
		localMembers.resultCallbacks[ATMEDatas.currentModuleName] = nil
	end
	
	function newPolygon.abortResultCallback(self)
		ATMEFct.C_Zone_abortResult(localMembers)
	end
	
	function newPolygon.getRandomPosition(self, _surfaceFilter, _isFlat, _altMin)
		local fctDefinition = ATMEFct.getErrorPrefix(newPolygon) .. "getRandomPosition([surfaceFilter], [flat], [altMin])"
	
		if (_surfaceFilter == nil or type(_surfaceFilter) == "string") and
					(_isFlat == nil or type(_isFlat) =="boolean") and (_altMin == nil or type(_altMin) =="number") then

			local surfaceFilter = "ALL"
			
			if _surfaceFilter ~= nil then
				surfaceFilter = string.upper(_surfaceFilter)
			end
			
			if ATMEConst.surfaceType[surfaceFilter] ~= nil then
				local datas = {
					zone = self,
					surfaceFilter = surfaceFilter,
					isFlat = false,
					altMin = 0,
					localMembers = localMembers,
					moduleName = ATMEDatas.currentModuleName,
				}
				
				if _isFlat ~= nil then
					datas.isFlat = _isFlat
				end
				
				if type(_altMin) == "number" then
					datas.altMin = _altMin
				end
				
				if localMembers.resultCallbacks[ATMEDatas.currentModuleName] == nil then
					ATMEFct.C_Zone_getRandomPosition(datas)
					return datas.point
				else
					if localMembers.coroutines[ATMEDatas.currentModuleName] == nil then
						ATMEFct.outputCore(nil, "newPolygon.getRandomPosition appel asynchrone", 2)
						localMembers.coroutines[ATMEDatas.currentModuleName] = coroutine.create(ATMEFct.C_Zone_getRandomPosition)
						coroutine.resume(localMembers.coroutines[ATMEDatas.currentModuleName], datas)
						localMembers.resultCallbacks[ATMEDatas.currentModuleName].idTimer = timer.scheduleFunction(ATMETimers.C_Zone_searchResultTimer, datas, timer.getTime() + 0.001)
						return
					else
						ATMEFct.setError("ALREADY_STARTED")
						ATMEFct.warningOutputCore(nil, fctDefinition, "search already started", false)
						return
					end
				end
			end
			
			ATMEFct.errorOutputCore(nil, fctDefinition, "surfaceFilter must be 'ALL', 'WATER' or 'LAND'", false)
		else
			if _surfaceFilter ~= nil and type(_surfaceFilter) ~= "string" then 
				ATMEFct.errorOutputCore(nil, fctDefinition, "surfaceFilter type must be nil or string", true)
			elseif _isFlat ~= nil and type(_isFlat) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "flat type must be nil or boolean", false)
			elseif _altMin ~= nil and type(_altMin) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "altMin type must be nil or number", false)
			end
		end
	end
	
	function newPolygon.addTrackingCallback(self, _fct, _object)
		local fctDefinition = ATMEFct.getErrorPrefix(newPolygon) .. "addTrackingCallback(fct, object)"

		if type(_fct) == "function" and (ATMEFct.isObjectClass(_object, ATME.C_Unit) == true or ATMEFct.isObjectClass(_object, ATME.C_Weapon) == true) then
			local className = _object:getClassName()
			local name = _object:getName()
			if localMembers.tracking[name] == nil then
				localMembers.tracking[name] = {
					className = className,
					name = name,
					object = _object,
					localMembersObject = ATMEDatas.localMembers[_object:getIndex()],
					callbacks = {},
				}
			end
				
			localMembers.tracking[name].pos = _object:getPosition()
			
			localMembers.tracking[name].callbacks[ATMEDatas.currentModuleName] = _fct
			
			local datas = {
				name = name,
				localMembers = localMembers,
				zone = self,
			}
			
			timer.scheduleFunction(ATMETimers.C_Zone_timer, datas, timer.getTime() + ATMEConst.offsetTracking[className])
		else
			if type(_fct) ~= "function" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
			elseif ATMEFct.isObjectClass(_object, ATME.C_Unit) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "object type must be ATME.C_Unit or ATME.C_Weapon", false)
			end
		end
	end
	
	function newPolygon.removeTrackingCallback(self, _object)
		local fctDefinition = ATMEFct.getErrorPrefix(newPolygon) .. "removeTrackingCallback(object)"
		
		if ATMEFct.isObjectClass(_object, ATME.C_Unit) == true or ATMEFct.isObjectClass(_object, ATME.C_Weapon) == true then
			if localMembers.tracking[_object:getName()] ~= nil then
				localMembers.tracking[_object:getName()].callbacks[ATMEDatas.currentModuleName] = nil
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "object type must be ATME.C_Unit or ATME.C_Weapon", true)
		end
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return newPolygon
end


ATMEDatas.C_Polygon_MT.__call = function(t, _vertices)
	local fctDefinition = "Constructor :\nATME.C_Polygon(vertices)\nATME.C_Polygon(DCSZoneName)"
	
	if ATMEDatas.startMissionOk < 2 then
		ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
		return
	end
	
	if type(_vertices) == "table" or type(_vertices) == "string" then
		local pErr
		local points = {}
		if type(_vertices) == "table" then
			for _key, _value in ipairs(_vertices) do
				if pErr == nil then
					local point
					
					if ATME.isPoint(_value) ~= POINT_BAD then
						point = ATMEFct.convertToPoint3D(_value)
					elseif type(_value) == "string" then
						local zone = ATMEFct.getDCSZone(_value)
						
						if zone ~= nil then
							point = zone.point
						else
							pErr = "Bad zone in item key " .. _key .. " : DCS Zone " .. _value .. " doesn't exist"
						end
					else
						pErr = "Bad type for item key " .. _key
					end
					
					if pErr == nil then
						table.insert(points, point)
					end
				end
			end
		else
			-- Le '-' devient un séparateur suite à modif DCS, report ici pour uniformisation (" #%03d" remplacé par "-%d")
			for _i = 1, ATMEConst.polygonMaxVertices do
				local currentName = _vertices .. string.format("-%d", _i)
				local zone = ATMEFct.getDCSZone(currentName)
				
				if zone ~= nil then
					table.insert(points, zone.point)
				end
			end
		end
		
		if pErr ~= nil then
			ATMEFct.errorOutputCore(nil, fctDefinition, pErr, false)
		else
			if #points > 2 and #points <= ATMEConst.polygonMaxVertices then
			
				-- Vérification de polygone croisé
				local intersect = false
				
				for _i = 1, #points - 1 do
					for _j = _i + 1, #points - 1 do
						local nr = (points[_i].z - points[_j].z) * (points[_j + 1].x - points[_j].x) - (points[_i].x - points[_j].x) * (points[_j + 1].z - points[_j].z)
						local ns = (points[_i].z - points[_j].z) * (points[_i + 1].x - points[_i].x) - (points[_i].x - points[_j].x) * (points[_i + 1].z - points[_i].z)
						local d = (points[_i + 1].x - points[_i].x) * (points[_j + 1].z - points[_j].z) - (points[_i + 1].z - points[_i].z) * (points[_j + 1].x - points[_j].x)

						if d ~= 0 then
							local r = nr / d
							local s = ns / d
							if r > 0 and r < 1 and s > 0 and s < 1 then
								intersect = true
							end
						end
					end
				end


				if intersect == false then
					local pErr
					-- Vérification des distances entre points consécutifs
					for _i = 1, #points - 1 do
						for _j = _i + 1, #points do
							local dX2 = (points[_i].x - points[_j].x) * (points[_i].x - points[_j].x)
							local dY2 = (points[_i].z - points[_j].z) * (points[_i].z - points[_j].z)
							local dist2 = dX2 + dY2

							if dist2 < 1 then
								-- Points trop proches distance < 1
								
								pErr = {
									first = _i,
									last = _j,
								}
							end
						end
					end

					if pErr == nil then
						return ATMEFct.C_Polygon_create(nil, nil, points)
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Bad polygon : Vertex " .. pErr.first .. " and " .. pErr.last .. " too near (distance < 1m)", false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "Bad polygon : crossed vertices", false)
				end
			else
				if #points < 3 then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Bad polygon : Less than 3 vertices", false)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "Bad polygon : More than limit of " .. ATMEConst.polygonMaxVertices .. " vertices", false)
				end
			end
		end
	else
		ATMEFct.errorOutputCore(nil, fctDefinition, "vertices type must be table or DCSZoneName type must be string", false)
	end
end

setmetatable(ATME.C_Polygon, ATMEDatas.C_Polygon_MT)

ATMEDatas.C_Polygon_MT.mt.__metatable = "Private"
ATMEDatas.C_Polygon_MT.mt.__index = ATME.C_Polygon

ATMEDatas.C_Polygon_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Polygon_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Polygon : Bad key " .. k, "table locked", false)
	end
end

ATMEFct.C_Rectangle_isInside = function(_point, _datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	local dx = _point.x - _datas.actualCenter.x
	local dz = _point.z - _datas.actualCenter.z
	
	local angleRad = math.rad(_datas.actualAzimuth)
	
	local lPoint = {
		x = dx * math.cos(angleRad) + dz * math.sin(angleRad),
		y = -dx * math.sin(angleRad) + dz * math.cos(angleRad),
	}
	
	local ok = (math.abs(lPoint.x) >= localMembers.halfLengthMin and math.abs(lPoint.x) <= localMembers.halfLength and 
					math.abs(lPoint.y) <= localMembers.halfWidth and localMembers.isRounded == false) or
			 (math.abs(lPoint.y) >= localMembers.halfWidthMin and math.abs(lPoint.y) <= localMembers.halfWidth and 
					math.abs(lPoint.x) <= localMembers.halfLength)
					
	if ok == false and localMembers.isRounded == true then	
		local dxPoint = lPoint.x - localMembers.halfLength
		local dyPoint = lPoint.y
		local d = math.sqrt(dxPoint * dxPoint + dyPoint * dyPoint)
		local angle = math.atan2(dyPoint, dxPoint)
	
		ok = angle > -math.pi / 2 and angle < math.pi / 2 and d >= localMembers.halfWidthMin and d <= localMembers.halfWidth
		
		if ok == false then
			dxPoint = lPoint.x + localMembers.halfLength
			local d = math.sqrt(dxPoint * dxPoint + dyPoint * dyPoint)
			local angle = math.atan2(dyPoint, dxPoint)
		
			ok = (angle < -math.pi / 2 or angle > math.pi / 2) and d >= localMembers.halfWidthMin and d <= localMembers.halfWidth
		end
	end

	return ok   
end

ATMEFct.C_Rectangle_getRandomPosition = function(_datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	local angleRad = math.rad(_datas.actualAzimuth)
	local point
	
	local rand = math.random()
	if rand < localMembers.probLength then
		-- Soit la surface correspondant au grand coté avec l'épaisseur
		
		local dx = ATMEConst.randomSign[math.random(1,2)] * localMembers.halfLength * math.random()
		local dy = ATMEConst.randomSign[math.random(1,2)] * (localMembers.deltaWidth * math.random() + localMembers.halfWidthMin)
		
		
		point = {
				x = _datas.actualCenter.x + dx * math.cos(angleRad) - dy * math.sin(angleRad),
				y = _datas.actualCenter.z + dx * math.sin(angleRad) + dy * math.cos(angleRad),
		}
	else 
		-- Soit la surface correspondant au petit coté moins l'épaisseur
		-- Soit sur 2x le demi-cercle de part et d'autre du centre sur l'axe x.
		
		if localMembers.isRounded == true then
			local angle = 2 * math.pi * math.random()
			local radius = localMembers.deltaWidth * math.random() + localMembers.halfWidthMin
			local i = 1
			
			if angle >= math.pi then
				i = -1
			end
			
			local localPoint = {
				x = i * localMembers.halfLength + radius * math.sin(angle),
				y = radius * math.cos(angle),
			}
			
			point = {
				x = _datas.actualCenter.x + localPoint.x * math.cos(angleRad) - localPoint.y * math.sin(angleRad),
				y = _datas.actualCenter.z + localPoint.x * math.sin(angleRad) + localPoint.y * math.cos(angleRad),
			}
		else
			local dx = ATMEConst.randomSign[math.random(1,2)] * (localMembers.deltaLength * math.random() + localMembers.halfLengthMin)
			local dy = ATMEConst.randomSign[math.random(1,2)] * localMembers.halfWidthMin * math.random()
			
			point = {
				x = _datas.actualCenter.x + dx * math.cos(angleRad) - dy * math.sin(angleRad),
				y = _datas.actualCenter.z + dx * math.sin(angleRad) + dy * math.cos(angleRad),
			}
		end
	end
	
	return ATMEFct.convertToPoint3D(point)
end

ATMEFct.C_Rectangle_searchVolume = function(_datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	local lPoints = {}
	local points = {}
	
	local angleRad = math.rad(_datas.actualAzimuth)
	
	local length = localMembers.halfLength
	if localMembers.isRounded == true then
		length = length + localMembers.halfWidth
	end

	table.insert(lPoints, {x = length, y = localMembers.halfWidth})
	table.insert(lPoints, {x = -length, y = localMembers.halfWidth})
	table.insert(lPoints, {x = -length, y = -localMembers.halfWidth})
	table.insert(lPoints, {x = length, y = -localMembers.halfWidth})

	for _index, _point in ipairs(lPoints) do
		points[_index] =  {
			x = _datas.actualCenter.x + _point.x * math.cos(angleRad) - _point.y * math.sin(angleRad),
			y = _datas.actualCenter.z + _point.x * math.sin(angleRad) + _point.y * math.cos(angleRad),
		}
	end
	
	local minPoint = {
		x = points[1].x,
		y = 0,
		z = points[1].y,
	}
		
	local maxPoint = {
		x = points[1].x,
		y = 200000,
		z = points[1].y,
	}
	
	for _, _point in ipairs(points) do
		if minPoint.x > _point.x then
			minPoint.x = _point.x
		end
		
		if minPoint.z > _point.y then
			minPoint.z = _point.y
		end
		
		if maxPoint.x < _point.x then
			maxPoint.x = _point.x
		end
		
		if maxPoint.z < _point.y then
			maxPoint.z = _point.y
		end
	end
	
	return 	world.VolumeType.BOX,
			"C_FindContext_search",
			{
				min = minPoint,
				max = maxPoint,
			}
end

ATMEDatas.C_Rectangle_MT = {
	className = "C_Rectangle",
	parentClass = ATME.C_MovableZone2D,
	
	mt = {},
	
	keys = {
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["isInside"] = true,
		["getRandomPosition"] = true,
		["addTrackingCallback"] = true,
		["removeTrackingCallback"] = true,
		["addResultCallback"] = true,
		["removeResultCallback"] = true,
		["abortResultCallback"] = true,
		["getIndex"] = true,
		["getUserPrivateDatas"] = true,
		["getLength"] = true,
		["getWidth"] = true,
		["getAzimuth"] = true,
		["getCenter"] = true,
		["changeThickness"] = true,
		["getThickness"] = true,
		["isRounded"] = true,
		["bindUnit"] = true,
		["getBoundUnit"] = true,
	},
	
	add = function(self, _center, _azimuth, _length, _width, _isRounded)
		local item = {
			center = _center,
			azimuth = _azimuth,
			length = _length,
			width = _width,
			halfLength = _length / 2,
			halfWidth = _width / 2,
			halfLengthMin = 0,
			halfWidthMin = 0,
			thickness = 0,
			isRounded = _isRounded,
			randomRatio = _length * _width,
			is3D = false,
			userPrivateDatas = {},
			tracking = {},
			resultCallbacks = {},
			coroutines = {},
		}
		
		-- item.box est initialisé dans C_Rectangle_create
		
		item.thicknessMax = math.min(item.halfLength, item.halfWidth)
		item.deltaLength = item.halfLength
		item.deltaWidth = item.halfWidth
		
		
		local sCircle = 0
		if _isRounded == true then
			sCircle = math.pi * (item.halfWidth * item.halfWidth)
		end
		
		local sLength = item.deltaWidth * item.length * 2
		local sWidth = item.deltaLength * (item.width - 2 * item.deltaWidth) * 2
		
		local sTotal = sCircle + sLength + sWidth
		
		item.probLength = sLength/sTotal
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].tracking = nil
		ATMEDatas.localMembers[_id].box = nil
		ATMEDatas.localMembers[_id].resultCallbacks = nil
		ATMEDatas.localMembers[_id].coroutines = nil
	end,
}

ATME.C_Rectangle = {
	getClassName = function(self)
		return ATMEDatas.C_Rectangle_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Rectangle_MT.parentClass
	end,
	
	createNamed = function(_name, _reference, _azimuth, _length, _width, _isRounded)
		local fctDefinition = "Function :\nATME.C_Rectangle.createNamed(name, reference, azimuth, length, width, [isRounded])"
		
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" and ATMEFct.isReference(_reference) == true and type(_azimuth) == "number" and type(_length) == "number" and type(_width) == "number" and
						(_isRounded == nil or type(_isRounded) == "boolean") then
			local center = ATMEFct.getPointFromReference(_reference)
			if _length > 0 then
				if _width > 0 then
					local isRounded = false
					if type(_isRounded) == "boolean" then
						isRounded = _isRounded
					end
					
					local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
					local err, newRectangle = ATMEFct.C_Zone_createNamed(localMembersModule.zones, _name, ATMEFct.C_Rectangle_create, center, ATMEFct.getAzimuthMod(_azimuth, 360), _length, _width, isRounded)				
			
					if err == true then
						ATMEFct.warningOutputCore(nil, fctDefinition, newRectangle, false)
						return
					end

					return newRectangle
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "width must be > 0", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "length must be > 0", false)
			end
		else
			if type(_name) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
			elseif ATMEFct.isReference(_reference) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
			elseif type(_azimuth) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "azimuth type must be number", false)
			elseif type(_length) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "length type must be number", false)
			elseif type(_width) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "width type must be number", false)
			elseif _isRounded ~= nil and type(_isRounded) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "isRounded type must be nil or boolean", false)
			end
		end
	end,
	
	getByName = function(_name)
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Rectangle.getByName(name)", "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			if ATMEFct.isObjectClass(ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name], ATME.C_Rectangle) == true then
				return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name]
			end
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Rectangle.getByName(name)", "name type must be string", false)
		end
	end,
}

ATMEFct.C_Rectangle_create = function(_name, _zones, _center, _heading, _length, _width, _isRounded)
	local newRectangle = {}

	setmetatable(newRectangle, ATMEDatas.C_Rectangle_MT.mt)
	local index = ATMEDatas.C_Rectangle_MT:add(_center, _heading, _length, _width, _isRounded)
	local localMembers = ATMEDatas.localMembers[index]

	local name
	local tagsName

	if _name ~= nil and _zones ~= nil then
		name = _name
		_zones[name] = newRectangle
		tagsName = ATMEFct.decompString(_name)
	end

	function newRectangle.getIndex(self)
		return index
	end
	
	function newRectangle.getName(self)
		return name
	end
	
	function newRectangle.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newRectangle) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newRectangle.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newRectangle.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newRectangle) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newRectangle.getLength(self)
		return localMembers.length
	end
	
	function newRectangle.getWidth(self)
		return localMembers.width
	end
	
	function newRectangle.getAzimuth(self)
		return ATMEFct.C_Zone_getActualAzimuth(localMembers)
	end

	function newRectangle.getCenter(self)
		return ATMEFct.C_Zone_getActualCenter(localMembers)
	end
	
	function newRectangle.bindUnit(self, _unit, _azimuth, _distance, _deltaAltitude)
		local fctDefinition = ATMEFct.getErrorPrefix(newRectangle) .. "bindUnit()\nbindUnit(unit, [azimuth, distance, [deltaAltitude]])"
		
		if _unit == nil or ATMEFct.isObjectClass(_unit, ATME.C_Unit) == true then
			if _unit == nil then
				localMembers.boundUnit = nil
			else
				local localMembersUnit = ATMEDatas.localMembers[_unit:getIndex()]
			
				if localMembersUnit.objectDCS:isExist() == true then
					localMembers.boundUnit = {
						unit = _unit,
						azimuthRad = 0,
						distance = 0,
						deltaAltitude = 0
					}
					
					if (_azimuth == nil and _distance == nil) or (type(_azimuth) == "number" and type(_distance) == "number") then
						local azimuth = 0
						local distance = 0
						
						if _distance ~= nil then
							distance = _distance
							azimuth = _azimuth
						end
						
						if distance >= 0 then
							if _deltaAltitude == nil or type(_deltaAltitude) == "number" then
								localMembers.boundUnit.azimuthRad = math.rad(ATMEFct.getAzimuthMod(azimuth, 360))
								localMembers.boundUnit.distance = math.floor(distance)
								
								if type(_deltaAltitude) == "number" then
									localMembers.boundUnit.deltaAltitude = _deltaAltitude
								end
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "deltaAltitude type must be nil or number", false)
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "distance must be >= 0", false)
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "azimuth and distance type must be both nil or number", false)
					end
				else
					localMembers.boundUnit = nil
					ATMEFct.setError("ALREADY_DISABLED")
					ATMEFct.warningOutputCore(nil, fctDefinition, "bound unit already disabled", false)
				end
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "unit type must be nil or ATME.C_Unit", true)
		end
	end
	
	function newRectangle.getBoundUnit(self)
		if localMembers.boundUnit ~= nil then
			local localMembersUnit = ATMEDatas.localMembers[localMembers.boundUnit.unit:getIndex()]
			
			if localMembersUnit.objectDCS:isExist() == true then
				return localMembers.boundUnit.unit
			else
				localMembers.boundUnit = nil
			end
		end
	end
	
	function newRectangle.getUserPrivateDatas(self, _moduleName)
		local fctDefinition = ATMEFct.getErrorPrefix(newRectangle) .. "getUserPrivateDatas([moduleName])"
		local moduleName
		
		if _moduleName ~= nil then
			if type(_moduleName) == "string" then
				if localMembers.userPrivateDatas[_moduleName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
					return
				else
					moduleName = _moduleName
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
				return
			end
		else
			moduleName = ATMEDatas.currentModuleName
		end
		
		return localMembers.userPrivateDatas[moduleName]
	end
		
	function newRectangle.changeThickness(self, _thickness)
		local fctDefinition = ATMEFct.getErrorPrefix(newRectangle) .. "changeThickness([thickness])"
				
		if _thickness == nil or type(_thickness) == "number" then
			local thickness = 0
			if type(_thickness) == "number" then
				thickness = _thickness
			end
			
			if thickness >= localMembers.thicknessMax then
				thickness = 0
			elseif thickness < 0 then
				ATMEFct.errorOutputCore(nil, fctDefinition, "thickness must be >= 0", false)
				return
			end
			
			if thickness > 0 then
				localMembers.halfLengthMin = localMembers.halfLength - thickness
				localMembers.halfWidthMin = localMembers.halfWidth - thickness
			else
				localMembers.halfLengthMin = 0
				localMembers.halfWidthMin = 0
			end
			
			localMembers.thickness = thickness
			localMembers.deltaLength = localMembers.halfLength - localMembers.halfLengthMin
			localMembers.deltaWidth = localMembers.halfWidth - localMembers.halfWidthMin
			
			local sLength = localMembers.deltaWidth * localMembers.length * 2
			local sCircle = 0
			local sWidth = 0

			if localMembers.isRounded == true then
				sCircle = math.pi * (localMembers.halfWidth * localMembers.halfWidth - localMembers.halfWidthMin * localMembers.halfWidthMin)
				sWidth = 0
			else
				sCircle = 0
				sWidth = localMembers.deltaLength * (localMembers.width - 2 * localMembers.deltaWidth) * 2
			end
			
			local sTotal = sCircle + sLength + sWidth
			
			localMembers.probLength = sLength/sTotal

			localMembers.randomRatio = localMembers.length * localMembers.width - localMembers.halfLengthMin * localMembers.halfWidthMin * 4 + sCircle
		else
			if _thickness ~= nil and type(_thickness) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "thickness type must be nil or number", true)
			end
		end
	end
	
	function newRectangle.getThickness(self)	
		if localMembers.thickness ~= 0 then
			return localMembers.thickness
		end
	end
	
	function newRectangle.isRounded(self)
		return localMembers.isRounded
	end
	
	function newRectangle.isInside(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newRectangle) .. "isInside(reference)"
		local isInside = false
		
		if ATMEFct.isReference(_reference) == true then		
			local point = ATMEFct.getPointFromReference(_reference)
			
			local datas = {
				zone = self,
				actualCenter = ATMEFct.C_Zone_getActualCenter(localMembers),
				actualAzimuth = ATMEFct.C_Zone_getActualAzimuth(localMembers),
			}
			
			ATMEFct.C_Rectangle_isInside(point, datas)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
		end
		
		return isInside
	end
	
	function newRectangle.addResultCallback(self, _fct)
		local fctDefinition = ATMEFct.getErrorPrefix(newRectangle) .. "addResultCallback(fct)"

		if type(_fct) == "function" then
			localMembers.resultCallbacks[ATMEDatas.currentModuleName] = {
				fct = _fct,
				aborted = false,
			}
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
		end
	end
	
	function newRectangle.removeResultCallback(self)
		ATMEFct.C_Zone_abortResult(localMembers)
		localMembers.resultCallbacks[ATMEDatas.currentModuleName] = nil
	end
	
	function newRectangle.abortResultCallback(self)
		ATMEFct.C_Zone_abortResult(localMembers)
	end
	
	function newRectangle.getRandomPosition(self, _surfaceFilter, _isFlat, _altMin)
		local fctDefinition = ATMEFct.getErrorPrefix(newRectangle) .. "getRandomPosition([surfaceFilter], [flat], [altMin])"
	
		if (_surfaceFilter == nil or type(_surfaceFilter) == "string") and
					(_isFlat == nil or type(_isFlat) =="boolean") and (_altMin == nil or type(_altMin) =="number") then

			local surfaceFilter = "ALL"
			
			if _surfaceFilter ~= nil then
				surfaceFilter = string.upper(_surfaceFilter)
			end
			
			if ATMEConst.surfaceType[surfaceFilter] ~= nil then
				local datas = {
					zone = self,
					surfaceFilter = surfaceFilter,
					isFlat = false,
					altMin = 0,
					localMembers = localMembers,
					moduleName = ATMEDatas.currentModuleName,
				}
				
				if _isFlat ~= nil then
					datas.isFlat = _isFlat
				end
				
				if type(_altMin) == "number" then
					datas.altMin = _altMin
				end
				
				if localMembers.resultCallbacks[ATMEDatas.currentModuleName] == nil then
					ATMEFct.C_Zone_getRandomPosition(datas)
					return datas.point
				else
					if localMembers.coroutines[ATMEDatas.currentModuleName] == nil then
						ATMEFct.outputCore(nil, "newRectangle.getRandomPosition appel asynchrone", 2)
						localMembers.coroutines[ATMEDatas.currentModuleName] = coroutine.create(ATMEFct.C_Zone_getRandomPosition)
						coroutine.resume(localMembers.coroutines[ATMEDatas.currentModuleName], datas)
						localMembers.resultCallbacks[ATMEDatas.currentModuleName].idTimer = timer.scheduleFunction(ATMETimers.C_Zone_searchResultTimer, datas, timer.getTime() + 0.001)
						return
					else
						ATMEFct.setError("ALREADY_STARTED")
						ATMEFct.warningOutputCore(nil, fctDefinition, "search already started", false)
						return
					end
				end
			end
			
			ATMEFct.errorOutputCore(nil, fctDefinition, "surfaceFilter must be 'ALL', 'WATER' or 'LAND'", false)
		else
			if _surfaceFilter ~= nil and type(_surfaceFilter) ~= "string" then 
				ATMEFct.errorOutputCore(nil, fctDefinition, "surfaceFilter type must be nil or string", true)
			elseif _isFlat ~= nil and type(_isFlat) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "flat type must be nil or boolean", false)
			elseif _altMin ~= nil and type(_altMin) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "altMin type must be nil or number", false)
			end
		end
	end
	
	function newRectangle.addTrackingCallback(self, _fct, _object)
		local fctDefinition = ATMEFct.getErrorPrefix(newRectangle) .. "addTrackingCallback(fct, object)"

		if type(_fct) == "function" and (ATMEFct.isObjectClass(_object, ATME.C_Unit) == true or ATMEFct.isObjectClass(_object, ATME.C_Weapon) == true) then
			local className = _object:getClassName()
			local name = _object:getName()
			if localMembers.tracking[name] == nil then
				localMembers.tracking[name] = {
					className = className,
					name = name,
					object = _object,
					localMembersObject = ATMEDatas.localMembers[_object:getIndex()],
					callbacks = {},
				}
			end
				
			localMembers.tracking[name].pos = _object:getPosition()
			
			localMembers.tracking[name].callbacks[ATMEDatas.currentModuleName] = _fct
			
			local datas = {
				name = name,
				localMembers = localMembers,
				zone = self,
			}
			
			timer.scheduleFunction(ATMETimers.C_Zone_timer, datas, timer.getTime() + ATMEConst.offsetTracking[className])
		else
			if type(_fct) ~= "function" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
			elseif ATMEFct.isObjectClass(_object, ATME.C_Unit) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "object type must be ATME.C_Unit or ATME.C_Weapon", false)
			end
		end
	end
	
	function newRectangle.removeTrackingCallback(self, _object)
		local fctDefinition = ATMEFct.getErrorPrefix(newRectangle) .. "removeTrackingCallback(object)"
		
		if ATMEFct.isObjectClass(_object, ATME.C_Unit) == true or ATMEFct.isObjectClass(_object, ATME.C_Weapon) == true then
			if localMembers.tracking[_object:getName()] ~= nil then
				localMembers.tracking[_object:getName()].callbacks[ATMEDatas.currentModuleName] = nil
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "object type must be ATME.C_Unit or ATME.C_Weapon", true)
		end
	end
	
	local datas = {
		zone = newRectangle,
		actualCenter = _center,
		actualAzimuth = _heading,
	}
	

	local _, _, cbox = ATMEFct.C_Rectangle_searchVolume(datas)
	localMembers.box = cbox
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return newRectangle
end

ATMEDatas.C_Rectangle_MT.__call = function(t, _reference, _azimuth, _length, _width, _isRounded)
	local fctDefinition = "Constructor :\nATME.C_Rectangle(reference, azimuth, length, width, [isRounded])"
	
	if ATMEDatas.startMissionOk < 2 then
		ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
		return
	end
	
	if ATMEFct.isReference(_reference) == true and type(_azimuth) == "number" and type(_length) == "number" and type(_width) == "number" and
					(_isRounded == nil or type(_isRounded) == "boolean") then
		local center = ATMEFct.getPointFromReference(_reference)
		if _length > 0 then
			if _width > 0 then
				local isRounded = false
				if type(_isRounded) == "boolean" then
					isRounded = _isRounded
				end

				return ATMEFct.C_Rectangle_create(nil, nil, center, ATMEFct.getAzimuthMod(_azimuth, 360), _length, _width, isRounded)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "width must be > 0", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "length must be > 0", false)
		end
	else
		if ATMEFct.isReference(_reference) == false then
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
		elseif type(_azimuth) ~= "number" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "azimuth type must be number", false)
		elseif type(_length) ~= "number" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "length type must be number", false)
		elseif type(_width) ~= "number" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "width type must be number", false)
		elseif _isRounded ~= nil and type(_isRounded) ~= "boolean" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "isRounded type must be nil or boolean", false)
		end
	end
end

setmetatable(ATME.C_Rectangle, ATMEDatas.C_Rectangle_MT)

ATMEDatas.C_Rectangle_MT.mt.__metatable = "Private"
ATMEDatas.C_Rectangle_MT.mt.__index = ATME.C_Rectangle

ATMEDatas.C_Rectangle_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Rectangle_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Rectangle : Bad key " .. k, "table locked", false)
	end
end

ATMEFct.C_Sphere_isInside = function(_point, _datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	
	local dx2 = (_point.x - _datas.actualCenter.x) * (_point.x - _datas.actualCenter.x)
	local dy2 = (_point.y - _datas.actualCenter.y) * (_point.y - _datas.actualCenter.y)
	local dz2 = (_point.z - _datas.actualCenter.z) * (_point.z - _datas.actualCenter.z)
	local d = math.sqrt(dx2 + dy2 + dz2) 
	
	return d >= localMembers.radiusMin and d <= localMembers.radiusMax
end

ATMEFct.C_Sphere_getRandomPosition = function(_datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	local angle1 = math.pi - 2 * math.pi * math.random()
	local angle2 = math.pi / 2 - math.pi * math.random()
	local radius = (localMembers.radiusMax - localMembers.radiusMin) * math.random() + localMembers.radiusMin

	return {
		x = _datas.actualCenter.x + radius * math.cos(angle1) * cos(angle2),
		y = _datas.actualCenter.y + radius * math.sin(angle2), 
		z = _datas.actualCenter.z + radius * math.sin(angle1) * cos(angle2),
	}
end

ATMEFct.C_Sphere_searchVolume = function(_datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	
	return 	world.VolumeType.SPHERE, 
			"C_FindContext_search",
			{
				point = {
					x = _datas.actualCenter.x,
					y = _datas.actualCenter.y,
					z = _datas.actualCenter.z,
				},
				radius = localMembers.radiusMax,
			}
end

ATMEDatas.C_Sphere_MT = {
	className = "C_Sphere",
	parentClass = ATME.C_MovableZone3D,
	
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["getUserPrivateDatas"] = true,
		["isInside"] = true,
		["getRandomPosition"] = true,
		["addTrackingCallback"] = true,
		["removeTrackingCallback"] = true,
		["addResultCallback"] = true,
		["removeResultCallback"] = true,
		["abortResultCallback"] = true,
		["getCenter"] = true,
		["getRadius"] = true,
		["changeThickness"] = true,
		["getThickness"] = true,
		["bindUnit"] = true,
		["getBoundUnit"] = true,
	},
	
	add = function(self, _center, _radiusMax)
		local item = {
			center = _center,
			radiusMax = _radiusMax,
			radiusMin = 0,
			thickness = 0,
			randomRatio = math.pi * _radiusMax * _radiusMax,
			is3D = true,
			sphere = {
				point = _center,
				radius = _radiusMax,
			},
			box = {
				min = {
					x = _center.x - _radiusMax,
					y = _center.y - _radiusMax,
					z = _center.z - _radiusMax,
				},
				max = {
					x = _center.x + _radiusMax,
					y = _center.y + _radiusMax,
					z = _center.z + _radiusMax,
				},
			},
			userPrivateDatas = {},
			tracking = {},
			resultCallbacks = {},
			coroutines = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].tracking = nil
		ATMEDatas.localMembers[_id].box = nil
		ATMEDatas.localMembers[_id].resultCallbacks = nil
		ATMEDatas.localMembers[_id].coroutines = nil
	end,
}

ATME.C_Sphere = {
	getClassName = function(self)
		return ATMEDatas.C_Sphere_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Sphere_MT.parentClass
	end,
	
	createNamed = function(_name, ...)
		local args = {...}
		local fctDefinition = "Function :\nATME.C_Sphere.createNamed(name, DCSZoneName, [altitudeMSL])\nATME.C_Sphere.createNamed(name, reference, radius)"
		local point, radius
		
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			if type(args[1]) == "string" then
				local zone = ATMEFct.getDCSZone(args[1])
				
				if zone ~=  nil then
					local altitude = zone.point.y
					
					if #args > 1 then
						if type(args[2]) == "number" then
							altitude = args[2]
							
							if altitude < 0 then
								ATMEFct.errorOutputCore(nil, fctDefinition, "altitude must be >= 0", false)
								return
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "altitude type must be nil or number", false)
							return
						end
					end
					
					point = {
						x = zone.point.x,
						y = altitude,
						z = zone.point.z,
					}

					radius = zone.radius
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone '" .. args[1] .. "' unknown", false)
					return
				end
			elseif ATMEFct.isReference(args[1]) == true then
				if type(args[2]) == "number" then
					radius = args[2]
					if radius > 0 then
						point = ATMEFct.getPointFromReference(args[1])
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "radius must be > 0", false)
						return
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "radius type must be number", false)
					return
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 type must be string," .. ATMEConst.errorLabelReference, false)
				return
			end
			
			local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
			local err, newSphere = ATMEFct.C_Zone_createNamed(localMembersModule.zones, _name, ATMEFct.C_Sphere_create, point, radius)				
		
			if err == true then
				ATMEFct.warningOutputCore(nil, fctDefinition, newSphere, false)
				return
			end

			return newSphere
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
		end
	end,
	
	getByName = function(_name)
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Sphere.getByName(name)", "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			if ATMEFct.isObjectClass(ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name], ATME.C_Sphere) == true then
				return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name]
			end
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Sphere.getByName(name)", "name type must be string", false)
		end
	end,
}

ATMEFct.C_Sphere_create = function(_name, _zones, _center, _radiusMax)
	local newSphere = {}

	setmetatable(newSphere, ATMEDatas.C_Sphere_MT.mt)
	local index = ATMEDatas.C_Sphere_MT:add(_center, _radiusMax)
	local localMembers = ATMEDatas.localMembers[index]

	local name
	local tagsName

	if _name ~= nil and _zones ~= nil then
		name = _name
		_zones[name] = newSphere
		tagsName = ATMEFct.decompString(_name)
	end
	
	function newSphere.getIndex(self)
		return index
	end
	
	function newSphere.getName(self)
		return name
	end
	
	function newSphere.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newSphere) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newSphere.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newSphere.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newSphere) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end

	function newSphere.changeThickness(self, _thickness)
		local fctDefinition = ATMEFct.getErrorPrefix(newSphere) .. "changeThickness([thickness])"
				
		if _thickness == nil or type(_thickness) == "number" then
			local thickness = 0
			if type(_thickness) == "number" then
				thickness = _thickness
			end
			
			if thickness >= localMembers.radiusMax then
				thickness = 0
			elseif thickness < 0 then
				ATMEFct.errorOutputCore(nil, fctDefinition, "thickness must be >= 0", false)
				return
			end
			
			localMembers.thickness = thickness
			
			if thickness == 0 then
				localMembers.radiusMin = 0
			else
				localMembers.radiusMin = localMembers.radiusMax - thickness
			end
			
			localMembers.randomRatio = math.pi * (localMembers.radiusMax * localMembers.radiusMax - localMembers.radiusMin * localMembers.radiusMin)
		else
			if _thickness ~= nil and type(_thickness) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "thickness type must be nil or number", true)
			end
		end
	end
	
	function newSphere.getThickness(self)	
		if localMembers.thickness ~= 0 then
			return localMembers.thickness
		end
	end
	
	function newSphere.getUserPrivateDatas(self, _moduleName)
		local fctDefinition = ATMEFct.getErrorPrefix(newSphere) .. "getUserPrivateDatas([moduleName])"
		local moduleName
		
		if _moduleName ~= nil then
			if type(_moduleName) == "string" then
				if localMembers.userPrivateDatas[_moduleName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
					return
				else
					moduleName = _moduleName
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
				return
			end
		else
			moduleName = ATMEDatas.currentModuleName
		end
		
		return localMembers.userPrivateDatas[moduleName]
	end
	
	function newSphere.isInside(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newSphere) .. "isInside(reference)"
		
		if ATMEFct.isReference(_reference) == true then
			local point = ATMEFct.getPointFromReference(_reference)
			
			local datas = {
				zone = self,
				actualCenter = ATMEFct.C_Zone_getActualCenter(localMembers),
				actualAzimuth = ATMEFct.C_Zone_getActualAzimuth(localMembers),
			}
			
			return ATMEFct.C_Sphere_isInside(point, datas)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
		end
		
		return false
	end
	
	function newSphere.addResultCallback(self, _fct)
		local fctDefinition = ATMEFct.getErrorPrefix(newSphere) .. "addResultCallback(fct)"

		if type(_fct) == "function" then
			localMembers.resultCallbacks[ATMEDatas.currentModuleName] = {
				fct = _fct,
				aborted = false,
			}
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
		end
	end
	
	function newSphere.removeResultCallback(self)
		ATMEFct.C_Zone_abortResult(localMembers)
		localMembers.resultCallbacks[ATMEDatas.currentModuleName] = nil
	end
	
	function newSphere.abortResultCallback(self)
		ATMEFct.C_Zone_abortResult(localMembers)
	end
	
	function newSphere.getRandomPosition(self, _surfaceFilter, _isFlat, _altMin)
		local fctDefinition = ATMEFct.getErrorPrefix(newSphere) .. "getRandomPosition([surfaceFilter], [flat], [altMin])"
	
		if (_surfaceFilter == nil or type(_surfaceFilter) == "string") and
					(_isFlat == nil or type(_isFlat) =="boolean") and (_altMin == nil or type(_altMin) =="number") then

			local surfaceFilter = "ALL"
			
			if _surfaceFilter ~= nil then
				surfaceFilter = string.upper(_surfaceFilter)
			end
			
			if ATMEConst.surfaceType[surfaceFilter] ~= nil then
				local datas = {
					zone = self,
					surfaceFilter = surfaceFilter,
					isFlat = false,
					altMin = 0,
					localMembers = localMembers,
					moduleName = ATMEDatas.currentModuleName,
				}
				
				if _isFlat ~= nil then
					datas.isFlat = _isFlat
				end
				
				if type(_altMin) == "number" then
					datas.altMin = _altMin
				end
				
				if localMembers.resultCallbacks[ATMEDatas.currentModuleName] == nil then
					ATMEFct.C_Zone_getRandomPosition(datas)
					return datas.point
				else
					if localMembers.coroutines[ATMEDatas.currentModuleName] == nil then
						ATMEFct.outputCore(nil, "newSphere.getRandomPosition appel asynchrone", 2)
						localMembers.coroutines[ATMEDatas.currentModuleName] = coroutine.create(ATMEFct.C_Zone_getRandomPosition)
						coroutine.resume(localMembers.coroutines[ATMEDatas.currentModuleName], datas)
						localMembers.resultCallbacks[ATMEDatas.currentModuleName].idTimer = timer.scheduleFunction(ATMETimers.C_Zone_searchResultTimer, datas, timer.getTime() + 0.001)
						return
					else
						ATMEFct.setError("ALREADY_STARTED")
						ATMEFct.warningOutputCore(nil, fctDefinition, "search already started", false)
						return
					end
				end
			end
			
			ATMEFct.errorOutputCore(nil, fctDefinition, "surfaceFilter must be 'ALL', 'WATER' or 'LAND'", false)
		else
			if _surfaceFilter ~= nil and type(_surfaceFilter) ~= "string" then 
				ATMEFct.errorOutputCore(nil, fctDefinition, "surfaceFilter type must be nil or string", true)
			elseif _isFlat ~= nil and type(_isFlat) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "flat type must be nil or boolean", false)
			elseif _altMin ~= nil and type(_altMin) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "altMin type must be nil or number", false)
			end
		end
	end
	
	function newSphere.getRadius(self)
		return localMembers.radiusMax
	end
	
	function newSphere.getCenter(self)
		return ATMEFct.C_Zone_getActualCenter(localMembers)
	end
	
	function newSphere.bindUnit(self, _unit, _azimuth, _distance, _deltaAltitude)
		local fctDefinition = ATMEFct.getErrorPrefix(newSphere) .. "bindUnit()\nbindUnit(unit, [azimuth, distance, [deltaAltitude]])"
		
		if _unit == nil or ATMEFct.isObjectClass(_unit, ATME.C_Unit) == true then
			if _unit == nil then
				localMembers.boundUnit = nil
			else
				local localMembersUnit = ATMEDatas.localMembers[_unit:getIndex()]
			
				if localMembersUnit.objectDCS:isExist() == true then
					localMembers.boundUnit = {
						unit = _unit,
						azimuthRad = 0,
						distance = 0,
						deltaAltitude = 0
					}
					
					if (_azimuth == nil and _distance == nil) or (type(_azimuth) == "number" and type(_distance) == "number") then
						local azimuth = 0
						local distance = 0
						
						if _distance ~= nil then
							distance = _distance
							azimuth = _azimuth
						end
						
						if distance >= 0 then
							if _deltaAltitude == nil or type(_deltaAltitude) == "number" then
								localMembers.boundUnit.azimuthRad = math.rad(ATMEFct.getAzimuthMod(azimuth, 360))
								localMembers.boundUnit.distance = math.floor(distance)
								
								if type(_deltaAltitude) == "number" then
									localMembers.boundUnit.deltaAltitude = _deltaAltitude
								end
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "deltaAltitude type must be nil or number", false)
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "distance must be >= 0", false)
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "azimuth and distance type must be both nil or number", false)
					end
				else
					localMembers.boundUnit = nil
					ATMEFct.setError("ALREADY_DISABLED")
					ATMEFct.warningOutputCore(nil, fctDefinition, "bound unit already disabled", false)
				end
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "unit type must be nil or ATME.C_Unit", true)
		end
	end
	
	function newSphere.getBoundUnit(self)
		if localMembers.boundUnit ~= nil then
			local localMembersUnit = ATMEDatas.localMembers[localMembers.boundUnit.unit:getIndex()]
			
			if localMembersUnit.objectDCS:isExist() == true then
				return localMembers.boundUnit.unit
			else
				localMembers.boundUnit = nil
			end
		end
	end
	
	function newSphere.addTrackingCallback(self, _fct, _object)
		local fctDefinition = ATMEFct.getErrorPrefix(newSphere) .. "addTrackingCallback(fct, object)"

		if type(_fct) == "function" and (ATMEFct.isObjectClass(_object, ATME.C_Unit) == true or ATMEFct.isObjectClass(_object, ATME.C_Weapon) == true) then
			local className = _object:getClassName()
			local name = _object:getName()
			if localMembers.tracking[name] == nil then
				localMembers.tracking[name] = {
					className = className,
					name = name,
					object = _object,
					localMembersObject = ATMEDatas.localMembers[_object:getIndex()],
					callbacks = {},
				}
			end
				
			localMembers.tracking[name].pos = _object:getPosition()
			
			localMembers.tracking[name].callbacks[ATMEDatas.currentModuleName] = _fct
			
			local datas = {
				name = name,
				localMembers = localMembers,
				zone = self,
			}
			
			timer.scheduleFunction(ATMETimers.C_Zone_timer, datas, timer.getTime() + ATMEConst.offsetTracking[className])
		else
			if type(_fct) ~= "function" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
			elseif ATMEFct.isObjectClass(_object, ATME.C_Unit) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "object type must be ATME.C_Unit or ATME.C_Weapon", false)
			end
		end
	end
	
	function newSphere.removeTrackingCallback(self, _object)
		local fctDefinition = ATMEFct.getErrorPrefix(newSphere) .. "removeTrackingCallback(object)"
		
		if ATMEFct.isObjectClass(_object, ATME.C_Unit) == true or ATMEFct.isObjectClass(_object, ATME.C_Weapon) == true then
			if localMembers.tracking[_object:getName()] ~= nil then
				localMembers.tracking[_object:getName()].callbacks[ATMEDatas.currentModuleName] = nil
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "object type must be ATME.C_Unit or ATME.C_Weapon", true)
		end
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return newSphere
end


ATMEDatas.C_Sphere_MT.__call = function(t, ...)
	local args = {...}
	local fctDefinition = "Constructor :\nATME.C_Sphere(DCSZoneName, [altitudeMSL])\nATME.C_Sphere(reference, radius)"
	local point, radius
	
	if ATMEDatas.startMissionOk < 2 then
		ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
		return
	end
	
	if type(args[1]) == "string" then
		local zone = ATMEFct.getDCSZone(args[1])
		
		if zone ~=  nil then
			local altitude = zone.point.y
			
			if #args > 1 then
				if type(args[2]) == "number" then
					altitude = args[2]
					
					if altitude < 0 then
						ATMEFct.errorOutputCore(nil, fctDefinition, "altitude must be >= 0", false)
						return
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "altitude type must be nil or number", false)
					return
				end
			end
			
			point = {
				x = zone.point.x,
				y = altitude,
				z = zone.point.z,
			}

			radius = zone.radius
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone '" .. args[1] .. "' unknown", false)
			return
		end
	elseif ATMEFct.isReference(args[1]) == true then
		if type(args[2]) == "number" then
			radius = args[2]
			if radius > 0 then
				point = ATMEFct.getPointFromReference(args[1])
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "radius must be > 0", false)
				return
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "radius type must be number", false)
			return
		end
	else
		ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #1 type must be string," .. ATMEConst.errorLabelReference, false)
		return
	end
	
	return ATMEFct.C_Sphere_create(nil, nil, point, radius)
end

setmetatable(ATME.C_Sphere, ATMEDatas.C_Sphere_MT)

ATMEDatas.C_Sphere_MT.mt.__metatable = "Private"
ATMEDatas.C_Sphere_MT.mt.__index = ATME.C_Sphere

ATMEDatas.C_Sphere_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Sphere_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Sphere : Bad key " .. k, "table locked", false)
	end
end


ATMEFct.C_Cylinder_isInside = function(_point, _datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]

	local dx2 = (_point.x - _datas.actualCenter.x) * (_point.x - _datas.actualCenter.x)
	local dy = _point.y - _datas.actualCenter.y
	local dz2 = (_point.z - _datas.actualCenter.z) * (_point.z - _datas.actualCenter.z)
	local d = math.sqrt(dx2 + dz2)
	
	return d >= localMembers.radiusMin and d <= localMembers.radiusMax and math.abs(dy) < localMembers.deltaAltitude
end

ATMEFct.C_Cylinder_getRandomPosition = function(_datas)
	local angle = 2 * math.pi * math.random()
	
	local radius = (localMembers.radiusMax - localMembers.radiusMin) * math.random() + localMembers.radiusMin
	local dAlt = ATMEConst.randomSign[math.random(1,2)] * localMembers.deltaAltitude * math.random()
	
	return {
		x = _datas.actualCenter.x + radius * math.sin(angle),
		y = _datas.actualCenter.y + dAlt,
		z = _datas.actualCenter.z + radius * math.cos(angle),
	}
end

ATMEFct.C_Cylinder_searchVolume = function(_datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	
	return world.VolumeType.BOX, 
			"C_FindContext_search",
			{
				min = {
					x = _datas.actualCenter.x - localMembers.radiusMax,
					y = _datas.actualCenter.y - localMembers.deltaAltitude,
					z = _datas.actualCenter.z - localMembers.radiusMax,
				},
				max = {
					x = _datas.actualCenter.x + localMembers.radiusMax,
					y = _datas.actualCenter.y + localMembers.deltaAltitude,
					z = _datas.actualCenter.z + localMembers.radiusMax,
				},
			}
end

ATMEDatas.C_Cylinder_MT = {
	className = "C_Cylinder",
	parentClass = ATME.C_MovableZone3D,
	
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["getUserPrivateDatas"] = true,
		["isInside"] = true,
		["getRandomPosition"] = true,
		["addTrackingCallback"] = true,
		["removeTrackingCallback"] = true,
		["addResultCallback"] = true,
		["removeResultCallback"] = true,
		["abortResultCallback"] = true,
		["getCenter"] = true,
		["getRadius"] = true,
		["getHeight"] = true,
		["changeThickness"] = true,
		["getThickness"] = true,
		["bindUnit"] = true,
		["getBoundUnit"] = true,
	},
	
	add = function(self, _center, _radiusMax, _deltaAltitude)
		local item = {
			center = _center,
			radiusMax = _radiusMax,
			radiusMin = 0,
			thickness = 0,
			deltaAltitude = _deltaAltitude,
			randomRatio = math.pi * _radiusMax * _radiusMax,
			is3D = true,
			box = {
				min = {
					x = _center.x - _radiusMax,
					y = _center.y - _deltaAltitude,
					z = _center.z - _radiusMax,
				},
				max = {
					x = _center.x + _radiusMax,
					y = _center.y + _deltaAltitude,
					z = _center.z + _radiusMax,
				},
			},
			userPrivateDatas = {},
			tracking = {},
			resultCallbacks = {},
			coroutines = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].tracking = nil
		ATMEDatas.localMembers[_id].box = nil
		ATMEDatas.localMembers[_id].resultCallbacks = nil
		ATMEDatas.localMembers[_id].coroutines = nil
	end,
}

ATME.C_Cylinder = {
	getClassName = function(self)
		return ATMEDatas.C_Cylinder_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Cylinder_MT.parentClass
	end,
	
	createNamed = function(_name, ...)
		local args = {...}
		local fctDefinition = "Function :\nATME.C_Cylinder.createNamed(name, DCSZoneName, height, [altitudeMSL])\nATME.C_Cylinder.createNamed(name, reference, radius, height)"
		local point, radius, deltaAltitude
		
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			if type(args[1]) == "string" then
				local zone = ATMEFct.getDCSZone(args[1])
				if zone ~=  nil then
					local altitude = zone.point.y
					
					if type(args[2]) ~= "number" then
						ATMEFct.errorOutputCore(nil, fctDefinition, "height type must be number", false)
						return
					end

					deltaAltitude = args[2] / 2
					if deltaAltitude <= 0 then
						ATMEFct.errorOutputCore(nil, fctDefinition, "height must be > 0", false)
						return
					end
					
					if #args > 2 then
						if type(args[3]) == "number" then
							if args[3] >= 0 then
								altitude = args[3]
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "altitude must be >= 0", false)
								return
							end
						else	
							ATMEFct.errorOutputCore(nil, fctDefinition, "altitude type must be nil or number", false)
							return
						end
					end
					
					point = {
						x = zone.point.x,
						y = altitude,
						z = zone.point.z,
					}

					radius = zone.radius
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone '" .. args[1] .. "' unknown", false)
					return
				end
			elseif ATMEFct.isReference(args[1]) == true then
				if type(args[2]) == "number" then
					radius = args[2]
					if radius > 0 then
						point = ATMEFct.getPointFromReference(args[1])
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "radius must be > 0", false)
						return
					end
					
					if type(args[3]) ~= "number" then
						ATMEFct.errorOutputCore(nil, fctDefinition, "deltaAltitude type must be number", false)
						return
					end
					
					deltaAltitude = args[3] / 2
					if deltaAltitude <= 0 then
						ATMEFct.errorOutputCore(nil, fctDefinition, "deltaAltitude must be > 0", false)
						return
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "radius type must be number", false)
					return
				end
			else

				ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 type must be string," .. ATMEConst.errorLabelReference, false)
				return
			end
			
			local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
			local err, newCylinder = ATMEFct.C_Zone_createNamed(localMembersModule.zones, _name, ATMEFct.C_Cylinder_create, point, radius, deltaAltitude)				
		
			if err == true then
				ATMEFct.warningOutputCore(nil, fctDefinition, newCylinder, false)
				return
			end

			return newCylinder
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
		end
	end,
	
	getByName = function(_name)	
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Cylinder.getByName(name)", "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			if ATMEFct.isObjectClass(ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name], ATME.C_Cylinder) == true then
				return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name]
			end
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Cylinder.getByName(name)", "name type must be string", false)
		end
	end,
}

ATMEFct.C_Cylinder_create = function(_name, _zones, _center, _radiusMax, _deltaAltitude)
	local newCylinder = {}
	
	setmetatable(newCylinder, ATMEDatas.C_Cylinder_MT.mt)
	local index = ATMEDatas.C_Cylinder_MT:add(_center, _radiusMax, _deltaAltitude)
	local localMembers = ATMEDatas.localMembers[index]

	local name
	local tagsName

	if _name ~= nil and _zones ~= nil then
		name = _name
		_zones[name] = newCylinder
		tagsName = ATMEFct.decompString(_name)
	end
	
	function newCylinder.getIndex(self)
		return index
	end
	
	function newCylinder.getName(self)
		return name
	end
	
	function newCylinder.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newCylinder) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newCylinder.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newCylinder.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newCylinder) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newCylinder.getUserPrivateDatas(self, _moduleName)
		local fctDefinition = ATMEFct.getErrorPrefix(newCylinder) .. "getUserPrivateDatas([moduleName])"
		local moduleName
		
		if _moduleName ~= nil then
			if type(_moduleName) == "string" then
				if localMembers.userPrivateDatas[_moduleName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
					return
				else
					moduleName = _moduleName
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
				return
			end
		else
			moduleName = ATMEDatas.currentModuleName
		end
		
		return localMembers.userPrivateDatas[moduleName]
	end
	
	function newCylinder.changeThickness(self, _thickness)
		local fctDefinition = ATMEFct.getErrorPrefix(newCylinder) .. "changeThickness([thickness])"
				
		if _thickness == nil or type(_thickness) == "number" then
			local thickness = 0
			if type(_thickness) == "number" then
				thickness = _thickness
			end
			
			
			if thickness >= localMembers.radiusMax then
				thickness = 0
			elseif thickness < 0 then
				ATMEFct.errorOutputCore(nil, fctDefinition, "thickness must be >= 0", false)
				return
			end
			
			localMembers.thickness = thickness
			
			if thickness == 0 then
				localMembers.radiusMin = 0
			else
				localMembers.radiusMin = localMembers.radiusMax - thickness
			end
			
			localMembers.randomRatio = math.pi * (localMembers.radiusMax * localMembers.radiusMax - localMembers.radiusMin * localMembers.radiusMin)
		else
			if _thickness ~= nil and type(_thickness) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "thickness type must be nil or number", true)
			end
		end
	end
	
	function newCylinder.getThickness(self)	
		if localMembers.thickness ~= 0 then
			return localMembers.thickness
		end
	end
	
	function newCylinder.isInside(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newCylinder) .. "isInside(reference)"
		
		if ATMEFct.isReference(_reference) == true then
			local point = ATMEFct.getPointFromReference(_reference)
			
			local datas = {
				zone = self,
				actualCenter = ATMEFct.C_Zone_getActualCenter(localMembers),
				actualAzimuth = ATMEFct.C_Zone_getActualAzimuth(localMembers),
			}
			
			return ATMEFct.C_Cylinder_isInside(point, datas)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
		end
		
		return false
	end
	
	function newCylinder.addResultCallback(self, _fct)
		local fctDefinition = ATMEFct.getErrorPrefix(newCylinder) .. "addResultCallback(fct)"

		if type(_fct) == "function" then
			localMembers.resultCallbacks[ATMEDatas.currentModuleName] = {
				fct = _fct,
				aborted = false,
			}
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
		end
	end
	
	function newCylinder.removeResultCallback(self)
		ATMEFct.C_Zone_abortResult(localMembers)
		localMembers.resultCallbacks[ATMEDatas.currentModuleName] = nil
	end
	
	function newCylinder.abortResultCallback(self)
		ATMEFct.C_Zone_abortResult(localMembers)
	end
	
	function newCylinder.getRandomPosition(self, _surfaceFilter, _isFlat, _altMin)
		local fctDefinition = ATMEFct.getErrorPrefix(newCylinder) .. "getRandomPosition([surfaceFilter], [flat], [altMin])"
	
		if (_surfaceFilter == nil or type(_surfaceFilter) == "string") and
					(_isFlat == nil or type(_isFlat) =="boolean") and (_altMin == nil or type(_altMin) =="number") then

			local surfaceFilter = "ALL"
			
			if _surfaceFilter ~= nil then
				surfaceFilter = string.upper(_surfaceFilter)
			end
			
			if ATMEConst.surfaceType[surfaceFilter] ~= nil then
				local datas = {
					zone = self,
					surfaceFilter = surfaceFilter,
					isFlat = false,
					altMin = 0,
					localMembers = localMembers,
					moduleName = ATMEDatas.currentModuleName,
				}
				
				if _isFlat ~= nil then
					datas.isFlat = _isFlat
				end
				
				if type(_altMin) == "number" then
					datas.altMin = _altMin
				end
				
				if localMembers.resultCallbacks[ATMEDatas.currentModuleName] == nil then
					ATMEFct.C_Zone_getRandomPosition(datas)
					return datas.point
				else
					if localMembers.coroutines[ATMEDatas.currentModuleName] == nil then
						ATMEFct.outputCore(nil, "newCylinder.getRandomPosition appel asynchrone", 2)
						localMembers.coroutines[ATMEDatas.currentModuleName] = coroutine.create(ATMEFct.C_Zone_getRandomPosition)
						coroutine.resume(localMembers.coroutines[ATMEDatas.currentModuleName], datas)
						localMembers.resultCallbacks[ATMEDatas.currentModuleName].idTimer = timer.scheduleFunction(ATMETimers.C_Zone_searchResultTimer, datas, timer.getTime() + 0.001)
						return
					else
						ATMEFct.setError("ALREADY_STARTED")
						ATMEFct.warningOutputCore(nil, fctDefinition, "search already started", false)
						return
					end
				end
			end
			
			ATMEFct.errorOutputCore(nil, fctDefinition, "surfaceFilter must be 'ALL', 'WATER' or 'LAND'", false)
		else
			if _surfaceFilter ~= nil and type(_surfaceFilter) ~= "string" then 
				ATMEFct.errorOutputCore(nil, fctDefinition, "surfaceFilter type must be nil or string", true)
			elseif _isFlat ~= nil and type(_isFlat) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "flat type must be nil or boolean", false)
			elseif _altMin ~= nil and type(_altMin) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "altMin type must be nil or number", false)
			end
		end
	end
	
	function newCylinder.getRadius(self)
		return localMembers.radiusMax
	end
	
	function newCylinder.getCenter(self)
		return ATMEFct.C_Zone_getActualCenter(localMembers)
	end
	
	function newCylinder.getHeight(self)
		return localMembers.deltaAltitude * 2
	end
	
	function newCylinder.bindUnit(self, _unit, _azimuth, _distance, _deltaAltitude)
		local fctDefinition = ATMEFct.getErrorPrefix(newCylinder) .. "bindUnit()\nbindUnit(unit, [azimuth, distance, [deltaAltitude]])"
		
		if _unit == nil or ATMEFct.isObjectClass(_unit, ATME.C_Unit) == true then
			if _unit == nil then
				localMembers.boundUnit = nil
			else
				local localMembersUnit = ATMEDatas.localMembers[_unit:getIndex()]
			
				if localMembersUnit.objectDCS:isExist() == true then
					localMembers.boundUnit = {
						unit = _unit,
						azimuthRad = 0,
						distance = 0,
						deltaAltitude = 0
					}
					
					if (_azimuth == nil and _distance == nil) or (type(_azimuth) == "number" and type(_distance) == "number") then
						local azimuth = 0
						local distance = 0
						
						if _distance ~= nil then
							distance = _distance
							azimuth = _azimuth
						end
						
						if distance >= 0 then
							if _deltaAltitude == nil or type(_deltaAltitude) == "number" then
								localMembers.boundUnit.azimuthRad = math.rad(ATMEFct.getAzimuthMod(azimuth, 360))
								localMembers.boundUnit.distance = math.floor(distance)
								
								if type(_deltaAltitude) == "number" then
									localMembers.boundUnit.deltaAltitude = _deltaAltitude
								end
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "deltaAltitude type must be nil or number", false)
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "distance must be >= 0", false)
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "azimuth and distance type must be both nil or number", false)
					end
				else
					localMembers.boundUnit = nil
					ATMEFct.setError("ALREADY_DISABLED")
					ATMEFct.warningOutputCore(nil, fctDefinition, "bound unit already disabled", false)
				end
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "unit type must be nil or ATME.C_Unit", true)
		end
	end
	
	function newCylinder.getBoundUnit(self)
		if localMembers.boundUnit ~= nil then
			local localMembersUnit = ATMEDatas.localMembers[localMembers.boundUnit.unit:getIndex()]
			
			if localMembersUnit.objectDCS:isExist() == true then
				return localMembers.boundUnit.unit
			else
				localMembers.boundUnit = nil
			end
		end
	end
	
	function newCylinder.addTrackingCallback(self, _fct, _object)
		local fctDefinition = ATMEFct.getErrorPrefix(newCylinder) .. "addTrackingCallback(fct, object)"

		if type(_fct) == "function" and (ATMEFct.isObjectClass(_object, ATME.C_Unit) == true or ATMEFct.isObjectClass(_object, ATME.C_Weapon) == true) then
			local className = _object:getClassName()
			local name = _object:getName()
			if localMembers.tracking[name] == nil then
				localMembers.tracking[name] = {
					className = className,
					name = name,
					object = _object,
					localMembersObject = ATMEDatas.localMembers[_object:getIndex()],
					callbacks = {},
				}
			end
				
			localMembers.tracking[name].pos = _object:getPosition()
			
			localMembers.tracking[name].callbacks[ATMEDatas.currentModuleName] = _fct
			
			local datas = {
				name = name,
				localMembers = localMembers,
				zone = self,
			}
			
			timer.scheduleFunction(ATMETimers.C_Zone_timer, datas, timer.getTime() + ATMEConst.offsetTracking[className])
		else
			if type(_fct) ~= "function" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
			elseif ATMEFct.isObjectClass(_object, ATME.C_Unit) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "object type must be ATME.C_Unit or ATME.C_Weapon", false)
			end
		end
	end
	
	function newCylinder.removeTrackingCallback(self, _object)
		local fctDefinition = ATMEFct.getErrorPrefix(newCylinder) .. "removeTrackingCallback(object)"
		
		if ATMEFct.isObjectClass(_object, ATME.C_Unit) == true or ATMEFct.isObjectClass(_object, ATME.C_Weapon) == true then
			if localMembers.tracking[_object:getName()] ~= nil then
				localMembers.tracking[_object:getName()].callbacks[ATMEDatas.currentModuleName] = nil
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "object type must be ATME.C_Unit or ATME.C_Weapon", true)
		end
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end

	return newCylinder
end


ATMEDatas.C_Cylinder_MT.__call = function(t, ...)
	local args = {...}
	local fctDefinition = "Constructor :\nATME.C_Cylinder(DCSZoneName, height, [altitudeMSL])\nATME.C_Cylinder(reference, radius, height)"
	local point, radius, deltaAltitude
	
	if ATMEDatas.startMissionOk < 2 then
		ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
		return
	end
	
	if type(args[1]) == "string" then
		local zone = ATMEFct.getDCSZone(args[1])
		if zone ~=  nil then
			local altitude = zone.point.y
			
			if type(args[2]) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "height type must be number", false)
				return
			end

			deltaAltitude = args[2] / 2
			if deltaAltitude <= 0 then
				ATMEFct.errorOutputCore(nil, fctDefinition, "height must be > 0", false)
				return
			end
			
			if #args > 2 then
				if type(args[3]) == "number" then
					if args[3] >= 0 then
						altitude = args[3]
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "altitude must be >= 0", false)
						return
					end
				else	
					ATMEFct.errorOutputCore(nil, fctDefinition, "altitude type must be nil or number", false)
					return
				end
			end
			
			point = {
				x = zone.point.x,
				y = altitude,
				z = zone.point.z,
			}

			radius = zone.radius
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone '" .. args[1] .. "' unknown", false)
			return
		end
	elseif ATMEFct.isReference(args[1]) == true then
		if type(args[2]) == "number" then
			radius = args[2]
			if radius > 0 then
				point = ATMEFct.getPointFromReference(args[1])
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "radius must be > 0", false)
				return
			end
			
			if type(args[3]) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "deltaAltitude type must be number", false)
				return
			end
			
			deltaAltitude = args[3] / 2
			if deltaAltitude <= 0 then
				ATMEFct.errorOutputCore(nil, fctDefinition, "deltaAltitude must be > 0", false)
				return
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "radius type must be number", false)
			return
		end
	else

		ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #1 type must be string," .. ATMEConst.errorLabelReference, false)
		return
	end
	
	return ATMEFct.C_Cylinder_create(nil, nil, point, radius, deltaAltitude)
end

setmetatable(ATME.C_Cylinder, ATMEDatas.C_Cylinder_MT)

ATMEDatas.C_Cylinder_MT.mt.__metatable = "Private"
ATMEDatas.C_Cylinder_MT.mt.__index = ATME.C_Cylinder

ATMEDatas.C_Cylinder_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Cylinder_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Cylinder : Bad key " .. k, "table locked", false)
	end
end

ATMEFct.C_Box_isInside = function(_point, _datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	local dx = _point.x - _datas.actualCenter.x
	local dy = _point.y - _datas.actualCenter.y
	local dz = _point.z - _datas.actualCenter.z
	
	local angleRad = math.rad(_datas.actualAzimuth)
	
	local lPoint = {
		x = dx * math.cos(angleRad) + dz * math.sin(angleRad),
		y = -dx * math.sin(angleRad) + dz * math.cos(angleRad),
	}

	local ok = (math.abs(lPoint.x) >= localMembers.halfLengthMin and math.abs(lPoint.x) <= localMembers.halfLength and 
					math.abs(lPoint.y) <= localMembers.halfWidth and localMembers.isRounded == false) or
			 (math.abs(lPoint.y) >= localMembers.halfWidthMin and math.abs(lPoint.y) <= localMembers.halfWidth and 
					math.abs(lPoint.x) <= localMembers.halfLength)
					
	if ok == false and localMembers.isRounded == true then	
		local dxPoint = lPoint.x - localMembers.halfLength
		local dyPoint = lPoint.y
		local d = math.sqrt(dxPoint * dxPoint + dyPoint * dyPoint)
		local angle = math.atan2(dyPoint, dxPoint)
	
		ok = angle > -math.pi / 2 and angle < math.pi / 2 and d >= localMembers.halfWidthMin and d <= localMembers.halfWidth
		
		if ok == false then
			dxPoint = lPoint.x + localMembers.halfLength
			local d = math.sqrt(dxPoint * dxPoint + dyPoint * dyPoint)
			local angle = math.atan2(dyPoint, dxPoint)
		
			ok = (angle < -math.pi / 2 or angle > math.pi / 2) and d >= localMembers.halfWidthMin and d <= localMembers.halfWidth
		end
	end
	
	dy = lPoint.x * localMembers.a - dy
	
	return ok == true and math.abs(dy) <= localMembers.deltaAltitude
end

ATMEFct.C_Box_getRandomPosition = function(_datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	local angleRad = math.rad(_datas.actualAzimuth)
	local point
	
	local rand = math.random()
	if rand < localMembers.probLength then
		-- Soit la surface correspondant au grand coté avec l'épaisseur
		
		local dx = ATMEConst.randomSign[math.random(1,2)] * localMembers.halfLength * math.random()
		local dy = ATMEConst.randomSign[math.random(1,2)] * (localMembers.deltaWidth * math.random() + localMembers.halfWidthMin)
		local dAlt = ATMEConst.randomSign[math.random(1,2)] * localMembers.deltaAltitude * math.random()
		
		return {
			x = _datas.actualCenter.x + dx * math.cos(angleRad) - dy * math.sin(angleRad),
			y = _datas.actualCenter.y + dAlt - dx * localMembers.a,
			z = _datas.actualCenter.z + dx * math.sin(angleRad) + dy * math.cos(angleRad),
		}
	else
		-- Soit la surface correspondant au petit coté moins l'épaisseur
		-- Soit sur 2x le demi-cercle de part et d'autre du centre sur l'axe x
		
		local dAlt = ATMEConst.randomSign[math.random(1,2)] * localMembers.deltaAltitude * math.random()
		
		if localMembers.isRounded == true then
			local angle = 2 * math.pi * math.random()
			local radius = localMembers.deltaWidth * math.random() + localMembers.halfWidthMin
			local i = 1
			
			if angle >= math.pi then
				i = -1
			end
			
			local localPoint = {
				x = i * localMembers.halfLength + radius * math.sin(angle),
				y = radius * math.cos(angle),
			}
			
			return {
				x = _datas.actualCenter.x + localPoint.x * math.cos(angleRad) - localPoint.y * math.sin(angleRad),
				y = _datas.actualCenter.y + dAlt - localPoint.x * localMembers.a,
				z = _datas.actualCenter.z + localPoint.x * math.sin(angleRad) + localPoint.y * math.cos(angleRad),
			}
		else
			local dx = ATMEConst.randomSign[math.random(1,2)] * (localMembers.deltaLength * math.random() + localMembers.halfLengthMin)
			local dy = ATMEConst.randomSign[math.random(1,2)] * localMembers.halfWidthMin * math.random()
			
			return {
				x = _datas.actualCenter.x + dx * math.cos(angleRad) - dy * math.sin(angleRad),
				y = _datas.actualCenter.y + dAlt - dx * localMembers.a,
				z = _datas.actualCenter.z + dx * math.sin(angleRad) + dy * math.cos(angleRad),
			}
		end
	end
end

ATMEFct.C_Box_searchVolume = function(_datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	local lPoints = {}
	local points = {}
	
	local angleRad = math.rad(_datas.actualAzimuth)

	local length = localMembers.halfLength
	if localMembers.isRounded == true then
		length = length + localMembers.halfWidth
	end

	table.insert(lPoints, {x = length, y = localMembers.halfWidth})
	table.insert(lPoints, {x = -length, y = localMembers.halfWidth})
	table.insert(lPoints, {x = -length, y = -localMembers.halfWidth})
	table.insert(lPoints, {x = length, y = -localMembers.halfWidth})

	for _index, _point in ipairs(lPoints) do
		points[_index] =  {
			x = _datas.actualCenter.x + _point.x * math.cos(angleRad) - _point.y * math.sin(angleRad),
			y = _datas.actualCenter.z + _point.x * math.sin(angleRad) + _point.y * math.cos(angleRad),
		}
	end
	
	local dy = math.abs(localMembers.halfLength * localMembers.a)

	local minPoint = {
		x = points[1].x,
		y = _datas.actualCenter.y - localMembers.deltaAltitude - dy,
		z = points[1].y,
	}
		
	local maxPoint = {
		x = points[1].x,
		y = _datas.actualCenter.y + localMembers.deltaAltitude + dy,
		z = points[1].y,
	}

	for _, _point in ipairs(points) do
		if minPoint.x > _point.x then
			minPoint.x = _point.x
		end
		
		if minPoint.z > _point.y then
			minPoint.z = _point.y
		end
		
		if maxPoint.x < _point.x then
			maxPoint.x = _point.x
		end
		
		if maxPoint.z < _point.y then
			maxPoint.z = _point.y
		end
	end

	return  world.VolumeType.BOX,
			"C_FindContext_search",
			{
				min = minPoint,
				max = maxPoint,
			}
end

ATMEDatas.C_Box_MT = {
	className = "C_Box",
	parentClass = ATME.C_MovableZone3D,
	
	mt = {},
	
	keys = {
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["isInside"] = true,
		["getRandomPosition"] = true,
		["addTrackingCallback"] = true,
		["removeTrackingCallback"] = true,
		["addResultCallback"] = true,
		["removeResultCallback"] = true,
		["abortResultCallback"] = true,
		["getIndex"] = true,
		["getUserPrivateDatas"] = true,
		["getLength"] = true,
		["getWidth"] = true,
		["getAzimuth"] = true,
		["getCenter"] = true,
		["getGlide"] = true,
		["getHeight"] = true,
		["changeThickness"] = true,
		["getThickness"] = true,
		["isRounded"] = true,
		["bindUnit"] = true,
		["getBoundUnit"] = true,
	},
	
	add = function(self, _center, _azimuth, _length, _width, _deltaAltitude, _glide, _isRounded)
		local lPoints = {}
		local points = {}
		local glideRadValue = math.rad(_glide)
		local offset = 0
		
		if _isRounded == true then
			offset = math.pi * _width * _width / 4
		end
		
		local item = {
			points = _points,
			localPoints = lPoints,
			a = math.sin(glideRadValue),
			center = _center,
			azimuth = _azimuth,
			glideRad = glideRadValue,
			glide = _glide,
			length = _length,
			width = _width,
			halfLength = _length / 2,
			halfWidth = _width / 2,
			halfLengthMin = 0,
			halfWidthMin = 0,
			deltaAltitude = _deltaAltitude,
			thickness = 0,
			isRounded = _isRounded,
			randomRatio = _length * _width + offset,
			is3D = true,
			userPrivateDatas = {},
			tracking = {},
			resultCallbacks = {},
			coroutines = {},
		}
		
		-- item.box est initialisé dans C_Box_create
		
		item.thicknessMax = math.min(item.halfLength, item.halfWidth)
		item.deltaLength = item.halfLength - item.halfLengthMin
		item.deltaWidth = item.halfWidth - item.halfWidthMin
		
		local vCircle = 0
		local vWidth = 0
		if _isRounded == true then
			vCircle = math.pi * (item.halfWidth * item.halfWidth)
		else
			vWidth = item.deltaLength * (item.width - 2 * item.deltaWidth) * 2
		end
		
		local vLength = item.deltaWidth * item.length * 2
		
		local vTotal = vCircle + vLength + vWidth
		
		item.probLength = vLength/vTotal
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].tracking = nil
		ATMEDatas.localMembers[_id].box = nil
		ATMEDatas.localMembers[_id].resultCallbacks = nil
		ATMEDatas.localMembers[_id].coroutines = nil
	end,
}

ATME.C_Box = {
	getClassName = function(self)
		return ATMEDatas.C_Box_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Box_MT.parentClass
	end,
	
	createNamed = function(_name, _reference, _azimuth, _length, _width, _height, ...)
		local fctDefinition = "Function :\nATME.C_Box.createNamed(name, reference, azimuth, length, width, height)\nATME.C_Box.createNamed(name, reference, azimuth, length, width, height, glide)\nATME.C_Box.createNamed(name, reference, azimuth, length, width, height, isRounded)ATME.C_Box(reference, azimuth, length, width, height, glide, isRounded)"
		local args = {...}
		
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" and ATMEFct.isReference(_reference) == true and type(_azimuth) == "number" and  
									type(_length) == "number" and type(_width) == "number" and 
									type(_height) == "number" then
			local center = ATMEFct.getPointFromReference(_reference)

			if _length > 0 then
				if _width > 0 then
					if _height > 0 then
						local glide = 0
						local isRounded = false

						if #args == 1 then
							if type(args[1]) == "number" then
								glide = ATMEFct.getAzimuthMod(args[1], 180)
								if glide <= -90 or glide >= 90 then
									ATMEFct.errorOutputCore(nil, fctDefinition, "glide must be > -90 and < 90", false)
									return
								end
							elseif type(args[1]) == "boolean" then
								isRounded = args[1]
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #6 type must be nil, number or boolean", false)
								return
							end
						elseif #args > 1 then
							if type(args[1]) == "number" then
								glide = ATMEFct.getAzimuthMod(args[1], 180)
								
								if glide <= -90 or glide >= 90 then
									ATMEFct.errorOutputCore(nil, fctDefinition, "glide must be > -90 and < 90", false)
									return
								end
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "glide type must be nil or number", false)
								return
							end
							
							if type(args[2]) == "boolean" then
								isRounded = args[2]
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "isRounded type must be nil or boolean", false)
								return
							end

						end
						
						local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
						local err, newBox = ATMEFct.C_Zone_createNamed(localMembersModule.zones, _name, ATMEFct.C_Box_create, center, ATMEFct.getAzimuthMod(_azimuth, 360), _length, _width, _height / 2, glide, isRounded)				
			
						if err == true then
							ATMEFct.warningOutputCore(nil, fctDefinition, newBox, false)
							return
						end

						return newBox
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "height must be > 0", false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "width must be > 0", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "length must be > 0", false)
			end
		else
			if type(_name) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
			elseif ATMEFct.isReference(_reference) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
			elseif type(_azimuth) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "azimuth type must be number", false)
			elseif type(_length) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "length type must be number", false)
			elseif type(_width) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "width type must be number", false)
			elseif type(_height) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "height type must be number", false)
			end
		end
	end,
	
	getByName = function(_name)
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Box.getByName(name)", "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			if ATMEFct.isObjectClass(ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name], ATME.C_Box) == true then
				return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name]
			end
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Box.getByName(name)", "name type must be string", false)
		end
	end,
}

ATMEFct.C_Box_create = function(_name, _zones, _center, _heading, _length, _width, _deltaAltitude, _glide, _isRounded)
	local newBox = {}
	
	setmetatable(newBox, ATMEDatas.C_Box_MT.mt)
	local index = ATMEDatas.C_Box_MT:add(_center, _heading, _length, _width, _deltaAltitude, _glide, _isRounded)
	local localMembers = ATMEDatas.localMembers[index]
			
	local name
	local tagsName

	if _name ~= nil and _zones ~= nil then
		name = _name
		_zones[name] = newBox
		tagsName = ATMEFct.decompString(_name)
	end
	
	function newBox.getIndex(self)
		return index
	end
	
	function newBox.getName(self)
		return name
	end
	
	function newBox.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newBox) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newBox.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newBox.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newBox) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newBox.getLength(self)
		return localMembers.length
	end
	
	function newBox.getHeight(self)
		return localMembers.deltaAltitude * 2
	end
	
	function newBox.getWidth(self)
		return localMembers.width
	end
	
	function newBox.getAzimuth(self)
		return ATMEFct.C_Zone_getActualAzimuth(localMembers)
	end

	function newBox.getGlide(self)
		return localMembers.glide
	end
	
	function newBox.getCenter(self)
		return ATMEFct.C_Zone_getActualCenter(localMembers)
	end
	
	function newBox.getUserPrivateDatas(self, _moduleName)
		local fctDefinition = ATMEFct.getErrorPrefix(newBox) .. "getUserPrivateDatas([moduleName])"
		local moduleName
		
		if _moduleName ~= nil then
			if type(_moduleName) == "string" then
				if localMembers.userPrivateDatas[_moduleName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
					return
				else
					moduleName = _moduleName
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
				return
			end
		else
			moduleName = ATMEDatas.currentModuleName
		end
		
		return localMembers.userPrivateDatas[moduleName]
	end
	
	function newBox.changeThickness(self, _thickness)
		local fctDefinition = ATMEFct.getErrorPrefix(newBox) .. "changeThickness([thickness])"
				
		if _thickness == nil or type(_thickness) == "number" then
			local thickness = 0
			if type(_thickness) == "number" then
				thickness = _thickness
			end
			
			
			if thickness >= localMembers.thicknessMax then
				thickness = 0
			elseif thickness < 0 then
				ATMEFct.errorOutputCore(nil, fctDefinition, "thickness must be >= 0", false)
				return
			end
			
			localMembers.thickness = thickness
			
			if thickness > 0 then
				localMembers.halfLengthMin = localMembers.halfLength - thickness
				localMembers.halfWidthMin = localMembers.halfWidth - thickness
			else
				localMembers.halfLengthMin = 0
				localMembers.halfWidthMin = 0
			end
			
			localMembers.deltaLength = localMembers.halfLength - localMembers.halfLengthMin
			localMembers.deltaWidth = localMembers.halfWidth - localMembers.halfWidthMin
			
			local vLength = localMembers.deltaWidth * localMembers.length * 2 * localMembers.deltaAltitude
			local vCircle = 0
			local vWidth = 0
			if localMembers.isRounded == true then
				vCircle = math.pi * (localMembers.halfWidth * localMembers.halfWidth - localMembers.halfWidthMin * localMembers.halfWidthMin) * 2 * localMembers.deltaAltitude
			else
				vWidth = localMembers.deltaLength * (localMembers.width - 2 * localMembers.deltaWidth) * 2 * localMembers.deltaAltitude
			end
			
			local vTotal = vCircle + vLength + vWidth
			
			localMembers.probLength = vLength/vTotal
			
			localMembers.randomRatio = localMembers.length * localMembers.width - localMembers.halfLengthMin * localMembers.halfWidthMin * 4 + math.pi * (localMembers.halfWidth * localMembers.halfWidth - localMembers.halfWidthMin * localMembers.halfWidthMin)
		else
			if _thickness ~= nil and type(_thickness) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "thickness type must be nil or number", true)
			end
		end
	end
	
	function newBox.getThickness(self)	
		if localMembers.thickness ~= 0 then
			return localMembers.thickness
		end
	end
	
	function newBox.isRounded(self)
		return localMembers.isRounded
	end
	
	function newBox.isInside(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newBox) .. "isInside(reference)"
		local isInside = false
		
		if ATMEFct.isReference(_reference) == true then		
			local point = ATMEFct.getPointFromReference(_reference)
			
			local datas = {
				zone = self,
				actualCenter = ATMEFct.C_Zone_getActualCenter(localMembers),
				actualAzimuth = ATMEFct.C_Zone_getActualAzimuth(localMembers),
			}
			
			ATMEFct.C_Box_isInside(point, datas)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
		end
		
		return isInside
	end
	
	function newBox.addResultCallback(self, _fct)
		local fctDefinition = ATMEFct.getErrorPrefix(newBox) .. "addResultCallback(fct)"

		if type(_fct) == "function" then
			localMembers.resultCallbacks[ATMEDatas.currentModuleName] = {
				fct = _fct,
				aborted = false,
			}
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
		end
	end
	
	function newBox.removeResultCallback(self)
		ATMEFct.C_Zone_abortResult(localMembers)
		localMembers.resultCallbacks[ATMEDatas.currentModuleName] = nil
	end
	
	function newBox.abortResultCallback(self)
		ATMEFct.C_Zone_abortResult(localMembers)
	end
	
	function newBox.getRandomPosition(self, _surfaceFilter, _isFlat, _altMin)
		local fctDefinition = ATMEFct.getErrorPrefix(newBox) .. "getRandomPosition([surfaceFilter], [flat], [altMin])"
	
		if (_surfaceFilter == nil or type(_surfaceFilter) == "string") and
					(_isFlat == nil or type(_isFlat) =="boolean") and (_altMin == nil or type(_altMin) =="number") then

			local surfaceFilter = "ALL"
			
			if _surfaceFilter ~= nil then
				surfaceFilter = string.upper(_surfaceFilter)
			end
			
			if ATMEConst.surfaceType[surfaceFilter] ~= nil then
				local datas = {
					zone = self,
					surfaceFilter = surfaceFilter,
					isFlat = false,
					altMin = 0,
					localMembers = localMembers,
					moduleName = ATMEDatas.currentModuleName,
				}
				
				if _isFlat ~= nil then
					datas.isFlat = _isFlat
				end
				
				if type(_altMin) == "number" then
					datas.altMin = _altMin
				end
				
				if localMembers.resultCallbacks[ATMEDatas.currentModuleName] == nil then
					ATMEFct.C_Zone_getRandomPosition(datas)
					return datas.point
				else
					if localMembers.coroutines[ATMEDatas.currentModuleName] == nil then
						ATMEFct.outputCore(nil, "newBox.getRandomPosition appel asynchrone", 2)
						localMembers.coroutines[ATMEDatas.currentModuleName] = coroutine.create(ATMEFct.C_Zone_getRandomPosition)
						coroutine.resume(localMembers.coroutines[ATMEDatas.currentModuleName], datas)
						localMembers.resultCallbacks[ATMEDatas.currentModuleName].idTimer = timer.scheduleFunction(ATMETimers.C_Zone_searchResultTimer, datas, timer.getTime() + 0.001)
						return
					else
						ATMEFct.setError("ALREADY_STARTED")
						ATMEFct.warningOutputCore(nil, fctDefinition, "search already started", false)
						return
					end
				end
			end
			
			ATMEFct.errorOutputCore(nil, fctDefinition, "surfaceFilter must be 'ALL', 'WATER' or 'LAND'", false)
		else
			if _surfaceFilter ~= nil and type(_surfaceFilter) ~= "string" then 
				ATMEFct.errorOutputCore(nil, fctDefinition, "surfaceFilter type must be nil or string", true)
			elseif _isFlat ~= nil and type(_isFlat) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "flat type must be nil or boolean", false)
			elseif _altMin ~= nil and type(_altMin) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "altMin type must be nil or number", false)
			end
		end
	end
	
	function newBox.bindUnit(self, _unit, _azimuth, _distance, _deltaAltitude)
		local fctDefinition = ATMEFct.getErrorPrefix(newBox) .. "bindUnit()\nbindUnit(unit, [azimuth, distance, [deltaAltitude]])"
		
		if _unit == nil or ATMEFct.isObjectClass(_unit, ATME.C_Unit) == true then
			if _unit == nil then
				localMembers.boundUnit = nil
			else
				local localMembersUnit = ATMEDatas.localMembers[_unit:getIndex()]
			
				if localMembersUnit.objectDCS:isExist() == true then
					localMembers.boundUnit = {
						unit = _unit,
						azimuthRad = 0,
						distance = 0,
						deltaAltitude = 0
					}
					
					if (_azimuth == nil and _distance == nil) or (type(_azimuth) == "number" and type(_distance) == "number") then
						local azimuth = 0
						local distance = 0
						
						if _distance ~= nil then
							distance = _distance
							azimuth = _azimuth
						end
						
						if distance >= 0 then
							if _deltaAltitude == nil or type(_deltaAltitude) == "number" then
								localMembers.boundUnit.azimuthRad = math.rad(ATMEFct.getAzimuthMod(azimuth, 360))
								localMembers.boundUnit.distance = math.floor(distance)
								
								if type(_deltaAltitude) == "number" then
									localMembers.boundUnit.deltaAltitude = _deltaAltitude
								end
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "deltaAltitude type must be nil or number", false)
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "distance must be >= 0", false)
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "azimuth and distance type must be both nil or number", false)
					end
				else
					localMembers.boundUnit = nil
					ATMEFct.setError("ALREADY_DISABLED")
					ATMEFct.warningOutputCore(nil, fctDefinition, "bound unit already disabled", false)
				end
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "unit type must be nil or ATME.C_Unit", true)
		end
	end
	
	function newBox.getBoundUnit(self)
		if localMembers.boundUnit ~= nil then
			local localMembersUnit = ATMEDatas.localMembers[localMembers.boundUnit.unit:getIndex()]
			
			if localMembersUnit.objectDCS:isExist() == true then
				return localMembers.boundUnit.unit
			else
				localMembers.boundUnit = nil
			end
		end
	end
	
	function newBox.addTrackingCallback(self, _fct, _object)
		local fctDefinition = ATMEFct.getErrorPrefix(newBox) .. "addTrackingCallback(fct, object)"

		if type(_fct) == "function" and (ATMEFct.isObjectClass(_object, ATME.C_Unit) == true or ATMEFct.isObjectClass(_object, ATME.C_Weapon) == true) then
			local className = _object:getClassName()
			local name = _object:getName()
			if localMembers.tracking[name] == nil then
				localMembers.tracking[name] = {
					className = className,
					name = name,
					object = _object,
					localMembersObject = ATMEDatas.localMembers[_object:getIndex()],
					callbacks = {},
				}
			end
				
			localMembers.tracking[name].pos = _object:getPosition()
			
			localMembers.tracking[name].callbacks[ATMEDatas.currentModuleName] = _fct
			
			local datas = {
				name = name,
				localMembers = localMembers,
				zone = self,
			}
			
			timer.scheduleFunction(ATMETimers.C_Zone_timer, datas, timer.getTime() + ATMEConst.offsetTracking[className])
		else
			if type(_fct) ~= "function" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
			elseif ATMEFct.isObjectClass(_object, ATME.C_Unit) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "object type must be ATME.C_Unit or ATME.C_Weapon", false)
			end
		end
	end
	
	function newBox.removeTrackingCallback(self, _object)
		local fctDefinition = ATMEFct.getErrorPrefix(newBox) .. "removeTrackingCallback(object)"
		
		if ATMEFct.isObjectClass(_object, ATME.C_Unit) == true or ATMEFct.isObjectClass(_object, ATME.C_Weapon) == true then
			if localMembers.tracking[_object:getName()] ~= nil then
				localMembers.tracking[_object:getName()].callbacks[ATMEDatas.currentModuleName] = nil
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "object type must be ATME.C_Unit or ATME.C_Weapon", true)
		end
	end
	
	local datas = {
		zone = newBox,
		actualCenter = _center,
		actualAzimuth = _heading,
	}
	
	local _, _, cbox = ATMEFct.C_Box_searchVolume(datas)
	localMembers.box = cbox
		
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return newBox
end

ATMEDatas.C_Box_MT.__call = function(t, _reference, _azimuth, _length, _width, _height, ...)
	local fctDefinition = "Constructor :\nATME.C_Box(reference, azimuth, length, width, height)\nATME.C_Box(reference, azimuth, length, width, height, glide)\nATME.C_Box(reference, azimuth, length, width, height, isRounded)ATME.C_Box(reference, azimuth, length, width, height, glide, isRounded)"
	local args = {...}
	
	if ATMEDatas.startMissionOk < 2 then
		ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
		return
	end
	
	if ATMEFct.isReference(_reference) == true and type(_azimuth) == "number" and  
								type(_length) == "number" and type(_width) == "number" and 
								type(_height) == "number" then
		local center = ATMEFct.getPointFromReference(_reference)

		if _length > 0 then
			if _width > 0 then
				if _height > 0 then
					local glide = 0
					local isRounded = false

					if #args == 1 then
						if type(args[1]) == "number" then
							glide = ATMEFct.getAzimuthMod(args[1], 180)
							if glide <= -90 or glide >= 90 then
								ATMEFct.errorOutputCore(nil, fctDefinition, "glide must be > -90 and < 90", false)
								return
							end
						elseif type(args[1]) == "boolean" then
							isRounded = args[1]
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #6 type must be nil, number or boolean", false)
							return
						end
					elseif #args > 1 then
						if type(args[1]) == "number" then
							glide = ATMEFct.getAzimuthMod(args[1], 180)
							
							if glide <= -90 or glide >= 90 then
								ATMEFct.errorOutputCore(nil, fctDefinition, "glide must be > -90 and < 90", false)
								return
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "glide type must be nil or number", false)
							return
						end
						
						if type(args[2]) == "boolean" then
							isRounded = args[2]
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "isRounded type must be nil or boolean", false)
							return
						end

					end
					
					return ATMEFct.C_Box_create(nil, nil, center, ATMEFct.getAzimuthMod(_azimuth, 360), _length, _width, _height / 2, glide, isRounded)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "height must be > 0", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "width must be > 0", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "length must be > 0", false)
		end
	else
		if ATMEFct.isReference(_reference) == false then
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
		elseif type(_azimuth) ~= "number" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "azimuth type must be number", false)
		elseif type(_length) ~= "number" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "length type must be number", false)
		elseif type(_width) ~= "number" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "width type must be number", false)
		elseif type(_height) ~= "number" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "height type must be number", false)
		end
	end
end

setmetatable(ATME.C_Box, ATMEDatas.C_Box_MT)

ATMEDatas.C_Box_MT.mt.__metatable = "Private"
ATMEDatas.C_Box_MT.mt.__index = ATME.C_Box

ATMEDatas.C_Box_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Box_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Box : Bad key " .. k, "table locked", false)
	end
end


----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Airbase : gère les bases aériennes (aéroports, bateaux ou FARP). 
--                         Pas de contructeur accessible aux modules externes.
-- 
-- Deux listes globales sont créés et initialisées :
--			-> airbasesById pour le référencement par ID (number pour les bases de la carte, string sinon)
--			-> airbases pour le référencement par nom
----------------------------------------------------------------------------------------------------------

ATMEDatas.C_Airbase_MT = {
	className = "C_Airbase",
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["getUserPrivateDatas"] = true,
		["getName"] = true,
		["getCategoryName"] = true,
		["getCoalitionName"] = true,
		["getPosition"] = true,
		["getDCSHelipadID"] = true,
		["getDCSAirdromeID"] = true,
		["getDCSObject"] = true,
		["getDCSID"] = true,
		["getAttributes"] = true,
		["hasAttributes"] = true,
		["isHelipad"] = true,
		["isInZones"] = true,
		["isInDCSZone"] = true,
		["isInZone"] = true,
		["getParking"] = true,
	},
	
	add = function(self, _airbase, _categoryName, _id)
		-- Bug DCS V2.5.6 
		-- -> Attention getID pour une unité ou un static est un string mais pour la base des cartes c'est un nombre 
		--    du coup : airdromeId est un nombre mais utilisé pour les bases de la carte uniquement,
		--	            helipadId est un nombre mais utilisé par les farps et bateaux, d'où conversion
		-- Toutes les bases ne sont pas unités, donc isExist de DCS n'est pas suffisant. Utilisation d'un flag destroyed
		
		local id = _id
		if _id == nil then
			id = _airbase:getID()
		end
		
		local item = {
			objectDCS = _airbase,
			objectDCSId = id,
			categoryName = _categoryName,
			userPrivateDatas = {},
			destroyed = false,					-- voir commentaire ci dessus			
		}

		if _id == nil then
			item.helipadId = id
		else
			item.airdromeId = id
		end
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
	end,
}

ATME.C_Airbase = {
	getClassName = function(self)
		return ATMEDatas.C_Airbase_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Airbase_MT.parentClass
	end,
	
	getByName = function(_name, _categoryName)
		if type(_name) == "string" and (_categoryName == nil or type(_categoryName) == "string") then
			local categoryName = _categoryName
			
			if categoryName == nil then
				return ATMEDatas.airbases["AIRBASE"][_name]
			elseif type(categoryName) == "string" then
				if categoryName ~= "SHIP" and categoryName ~= "HELIPAD" and categoryName ~= "AIRBASE" then
					ATMEFct.errorOutputCore(nil, "Function ATME.C_Airbase.getByName(name, [categoryName])", "categoryName must be 'SHIP', 'HELIPAD' or 'AIRBASE'", false)
				else
					return ATMEDatas.airbases[categoryName][_name]
				end
			else
				ATMEFct.errorOutputCore(nil, "Function ATME.C_Airbase.getByName(name, [categoryName])", "categoryName type must be string", false)
			end
		else
			if type(_name) ~= "string" then
				ATMEFct.errorOutputCore(nil, "Function ATME.C_Airbase.getByName(name, [categoryName])", "name type must be string", false)
			elseif _categoryName ~= nil and type(_categoryName) ~= "string" then
				ATMEFct.errorOutputCore(nil, "Function ATME.C_Airbase.getByName(name, [categoryName])", "categoryName type must be nil or string", false)
			end
		end
	end,
	
	getByID = function(_id)
		if type(_id) == "string" or type(_id) == "number" then
			local id = _id
			if type(_id) == "number" then
				id = math.floor(_id)
			end
			
			return ATMEDatas.airbasesById[id]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Airbase.getByID(id)", "id type must be string or number", false)
		end
	end,
}

ATMEFct.C_Airbase_create_basic = function(_categoryName, _helicoptersOnly, _airbase, _id)
	local name = _airbase:getName()
	local newAirbase = {}
	local position = _airbase:getPosition().p

	local helicoptersOnly = _helicoptersOnly

	setmetatable(newAirbase, ATMEDatas.C_Airbase_MT.mt)
	local index = ATMEDatas.C_Airbase_MT:add(_airbase, _categoryName, _id)
	local localMembers = ATMEDatas.localMembers[index]

	ATMEDatas.airbasesById[localMembers.objectDCSId] = newAirbase
	ATMEDatas.airbases[localMembers.categoryName][name] = newAirbase
	
	function newAirbase.getIndex(self)
		return index
	end
	
	function newAirbase.getName(self)
		return name
	end

	function newAirbase.getDCSHelipadID(self)	
		return localMembers.helipadId
	end

	function newAirbase.getDCSAirdromeID(self)
		return localMembers.airdromeId
	end

	function newAirbase.getDCSObject(self)
		return localMembers.objectDCS
	end

	function newAirbase.getDCSID(self)
		return localMembers.objectDCSId
	end

	function newAirbase.getParking(self, _available)
		local fctDefinition = ATMEFct.getErrorPrefix(newAirbase) .. "getParking([available])"
		
		if localMembers.destroyed == false then
			if _available == nil or type(_available) == "boolean" then
				local available = true
				
				if type(_available) == "boolean" then
					available = _available
				end
				
				return localMembers.objectDCS:getParking(available)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "available type must be nil or boolean", true)
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end
	
	function newAirbase.getUserPrivateDatas(self, _moduleName)
		local fctDefinition = ATMEFct.getErrorPrefix(newAirbase) .. "getUserPrivateDatas([moduleName])"
		local moduleName
		
		if _moduleName ~= nil then
			if type(_moduleName) == "string" then
				if localMembers.userPrivateDatas[_moduleName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
					return
				else
					moduleName = _moduleName
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
				return
			end
		else
			moduleName = ATMEDatas.currentModuleName
		end
		
		return localMembers.userPrivateDatas[moduleName]
	end
		
	function newAirbase.isHelipad(self)
		if localMembers.destroyed == false then
			return helicoptersOnly
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newAirbase) .. "isHelipad()", "already disabled", false)
		end
	end
	
	function newAirbase.getCoalitionName(self)
		if localMembers.destroyed == false then
			return ATMEConst.coalitionNames[localMembers.objectDCS:getCoalition()]
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newAirbase) .. "getCoalitionName()", "already disabled", false)
		end
	end
	
	function newAirbase.getCategoryName(self)
		if localMembers.destroyed == false then
			return localMembers.categoryName
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newAirbase) .. "getCategoryName()", "already disabled", false)
		end
	end
	
	function newAirbase.getPosition(self)
		if localMembers.categoryName == "SHIP" then
			if localMembers.destroyed == false then
				return localMembers.objectDCS:getPoint()
			else
				return localMembers.lastPosition.p
			end
		else
			return position
		end
	end
	
	function newAirbase.isInZones(self, _zones)
		local fctDefinition = ATMEFct.getErrorPrefix(newAirbase) .. "isInZones(zones)"
	
		if ATMEFct.isObjectClass(_zones, ATME.C_Zones) == true then
			local point = self:getPosition()
		
			local isInside, zonesId = ATMEFct.C_Zones_isInside(self, point)
			return isInside
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "zones type must be ATME.C_Zones", true)
		end
	end				
	
	function newAirbase.isInDCSZone(self, _name)
		local fctDefinition = ATMEFct.getErrorPrefix(newAirbase) .. "isInDCSZone(name)"
	
		if type(_name) == "string" then
			local zone = ATMEFct.getDCSZone(_name)
			
			if zone ~= nil then
				local vector = ATMEFct.vectors["3D"].createFromReference(nil, nil, self:getPosition(), zone.point)
				if vector:getHModule() <= zone.radius then
					return true
				else
					return false
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone " .. _name .. " unknown", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", true)
		end
	end
	
	function newAirbase.isInZone(self, _zone)
		local fctDefinition = ATMEFct.getErrorPrefix(newAirbase) .. "isInZone(zone)"
		
		if ATMEFct.isObjectClass(_zone, ATME.C_Zone) == true then
			local point = self:getPosition()
			local localMembersZone = ATMEDatas.localMembers[_zone:getIndex()]
			
			local datas = {
				zone = _zone,
				actualCenter = ATMEFct.C_Zone_getActualCenter(localMembersZone),
				azimuth = ATMEFct.C_Zone_getActualAzimuth(localMembersZone),
			}
		
			return ATMEFct[ATMEConst.indZoneConfig[_zone:getClassName()].fctIsInside](point, datas)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "zone type must be ATME.C_Zone", true)
		end
	end
	
	function newAirbase.getAttributes(self)
		if localMembers.destroyed == false then
			local datas = Airbase.getDesc(localMembers.objectDCS)
			local attributes = {}
			
			for _id, _ in pairs(datas.attributes) do
				if ATMEConst.backTargetsAttributes[_id] ~= nil then
					table.insert(attributes, ATMEConst.backTargetsAttributes[_id].value)
				end
			end
			
			return attributes
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newAirbase) .. "getAttributes()", "already disabled", false)
		end
	end
	
	function newAirbase.hasAttributes(self, _attributes)
		if localMembers.destroyed == false then
			local fctDefinition = ATMEFct.getErrorPrefix(newAirbase) .. "hasAttributes(attributes)"
			
			if type(_attributes) == "string" or type(_attributes) == "table" then
				local err, datas = ATMEFct.extractDatasFromStringOrTable(_attributes)
		
				if err == false then
					return ATMEFct.hasAttributes(localMembers.objectDCS, datas)
				end
				
				ATMEFct.errorOutputCore(nil, fctDefinition, datas, false)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "attributes type must be string or table", true)
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end

	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end

	ATMEFct.outputCore(nil, "Create new airbase .. " .. name .. " " .. localMembers.categoryName .. " -> " .. localMembers.objectDCSId .. "(type " .. type(localMembers.objectDCSId) .. ") " .. ATMEConst.coalitionNames[localMembers.objectDCS:getCoalition()],2)
	
	return newAirbase
end

ATMEFct.C_Airbase_delete = function(_unit)
	if _unit:isExist() == true then
		if ATMEDatas.airbasesById[_unit:getID()] ~= nil then
			-- L'unité détruite est un FARP ou un bateau avec FARP ou porte-avion
			local localMembers = ATMEDatas.localMembers[ATMEDatas.airbasesById[_unit:getID()]]
			
			if localMembers.destroyed == false then
				localMembers.destroyed = true
				ATMEDatas.airbasesById[_unit:getID()] = nil
				
				ATMEFct.outputCore(nil, "Suppression entrée airbasesById .. " .. _unit:getName(), 2)
			end
		end
	end
end

setmetatable(ATME.C_Airbase,ATMEDatas.C_Airbase_MT)

ATMEDatas.C_Airbase_MT.mt.__metatable = "Private"
ATMEDatas.C_Airbase_MT.mt.__index = ATME.C_Airbase
ATMEDatas.C_Airbase_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Airbase_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Airbase : Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Zones : Gère les ensembles de zones DCS ou zones 2D/3D. 
-- Contructeur accessible aux modules externes.
-- Cette classe permet d'étendre la notion de zones par un ensemble de zones.
-- Une zone peut être liée à un objet pour les classes dérivées de ATME.C_MovableZone2D 
-- et ATME.C_MovableZone3D.
----------------------------------------------------------------------------------------------------------

ATMEFct.C_Zones_isInside = function(_zones, _point)
	local localMembersZones = ATMEDatas.localMembers[_zones:getIndex()]
	local isInside = false
	local zones = {}
	
	-- La table ATMEConst.indZoneConfig optimise les tests, les appels de fonctions
	-- et le passage de paramètres.
	
	if #localMembersZones.zonesList > 0 then
		for _id, _item in ipairs(localMembersZones.zonesList) do
			local localMembers = ATMEDatas.localMembers[_item.zone:getIndex()]
			local datas = {
				zone = _item.zone,
				actualCenter = ATMEFct.C_Zone_getActualCenter(localMembers),
				actualAzimuth = ATMEFct.C_Zone_getActualAzimuth(localMembers),
			}
			
			if ATMEFct[ATMEConst.indZoneConfig[_item.zone:getClassName()].fctIsInside](_point, datas) == true then
				isInside = true
				table.insert(zones, _id)
			end
		end
	end
	
	return isInside, zones
end

ATMEFct.C_Zones_getRandomPosition = function(_datas)
	local nbTries = 0
	local aborted = false

	repeat
		repeat
			local randomValue = math.random()
			local randomIndex = 1
			local found = false
			while found == false do
				if randomValue < _datas.localMembers.zonesList[randomIndex].max then
					found = true
				else
					randomIndex = randomIndex + 1
				end
			end

			_datas.point = ATMEFct.C_Zone_randomPositionTest(_datas.localMembers.zonesList[randomIndex].zone, _datas)
			nbTries = nbTries + 1
		until _datas.point ~= nil or nbTries % 10 == 0
		
		if _datas.localMembers.coroutines[_datas.moduleName] ~= nil and _datas.point == nil then
			aborted = _datas.localMembers.resultCallbacks[_datas.moduleName].aborted
			coroutine.yield()
		elseif _datas.localMembers.coroutines[_datas.moduleName] == nil and nbTries >= 100 then
			aborted = true
		end
	until _datas.point ~= nil or aborted == true
end

ATMEFct.C_Zones_createNamed = function(_namedTable, _name)
	if _namedTable[_name] ~= nil then
		ATMEFct.setError("ALREADY_CREATED")
		return true, "named zones manager already exists"
	else
		local d,f = string.find(_name, "|", 1)
		if d == nil then
			return false, ATMEFct.C_Zones_create(_name, _namedTable)
		else
			ATMEFct.setError("BAD_FORMAT")
			return true, "name can't contain '|' character"
		end
	end
end

ATMEFct.C_Zones_abortResult = function(_localMembers)
	if _localMembers.resultCallbacks[ATMEDatas.currentModuleName] ~= nil then
		if _localMembers.coroutines[ATMEDatas.currentModuleName] ~= nil then
			if _localMembers.resultCallbacks[ATMEDatas.currentModuleName].idTimer ~= nil then
				timer.removeFunction(_localMembers.resultCallbacks[ATMEDatas.currentModuleName].idTimer)
				_localMembers.resultCallbacks[ATMEDatas.currentModuleName].idTimer = nil
			end
			
			_localMembers.resultCallbacks[ATMEDatas.currentModuleName].aborted = true
			while coroutine.status(_localMembers.coroutines[ATMEDatas.currentModuleName]) ~= "dead" do
				if coroutine.status(_localMembers.coroutines[ATMEDatas.currentModuleName]) == "suspended" then
					coroutine.resume(_localMembers.coroutines[ATMEDatas.currentModuleName])
				end
			end

			_localMembers.coroutines[ATMEDatas.currentModuleName] = nil
			_localMembers.resultCallbacks[ATMEDatas.currentModuleName].aborted = false
		end
	end
end

ATMEFct.C_Zones_searchResultFct = {
	["dead"] = function(_datas)
		if _datas.localMembers.resultCallbacks[_datas.moduleName] ~= nil then
			xpcall(function() _datas.localMembers.resultCallbacks[_datas.moduleName].fct(_datas.zone, _datas.point) end, function() ATMEFct.errUserHandler('Result callback') end)
		end

		_datas.localMembers.coroutines[_datas.moduleName] = nil
		return false
	end,
	["suspended"]= function(_datas)
		local ok = coroutine.resume(_datas.localMembers.coroutines[_datas.moduleName])
		return ok
	end,
	["running"] = function(_datas)
		return true
	end
}

ATMETimers.C_Zones_searchResultTimer = function(_datas)
	if ATMEDatas.exitATME == false then
		local ok = ATMEFct.C_Zones_searchResultFct[coroutine.status(_datas.localMembers.coroutines[_datas.moduleName])](_datas)

		if ok == true then
			return timer.getTime() + 0.01
		end
		
		_datas.localMembers.resultCallbacks[_datas.moduleName].idTimer = nil
	end
end

ATMETimers.C_Zones_timer = function(_datas)
	if ATMEDatas.exitATME == false then
		if _datas.localMembers.tracking[_datas.name].localMembersObject.objectDCS:isExist() == true then
			ATMEFct.outputCore(nil, "Tracking sur zones ".. _datas.localMembers.tracking[_datas.name].name .. " - classe " .. _datas.localMembers.tracking[_datas.name].className , 2)
			
			if _datas.localMembers.tracking[_datas.name].localMembersObject.objectDCS:isActive() == true then
				local isInside
				local zonesId
				isInside, zonesId = ATMEFct.C_Zones_isInside(_datas.zones, _datas.localMembers.tracking[_datas.name].object:getPosition())
				
				if isInside == true then
					if _datas.localMembers.tracking[_datas.name].flag == nil or _datas.localMembers.tracking[_datas.name].flag  == false then
						_datas.localMembers.tracking[_datas.name].flag = true
						local datas = {
							isInside = true,
							object = _datas.localMembers.tracking[_datas.name].object,
							zonesID = zonesId,
						}
						
						for _moduleName, _callback in pairs(_datas.localMembers.tracking[_datas.name].callbacks) do
							ATMEDatas.currentModuleName = _moduleName
							
							ATMEFct.outputCore(nil,"Appel callback - Entrée zones pour unité " .. _datas.localMembers.tracking[_datas.name].name .. " pour le module " .. _moduleName, 2)
							xpcall(function() _callback(_datas.zones, datas) end, function() ATMEFct.errUserHandler('Tracking callback') end)
						end
					end
				else
					if _datas.localMembers.tracking[_datas.name].flag  == true then
						_datas.localMembers.tracking[_datas.name].flag = false
						local datas = {
							isInside = false,
							object = _datas.localMembers.tracking[_datas.name].object,
						}
						
						for _moduleName, _callback in pairs(_datas.localMembers.tracking[_datas.name].callbacks) do
							ATMEDatas.currentModuleName = _moduleName
							
							ATMEFct.outputCore(nil,"Appel callback - Sortie zones pour unité " .. _datas.localMembers.tracking[_datas.name].name .. " pour le module " .. _moduleName, 2)
							xpcall(function() _callback(_datas.zones, datas) end, function() ATMEFct.errUserHandler('Tracking callback') end)
						end
					end
				end
			end
			
			return timer.getTime() + 1
		else
			_datas.localMembers.tracking[_datas.name] = nil
		end
	end
end

ATMEDatas.C_Zones_MT = {
	className = "C_Zones",
	mt = {},
	
	keys = {
		["add"] = true,
		["isInside"] = true,
		["getIndex"] = true,
		["getRandomPosition"] = true,
		["getCount"] = true,
		["addTrackingCallback"] = true,
		["removeTrackingCallback"] = true,
		["addResultCallback"] = true,
		["removeResultCallback"] = true,
		["abortResultCallback"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["getZone"] = true,
		["getUserPrivateDatas"] = true,
	},
	
	add = function(self)
		local item = {
			randomRatio = 0,
			zonesList = {},
			tracking = {},
			userPrivateDatas = {},
			resultCallbacks = {},
			coroutines = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].tracking = nil
		ATMEDatas.localMembers[_id].zonesList = nil
		ATMEDatas.localMembers[_id].resultCallbacks = nil
		ATMEDatas.localMembers[_id].coroutines = nil
	end,
}

ATME.C_Zones = {
	getClassName = function(self)
		return ATMEDatas.C_Zones_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Zones_MT.parentClass
	end,
	
	createNamed = function(_name)
		local fctDefinition = "Function ATME.C_Zones.createNamed(name)"
		
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
			local err, newZones = ATMEFct.C_Zones_createNamed(localMembersModule.zones, _name)				
			
			if err == true then
				ATMEFct.warningOutputCore(nil, fctDefinition, newZones, false)
				return
			end
			
			return newZones
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
		end
	end,
	
	getByName = function(_name)
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Zones.getByName(name)", "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Zones.getByName(name)", "name type must be string", false)
		end
	end,
}

ATMEFct.C_Zones_create = function(_name, _zones)
	local newZones = {}

	setmetatable(newZones, ATMEDatas.C_Zones_MT.mt)
	
	local name
	local tagsName
	
	if _name ~= nil and _zones ~= nil then
		name = _name
		_zones[name] = newZones
		tagsName = ATMEFct.decompString(_name)
	end
	
	local index = ATMEDatas.C_Zones_MT:add()
	local localMembers = ATMEDatas.localMembers[index]
	
	function newZones.getName(self)
		return name
	end
	
	function newZones.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newZones) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newZones.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newZones.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newZones) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newZones.getUserPrivateDatas(self, _moduleName)
		local fctDefinition = ATMEFct.getErrorPrefix(newZones) .. "getUserPrivateDatas([moduleName])"
		local moduleName
		
		if _moduleName ~= nil then
			if type(_moduleName) == "string" then
				if localMembers.userPrivateDatas[_moduleName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
					return
				else
					moduleName = _moduleName
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
				return
			end
		else
			moduleName = ATMEDatas.currentModuleName
		end
		
		return localMembers.userPrivateDatas[moduleName]
	end
	
	function newZones.add(self, _zone)
		local item = {}
		local fctDefinition = ATMEFct.getErrorPrefix(newZones) .. "add(zone)"
	
		if type(_zone) == "string" then
			local zone = ATMEFct.getDCSZone(_zone)
			if zone ~= nil then
				item.zone = ATMEFct.C_Circle_create(nil, nil, zone.point, zone.radius)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone '" .. _zone .. "' unknown", false)
				return
			end
		elseif ATMEFct.isObjectClass(_zone, ATME.C_Zone) == true then	
			item.zone = _zone
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "zone type must be string, ATME.C_Circle, ATME.C_PickupZone, ATME.C_Rectangle, ATME.C_Polygon, ATME.C_Sphere or ATME.C_Box", false)
			return
		end
		
		local localMembersZone = ATMEDatas.localMembers[item.zone:getIndex()]
		
		localMembers.randomRatio = localMembers.randomRatio + localMembersZone.randomRatio
		table.insert(localMembers.zonesList, item)
		
		localMembersZone = ATMEDatas.localMembers[localMembers.zonesList[1].zone:getIndex()]
		localMembers.zonesList[1].max = localMembersZone.randomRatio / localMembers.randomRatio
		localMembersZone = ATMEDatas.localMembers[localMembers.zonesList[#localMembers.zonesList].zone:getIndex()]
		localMembers.zonesList[#localMembers.zonesList].max = 1
		
		for _i = 2, #localMembers.zonesList - 1 do
			localMembersZone = ATMEDatas.localMembers[localMembers.zonesList[_i].zone:getIndex()]
			localMembers.zonesList[_i].max = localMembers.zonesList[_i - 1].max + localMembersZone.randomRatio / localMembers.randomRatio
		end
		
		return #localMembers.zonesList
	end
	
	function newZones.getZone(self, _index)
		local fctDefinition = ATMEFct.getErrorPrefix(newZones) .. "getZone(index)"
		
		if type(_index) == "number" then
			local index = math.floor(_index)
			if index > 0 then
				if localMembers.zonesList[index] ~= nil then
					return localMembers.zonesList[index].zone
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "index must be > 0", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "index type must be number", true)
		end
	end

	function newZones.isInside(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newZones) .. "isInside(reference)"
		local isInside = false
		local zonesId
		
		if ATMEFct.isReference(_reference) == true then		

			isInside, zonesId = ATMEFct.C_Zones_isInside(self, ATMEFct.getPointFromReference(_reference))

			return isInside, zonesId
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
		end
	end
	
	function newZones.getCount(self)
		return #localMembers.zonesList
	end
	
	function newZones.getIndex(self)
		return index
	end
	
	function newZones.addResultCallback(self, _fct)
		local fctDefinition = ATMEFct.getErrorPrefix(newZones) .. "addResultCallback(fct)"

		if type(_fct) == "function" then
			localMembers.resultCallbacks[ATMEDatas.currentModuleName] = {
				fct = _fct,
				aborted = false,
			}
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
		end
	end
	
	function newZones.removeResultCallback(self)
		ATMEFct.C_Zones_abortResult(localMembers)
		localMembers.resultCallbacks[ATMEDatas.currentModuleName] = nil
	end
	
	function newZones.abortResultCallback(self)
		ATMEFct.C_Zones_abortResult(localMembers)
	end
	
	function newZones.getRandomPosition(self, _surfaceFilter, _isFlat, _altMin)
		local fctDefinition = ATMEFct.getErrorPrefix(newZones) .. "getRandomPosition([surfaceFilter], [flat], [altMin])"
	
		if (_surfaceFilter == nil or type(_surfaceFilter) == "string") and
					(_isFlat == nil or type(_isFlat) =="boolean") and (_altMin == nil or type(_altMin) =="number") then

			if #localMembers.zonesList > 0 then
				local surfaceFilter = "ALL"
				
				if _surfaceFilter ~= nil then
					surfaceFilter = string.upper(_surfaceFilter)
				end
				
				if ATMEConst.surfaceType[surfaceFilter] ~= nil then
					local datas = {
						zone = self,
						surfaceFilter = surfaceFilter,
						isFlat = false,
						altMin = 0,
						localMembers = localMembers,
						moduleName = ATMEDatas.currentModuleName,
					}
					
					if _isFlat ~= nil then
						datas.isFlat = _isFlat
					end
					
					if type(_altMin) == "number" then
						datas.altMin = _altMin
					end
					
					if localMembers.resultCallbacks[ATMEDatas.currentModuleName] == nil then
						ATMEFct.C_Zones_getRandomPosition(datas)
						return datas.point
					else
						if localMembers.coroutines[ATMEDatas.currentModuleName] == nil then
							ATMEFct.outputCore(nil, "newZones.getRandomPosition appel asynchrone", 2)
							localMembers.coroutines[ATMEDatas.currentModuleName] = coroutine.create(ATMEFct.C_Zones_getRandomPosition)
							coroutine.resume(localMembers.coroutines[ATMEDatas.currentModuleName], datas)
							localMembers.resultCallbacks[ATMEDatas.currentModuleName].idTimer = timer.scheduleFunction(ATMETimers.C_Zones_searchResultTimer, datas, timer.getTime() + 0.001)
							return
						else
							ATMEFct.setError("ALREADY_STARTED")
							ATMEFct.warningOutputCore(nil, fctDefinition, "search already started", false)
							return
						end
					end
				end
				
				ATMEFct.errorOutputCore(nil, fctDefinition, "surfaceFilter must be 'ALL', 'WATER' or 'LAND'", false)
			end
		else
			if _surfaceFilter ~= nil and type(_surfaceFilter) ~= "string" then 
				ATMEFct.errorOutputCore(nil, fctDefinition, "surfaceFilter type must be nil or string", true)
			elseif _isFlat ~= nil and type(_isFlat) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "flat type must be nil or boolean", false)
			elseif _altMin ~= nil and type(_altMin) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "altMin type must be nil or number", false)
			end
		end
	end
	
	function newZones.addTrackingCallback(self, _fct, _object)
		local fctDefinition = ATMEFct.getErrorPrefix(newZones) .. "addTrackingCallback(fct, object)"

		if type(_fct) == "function" and (ATMEFct.isObjectClass(_object, ATME.C_Unit) == true or ATMEFct.isObjectClass(_object, ATME.C_Weapon) == true) then
			local className = _object:getClassName()
			local name = _object:getName()
			if localMembers.tracking[name] == nil then
				localMembers.tracking[name] = {
					className = className,
					name = name,
					object = _object,
					localMembersObject = ATMEDatas.localMembers[_object:getIndex()],
					callbacks = {},
				}
			end
				
			localMembers.tracking[name].pos = _object:getPosition()
			
			localMembers.tracking[name].callbacks[ATMEDatas.currentModuleName] = _fct
			
			local datas = {
				name = name,
				localMembers = localMembers,
				zones = self,
			}
			
			timer.scheduleFunction(ATMETimers.C_Zones_timer, datas, timer.getTime() + ATMEConst.offsetTracking[className])
		else
			if type(_fct) ~= "function" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
			elseif ATMEFct.isObjectClass(_object, ATME.C_Unit) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "object type must be ATME.C_Unit or ATME.C_Weapon", false)
			end
		end
	end
	
	function newZones.removeTrackingCallback(self, _object)
		local fctDefinition = ATMEFct.getErrorPrefix(newZones) .. "removeTrackingCallback(object)"
		
		if ATMEFct.isObjectClass(_object, ATME.C_Unit) == true or ATMEFct.isObjectClass(_object, ATME.C_Weapon) == true then
			if localMembers.tracking[_object:getName()] ~= nil then
				localMembers.tracking[_object:getName()].callbacks[ATMEDatas.currentModuleName] = nil
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "object type must be ATME.C_Unit or ATME.C_Weapon", true)
		end
	end

	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end

	return newZones
end

setmetatable(ATME.C_Zones,ATMEDatas.C_Zones_MT)

ATMEDatas.C_Zones_MT.__call = function(t)
	if ATMEDatas.startMissionOk < 2 then
		ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
		return
	end
	
	return ATMEFct.C_Zones_create(nil, nil)
end

ATMEDatas.C_Zones_MT.mt.__metatable = "Private"
ATMEDatas.C_Zones_MT.mt.__index = ATME.C_Zones
ATMEDatas.C_Zones_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Zones_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Zones : Bad key " .. k, "table locked", false)
	end
end


-----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Sound : gère les fichiers son définis dans la mission
--
-----------------------------------------------------------------------------------------------------------

ATMEFct.C_Sound_getPlaylist = function(_text, _random, _special, _moduleName)
	local tokens = ATMEFct.decompStringSpaceOnly(_text)
	local moduleName = ATMEDatas.currentModuleName
	if _moduleName ~= nil then
		moduleName = _moduleName
	end
	
	if tokens ~= nil then
		local playlist = {}

		for _, _token in ipairs(tokens) do
			local _, _, temp = _token:find("([%d%a_]+)")
			
			if temp == _token then
				local idToken = _token:upper()
				if ATMEDatas.modulesLocalMembers[moduleName].sounds[idToken] ~= nil then
					local localMembers = ATMEDatas.localMembers[ATMEDatas.modulesLocalMembers[moduleName].sounds[idToken]:getIndex()]

					table.insert(playlist, {
												fileName = localMembers.fileName,
												tipsBase =  localMembers.tipsBase,
												tips = localMembers.tips,
												pause = false,
												spell = false,
												subtitle = {
													text = localMembers.subtitle.text,
													duration = localMembers.subtitle.duration,
												},
											})
				else
					ATMEFct.setError("BAD_TOKEN")
					return true, "token " .. _token .. " - Sound doesn't exist"
				end
			elseif _special == nil then
				_, _, temp = _token:find("%((%d+.?%d*)%)")

				if temp ~= nil then  -- Cas de la pause
					table.insert(playlist, {
												fileName = "",
												pause = true,
												spell = false,
												tipsBase =  tonumber(temp),
												tips =  math.ceil(tonumber(temp)),
												subtitle = {
													text = "",
													duration = 0,
												},
											})
				else
					_, _, temp = _token:find("{(%d+.?%d*)}")
					local ok = false
					
					if temp ~= nil then -- Cas nombre à épeler
						ok = true
					else
						_, _, temp = _token:find("{([%a%d]+)}")
						
						if temp ~= nil then -- Cas lettres et chiffres à épeler 
							temp = temp:upper()
							ok = true
						end
					end
					
					if ok == true then
						local lg = temp:len()
						local endCar = ""
						for i = 1, lg do
							local id = temp:sub(i,i)
							
							if id == "." then
								id = "DECIMAL"
							end
							
							if ATMEDatas.modulesLocalMembers[moduleName].sounds[id] ~= nil then
								if i == lg then
									endCar = " "
								end
								local localMembers = ATMEDatas.localMembers[ATMEDatas.modulesLocalMembers[moduleName].sounds[id]:getIndex()]
								table.insert(playlist, {
															fileName = localMembers.fileName,
															tipsBase =  localMembers.tipsBase,
															tips = localMembers.tips,
															pause = false,
															spell = true,
															subtitle = {
																text = localMembers.subtitle.text .. endCar,
																duration = localMembers.subtitle.duration,
															},
														})
							else
								ATMEFct.setError("BAD_TOKEN")
								return true, "Token " .. _token .. " spelling '" .. temp:sub(i,i) .. "' - Sound " .. id .. " is missing"
							end
						end
						
						table.insert(playlist, {
											fileName = "",
											pause = true,
											spell = true,
											tipsBase =  0.3,
											tips =  1,
											subtitle = {
												text = "",
												duration = 0,
											},
										})
					else
						local dirlatS, dlatS, mlatS, slatS, ddlatS, dirlonS, dlonS, mlonS, slonS, ddlonS, tOut, pOut
						local dirlat, dlat, mlat, slat, ddlat, dirlon, dlon, mlon, slon, ddlon
						
						_, _, dirlatS, dlatS, mlatS, slatS, ddlatS, dirlonS, dlonS, mlonS, slonS, ddlonS, pOut = _token:find("%[([-+]?)(%d+):(%d+):(%d+).?(%d*)/([+-]?)(%d+):(%d+):(%d+).?(%d*),([sp])%]")

						if dirlatS == nil then
							_, _, dirlatS, dlatS, mlatS, slatS, dirlonS, dlonS, mlonS, slonS, pOut = _token:find("%[([-+]?)(%d+):(%d+).?(%d*)/([+-]?)(%d+):(%d+).?(%d*),([sp])%]")
							
							if dirlatS == nil then
								local nb1, nb2
								
								_, _, nb1, nb2, tOut, pOut = _token:find("%[([+-]?%d+.?%d*)/([+-]?%d+.?%d*),([MS])([sp])%]")
								
								if nb1 ~= nil then
									if tOut == "S" then
										dirlat, dlat, mlat, slat, ddlat  = ATMEFct.convertToDMS(tonumber(nb1))
										dirlon, dlon, mlon, slon, ddlon  = ATMEFct.convertToDMS(tonumber(nb2))
									else
										dirlat, dlat, mlat, slat  = ATMEFct.convertToDMd(tonumber(nb1), true)
										dirlon, dlon, mlon, slon  = ATMEFct.convertToDMd(tonumber(nb2), true)
									end
								end
							end
						end
						
						if dirlatS ~= nil then
							dirlat = 1
							if dirlatS == "-" then
								dirlat = -1
							end
							
							dirlon = 1
							if dirlonS == "-" then
								dirlon = -1
							end

							if ddlatS ~= nil then
								if ddlatS == "" then
									ddlat = 0
								else
									ddlat = tonumber(ddlatS)
								end
							else
								local lg = slatS:len()
								for _ = lg, 3 do
									slatS = slatS .. "0"
								end
							end
							
							dlat = tonumber(dlatS)
							mlat = tonumber(mlatS)
							slat = tonumber(slatS)
							
							if ddlonS ~= nil then
								if ddlonS == "" then
									ddlon = 0
								else
									ddlon = tonumber(ddlonS)
								end
							else
								local lg = slonS:len()
								for _ = lg, 3 do
									slonS = slonS .. "0"
								end
							end
							
							dlon = tonumber(dlonS)
							mlon = tonumber(mlonS)
							slon = tonumber(slonS)
						end
						
						
						if dirlat ~= nil then
							if dlon <= 180 and dlat <= 90 and mlat < 60 and mlon < 60 and ((ddlon == nil and slat < 10000 and slon < 10000) or (ddlon ~= nil and slat < 60 and slon < 60)) then
								-- OK coordonnées
								local function readDigits(_coord, _pause, _extraCar, _endCar)
									local lg = _coord:len()

									for i = 1, lg do
										local id = _coord:sub(i,i)

										if id == "." then
											id = "DECIMAL"
										end
															
										if ATMEDatas.modulesLocalMembers[moduleName].sounds[id] ~= nil then
											local localMembers = ATMEDatas.localMembers[ATMEDatas.modulesLocalMembers[moduleName].sounds[id]:getIndex()]
											
											local txt = localMembers.subtitle.text
											
											if i == lg then
												txt = txt .. _extraCar
											end

											table.insert(playlist, {
																		fileName = localMembers.fileName,
																		tipsBase =  localMembers.tipsBase,
																		tips = localMembers.tips,
																		pause = false,
																		spell = true,
																		subtitle = {
																			text = txt,
																			duration = localMembers.subtitle.duration,
																		},
																	})
										else
											ATMEFct.setError("BAD_TOKEN")
											return true, "Token '" .. _token .. "' spelling '" .. _token:sub(i,i) .. "' - Sound " .. id .. " is missing"
										end
									end

									if _pause ~= nil then
										table.insert(playlist, {
											fileName = "",
											pause = true,
											spell = true,
											tipsBase =  _pause,
											tips =  math.ceil(_pause),
											subtitle = {
												text = "",
												duration = 0,
											},
										})
									end
									
									return false
								end
								
								local function buildMS(_ml, _sl, _ddl, _pause, _extraCar)
									if _ddl ~= nil then
										local err, errText = readDigits(string.format("%02d", _ml), 0.5, "'")
										if err == true then
											return err, errText
										end

										if pOut == "p" then
											local err, errText = readDigits(string.format("%02d", _sl) .. "." .. string.format("%02d", _ddl), _pause, _extraCar)
											if err == true then
												return err, errText
											end
										else
											local err, errText = readDigits(string.format("%02d", _sl), _pause, _extraCar)
											if err == true then
												return err, errText
											end
										end
									else
										if pOut == "p" then
											local err, errText = readDigits(string.format("%02d", _ml) .. "." .. string.format("%04d", _sl), _pause, _extraCar)
											if err == true then
												return err, errText
											end
										else
											local err, errText = readDigits(string.format("%02d", _ml) .. "." .. string.format("%02d", math.floor(_sl/100)), _pause, _extraCar)
											if err == true then
												return err, errText
											end
										end
									end
									
									return false
								end

								local cardinal = "N"
								if dirlat == -1 then
									cardinal = "S"
								end
								
								if ATMEDatas.modulesLocalMembers[moduleName].sounds[ATMEConst.cardinal[cardinal]] == nil then
									ATMEFct.setError("BAD_TOKEN")
									return true, "Token '" .. _token .. "' - Sound " .. ATMEConst.cardinal[cardinal] .. " is missing"
								end
								
								local localMembers = ATMEDatas.localMembers[ATMEDatas.modulesLocalMembers[moduleName].sounds[ATMEConst.cardinal[cardinal]]:getIndex()]
								table.insert(playlist, {
									fileName = localMembers.fileName,
									tipsBase =  localMembers.tipsBase,
									tips = localMembers.tips,
									pause = false,
									spell = true,
									subtitle = {
										text = localMembers.subtitle.text,
										duration = localMembers.subtitle.duration,
									},
								})
								
								table.insert(playlist, {
											fileName = "",
											pause = true,
											spell = true,
											tipsBase =  0.5,
											tips =  1,
											subtitle = {
												text = "",
												duration = 0,
											},
										})

								local err, errText = readDigits(string.format("%02d", dlat), 0.5, "°")
								if err == true then
									return err, errText
								end

								buildMS(mlat, slat, ddlat, 0.5, " / ")

								cardinal = "E"
								if dirlon == -1 then
									cardinal = "W"
								end
								
								if ATMEDatas.modulesLocalMembers[moduleName].sounds[ATMEConst.cardinal[cardinal]] == nil then
									ATMEFct.setError("BAD_TOKEN")
									return true, "Token '" .. _token .. "' - Sound " .. ATMEConst.cardinal[cardinal] .. " is missing"
								end
								
								local localMembers = ATMEDatas.localMembers[ATMEDatas.modulesLocalMembers[moduleName].sounds[ATMEConst.cardinal[cardinal]]:getIndex()]
								table.insert(playlist, {
									fileName = localMembers.fileName,
									tipsBase =  localMembers.tipsBase,
									tips = localMembers.tips,
									pause = false,
									spell = true,
									subtitle = {
										text = localMembers.subtitle.text,
										duration = localMembers.subtitle.duration,
									},
								})
								
								table.insert(playlist, {
											fileName = "",
											pause = true,
											spell = true,
											tipsBase =  0.5,
											tips =  1,
											subtitle = {
												text = "",
												duration = 0,
											},
										})
								
								local err, errText = readDigits(string.format("%03d", dlon), 0.5, "°")
								

								if err == true then
									return err, errText
								end
								
								buildMS(mlon, slon, ddlon, nil, " ")

								table.insert(playlist, {
											fileName = "",
											pause = true,
											spell = true,
											tipsBase =  0.5,
											tips =  1,
											subtitle = {
												text = "",
												duration = 0,
											},
										})
							else
								ATMEFct.setError("BAD_FORMAT")
								return true, "Token '" .. _token .. "' - bad coordinates"
							end
						else
							ATMEFct.setError("BAD_FORMAT")
							return true, "Token '" .. _token .. "' - bad format"
						end
					end
				end
			end

			if _special == true then
				break
			end
		end
						
		if #playlist > 1 and _random == false then
			-- Si playlist > 1 alors les sous titres sont regroupés dans le premier item
			local spelling = playlist[1].spell
			for i = 2, #playlist do
				if (spelling == true and playlist[i].spell == true) or playlist[1].subtitle.text:sub(-1, -1) == "\n" or playlist[i].subtitle.text:sub(1, 1) == "\n" or
					playlist[1].subtitle.text:sub(-1, -1) == " " or playlist[i].subtitle.text:sub(1, 1) == " " then
					playlist[1].subtitle.text = playlist[1].subtitle.text .. playlist[i].subtitle.text
				elseif playlist[i].subtitle.text ~= "" then
					playlist[1].subtitle.text = playlist[1].subtitle.text .. " " .. playlist[i].subtitle.text
				end
				
				playlist[i].subtitle.text = ""
				playlist[1].subtitle.duration = playlist[1].subtitle.duration + playlist[i].subtitle.duration
				playlist[i].subtitle.duration = 0
				
				spelling = playlist[i].spell
			end
		elseif #playlist == 0 then
			ATMEFct.setError("NO_TOKEN")
			return true, "text string has no sound token"
		end
		
		return false, playlist
	end
	
	ATMEFct.setError("NO_TOKEN")
	return true, "text string has no sound token"
end

ATMEDatas.C_Sound_MT = {
	className = "C_Sound",
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["getID"] = true,
		["getSoundDatas"] = true,
		["getSubtitleDatas"] = true,
		["getUserPrivateDatas"] = true,
	},
	
	add = function(self, _fileName, _tips, _tipsBase)
		local item = {
			fileName = _fileName,
			tipsBase = _tipsBase,
			tips = _tips,
			subtitle = {
				text = "",
				duration = 0,
			},
			userPrivateDatas = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
	end,
}

ATME.C_Sound = {
	getClassName = function(self)
		return ATMEDatas.C_Sound_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Sound_MT.parentClass
	end,
	
	getByID = function(_id)
		if type(_id) == "string" then
			return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].sounds[string.upper(_id)]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Sound.getByID(id)", "id type must be string", false)
		end
	end,
}

ATMEFct.C_Sound_create = function(_id, _tips, _tipsBase, _fileName, _subtitle)	
	local newSound = {}
	local id = _id

	
	local index = ATMEDatas.C_Sound_MT:add(_fileName, _tips, _tipsBase)
	local localMembers = ATMEDatas.localMembers[index]

	local durationSec = _tips
	local roundDurationSec = math.ceil(durationSec)
	
	setmetatable(newSound, ATMEDatas.C_Sound_MT.mt)

	function newSound.getIndex(self)
		return index
	end
	
	function newSound.getID(self)
		return id
	end
	
	function newSound.getSoundDatas(self)
		return localMembers.fileName, durationSec
	end
	
	function newSound.getSubtitleDatas(self)
		return localMembers.subtitle.text, localMembers.subtitle.duration
	end
	
	function newSound.getUserPrivateDatas(self, _moduleName)
		local fctDefinition = ATMEFct.getErrorPrefix(newSound) .. "getUserPrivateDatas([moduleName])"
		local moduleName
		
		if _moduleName ~= nil then
			if type(_moduleName) == "string" then
				if localMembers.userPrivateDatas[_moduleName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
					return
				else
					moduleName = _moduleName
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
				return
			end
		else
			moduleName = ATMEDatas.currentModuleName
		end
		
		return localMembers.userPrivateDatas[moduleName]
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	localMembers.subtitle.text = _subtitle.text
	localMembers.subtitle.duration = _subtitle.duration

	ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].sounds[id] = newSound
	
	return newSound
end

setmetatable(ATME.C_Sound,ATMEDatas.C_Sound_MT)

ATMEDatas.C_Sound_MT.__call = function(t, _id, ...)
	local fctDefinition = "Constructor ATME.C_Sound(ID, fileName, duration)\nATME.C_Sound(ID, fileName, duration, subtitle, subtitleDuration)"
	local args = {...}
	
	if ATMEDatas.startMissionOk < 2 then
		ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
		return
	end
	if type(_id) == "string" then
		local tokens = ATMEFct.decompStringSpaceOnly(_id)
			
		if tokens ~= nil then
			local d, f, temp = tokens[1]:find("([%d%a_]+)")
			
			if temp == tokens[1] then
				if #tokens > 1 then
					ATMEFct.errorOutputCore(nil, fctDefinition, "ID consist of letters, numbers and '_' char", false)
				else
					local tokenName = string.upper(tokens[1])
					if ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].sounds[tokenName] ~= nil then
						ATMEFct.errorOutputCore(nil, fctDefinition, "ID " .. tokens[1] .. " already exists", false)
					else
						local tips
						local fileName
						local subtitle
						
						if #args >= 2 then
							if type(args[1]) == "string" and type(args[2]) == "number" then
								fileName = args[1]
								tips = math.floor(args[2] + 1)
								
								if #args == 4 then
									if type(args[3]) == "string" and type(args[4]) == "number" then
										local subtitleDuration = math.floor(args[4])
										
										if subtitleDuration >= 1 then
											subtitle = {
												text = args[3],
												duration = args[4],
											}
										else
											ATMEFct.errorOutputCore(nil, fctDefinition, "durationSubtitle must be >= 1", false)
											return
										end
									else
										if type(args[3]) ~= "string" then
											ATMEFct.errorOutputCore(nil, fctDefinition, "subtitle type must be string", false)
											return
										elseif type(args[4]) ~= "number" then
											ATMEFct.errorOutputCore(nil, fctDefinition, "durationSubtitle type must be number", false)
											return
										end
									end
								elseif #args == 2 then
									subtitle = {
										text = "",
										duration = 0,
									}
								else
									ATMEFct.errorOutputCore(nil, fctDefinition, "bad parameters", false)
									return
								end
							else
								if type(args[1]) ~= "string" then
									ATMEFct.errorOutputCore(nil, fctDefinition, "fileName type must be string", false)
									return
								elseif type(args[2]) ~= "number" then
									ATMEFct.errorOutputCore(nil, fctDefinition, "duration type must be number", false)
									return
								end
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "bad parameters", false)
							return
						end

						if args[2] > 0 then
							return ATMEFct.C_Sound_create(tokenName, tips, args[2], fileName, subtitle)
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "duration must be >= 0.1", false)
						end
					end
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "ID " .. token[1] .. " - bad format", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "no ID defined", false)
		end
	else
		if type(_id) ~= "string" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "ID type must be string", false)
		elseif type(_duration) ~= "number" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "duration type must be number", false)
		elseif _fileName ~= nil and type(_fileName) ~= "string" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "fileName type must be nil or string", false)
		end
	end
end

ATMEDatas.C_Sound_MT.mt.__metatable = "Private"
ATMEDatas.C_Sound_MT.mt.__index = ATME.C_Sound
ATMEDatas.C_Sound_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Sound_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Sound : Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Mark : gère les marqueurs créés au cours de la mission
----------------------------------------------------------------------------------------------------------

ATMEDatas.C_Mark_MT = {
	className = "C_Mark",
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["getID"] = true,
		["getDatas"] = true,
		["getUserPrivateDatas"] = true,
	},
	
	add = function(self, _text, _point)
		local item = {
			text = _text,
			point = {
				x = _point.x,
				y = _point.y,
				z = _point.z,
			},
			userPrivateDatas = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
	end,
}

ATME.C_Mark = {
	getClassName = function(self)
		return ATMEDatas.C_Mark_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Mark_MT.parentClass
	end,
	
	getByID = function(_id)
		if type(_id) == "number" then
			return ATMEDatas.marks[math.floor(_id)]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Mark.getByID(id)", "id type must be number", false)
		end
	end,
}

ATMEFct.C_Mark_create = function(_id, _text, _point)	
	local newMark = {}
	local id = _id

	
	local index = ATMEDatas.C_Mark_MT:add(_text, _point)
	local localMembers = ATMEDatas.localMembers[index]
	
	setmetatable(newMark, ATMEDatas.C_Mark_MT.mt)

	function newMark.getIndex(self)
		return index
	end
	
	function newMark.getID(self)
		return id
	end
	
	function newMark.getDatas(self)
		return localMembers.text, {
			x = localMembers.point.x,
			y = localMembers.point.y,
			z = localMembers.point.z,
		}
	end
	

	function newMark.getUserPrivateDatas(self, _moduleName)
		local fctDefinition = ATMEFct.getErrorPrefix(newMark) .. "getUserPrivateDatas([moduleName])"
		local moduleName
		
		if _moduleName ~= nil then
			if type(_moduleName) == "string" then
				if localMembers.userPrivateDatas[_moduleName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
					return
				else
					moduleName = _moduleName
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
				return
			end
		else
			moduleName = ATMEDatas.currentModuleName
		end
		
		return localMembers.userPrivateDatas[moduleName]
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	

	ATMEDatas.marks[id] = newMark
	
	ATMEFct.outputCore(nil,"Create Mark id = " .. id, 2)
	
	return newMark
end

ATMEFct.C_Mark_delete = function(_id)
	ATMEDatas.C_Mark_MT:remove(ATMEDatas.marks[_id]:getIndex())
	ATMEDatas.marks[_id] = nil
	
	ATMEFct.outputCore(nil,"Delete Mark id = " .. _id, 2)
end

setmetatable(ATME.C_Mark,ATMEDatas.C_Mark_MT)

ATMEDatas.C_Mark_MT.mt.__metatable = "Private"
ATMEDatas.C_Mark_MT.mt.__index = ATME.C_Mark
ATMEDatas.C_Mark_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Mark_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Mark : Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Menu : gère les menus accessibles par F10. Constructeur accessible aux modules externes 
-- Cette classe est complexe et gère :
--			- Les entrées d'un menu (menu de base root ou sous menu). Un entrée engendre une action par l'appel
--			  d'une fonction.
--			- Les sous-menus rattachés :
--					-> au menu de base root (accessible directement par F10)
--					-> à un menu, lui même sous-menu du menu de base rout ou d'un autre menu.
--			- L'affichage dans l'ordre choisi (alpha par défaut) des sous-menus du menu courant et de ses entrées
--					-> sous menus d'abord
--					-> entrées ensuite
--			- La gestion des entrées par une identification de type groupId pour une gestion en groupe :
--					-> lors de l'affichage dans l'ordre par groupId puis alpha si groupId identique
--					-> pour la suppression de plusieurs entrées  simultanément.
--					-> le groupId d'une entrée s'applique à toutes les entrées d'un menu donné, y compris 
--					   pour les entrées de ses sous-menus
--			- Un menu vide n'est plus affiché. Un menu est vide si lui-même et ses sous-menus ne contiennent
--			  aucune entrée
----------------------------------------------------------------------------------------------------------

ATMEDatas.C_Menu_MT = {
	className = "C_Menu",
	mt = {},
	
	keys = {
		["getParent"] = true,
		["isRoot"] = true,
		["getLabel"] = true,
		["getIndex"] = true,
		["getGroup"] = true,
		["append"] = true,
		["removeAll"] = true,
		["remove"] = true,
		["getNbEntries"] = true,
		["getUserPrivateDatas"] = true,
	},
	
	add = function(self, _group)
		local item = {
			itemCommandsList = {},
			itemSubMenusList = {},
			isActive = false,
			path = nil,
			nbSubmenus = 0,
			erased = false,									-- Utilisé pour les données privées. Passe à true si le menu n'est pas affiché
			handlers = ATMEFct.C_List_create(),		-- Table des handlers définis (proxyHandler d'appel)
			userPrivateDatas = {},
			group = _group,
			localMembersGroup = ATMEDatas.localMembers[_group:getIndex()],
		}

		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].group = nil
		ATMEDatas.localMembers[_id].localMembersGroup = nil
	end,
}


ATME.C_Menu = {
	getClassName = function(self)
		return ATMEDatas.C_Menu_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Menu_MT.parentClass
	end,
}

ATMEFct.C_Menu_handler = function(_args)
	local localMembers = ATMEDatas.localMembers[_args.indexMenu] 
	local tempModuleName = ATMEDatas.currentModuleName
	
	ATMEDatas.currentModuleName = _args.moduleName
	ATMEFct.outputCore(nil,"Appel handler du menu '" .. _args.labelMenu .. "' du module " .. ATMEDatas.currentModuleName, 2)
	xpcall(function() localMembers.handlers:get(_args.idHandler)(_args.menu, _args.args) end, function() ATMEFct.errUserHandler(_args.labelMenu, true) end)
	
	ATMEDatas.currentModuleName = tempModuleName
end

ATMEFct.C_Menu_create = function(_group, _label, _parent, _moduleName)
	local found = false
	local nbEntries = 0

	if _parent ~= nil then
		local localMembersParent = ATMEDatas.localMembers[_parent:getIndex()]
		if _parent:isRoot() == true then -- Root menu direct avec _moduleName défini
			for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
				for _k, _itemSubMenu in pairs(localMembersParent.itemSubMenusList) do
					if _itemSubMenu:getLabel() == _label then
						found = true
					end
				end
			end
		else -- Sous menu
			for _k, _itemSubMenu in pairs(localMembersParent.itemSubMenusList) do
				if _itemSubMenu:getLabel() == _label then
					found = true
				end
			end
		end
		nbEntries = _parent:getNbEntries()
	end

	if nbEntries < 10 and found == false then
		local parent = _parent
		local isRoot = false
		local label = _label
		
		local newF10SubMenu = {}
		
		setmetatable(newF10SubMenu, ATMEDatas.C_Menu_MT.mt)
			
		local index = ATMEDatas.C_Menu_MT:add(_group)
		local localMembers = ATMEDatas.localMembers[index]
		
		-- Si _moduleName ~= nil -> Menu root, à mémoriser pour le joueur
		-- Plusieurs arborescences root peuvent être créées pour un joueur donné
		if _parent == nil then
			isRoot = true
			label = "Root"
		end
		
		function newF10SubMenu.getParent(self)
			return parent
		end
		
		function newF10SubMenu.isRoot(self)
			return isRoot
		end
		
		function newF10SubMenu.getLabel(self)
			return label
		end
		
		function newF10SubMenu.getIndex(self)
			return index
		end
		
		function newF10SubMenu.getGroup(self)
			return localMembers.group
		end
		
		function newF10SubMenu.getNbEntries(self)
			local nbEntries = 0
			
			if self:isRoot() == true then
				for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
					local menuRoot = localMembers.localMembersGroup.itemsMenuCommandRoot[_localModuleMembers.name]
					local localMembersRoot = ATMEDatas.localMembers[menuRoot:getIndex()]
					nbEntries = nbEntries + localMembersRoot.nbSubmenus + #localMembersRoot.itemCommandsList
				end
			else
				nbEntries = localMembers.nbSubmenus + #localMembers.itemCommandsList
			end
			
			return nbEntries
		end
		
		local function C_Menu_deleteAllCommands(_itemMenu, _deleteSubMenu)
			ATMEFct.outputCore(nil, "C_Menu_deleteAllCommands Start ->" .. _itemMenu:getLabel(), 3)
			
			local localMembersSub = ATMEDatas.localMembers[_itemMenu:getIndex()]
			while #localMembersSub.itemCommandsList > 0 do
				local toRemove= table.remove(localMembersSub.itemCommandsList)
				missionCommands.removeItemForGroup(localMembersSub.localMembersGroup.objectDCSId, toRemove.itemMenuCommandId)
				localMembersSub.handlers:remove(toRemove.idHandler)
			end
			
			for _k, _itemSubMenu in pairs(localMembersSub.itemSubMenusList) do
				C_Menu_deleteAllCommands(_itemSubMenu, true)
				localMembersSub.itemSubMenusList[_k] = nil
				localMembersSub.nbSubmenus = localMembersSub.nbSubmenus - 1
			end
		  
			if _deleteSubMenu == true and _itemMenu:isRoot() == false then
				missionCommands.removeItemForGroup(localMembersSub.localMembersGroup.objectDCSId, localMembersSub.path)
				localMembersSub.erased = true
			end

			ATMEFct.outputCore(nil, "C_Menu_deleteAllCommands End -> " .. _itemMenu:getLabel(), 3)
		end

		local function C_Menu_deleteCommands(_itemMenu, _groupId, _deleteSubMenu)
			local function deleteCommands(_itemMenu, _groupId)
				local erase = false

				ATMEFct.outputCore(nil, "  deleteCommands Start on Id " .. _groupId .. " -> " .. _itemMenu:getLabel(), 3)
			  
				local localMembersSub = ATMEDatas.localMembers[_itemMenu:getIndex()]
				local i = 1
				while i <= #localMembersSub.itemCommandsList do
					if localMembersSub.itemCommandsList[i].groupId == _groupId then    
						missionCommands.removeItemForGroup(localMembersSub.localMembersGroup.objectDCSId, localMembersSub.itemCommandsList[i].itemMenuCommandId)
						localMembersSub.handlers:remove(localMembersSub.itemCommandsList[i].idHandler)
						table.remove(localMembersSub.itemCommandsList, i)
					else
						i = i + 1
					end
				end
				
				for _k, _itemSubMenu in pairs(localMembersSub.itemSubMenusList) do
					if deleteCommands(_itemSubMenu, _groupId) == true then
						localMembersSub.itemSubMenusList[_k] = nil
						localMembersSub.nbSubmenus = localMembersSub.nbSubmenus - 1
					end
				end
				
				if #localMembersSub.itemCommandsList == 0 and _deleteSubMenu == true and _itemMenu:isRoot() == false then
					missionCommands.removeItemForGroup(localMembersSub.localMembersGroup.objectDCSId, localMembersSub.path)
					localMembersSub.erased = true
					erase = true
					
					ATMEFct.outputCore(nil, "    Submenu deleted " .. _itemMenu:getLabel(), 3)
				end
				
				ATMEFct.outputCore(nil, "  deleteCommands End -> " .. _itemMenu:getLabel(), 3)
					
				return erase
			end
			
			ATMEFct.outputCore(nil, "C_Menu_deleteCommands Start on Id " .. _groupId .. " -> " .. _itemMenu:getLabel(), 3)
			deleteCommands(_itemMenu, _groupId)
			ATMEFct.outputCore(nil, "C_Menu_deleteCommands End -> " .. _itemMenu:getLabel(), 3)
		end
		
		function newF10SubMenu.append(self, _groupId, _label, _handler, _args)
			local fctDefinition = ATMEFct.getErrorPrefix(newF10SubMenu) .. "append(groupId, label, handler, [args])"
			if type(_groupId) == "number" and type(_label) == "string" and type(_handler) == "function" then
				if self:getNbEntries() < 10 then
					if _groupId >= 1 then
						ATMEFct.outputCore(nil, "Start append -> " .. _label, 3)
						
						local id = localMembers.handlers:add(_handler)
						
						local args = {
							idHandler = id,
							indexMenu = index,
							labelMenu = _label,
							menu = self,
							moduleName = ATMEDatas.currentModuleName,
							args = _args,
						}
						
						local itemRadio = {
								groupId = math.floor(_groupId),
								label = _label,
								idHandler = id,
								itemMenuCommandId = missionCommands.addCommandForGroup(localMembers.localMembersGroup.objectDCSId, _label, localMembers.path, ATMEFct.C_Menu_handler , args)
						}

						table.insert(localMembers.itemCommandsList,itemRadio)
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "groupId must be >= 1", false)
					end
				else
					ATMEFct.setError("MENU_MAX_ENTRIES")
					ATMEFct.warningOutputCore(nil, fctDefinition, "Max of entries per menu is 10. Can't add new item", false)
				end
			else
				if type(_groupId) ~= "number" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "groupId type must be number", true)
				elseif type(_label) ~= "string" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "label type must be string", false)
				elseif type(_handler) ~= "function" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "handler type must be function", false)
				end
			end
		end
		
		function newF10SubMenu.removeAll(self, _deleteSubMenu)
			local fctDefinition = ATMEFct.getErrorPrefix(newF10SubMenu) .. "removeAll([deleteSubMenu])"
			
			if  _deleteSubMenu == nil or type(_deleteSubMenu) == "boolean" then
				local flag = false
				if _deleteSubMenu ~= nil then
					flag = _deleteSubMenu
				end

				C_Menu_deleteAllCommands(self, flag)
				if flag == true and parent ~= nil then
					local localParentMembers = ATMEDatas.localMembers[parent:getIndex()]
					
					localParentMembers.itemSubMenusList[label] = nil
					localParentMembers.nbSubmenus = localParentMembers.nbSubmenus - 1
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "deleteSubMenu type must be nil or boolean", true)
			end
		end
		
		function newF10SubMenu.remove(self, _groupId, _deleteSubMenu)
			local fctDefinition = ATMEFct.getErrorPrefix(newF10SubMenu) .. "remove(groupId, [deleteSubMenu])"
			
			if type(_groupId) == "number" and (_deleteSubMenu == nil or type(_deleteSubMenu) == "boolean") then
				local flag = false
				if _deleteSubMenu ~= nil then
					flag = _deleteSubMenu
				end
				
				if _groupId >= 1 then
					C_Menu_deleteCommands(self, math.floor(_groupId), flag) 
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "groupId must be >= 1", false)
				end
			else
				if type(_groupId) ~= "number" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "groupId type must be number", true)
				elseif _deleteSubMenu ~= nil and type(_deleteSubMenu) ~= "boolean" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "deleteSubMenu type must be nil or boolean", false)
				end
			end
		end
		
		function newF10SubMenu.getUserPrivateDatas(self, _moduleName)
			local fctDefinition = ATMEFct.getErrorPrefix(newF10SubMenu) .. "getUserPrivateDatas([moduleName])"
			local moduleName
			
			if _moduleName ~= nil then
				if type(_moduleName) == "string" then
					if localMembers.userPrivateDatas[_moduleName] == nil then
						ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
						return
					else
						moduleName = _moduleName
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
					return
				end
			else
				moduleName = ATMEDatas.currentModuleName
			end
			
			return localMembers.userPrivateDatas[moduleName]
		end

		for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
			localMembers.userPrivateDatas[_localModuleMembers.name] = {}
		end

		if isRoot == false then
			local localMembersParent = ATMEDatas.localMembers[parent:getIndex()]
			localMembersParent.itemSubMenusList[label] = newF10SubMenu
			
			localMembersParent.nbSubmenus = localMembersParent.nbSubmenus + 1
			
			ATMEFct.outputCore(nil, "Création subMenu " .. label .. " à partir de " .. parent:getLabel(), 3)
		end
		
		
		
		return nbEntries, newF10SubMenu
	end
	
	return nbEntries
end


ATMEDatas.C_Menu_MT.__call = function(t, _label, _parent)
	local fctDefinition = "Constructor ATME.C_Menu(label, parent)"
	
	if ATMEDatas.startMissionOk < 2 then
		ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
		return
	end
	
	if type(_label) == "string" and ATMEFct.isObjectClass(_parent, ATME.C_Menu) == true then
		local nbEntries, menu = ATMEFct.C_Menu_create(_parent:getGroup(), _label, _parent, nil)
		
		if menu ~= nil then
			local parentPath = nil
			if _parent:isRoot() == false then
				local localParentMembers = ATMEDatas.localMembers[_parent:getIndex()]
				parentPath = localParentMembers.path
			end
			
			local localMembers = ATMEDatas.localMembers[menu:getIndex()]
			localMembers.path = missionCommands.addSubMenuForGroup(localMembers.localMembersGroup.objectDCSId, menu:getLabel(), parentPath)
			
			return menu
		else
			if nbEntries == 10 then
				ATMEFct.setError("MENU_MAX_ENTRIES")
				ATMEFct.warningOutputCore(nil, fctDefinition, "Max of entries per menu is 10. Can't add new submenu", false)
			else
				ATMEFct.setError("MENU_DUB")
				ATMEFct.warningOutputCore(nil, fctDefinition, "Label '" .. _label .. "' already exist in parent or root menu", false)
			end
		end
	else
		if type(_label) ~= "string" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "label type must be string", false)
		elseif ATMEFct.isObjectClass(_parent, ATME.C_Menu) == false then
			ATMEFct.errorOutputCore(nil, fctDefinition, "parent type must be ATME.C_Menu", false)
		end
	end
end

setmetatable(ATME.C_Menu,ATMEDatas.C_Menu_MT)

ATMEDatas.C_Menu_MT.mt.__metatable = "Private"
ATMEDatas.C_Menu_MT.mt.__index = ATME.C_Menu
ATMEDatas.C_Menu_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Menu_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Menu : Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Alarm : gère les alarmes définies par l'utilisateur.
-- Pas de contructeur accessible aux modules externes.
-- En effet, la création d'une alarme passe par deux méthodes de la classe ATME.C_Module.
-- Il existe deux types d'alarme :
--		- Les alarmes absolues se déclenchant au moment programmé.
--		- Les alarmes relative à un flag et donc associée à l'activation de ce flag pour se déclencher.
-- La condition de temps est donné par un string au format particulier :
--		- "X..X" pour un temps en secondes depuis le début de mission (ex : "10" ou "100" pour 10s ou 100s)
--		- "+X..X" pour un temps en seconde depuis la création de l'alarme (ex : "+10" pour 10s après la création)
--		- "HH:MM" ou "HH:MM:SS" pour remplacer une expression en secondes par une expression en m/s ou h/m/s
----------------------------------------------------------------------------------------------------------

ATMEFct.C_Alarm_createNamed = function(_namedTable, _name, _flagID, _rules)
	if _namedTable[_name] ~= nil then
		ATMEFct.setError("ALREADY_CREATED")
		return true, "named alarm already exists"
	else
		local d,f = string.find(_name, "|", 1)
		if d == nil then
			return ATMEFct.C_Alarm_create(_namedTable, _name, _flagID, _rules)
		else
			ATMEFct.setError("BAD_FORMAT")
			return true, "name can't contain '|' character"
		end
	end
end

ATMEDatas.C_Alarm_MT = {
	className = "C_Alarm",
	mt = {},
	
	keys = {
		["getFlag"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["getIndex"] = true,
		["getUserPrivateDatas"] = true, 
		["reschedule"] = true,
		["addTriggeringsLimit"] = true,
		["removeTriggeringsLimit"] = true,
		["addCallback"] = true,
		["removeCallback"] = true,
		["remove"] = true,
	},
	
	add = function(self)
		local item = {
			triggered = false,
			restarted = false,
			eventCallback = {},
			maxTriggerings = -1,
			maxTriggeringsCallback = {},
			userPrivateDatas = {},
			destroyed = false,
			moduleName = ATMEDatas.currentModuleName,
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].eventCallback = nil
		ATMEDatas.localMembers[_id].maxTriggeringsCallback = nil
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
	end,
}

----------------------------------------------------------------------------------------------------------
-- Fonctions timers appelées pour les alarmes
----------------------------------------------------------------------------------------------------------

ATMETimers.C_Alarm_timer = function(_alarm)
	if ATMEDatas.exitATME == false then
		local localMembers = ATMEDatas.localMembers[_alarm:getIndex()]
		local localMembersModule = ATMEDatas.modulesLocalMembers[localMembers.moduleName]
		
		localMembers.triggered = true
		if localMembers.flag ~= nil then
			if trigger.misc.getUserFlag(localMembers.flag) == 0 then
				ATMEFct.C_Alarm_delete(localMembersModule.alarms, _alarm)
				ATMEFct.outputCore(nil,"Alarme annulée -> flag non défini" , 2)
				return
			end
		end
		
		ATMEFct.outputCore(nil,"Alarme déclenchée" , 2)
		
		for _moduleName, _callback in pairs(localMembers.eventCallback) do
			ATMEDatas.currentModuleName = _moduleName

			ATMEFct.outputCore(nil,"Appel handler d'alarme pour le module " .. _moduleName, 2)
			xpcall(function() _callback(_alarm) end, function() ATMEFct.errUserHandler('alarmEventCallbackHandler') end)
		end
			
		if localMembers.maxTriggerings > 0 then
			localMembers.maxTriggerings = localMembers.maxTriggerings - 1
			
			if localMembers.maxTriggerings == 0 then
				for _moduleName, _callback in pairs(localMembers.maxTriggeringsCallback) do
					ATMEDatas.currentModuleName = _moduleName
						
					ATMEFct.outputCore(nil,"Appel handler maxTriggerings d'alarme pour le module " .. _moduleName, 2)
					xpcall(function() _callback(_alarm) end, function() ATMEFct.errUserHandler('maxTriggeringsCallbackHandler') end)
				end
				
				if localMembers.maxTriggerings ~= - 1 then  -- removeTriggeringLimit appelé dans les callbacks
					localMembers.restarted = false
				end
			end
		end
		
		if localMembers.restarted == false then
			ATMEFct.C_Alarm_delete(localMembersModule.alarms, _alarm)
		else
			localMembers.restarted = false
		end
	end
end	
				
ATMETimers.C_Alarm_timerSchedule = function(_args)
	if ATMEDatas.exitATME == false then
		ATMEDatas.currentModuleName = _args[3]
		xpcall(function() _args[1](unpack(_args[2])) end, function() ATMEFct.errUserHandler('Scheduled function') end)
	end
end

ATME.C_Alarm = {
	getClassName = function(self)
		return ATMEDatas.C_Alarm_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Alarm_MT.parentClass
	end,
	
	schedule = function(_rules, _fct, ...)
		local fctDefinition = "Function ATME.C_Alarm.schedule(rules, fct, fctArgs)"
		local args = {...}
		
		if (type(_rules) == "number" or type(_rules) == "string") and type(_fct) == "function" then
			local err, timeStart, isRelative = ATMEFct.getTimeValue(_rules)
			if err == false then
				local currentTime = timer.getTime()
				
				if isRelative == true then
					timeStart = timeStart + currentTime
				end
				
				if timeStart > currentTime then
					timer.scheduleFunction(ATMETimers.C_Alarm_timerSchedule, {_fct, args, ATMEDatas.currentModuleName}, timeStart)
					return true
				else
					ATMEFct.outputCore(nil, "Schedule refusé -> startTime < current time", 2)
					return false
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, timeStart, false)
			end
		else
			if type(args[1]) ~= "number" and type(args[1]) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "flagID type must be number or string", false)
			elseif type(_fct) ~= "function" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", false)
			end
		end
	end,
	
	createNamed = function(_name, ...)
		local fctDefinition = "Function ATME.C_Alarm.createNamed(name, rules)\nATME.C_Alarm.createNamed(name, flagID, rules)"
		local args = {...}
		
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			local d,f = string.find(_name, "|", 1)
			if d == nil then
				if #args == 1 then
					if type(args[1]) == "number" or type(args[1]) == "string" then
						local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
						local err, newAlarm = ATMEFct.C_Alarm_createNamed(localMembersModule.alarms, _name, nil, args[1])
							
						if err == true then
							ATMEFct.warningOutputCore(nil, fctDefinition, newAlarm, false)
							return
						end
						
						return newAlarm
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "rules type must be number or string", false)
					end
				elseif #args >= 2 then
					if (type(args[1]) == "number" or type(args[1]) == "string") and (type(args[2]) == "number" or type(args[2]) == "string") then
						local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
						local flag
						
						if type(args[1]) == "number" then
							flag = math.floor(args[1])
							
							if flag > 0 then
								flag = tostring(flag)
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "flagID must be > 0", false)
								return
							end
						else
							flag = args[1]
						end
						
						local err, newAlarm = ATMEFct.C_Alarm_createNamed(localMembersModule.alarms, _name, flag, args[2])
						
						if err == true then
							ATMEFct.warningOutputCore(nil, fctDefinition, newAlarm, false)
							return
						end
						
						return newAlarm
					else
						if type(args[1]) ~= "number" and type(args[1]) ~= "string" then
							ATMEFct.errorOutputCore(nil, fctDefinition, "flagID type must be number or string", false)
						elseif type(args[2]) ~= "number" and type(args[2]) ~= "string" then
							ATMEFct.errorOutputCore(nil, fctDefinition, "rules type must be number or string", false)
						end
					end
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "name can't contain '|' character", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", true)
		end
	end,
	
	getByName = function(_name)
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Alarm.getByName(name)", "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].alarms[_name]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Alarm.getByName(name)", "name type must be string", false)
		end
	end,
}

ATMEFct.C_Alarm_create = function(_alarms, _name, _flagID, _rules)
	local name
	local tagsName
	local index
	local localMembers
	local err, timeStart, isRelative = ATMEFct.getTimeValue(_rules)
	
	if err == false then
		local currentTime = timer.getTime()

		if isRelative == true then
			timeStart = timeStart + currentTime
		end
		
		if timeStart > currentTime then
			local newAlarm = {}
			
			if _name ~= nil then
				name = _name
				_alarms[name] = newAlarm
				tagsName = ATMEFct.decompString(_name)
			end
		
			setmetatable(newAlarm, ATMEDatas.C_Alarm_MT.mt)
	
			index = ATMEDatas.C_Alarm_MT:add()
			localMembers = ATMEDatas.localMembers[index]
			localMembers.flag = _flagID

			function newAlarm.getIndex(self)
				return index
			end
			
			function newAlarm.getFlag(self)
				if localMembers.destroyed == false then
					if localMembers.flag ~= nil then
						return trigger.misc.getUserFlag(localMembers.flag)
					end
				else
					ATMEFct.setError("ALREADY_DISABLED")
					ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newAlarm) .. "getFlag()", "already disabled", false)
				end
			end
			
			function newAlarm.getName(self)
				return name
			end
			
			function newAlarm.tagGetByID(self, _position)
				local fctDefinition = ATMEFct.getErrorPrefix(newAlarm) .. "tagGetByID(position)"
				
				if localMembers.destroyed == false then
					if type(_position) == "number" then
						local pos = math.floor(_position)
						if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
							return tagsName[pos]
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
					end
				else
					ATMEFct.setError("ALREADY_DISABLED")
					ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
				end
			end
			
			function newAlarm.tagGetCount(self)
				if localMembers.destroyed == false then
					if tagsName ~= nil then
						return #tagsName
					else
						return 0
					end
				else
					ATMEFct.setError("ALREADY_DISABLED")
					ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newAlarm) .. "tagGetCount()", "already disabled", false)
				end
			end
			
			function newAlarm.tagCompare(self, _tagsToCompare)
				local fctDefinition = ATMEFct.getErrorPrefix(newAlarm) .. "tagCompare(tagsToCompare)"
				
				if localMembers.destroyed == false then
					if type(_tagsToCompare) == "string" then
						if tagsName ~= nil then
							local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
							
							if ok ~= nil then
								return ok
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
							end
						else
							return false
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
					end
				else
					ATMEFct.setError("ALREADY_DISABLED")
					ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
				end
			end
			
			function newAlarm.getUserPrivateDatas(self, _moduleName)
				local fctDefinition = ATMEFct.getErrorPrefix(newAlarm) .. "getUserPrivateDatas([moduleName])"
				local moduleName
				
				if _moduleName ~= nil then
					if type(_moduleName) == "string" then
						if localMembers.userPrivateDatas[_moduleName] == nil then
							ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
							return
						else
							moduleName = _moduleName
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
						return
					end
				else
					moduleName = ATMEDatas.currentModuleName
				end
				
				return localMembers.userPrivateDatas[moduleName]
			end
			
			function newAlarm.addCallback(self, _fct)
				local fctDefinition = ATMEFct.getErrorPrefix(newAlarm) .. "addCallback(fct)"
				
				if localMembers.destroyed == false then
					if type(_fct) == "function" then
						localMembers.eventCallback[ATMEDatas.currentModuleName] = _fct
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
					end
				else
					ATMEFct.setError("ALREADY_DISABLED")
					ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
				end
			end
			
			function newAlarm.removeCallback(self)
				if localMembers.destroyed == false then
					localMembers.eventCallback[ATMEDatas.currentModuleName] = nil
				else
					ATMEFct.setError("ALREADY_DISABLED")
					ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newAlarm) .. "removeCallback()", "already disabled", false)
				end
			end
			
			function newAlarm.addTriggeringsLimit(self, _value, _callback)
				local fctDefinition = ATMEFct.getErrorPrefix(newAlarm) .. "addTriggeringsLimit(value, [callback])"
				
				if localMembers.destroyed == false then
					if type(_value) == "number" and (_callback == nil or type(_callback) == "function") then
						local value = math.floor(_value)
						
						if value > 0 then
							localMembers.maxTriggerings = value
							localMembers.maxTriggeringsCallback[ATMEDatas.currentModuleName] = _callback
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "value must be > 0", true)
						end
					else
						if type(_value) ~= "number" then
							ATMEFct.errorOutputCore(nil, fctDefinition, "value type must be number", true)
						elseif _callback ~= nil and type(_callback) ~= "function" then
							ATMEFct.errorOutputCore(nil, fctDefinition, "callback type must be nil or function", false)
						end
					end
				else
					ATMEFct.setError("ALREADY_DISABLED")
					ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
				end
			end
			
			function newAlarm.removeTriggeringsLimit(self)
				if localMembers.destroyed == false then
					localMembers.maxTriggerings = -1
					localMembers.maxTriggeringsCallback[ATMEDatas.currentModuleName] = nil
				else
					ATMEFct.setError("ALREADY_DISABLED")
					ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newAlarm) .. "removeTriggeringsLimit()", "already disabled", false)
				end
			end
			
			function newAlarm.reschedule(self, _rules)
				local fctDefinition = ATMEFct.getErrorPrefix(newAlarm) .. "reschedule(rules)"
				
				if localMembers.destroyed == false then
					if type(_rules) == "number" or type(_rules) == "string" then
						local err, timeStart, isRelative = ATMEFct.getTimeValue(_rules)
						
						if err == false then
							local currentTime = timer.getTime()
				
							if isRelative == true then
								timeStart = timeStart + currentTime
							end
							
							if timeStart > currentTime then
								localMembers.restarted = true
								
								if localMembers.triggered == false then
									timer.setFunctionTime(localMembers.idTimer, timeStart)
								else
									localMembers.idTimer = timer.scheduleFunction(ATMETimers.C_Alarm_timer, self, timeStart)
								end
								
								localMembers.triggered = false
								

								ATMEFct.outputCore(nil, "Reschedule alarme -> alarme reprogrammée (timeStart = ".. timeStart .. ")" , 2)
							else
								ATMEFct.outputCore(nil, "Reschedule alarme -> startTime < current time : alarme non prise en compte", 2)
								return false
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, timeStart, false)
						end
							
						return localMembers.restarted
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "rules type must be number or string", true)
					end
				else
					ATMEFct.setError("ALREADY_DISABLED")
					ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
				end
			end
			
			function newAlarm.remove(self)
				if localMembers.destroyed == false then
					local localMembersModule = ATMEDatas.modulesLocalMembers[localMembers.moduleName]
					ATMEFct.C_Alarm_delete(localMembersModule.alarms, self)
				else
					ATMEFct.setError("ALREADY_DISABLED")
					ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newAlarm) .. "remove()", "already disabled", false)
				end
			end

			for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
				localMembers.userPrivateDatas[_localModuleMembers.name] = {}
			end

			localMembers.idTimer = timer.scheduleFunction(ATMETimers.C_Alarm_timer, newAlarm, timeStart)
			ATMEFct.outputCore(nil, "Alarme créée" , 2)

			return false, newAlarm
		else
			ATMEFct.outputCore(nil, "Alarme startTime < current time : Alarme non crée", 2)
			return false
		end
	else
		return true, timeStart
	end
end

ATMEFct.C_Alarm_delete = function(_alarms, _alarm)
	local name = _alarm:getName()
	
	if name ~= nil then
		_alarms[name] = nil
	end
	
	local localMembers = ATMEDatas.localMembers[_alarm:getIndex()]
	localMembers.destroyed = true
	
	if localMembers.triggered == false then
		timer.removeFunction(localMembers.idTimer)
	end
	
	localMembers.triggered = false
	ATMEDatas.C_Alarm_MT:remove(_alarm:getIndex())
	
	ATMEFct.outputCore(nil, "Alarme détruite", 2)
end


setmetatable(ATME.C_Alarm,ATMEDatas.C_Alarm_MT)

ATMEDatas.C_Alarm_MT.__call = function(t, ...)
	local fctDefinition = "Constructor ATME.C_Alarm(rules)\nATME.C_Alarm(flagID, rules)"
	local args = {...}
	
	if ATMEDatas.startMissionOk < 2 then
		ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
		return
	end
	
	if #args == 1 then
		if type(args[1]) == "number" or type(args[1]) == "string" then
			local err, newAlarm = ATMEFct.C_Alarm_create(nil, nil, nil, args[1])
			if err == true then
				ATMEFct.errorOutputCore(nil, fctDefinition, newAlarm, false)
				return
			end
			return newAlarm
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "rules type must be number or string", false)
		end
	elseif #args >= 2 then
		if (type(args[1]) == "number" or type(args[1]) == "string") and (type(args[2]) == "number" or type(args[2]) == "string") then
			local flag
			
			if type(args[1]) == "number" then
				flag = math.floor(args[1])
				
				if flag > 0 then
					flag = tostring(flag)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "flagID must be > 0", false)
					return
				end
			else
				flag = args[1]
			end
			
			local err, newAlarm = ATMEFct.C_Alarm_create(nil, nil, flag, args[2])
		
			if err == true then
				ATMEFct.errorOutputCore(nil, fctDefinition, newAlarm, false)
				return
			end
			
			return newAlarm
		else
			if type(args[1]) ~= "number" and type(args[1]) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "flagID type must be number or string", false)
			elseif type(args[2]) ~= "number" and type(args[2]) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "rules type must be number or string", false)
			end
		end
	end
end

ATMEDatas.C_Alarm_MT.mt.__metatable = "Private"
ATMEDatas.C_Alarm_MT.mt.__index = ATME.C_Alarm
ATMEDatas.C_Alarm_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Alarm_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Alarm : Bad key " .. k, "table locked", false)
	end
end


----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Event : La classe ATME.C_Event permet la création d'instance events Core.
-- Pas de contructeur accessible aux modules externes. instances entièrement gérées par ATME.
---------------------------------------------------------------------------------------------------------- 

ATMEDatas.C_Event_MT = {
	className = "C_Event",
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["getType"] = true,
		["getDatas"] = true,
	},
	
	add = function(self, _item)
		local item = {
			type = _item.type,
			datas = _item.datas,
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
	end,
}

ATME.C_Event = {
	getClassName = function(self)
		return ATMEDatas.C_Event_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Event_MT.parentClass
	end,
}

setmetatable(ATME.C_Event, ATMEDatas.C_Event_MT)

ATMEDatas.C_Event_MT.mt.__metatable = "Private"
ATMEDatas.C_Event_MT.mt.__index = ATME.C_Event
ATMEDatas.C_Event_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Event_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Event : Bad key " .. k, "table locked", false)
	end
end

ATMEFct.C_Event_create = function(_item)
	local newEvent = {}
	
	setmetatable(newEvent, ATMEDatas.C_Event_MT.mt)
	local index = ATMEDatas.C_Event_MT:add(_item)
	local localMembers = ATMEDatas.localMembers[index]
	
	function newEvent.getIndex()
		return index
	end
	
	function newEvent.getType(self)
		return localMembers.type
	end
	
	function newEvent.getDatas(self)
		return ATMEConst.coreEventTypes[localMembers.type](localMembers)
	end
	
	return newEvent
end

-- Définition du deux events mgr : la bascule se fait dans la fonction main timer
-- Ceci permet si des événements sont créés dans cette fonction, de les traiter
-- au coup suivant. Dans main timer, activeEvents passe de 1 à 0 et vice versa.

local events = {
	[0] = ATMEFct.C_List_create(),
	[1] = ATMEFct.C_List_create(),
}

local activeEvents = 0

ATMEFct.C_Event_add = function(_item)
	local newEvent = ATMEFct.C_Event_create(_item)
	events[activeEvents]:add(newEvent)
end

ATMEFct.C_Event_removeAll = function(_id)
	events[_id]:removeAll()
end

ATMEFct.C_Event_fire = function(_event)
	local localEventMembers = ATMEDatas.localMembers[_event:getIndex()]
			
	local name = localEventMembers.datas.unitName
	if ATMEDatas.units[name] ~= nil then
		local localMembersUnit = ATMEDatas.localMembers[ATMEDatas.units[name]:getIndex()]
		
		if localEventMembers.datas.action == "UNLOADED" then
			local localMembersGroup = ATMEDatas.localMembers[localEventMembers.datas.linkedGroup:getIndex()]
			localMembersGroup.disembarking = false
		end

		local moduleName = ATMEDatas.currentModuleName
		for _moduleName, _callbacks in pairs(localMembersUnit.eventCallbacks) do
			ATMEDatas.currentModuleName = _moduleName

			if _callbacks[localEventMembers.type] ~= nil then
				ATMEFct.outputCore(nil,"Appel event (type " .. localEventMembers.type .. ") callback de l'unité " .. name .. " pour le module " .. _moduleName, 2)
				xpcall(function() _callbacks[localEventMembers.type](ATMEDatas.units[name], _event) end, function() ATMEFct.errUserHandler('eventCallbackHandler') end)
			end
		end
		
		ATMEDatas.currentModuleName = moduleName
	end
end

ATMEFct.C_Event_target_class = {
	["C_Unit"] = function(_targetName)
		return ATMEDatas.units[_targetName]
	end,
	["C_Weapon"] = function(_targetName)
		return ATMEDatas.weapons[_targetName]
	end,
	["C_Static"] = function(_targetName)
		return ATMEDatas.statics[_targetName]
	end,
	["C_Cargo"] = function(_targetName)
		return ATMEDatas.cargos[_targetName]
	end,
	["C_SceneryObject"] = function(_targetName)
		return ATMEDatas.sceneryObjects[_targetName]
	end,
	["C_Airbase"] = function(_targetName)
		-- cas "SHIP" traité par C_Unit
		
		if ATMEDatas.airbases["AIRBASE"][_targetName] ~= nil then
			return ATMEDatas.airbases["AIRBASE"][_targetName]
		else
			return ATMEDatas.airbases["HELIPAD"][_targetName]
		end
	end,
}

ATMEFct.C_Event_update = function(_event, _remove)
	local localEventMembers = ATMEDatas.localMembers[_event:getIndex()]

	if localEventMembers.datas.unitName ~= nil then
		if ATMEDatas.units[localEventMembers.datas.unitName] ~= nil then
			if localEventMembers.datas.linkedGroupName ~= nil then  -- Type string
				if ATMEDatas.groups[localEventMembers.datas.linkedGroupName] ~= nil then
					localEventMembers.datas.linkedGroup = ATMEDatas.groups[localEventMembers.datas.linkedGroupName]
				elseif _remove == true then
					events[activeEvents]:remove(_id)
				end
			end

			if localEventMembers.datas.targetName ~= nil then -- Type string ou number suivant les cas
				local target = ATMEFct.C_Event_target_class[localEventMembers.datas.targetClassName](localEventMembers.datas.targetName)
				if target ~= nil then
					localEventMembers.datas.target = target
				elseif _remove == true then
					events[activeEvents]:remove(_id)
				end
			end
			
			if localEventMembers.datas.weaponName ~= nil then -- Type number
				if ATMEDatas.weapons[localEventMembers.datas.weaponName] ~= nil then
					localEventMembers.datas.weapon = ATMEDatas.weapons[localEventMembers.datas.weaponName]
				end
			end
		end
	elseif _remove == true then
		events[activeEvents]:remove(_id)
	end
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Module : gère les modules externes. Contructeur accessible aux modules externes.
-- Tout module externe doit créer une instance de cette classe afin de pouvoir activer les handlers définis
-- dans le module externe. Un handler à nil, donc non défini, ne sera pas activé par ATME_Core.
-- Cette classe offre des fonction liées au event ATME et à leurs données, permet la création des triggers
-- utilisateurs dont les identifiants sont spécifiques à un module et ajoute des fonction de debug.
--
-- Il est à noter que tous les appels xpcall au travers de localMembers (Instance de ATME.C_List) se font
-- toujours dans l'ordre de création des instance de cette classe. Ainsi, un module ayant besoin d'un autre
-- module devra être défini dans les triggers après le module qu'il utilise pour que le bon ordre d'appel
-- soit respecté.
----------------------------------------------------------------------------------------------------------

ATMEDatas.C_Module_MT = {
	className = "C_Module",
	mt = {},
	
	keys = {
		["error"] = true,
		["output"] = true,
		["outputVar"] = true,
		["getName"] = true,
		["getIndex"] = true,
		["isDebugOn"] = true,
	},
	
	add = function(self, _name, _newHandlers)
		local item = {
			alarms = {},
			zones = {},
			tasks = {},
			lines2D = {},
			vectors3D = {},
			vectors2D = {},
			segments2D = {},
			zones = {},
			groupCloningContexts = {},
			staticCloningContexts = {},
			cargoCloningContexts = {},
			findContexts = {},
			radioChannels = {},
			dataSharings = {},
			sounds = {},
			routes = {},
			name = _name,
			id = #ATMEDatas.modulesNames + 1,
			handlers = _newHandlers,
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].alarms = nil
		ATMEDatas.localMembers[_id].namedAlarms = nil
		ATMEDatas.localMembers[_id].handlers = nil
		ATMEDatas.localMembers[_id].zones = nil
		ATMEDatas.localMembers[_id].tasks = nil
		ATMEDatas.localMembers[_id].lines2D = nil
		ATMEDatas.localMembers[_id].vectors3D = nil
		ATMEDatas.localMembers[_id].vectors2D = nil
		ATMEDatas.localMembers[_id].segments2D = nil
		ATMEDatas.localMembers[_id].groupCloningContexts = nil
		ATMEDatas.localMembers[_id].zones = nil
		ATMEDatas.localMembers[_id].staticCloningContexts = nil
		ATMEDatas.localMembers[_id].cargoCloningContexts = nil
		ATMEDatas.localMembers[_id].findContexts = nil
		ATMEDatas.localMembers[_id].radioChannels = nil
		ATMEDatas.localMembers[_id].dataSharings = nil
		ATMEDatas.localMembers[_id].sounds = nil
		ATMEDatas.localMembers[_id].routes = nil
	end,
}

ATME.C_Module = {
	getClassName = function(self)
		return ATMEDatas.C_Module_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Module_MT.parentClass
	end,
	
	isLoaded = function(_name)
		local fctDefinition = "ATME.C_Module.isLoaded(name)"
		
		if type(_name) == "string" then
			return ATMEDatas.modulesLocalMembers[_name] ~= nil
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", true)
		end
	end,
	
	getNames = function()
		local names = {}
		for _i, _name in ipairs(ATMEDatas.modulesNames) do
			names[_i] = _name
		end
		
		return names
	end
}

ATMEFct.C_Module_create = function(_name, _handlers, _debugSwitch)		   
    local newHandlers = {}
   
    for _id, _ in pairs(_handlers) do
		newHandlers[_id] = _handlers[_id]
	end
	
	local newModule = {}
	setmetatable(newModule, ATMEDatas.C_Module_MT.mt)
	
	local debugSwitch = _debugSwitch
	local index = ATMEDatas.C_Module_MT:add(_name, newHandlers)
	local localMembers = ATMEDatas.localMembers[index]
	
	local err = debug.traceback()

	local referenceFile = ATMEFct.searchMainChunk(err)
	
	ATMEDatas.refModules[referenceFile] = _name
	ATMEDatas.modules[_name] = newModule
	ATMEDatas.modulesLocalMembers[_name] = localMembers
	table.insert(ATMEDatas.modulesNames, _name)
	
	if ATME.modules[_name] == nil then
		ATME.modules[_name] = {}
	end
	
	function newModule.getName(self)	
		return localMembers.name
	end
	
	function newModule.getIndex(self)
		return index
	end
	
	function newModule.isDebugOn(self)
		return debugSwitch
	end
	
	function newModule.output(self, _text, _level)
		local fctDefinition = ATMEFct.getErrorPrefix(newModule) .. "output(text, [level])"
		
		if type(_text) == "string" and (_level == nil or type(_level) == "number") then
			local level = 1
			if type(_level) == "number" then
				level = math.floor(_level)
			end
			
			if level >= 0 then
				ATMEFct.outputCore(self, _text, level)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "level must be >= 0", false)
			end
		else
			if type(_text) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "text type must be string", true)
			elseif _level ~= nil and type(_level) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "level type must be nil or number", false)
			end
		end
	end
	
	function newModule.outputVar(self, _title, _var, _level)
		local fctDefinition = ATMEFct.getErrorPrefix(newModule) .. "outputVar(text, var, [level])"
		
		if type(_title) == "string" and (_level == nil or type(_level) == "number") then
			local level = 1
			if type(_level) == "number" then
				level = math.floor(_level)
			end
			
			if level >= 0 then
				ATMEFct.outputVarCore(self, _title, _var, level)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "level must be >= 0", false)
			end
		else
			if type(_title) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "_title type must be string", true)
			elseif _vel ~= nil and type(_level) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "level type must be nil or number", false)
			end
		end
	end
	
	function newModule.error(self, _text)
		local fctDefinition = ATMEFct.getErrorPrefix(newModule) .. "error(text)"
		
		if type(_text) == "string" then
			ATMEFct.errorOutputCore(self, "test", _text, false)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "text type must be string", true)
		end
	end
	
	ATMEFct.outputCore(nil, "Module " .. localMembers.name .." -> loaded", 2)
		
	return newModule			
end

ATMEDatas.C_Module_MT.__call = function(t, _name, _handlers, _debugSwitch)
	local fctDefinition = "Constructor ATME.C_Module(name, handlers, debugSwitch)"
	
	if type(_name) == "string" and type(_handlers) == "table" and type(_debugSwitch) == "boolean" then
		if ATMEDatas.isRunning == false then
			if ATMEDatas.modulesLocalMembers[_name] == nil then
				for _id, _ in pairs(_handlers) do
					if ATMEConst.moduleHandlers[_id] ~= nil then
						if type(_handlers[_id]) ~= "function" then
							ATMEFct.errorOutputCore(nil, fctDefinition, "Module " .. _name .. " - Handler '" .. _id .. " must be nil or function", false)
							return
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Module " .. _name .. " - Handler bad id : '" .. _id, false)
						return
					end
				end
					
				return ATMEFct.C_Module_create(_name, _handlers, _debugSwitch)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Module " .. _name .. " already loaded", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "Module " .. _name .. " is not created - ATME is running", false)
		end
	else
		if type(_name) ~= "string" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
		elseif type(_handlers) ~= "table" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "handlers type must be table of functions", false)
		elseif type(_debugSwitch) ~= "boolean" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "debugSwitch type must be boolean", false)
		end
	end
end

setmetatable(ATME.C_Module,ATMEDatas.C_Module_MT)

ATMEDatas.C_Module_MT.mt.__metatable = "Private"
ATMEDatas.C_Module_MT.mt.__index = ATME.C_Module
ATMEDatas.C_Module_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Module_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Module : Bad key " .. k, "table locked", false)
	end
end


----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Smoke : gère les fumées. Contructeur accessible aux modules externes.
-- Une fumée peut être déclenché simplement. DCS l'arrêtera au bout de 300s (5 minutes). Elle peut être également
-- redéclenchée au bout de cette durée jusqu'à un arrêt volontaire. 
-- La couleur est définie par une string dans ATME.
-- On notera qu'il est aussi possible de déclencher une fumée à partir d'un unité joueur, AI, ou d'un objet statique.
----------------------------------------------------------------------------------------------------------

ATMETimers.C_Smoke_timer = function(_localMembers)
	if ATMEDatas.exitATME == false then
		if _localMembers.activated == true then
			if _localMembers.always == true then
				trigger.action.smoke(_localMembers.point, _localMembers.color)
				
				return timer.getTime() + ATMEConst.maxTimeSmoke
			else
				_localMembers.activated = false
			end
		end
	end
end

ATMEDatas.C_Smoke_MT = {
	className = "C_Smoke",
	mt = {},
	
	keys = {
		["activate"] = true,
		["stop"] = true,
		["getIndex"] = true,
		["getUserPrivateDatas"]= true,
	},
	
	add = function(self, _point, _color)
		local item = {
			point = _point,
			color = _color,
			always = false,
			activated = false,
			userPrivateDatas = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
	end,
}

ATME.C_Smoke = {
	getClassName = function(self)
		return ATMEDatas.C_Smoke_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Smoke_MT.parentClass
	end,
}

ATMEFct.C_Smoke_create = function(_color, _point)
	local newSmoke = {}
	local color

	setmetatable(newSmoke, ATMEDatas.C_Smoke_MT.mt)

	if ATMEConst.smokeColors[_color] == -1 then
		color = ATMEConst.smokeColors[ATMEConst.smokeColorsIndex[math.random(1,#ATMEConst.smokeColorsIndex)]]
	else
		color = ATMEConst.smokeColors[_color]
	end
	
	local index = ATMEDatas.C_Smoke_MT:add(_point, color)
	local localMembers = ATMEDatas.localMembers[index]
	
	function newSmoke.activate(self, _always)
		local fctDefinition = ATMEFct.getErrorPrefix(newSmoke) .. "activate([always])"
		
		if _always == nil or type(_always) == "boolean" then
			localMembers.always = _always
			if _always == nil then
				localMembers.always = false
			end
			
			if localMembers.activated == false then
				localMembers.activated = true
				trigger.action.smoke(localMembers.point, localMembers.color)
				timer.scheduleFunction(ATMETimers.C_Smoke_timer, localMembers, timer.getTime() + ATMEConst.maxTimeSmoke)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "always type must be nil or boolean", true)
		end
	end

	function newSmoke.stop(self)
		localMembers.activated = false
		localMembers.always = false
	end
	
	function newSmoke.getIndex(self)
		return index
	end
	
	function newSmoke.getUserPrivateDatas(self, _moduleName)
		local fctDefinition = ATMEFct.getErrorPrefix(newSmoke) .. "getUserPrivateDatas([moduleName])"
		local moduleName
		
		if _moduleName ~= nil then
			if type(_moduleName) == "string" then
				if localMembers.userPrivateDatas[_moduleName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
					return
				else
					moduleName = _moduleName
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
				return
			end
		else
			moduleName = ATMEDatas.currentModuleName
		end
		
		return localMembers.userPrivateDatas[moduleName]
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return newSmoke
end

setmetatable(ATME.C_Smoke,ATMEDatas.C_Smoke_MT)

ATMEDatas.C_Smoke_MT.__call = function(t, _color, _reference)
	local fctDefinition = "Constructor ATME.C_Smoke(color, reference)"
	
	if ATMEDatas.startMissionOk < 2 then
		ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
		return
	end
	
	if type(_color) == "string" and ATMEFct.isReference(_reference) == true then
		local color = string.upper(_color)
		if ATMEConst.smokeColors[color]~= nil then
			return ATMEFct.C_Smoke_create(color, ATMEFct.getPointFromReference(_reference))
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "color must be a smoke color or 'RANDOM'", false)
		end
	else
		if type(_color) ~= "string" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "color type must be string", false)
		elseif ATMEFct.isReference(_reference) == false then
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
		end
	end
end

ATMEDatas.C_Smoke_MT.mt.__metatable = "Private"
ATMEDatas.C_Smoke_MT.mt.__index = ATME.C_Smoke
ATMEDatas.C_Smoke_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Smoke_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Smoke : Bad key " .. k, "table locked", false)
	end
end


----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Flare : gère les flares de couleur. Contructeur accessible aux modules externes.
-- Un tir de flare peut être déclenché unitairement. Il peut être aussi redéclenché automatiquement et aléatoirement
-- sur une durée souhaitée ou jusqu'à un arrêt volontaire.
-- La couleur est définie par une string dans ATME. Il existe une couleur "RANDOM" pour un choix aléatoire.
-- On notera qu'il est aussi possible de déclencher un flare à partir d'un unité joueur, AI, ou d'un objet statique.
----------------------------------------------------------------------------------------------------------

ATMETimers.C_Flare_timer = function(_localMembers)
	if ATMEDatas.exitATME == false then
		if _localMembers.activated == true then
			local color
				
			if ATMEConst.flareColors[_localMembers.colorName] == -1 then
				color = ATMEConst.flareColors[ATMEConst.flareColorsIndex[math.random(1, #ATMEConst.flareColorsIndex)]]
			else
				color = ATMEConst.flareColors[_localMembers.colorName]
			end
			
			trigger.action.signalFlare(_localMembers.point, color, math.random() * 2 * math.pi)
			
			if _localMembers.always == true then
				return timer.getTime() + math.random(_localMembers.minTimer, _localMembers.maxTimer)
			else
				_localMembers.activated = false
			end
		end
	end
end

ATMEDatas.C_Flare_MT = {
	className = "C_Flare",
	mt = {},
	
	keys = {
		["activate"] = true,
		["stop"] = true,
		["getIndex"] = true,
		["getUserPrivateDatas"]= true,
	},
	
	add = function(self, _point, _colorName)
		local item = {
			point = _point,
			colorName = _colorName,
			activated = false,
			always = false,
			minTimer = 1,
			maxTimer = 4,
			userPrivateDatas = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
	end,
}

ATME.C_Flare = {
	getClassName = function(self)
		return ATMEDatas.C_Flare_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Flare_MT.parentClass
	end,
}

ATMEFct.C_Flare_create = function(_color, _point)	
	local newFlare = {}
	
	setmetatable(newFlare, ATMEDatas.C_Flare_MT.mt)
	local index = ATMEDatas.C_Flare_MT:add(_point, string.upper(_color))
	local localMembers = ATMEDatas.localMembers[index]
	
	function newFlare.activate(self, _always)
		local fctDefinition = ATMEFct.getErrorPrefix(newSmoke) .. "activate([always])"
		
		if _always == nil or type(_always) == "boolean" then
			localMembers.always = _always
			if _always == nil then
				localMembers.always = false
			end
			
			if localMembers.activated == false then
				localMembers.activated = true
				timer.scheduleFunction(ATMETimers.C_Flare_timer, localMembers, timer.getTime() + 1)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "always type must be nil or boolean", true)
		end
	end
	
	function newFlare.stop(self)
		localMembers.activated = false
		localMembers.always = false
	end
	
	function newFlare.getIndex(self)
		return index
	end
	
	function newFlare.getUserPrivateDatas(self, _moduleName)
		local fctDefinition = ATMEFct.getErrorPrefix(newFlare) .. "getUserPrivateDatas([moduleName])"
		local moduleName
		
		if _moduleName ~= nil then
			if type(_moduleName) == "string" then
				if localMembers.userPrivateDatas[_moduleName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
					return
				else
					moduleName = _moduleName
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
				return
			end
		else
			moduleName = ATMEDatas.currentModuleName
		end
		
		return localMembers.userPrivateDatas[moduleName]
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return newFlare
end

setmetatable(ATME.C_Flare,ATMEDatas.C_Flare_MT)

ATMEDatas.C_Flare_MT.__call = function(t, _color, _reference)
	local fctDefinition = "Constructor ATME.C_Flare(color, reference)"
	
	if ATMEDatas.startMissionOk < 2 then
		ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
		return
	end
	
	if type(_color) == "string" and ATMEFct.isReference(_reference) == true then
		if ATMEConst.flareColors[string.upper(_color)] ~= nil then
			return ATMEFct.C_Flare_create(_color, ATMEFct.getPointFromReference(_reference))
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "color must be a flare color or 'RANDOM'", false)
		end
	else
		if type(_color) ~= "string" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "color type must be string", false)
		elseif ATMEFct.isReference(_reference) == false then
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
		end
	end
end


ATMEDatas.C_Flare_MT.mt.__metatable = "Private"
ATMEDatas.C_Flare_MT.mt.__index = ATME.C_Flare
ATMEDatas.C_Flare_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Flare_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Flare : Bad key " .. k, "table locked", false)
	end
end


-----------------------------------------------------------------------------------------------------------
-- Définition des fonctions indirectes ATMEAIIndFct liées aux tâches et définies dans la table 
-- ATMEConst.taskAI. 
-- Ces fonctions retournent soit :
--     - une table, qui comporte un ou plusieurs éléments indexés. Si plusieurs éléments sont 
--       retournés, les paramètres associés à linkedParams, définis dans la table ATMEConst.taskAI 
--       seront initialisés.
--	   - une erreur ou non en cas de validation finale. Dans ce cas la fonction débute par C_TaskValidate
-----------------------------------------------------------------------------------------------------------

ATMEAIIndFct.C_TaskValidate_Altitude = function(_taskParams)
	if _taskParams["altitude"] ~= nil and _taskParams["alt_type"] == nil then
		_taskParams["alt_type"] = 0
	elseif _taskParams["altitude"] == nil and _taskParams["alt_type"] ~= nil then
		_taskParams["alt_type"] = nil
	end
	
	return false
end

ATMEAIIndFct.C_getNumberWithFlag = function(_arg)
	return {_arg, true}, 0
end

ATMEAIIndFct.C_getWaypointNumberWithFlag = function(_arg)
	return {_arg + 1, true}, 0
end

ATMEAIIndFct.C_getAltitudeType = function(_arg)
	local typeAlt = 0
	
	if _arg == "AGL" then
		typeAlt = 1
	end
	
	return {typeAlt}, 0
end

ATMEAIIndFct.C_Circle_getCenterRadius = function(_arg)
	local point = _arg:getCenter()
	local rMax = _arg:getRadius()
	
	return {{x = point.x, y = point.z}, rMax}, 0
end

ATMEAIIndFct.C_Group_getID = function(_arg)
	return {_arg:getDCSObject():getID()}, 0
end

ATMEAIIndFct.C_Group_getIDWithWP = function(_arg)
	return {_arg:getDCSObject():getID(), false}, 0
end

ATMEAIIndFct.C_Unit_getID = function(_arg)
	return {_arg:getDCSObject():getID()}, 0
end

ATMEAIIndFct.C_getWeaponFlag = function(_arg)
	local value = 0
	
	for _item in _arg:gmatch("([^%s]+)") do
		if ATMEConst.weaponFlags[string.upper(_item)] ~= nil then
			value = value + ATMEConst.weaponFlags[string.upper(_item)]
		else
			return {}, 0
		end
	end

	return {value}, 0
end

ATMEAIIndFct.C_getExpend = function(_arg)	
	return {ATMEConst.weaponExpend[string.upper(_arg)]}, 0
end

ATMEAIIndFct.C_getDesignation = function(_arg)	
	return {ATMEConst.designationAI[string.upper(_arg)]}, 0
end

ATMEAIIndFct.C_getTargetsType = function(_arg)
	local values = {}

	for _item in _arg:gmatch("([^%s]+)") do
		local attr = ATMEConst.targetsAttributes[string.upper(_item)]

		if attr ~= nil and ATMEConst.backTargetsAttributes[attr].target ~= nil then
			table.insert(values, attr)
		end
	end
	
	return values, 0
end

ATMEAIIndFct.C_getAirTargetsType = function(_arg)
	local values = {}

	for _item in _arg:gmatch("([^%s]+)") do
		local attr = ATMEConst.targetsAttributes[string.upper(_item)]

		if attr ~= nil and ATMEConst.backTargetsAttributes[attr].airTarget ~= nil then
			table.insert(values, attr)
		end
	end
	
	return values, 0
end

ATMEAIIndFct.C_getOptionTargetsType = function(_arg)
	local values = {}

	for _item in _arg:gmatch("([^%s]+)") do
		local attr = ATMEConst.targetsAttributes[string.upper(_item)]

		if attr ~= nil and ATMEConst.backTargetsAttributes[attr].optionTarget ~= nil then
			table.insert(values, attr)
		end
	end
	
	return values, 0
end

ATMEAIIndFct.C_getModulation = function(_arg)
	local value = string.upper(_arg)
	
	if value == "FM" then
		return {radio.modulation.FM}, 0
	elseif value == "AM" then
		return {radio.modulation.AM}, 0
	end
end

ATMEAIIndFct.C_getCallname = function(_arg)
	local value = string.gsub(_arg, "-", " ")
	
	local tokens = ATMEFct.analyzeString(string.upper(value))
	
	if #tokens == 2 and tokens[1].type == "TOKEN_JTAC_CALLSIGN_NAME" and tokens[2].type == "TOKEN_INTEGER" then
		return {ATMEConst.blueCallsigns.jtacs[tokens[1].value].id, tokens[2].value}, 1
	elseif #tokens == 2 and tokens[1].type == "TOKEN_AIRCRAFT_CALLSIGN_NAME" and tokens[2].type == "TOKEN_INTEGER" then
		return {ATMEConst.blueCallsigns.aircrafts[tokens[1].value].id, tokens[2].value}, 2
	end
end

ATMEAIIndFct.C_getPoint2D = function(_arg)
	if ATMEFct.isObjectClass(_arg, ATME.C_Unit) or ATMEFct.isObjectClass(_arg, ATME.C_Group) then
		local pos = _arg:getPosition()
		return {{x = pos.x, y = pos.z}}, 0
	elseif _arg.x ~= nil and _arg.y ~= nil then
		local point = ATMEFct.convertToPoint3D(_arg)
		return {{x = point.x, y = point.z}}, 0
	end
end

ATMEAIIndFct.C_getOrbitPattern = function(_arg)
	return {ATMEConst.orbitPattern[string.upper(_arg)]}, 0
end	

ATMEAIIndFct.C_getAirbase = function(_arg)
	if _arg:getDCSAirdromeID() ~= nil then
		return {_arg:getDCSAirdromeID()}, 0
	end
end

ATMEAIIndFct.C_Vector_getPos = function(_arg)
	return {{x = _arg:getX(), y = _arg:getY(), z = _arg:getZ()}}, 0
end

ATMEFct.testDatas = function(_data, _param)
	if _data ~= nil then
		if _param.enum ~= nil then
			local data = _data
			if type(data) == "string" then
				data = string.upper(_data)
			end
			if _param.enum[data] ~= nil then
				return _param.enum[data]
			end
		elseif type(_data) == "number" then
			if _param.max ~= nil and type(_data) == "number" then
				if _data > _param.max then
					_data = _param.max
				end
			end
			
			if _param.min ~= nil and type(_data) == "number" then
				if _data < _param.min then
					_data = _param.min
				end
			end
			
			return _data
		else
			return _data
		end
	end
end

-----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Task : gère la création dynamique des taches associées ensuite à des WP ou à des groupes
-- d'unités.
-----------------------------------------------------------------------------------------------------------

-- Analyse des paramètres pour une tache :
-- _id = 1 pour les paramètres obligatoires (voir déclaration de taskAI)
-- _id = 2 pour les paramètres optionnels (voir déclaration de taskAI)
--
-- Cette analyse se fait argument par argument lié à {...}
-- Le type des arguments est testé.
-- Un argument peut représenter plusieurs paramètres de la tâche (exemple argument cercle qui renseigne
-- un centre et un rayon. Chaque paramètre peut faire appel à une fonction indirecte spécifique ou pas.
-- Un argument peut également lié plusieurs paramètres de la tâche, dans ce cas, la fonction indirecte
-- doit exister et retourner une table avec plusieurs valeurs, une par paramètre.
-- Enfin, chaque paramètre est testé en terme de valeur sur une plage min/max ou énumération.

ATMEFct.C_Task_createNamed = function(_namedTable, _name, _task, _datas)
	if _namedTable[_name] ~= nil then
		ATMEFct.setError("ALREADY_CREATED")
		return true, "named task already exists"
	else
		local d,f = string.find(_name, "|", 1)
		if d == nil then
			local taskDatas = {}
			local err, usedOnZoneOnly, filters, taskType = ATMEFct.C_Task_setParams(_task, taskDatas, _datas)
			
			if err == nil then
				return false, ATMEFct.C_Task_create(_name, _namedTable, _task, usedOnZoneOnly, taskDatas, filters, taskType)
			else
				ATMEFct.setError("BAD_PARAM")
				return true, err
			end
		else
			ATMEFct.setError("BAD_FORMAT")
			return true, "name can't contain '|' character"
		end
	end
end

ATMEFct.C_Task_setArg = function(_name, _arg, _item, _datas)
	if _arg.type == type(_item) or ATMEFct.isObjectClass(_item, _arg.type) == true then
		if _arg.parameter.fct ~= nil then
			local values, taskT = ATMEAIIndFct[_arg.parameter.fct](_item)	-- taskT : Uniquement utilisé pour le callsign aircraft/jtacs sol
																			-- 		   taches liées au JTAC (1 sol, 2 air), sinon 0
			if #values == 0 then
				return true, "datas." .. _name .. " - Bad token '" .. _item .. "'"
			end

			local value = ATMEFct.testDatas(values[1], _arg)
			if value == nil then
				return true, "datas." .. _name .. " - Bad value"
			end
			
			table.insert(_datas, value)
			
			return false, taskT
		else
			local value = ATMEFct.testDatas(_item, _arg)
			if  value == nil then
				return true, "datas." .. _name .. " - Bad value"
			end
			
			table.insert(_datas, value)
			
			return false, 0
		end
	else
		return true, "datas." .. _name .. " - Bad type"
	end
end

ATMEFct.C_Task_setParams = function(_task, _taskDatas, _datas)
	local taskID = string.upper(_task)
	local taskType = 0
	local usedOnZoneOnly = false

	if ATMEConst.taskAI[taskID] ~= nil then
		local index = 1
		
		_taskDatas["id"] = ATMEConst.taskAI[taskID].id
		_taskDatas["params"] = {}
		
		if ATMEConst.taskAI[taskID].const ~= nil then
			for _name, _value in pairs(ATMEConst.taskAI[taskID].const) do
				_taskDatas["params"][_name] = _value
			end
		end
			
		for _argName, _arg in pairs(ATMEConst.taskAI[taskID].args) do
			if _datas ~= nil and _datas[_argName] ~= nil then
				if _arg.multi == true then
					local datas = {}
					if type(_datas[_argName]) == "string" then
						for _item in _datas[_argName]:gmatch("([^%s]+)") do
							local err, value = ATMEFct.C_Task_setArg(_argName, _arg, _item, datas)
							
							if err == true then
								return value
							else
								taskType = value
							end
						end
					elseif type(_datas[_argName]) == "table" then
						for _, _item in ipairs(_datas[_argName]) do
							local err, value = ATMEFct.C_Task_setArg(_argName, _arg, _item, datas)
							
							if err == true then
								return value
							else
								taskType = value
							end
						end
					else
						return "datas." .. _argName .. " - Bad type"
					end

					_taskDatas["params"][_arg.parameter.name] = datas
				elseif type(_datas[_argName]) == _arg.type or ATMEFct.isObjectClass(_datas[_argName], _arg.type) == true then
					if _arg.parameter.fct ~= nil then
						local values, taskT = ATMEAIIndFct[_arg.parameter.fct](_datas[_argName])	-- taskT : Uniquement utilisé pour le callsign aircraft/jtacs sol
																									-- 		   taches liées au JTAC (1 sol, 2 air), sinon 0

						taskType = taskT
						
						if #values == 0 then
							return "datas." .. _argName .. " - Bad value"
						end
						
						for _valueId, _value in ipairs(values) do
							local value = ATMEFct.testDatas(_value, _arg)
							if value ~= nil then
								if _valueId == 1 then
									_taskDatas["params"][_arg.parameter.name] = value
								else
									_taskDatas["params"][_arg.parameter.linkedParams[_valueId - 1]] = value
								end
							else
								return "datas." .. _argName .. " - Bad value"
							end
						end
					elseif type(_datas[_argName]) == _arg.type then
						local value = ATMEFct.testDatas(_datas[_argName], _arg)
						if value ~= nil then
							_taskDatas["params"][_arg.parameter.name] = value
						else
							return "datas." .. _argName .. " - Bad value"
						end
					else
						return "datas." .. _argName .. " - Bad value"
					end
				else
					return "datas." .. _argName .. " - Bad type"
				end
			elseif _arg.mandatory == true then
				return "datas." .. _argName .. " is nil - Not optional"
			end
		end

		if ATMEConst.taskAI[taskID].fctValidate ~= nil then
			local err, value = ATMEAIIndFct[ATMEConst.taskAI[taskID].fctValidate](_taskDatas.params)

			if err == true then
				return value
			end
		end

		return nil, usedOnZoneOnly, ATMEConst.taskAI[taskID].filters, taskType
	else
		return "Task '" .. _task .. "' doesn't exist"
	end
end

ATMEFct.C_Task_getTaskDatas = function(_task, _type)
	local localMembers = ATMEDatas.localMembers[_task:getIndex()]
	

	if localMembers.taskType == 0 or ATMEConst.JTACStatus[_type] == localMembers.taskType then
		local ok = false
		local taskDatas = ATMEFct.copyTable(localMembers.datas)
	
		for _id, _ in pairs(localMembers.filters.type) do
			if _id == _type then
				ok = true
			end
		end
		
		if ok == true then
			if localMembers.stopCondition ~= nil or localMembers.startCondition ~= nil then
				local temp = taskDatas
				taskDatas = { 
					["id"] = "ControlledTask", 
					["params"] = { 
						task = temp,
					} 
				}
				
				if localMembers.stopCondition ~= nil then
					taskDatas.params.stopCondition = ATMEFct.copyTable(localMembers.stopCondition) 
				end
				
				if localMembers.startCondition ~= nil then
					taskDatas.params.condition = ATMEFct.copyTable(localMembers.startCondition)
				end
			end
			
			return true, taskDatas
		else
			return false, "'" .. taskDatas.id .. "' is not allowed"
		end
	end
	
	return false, "'" .. taskDatas.id .. "' bad JTAC callsign"
end

ATMEDatas.C_Task_MT = {
	className = "C_Task",
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["addStopCondition"] = true,
		["removeStopCondition"] = true,
		["addStartCondition"] = true,
		["removeStartCondition"] = true,
		["getUserPrivateDatas"] = true,
	},
	
	add = function(self, _taskID, _usedOnZoneOnly, _taskDatas, _filters, _taskType)
		local item = {
			taskID = _taskID,
			datas = _taskDatas,
			filters = _filters,
			taskType = _taskType,
			usedOnZoneOnly = _usedOnZoneOnly,
			userPrivateDatas = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
	end,
}

ATME.C_Task = {
	getClassName = function(self)
		return ATMEDatas.C_Task_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Task_MT.parentClass
	end,
	
	createNamed = function(_name, _task, _datas)
		local fctDefinition = "Function ATME.C_Task.createNamed(name, taskID, datas)"
		
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" and type(_task) == "string" then
			local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
			local err, newTask = ATMEFct.C_Task_createNamed(localMembersModule.tasks, _name, _task, _datas)				
			
			if err == true then
				ATMEFct.warningOutputCore(nil, fctDefinition, newTask, false)
				return
			end
			
			return newTask
		else
			if type(_name) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
			elseif type(_task) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "taskID type must be string", false)
			end
		end
	end,
	
	getByName = function(_name)
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Task.getByName(name)", "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].tasks[_name]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Task.getByName(name)", "name type must be string", false)
		end
	end,
}

ATMEFct.C_Task_create = function(_name, _tasks, _taskID, _usedOnZoneOnly, _taskDatas, _filters, _taskType)
	local newTask = {}

	setmetatable(newTask, ATMEDatas.C_Task_MT.mt)
	
	local name
	local tagsName
	
	if _name ~= nil and _tasks ~= nil then
		name = _name
		_tasks[name] = newTask
		tagsName = ATMEFct.decompString(_name)
	end
	
	local index = ATMEDatas.C_Task_MT:add(_taskID, _usedOnZoneOnly, _taskDatas, _filters, _taskType)
	local localMembers = ATMEDatas.localMembers[index]
	
	function newTask.getName(self)
		return name
	end
	
	function newTask.getIndex(self)
		return index
	end
	
	function newTask.addStopCondition(self, _condition)
		local fctDefinition = ATMEFct.getErrorPrefix(newTask) .. "addStopCondition(condition)"
		
		if type(_condition) == "string" then
			local stopCondition = {}
			local tokens = ATMEFct.analyzeString(_condition)

			if #tokens == 2 then
				if tokens[1].type == "TOKEN_TYPE_TIME" then
					if tokens[2].type == "TOKEN_INTEGER" then
						if  tokens[2].value > 0 then
							if tokens[2].status == "RELATIVE" then
								stopCondition.duration = tokens[2].value
							else
								stopCondition.time = tokens[2].value
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be > 0 in string format condition", false)
							return
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be integer or time value in string format condition", false)
						return
					end
				elseif tokens[1].type == "TOKEN_TYPE_WAYPOINT" then
					if tokens[2].type == "TOKEN_INTEGER" then
						if tokens[2].value >= 0 then
							stopCondition.lastWaypoint = tokens[2].value
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be >= 0 in string format condition", false)
							return
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be integer or string in string format condition", false)
						return
					end
				elseif tokens[1].type == "TOKEN_TYPE_FLAG" then
					stopCondition.userFlagValue = true
			
					if tokens[2].type == "TOKEN_INTEGER" then
						if tokens[2].value > 0 then
							stopCondition.userFlag = string.format("%d", tokens[2].value)
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be > 0 in string format condition", false)
							return
						end
					elseif tokens[2].type == "TOKEN_STRING" then
						stopCondition.userFlag = tokens[2].value
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be integer or string in string format condition", false)
						return
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "Bad string format condition", false)
					return
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Bad string format condition", false)
				return
			end
			
			if localMembers.stopCondition == nil then
				localMembers.stopCondition = stopCondition
			else
				for _id, _item in pairs(stopCondition) do
					localMembers.stopCondition[_id] = _item
				end
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "condition type must be string", true)
		end
	end
	
	function newTask.removeStopCondition(self)
		localMembers.stopCondition = nil
	end
	
	function newTask.addStartCondition(self,_condition)
		if type(_condition) == "string" then
			local startCondition = {}
			local tokens = ATMEFct.analyzeString(_condition)

			if #tokens == 2 then
				if tokens[1].type == "TOKEN_TYPE_TIME" then
					if tokens[2].type == "TOKEN_INTEGER" then
						if  tokens[2].value > 0 then
							if tokens[2].status == "RELATIVE" then
								startCondition.duration = tokens[2].value
							else
								startCondition.time = tokens[2].value
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be > 0 in string format condition", false)
							return
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be integer or time value in string format condition", false)
						return
					end
				elseif tokens[1].type == "TOKEN_TYPE_WAYPOINT" then
					if tokens[2].type == "TOKEN_INTEGER" then
						if tokens[2].value >= 0 then
							startCondition.lastWaypoint = tokens[2].value
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be >= 0 in string format condition", false)
							return
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be integer or string in string format condition", false)
						return
					end
				elseif tokens[1].type == "TOKEN_TYPE_FLAG" then
					startCondition.userFlagValue = true
			
					if tokens[2].type == "TOKEN_INTEGER" then
						if tokens[2].value > 0 then
							startCondition.userFlag = string.format("%d", tokens[2].value)
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be > 0 in string format condition", false)
							return
						end
					elseif tokens[2].type == "TOKEN_STRING" then
						startCondition.userFlag = tokens[2].value
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be integer or string in string format condition", false)
						return
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "Bad string format condition", false)
					return
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Bad string format condition", false)
				return
			end
			
			if localMembers.startCondition == nil then
				localMembers.startCondition = startCondition
			else
				for _id, _item in pairs(startCondition) do
					localMembers.startCondition[_id] = _item
				end
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "condition type must be string", true)
		end
	end
	
	function newTask.removeStartCondition(self)
		localMembers.startCondition = nil
	end
	
	function newTask.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newTask) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newTask.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newTask.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newTask) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newTask.getUserPrivateDatas(self, _moduleName)
		local fctDefinition = ATMEFct.getErrorPrefix(newTask) .. "getUserPrivateDatas([moduleName])"
		local moduleName
		
		if _moduleName ~= nil then
			if type(_moduleName) == "string" then
				if localMembers.userPrivateDatas[_moduleName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
					return
				else
					moduleName = _moduleName
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
				return
			end
		else
			moduleName = ATMEDatas.currentModuleName
		end
		
		return localMembers.userPrivateDatas[moduleName]
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return newTask
end

setmetatable(ATME.C_Task,ATMEDatas.C_Task_MT)

ATMEDatas.C_Task_MT.__call = function(t, _task, _datas)
	local fctDefinition = "Constructor ATME.C_Task(taskID, datas)"
	
	if ATMEDatas.startMissionOk < 2 then
		ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
		return
	end
	
	if type(_task) == "string" then
		local taskDatas = {}

		local err, usedOnZoneOnly, filters, taskType = ATMEFct.C_Task_setParams(_task, taskDatas, _datas)
		
		if err == nil then
			return ATMEFct.C_Task_create(nil, nil, _task, usedOnZoneOnly, taskDatas, filters, taskType)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "Task '" .. _task .. "' : " .. err, false)
		end
	else
		if type(_task) ~= "string" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "taskID type must be string", false)
		end
	end
end

ATMEDatas.C_Task_MT.mt.__metatable = "Private"
ATMEDatas.C_Task_MT.mt.__index = ATME.C_Task
ATMEDatas.C_Task_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Task_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Task : Bad key " .. k, "table locked", false)
	end
end

-----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Route : gère la création dynamique de route composée de WP
--
-- ATTENTION : WP 0 dans l'éditeur de mission est waypoints[1] de cette classe avec les coordonnées de la
-- première unité du groupe.
-----------------------------------------------------------------------------------------------------------
-- Set option des AI
ATMEFct.C_setOption = function(_type, _option, _param)
	local optionName = string.upper(_option)

	if ATMEConst.optionAI[_type][optionName] ~= nil then
		local tokens
		local value
		
		if type(_param) == "string" then
			tokens = ATMEFct.decompStringSpaceOnly(_param)
			
			if #tokens > 0 then
				local tokenName = string.upper(tokens[1])
				
				if ATMEConst.optionAI[_type][optionName].enum[tokenName] ~= nil then
					if #tokens ~= #ATMEConst.optionAI[_type][optionName].enum[tokenName].params + 1 then
						return true, "Option '" .. _option .. "' : string value must have " .. #ATMEConst.optionAI[_type][optionName].enum[tokenName].params + 1 .. " parameter(s) if parameter #1 is '" .. tokenName .. "'"
					end
				elseif ATMEConst.optionAI[_type][optionName].fct ~= nil then
					local values, taskT = ATMEAIIndFct[ATMEConst.optionAI[_type][optionName].fct](_param)

					if #values == 0 then
						return true, "Option '" .. _option .. "' : Bad Value"
					else
						if type(values[1]) == "number" then
							return false, optionName, values[1]
						else
							return false, optionName, values
						end
					end
				else
					return true, "Option '" .. _option .. "' : parameter #1 - Bad value '" .. tokenName .. "'"
				end
				
				value = ATMEConst.optionAI[_type][optionName].enum[tokenName].value
				for _i, _item in ipairs(ATMEConst.optionAI[_type][optionName].enum[tokenName].params) do
					local subTokenName = string.upper(tokens[_i + 1])
					
					if ATMEConst.optionAI[_type][optionName].enum[tokenName].params[_i].enum[subTokenName] ~= nil then
						value = value + ATMEConst.optionAI[_type][optionName].enum[tokenName].params[_i].enum[subTokenName]
					else
						return true, "Option '" .. _option .. "' : parameter #" .. _i + 1 .. " - Bad value '" .. subTokenName .. "'"
					end
				end
			else
				return true, "Option '" .. _option .. "' : string value must have " .. #ATMEConst.optionAI[_type][optionName].enum.params + 1 .. " parameter(s)"
			end
		else
			if type(_param) == "number" then
				if ATMEConst.optionAI[_type][optionName].max ~= nil then
					if _param > ATMEConst.optionAI[_type][optionName].max then
						_param = ATMEConst.optionAI[_type][optionName].max
					end
				end
				
				if ATMEConst.optionAI[_type][optionName].min ~= nil then
					if _param < ATMEConst.optionAI[_type][optionName].min then
						_param = ATMEConst.optionAI[_type][optionName].min
					end
				end
				
				value = _param
			else
				value = _param
			end
		end
		
		return false, optionName, value
	else
		return true, "Option '" .. _option .. "' doesn't exist"
	end
end

ATMEFct.C_Route_createNamed = function(_namedTable, _name, _typeRoute, _rebuild)
	if _namedTable[_name] ~= nil then
		ATMEFct.setError("ALREADY_CREATED")
		return true, "named task already exists"
	else
		local d,f = string.find(_name, "|", 1)
		if d == nil then
			if ATMEConst.categoryNameFilters[_typeRoute] ~= nil then
				local rebuild = _rebuild
				if rebuild == nil then
					rebuild = false
				end
				
				return false, ATMEFct.C_Route_create(_name, _namedTable, ATMEConst.categoryNameFilters[_typeRoute], rebuild)
			else
				return true, "type must be 'AIRPLANE', 'HELICOPTER', 'GROUND' or 'SHIP'"
			end
		else
			ATMEFct.setError("BAD_FORMAT")
			return true, "name can't contain '|' character"
		end
	end
end

ATMEFct.C_Route_getWaypointPosition = function(_route, _position, _isGroundForAircraft)
	local localMembers = ATMEDatas.localMembers[_route:getIndex()]
	local surfaceFilter = "LAND"
	local inWater = false
	local position
	
	if localMembers.type == "ship" then
		surfaceFilter = "WATER"
		inWater = true
	elseif localMembers.type == "plane" or localMembers.type == "helicopter" then
		if _isGroundForAircraft == false then
			surfaceFilter = "ALL"
		end
	end
	
	if ATME.isPoint(_position) ~= POINT_BAD then
		position = ATMEFct.convertToPoint3D(_position)
		
		if surfaceFilter ~= "ALL" and ATMEFct.isSurfaceWater(position) ~= inWater then
			return true, "Bad surface for position"
		end
	elseif type(_position) == "string" then
		local zone = ATMEFct.getDCSZone(_position)
		
		if zone ~= nil then
			position = ATMEFct.C_Circle_create(nil, nil, zone.point, zone.radius)
		else
			return true, "DCS zone " .. _position .. " unknown"
		end
	else
		position = _position
	end
	
	return false, position, surfaceFilter
end								

ATMEFct.C_Route_createWaypoints = function(_route)
	local localMembers = ATMEDatas.localMembers[_route:getIndex()]

	if localMembers.waypoints == nil or localMembers.forced == true then
		localMembers.waypoints = {}
		if #localMembers.initWP > 0 then
			for _index, _initWP in ipairs(localMembers.initWP) do
				local absPoint
				local altitude = _initWP.alt
				
				if _initWP.type ~= "WAYPOINT" then
					if _initWP.position ~= nil then
						local localMembersZone = ATMEDatas.localMembers[_initWP.position:getIndex()]
						local datas = {
							zone = _initWP.position,
							surfaceFilter = _initWP.surfaceFilter,
							isFlat = false,
							altMin = 0,
							localMembers = localMembersZone,
						}

						local position = ATMEFct.C_Zone_randomPositionTest(_initWP.position, datas)

						if position ~= nil then
							absPoint = {
								x = position.x,
								y = _initWP.alt,
								z = position.z,
							}
						else
							return true
						end
					elseif _initWP.localMembersUnit ~= nil then
						if _initWP.localMembersUnit.objectDCS:isExist() == true then
							local point = {
								x = _initWP.x,
								y = _initWP.alt,
								z = _initWP.y,
							}

							local vector = ATMEFct.vectors["3D"].createFromTable(nil, nil, _initWP.localMembersUnit.objectDCS:getPosition().x)
							local pos = ATMEFct.coordLocalToDCS(point, _initWP.localMembersUnit.objectDCS:getPoint(), vector:getAzimuth())

							absPoint = {
								x = pos.x,
								y = _initWP.alt,
								z = pos.z,
							}
						else
							absPoint = {
								x = _initWP.x0,
								y = _initWP.alt,
								z = _initWP.y0,
							}
						end
					else
						absPoint = {
							x = _initWP.x,
							y = _initWP.alt,
							z = _initWP.y,
						}
					end
					
					if _initWP.action == "On Road" then
						absPoint.x, absPoint.z = land.getClosestPointOnRoads("roads", absPoint.x, absPoint.z)
					end
					
					if  _initWP.alt == nil or ATMEConst.altRouteSet[localMembers.type] == true then
						absPoint.y = ATMEFct.getSurfaceAltitude(absPoint)
					end

					table.insert(localMembers.waypoints, {
						x = absPoint.x,
						y = absPoint.z,
						alt = absPoint.y,
						name = _initWP.name,
						speed = _initWP.speed,
						speed_locked = false,
						alt_type = "BARO",
						formation_template = "",
						type = _initWP.typeWP,
						action = _initWP.action,
						helipadId = _initWP.helipadId,
						airdromeId = _initWP.airdromeId,
						linkUnit = _initWP.linkUnit,
						task = {
							["id"] = "ComboTask",
							["params"] = 
							{
								["tasks"] = 
								{
								},
							},
						},
					})
				else
					table.insert(localMembers.waypoints, ATMEFct.copyTable(_initWP.datas))
				end
				
				for _, _task in ipairs(_initWP.tasks) do
					table.insert(localMembers.waypoints[#localMembers.waypoints].task.params.tasks, _task)
				end
			end
			
			if #localMembers.waypoints == 0 then
				return true
			end
		else
			return true
		end
	end

	local waypoints = {}
	local waypointsName = {}
	
	for _i, _waypoint in ipairs(localMembers.waypoints) do
		if _waypoint.name ~= "" and waypointsName[_waypoint.name] == nil then
			waypointsName[_waypoint.name] = _i - 1
		end
		
		table.insert(waypoints, _waypoint)
	end
	
	return false, waypoints, waypointsName
end

ATMEFct.C_Route_addCurrentPositionWaypoint = function(_waypoints, _group)
	local waypoint = ATMEFct.copyTable(_waypoints[1])
	
	local localMembers = ATMEDatas.localMembers[_group:getIndex()]
	
	if ATMEConst.canFly[localMembers.category] == true and ATMEFct.isTakeOffWaypoint(_waypoints[1].type) == true then
		table.remove(_waypoints, 1)
	end

	if #_waypoints > 0 then
		local unit = ATMEFct.getFirstUnit(_group)
		local localMembers = ATMEDatas.localMembers[unit:getIndex()]
		local pos = localMembers.objectDCS:getPoint()
		
		waypoint.x = pos.x
		waypoint.y = pos.z
		waypoint.alt = pos.y
		waypoint.speed_locked = false
		waypoint.name = nil
		waypoint.task = {
			["id"] = "ComboTask",
			["params"] = 
			{
				["tasks"] = 
				{
				},
			},
		}
		
		table.insert(_waypoints, 1, waypoint)
	end
end

ATMEFct.setAirbaseToNewPoint = function(_newPoint, _airbase)
	local localMembersAirbase = ATMEDatas.localMembers[_airbase:getIndex()]
			
	_newPoint.airdromeId = localMembersAirbase.airdromeId
	_newPoint.helipadId = localMembersAirbase.helipadId
	_newPoint.linkUnit = localMembersAirbase.helipadId
	_newPoint.airbaseName = _airbase:getName()
	_newPoint.airbaseCategoryName = _airbase:getCategoryName()
end

ATMEFct.setGroundActionFunctions = {
	["Turning Point"] = function(_newPoint, _datas, _localMembers)
		if _datas.action == nil then
			_newPoint.action = "Turning Point"
			if _localMembers.type == "vehicle" then
				_newPoint.action = "Off Road"
			elseif _localMembers.type == "ship" then
				_newPoint.action = "Turning Point"
			end
		else
			_newPoint.action = _datas.action
		end
	end,
	["TakeOffParkingHot"] = function(_newPoint, _datas, _localMembers)
		ATMEFct.setAirbaseToNewPoint(_newPoint, _datas.airbase)
		_newPoint.action = "From Parking Area"
	end,
	["TakeOffParking"] = function(_newPoint, _datas, _localMembers)
		ATMEFct.setAirbaseToNewPoint(_newPoint, _datas.airbase)
		_newPoint.action = "From Parking Area"
	end,
	["TakeOffGroundHot"] = function(_newPoint, _datas, _localMembers)
		_newPoint.action = "From Ground Area"
	end,
	["TakeOffGround"] = function(_newPoint, _datas, _localMembers)
		_newPoint.action = "From Ground Area"
	end,
	["Land"] = function(_newPoint, _datas, _localMembers)
		ATMEFct.setAirbaseToNewPoint(_newPoint, _datas.airbase)
		_newPoint.action = "Landing"
	end,
}

ATMEFct.addInitWP = function(_route, _fctDefinition, _datas, _copy)
	local newPoint
	local localMembers = ATMEDatas.localMembers[_route:getIndex()]

	if _copy == false then	
		newPoint = {
			type = "POINT",
			x = _datas.x,
			y = _datas.y,
			alt = _datas.altitude,
			speed = _datas.speed,
			typeWP = _datas.typeWP,
			localMembersUnit = _datas.localMembersUnit,
			position = _datas.position,
			surfaceFilter = _datas.surfaceFilter,
		}

		if ATMEFct.setGroundActionFunctions[_datas.typeWP] ~= nil then
			ATMEFct.setGroundActionFunctions[_datas.typeWP](newPoint, _datas, localMembers)
		end

		newPoint.name = ""
		newPoint.tasks = {}
	else
		newPoint = {
			type = "WAYPOINT",
			datas = _datas,
			tasks = {},
		}
	end
	
	
	-- Si le dernier WP est landing, il est supprimé.
	if #localMembers.initWP > 0 then
		if localMembers.initWP[#localMembers.initWP].typeWP == "Land" then
			table.remove(localMembers.initWP)
		end
	end
	
	table.insert(localMembers.initWP, newPoint)

	-- On retourne le numéro du waypoint qui débute à 1, le waypoint 0 est créé à l'affectation de la route.
	return #localMembers.initWP
end		

ATMEFct.addWPFrom = function(_route, _type, _waypoints, _withTask, _fromWP, _nbWP)
	local waypoints = ATMEFct.copyAllWaypoints(_waypoints)
	local localMembers = ATMEDatas.localMembers[_route:getIndex()]
		
	if localMembers.type == _type then
		ATMEFct.C_GroupCloningContext_startRouteAt(waypoints, _fromWP, 0)

		if ATMEFct.isTakeOffWaypoint(waypoints[1].type) == true then
			if #localMembers.initWP ~= 0 then
				return true, "can't add a takeoff waypoint after existing waypoints"
			end
		end
		
		local first = #localMembers.initWP
		local last
		local nb = _nbWP

		if nb > #waypoints then
			nb = #waypoints
		end
		
		for _index = 1, nb do
			if _withTask == false then
				waypoints[_index].task = {
					["id"] = "ComboTask",
					["params"] = {
						["tasks"] = {
						},
					},
				}
			end
			
			last = ATMEFct.addInitWP(_route, fctDefinition, waypoints[_index], true)
		end
		
		return false, first, last
	else
		return true, "Route types are different"
	end
end

ATMEDatas.C_Route_MT = {
	className = "C_Route",
	mt = {},
	
	keys = {
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["addRampHotWaypoint"] = true,
		["addGroundWaypoint"] = true,
		["addGroundHotWaypoint"] = true,
		["addRampWaypoint"] = true,
		["addUnitRelativePolarWaypoint"] = true,
		["addReferencePolarWaypoint"] = true,
		["addAbsoluteWaypoint"] = true,
		["addLandingWaypoint"] = true,
		["waypointAddTask"] = true,
		["waypointAddOption"] = true,
		["waypointSetName"] = true,
		["getIndex"] = true,
		["addWaypointsFrom"] = true,
		["removeAllWaypoints"] = true,
		["getLandingAirbase"] = true,
		["getNbWaypoints"] = true,
		["getWaypoint"] = true,
		["getUserPrivateDatas"] = true,
	},
	
	add = function(self, _type, _forced)
		local item = {
			initWP = {},
			waypoints = nil,
			waypointsName = {},
			type = _type,
			forced = _forced,
			userPrivateDatas = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].initWP = nil
		ATMEDatas.localMembers[_id].waypoints = nil
	end,
}

ATME.C_Route = {
	getClassName = function(self)
		return ATMEDatas.C_Route_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Route_MT.parentClass
	end,
	
	createNamed = function(_name, _typeRoute, _rebuild)
		local fctDefinition = "Function ATME.C_Route.createNamed(name, typeRoute, [rebuild])"
		
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
			return
		end
		
		if type(_typeRoute) == "string" and (_rebuild == nil or type(_rebuild) == "boolean") then
			local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
			
			local err, newRoute = ATMEFct.C_Route_createNamed(localMembersModule.routes, _name, _typeRoute, _rebuild)
			if err == true then
				if ATMEDatas.error ~= nil then
					ATMEFct.warningOutputCore(nil, fctDefinition, newRoute, false)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, newRoute, false)
				end
				
				return
			end
			
			return newRoute
		else
			if type(_typeRoute) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "typeRoute type must be string", false)
			elseif _rebuild ~= nil and type(_rebuild) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "rebuild type must be nil or boolean", false)
			end
		end
	end,
	
	getByName = function(_name)
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Route.getByName(name)", "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].routes[_name]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Route.getByName(name)", "name type must be string", false)
		end
	end,
}

ATMEFct.C_Route_create = function(_name, _routes, _type, _forced)	
	local newRoute = {}
	
	setmetatable(newRoute, ATMEDatas.C_Route_MT.mt)
	
	local name
	local tagsName
	
	if _name ~= nil and _routes ~= nil then
		name = _name
		_routes[name] = newRoute
		tagsName = ATMEFct.decompString(_name)
	end
	
	local index = ATMEDatas.C_Route_MT:add(_type, _forced)
	local localMembers = ATMEDatas.localMembers[index]
	local landingWaypointExist = false
	
	local function analyzeAltitudeAction(_typeRoute, ...)
		local args = {...}
		local altitude
		local action
		local err
		
		if type(args[1]) == "number" then
			altitude = args[1]
			if type(args[2]) == "string" then
				if ATMEConst.WPActions[_typeRoute][args[2]] ~= nil then
					action = ATMEConst.WPActions[_typeRoute][args[2]]
				else
					err = "action must be a correct value. See doc for further informations"
				end
			elseif args[2] ~= nil then
				err = "action type must be nil or string"
			end
		elseif type(args[1]) == "string" then
			if ATMEConst.WPActions[_typeRoute][args[1]] ~= nil then
				action = ATMEConst.WPActions[_typeRoute][args[1]]
			else
				err = "action must be a correct value. See doc for further informations"
			end
		elseif args[1] ~= nil then
			err = "altitude type must be nil or number"
		end

		return err, altitude, action
	end
	
	function newRoute.getIndex(self)
		return index
	end
	
	function newRoute.getName(self)
		return name
	end
	
	function newRoute.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newRoute) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newRoute.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newRoute.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newRoute) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newRoute.removeAllWaypoints(self)
		localMembers.initWP = {}
		localMembers.waypoints = nil
		localMembers.waypointsName = nil
		landingWaypointExist = false
	end
	
	function newRoute.addUnitRelativePolarWaypoint(self, _unit, _azimuth, _distance, _speed, ...)
		local fctDefinition = ATMEFct.getErrorPrefix(newRoute) .. "addUnitRelativePolarWaypoint(unit, azimuth, distance, speed)\naddUnitRelativePolarWaypoint(unit, azimuth, distance, speed, altitude)\naddUnitRelativePolarWaypoint(unit, azimuth, distance, speed, action)\naddUnitRelativePolarWaypoint(unit, azimuth, distance, speed, altitude, action)"
		
		if ATMEFct.isObjectClass(_unit, ATME.C_Unit) == true and type(_azimuth) == "number" and type(_distance) == "number" and 
				type(_speed) == "number" then
			
			local err, altitude, action = analyzeAltitudeAction(localMembers.type, ...)

			if err == nil then
				if _speed > 0 then
					local azimuthRad =  math.rad(ATMEFct.getAzimuthMod(_azimuth, 360))
					local localMembersUnit = ATMEDatas.localMembers[_unit:getIndex()]
					local pos = localMembersUnit.objectDCS:getPoint()
					
					local point = {
						x = _distance * math.cos(azimuthRad),
						y = _distance * math.sin(azimuthRad),
					}

					local datas = {
						x = point.x,
						y = point.y,
						speed = _speed,
						altitude = altitude,
						typeWP = "Turning Point",
						action = action,
						x0 = point.x + pos.x,
						y0 = point.y + pos.z,
						localMembersUnit = localMembersUnit,
					}
					
					return ATMEFct.addInitWP(self, fctDefinition, datas, false)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "speed must be > 0", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, err, false)
			end
		else
			if ATMEFct.isObjectClass(_unit, ATME.C_Unit) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "unit type must be ATME.C_Unit", true)
			elseif type(_azimuth) ~= "number" then 
				ATMEFct.errorOutputCore(nil, fctDefinition, "azimuth type must be number", false)
			elseif type(_distance) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "distance type must be number", false)
			elseif type(_speed) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "speed type must be number", false)
			end
		end
	end
	
	function newRoute.addReferencePolarWaypoint(self, _reference, _azimuth, _distance, _speed, ...)
		local fctDefinition = ATMEFct.getErrorPrefix(newRoute) .. "addReferencePolarWaypoint(reference, azimuth, distance, speed)\naddReferencePolarWaypoint(reference, azimuth, distance, speed, altitude)\naddReferencePolarWaypoint(reference, azimuth, distance, speed, action)\naddReferencePolarWaypoint(reference, azimuth, distance, speed, altitude, action)"
		
		if ATMEFct.isReference(_reference) == true and type(_azimuth) == "number" and type(_distance) == "number" and 
				type(_speed) == "number" then
			
			if landingWaypointExist == false then
				local err, altitude, action = analyzeAltitudeAction(localMembers.type, ...)

				if err == nil then
					if _speed > 0 then
						local azimuthRad =  math.rad(ATMEFct.getAzimuthMod(_azimuth, 360))
						local pos = ATMEFct.getPointFromReference(_reference)
						
						local datas = {
							x = _distance * math.cos(azimuthRad) + pos.x,
							y = _distance * math.sin(azimuthRad) + pos.z,
							speed = _speed,
							altitude = altitude,
							typeWP = "Turning Point",
							action = action,
						}
						
						return ATMEFct.addInitWP(self, fctDefinition, datas, false)
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "speed must be > 0", false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, err, false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "can't a new waypoint after a landing waypoint", false)
			end
		else
			if ATME.isReference(_reference) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
			elseif type(_azimuth) ~= "number" then 
				ATMEFct.errorOutputCore(nil, fctDefinition, "azimuth type must be number", false)
			elseif type(_distance) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "distance type must be number", false)
			elseif type(_speed) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "speed type must be number", false)
			end
		end
	end
	
	function newRoute.addAbsoluteWaypoint(self, _position, _speed, ...)
		local fctDefinition = ATMEFct.getErrorPrefix(newRoute) .. "addAbsoluteWaypoint(position, speed)\naddAbsoluteWaypoint(position, speed, altitude)\naddAbsoluteWaypoint(position, speed, action)\naddAbsoluteWaypoint(position, speed, altitude, action)"
		
		if (ATME.isPoint(_position) ~= POINT_BAD or type(_position) == "string" or ATMEFct.isObjectClass(_position, ATME.C_Zone) == true) and 
				type(_speed) == "number" then
			
			if landingWaypointExist == false then
				local err, altitude, action = analyzeAltitudeAction(localMembers.type, ...)
				
				if err == nil then
					local err1, position, surfaceFilter = ATMEFct.C_Route_getWaypointPosition(self, _position, false)
					
					if err1 == false then
						if _speed >= 0 then
							local datas
							
							if ATME.isPoint(position) ~= POINT_BAD then
								datas = {
									x = position.x,
									y = position.z,
									speed = _speed,
									altitude = altitude,
									typeWP = "Turning Point",
									action = action,
								}
							else
								datas = {
									speed = _speed,
									altitude = altitude,
									typeWP = "Turning Point",
									action = action,
									position = position,
									surfaceFilter = surfaceFilter,
								}
							end

							return ATMEFct.addInitWP(self, fctDefinition, datas, false)
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "speed must be >= 0", false)
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, position, false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, err, false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "can't a new waypoint after a landing waypoint", false)
			end
		else
			if ATME.isPoint(_position) == POINT_BAD and type(_position) ~= "string" and ATMEFct.isObjectClass(_position, ATME.C_Zone) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be string, ATME.C_Zone or table (x,y)/(x,y,z) ", true)
			elseif type(_speed) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "speed type must be number", false)
			end
		end
	end
	
	function newRoute.addRampWaypoint(self, _airbase)
		local fctDefinition = ATMEFct.getErrorPrefix(newRoute) .. "addRampWaypoint(airbase)"
		
		if #localMembers.initWP == 0 then
			if localMembers.type == "plane" or localMembers.type == "helicopter" then
				if ATMEFct.isObjectClass(_airbase, ATME.C_Airbase) == true then
					local point = _airbase:getPosition()
					
					local datas = {
						x = point.x,
						y = point.z,
						altitude = point.y,
						typeWP = "TakeOffParking",
						airbase = _airbase,
					}

					return ATMEFct.addInitWP(self, fctDefinition, datas, false)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "airbase type must be ATME.C_Airbase", true)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "reserved to 'AIRPLANE' or 'HELICOPTER' route", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "can't add a takeoff waypoint after existing waypoints", false)
		end
	end
	
	function newRoute.addRampHotWaypoint(self, _airbase)
		local fctDefinition = ATMEFct.getErrorPrefix(newRoute) .. "addRampHotWaypoint(airbase)"
		
		if #localMembers.initWP == 0 then
			if localMembers.type == "plane" or localMembers.type == "helicopter" then
				if ATMEFct.isObjectClass(_airbase, ATME.C_Airbase) == true then
					local point = _airbase:getPosition()
					
					local datas = {
						x = point.x,
						y = point.z,
						altitude = point.y,
						typeWP = "TakeOffParkingHot",
						airbase = _airbase,
					}

					return ATMEFct.addInitWP(self, fctDefinition, datas, false)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "airbase type must be ATME.C_Airbase", true)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "reserved to 'AIRPLANE' or 'HELICOPTER' route", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "can't add a takeoff waypoint after existing waypoints", false)
		end
	end
	
	function newRoute.addGroundWaypoint(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newRoute) .. "addGroundWaypoint(position)"

		if #localMembers.initWP == 0 then
			if localMembers.type == "plane" or localMembers.type == "helicopter" then
				if ATME.isPoint(_position) ~= POINT_BAD or type(_position) == "string" or ATMEFct.isObjectClass(_position, ATME.C_Zone) == true then
					local err, position, surfaceFilter = ATMEFct.C_Route_getWaypointPosition(self, _position, true)
					
					if err == false then
						local datas
						
						if ATME.isPoint(position) ~= POINT_BAD then
							datas = {
								x = position.x,
								y = position.z,
								typeWP = "TakeOffGround",
								action = action,
							}
						else
							datas = {
								typeWP = "TakeOffGround",
								position = position,
								surfaceFilter = surfaceFilter,
								action = action,
							}
						end
						
						return ATMEFct.addInitWP(self, fctDefinition, datas, false)
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, position, false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be string, ATME.C_Zone or table (x,y)/(x,y,z) ", true)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "reserved to 'AIRPLANE' or 'HELICOPTER' route", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "can't add a takeoff waypoint after existing waypoints", false)
		end
	end
	
	function newRoute.addGroundHotWaypoint(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newRoute) .. "addGroundHotWaypoint(position)"
		
		if #localMembers.initWP == 0 then
			if localMembers.type == "plane" or localMembers.type == "helicopter" then
				if ATME.isPoint(_position) ~= POINT_BAD or type(_position) == "string" or ATMEFct.isObjectClass(_position, ATME.C_Zone) == true then
					local err, position, surfaceFilter = ATMEFct.C_Route_getWaypointPosition(self, _position, true)
					
					if err == false then
						local datas
						
						if ATME.isPoint(position) ~= POINT_BAD then
							datas = {
								x = position.x,
								y = position.z,
								typeWP = "TakeOffGroundHot",
								action = action,
							}
						else
							datas = {
								typeWP = "TakeOffGroundHot",
								position = position,
								surfaceFilter = surfaceFilter,
								action = action,
							}
						end
						
						return ATMEFct.addInitWP(self, fctDefinition, datas, false)
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, position, false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be string, ATME.C_Zone or table (x,y)/(x,y,z) ", true)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "reserved to 'AIRPLANE' or 'HELICOPTER' route", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "can't add a takeoff waypoint after existing waypoints", false)
		end
	end
	
	function newRoute.addLandingWaypoint(self, _airbase)
		local fctDefinition = ATMEFct.getErrorPrefix(newRoute) .. "addLandingWaypoint(airbase)"
		
		if localMembers.type == "plane" or localMembers.type == "helicopter" then
			if ATMEFct.isObjectClass(_airbase, ATME.C_Airbase) == true then
				if _airbase:isHelipad() == true and localMembers.type == "plane" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Only helicopters can land on Helipad", false)
					return
				end
				
				local point = _airbase:getPosition()
				
				local datas = {
					x = point.x,
					y = point.z,
					altitude = point.y,
					typeWP = "Land",
					airbase = _airbase,
				}
				
				landingWaypointExist = true
				
				return ATMEFct.addInitWP(self, fctDefinition, datas, false)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "airbase type must be ATME.C_Airbase", true)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reserved to 'AIRPLANE' or 'HELICOPTER' route", false)
		end
	end
	
	function newRoute.addWaypointsFrom(self, _origin, ...)
		local fctDefinition = ATMEFct.getErrorPrefix(newRoute) .. "addWaypointsFrom(origin)\addWaypointsFrom(origin, fromWP, [nbWP])\addWaypointsFrom(origin, withTask)\addWaypointsFrom(origin, withTask, fromWP, [nbWP])"
		local args = {...}

		if type(_origin) == "string" or ATMEFct.isObjectClass(_origin, ATME.C_Group) == true or 
				ATMEFct.isObjectClass(_origin, ATME.C_GroupCloningContext) == true or ATMEFct.isObjectClass(_origin, ATME.C_Route) == true then
				
			local withTask = false
			local fromWP = 0
			local nbWP = 0
			local typeRoute
			local waypoints
			local waypointsName
			local fromRoute = false
			local localMembersToCopy
			
			if type(_origin) == "string" then
				if ATMEDatas.allGroupsInMission[_origin] ~= nil then
					typeRoute = ATMEDatas.allGroupsInMission[_origin].inMissionSpawnDatas.type
					waypoints = ATMEDatas.allGroupsInMission[_origin].inMissionSpawnDatas.datas.route.points
					waypointsName = ATMEDatas.allGroupsInMission[_origin].inMissionSpawnDatas.waypointsName
					nbWP = #waypoints
				else
					ATMEFct.setError("NO_GROUP")
					ATMEFct.warningOutputCore(nil, fctDefinition, "Group '" .. _origin .. "' not defined in Mission Editor", false)
					return
				end
			else
				if ATMEFct.isObjectClass(_origin, ATME.C_Group) == true then
					localMembersToCopy = ATMEDatas.localMembers[_origin:getIndex()]
					
					typeRoute = localMembersToCopy.spawnDatas.type
					waypoints = localMembersToCopy.initialSpawnDatas.datas.route.points
					waypointsName = localMembersToCopy.initialWaypointsName
					nbWP = #waypoints
				elseif ATMEFct.isObjectClass(_origin, ATME.C_GroupCloningContext) == true then
					localMembersToCopy = ATMEDatas.localMembers[_origin:getIndex()]
					
					typeRoute = localMembersToCopy.spawnDatas.type
					waypoints = localMembersToCopy.spawnDatas.datas.route.points
					waypointsName = localMembersToCopy.spawnDatas.waypointsName
					nbWP = #waypoints
				else
					localMembersToCopy = ATMEDatas.localMembers[_origin:getIndex()]
					typeRoute = localMembersToCopy.type
					nbWP = #localMembersToCopy.initWP
					fromRoute = true
				end
			end
			
			
			if #args > 0 then
				local index = 1
				if type(args[1]) == "boolean" then
					withTask = args[1]
					index = index + 1
				end
				
				if args[index] == nil or ATMEFct.isReference(args[index]) == true or 
						type(args[index]) == "string" or type(args[index]) == "number" then
				
					if args[index] ~= nil then
						if type(args[index]) == "string" then
							if waypointsName[args[index]] ~= nil then
								fromWP = waypoints[waypointsName[args[index]] + 1]
							else
								ATMEFct.setError("NO_COPY")
								ATMEFct.warningOutputCore(nil, fctDefinition, "No waypoint copied - fromWP '" .. args[index] .. "' doesn't exist in current route" , false)
								return
							end
						elseif type(args[index]) == "number" then
							fromWP = math.floor(args[index])
							
							if fromWP < 0 then
								fromWP = 0
							end
						end

						index = index + 1
						if type(args[index]) == "number" then
							nbWP = math.floor(args[index])
							
							if nbWP <= 0 then
								ATMEFct.setError("NO_COPY")
								ATMEFct.warningOutputCore(nil, fctDefinition, "No waypoint copied - Number of waypoints <= 0" , false)
								return
							end
						elseif args[index] ~= nil then
							ATMEFct.errorOutputCore(nil, fctDefinition, "nbWP type must be nil or number", false)
							return
						end
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "fromWP type must be nil, string, number," .. ATMEConst.errorLabelReference, false)
					return
				end
			end
			
			if fromRoute == false then
				local err, first, last = ATMEFct.addWPFrom(self, typeRoute, waypoints, withTask, fromWP, nbWP)

				if err == false then
					return first, last
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, first, false)
				end
			else
				local maxWP = fromWP + nbWP
				if maxWP > #localMembersToCopy.initWP then
					maxWP = #localMembersToCopy.initWP
				end

				for _i = fromWP + 1, maxWP do
					local currentWP = ATMEFct.copyTable(localMembersToCopy.initWP[_i])
					
					if withTask == false then
						currentWP.tasks = {}
					end
					
					table.insert(localMembers.initWP, currentWP)
				end
				
				return startIndex - 1, #localMembers.initWP - 1
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "origin type must be string, ATME.C_Route, ATME.C_GroupCloningContext or ATME.C_Group", true)
		end
	end
	
	function newRoute.waypointSetName(self, _numWaypoint, _name)
		local fctDefinition = ATMEFct.getErrorPrefix(newRoute) .. "waypointSetName(numWaypoint, name)"
		
		if type(_numWaypoint) == "number" and type(_name) == "string" then
			local numWaypoint = math.floor(_numWaypoint)
			if numWaypoint > 0 and numWaypoint <= #localMembers.initWP then
				if localMembers.waypointsName[_name] == nil then
					if localMembers.initWP[numWaypoint].name ~= "" then
						localMembers.waypointsName[localMembers.initWP[numWaypoint].name] = nil
					end
					
					localMembers.initWP[numWaypoint].name = _name
					localMembers.waypointsName[_name] = numWaypoint
					return true
				else
					ATMEFct.setError("WAYPOINT_NAME_USED")
					ATMEFct.warningOutputCore(nil, fctDefinition, "waypoint name '" .. _name .. "' already used for this route", false)
					return false
				end
			end
		else
			if type(_numWaypoint) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "numWaypoint type must be number", true)
			elseif type(_name) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
			end
		end
	end
	
	function newRoute.getNbWaypoints(self)
		return #localMembers.initWP
	end
	
	function newRoute.getWaypoint(self, _waypointId)
		local fctDefinition = ATMEFct.getErrorPrefix(newRoute) .. "getWaypoint(waypointId)"
		
		if type(_waypointId) == "number" or type(_waypointId) == "string" then
			local numWaypoint
			if type(_waypointId) == "number" then
				numWaypoint = math.floor(_waypointId)
			else
				if localMembers.waypointsName[_waypointId] ~= nil then
					numWaypoint = localMembers.waypointsName[_waypointId]
				else
					ATMEFct.setError("BAD_WAYPOINT")
					ATMEFct.warningOutputCore(nil, fctDefinition, "WaypointId '" .. _waypointId .. "' doesn't exist in current route" , false)
					return
				end
			end
			
			if numWaypoint > 0 and numWaypoint <= #localMembers.initWP then
				if localMembers.initWP[numWaypoint].localMembersUnit ~= nil then
					if localMembers.initWP[numWaypoint].localMembersUnit.objectDCS:isExist() == false then
						return {
							waypointID = numWaypoint,
							name = localMembers.initWP[numWaypoint].name,
							x = localMembers.initWP[numWaypoint].x0,
							y = localMembers.initWP[numWaypoint].alt,
							z = localMembers.initWP[numWaypoint].y0,
						}
					else
						return {
							waypointID = numWaypoint,
							name = localMembers.initWP[numWaypoint].name,
							unitName = localMembers.initWP[numWaypoint].localMembersUnit.name,
						}
					end
				elseif localMembers.initWP[numWaypoint].position ~= nil then
					return {
							waypointID = numWaypoint,
							name = localMembers.initWP[numWaypoint].name,
							zone = localMembers.initWP[numWaypoint].position,
					}
				else
					return {
							waypointID = numWaypoint,
							name = localMembers.initWP[numWaypoint].name,
							x = localMembers.initWP[numWaypoint].x,
							y = localMembers.initWP[numWaypoint].alt,
							z = localMembers.initWP[numWaypoint].y,
					}
				end
			else
				ATMEFct.setError("BAD_WAYPOINT")
				ATMEFct.warningOutputCore(nil, fctDefinition, "WaypointId '" .. _waypointId .. "' out of range" , false)
				return
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "waypointId type must be number or string", true)
		end
	end
	
	function newRoute.waypointAddTask(self, _waypointId, _task)
		local fctDefinition = ATMEFct.getErrorPrefix(newRoute) .. "waypointAddTask(waypointId, task)"
		
		if (type(_waypointId) == "number" or type(_waypointId) == "string") and ATMEFct.isObjectClass(_task, ATME.C_Task) == true then
			local numWaypoint
			if type(_waypointId) == "number" then
				numWaypoint = math.floor(_waypointId)
			else
				if localMembers.waypointsName[_waypointId] ~= nil then
					numWaypoint = localMembers.waypointsName[_waypointId]
				else
					ATMEFct.setError("NO_WAYPOINT")
					ATMEFct.warningOutputCore(nil, fctDefinition, "No task added - waypointId '" .. _waypointId .. "' doesn't exist in current route" , false)
					return
				end
			end
			
			if numWaypoint > 0 and numWaypoint <= #localMembers.initWP then
				local ok, taskDatas = ATMEFct.C_Task_getTaskDatas(_task, localMembers.type)
				
				if ok == true then
					taskDatas.enabled = true
					
					if taskDatas.id == "Orbit" then
						taskDatas.auto = false
					else
						taskDatas.auto = true
					end
					taskDatas.number = 1
					
					table.insert(localMembers.initWP[numWaypoint].tasks, taskDatas)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, taskDatas, false)
				end
			else
				ATMEFct.setError("NO_WAYPOINT")
				ATMEFct.warningOutputCore(nil, fctDefinition, "No task added - waypointId " .. _waypointId .. " out of range" , false)
			end
		else
			if type(_waypointId) ~= "number" and type(_waypointId) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "waypointId type must be number or string", true)
			elseif ATMEFct.isObjectClass(_task, ATME.C_Task) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "task type must be ATME.C_Task", false)
			end
		end
	end
	
	function newRoute.waypointAddOption(self, _waypointId, _option, _value)
		local fctDefinition = ATMEFct.getErrorPrefix(newRoute) .. "waypointAddOption(waypointId, option, value)"
		
		if (type(_waypointId) == "number" or type(_waypointId) == "string") and type(_option) == "string" and _value ~= nil then
			
			local numWaypoint
			if type(_waypointId) == "number" then
				numWaypoint = math.floor(_waypointId)
			else
				if localMembers.waypointsName[_waypointId] ~= nil then
					numWaypoint = localMembers.waypointsName[_waypointId]
				else
					ATMEFct.setError("NO_WAYPOINT")
					ATMEFct.warningOutputCore(nil, fctDefinition, "No option added - waypointId '" .. _waypointId .. "' doesn't exist in current route" , false)
					return
				end
			end
			
			if numWaypoint > 0 and numWaypoint <= #localMembers.initWP then
				local err, option, optionValue = ATMEFct.C_setOption(localMembers.type, _option, _value) 

				if err == false then
					local taskDatas = {
						["id"] = "WrappedAction",		   
						["params"] = {
							["action"] = 
							{
								["id"] = "Option",
								["params"] = 
								{
									["value"] = optionValue,
									["name"] = option,
								},
							},
						},
					}
					table.insert(localMembers.initWP[numWaypoint].tasks, taskDatas)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, option, false)
				end
			else
				ATMEFct.setError("NO_WAYPOINT")
				ATMEFct.warningOutputCore(nil, fctDefinition, "No option added - waypointId " .. _waypointId .. " out of range" , false)
				return
			end
		else
			if type(_waypointId) ~= "number" and type(_waypointId) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "waypointId type must be number or string", true)
			elseif type(_option) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "option type must be string", false)
			elseif _value == nil then
				ATMEFct.errorOutputCore(nil, fctDefinition, "value type can't be nil", false)
			end
		end
	end
	
	function newRoute.getUserPrivateDatas(self, _moduleName)
		local fctDefinition = ATMEFct.getErrorPrefix(newRoute) .. "getUserPrivateDatas([moduleName])"
		local moduleName
		
		if _moduleName ~= nil then
			if type(_moduleName) == "string" then
				if localMembers.userPrivateDatas[_moduleName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
					return
				else
					moduleName = _moduleName
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
				return
			end
		else
			moduleName = ATMEDatas.currentModuleName
		end
		
		return localMembers.userPrivateDatas[moduleName]
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return newRoute
end

setmetatable(ATME.C_Route,ATMEDatas.C_Route_MT)

ATMEDatas.C_Route_MT.__call = function(t, _typeRoute, _rebuild)
	local fctDefinition = "Constructor ATME.C_Route(typeRoute, [rebuild])"

	if ATMEDatas.startMissionOk < 2 then
		ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
		return
	end
	
	if type(_typeRoute) == "string" and (_rebuild == nil or type(_rebuild) == "boolean") then
		if ATMEConst.categoryNameFilters[_typeRoute] ~= nil then
			local rebuild = _rebuild
			if rebuild == nil then
				rebuild = false
			end
			
			return ATMEFct.C_Route_create(nil, nil, ATMEConst.categoryNameFilters[_typeRoute], rebuild)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "type must be 'AIRPLANE', 'HELICOPTER', 'GROUND' or 'SHIP'", false)
		end
	else
		if type(_typeRoute) ~= "string" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "typeRoute type must be string", false)
		elseif _rebuild ~= nil and type(_rebuild) ~= "boolean" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "rebuild type must be nil or boolean", false)
		end
	end
end


ATMEDatas.C_Route_MT.mt.__metatable = "Private"
ATMEDatas.C_Route_MT.mt.__index = ATME.C_Route
ATMEDatas.C_Route_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Route_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Route : Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_GroupCloningContext : gère les données de clonage nécessaires à la duplication de groupes
--
-- Cette classe permet de gérer les chargements/déchargements d'unités d'un véhicule (voir load/unload)
-- de la classe C_Unit. Cette fonctionnalité n'est pas accessible aux autres modules
--
-- Cette classe offre également le spawn d'unités définie à partir d'un groupe existant.
-- Le spawn sera en échec si le groupe existe. 
-- Pour l'attribution des noms :
-- 		* Pour le nom des groupes : Si le nom n'existe pas dans la mission de base (donnée initiale) ni 
--		  dans les groupes existants (actifs), le nom est celui donnée à la construction du contexte de 
--		  de clonage. Sinon, le num sera postfixé avec " - Spawned #XXX". La valeur XXX est incrémentée 
--        automatiquement un nom de groupe particulier. Elle est comprise en 001 et 999. Si le groupe
--		  est détruit, cette valeur sera réutilisée par le prochain spawn.
--		* Pour les noms d'unités : Le nom d'unité dépend du groupe initial (défini dans la mission ou créé
--	      directement "from scratch"). En conséquence, le nom de l'unité existe dès le clonage, aussi tous
--        tous les noms d'unités clonées sont postfixée avec " - Spawned #XXX". La valeur XXX est incrémentée 
--        automatiquement un nom de l'unité particulière. Elle est comprise en 001 et 999. Si l'unité
--		  est détruite, cette valeur sera réutilisée par le prochain spawn.
--
-- Il est possible de construire un contexte de clonage à partir d'un groupe existants (actifs) ou défini initialement
-- dans la mission. Le nom du groupe à cloner sera alors une entrée pour le contexte de clonage.  Il est également
-- possible de construire un groupe "from scratch", l'entrée du contexte de clonage sera alors la table datas de 
-- DCS correctement renseignée.
--
-- Pour les waypoints : La modification d'une route initiale par la fonction setRoute de ATME.C_Group n'a aucun incidence
-- sur la route du nouveau groupe cloné. Le groupe cloné peut avoir :
--			- Par défaut : les WP définis initialement pour le groupe source du clonage.
--			- Une route spécifique définie par une instance de ATME.C_Route
----------------------------------------------------------------------------------------------------------

ATMEFct.C_GroupCloningContext_createNamed = function(_namedTable, _name, _origin, _newGroupName)
	if _namedTable[_name] ~= nil then
		ATMEFct.setError("ALREADY_CREATED")
		return true, "named cloning context already exists"
	else
		local spawnDatas
		if type(_origin) == "table" then
			if ATMEFct.isObjectClass(_origin, ATME.C_Group) == true then
				local localMembersGroup = ATMEDatas.localMembers[_origin:getIndex()]
				spawnDatas = ATMEFct.C_GroupCloningContext_copyTable(localMembersGroup.spawnDatas)
			elseif ATMEFct.C_GroupCloningContext_controlTable(_origin, _newGroupName) == false then
				return true, "Bad cloning table structure"
			else
				spawnDatas = ATMEFct.C_GroupCloningContext_copyTable(_origin)
				spawnDatas.type = ATMEConst.categoryNameFilters[spawnDatas.type]
				spawnDatas.country = country.id[spawnDatas.country]
				spawnDatas.category = ATMEConst.groupCategories[spawnDatas.type]
				spawnDatas.typeName = nil
				spawnDatas.countryName = nil
			end
		else 
			if ATMEDatas.allGroupsInMission[_origin] ~= nil then
				spawnDatas = ATMEFct.C_GroupCloningContext_copyTable(ATMEDatas.allGroupsInMission[_origin].spawnDatas)
			else
				return true, "Group " .. _origin .. " not defined in mission editor"
			end
		end
		
		if spawnDatas.type == "plane" or spawnDatas.type == "helicopter" then
			for _, _unitDatas in ipairs(spawnDatas.datas.units) do
				if _unitDatas.skill == "Player" or _unitDatas.skill == "Client" then
					_unitDatas.skill = "Random"
				end
			end
		end

		if _newGroupName ~= nil then
			local d,f = string.find(_newGroupName, "|", 1)
			if d ~= nil then
				return true, "Name of group " .. _newGroupName .. " can't contain '|' character"
			end
			
			spawnDatas.datas.name = _newGroupName
		end
		
		return false, ATMEFct.C_GroupCloningContext_create(_name, _namedTable, spawnDatas)
	end
end

ATMEFct.C_GroupCloningContext_controlTable = function(_groupTable, _newGroupName)
	if type(_groupTable.country) == "string" and country.id[_groupTable.country] ~= nil and
		type(_groupTable.type) == "string" and ATMEConst.categoryNameFilters[_groupTable.type] ~= nil and
		type(_groupTable.datas) == "table" and 
		(type(_groupTable.datas.name) == "string" or type(_newGroupName) == "string") and
		type(_groupTable.datas.task) == "string" then

		if type(_groupTable.datas.units) ~= "table" then
			_groupTable.datas.units = {}
		end
		
		if type(_groupTable.datas.tasks) ~= "table" then
			_groupTable.datas.tasks = {}
		end

		if _groupTable.type == "AIRPLANE" or _groupTable.type == "HELICOPTER" then
			local task = ATMEConst.mainTask[_groupTable.datas.task]
			if task == nil then
				task = "Nothing"
			end
			_groupTable.datas.task = task
		else
			_groupTable.datas.task = ""
		end

		for _, _unit in ipairs(_groupTable.datas.units) do
			if type(_unit.name) ~= "string" or type(_unit.type) ~= "string" then
				return false
			end
			
			local skill = "RANDOM"
			
			if type(_unit.skill) == "string" then
				skill = string.upper(_unit.skill)
			end			
			
			if ATMEConst.skill[skill] ~= nil then
				return false
			end
			
			_unit.skill = ATMEConst.skill[skill]
			
			if _groupTable.type == "AIRPLANE" or _groupTable.type == "HELICOPTER" then
				if type(_unit.livery_id) ~= "string" then
					return false
				end
				if type(_unit.payload) ~= "table" or type(_unit.alt) ~= "number" or type(_unit.speed) ~= "string" then
					return false
				end
				
				if type(_unit.alt_type) ~= "string" or _unit.alt_type ~= "BARO" or _unit.alt_type ~= "RADIO" then
					return false
				end
			else
				if type(_unit.x) ~= "number" or type(_unit.y) ~= "number" then
					return false
				end
			end
		end

		return true
	else
		return false
	end
end

ATMEFct.C_GroupCloningContext_addUnit = function(_spawnDatas, _unit, _newName, _initialType)
	local newName = _newName
	
	if newName == nil then
		newName = _unit.name
		
		if type(newName) ~= "string" then
			return true, "Group " .. _spawnDatas.datas.name .. " - Bad new unit name"
		end
	end
	
	if type(_unit.type) ~= "string" then
		return true, "Group " .. _spawnDatas.datas.name .. " - Bad new unit type"
	end
		
	if _unit.skill == "Client" or _unit.skill == "Player" then
		_unit.skill = "Random"
	end
	
	local payload = _unit.payload
	
	if _initialType == "plane" or _initialType == "helicopter" then
		if type(payload) ~= "table" then
			payload = {
				["pylons"] = {},
			}
		end
		
		if type(_unit.alt) ~= "number" then
			return true, "Group " .. _spawnDatas.datas.name .. " - Bad new unit altitude"
		end
		
		if type(_unit.alt_type) ~= "string" and _unit.alt_type ~= "BARO" and _unit.alt_type ~= "RADIO" then
			return true, "Group " .. _spawnDatas.datas.name .. " - Bad new unit altitude type (BARO/RADIO)"
		end
	end
	
	_spawnDatas.unitsByName[_newName] = {
		name = newName,
		type = _unit.type,
		livery_id = _unit.livery_id,
		speed = _unit.speed,
		psi = _unit.psi,
		payload = payload,
		skill = _unit.skill,
		x = _unit.x,
		y = _unit.y,
		playerCanDrive = _unit.playerCanDrive,
		heading = _unit.heading,
		alt = _unit.alt,
		alt_type = _unit.alt_type,
		callsign = _unit.callsign,
		onboard_num = _unit.onboard_num,
		parking = _unit.parking,
		parking_id = _unit.parking_id,
	}

	_spawnDatas.datas.units[#_spawnDatas.datas.units + 1] = _spawnDatas.unitsByName[_newName]
	
	return false
end

ATMEFct.copyAllWaypoints = function(_waypoints)
	local newWaypoints = {}

	for _, _waypoint in ipairs(_waypoints) do
		table.insert(newWaypoints, ATMEFct.copyTable(_waypoint))
	end

	return newWaypoints
end

ATMEFct.C_GroupCloningContext_setWaypointComboTask = function(_waypoint)
	if _waypoint.task == nil then
		_waypoint.task = {
			["id"] = "ComboTask",
			["params"] = {
				["tasks"] = {
				},
			},
		}
	end
end

ATMETimers.C_GroupCloningContext_missionTimer = function(_datas)
	if _datas.localMembers.objectDCS:isExist() == true and #_datas.localMembers.objectDCS:getUnits() ~= 0 and ATMEDatas.groups[_datas.localMembers.objectDCS:getName()] ~= nil then
		if _datas.localMembers.initOk == true then
			local unit = ATMEFct.getFirstUnit(_datas.group)
			local localMembersUnit = ATMEDatas.localMembers[unit:getIndex()]
			
			if _datas.armed == true then
				_datas.localMembers.objectDCS:getController():setTask(_datas.task)
				
				if _datas.routeSet == true then
					timer.scheduleFunction(ATMETimers.C_Group_newRouteSet, _datas, timer.getTime() + 1)
				end
				
				return
			else
				_datas.armed = true
			end
		end
		
		return timer.getTime() + 2
	end
end

ATMEFct.C_GroupCloningContext_setMission = function(_group, _waypoints, _routeSet)
	local localMembers = ATMEDatas.localMembers[_group:getIndex()]
	local timerOffset = 0.1
	
	local datas = {
		localMembers = localMembers,
		group = _group,
		armed = false,
		routeSet = _routeSet,
		task = { 
			id = 'Mission', 
			params = {
				airborne = ATMEConst.canFly[localMembers.category],
				route = {
					points = _waypoints,
				},
			},
		},
	}
	
	if _routeSet == true then
		localMembers.objectDCS:getController():resetTask()
		timerOffset = 3
	end
	
	timer.scheduleFunction(ATMETimers.C_GroupCloningContext_missionTimer, datas, timer.getTime() + timerOffset)
end

ATMEFct.C_GroupCloningContext_setWaypoints = function(_group, _waypoints, _isOnRoad)
	local firstUnit = ATMEFct.getFirstUnit(_group)
	
	if firstUnit ~= nil then
		-- On ajoute la position courante à la nouvelle route

		local localMembers = ATMEDatas.localMembers[firstUnit:getIndex()]
		
		local newWP = ATMEFct.copyTable(_waypoints[1])
		local pos = localMembers.objectDCS:getPoint()
		newWP.x = pos.x
		newWP.y = pos.z
		
		if _waypoints[1].action == "On Road" and _isOnRoad == false then
			newWP.action = "Off Road"
		end
		
		table.insert(_waypoints, 1, newWP)
		
		ATMEFct.C_GroupCloningContext_setMission(_group, _waypoints, false)
	end
end

ATMEFct.C_GroupCloningContext_startRouteAt = function(_waypoints, _reference, _radius, _groupName, _delta)
	-- _radius défini la zone d'exclusion par rapport à la référence de type objet. Non pris en compte sinon.
	-- Dans ce cas, la route débutera au waypoint le plus proche de la référence, mais en excluant un cercle de rayon 
	-- _radius autour de la référence.
	-- Pour les autres cas, l'offset ou le nom du waypoint est recherché. S'il n'est pas trouvé, la route reste la 
	-- route par défaut

	if type(_reference) == "number" then -- relatif au numéro de waypoint (début à 0 pour rappel)
		-- Si le numéro est plus grand  ou egal au dernier waypoint, il s'agira du dernier waypoint.
		-- radius contient alors waypointId (si débarquement) ou 0
		
		local nb = _reference - _radius
		if nb < #_waypoints then
			while nb > 0 do 
				table.remove(_waypoints, 1)
				nb = nb - 1
			end

			return _reference - _radius 
		end
		
		return #_waypoints - _radius
	else -- Relatif à un objet
		local point = ATMEFct.getPointFromReference(_reference)

		local index = #_waypoints
		local v = ATMEFct.vectors["3D"].createFromReference(nil, nil, point, _waypoints[index])
		local minDistance = v:getHModule()
		
		local found = index

		while index > 1 do
			index = index - 1
			
			local v = ATMEFct.vectors["3D"].createFromReference(nil, nil, point, _waypoints[index])
			
			if v:getHModule() >= _radius then
				if minDistance > v:getHModule() then
					minDistance = v:getHModule()
					found = index
				end
			elseif _groupName ~= nil then
				ATMEFct.outputCore(nil, "Group '" .. _groupName .. "' disembark : Exclude WP #" .. index + _delta - 2, 0)
			end
		end
		
		for _i = 1, found - 1 do
			table.remove(_waypoints, 1)
		end

		return found - 1
	end
end
								

-- Fonction spawn des objets DCS avec gestion ATME
ATMEFct.C_GroupCloningContext_spawn = function(_clonigContext, _localMembers, _route)
	local err, waypoints, waypointsName = ATMEFct.C_Route_createWaypoints(_route)

	if err == false then
		local idSpawnedList = {
			units = {},
		}
		
		local newGroupName

		local newId
		if ATMEDatas.spawnedGroups[_localMembers.spawnDatas.datas.name] == nil then
			ATMEDatas.spawnedGroups[_localMembers.spawnDatas.datas.name] = {
				max = 1,
				freeIds = {},
			}
			
			newId = ATMEDatas.spawnedGroups[_localMembers.spawnDatas.datas.name].max
		else
			if #ATMEDatas.spawnedGroups[_localMembers.spawnDatas.datas.name].freeIds == 0 then
				ATMEDatas.spawnedGroups[_localMembers.spawnDatas.datas.name].max = ATMEDatas.spawnedGroups[_localMembers.spawnDatas.datas.name].max + 1
				newId = ATMEDatas.spawnedGroups[_localMembers.spawnDatas.datas.name].max
			else
				newId = table.remove(ATMEDatas.spawnedGroups[_localMembers.spawnDatas.datas.name].freeIds)
			end
		end
	
		-- Le '-' devient un séparateur pour la composition des noms suite à modif DCS 2.5.6 des copies ('#%03d' remplacé par '-%d')
		-- Ainsi "Naval-1" aura bien 2 tags "Naval" et "1"
		newGroupName = string.format("%s-%d", _localMembers.spawnDatas.datas.name, newId)
		
		if ATMEDatas.reservedGroupNames[newGroupName] == nil then
			idSpawnedList.groupSpawnedId = newId
			idSpawnedList.freeIds = ATMEDatas.spawnedGroups[_localMembers.spawnDatas.datas.name].freeIds

			local spawnDatas = {
				idSpawnedList = idSpawnedList,
				country = _localMembers.spawnDatas.country,
				category = _localMembers.spawnDatas.category,
				type = _localMembers.spawnDatas.type,
				datas = {
					name = newGroupName,
					hidden = _localMembers.spawnDatas.datas.hidden,
					start_time = 0,
					task = _localMembers.spawnDatas.datas.task,
					visible = false,
					taskSelected = true,
					lateActivation = false,
					tasks = {},
					units = {},
					route = {},
				},
				unitsByName = {},
				spawnCallback = _localMembers.spawnCallback,
				cloningContext = _clonigContext,
			}
			
			local position0 = {
				x = 0,
				y = 0,
			}
			
			if _localMembers.spawnDatas.datas.units[1].x ~= nil then
				position0.x = _localMembers.spawnDatas.datas.units[1].x
			end
			
			if _localMembers.spawnDatas.datas.units[1].y ~= nil then
				position0.y = _localMembers.spawnDatas.datas.units[1].y
			end
		
			for _i = 1, #_localMembers.spawnDatas.datas.units do
				local unitName = _localMembers.spawnDatas.datas.units[_i].name

				local newId
				if ATMEDatas.spawnedUnits[unitName] == nil then
					ATMEDatas.spawnedUnits[unitName] = {
						max = 1,
						freeIds = {},
					}
					newId = ATMEDatas.spawnedUnits[unitName].max
				else
					if #ATMEDatas.spawnedUnits[unitName].freeIds == 0 then
						ATMEDatas.spawnedUnits[unitName].max = ATMEDatas.spawnedUnits[unitName].max + 1
						newId = ATMEDatas.spawnedUnits[unitName].max
					else
						newId = table.remove(ATMEDatas.spawnedUnits[unitName].freeIds)
					end
				end
				
				-- Le '-' devient un séparateur pour la composition des noms suite à modif DCS des copies (' #%03d' remplacé par '-%d')
				-- Ainsi "Naval-1" aura bien 2 tags "Naval" et "1"
				unitName = string.format("%s-%d", unitName, newId)
				
				if ATMEDatas.reservedUnitNames[unitName] ~= nil then
					-- Warning doublon de nom
					ATMEFct.setError("UNIT_NAME_DUPLICATE")
					return "Unit " .. unitName .. " already exist - No spawn"
				end
				
				idSpawnedList.units[unitName] = {
					id = newId,
					freeIds = ATMEDatas.spawnedUnits[_localMembers.spawnDatas.datas.units[_i].name].freeIds,
				}
				
				local heading = _localMembers.spawnDatas.datas.units[_i].heading
				local x = 0
				local y = 0
				
				if _localMembers.spawnDatas.datas.units[_i].x ~= nil then
					x = _localMembers.spawnDatas.datas.units[_i].x
				end
				
				if _localMembers.spawnDatas.datas.units[_i].y ~= nil then
					y = _localMembers.spawnDatas.datas.units[_i].y
				end

				x = waypoints[1].x + x - position0.x
				y = waypoints[1].y + y - position0.y

				if #waypoints > 1 then
					local dx = waypoints[2].x - x
					local dy = waypoints[2].y - y
					
					if dx == 0 and dy == 0 then
						if heading == nil then
							heading = math.rad(math.random(-179, 180))
						end
					else
						heading = math.atan2(dy, dx)
					end
				else
					if heading == nil then
						heading = math.rad(math.random(-179, 180))
					end
				end

				spawnDatas.datas.units[_i] = {
					type = _localMembers.spawnDatas.datas.units[_i].type,
					name = unitName,
					skill = _localMembers.spawnDatas.datas.units[_i].skill,
					alt = waypoints[1].alt,
					x = x,
					y = y,
					alt_type = _localMembers.spawnDatas.datas.units[_i].alt_type,
					playerCanDrive = _localMembers.spawnDatas.datas.units[_i].playerCanDrive,
					heading = heading,
					payload = _localMembers.spawnDatas.datas.units[_i].payload,
					speed = _localMembers.spawnDatas.datas.units[_i].speed,
					callsign = _localMembers.spawnDatas.datas.units[_i].callsign,
					onboard_num = _localMembers.spawnDatas.datas.units[_i].onboard_num,
					livery_id = _localMembers.spawnDatas.datas.units[_i].livery_id,
				}
				
				spawnDatas.unitsByName[unitName] = spawnDatas.datas.units[_i]
			end

			if initialType == "plane" or initialType == "helicopter" then
				if _localMembers.callsign ~= nil then
					if initialCoalition == "BLUE" then
						for _i, _unit in ipairs(spawnDatas.datas.units) do
							_unit.callsign = {
								[1] = _localMembers.callsign[1],
								[2] = _localMembers.callsign[2],
								[3] = _i,
								name = _localMembers.callsign.name .. _i,
							}
						end
						
						_localMembers.callsign[2] = _localMembers.callsign[2] + 1
						if _localMembers.callsign[2] > 9 then
							_localMembers.callsign[2] = 1
						end
					else -- RED
						for _i, _unit in ipairs(spawnDatas.datas.units) do
							_unit.callsign = string.format("%03d", _localMembers.callsign + _i - 1)
						end
						
						_localMembers.callsign = _localMembers.callsign + 10
					end
				end
			end

			spawnDatas.datas.route.points = waypoints
			spawnDatas.waypointsName = waypointsName
			
			-- Mémorisation des GroupDatas pour un éventuel spawn multiple
			if ATMEDatas.spawnGroupDatasRepeat[_localMembers.spawnDatas.datas.name] ~= nil then
				ATMEDatas.spawnGroupDatasRepeat[_localMembers.spawnDatas.datas.name].arg = _route
				local delay
				
				if ATMEDatas.spawnGroupDatasRepeat[_localMembers.spawnDatas.datas.name].repeatRandom == true then
					delay = math.random(ATMEDatas.spawnGroupDatasRepeat[_localMembers.spawnDatas.datas.name].repeatPeriodMin, 
										ATMEDatas.spawnGroupDatasRepeat[_localMembers.spawnDatas.datas.name].repeatPeriodMax)
					
				else
					delay = ATMEDatas.spawnGroupDatasRepeat[_localMembers.spawnDatas.datas.name].repeatPeriodMin
				end

				timer.scheduleFunction(ATMETimers.C_GroupCloningContext_respawnMultiple, ATMEDatas.spawnGroupDatasRepeat[_localMembers.spawnDatas.datas.name], timer.getTime() + delay)
				ATMEDatas.spawnGroupDatasRepeat[_localMembers.spawnDatas.datas.name] = nil
			end

			ATMEFct.C_GroupCloningContext_finalizeSpawn(spawnDatas, true)
		else
			-- Warning doublon de nom
			ATMEFct.setError("GROUP_NAME_DUPLICATE")
			return "Group " .. newGroupName .. " already exist - No spawn"
		end
	else
		ATMEFct.setError("BAD_WAYPOINTS")
		return "bad waypoints or no waypoint defined"
	end
end

ATMEFct.C_GroupCloningContext_finalizeSpawn = function(_spawnDatas, _isSpawned)
	-- ATTENTION deux cas :
	--	*	Pour les groupes d'unités au sol ou les bateaux, les unités sont créées (event S_EVENT_BIRTH) 
	--		pendant le addgroup mais parfois leur groupe associé ne l'est pas encore ... 
	--		Donc on mémorise lors de l'événement DCS et on traite ci après
	--	*	Pour les groupes aircraft, l'event S_EVENT_BIRTH est appelé bien après. On mémorise ici pour un
	--		traitement lors de l'événement DCS.

	ATMEDatas.spawnedGroupsWaypoints[_spawnDatas.datas.name] = {
		waypoints = _spawnDatas.datas.route.points,
		initialWaypoints = _spawnDatas.datas.route.points,
		initialWaypointsName = _spawnDatas.waypointsName,
		offsetBoard = false,
	}

	coalition.addGroup(_spawnDatas.country, _spawnDatas.category, _spawnDatas.datas)

	ATMEDatas.reservedGroupNames[_spawnDatas.datas.name] = true
	
	if #ATMEDatas.unitsReadyToCreate > 0 then
		while #ATMEDatas.unitsReadyToCreate > 0 do
			local newUnit = table.remove(ATMEDatas.unitsReadyToCreate)
			
			-- L'unité à créer est un bateau avec FARP ou porte-avion
			
			if newUnit:hasAttribute("AircraftCarrier") == true then
				ATMEFct.C_Airbase_create_basic("SHIP", false, newUnit)
			elseif newUnit:hasAttribute("HelicopterCarrier") == true then
				ATMEFct.C_Airbase_create_basic("SHIP", true, newUnit)
			end
			
			local datas = {
				uncontrolled = false,
				visible = true,
				spawnDatas = _spawnDatas,
			}
			
			ATMEDatas.reservedUnitNames[newUnit:getName()] = true
			ATMEFct.C_Unit_create(newUnit, datas, true, false)
		end
		
		ATMEDatas.newCreatedGroups[_spawnDatas.datas.name].isSpawned = _isSpawned
		ATMEDatas.newCreatedGroups[_spawnDatas.datas.name].spawnCallback = _spawnDatas.spawnCallback
		ATMEDatas.newCreatedGroups[_spawnDatas.datas.name].cloningContext = _spawnDatas.cloningContext
	else
		local itemDatas = {
			name = _spawnDatas.datas.name,
			units = {},
			nbUnits = 0,
			uncontrolled = false,
			visible = true,
			spawnDatas = _spawnDatas,
		}
		
		for _i, _unit in ipairs(_spawnDatas.datas.units) do
			itemDatas.units[_unit.name] = true
			itemDatas.nbUnits = itemDatas.nbUnits + 1
			ATMEDatas.reservedUnitNames[_unit.name] = true
			ATMEDatas.unitsReadyToBirth[_unit.name] = true
		end
		
		ATMEDatas.groupsReadyToBirth[_spawnDatas.datas.name] = {
			item = itemDatas,
			spawnCallback = _spawnDatas.spawnCallback,
			cloningContext = _spawnDatas.cloningContext,
		}
	end
end

-- Spawn sur débarquement des troupes embarquées dans un transport
ATMEFct.C_GroupCloningContext_spawnFromTransportUnit = function(_transportUnit, _currentOnBoardGroupDatas)
	local spawnDatas = _currentOnBoardGroupDatas.localMembers.spawnDatas
	local unitsOnBoard = _currentOnBoardGroupDatas.unitsOnBoard
	local initialWaypoints = _currentOnBoardGroupDatas.localMembers.spawnDatas.datas.route.points
	local initialWaypointsName = _currentOnBoardGroupDatas.localMembers.spawnDatas.waypointsName
	local localMembersTransport = ATMEDatas.localMembers[_transportUnit:getIndex()]
	
	local nb = localMembersTransport.nbUnitsOnBoard

	-- Min une unité par 3° soit 118 unités au maximum
	-- Calcul : pour 180° -> 180/3 - 1 unités, soit pour 360°, 2 x (180/3 - 1) = 118
	-- Max units par groupe terrestre = 99 < 118 donc pas de soucis actuellement
	-- sachant que la charge d'un hélico est bien inférieure à 99 unités.

	local point = localMembersTransport.objectDCS:getPoint()
	local vector = ATMEFct.vectors["3D"].createFromTable(nil, nil, localMembersTransport.objectDCS:getPosition().x)
	local azimuth = vector:getAzimuth()

	-- Vérification d'absence d'eau côté droit/gauche 
	
	local sides = {}
	local startAngle = 0
	local nbPointsWater = {0,0}
	local index = 1
	local distance = ATMEConst.personnelCarrierAttributes[localMembersTransport.typeName].minDistance
	local deltaRad = math.pi / 4
	for _side = 1,-1,-2 do
		for _angle = deltaRad / 4 * _side, 3 * deltaRad * _side, deltaRad * _side do
			local result = ATMEFct.coordLocalToDCS({x = distance * math.cos(_angle), y = distance * math.sin(_angle)}, point, azimuth)
			if ATMEFct.isSurfaceWater(result) == true then
				nbPointsWater[index] = nbPointsWater[index] + 1
			end
		end
		
		if nbPointsWater[index] < 2 then
			table.insert(sides, startAngle)
		end
		startAngle = startAngle + 180
		index = index + 1
	end
	
	if #sides > 0 then
		if #sides == 2 then
			if nbPointsWater[1] > nbPointsWater[2] then -- On permute les côtés pour privilégier le côté sans eau (1 unité de plus si impair)
				sides[1] = 180
				sides[2] = 0
			end
		end
		-- Ajustements en fonction du nombre d'unités et du nombre
		-- de côtés possibles.
		
		-- Si moins de 2 unités par côté pas de symétrie
		if nb/#sides <= 2 and #sides > 1 then
			table.remove(sides)
		end
		
		-- Si symétrie (2 côtés) et nombre impair d'unités -> On ajuste pour rendre pairs
		if nb/#sides ~= math.floor(nb/#sides) then
			-- Si le nombre d'unités est impair quand 2 côtés ok, on ajoute 1
			nb = nb + 1
		end

		-- Calcul de l'angle entre chaque unité
		
		local deltaAngle = math.floor(180 / (nb + 1))
		local unitAngle = startAngle
		local id = 1
		local currentSide = 1

		while id <= #_currentOnBoardGroupDatas.localMembers.spawnDatas.datas.units do
			if unitsOnBoard[_currentOnBoardGroupDatas.localMembers.spawnDatas.datas.units[id].name] ~= nil then
				local unit = _currentOnBoardGroupDatas.localMembers.spawnDatas.datas.units[id]
				unitAngle = unitAngle + deltaAngle


				local unitAngleRad = math.rad(unitAngle + sides[currentSide])
				currentSide = currentSide + 1
				if currentSide > #sides then
					currentSide = 1
				end
				
				local result = ATMEFct.coordLocalToDCS(
											{
												x = distance * math.cos(unitAngleRad), 
												y = distance  * math.sin(unitAngleRad),
											}, 
											point, 
											azimuth)
				
				-- Si l'unité est bien sur terre, alors ok
				if ATMEFct.isSurfaceWater(result) == false then
				
					unit.x = result.x
					unit.y = result.z

					unit.heading = unitAngleRad + math.rad(azimuth)
					id = id + 1
				else -- si elle est dans l'eau alors elle meure (elle ne sera pas réactivée).
					table.remove(_currentOnBoardGroupDatas.localMembers.spawnDatas.datas.units, id)
					_currentOnBoardGroupDatas.nbDeadUnits = _currentOnBoardGroupDatas.nbDeadUnits + 1
				end
			else
				table.remove(_currentOnBoardGroupDatas.localMembers.spawnDatas.datas.units, id)
			end
		end

		-- Une unité est forcément dans le transport si on débarque donc pas de tests, le spawn aura lieu
		local startWP = ATMEFct.copyTable(_currentOnBoardGroupDatas.localMembers.waypoints[_currentOnBoardGroupDatas.localMembers.waypointId])
		local newWaypoints = {}
		
		for i = _currentOnBoardGroupDatas.localMembers.waypointId, #_currentOnBoardGroupDatas.localMembers.waypoints do
			table.insert(newWaypoints, _currentOnBoardGroupDatas.localMembers.waypoints[i])
		end
		
		_currentOnBoardGroupDatas.waypoints = newWaypoints			

		local indexWP
		if _currentOnBoardGroupDatas.localMembers.nextWaypoint == nil then
			-- Dans ce cas, les unités débarquées suivent une route qui débute
			-- au waypoint le plus proche du transport (exclusion de 100m)
			-- sans prendre en compte les waypoints déjà passés.

			indexWP = ATMEFct.C_GroupCloningContext_startRouteAt(newWaypoints, _transportUnit, 100, 
										_currentOnBoardGroupDatas.localMembers.spawnDatas.datas.name, _currentOnBoardGroupDatas.localMembers.deltaWaypointId + _currentOnBoardGroupDatas.localMembers.waypointId)
		else
			-- Dans ce cas le numéro du waypoint est connu.
			-- Dans ce cas, les unités débarquées suivent une route qui débute au waypoint désigné.

			indexWP = ATMEFct.C_GroupCloningContext_startRouteAt(newWaypoints, _currentOnBoardGroupDatas.localMembers.nextWaypoint, 
																	_currentOnBoardGroupDatas.localMembers.deltaWaypointId + _currentOnBoardGroupDatas.localMembers.waypointId)
																	
		end

		table.insert(newWaypoints, 1, startWP)
		
		-- Mise à jour de la position du premier WP avec
		-- les nouvelles coordonnées de la première unité du groupe qui débarque.
		-- Si le groupe est en maintenir position (task = Hold), cette tache est supprimée
		
		newWaypoints[1].x = spawnDatas.datas.units[1].x
		newWaypoints[1].y = spawnDatas.datas.units[1].y
		newWaypoints[1].action = "Off Road"
		
		-- Pour le spawn, création de la bonne route
		spawnDatas.datas.route = {
			points = newWaypoints,
		}

		local indexTask = 0
		for _i, _task in ipairs(spawnDatas.datas.route.points[1].task.params.tasks) do
			if _task.id == "Hold" then -- Maintenir position activé
				indexTask = _i
			end
		end

		if indexTask ~= 0 then
			table.remove(spawnDatas.datas.route.points[1].task.params.tasks, indexTask)
		end
		
		-- Mémorisation du groupe spawné pour traitement à la création
		-- du groupe ATME.C_Group

		ATMEDatas.spawnedGroupsWaypoints[spawnDatas.datas.name] = {
			waypoints = newWaypoints,
			initialWaypoints = initialWaypoints,
			initialWaypointsName = initialWaypointsName,
			offBoard = true,
		}

		coalition.addGroup(spawnDatas.country, spawnDatas.category, spawnDatas.datas)
		
		
		-- Les unités sont créées (event S_EVENT_BIRTH) pendant le addgroup mais parfois
		-- leur groupe associé ne l'est pas encore ... Dans ce cas on mémorise et on traite ci après.

		while #ATMEDatas.unitsReadyToCreate > 0 do
			local newUnit = table.remove(ATMEDatas.unitsReadyToCreate)
			
			local datas = {
				uncontrolled = false,
				visible = true,
				spawnDatas = spawnDatas,
			}

			ATMEFct.C_Unit_create(newUnit, datas, false, false) -- 1er False car pas de mémo sur débarquement
		end
		
		return false, _currentOnBoardGroupDatas.localMembers.waypointId + indexWP - 2	-- offsetWaypoint ajusté cause 1 WP de débarquement -> 1er WP ok à 2.
	else
		return true
	end
end

ATMEFct.C_GroupCloningContext_copyTable = function(_spawnDatas)
	local function _copy(datas)
		if type(datas) ~= "table" then
			return datas
		end
		
		local newTable = {}
		for _index, _value in pairs(datas) do
			newTable[_index] = _copy(_value)
		end
			
		return newTable
	end

	local spawnDatas = {
		country = _spawnDatas.country,
		category = _spawnDatas.category,
		type = _spawnDatas.type,
		unitsByName = {},
		datas = {
			name = _spawnDatas.datas.name,
			groupId = _spawnDatas.datas.groupId,
			hidden = _spawnDatas.datas.hidden,
			start_time = _spawnDatas.datas.start_time,
			task = _spawnDatas.datas.task,
			visible = _spawnDatas.datas.visible,
			taskSelected = _spawnDatas.datas.taskSelected,
			lateActivation = false,
			units = {},
		},
	}
	
	for _id, _unit in ipairs(_spawnDatas.datas.units) do
		spawnDatas.datas.units[_id] = {
			type = _unit.type,
			name = _unit.name,
			skill = _unit.skill,
			x = _unit.x,
			y = _unit.y,
			alt = _unit.alt,
			alt_type = _unit.alt_type,
			playerCanDrive = _unit.playerCanDrive,
			heading = _unit.heading,
			payload = _unit.payload,
			speed = _unit.speed,
			callsign = _unit.callsign,
			onboard_num = _unit.onboard_num,
			psi = _unit.psi,
			livery_id = _unit.livery_id,
		}
		
		spawnDatas.unitsByName[_unit.name] = spawnDatas.datas.units[_id]
	end

	spawnDatas.datas.tasks = _copy(_spawnDatas.datas.tasks)
	spawnDatas.datas.route = _copy(_spawnDatas.datas.route)
	
	return spawnDatas
end

ATMETimers.C_GroupCloningContext_respawnMultiple = function(_datas)
	if ATMEDatas.exitATME == false then
		if _datas.nbSpawn < _datas.maxQuantity then
			local delay

			if _datas.repeatRandom == true then
				delay = math.random(_datas.repeatPeriodMin, _datas.repeatPeriodMax)
			else
				delay = _datas.repeatPeriodMin
			end

			_datas.nbSpawn = _datas.nbSpawn + 1
			
			local route = _datas.arg
			if route == nil then
				route = ATMEFct.C_Route_create(nil, nil, _datas.initialType)
				ATMEFct.addWPFrom(route, _datas.localMembers.spawnDatas.type, _datas.localMembers.spawnDatas.datas.route.points, true, 0, #_datas.localMembers.spawnDatas.datas.route.points)
			end
			
			local text = ATMEFct.C_GroupCloningContext_spawn(_datas.cloningContext, _datas.localMembers, route)
			if ATMEDatas.error ~= nil then
				ATMEFct.warningOutputCore(nil, "Multiple respawn", text, false)
			end
			
			return timer.getTime() + delay
		end
	end
end

-- Fonction de création des données de base nécessaire au spawn simple ou multiple.
--   Le groupe _groupeName s'il existe peut lui même être un groupe spawné.
--   Dans ce cas, il faudra utiliser les noms originaux du groupe et unités associés 
--   tels que définis dans la mission. D'où la variable originalMissionNames
--   Si ce groupe n'existe pas alors on se base sur les données de missions.

ATMEFct.C_GroupCloningContext_create = function(_name, _groupCloningContexts, _spawnDatas)
	local initialType = _spawnDatas.type
	local initialCoalition = ATMEConst.coalitionNames[coalition.getCountryCoalition(_spawnDatas.country)]

	local newGroupClonigContext = {}
	
	local name
	local tagsName
	local transportUnitName
	
	if _name ~= nil and _groupCloningContexts ~= nil then
		name = _name
		_groupCloningContexts[name] = newGroupClonigContext
		tagsName = ATMEFct.decompString(_name)
	end

	setmetatable(newGroupClonigContext, ATMEDatas.C_GroupCloningContext_MT.mt)
	
	local index = ATMEDatas.C_GroupCloningContext_MT:add(_spawnDatas)
	local localMembers = ATMEDatas.localMembers[index]
	
	function newGroupClonigContext.getIndex(self)
		return index
	end
	
	function newGroupClonigContext.setMainTask(self, _mainTask)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroupClonigContext) .. "setMainTask(mainTask)"
		
		if type(_mainTask) == "string" then
			if initialType == "plane" or initialType == "helicopter" then
				if ATMEConst.mainTask[_mainTask] ~= nil then
					localMembers.spawnDatas.datas.task = ATMEConst.mainTask[_mainTask]
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "mainTask bad value", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Model group '" .. name .. "' must be an aircraft group", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "mainTask type must be string", true)
		end
	end
	
	function newGroupClonigContext.setPosition(self, _unitName, _point)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroupClonigContext) .. "setPosition(unitName, point)"
		if type(_unitName) == "string" and ATME.isPoint(_point) ~= POINT_BAD then
			if initialType == "ship" or initialType == "vehicle" then
				if localMembers.spawnDatas.unitsByName[_unitName] ~= nil then
					local x, y, alt
					if _point.z ~= nil then
						x = _point.x
						y = _point.z
						alt = _point.y
					else
						x = _point.x
						y = _point.y
					end
					
					localMembers.spawnDatas.unitsByName[_unitName].x = x
					localMembers.spawnDatas.unitsByName[_unitName].y = y
					localMembers.spawnDatas.unitsByName[_unitName].alt = alt
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "Unit " .. _unitName .. " unknown", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Model group '" .. name .. "' must be a naval or ground group", false)
			end
		else
			if type(_unitName) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "unitName type must be string", true)
			elseif ATME.isPoint(_point) == POINT_BAD then
				ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be table (x,y)/(x,y,z)", false)
			end
		end
	end
	
	function newGroupClonigContext.adjustPosition(self, _unitName, _offset)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroupClonigContext) .. "adjustPosition(unitName, offset)"
		if type(_unitName) == "string" and ATME.isPoint(_offset) ~= POINT_BAD then
			if initialType == "ship" or initialType == "vehicle" then
				if localMembers.spawnDatas.unitsByName[_unitName] ~= nil then
					local x, y, alt
					if _offset.z ~= nil then
						x = _offset.x
						y = _offset.z
						alt = _offset.y
					else
						x = _offset.x
						y = _offset.y
					end
					
					localMembers.spawnDatas.unitsByName[_unitName].x = localMembers.spawnDatas.unitsByName[_unitName].x + x
					localMembers.spawnDatas.unitsByName[_unitName].y = localMembers.spawnDatas.unitsByName[_unitName].y + y
					
					if alt ~= nil then
						localMembers.spawnDatas.unitsByName[_unitName].alt = localMembers.spawnDatas.unitsByName[_unitName].alt + alt
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "Unit " .. _unitName .. " unknown", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Model group '" .. name .. "' must be a naval or ground group", false)
			end
		else
			if type(_unitName) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "unitName type must be string", true)
			elseif ATME.isPoint(_offset) == POINT_BAD then
				ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be table (x,y)/(x,y,z)", false)
			end
		end
	end
	
	function newGroupClonigContext.setHeading(self, _unitName, _heading)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroupClonigContext) .. "setHeading(unitName, [heading])"
		
		if type(_unitName) == "string" and (_heading == nil or type(_heading) == "number") then
			if initialType == "ship" or initialType == "vehicle" then
				if localMembers.spawnDatas.unitsByName[_unitName] ~= nil then
					if _heading ~= nil then
						localMembers.spawnDatas.unitsByName[_unitName].heading = math.rad(_heading)
					else
						localMembers.spawnDatas.unitsByName[_unitName].heading = math.rad(math.random(-179, 180))
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "Unit " .. _unitName .. " unknown", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Model group '" .. name .. "' must be a naval or ground group", false)
			end
		else
			if type(_unitName) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "unitName type must be string", true)
			elseif _heading ~= nil and type(_heading) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "heading type must be nil or number", false)
			end
		end
	end
	
	function newGroupClonigContext.setSkill(self, _unitName, _skill)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroupClonigContext) .. "setSkill(unitName, [skill])"
		
		if type(_unitName) == "string" and (_skill == nil or type(_skill) == "string") then
			local skill = "RANDOM"

			if _skill ~= nil then
				skill = string.upper(_skill)
			end
			
			if ATMEConst.skill[skill] ~= nil then
				if localMembers.spawnDatas.unitsByName[_unitName] ~= nil then
					localMembers.spawnDatas.unitsByName[_unitName].skill = skill
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "Unit " .. _unitName .. " unknown", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "skill must be 'RANDOM', 'AVERAGE', 'HIGH', 'GOOD' or 'EXCELLENT'", false)
			end
		else
			if type(_unitName) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "unitName type must be string", true)
			elseif type(_skill) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "skill type must be string", false)
			end
		end
	end
	
	function newGroupClonigContext.setCallsign(self, _callsign)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroupClonigContext) .. "setCallsign(callsign)"
		
		if type(_callsign) == "string" or type(_callsign) == "number" then
			if initialType == "plane" or initialType == "helicopter" then
				if #localMembers.spawnDatas.datas.units > 0 then
					if initialCoalition == "BLUE" then -- Coalition BLUE
						local aircraftType = localMembers.spawnDatas.datas.units[1].type
						
						if type(_callsign) == "string" then
							local tokens = ATMEFct.analyzeString(_callsign)
							if #tokens >= 1 and tokens[1].type == "TOKEN_AIRCRAFT_CALLSIGN_NAME" then
								local value
								if #tokens == 2 and tokens[2].type == "TOKEN_INTEGER" and tokens[2].status == "ABSOLUTE" and tokens[2].value <= 9 then
									value = tokens[2].value
								else
									value = 1
								end
								
								if (ATMEConst.specialAircraftTypes[aircraftType] == nil and 
											ATMEConst.blueCallsigns.aircrafts[tokens[1].value].type == "standard") or
									(ATMEConst.specialAircraftTypes[aircraftType] ~= nil and 
											ATMEConst.blueCallsigns.aircrafts[tokens[1].value].type == ATMEConst.specialAircraftTypes[aircraftType]) then
						
									localMembers.callsign = {
										[1] = ATMEConst.blueCallsigns.aircrafts[tokens[1].value].id,
										[2] = value,
										name = ATMEConst.blueCallsigns.aircrafts[tokens[1].value].name .. tokens[2].value,
									}
								else
									ATMEFct.errorOutputCore(nil, fctDefinition, "For BLUE coalition, callsign depends on aircraft type (tanker, awacs...)", false)
								end
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "For BLUE coalition, callsign format must be 'Callsign' or 'Callsign X' x < 9", false)
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "For BLUE coalition, callsign type must be string only", false)
						end
					else --Coalition RED
						local callsignId = 0
						if type(_callsign) == "string" then
							local tokens = ATMEFct.analyzeString(_callsign)
							if #tokens == 1 and tokens[1].type == "TOKEN_INTEGER" and tokens[1].status == "ABSOLUTE" then
								callsignId = math.floor(tonumber(tokens[1].value))
							end
						else
							callsignId = math.floor(_callsign)
						end
						
						if callsignId > 0 and callsignId <= 999 then
							localMembers.callsign[i] = callsignId
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "For RED coalition, callsign format must be 'XXX' or XXX (XXX > 0 and <= 999)", false)
						end
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "can't set callsign - no unit", false)
				end
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "callsign type must be string or number ('CALLSIGN X' or 'XXX'/XXX)", true)
		end
	end
	
	function newGroupClonigContext.repeatSpawn(self, _quantity, _repeatPeriod)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroupClonigContext) .. "repeatSpawn([quantity, repeatPeriod])"
		
		if (_quantity == nil or type(_quantity) == "number" or type(_quantity) == "string") and (_repeatPeriod == nil or type(_repeatPeriod) == "number" or type(_repeatPeriod) == "string") then
			
			if _quantity == nil or _repeatPeriod == nil then
				ATMEDatas.spawnGroupDatasRepeat[localMembers.spawnDatas.datas.name] = nil
				return
			end
			
			local maxQuantity = 0

			local err, qtyRandom, qtyMin, qtyMax = ATMEFct.getRandomDatas(_quantity)
			if err ~= nil then
				ATMEFct.errorOutputCore(nil, fctDefinition, err .. "quantity", false)
				return
			else
				if qtyRandom == true then
					if qtyMin >= 1 then
						if qtyMax ~= nil then 
							if qtyMax <= 20 then
								maxQuantity = math.random(qtyMin, qtyMax)
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #3 must be <= 20 in string format quantity", false)
								return
							end
						else
							maxQuantity = qtyMin
							if maxQuantity < 1 then
								ATMEFct.errorOutputCore(nil, fctDefinition, "quantity must be >= 1", false)
								return
							end
						end
						
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be >= 1 in string format quantity", false)
						return
					end
				else
					maxQuantity = qtyMin
					if maxQuantity < 1 then
						ATMEFct.errorOutputCore(nil, fctDefinition, "quantity must be >= 1", false)
						return
					end
				end
			end
			
			local err, repeatRandom, repeatPeriodMin, repeatPeriodMax = ATMEFct.getRandomDatas(_repeatPeriod)
			if err ~= nil then
				ATMEFct.errorOutputCore(nil, fctDefinition, err .. "repeatPeriod", false)
				return
			else
				if repeatRandom == true then
					if repeatPeriodMin < 1 then
						ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be >= 1 in string format repeatPeriod", false)
						return
					end
				else
					if repeatPeriodMin < 1 then
						ATMEFct.errorOutputCore(nil, fctDefinition, "repeatPeriod must be >= 1", false)
						return
					end
				end
			end

			ATMEDatas.spawnGroupDatasRepeat[localMembers.spawnDatas.datas.name] = {
				name = localMembers.spawnDatas.datas.name,
				cloningContext = self,
				localMembers = localMembers,
				initialType = initialType,
				repeatPeriodMin = repeatPeriodMin,
				repeatPeriodMax = repeatPeriodMax,
				repeatRandom = repeatRandom,
				maxQuantity = maxQuantity,
				nbSpawn = 1,
			}
		else
			if type(_maxQuantity) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "quantity type must be number or string", true)
			elseif type(_repeatPeriod) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "repeatPeriod type must be number or string", false)
			end
		end
	end
	
	function newGroupClonigContext.addUnit(self, _unitDatas, _newUnitName)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroupClonigContext) .. "addUnit(unitDatas, [newUnitName])"
		
		if type(_unitDatas) == "table" and (_newUnitName == nil or type(_newUnitName) == "string") then
			local err, value = ATMEFct.C_GroupCloningContext_addUnit(localMembers.spawnDatas, _unitDatas, _newUnitName, initialType)
			
			if err == true then
				ATMEFct.errorOutputCore(nil, fctDefinition, value, false)
			end
		else
			if type(_unitDatas) ~= "table" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "unitDatas type must be table", true)
			elseif _newUnitName ~= nil and type(_newUnitName) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "newUnitName type must be nil or string", true)
			end
		end
	end
	
	function newGroupClonigContext.getName(self)
		return name
	end
	
	function newGroupClonigContext.addWeapon(self, _unitName, _pylonID, _weapon)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroupClonigContext) .. "addWeapon(unitName, pylonID, weapon)"
		
		if type(_unitName) == "string" and type(_pylonID) == "number" and (_weapon == nil or type(_weapon) == "table") then
			if initialType == "plane" or initialType == "helicopter" then
				if localMembers.spawnDatas.unitsByName[_unitName] ~= nil then
					if _pylonID > 0 then
						if type(_weapon) == "table" then
							localMembers.spawnDatas.unitsByName[_unitName].payload.pylons[_pylonID] = ATMEFct.copyTable(_weapon)
						else
							localMembers.spawnDatas.unitsByName[_unitName].payload.pylons[_pylonID] = nil
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "pylonID must be > 0", false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "unit " .. _unitName .. " doesn't exist", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "unit " .. _unitName .. " must be an aircraft unit", false)
			end
		else
			if type(_unitName) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "unitName type must be string", true)
			elseif type(_pylonID) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "pylonID type must be number", false)
			elseif _weapon ~= nil and type(_weapon) ~= "table" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "weapon type must be nil or table", true)
			end
		end
	end
	
	function newGroupClonigContext.getGroupName(self)
		return localMembers.spawnDatas.datas.name
	end
	
	function newGroupClonigContext.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroupClonigContext) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newGroupClonigContext.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newGroupClonigContext.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroupClonigContext) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newGroupClonigContext.addSpawnCallback(self, _fct)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroupClonigContext) .. "addSpawnCallback(fct)"
		
		if type(_fct) == "function" then
			localMembers.spawnCallback[ATMEDatas.currentModuleName] = _fct
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
		end
	end
	
	function newGroupClonigContext.removeSpawnCallback(self)
		localMembers.spawnCallback[ATMEDatas.currentModuleName] = nil
	end
	
	function newGroupClonigContext.getUserPrivateDatas(self, _moduleName)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroupClonigContext) .. "getUserPrivateDatas([moduleName])"
		local moduleName
		
		if _moduleName ~= nil then
			if type(_moduleName) == "string" then
				if localMembers.userPrivateDatas[_moduleName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
					return
				else
					moduleName = _moduleName
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
				return
			end
		else
			moduleName = ATMEDatas.currentModuleName
		end
		
		return localMembers.userPrivateDatas[moduleName]
	end
	
	function newGroupClonigContext.spawn(self, _route)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroupClonigContext) .. "spawn([route])"

		if _route == nil or ATMEFct.isObjectClass(_route, ATME.C_Route) == true then
			if ATMEDatas.exitATME == false then 
				local route
				
				if _route ~= nil then
					route = _route
					local localMembersRoute = ATMEDatas.localMembers[route:getIndex()]

					if initialType ~= localMembersRoute.type then
						ATMEFct.errorOutputCore(nil, fctDefinition, "route bad type", false)
						return
					end
				else
					route = ATMEFct.C_Route_create(nil, nil, initialType)

					ATMEFct.addWPFrom(route, localMembers.spawnDatas.type, localMembers.spawnDatas.datas.route.points, true, 0, #localMembers.spawnDatas.datas.route.points)
				end

				local text = ATMEFct.C_GroupCloningContext_spawn(self, localMembers, route)

				if ATMEDatas.error ~= nil then
					ATMEFct.warningOutputCore(nil, fctDefinition, text, false)
					return false
				else
					return true
				end
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "route type must be nil or ATME.C_Route", true)
		end
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return newGroupClonigContext
end

ATMEDatas.C_GroupCloningContext_MT = {
	className = "C_GroupCloningContext",
	
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["spawn"] = true,
		["setMainTask"] = true,
		["repeatSpawn"] = true,
		["setPosition"] = true,
		["adjustPosition"] = true,
		["setHeading"] = true,
		["setSkill"] = true,
		["setCallsign"] = true,
		["addUnit"] = true,
		["addWeapon"] = true,
		["getName"] = true,
		["getGroupName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["addSpawnCallback"] = true,
		["removeSpawnCallback"] = true,
		["getUserPrivateDatas"] = true,
	},
	
	add = function(self, _spawnDatas)
		local item = {
			spawnDatas = _spawnDatas,
			callsigns = nil,
			userPrivateDatas = {},		-- Point d'entrée des données privées : un par module
			spawnCallback = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].spawnCallback = nil
	end,
}

ATME.C_GroupCloningContext = {
	getClassName = function(self)
		return ATMEDatas.C_GroupCloningContext_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_GroupCloningContext_MT.parentClass
	end,
	
	createNamed = function(_name, _origin, _newGroupName)
		local fctDefinition = "Function ATME.C_GroupCloningContext.createNamed(name, origin, [newGroupName])"

		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" and 
				(type(_origin) == "string" or type(_origin) == "table") and (type(_newGroupName) == "string" or _newGroupName == nil) then
			local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
			local err, newContext = ATMEFct.C_GroupCloningContext_createNamed(localMembersModule.groupCloningContexts, _name, _origin, _newGroupName)				
			
			if err == true then
				if ATMEDatas.error ~= nil then
					ATMEFct.warningOutputCore(nil, fctDefinition, newContext, false)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, newContext, false)
				end
				
				return
			end
			
			return newContext
		else
			if type(_name) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
			elseif _newGroupName ~= nil and type(_newGroupName) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "newGroupName type must be nil or string", false)
			end
		end
	end,
	
	getByName = function(_name)
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, "Function ATME.C_GroupCloningContext.getByName(name)", "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].groupCloningContexts[_name]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_GroupCloningContext.getByName(name)", "name type must be string", false)
		end
	end,
}

setmetatable(ATME.C_GroupCloningContext,ATMEDatas.C_GroupCloningContext_MT)

ATMEDatas.C_GroupCloningContext_MT.__call = function(t, _origin, _newGroupName)
	local fctDefinition = "Constructor ATME.C_GroupCloningContext(origin, [newGroupName])"
	
	if ATMEDatas.startMissionOk < 2 then
		ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
		return
	end
	
	if (type(_origin) == "string" or type(_origin) == "table") and (type(_newGroupName) == "string" or _newGroupName == nil) then
		local spawnDatas
		if type(_origin) == "table" then
			if ATMEFct.isObjectClass(_origin, ATME.C_Group) == true then
				local localMembersGroup = ATMEDatas.localMembers[_origin:getIndex()]
				spawnDatas = ATMEFct.C_GroupCloningContext_copyTable(localMembersGroup.spawnDatas)
			elseif ATMEFct.C_GroupCloningContext_controlTable(_origin, _newGroupName) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "Bad cloning table structure", false)
				return
			else
				spawnDatas = ATMEFct.C_GroupCloningContext_copyTable(_origin)
				spawnDatas.type = ATMEConst.categoryNameFilters[spawnDatas.type]
				spawnDatas.country = country.id[spawnDatas.country]
				spawnDatas.category = ATMEConst.groupCategories[spawnDatas.type]
				spawnDatas.typeName = nil
				spawnDatas.countryName = nil
			end
		else
			if ATMEDatas.allGroupsInMission[_origin] ~= nil then
				spawnDatas = ATMEFct.C_GroupCloningContext_copyTable(ATMEDatas.allGroupsInMission[_origin].spawnDatas)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Group " .. _origin .. " not defined in mission editor", false)
				return
			end
		end
		
		if spawnDatas.type == "plane" or spawnDatas.type == "helicopter" then
			for _, _unitDatas in ipairs(spawnDatas.datas.units) do
				if _unitDatas.skill == "Player" or _unitDatas.skill == "Client" then
					_unitDatas.skill = "Random"
				end
			end
		end

		if _newGroupName ~= nil then
			local d,f = string.find(_newGroupName, "|", 1)
			if d ~= nil then
				ATMEFct.errorOutputCore(nil, fctDefinition, "Name of group " .. _newGroupName .. " can't contain '|' character", false)
				return
			end
			
			spawnDatas.datas.name = _newGroupName
		end

		return ATMEFct.C_GroupCloningContext_create(nil, nil, spawnDatas)
	else
		if type(_origin) ~= "string" and type(_origin) ~= "table" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "origin type must be string or table (DCS format)", false)
		elseif _newGroupName ~= nil and type(_newGroupName) ~= "string" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "newGroupName type must be nil or string", false)
		end
	end
end

ATMEDatas.C_GroupCloningContext_MT.mt.__metatable = "Private"
ATMEDatas.C_GroupCloningContext_MT.mt.__index = ATME.C_GroupCloningContext
ATMEDatas.C_GroupCloningContext_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_GroupCloningContext_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_GroupCloningContext : Bad key " .. k, "table locked", false)
	end
end


----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_StaticCloningContext : gère les données de clonage nécessaires à la duplication des objets 
-- statiques
----------------------------------------------------------------------------------------------------------

ATMEFct.C_StaticCloningContext_getName = function(_name)
	local newId

	if ATMEDatas.spawnedStatics[_name] == nil then
		ATMEDatas.spawnedStatics[_name] = {
			max = 1,
			freeIds = {},
		}	

		newId = ATMEDatas.spawnedStatics[_name].max
	else
		if #ATMEDatas.spawnedStatics[_name].freeIds == 0 then
			ATMEDatas.spawnedStatics[_name].max = ATMEDatas.spawnedStatics[_name].max + 1
			newId = ATMEDatas.spawnedStatics[_name].max
		else
			newId = table.remove(ATMEDatas.spawnedStatics[_name].freeIds)
		end
	end

	-- Le '-' devient un séparateur pour la composition des noms suite à modif DCS des copies (' #%03d' remplacé par '-%d')
	-- Ainsi "Naval-1" aura bien 2 tags "Naval" et "1"
	local newName = string.format("%s-%d", _name, newId)
	
	if ATMEDatas.statics[newName] == nil or ATMEDatas.newCreatedStatics[newName] == nil then
		return newName
	else
		ATMEFct.setError("STATIC_NAME_DUPLICATE")
		return "Static " .. newName .. " already exist - No spawn"
	end
end

ATMEFct.C_StaticCloningContext_createNamed = function(_namedTable, _name, _staticName, _staticDatas)
	if _namedTable[_name] ~= nil then
		ATMEFct.setError("ALREADY_CREATED")
		return true, "named static cloning context already exists"
	else
		local d,f = string.find(_staticName, "|", 1)
		if d == nil then
			return false, ATMEFct.C_StaticCloningContext_create(_name, _namedTable, _staticName, _staticDatas)
		else
			return true, "name " .. _staticName .. " can't contain '|' character"
		end
	end
end

ATMEFct.C_StaticCloningContext_create = function(_name, _staticCloningContexts, _staticName, _staticDatas)
	local newStaticCloningContext = {}
	
	local name
	local tagsName
	
	if _name ~= nil and _staticCloningContexts ~= nil then
		name = _name
		_staticCloningContexts[name] = newStaticCloningContext
		tagsName = ATMEFct.decompString(_name)
	end

	
	setmetatable(newStaticCloningContext, ATMEDatas.C_StaticCloningContext_MT.mt)

	local index = ATMEDatas.C_StaticCloningContext_MT:add(
		_staticDatas.country,
		{
			name = _staticName,
			type = _staticDatas.datas.type,
			category = _staticDatas.datas.category,
			shape_name = _staticDatas.datas.shape_name,
			livery_id = _staticDatas.datas.livery_id,
		}
	)
	
	local localMembers = ATMEDatas.localMembers[index]
						
	function newStaticCloningContext.getIndex(self)
		return index
	end
	
	function newStaticCloningContext.getName(self)
		return name
	end
	
	function newStaticCloningContext.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newStaticCloningContext) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newStaticCloningContext.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newStaticCloningContext.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newStaticCloningContext) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newStaticCloningContext.getUserPrivateDatas(self, _moduleName)
		local fctDefinition = ATMEFct.getErrorPrefix(newStaticCloningContext) .. "getUserPrivateDatas([moduleName])"
		local moduleName
		
		if _moduleName ~= nil then
			if type(_moduleName) == "string" then
				if localMembers.userPrivateDatas[_moduleName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
					return
				else
					moduleName = _moduleName
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
				return
			end
		else
			moduleName = ATMEDatas.currentModuleName
		end
		
		return localMembers.userPrivateDatas[moduleName]
	end
	
	function newStaticCloningContext.addSpawnCallback(self, _fct)
		local fctDefinition = ATMEFct.getErrorPrefix(newStaticCloningContext) .. "addSpawnCallback(fct)"
		
		if type(_fct) == "function" then
			localMembers.spawnCallback[ATMEDatas.currentModuleName] = _fct
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
		end
	end
	
	function newStaticCloningContext.removeSpawnCallback(self)
		localMembers.spawnCallback[ATMEDatas.currentModuleName] = nil
	end
	
	function newStaticCloningContext.spawn(self, _position, _data1, _data2, _data3, _data4)
		local fctDefinition = ATMEFct.getErrorPrefix(newStaticCloningContext) .. "spawn(position, heading, [dead])\nspawn(boat, x, y, azimuth)"

		if ATMEDatas.exitATME == false then -- Test nécessaire car sinon problème sur l'événement du spawn (S_EVENT_BIRTH) 
											-- désactivé par ATMEDatas.exitATME à true
											
			if type(_position) == "string" or ATME.isPoint(_position) ~= POINT_BAD then 
				if type(_data1) == "number" and (_data2 == nil or type(_data2) == "boolean") then		
					local point
					local newName = ATMEFct.C_StaticCloningContext_getName(localMembers.staticDatas.name)
					
					if ATMEDatas.error ~= nil then
						-- Warning doublon de nom
						ATMEFct.warningOutputCore(nil, fctDefinition, newName, false)
						return false
					end

					if type(_position) == "string" then
						local zone = ATMEFct.getDCSZone(_position)
						if zone ==  nil then
							ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone " .. _position .. " unknown", false)
							return
						end

						point = {}
						local angle = 2 * math.pi * math.random()
						local radius = math.random(0, zone.radius)
						point.x = zone.point.x + radius * math.sin(angle)
						point.z = zone.point.z + radius * math.cos(angle)
					else
						point = ATMEFct.convertToPoint3D(_position) 
					end
					
					if _data1 >= 0 and _data1 < 360 then
						local staticDatas = {
							name = newName,
							type = localMembers.staticDatas.type,
							canCargo = false,
							category = localMembers.staticDatas.category,
							shape_name = localMembers.staticDatas.shape_name,
							livery_id = localMembers.staticDatas.livery_id,
							x = point.x,
							y = point.z,
							heading = math.rad(_data1),
							dead = _data2,
						}

						-- Permet de transmettre la catégorie lors de la création de l'objet. A ce moment
						-- là sera remis à nil
						
						ATMEDatas.staticsToBirth[newName] = {
							isSpawned = true,
							spawnCallback = localMembers.spawnCallback,
							staticCloningContext = self,
							category = localMembers.staticDatas.category,
							idSpawnedList = ATMEDatas.spawnedStatics[localMembers.staticDatas.name],
							id = newId,
						}
						
						coalition.addStaticObject(country.id[localMembers.country], staticDatas)
						
						return true
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "heading must be >= 0 and < 360", false)
					end
				else
					if type(_data1) ~= "number" then
						ATMEFct.errorOutputCore(nil, fctDefinition, "heading type must be number", false)
					elseif _data2 ~= nil and type(_data2) ~= "boolean" then
						ATMEFct.errorOutputCore(nil, fctDefinition, "dead type must be nil or boolean", false)
					end
				end
			elseif ATMEFct.isObjectClass(_position, ATME.C_Unit) == true then
				local localMembersUnit = ATMEDatas.localMembers[_position:getIndex()]
				
				if localMembersUnit.objectDCS:hasAttribute("AircraftCarrier") == true or localMembersUnit.objectDCS:hasAttribute("HelicopterCarrier") == true then
					if localMembersUnit.objectDCS:isExist() == true then
						if type(_data1) == "number" and type(_data2) == "number" and type(_data3) == "number" then
							local x = _data1
							local y = _data2
							local angle
							local dead
							
							local newName = ATMEFct.C_StaticCloningContext_getName(localMembers.staticDatas.name)
					
							if ATMEDatas.error ~= nil then
								-- Warning doublon de nom
								ATMEFct.warningOutputCore(nil, fctDefinition, newName, false)
								return false
							end
							
							if _data3 >= 0 and _data3 < 360 then
								angle = math.rad(_data3)
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "azimuth must be >= 0 and < 360", false)
								return
							end
							
							local staticDatas = {
								name = newName,
								type = localMembers.staticDatas.type,
								canCargo = false,
								category = localMembers.staticDatas.category,
								shape_name = localMembers.staticDatas.shape_name,
								livery_id = localMembers.staticDatas.livery_id,
								linkUnit = localMembersUnit.objectDCSId,
								offsets = {
									x = x,
									y = y,
									angle = angle,
								},
								x = 0,
								y = 0,
								heading = 0,
								linkOffset = true,
								dead = false,
							}

							-- Permet de transmettre la catégorie lors de la création de l'objet. A ce moment
							-- là sera remis à nil
					
							ATMEDatas.staticsToBirth[newName] = {
								isSpawned = true,
								spawnCallback = localMembers.spawnCallback,
								staticCloningContext = self,
								category = localMembers.staticDatas.category,
								idSpawnedList = ATMEDatas.spawnedStatics[localMembers.staticDatas.name],
								id = newId,
							}

							coalition.addStaticObject(country.id[localMembers.country], staticDatas)
							
							return true
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "x, y and azimuth type must all be number", false)
						end
					else
						ATMEFct.setError("ALREADY_DISABLED")
						ATMEFct.warningOutputCore(nil, fctDefinition, "bound boat already disabled", false)
						
						return false
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "boat must be aircraft carrier or helicopter carrier", true)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "position/boat type must be ATME.C_Unit or string or table (x,y)/(x,y,z)", true)
			end
		end
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end

	return newStaticCloningContext
end

ATMEDatas.C_StaticCloningContext_MT = {
	className = "C_StaticCloningContext",
	
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["spawn"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["getUserPrivateDatas"] = true,
		["addSpawnCallback"] = true,
		["removeSpawnCallback"] = true,
	},

	add = function(self, _country, _staticDatas)
		local item = {
			country = _country,
			staticDatas = _staticDatas,
			userPrivateDatas = {},						-- Point d'entrée des données privées : un par module
			spawnCallback = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].spawnCallback = nil
	end,
}

ATME.C_StaticCloningContext = {
	getClassName = function(self)
		return ATMEDatas.C_StaticCloningContext_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_StaticCloningContext_MT.parentClass
	end,
	
	createNamed = function(_name, _origin, _staticName)
		local fctDefinition = "Function ATME.C_StaticCloningContext.createNamed(name, origin, [staticName])"
		
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" and (type(_origin) == "table" or type(_origin) == "string") and (_staticName == nil or type(_staticName) == "string") then
			local staticDatas
			local staticName = _staticName
			
			if type(_origin) == "string" then
				if ATMEDatas.allStaticsInMission[_origin] ~= nil then
					staticDatas = {
						country = ATMEDatas.allStaticsInMission[_origin].country,
						datas = {
							type = ATMEDatas.allStaticsInMission[_origin].type,
							category = ATMEDatas.allStaticsInMission[_origin].category,
							shape_name = ATMEDatas.allStaticsInMission[_origin].shape_name,
							livery_id = ATMEDatas.allStaticsInMission[_origin].livery_id,
						},
					}
					
					if staticName == nil then
						staticName = ATMEDatas.allStaticsInMission[_origin].name
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "Static " .. _origin .. " unknown in mission", false)
					return
				end
			elseif ATMEFct.isObjectClass(_origin, ATME.C_Static) == true then
				local localMembers = ATMEDatas.localMembers[_origin:getIndex()]
				staticDatas = ATMEFct.copyTable(localMembers.spawnDatas)
				
				if staticName == nil then
					staticName = localMembers.spawnDatas.name
				end
			else
				if type(_origin.country) == "string" then
					local countryName = string.upper(_origin.country)
					if country.id[countryName] ~= nil then
						if type(_origin.datas) == "table" then
							if type(_origin.datas.category) == "string" and type(_origin.datas.type) == "string" then
								local category = string.upper(_origin.datas.category)
								if ATMEConst.staticCategories[category] ~= nil then
									staticDatas = {
										country = countryName,
										datas = {
											type = _origin.datas.type,
											category = ATMEConst.staticCategories[category].category,
											shape_name = _origin.datas.shape_name,
											livery_id = _origin.datas.livery_id,
										},
									}
									
									if staticName == nil then
										if _origin.datas.name == nil then
											ATMEFct.errorOutputCore(nil, fctDefinition, "static name can't be nil", false)
											return
										end
										
										staticName = _origin.datas.name
									end
								else
									ATMEFct.errorOutputCore(nil, fctDefinition, "Category " .. _origin.datas.category .. " unknown", false)
									return
								end
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "Bad cloning table structure", false)
								return
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "Bad cloning table structure", false)
							return
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Country " .. _origin.country .. " unknown", false)
						return
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "Bad cloning table structure", false)
					return
				end
			end
			
			local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
			local err, newContext = ATMEFct.C_StaticCloningContext_createNamed(localMembersModule.staticCloningContexts, _name, staticName, staticDatas)				
			
			if err == true then
				if ATMEDatas.error ~= nil then
					ATMEFct.warningOutputCore(nil, fctDefinition, newContext, false)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, newContext, false)
				end
				
				return
			end
			
			return newContext
		else
			if type(_name) ~= "string" then 
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
			elseif type(_origin) ~= "table" and type(_origin) ~=  "string" then 
				ATMEFct.errorOutputCore(nil, fctDefinition, "origin type must be string, table or ATME.C_Static", false)
			elseif _staticName ~= nil and type(_staticName) ~= "string" then 
				ATMEFct.errorOutputCore(nil, fctDefinition, "staticName type must be nil or string", false)
			end
		end
	end,
	
	getByName = function(_name)
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, "Function ATME.C_StaticCloningContext.getByName(name)", "Modules loading ... ATME.run() function not called", false)
			return
		end
	
		if type(_name) == "string" then
			return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].staticCloningContexts[_name]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_StaticCloningContext.getByName(name)", "name type must be string", false)
		end
	end,
}

setmetatable(ATME.C_StaticCloningContext,ATMEDatas.C_StaticCloningContext_MT)

ATMEDatas.C_StaticCloningContext_MT.__call = function(t, _origin, _staticName)
	local fctDefinition = "Constructor ATME.C_StaticCloningContext(origin, [staticName])"
	
	if ATMEDatas.startMissionOk < 2 then
		ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
		return
	end
	
	if (type(_origin) == "table" or type(_origin) == "string") and (_staticName == nil or type(_staticName) == "string") then
		local staticDatas
		local staticName = _staticName
		
		if type(_origin) == "string" then
			if ATMEDatas.allStaticsInMission[_origin] ~= nil then
				staticDatas = {
					country = ATMEDatas.allStaticsInMission[_origin].country,
					datas = {
						type = ATMEDatas.allStaticsInMission[_origin].type,
						category = ATMEDatas.allStaticsInMission[_origin].category,
						shape_name = ATMEDatas.allStaticsInMission[_origin].shape_name,
						livery_id = ATMEDatas.allStaticsInMission[_origin].livery_id,
					},
				}
				
				if staticName == nil then
					staticName = ATMEDatas.allStaticsInMission[_origin].name
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Static " .. _origin .. " unknown in mission", false)
				return
			end
		elseif ATMEFct.isObjectClass(_origin, ATME.C_Static) == true then
			local localMembers = ATMEDatas.localMembers[_origin:getIndex()]
			staticDatas = ATMEFct.copyTable(localMembers.spawnDatas)
			
			if staticName == nil then
				staticName = localMembers.spawnDatas.name
			end
		else
			if type(_origin.country) == "string" then
				local countryName = string.upper(_origin.country)
				if country.id[countryName] ~= nil then
					if type(_origin.datas) == "table" then
						if type(_origin.datas.category) == "string" and type(_origin.datas.type) == "string" then
							local category = string.upper(_origin.datas.category)

							if ATMEConst.staticCategories[category] ~= nil then
								staticDatas = {
									country = countryName,
									datas = {
										type = _origin.datas.type,
										category = ATMEConst.staticCategories[category].category,
										shape_name = _origin.datas.shape_name,
										livery_id = _origin.datas.livery_id,
									},
								}
								
								if staticName == nil then
									if _origin.datas.name == nil then
										ATMEFct.errorOutputCore(nil, fctDefinition, "static name can't be nil", false)
										return
									end
					
									staticName = _origin.datas.name
								end
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "Category " .. _origin.datas.category .. " unknown", false)
								return
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "Bad cloning table structure", false)
							return
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Bad cloning table structure", false)
						return
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "Country " .. _origin.country .. " unknown", false)
					return
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Bad cloning table structure", false)
				return
			end
		end
		
		local d,f = string.find(staticName, "|", 1)
		if d == nil then
			return ATMEFct.C_StaticCloningContext_create(nil, nil, staticName, staticDatas)	
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name " .. staticName .. " can't contain '|' character", false)
		end
	else
		if type(_origin) ~= "table" and type(_origin) ~= "string" then 
			ATMEFct.errorOutputCore(nil, fctDefinition, "origin type must be string, table or ATME.C_Static", false)
		elseif _staticName ~= nil and type(_staticName) ~= "string" then 
			ATMEFct.errorOutputCore(nil, fctDefinition, "staticName type must be nil or string", false)
		end
	end
end

ATMEDatas.C_StaticCloningContext_MT.mt.__metatable = "Private"
ATMEDatas.C_StaticCloningContext_MT.mt.__index = ATME.C_StaticCloningContext
ATMEDatas.C_StaticCloningContext_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_StaticCloningContext_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_StaticCloningContext : Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_CargoCloningContext : gère les données de clonage nécessaires à la duplication 
-- des cargos
----------------------------------------------------------------------------------------------------------

ATMEFct.C_CargoCloningContext_getName = function(_name)
	local newId

	if ATMEDatas.spawnedCargos[_name] == nil then
		ATMEDatas.spawnedCargos[_name] = {
			max = 1,
			freeIds = {},
		}	

		newId = ATMEDatas.spawnedCargos[_name].max
	else
		if #ATMEDatas.spawnedCargos[_name].freeIds == 0 then
			ATMEDatas.spawnedCargos[_name].max = ATMEDatas.spawnedCargos[_name].max + 1
			newId = ATMEDatas.spawnedCargos[_name].max
		else
			newId = table.remove(ATMEDatas.spawnedCargos[_name].freeIds)
		end
	end

	-- Le '-' devient un séparateur pour la composition des noms suite à modif DCS des copies (' #%03d' remplacé par '-%d')
	-- Ainsi "Naval-1" aura bien 2 tags "Naval" et "1"
	local newName = string.format("%s-%d", _name, newId)
	
	if ATMEDatas.cargos[newName] == nil and ATMEDatas.newCreatedCargos[newName] == nil then
		return newName
	else
		ATMEFct.setError("CARGO_NAME_DUPLICATE")
		return "Cargo " .. newName .. " already exist - No spawn"
	end
end

ATMEFct.C_CargoCloningContext_createNamed = function(_namedTable, _name, _cargoName, _cargoDatas)
	if _namedTable[_name] ~= nil then
		ATMEFct.setError("ALREADY_CREATED")
		return true, "named cargo cloning context already exists"
	else
		local d,f = string.find(_cargoName, "|", 1)
		if d == nil then
			return false, ATMEFct.C_CargoCloningContext_create(_name, _namedTable, _cargoName, _cargoDatas)	
		else
			return true, "name " .. _cargoName .. " can't contain '|' character"
		end
	end
end

ATMEFct.C_CargoCloningContext_create = function(_name, _cargoCloningContexts, _cargoName, _cargoDatas)
	local newCargoCloningContext = {}
	
	local name
	local tagsName
	
	if _name ~= nil and _cargoCloningContexts ~= nil then
		name = _name
		_cargoCloningContexts[name] = newCargoCloningContext
		tagsName = ATMEFct.decompString(_name)
	end

	setmetatable(newCargoCloningContext, ATMEDatas.C_CargoCloningContext_MT.mt)
	
	local index = ATMEDatas.C_CargoCloningContext_MT:add(
		_cargoDatas.country,
		{
			name = _cargoName,
			type = _cargoDatas.datas.type,
			shape_name = _cargoDatas.datas.shape_name,
			rate = 100,
			mass = 100,
		}
	)

	local localMembers = ATMEDatas.localMembers[index]
						
	function newCargoCloningContext.getIndex(self)
		return index
	end
	
	function newCargoCloningContext.getName(self)
		return name
	end
	
	function newCargoCloningContext.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargoCloningContext) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newCargoCloningContext.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newCargoCloningContext.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargoCloningContext) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newCargoCloningContext.getUserPrivateDatas(self, _moduleName)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargoCloningContext) .. "getUserPrivateDatas([moduleName])"
		local moduleName
		
		if _moduleName ~= nil then
			if type(_moduleName) == "string" then
				if localMembers.userPrivateDatas[_moduleName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
					return
				else
					moduleName = _moduleName
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
				return
			end
		else
			moduleName = ATMEDatas.currentModuleName
		end
		
		return localMembers.userPrivateDatas[moduleName]
	end
	
	function newCargoCloningContext.addSpawnCallback(self, _fct)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargoCloningContext) .. "addSpawnCallback(fct)"
		
		if type(_fct) == "function" then
			localMembers.spawnCallback[ATMEDatas.currentModuleName] = _fct
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
		end
	end
	
	function newCargoCloningContext.removeSpawnCallback(self)
		localMembers.spawnCallback[ATMEDatas.currentModuleName] = nil
	end
	
	function newCargoCloningContext.setWeight(self, _mass)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargoCloningContext) .. "setWeight(mass)"
		
		if type(_mass) == "number" then
			if math.floor(_mass) >= 0 then
				localMembers.cargoDatas.mass = math.floor(_mass)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "mass must be >= 0", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "mass type must be number", true)
		end
	end
	
	function newCargoCloningContext.spawn(self, _position, _data1, _data2, _data3, _data4)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargoCloningContext) .. "spawn(position, heading, [dead])\nspawn(boat, x, y, azimuth)"

		if ATMEDatas.exitATME == false then -- Test nécessaire car sinon problème sur l'événement du spawn (S_EVENT_BIRTH) 
											-- désactivé par ATMEDatas.exitATME à true		
			if type(_position) == "string" or ATME.isPoint(_position) ~= POINT_BAD then 
				if type(_data1) == "number" and (_data2 == nil or type(_data2) == "boolean") then		
					local point
					local newName = ATMEFct.C_CargoCloningContext_getName(localMembers.cargoDatas.name)
					if ATMEDatas.error ~= nil then
						-- Warning doublon de nom
						ATMEFct.warningOutputCore(nil, fctDefinition, newName, false)
						return false
					end

					if type(_position) == "string" then
						local zone = ATMEFct.getDCSZone(_position)
						if zone ==  nil then
							ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone " .. _position .. " unknown", false)
							return
						end

						point = {}
						local angle = 2 * math.pi * math.random()
						local radius = math.random(0, zone.radius)
						point.x = zone.point.x + radius * math.sin(angle)
						point.z = zone.point.z + radius * math.cos(angle)
					else
						point = ATMEFct.convertToPoint3D(_position) 
					end

					if _data1 >= 0 and _data1 < 360 then
						local cargoDatas = {
							name = newName,
							type = localMembers.cargoDatas.type,
							canCargo = true,
							shape_name = localMembers.cargoDatas.shape_name,
							rate = localMembers.cargoDatas.rate,
							mass = localMembers.cargoDatas.mass,
							x = point.x,
							y = point.z,
							heading = math.rad(_data1),
							dead = _data2,
						}
						
						-- Permet de transmettre la catégorie lors de la création de l'objet. A ce moment
						-- là sera remis à nil

						ATMEDatas.cargosToBirth[newName] = {
							isSpawned = true,
							spawnCallback = localMembers.spawnCallback,
							cargoCloningContext = self,
							idSpawnedList = ATMEDatas.spawnedCargos[localMembers.cargoDatas.name],
							id = newId,
						}

						coalition.addStaticObject(country.id[localMembers.country], cargoDatas)
						
						return true
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "heading must be >= 0 and < 360", false)
					end
				else
					if type(_data1) ~= "number" then
						ATMEFct.errorOutputCore(nil, fctDefinition, "heading type must be number", false)
					elseif _data2 ~= nil and type(_data2) ~= "boolean" then
						ATMEFct.errorOutputCore(nil, fctDefinition, "dead type must be nil or boolean", false)
					end
				end
			elseif ATMEFct.isObjectClass(_position, ATME.C_Unit) == true then
				local localMembersUnit = ATMEDatas.localMembers[_position:getIndex()]
				
				if localMembersUnit.objectDCS:hasAttribute("AircraftCarrier") == true or localMembersUnit.objectDCS:hasAttribute("HelicopterCarrier") == true then
					if localMembersUnit.objectDCS:isExist() == true then
						if type(_data1) == "number" and type(_data2) == "number" and type(_data3) == "number" then
							local x = _data1
							local y = _data2
							local angle
							local dead
							
							local newName = ATMEFct.C_CargoCloningContext_getName(localMembers.cargoDatas.name)
							if ATMEDatas.error ~= nil then
								-- Warning doublon de nom
								ATMEFct.warningOutputCore(nil, fctDefinition, newName, false)
								return false
							end
							
							if _data3 >= 0 and _data3 < 360 then
								angle = math.rad(_data3)
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "azimuth must be >= 0 and < 360", false)
								return
							end
							
							local cargoDatas = {
								name = newName,
								type = localMembers.cargoDatas.type,
								canCargo = true,
								shape_name = localMembers.cargoDatas.shape_name,
								rate = localMembers.cargoDatas.rate,
								mass = localMembers.cargoDatas.mass,
								linkUnit = localMembersUnit.objectDCSId,
								offsets = {
									x = x,
									y = y,
									angle = angle,
								},
								x = 0,
								y = 0,
								heading = 0,
								linkOffset = true,
								dead = false,
							}

							-- Permet de transmettre la catégorie lors de la création de l'objet. A ce moment
							-- là sera remis à nil
					
							ATMEDatas.cargosToBirth[newName] = {
								isSpawned = true,
								spawnCallback = localMembers.spawnCallback,
								cargoCloningContext = self,
								idSpawnedList = ATMEDatas.spawnedCargos[localMembers.cargoDatas.name],
								id = newId,
							}

							coalition.addStaticObject(country.id[localMembers.country], cargoDatas)
							
							return true
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "x, y and azimuth type must all be number", false)
						end
					else
						ATMEFct.setError("ALREADY_DISABLED")
						ATMEFct.warningOutputCore(nil, fctDefinition, "bound boat already disabled", false)
						return false
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "boat must be aircraft carrier or helicopter carrier", true)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "position/boat type must be ATME.C_Unit or string or table (x,y)/(x,y,z)", true)
			end
		end
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return newCargoCloningContext
end

ATMEDatas.C_CargoCloningContext_MT = {
	className = "C_CargoCloningContext",
	
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["setWeight"] = true,
		["spawn"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["getUserPrivateDatas"] = true,
		["addSpawnCallback"] = true,
		["removeSpawnCallback"] = true,
	},
	
	add = function(self, _country, _cargoDatas)
		local item = {
			country = _country,
			cargoDatas = _cargoDatas,
			userPrivateDatas = {},						-- Point d'entrée des données privées : un par module
			spawnCallback = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].spawnCallback = nil
	end,
}

ATME.C_CargoCloningContext = {
	getClassName = function(self)
		return ATMEDatas.C_CargoCloningContext_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_CargoCloningContext_MT.parentClass
	end,
	
	createNamed = function(_name, _origin, _cargoName)
		local fctDefinition = "Function ATME.C_CargoCloningContext.createNamed(name, origin, [cargoName])"
		
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
			return
		end
		
		if type(_name) == "string" and (type(_origin) == "table" or type(_origin) == "string") and (_cargoName == nil or type(_cargoName) == "string") then
			local cargoDatas
			local cargoName = _cargoName
			
			if type(_origin) == "string" then
				if ATMEDatas.allCargosInMission[_origin] ~= nil then
					cargoDatas = {
						country = ATMEDatas.allCargosInMission[_origin].country,
						datas = {
							type = ATMEDatas.allCargosInMission[_origin].type,
							shape_name = ATMEDatas.allCargosInMission[_origin].shape_name,
						},
					}
					
					if cargoName == nil then
						cargoName = ATMEDatas.allCargosInMission[_origin].name
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "Cargo " .. _origin .. " unknown in mission", false)
					return
				end
			elseif ATMEFct.isObjectClass(_origin, ATME.C_Cargo) == true then
				local localMembers = ATMEDatas.localMembers[_origin:getIndex()]
				cargoDatas = ATMEFct.copyTable(localMembers.spawnDatas)
				
				if cargoName == nil then
					cargoName = localMembers.spawnDatas.name
				end
			else
				if type(_origin.country) == "string" then
					local countryName = string.upper(_origin.country)
					if country.id[countryName] ~= nil then
						if type(_origin.datas) == "table" then
						if type(_origin.datas.type) == "string" then
							if type(_origin.datas.shape_name) == "string" then
								cargoDatas = {
									country = countryName,
									datas = {
										type = _origin.datas.type,
										shape_name = _origin.datas.shape_name,
									},
								}
								
								if cargoName == nil then
									if _origin.datas.name == nil then
										ATMEFct.errorOutputCore(nil, fctDefinition, "cargo name can't be nil", false)
										return
									end
					
									cargoName = _origin.datas.name
								end
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "Bad cloning table structure", false)
								return
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "Bad cloning table structure", false)
							return
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Bad cloning table structure", false)
						return
					end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Country " .. _origin.country .. " unknown", false)
						return
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "Bad cloning table structure", false)
					return
				end
			end
		
			local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
			local err, newContext = ATMEFct.C_CargoCloningContext_createNamed(localMembersModule.cargoCloningContexts, _name, cargoName, cargoDatas)				
			
			if err == true then
				if ATMEDatas.error ~= nil then
					ATMEFct.warningOutputCore(nil, fctDefinition, newContext, false)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, newContext, false)
				end
				
				return
			end
			
			return newContext
		else
			if type(_name) ~= "string" then 
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
			elseif type(_countryName) ~= "string" then 
				ATMEFct.errorOutputCore(nil, fctDefinition, "countryName type must be string", false)
			elseif _cargoName ~= nil and type(_cargoName) ~= "string" then 
				ATMEFct.errorOutputCore(nil, fctDefinition, "cargoName type must be nil or string", false)
			end
		end
	end,
	
	getByName = function(_name)
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, "Function ATME.C_CargoCloningContext.getByName(name)", "Modules loading ... ATME.run() function not called", false)
			return
		end
		
		if type(_name) == "string" then
			return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].cargoCloningContexts[_name]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_CargoCloningContext.getByName(name)", "name type must be string", false)
		end
	end,
}

setmetatable(ATME.C_CargoCloningContext,ATMEDatas.C_CargoCloningContext_MT)

ATMEDatas.C_CargoCloningContext_MT.__call = function(t, _origin, _cargoName)
	local fctDefinition = "Constructor ATME.C_CargoCloningContext(origin, [cargoName])"

	if ATMEDatas.startMissionOk < 2 then
		ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
		return
	end
	
	if (type(_origin) == "table" or type(_origin) == "string") and (_cargoName == nil or type(_cargoName) == "string") then
		local cargoDatas
		local cargoName = _cargoName
		
		if type(_origin) == "string" then
			if ATMEDatas.allCargosInMission[_origin] ~= nil then
				cargoDatas = {
					country = ATMEDatas.allCargosInMission[_origin].country,
					datas = {
						type = ATMEDatas.allCargosInMission[_origin].type,
						shape_name = ATMEDatas.allCargosInMission[_origin].shape_name,
					},
				}
				
				if cargoName == nil then
					cargoName = ATMEDatas.allCargosInMission[_origin].name
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Cargo " .. _origin .. " unknown in mission", false)
				return
			end
		elseif ATMEFct.isObjectClass(_origin, ATME.C_Cargo) == true then
			local localMembers = ATMEDatas.localMembers[_origin:getIndex()]
			cargoDatas = ATMEFct.copyTable(localMembers.spawnDatas)
			
			if cargoName == nil then
				cargoName = localMembers.spawnDatas.name
			end
		else
			if type(_origin.country) == "string" then
				local countryName = string.upper(_origin.country)
				if country.id[countryName] ~= nil then
					if type(_origin.datas) == "table" then
						if type(_origin.datas.type) == "string" then
							if type(_origin.datas.shape_name) == "string" then
								cargoDatas = {
									country = countryName,
									datas = {
										type = _origin.datas.type,
										shape_name = _origin.datas.shape_name,
									},
								}
								
								if cargoName == nil then
									if _origin.datas.name == nil then
										ATMEFct.errorOutputCore(nil, fctDefinition, "cargo name can't be nil", false)
										return
									end
					
									cargoName = _origin.datas.name
								end
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "Bad cloning table structure", false)
								return
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "Bad cloning table structure", false)
							return
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Bad cloning table structure", false)
						return
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "Country " .. _origin.country .. " unknown", false)
					return
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Bad cloning table structure", false)
				return
			end
		end

		local d,f = string.find(cargoName, "|", 1)
		if d == nil then
			return ATMEFct.C_CargoCloningContext_create(nil, nil, cargoName, cargoDatas)	
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name " .. cargoName .. " can't contain '|' character", false)
		end
	else
		if type(_origin) ~= "table" and type(_origin) ~= "string" then 
			ATMEFct.errorOutputCore(nil, fctDefinition, "origin type must be string, table or ATME.C_Cargo", false)
		elseif _cargoName ~= nil and type(_cargoName) ~= "string" then 
			ATMEFct.errorOutputCore(nil, fctDefinition, "cargoName type must be nil or string", false)
		end
	end
end

ATMEDatas.C_CargoCloningContext_MT.mt.__metatable = "Private"
ATMEDatas.C_CargoCloningContext_MT.mt.__index = ATME.C_CargoCloningContext
ATMEDatas.C_CargoCloningContext_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_CargoCloningContext_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_CargoCloningContext : Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_FindContext : Gestion des contextes de recherche 
--                   .
-- Un context de recherche permet de fixer les critères, le type d'unités et les filtres. Une callback peut
-- être associée
----------------------------------------------------------------------------------------------------------

ATMEFct.C_FindContext_createNamed = function(_namedTable, _name, _findType, _findCategories)
	if _namedTable[_name] ~= nil then
		ATMEFct.setError("ALREADY_CREATED")
		return true, "named find context already exists"
	else
		local d,f = string.find(_name, "|", 1)
		if d == nil then
			return false, ATMEFct.C_FindContext_create(_name, _namedTable, _findType, _findCategories)
		else
			ATMEFct.setError("BAD_FORMAT")
			return true, "name can't contain '|' character"
		end
	end
end

ATMEFct.C_FindContext_Object_isInZone = function(_object, _datas)
	if _datas.zone ~= nil then
		local point = _object:getPosition()
		return ATMEFct[ATMEConst.indZoneConfig[_datas.zone:getClassName()].fctIsInside](point, _datas)
	end
	
	return true
end

ATMEFct.C_FindContext_setToList = function(_datas, _group, _id)
	ATMEConst.findClasses[_datas.localMembers.type].fct(_datas, _group, _id)
end

ATMEFct.C_FindContext_search = function(_object, _datas)
	ATMEConst.findClasses[_datas.localMembers.type].search_fct(_object, _datas)
end

ATMEFct.C_FindContext_findObjects = function(_object, _datas)
	table.insert(_datas.foundObjects, _object)
end

ATMEFct.C_FindContext_findAirbase = function(_all, _point, _airbase, _datas)
	-- Traitement cause bug DCS search_object sur les airbases - V2.5.6
	
	if _all == true or ATMEFct[ATMEConst.indZoneConfig[_datas.zone:getClassName()].fctIsInside](_point, _datas) == true then
		local record
		local status
		
		if _datas.localMembers.callbacks[_datas.moduleName] ~= nil then
			local datas = {
				zone = _zone,
				object = _airbase,
			}
			status, record = xpcall(function() return _datas.localMembers.callbacks[_datas.moduleName](_datas.findContext, datas) end, function() ATMEFct.errUserHandler('Filter callback') end)
		end
		
		if record == nil or (type(record) == "boolean" and record == true) then
			table.insert(_datas.objectsIndex, _airbase:getName())
			_datas.objects[_airbase:getName()] = _airbase
		end
	end
end

ATMEFct.C_FindContext_find = function(_type, _zone, _datas)
	if _type ~= "AIRBASES" then -- world.searchObjects ne fonctionne pas avec BASE - Bug DCS pas d'objet retourné.
		if ATMEFct.isObjectClass(_zone, ATME.C_Zones) == true then
			local localMembersZones = ATMEDatas.localMembers[_zone:getIndex()]
			
			for _, _zoneItem in ipairs(localMembersZones.zonesList) do		
				local localMembersZone = ATMEDatas.localMembers[_zoneItem.zone:getIndex()]

				_datas.zone = _zoneItem.zone
				_datas.actualCenter = ATMEFct.C_Zone_getActualCenter(localMembersZone)
				_datas.actualAzimuth = ATMEFct.C_Zone_getActualAzimuth(localMembersZone)

				local idVolume, searchFct, paramsVolume = ATMEFct[ATMEConst.indZoneConfig[_zoneItem.zone:getClassName()].fctEvalSearchVolume](_datas)
				
				local volume = {
					id = idVolume,
					params = paramsVolume,
				}
				
				if _datas.localMembers.coroutines[_datas.moduleName] ~= nil then
					_datas.foundObjects = {}

					world.searchObjects(ATMEConst.findClasses[_datas.localMembers.type].category, volume, ATMEFct.C_FindContext_findObjects, _datas)
					
					for _i, _object in ipairs(_datas.foundObjects) do
						ATMEFct[searchFct](_object, _datas)
						coroutine.yield()
						
						if _datas.localMembers.resultCallbacks[_datas.moduleName].aborted == true then
							return
						end
					end
				else
					world.searchObjects(ATMEConst.findClasses[_datas.localMembers.type].category, volume, ATMEFct[searchFct], _datas)
					ATMEFct.C_FindContext_finalize(_datas)
				end
			end
		elseif _zone ~= nil then
			local localMembersZone = ATMEDatas.localMembers[_zone:getIndex()]

			_datas.zone = _zone
			_datas.actualCenter = ATMEFct.C_Zone_getActualCenter(localMembersZone)
			_datas.actualAzimuth = ATMEFct.C_Zone_getActualAzimuth(localMembersZone)

			local idVolume, searchFct, paramsVolume = ATMEFct[ATMEConst.indZoneConfig[_zone:getClassName()].fctEvalSearchVolume](_datas)
					
			local volume = {
				id = idVolume,
				params = paramsVolume,
			}
			
			if _datas.localMembers.coroutines[_datas.moduleName] ~= nil then
				_datas.foundObjects = {}

				world.searchObjects(ATMEConst.findClasses[_datas.localMembers.type].category, volume, ATMEFct.C_FindContext_findObjects, _datas)
				
				for _i, _object in ipairs(_datas.foundObjects) do
					ATMEFct[searchFct](_object, _datas)
					coroutine.yield()
					
					if _datas.localMembers.resultCallbacks[_datas.moduleName].aborted == true then
						return
					end
				end
			else
				world.searchObjects(ATMEConst.findClasses[_datas.localMembers.type].category, volume, ATMEFct[searchFct], _datas)
				ATMEFct.C_FindContext_finalize(_datas)
			end
		else
			local listObjects = ATMEConst.findClasses[_datas.localMembers.type].get_objects()
			for _, _object in pairs(listObjects) do
				ATMEFct.C_FindContext_search(_object:getDCSObject(), _datas)
				
				if _datas.localMembers.coroutines[_datas.moduleName] ~= nil then
					coroutine.yield()
					
					if _datas.localMembers.resultCallbacks[_datas.moduleName].aborted == true then
						return
					end
				end
			end
		end
	else -- Recherche AIRBASES
		_datas.objects = {}

		for _, _airbase in pairs(ATMEDatas.airbasesById) do
			local localMembersAirbase = ATMEDatas.localMembers[_airbase:getIndex()]
			local id = localMembersAirbase.objectDCS:getCoalition()
			if _datas.coalitionName == nil or ATMEConst.coalitionNames[id] == _datas.coalitionName then
				local point = localMembersAirbase.objectDCS:getPoint()

				if ATMEFct.isObjectClass(_zone, ATME.C_Zones) == true then
					local localMembersZones = ATMEDatas.localMembers[_zone:getIndex()]

					for _, _zoneItem in ipairs(localMembersZones.zonesList) do
						local localMembersZone = ATMEDatas.localMembers[_zoneItem.zone:getIndex()]

						_datas.zone = _zoneItem.zone
						_datas.actualCenter = ATMEFct.C_Zone_getActualCenter(localMembersZone)
						_datas.actualAzimuth = ATMEFct.C_Zone_getActualAzimuth(localMembersZone)
						
						ATMEFct.C_FindContext_findAirbase(false, point, _airbase, _datas)
						
						if _datas.localMembers.coroutines[_datas.moduleName] ~= nil then
							coroutine.yield()
							
							if _datas.localMembers.resultCallbacks[_datas.moduleName].aborted == true then
								return
							end
						end
					end
				elseif _zone ~= nil then
					local localMembersZone = ATMEDatas.localMembers[_zone:getIndex()]

					_datas.zone = _zone
					_datas.actualCenter = ATMEFct.C_Zone_getActualCenter(localMembersZone)
					_datas.actualAzimuth = ATMEFct.C_Zone_getActualAzimuth(localMembersZone)
					
					ATMEFct.C_FindContext_findAirbase(false, point, _airbase, _datas)

					if _datas.localMembers.coroutines[_datas.moduleName] ~= nil then
						coroutine.yield()
						
						if _datas.localMembers.resultCallbacks[_datas.moduleName].aborted == true then
							return
						end
					end
				else
					ATMEFct.C_FindContext_findAirbase(true, point, _airbase, _datas)
					
					if _datas.localMembers.coroutines[_datas.moduleName] ~= nil then
						coroutine.yield()
						
						if _datas.localMembers.resultCallbacks[_datas.moduleName].aborted == true then
							return
						end
					end
				end
			end
		end
	end
end

ATMEFct.C_FindContext_finalize = function(_datas)
	for _name, _ in pairs(_datas.allGroupIn) do -- allGroupIn n'est rempli que pour GROUPS
		local record
		local status
		if _datas.localMembers.callbacks[_datas.moduleName] ~= nil then
			local datas = {
				zone = _datas.zone,
				object = ATMEDatas.groups[_name],
				unitsName = _datas.allGroupIn[_name],
			}
			status, record = xpcall(function() return _datas.localMembers.callbacks[_datas.moduleName](_datas.findContext, datas) end, function() ATMEFct.errUserHandler('Filter callback') end)
		end
		
		if type(record) ~= "boolean" or record == true then
			_datas.objects[_name] = _group
			table.insert(_datas.objectsIndex, _name)
		end
		
		_datas.allGroupIn[_name] = nil
	end
	
	_datas.nbUnitsGroup = nil
end

ATMEFct.C_FindContext_abortResult = function(_localMembers)
	if _localMembers.resultCallbacks[ATMEDatas.currentModuleName] ~= nil then
		if _localMembers.coroutines[ATMEDatas.currentModuleName] ~= nil then
			if _localMembers.resultCallbacks[ATMEDatas.currentModuleName].idTimer ~= nil then
				timer.removeFunction(_localMembers.resultCallbacks[ATMEDatas.currentModuleName].idTimer)
				_localMembers.resultCallbacks[ATMEDatas.currentModuleName].idTimer = nil
			end
			
			_localMembers.resultCallbacks[ATMEDatas.currentModuleName].aborted = true
			while coroutine.status(_localMembers.coroutines[ATMEDatas.currentModuleName]) ~= "dead" do
				if coroutine.status(_localMembers.coroutines[ATMEDatas.currentModuleName]) == "suspended" then
					coroutine.resume(_localMembers.coroutines[ATMEDatas.currentModuleName])
				end
			end

			_localMembers.coroutines[ATMEDatas.currentModuleName] = nil
			_localMembers.resultCallbacks[ATMEDatas.currentModuleName].aborted = false
		end
	end
end

ATMEFct.C_FindContext_searchResultFct = {
	["dead"] = function(_datas)
		if _datas.localMembers.resultCallbacks[_datas.moduleName] ~= nil then
			ATMEFct.C_FindContext_finalize(_datas)
			
			local datas = {
				objectsIndex = _datas.objectsIndex, 
				objects = _datas.objects,
			}
			
			xpcall(function() _datas.localMembers.resultCallbacks[_datas.moduleName].fct(_datas.findContext, datas) end, function() ATMEFct.errUserHandler('Result callback') end)
		end
		
		_datas.localMembers.coroutines[_datas.moduleName] = nil
		return false
	end,
	["suspended"]= function(_datas)
		local ok = coroutine.resume(_datas.localMembers.coroutines[_datas.moduleName])
		return ok
	end,
	["running"] = function(_datas)
		return true
	end,
}

ATMETimers.C_FindContext_searchResultTimer = function(_datas)
	if ATMEDatas.exitATME == false then
		local ok = ATMEFct.C_FindContext_searchResultFct[coroutine.status(_datas.localMembers.coroutines[_datas.moduleName])](_datas)

		if ok == true then
			return timer.getTime() + 0.01
		end
		
		_datas.localMembers.resultCallbacks[_datas.moduleName].idTimer = nil
	end
end

ATMEDatas.C_FindContext_MT = {
	className = "C_FindContext",
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["getUserPrivateDatas"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["addFilterCallback"] = true,
		["removeFilterCallback"] = true,
		["addResultCallback"] = true,
		["removeResultCallback"] = true,
		["abortResultCallback"] = true,
		["findForAll"] = true,
		["findForCoalition"] = true,
	},
	
	add = function(self, _type)
		local item = {
			type = _type,
			userPrivateDatas = {},
			callbacks = {},
			resultCallbacks = {},
			coroutines = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].callbacks = nil
		ATMEDatas.localMembers[_id].resultCallbacks = nil
		ATMEDatas.localMembers[_id].coroutines = nil
	end,
}

ATME.C_FindContext = {
	getClassName = function(self)
		return ATMEDatas.C_FindContext_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_FindContext_MT.parentClass
	end,
	
	createNamed = function(_name, _findType)
		local fctDefinition = "Function ATME.C_FindContext.createNamed(name, findType)"
		
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
			return
		end
		
		if type(_name) == "string" and type(_findType) == "string" then
			local findType = string.upper(_findType) 
			if ATMEConst.findClasses[findType] ~= nil then
				local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
				
				local err1, newFindContext = ATMEFct.C_FindContext_createNamed(localMembersModule.findContexts, _name, findType)				

				if err1 == true then
					ATMEFct.warningOutputCore(nil, fctDefinition, newFindContext, false)
					return
				end
				
				return newFindContext
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "findType must be 'GROUPS', 'UNITS', 'PLAYERS', 'WEAPONS', 'STATICS', 'CARGOS', 'SCENERY_OBJECTS' or 'AIRBASES'", false)
			end
		else
			if type(_name) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
			elseif type(_findType) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "findType type must be string", false)
			end
		end
	end,
	
	getByName = function(_name)
		if ATMEDatas.startMissionOk < 2 then
			ATMEFct.errorOutputCore(nil, "ATME.C_FindContext.getByName(name)", "Modules loading ... ATME.run() function not called", false)
			return
		end
		
		if type(_name) == "string" then
			return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].findContexts[_name]
		else
			ATMEFct.errorOutputCore(nil, "ATME.C_FindContext.getByName(name)", "name type must be string", false)
		end
	end,
}

ATMEFct.C_FindContext_create = function(_name, _findContexts, _type)
	local newFindContext = {}

	setmetatable(newFindContext, ATMEDatas.C_FindContext_MT.mt)
	
	local name
	local tagsName
	
	if _name ~= nil and _findContexts ~= nil then
		name = _name
		_findContexts[name] = newFindContext
		tagsName = ATMEFct.decompString(_name)
	end

	local index = ATMEDatas.C_FindContext_MT:add(_type)
	local localMembers = ATMEDatas.localMembers[index]
	
	function newFindContext.getName(self)
		return name
	end
	
	function newFindContext.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newFindContext) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newFindContext.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newFindContext.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newFindContext) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newFindContext.getIndex(self)
		return index
	end
	
	function newFindContext.getUserPrivateDatas(self, _moduleName)
		local fctDefinition = ATMEFct.getErrorPrefix(newFindContext) .. "getUserPrivateDatas([moduleName])"
		local moduleName
		
		if _moduleName ~= nil then
			if type(_moduleName) == "string" then
				if localMembers.userPrivateDatas[_moduleName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
					return
				else
					moduleName = _moduleName
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
				return
			end
		else
			moduleName = ATMEDatas.currentModuleName
		end
		
		return localMembers.userPrivateDatas[moduleName]
	end
	
	function newFindContext.addFilterCallback(self, _fct)
		local fctDefinition = ATMEFct.getErrorPrefix(newFindContext) .. "addFilterCallback(fct)"
		if type(_fct) == "function" then
			localMembers.callbacks[ATMEDatas.currentModuleName] = _fct
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
		end
	end
	
	function newFindContext.removeFilterCallback(self)
		localMembers.callbacks[ATMEDatas.currentModuleName] = nil
	end
	
	function newFindContext.addResultCallback(self, _fct)
		local fctDefinition = ATMEFct.getErrorPrefix(newFindContext) .. "addResultCallback(fct)"
		if type(_fct) == "function" then
			localMembers.resultCallbacks[ATMEDatas.currentModuleName] = {
				fct = _fct,
				aborted = false,
			}
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
		end
	end
	
	function newFindContext.removeResultCallback(self)
		ATMEFct.C_FindContext_abortResult(localMembers)
		localMembers.resultCallbacks[ATMEDatas.currentModuleName] = nil
	end
	
	function newFindContext.abortResultCallback(self)
		ATMEFct.C_FindContext_abortResult(localMembers)
	end
	
	function newFindContext.findForAll(self, _zone)
		local fctDefinition = ATMEFct.getErrorPrefix(newFindContext) .. "findForAll([zone])"
		
		if _zone == nil or type(_zone) == "string" or ATMEFct.isObjectClass(_zone, ATME.C_Zones) == true or ATMEFct.isObjectClass(_zone, ATME.C_Zone) == true then
			local datas = {
				objects = {},
				objectsIndex = {},
				nbUnitsGroup = {},
				allGroupIn = {},
				localMembers = localMembers,
				findContext = self,
				moduleName = ATMEDatas.currentModuleName,
			}
			
			local zone = _zone
			if type(zone) == "string" then
				local zoneDCS = ATMEFct.getDCSZone(zone)
		
				if zoneDCS ~= nil then
					zone = ATMEFct.C_Circle_create(nil, nil, zoneDCS.point, zoneDCS.radius)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone " .. _zone .. " unknown", false)
					return
				end
			end
			
			if localMembers.resultCallbacks[ATMEDatas.currentModuleName] == nil then
				ATMEFct.C_FindContext_find(localMembers.type, zone, datas)
				return datas.objectsIndex, datas.objects
			else
				if localMembers.coroutines[ATMEDatas.currentModuleName] == nil then
					ATMEFct.outputCore(nil, "newFindContext.findForAll appel asynchrone", 2)
					localMembers.coroutines[ATMEDatas.currentModuleName] = coroutine.create(ATMEFct.C_FindContext_find)
					coroutine.resume(localMembers.coroutines[ATMEDatas.currentModuleName], localMembers.type, zone, datas)
					
					localMembers.resultCallbacks[ATMEDatas.currentModuleName].idTimer = timer.scheduleFunction(ATMETimers.C_FindContext_searchResultTimer, datas, timer.getTime() + 0.001)
				else
					ATMEFct.setError("ALREADY_STARTED")
					ATMEFct.warningOutputCore(nil, fctDefinition, "search already started", false)
				end
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "zone type must be nil, string, ATME.C_Zones or ATME.C_Zone", true)
		end
	end
	
	function newFindContext.findForCoalition(self, _coalitionName, _zone)
		local fctDefinition = ATMEFct.getErrorPrefix(newFindContext) .. "findForCoalition(coalitionName, [zone])"
		
		if type(_coalitionName) == "string" and 
				(_zone == nil or type(_zone) == "string" or ATMEFct.isObjectClass(_zone, ATME.C_Zones) == true or ATMEFct.isObjectClass(_zone, ATME.C_Zone) == true) then
			
			local coalitionName = string.upper(_coalitionName)
			if ATMEConst.coalitionIds[coalitionName] ~= nil then
				local datas = {
					objects = {},
					objectsIndex = {},
					nbUnitsGroup = {},
					allGroupIn = {},
					coalitionName = coalitionName,
					localMembers = localMembers,
					findContext = self,
					moduleName = ATMEDatas.currentModuleName,
				}
				
				if localMembers.type == "SCENERY_OBJECTS" then
					datas.coalitionName = nil
				end
				
				local zone = _zone
				if type(zone) == "string" then
					local zoneDCS = ATMEFct.getDCSZone(zone)
			
					if zoneDCS ~= nil then
						zone = ATMEFct.C_Circle_create(nil, nil, zoneDCS.point, zoneDCS.radius)
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone " .. _zone .. " unknown", false)
						return
					end
				end
				
				if localMembers.resultCallbacks[ATMEDatas.currentModuleName] == nil then
					ATMEFct.C_FindContext_find(localMembers.type, zone, datas)
					return datas.objectsIndex, datas.objects
				else
					if localMembers.coroutines[ATMEDatas.currentModuleName] == nil then
						ATMEFct.outputCore(nil, "newFindContext.findForCoalition appel asynchrone", 2)
						localMembers.coroutines[ATMEDatas.currentModuleName] = coroutine.create(ATMEFct.C_FindContext_find)
						coroutine.resume(localMembers.coroutines[ATMEDatas.currentModuleName], localMembers.type, zone, datas)
						
						localMembers.resultCallbacks[ATMEDatas.currentModuleName].idTimer = timer.scheduleFunction(ATMETimers.C_FindContext_searchResultTimer, datas, timer.getTime() + 0.001)
					else
						ATMEFct.setError("ALREADY_STARTED")
						ATMEFct.warningOutputCore(nil, fctDefinition, "search already started", false)
					end
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName must be 'RED', 'BLUE' or 'NEUTRAL'", false)
			end
		else
			if type(_coalitionName) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName type must be string", true)
			elseif _zone ~= nil and type(_zone) ~= "string" and ATMEFct.isObjectClass(_zone, ATME.C_Zones) == false and ATMEFct.isObjectClass(_zone, ATME.C_Zone) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "zone type must be nil, string, ATME.C_Zones or ATME.C_Zone", false)
			end
		end
	end

	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return newFindContext
end

setmetatable(ATME.C_FindContext,ATMEDatas.C_FindContext_MT)

ATMEDatas.C_FindContext_MT.__call = function(t, _findType)
	local fctDefinition = "Constructor ATME.C_FindContext(findType)"
	
	if ATMEDatas.startMissionOk < 2 then
		ATMEFct.errorOutputCore(nil, fctDefinition, "Modules loading ... ATME.run() function not called", false)
		return
	end
		
	if type(_findType) == "string" and
			(_findCategories == nil or type(_findCategories) == "string" or type(_findCategories) == "table") then 
		local findType = string.upper(_findType)
		if ATMEConst.findClasses[findType] ~= nil then
			return ATMEFct.C_FindContext_create(nil, nil, findType)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "findType must be 'GROUPS', 'UNITS', 'PLAYERS', 'WEAPONS', 'STATICS', 'CARGOS', 'SCENERY_OBJECTS' or 'AIRBASES'", false)
		end
	else
		if type(_findType) ~= "string" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "findType type must be string", false)
		end
	end
end

ATMEDatas.C_FindContext_MT.mt.__metatable = "Private"
ATMEDatas.C_FindContext_MT.mt.__index = ATME.C_FindContext
ATMEDatas.C_FindContext_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_FindContext_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_FindContext : Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Group : gère les groupes d'unités (AI ou joueur) au format ATME. Pas de contructeur accessible
-- pour les modules externes.
-- Les instances de cette classe sont intégralement gérées par le module ATME_Core. La création et la destruction
-- dépendent du contexte de début de mission et des événements qui se produisent dans DCS en cours de mission.
-- Les modules externes accèdent à cette gestion indirectement par les handlers définis dans ceux-ci et les méthodes
-- accessibles de cette classe. Ceci permet de garantir une cohérence sur les listes internes ATME.
-- ATME crée des instances de cette classe à partir de la création d'instance d'unité joueur ou AI.
-- Chaque unité AI ou joueur dispose d'une variable représentant le groupe ATME qui lui est associé.
-- En cas de destruction (suppression) d'une unité (AI ou joueur), le groupe ne sera détruit qu'après la destruction
-- de la dernière unité (AI ou joueur) du groupe.
----------------------------------------------------------------------------------------------------------

ATMEFct.C_Group_CreateTrackingWaypoints = function(_localMembers)
		
	for _i = 1, #_localMembers.waypoints do
		_localMembers.trackingDatas[_i] = {
			pos = {
				x = _localMembers.waypoints[_i].x,
				y = _localMembers.waypoints[_i].alt,
				z = _localMembers.waypoints[_i].y,
			},
			name = _localMembers.waypoints[_i].name,
			fctRadius = ATMEConst.fctRadius[_localMembers.category],
			action = _localMembers.waypoints[_i].action,
		}
	end
end

ATMETimers.C_Group_timerTrackingWaypoint = function(_group)
	if ATMEDatas.exitATME == false then
		local localMembersGroup = ATMEDatas.localMembers[_group:getIndex()]

		if localMembersGroup.objectDCS:isExist() == true and #localMembersGroup.objectDCS:getUnits() ~= 0 and ATMEDatas.groups[localMembersGroup.objectDCS:getName()] ~= nil then
			if localMembersGroup.disembarking == false then
				local unit = ATMEDatas.units[localMembersGroup.objectDCS:getUnits()[1]:getName()]	
				local localMembersUnit = ATMEDatas.localMembers[unit:getIndex()]
				
				if localMembersUnit.objectDCS:isExist() == true then		-- Sur changement de slot, l'unité est détruite puis recréée par DCS, 
																			-- il est possible d'avoir nil ponctuellement sur le timer qui continue à se dérouler si
																			-- l'unité sur lequel le changement a lieu est la première. (lié au déclenchement du timer).
				
					local pos = localMembersUnit.objectDCS:getPoint()
					localMembersGroup.lastPosition = pos
	
					local inZoneWP = false
					local dx = pos.x - localMembersGroup.trackingDatas[localMembersGroup.waypointId].pos.x
					local dz = pos.z - localMembersGroup.trackingDatas[localMembersGroup.waypointId].pos.z
					local d2 = dx * dx +  dz * dz
					local radius = localMembersGroup.trackingDatas[localMembersGroup.waypointId].fctRadius(localMembersGroup.trackingDatas[localMembersGroup.waypointId].action, localMembersUnit.objectDCS:getVelocity(), localMembersUnit.objectDCS:inAir())
					
					if d2 < radius * radius then
						if localMembersGroup.lastD2 ~= nil and d2 >= localMembersGroup.lastD2 then
							inZoneWP = true
							localMembersGroup.lastD2 = nil
						else
							localMembersGroup.lastD2 = d2
						end
					end
					
					if inZoneWP == true then
						ATMEFct.outputCore(nil, "Group " .. localMembersGroup.name .." => Route courante - Passage WP " .. localMembersGroup.waypointId - 1 .. "/" .. #localMembersGroup.waypoints - 1, 2)
						if localMembersGroup.waypointId > localMembersGroup.offsetWaypoint then
							-- Passage au waypoint
							-- Préparation de l'event core et WP suivante	
							local id = localMembersGroup.waypointId
							local datas = {
								waypointID = id + localMembersGroup.deltaWaypointId - 1,
								lastWaypointID = localMembersGroup.nbInitialWaypoints - 1,
								point = {
									x = localMembersGroup.trackingDatas[id].pos.x,
									y = localMembersGroup.trackingDatas[id].pos.y,
									z = localMembersGroup.trackingDatas[id].pos.z,
								},
								name = localMembersGroup.trackingDatas[id].name,
							}
							
							if localMembersGroup.offsetWaypoint > 0 then
								ATMEFct.outputCore(nil, "              -> Groupe sortant d'un véhicule, Route originale " .. datas.waypointID .. "/" .. localMembersGroup.nbInitialWaypoints - 1, 2)
							end
							
							localMembersGroup.lastPassedWaypoint = localMembersGroup.waypointId
							localMembersGroup.waypointId = localMembersGroup.waypointId + 1
							for _moduleName, _callbacks in pairs(localMembersGroup.trackingCallbacks) do
								ATMEDatas.currentModuleName = _moduleName
								
								if _callbacks[id] ~= nil then
									ATMEFct.outputCore(nil, "Appel callback Tracking waypoint du module " .. _moduleName .. " --> Group ".. localMembersGroup.name, 2)
									
									xpcall(function() return _callbacks[id](_group, datas) end, function() ATMEFct.errUserHandler('Tracking callback') end)
								end
							end
						else
							ATMEFct.outputCore(nil, "Group " .. localMembersGroup.name .." => Pas de traitement sur ce WP", 2)
							localMembersGroup.waypointId = localMembersGroup.waypointId + 1
						end

						if localMembersGroup.trackingDatas == nil or localMembersGroup.waypointId > #localMembersGroup.trackingDatas then
							localMembersGroup.endOfRoute = true
							localMembersGroup.idTimer = nil
							ATMEFct.outputCore(nil,  "Group " .. localMembersGroup.name .." => fin de tracking waypoint", 2)
							return
						end
					end
					
					return timer.getTime() + 2
				else
					-- La première unité n'a pas été trouvée suite à changement de slot, on réarme
					
					return timer.getTime() + 2
				end
			else
				ATMEFct.outputCore(nil, "Group ".. localMembersGroup.name .. " débarquement en cours", 2)
				return timer.getTime() + 2
			end
		else
			localMembersGroup.idTimer = nil
			ATMEFct.outputCore(nil, "Group " .. localMembersGroup.name .." => inexistant, plus de tracking", 2)
		end
	else
		local localMembersGroup = ATMEDatas.localMembers[_group:getIndex()]
		localMembersGroup.idTimer = nil
	end
end

ATMEFct.C_Group_disable = function(_group, _reserved, _flag)
	local localMembers = ATMEDatas.localMembers[_group:getIndex()]
	localMembers.isEnabled = false
	local flag = _flag
	
	if flag == nil then
		flag = "DISABLED"
	end
	
	
	for _name, _unit in pairs(localMembers.ATMEUnits) do
		ATMEDatas.flagDeleteUnit[_name] = flag
		ATMEFct.C_Unit_disable(_unit, _reserved)
	end
end


ATMETimers.C_Group_autoRespawn = function(_spawnDatas)
	if ATMEDatas.exitATME == false then
		-- Gestion du respawn automatique après destruction
		
		ATMEFct.outputCore(nil, "C_Group_autoRespawn pour " .. _spawnDatas.datas.name, 2)
		ATMEFct.C_GroupCloningContext_finalizeSpawn(_spawnDatas, false)
		ATMEDatas.groupsToRespawn[_spawnDatas.datas.name] = nil
	end
end

ATMETimers.C_Group_newRouteSet = function(_datas)
	if ATMEDatas.exitATME == false then
		if _datas.localMembers.objectDCS:isExist() == true and #_datas.localMembers.objectDCS:getUnits() ~= 0 and ATMEDatas.groups[_datas.localMembers.objectDCS:getName()] ~= nil then
			-- Gestion appel callback sur setRoute
			-- Si le setRoute est fait dès le début de mission avant l'init newPreparedGroups dans le timer principal
			if _datas.localMembers.initOk == true then
				_datas.localMembers.waypointId = 2
				_datas.localMembers.lastPassedWaypoint = 1
			else
				_datas.localMembers.waypointId = 1
				_datas.localMembers.lastPassedWaypoint = 0
			end
			
			if _datas.localMembers.idTimer == nil then
				_datas.localMembers.idTimer = timer.scheduleFunction(ATMETimers.C_Group_timerTrackingWaypoint, _datas.group, timer.getTime() + 0.5)
			end
			
			local temp = ATMEDatas.currentModuleName
			for _moduleName, _callback in pairs(_datas.localMembers.newRouteCallback) do
				ATMEDatas.currentModuleName = _moduleName
					
				ATMEFct.outputCore(nil,"Appel de newRouteCallback du groupe " .. _datas.localMembers.name .. " pour le module " .. _moduleName, 2)
				xpcall(function() _callback(ATMEDatas.groups[_datas.localMembers.name]) end, function() ATMEFct.errUserHandler('newRouteCallback du groupe ' .. _name) end)
			end
			
			ATMEDatas.currentModuleName = temp
		end
	end
end


ATMEFct.updateNbPlayers = function(_inc, _localMembersGroup)
	ATMEDatas.nbPlayers = ATMEDatas.nbPlayers + _inc
	ATMEFct.outputCore(nil,"Global Nb players = " .. ATMEDatas.nbPlayers, 2)
	
	_localMembersGroup.nbPlayers = _localMembersGroup.nbPlayers + _inc
	ATMEFct.outputCore(nil,"Groupe " .. _localMembersGroup.name .. " -> Nb players = " .. _localMembersGroup.nbPlayers, 2)
end

ATMEFct.C_Group_addPlayer = function(_group)
	local localMembers = ATMEDatas.localMembers[_group:getIndex()]

	if localMembers.itemsMenuCommandRoot == nil then
		localMembers.itemsMenuCommandRoot = {}
		localMembers.dataSharings = {}
		
		for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
			_, localMembers.itemsMenuCommandRoot[_localModuleMembers.name] = ATMEFct.C_Menu_create(_group, nil, nil, _localModuleMembers.name)
			_, localMembers.dataSharings[_localModuleMembers.name] = ATMEFct.C_DataSharing_create(nil, nil, "groups", ATMEDatas.localMembers[_group:getIndex()], nil, _localModuleMembers.name)
		end
	end

	ATMEFct.updateNbPlayers(1, localMembers)
end

ATMEFct.C_Group_deletePlayer = function(_localMembersGroup)
	if _localMembersGroup.itemsMenuCommandRoot ~= nil and _localMembersGroup.nbPlayers == 1 then
		for _k, _itemMenuCommandRoot in pairs(_localMembersGroup.itemsMenuCommandRoot) do
			_itemMenuCommandRoot:removeAll(true)
			ATMEDatas.C_Menu_MT:remove(_itemMenuCommandRoot:getIndex())
			_localMembersGroup.itemsMenuCommandRoot[_k] = nil
		end
		
		_localMembersGroup.itemsMenuCommandRoot = nil
		_localMembersGroup.dataSharings = nil
	end
	
	ATMEFct.updateNbPlayers(-1, _localMembersGroup)
end

ATMEDatas.C_Group_MT = {
	className = "C_Group",
	
	mtModules = {},
	mt = {},
	
	keys = {
		["disable"] = true,
		["isDisabled"] = true,
		["getNbPlayers"] = true,
		["getFirstUnit"] = true,
		["getNextUnit"] = true,
		["getPreviousUnit"] = true,
		["getLastUnit"] = true,
		["getUnitByName"] = true,
		["getUnits"] = true,
		["getDCSID"] = true,
		["getNbUnits"] = true,
		["getCoalitionName"] = true,
		["getCountryName"] = true,
		["isOnlyInfantry"] = true,
		["isInZones"] = true,
		["isInDCSZone"] = true,
		["isInZone"] = true,
		["getCategoryName"] = true,
		["isReadyToBoard"] = true,
		["setReadyToBoard"] = true,
		["resetReadyToBoard"]= true,
		["isBoarding"] = true,
		["move"] = true,
		["stop"] = true,
		["freeze"] = true,
		["getDCSObject"] = true,
		["respawn"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["getIndex"] = true,
		["activate"] = true,
		["isActivated"] = true,
		["isAlwaysVisible"] = true,
		["setRoute"] = true,
		["setTask"] = true,
		["resetTask"] = true,
		["pushTask"] = true,
		["popTask"] = true,
		["getMinDistance"] = true,
		["getMaxDistance"] = true, 
		["getMinHDistance"] = true,
		["getMaxHDistance"] = true,
		["getPosition"] = true,
		["modules"] = true,
		["start"] = true,
		["isControlled"] = true,
		["canFly"] = true,
		["getMenuRoot"] = true,
		["getUserPrivateDatas"] = true,
		["getDataSharing"] = true,
		["setInvisible"] = true,
		["setImmortal"] = true,
		["switchWaypoint"] = true,
		["setOption"] = true,
		["addTrackingCallback"] = true,
		["removeTrackingCallback"] = true,
		["addNewRouteCallback"] = true,
		["removeNewRouteCallback"] = true,
		["addReadyToBoardCallback"] = true,
		["removeReadyToBoardCallback"] = true,
		["getWaypoint"] = true,
		["getNextWaypoint"] = true,
		["getLastWaypoint"] = true,
		["radarOn"] = true,
		["radarOff"] = true,
	},
	
	add = function(self, _name, _group, _datas, _country)
		local delay = ATMEConst.basicRouteDelay

		local item = {
			name = _name,
			objectDCS = _group,
			initOk = false,
			objectDCSId = _group:getID(),
			alwaysVisible = _datas.visible,
			trackingDatas = {},							-- Données pour le tracking waypoints
			trackingCallbacks = {}, 					-- Liste des callbacks pour les waypoints trackés									
			maxWaypoints = 0, 							-- Nombre de waypoints tenant compte de l'ajustement si le dernier WP
														-- n'est pas de type "Land" pour les aéronefs.
			waypointId = 1,
			offsetWaypoint = 0,							-- Offset de 2 si débarquement, 0 sinon
			nbInitialWaypoints = 0,						-- Indique le nombre de WP initiaux (utile pour le load/unload de troupes)
			deltaWaypointId = 0, 						-- Indique l'écart d'ID entre la première route et la route actuelle pour troupes load/unload
			endOfRoute = false,							-- Route en cours ... Utilisé pour le suivi des waypoint sur modif de route
			lastPassedWaypoint = 0,
			disembarking = false,						-- True si débarquement en cours, permet de synchroniser les événements
			groupId = _group:getID(),
			nbPlayers = 0,
			onBoard = false,
			justActivated = 0,
			autoRespawn = {
				activated = false,
				time = 0,
				random = false,
				minTime = 0,
				maxTime = 0,
			},
			coalition = ATMEConst.coalitionNames[_group:getCoalition()],
			country = country.name[_country],
			category = ATMEConst.categoryNames[_group:getCategory()],
			routeDelay = delay,
			routeUncontrolled = _datas.uncontrolled,
			onlyInfantry = true,
			boardingStarted = false,
			readyToBoardSet = false,
			isEnabled = true,
			destroyed = false,
			nbUnitsToEnable = 0,
			ATMEUnits = {},
			ATMEUnitNames = ATME.C_List(),
			userPrivateDatas = {},						-- Point d'entrée des données privées : un par module
			newRouteCallback = {},
			tagsName = ATMEFct.decompString(_name),
			spawnDatas = ATMEFct.copyTable(_datas.spawnDatas),	-- On copie l'original cause modfié
			initialSpawnDatas = _datas.spawnDatas,				-- On conserve l'original
			readyToBoardCallback = {},
		}

		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].ATMEUnits = nil
		ATMEDatas.localMembers[_id].ATMEUnitNames = nil
		ATMEDatas.localMembers[_id].trackingDatas = nil
		ATMEDatas.localMembers[_id].trackingCallbacks = nil
		ATMEDatas.localMembers[_id].newRouteCallback = nil
		ATMEDatas.localMembers[_id].readyToBoardCallback = nil
	end,
}

ATME.C_Group = {
	getClassName = function(self)
		return ATMEDatas.C_Group_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Group_MT.parentClass
	end,
	
	getByName = function(_groupName)
		if type(_groupName) == "string" then
			return ATMEDatas.groups[_groupName]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Group.getByName(name)", "name type must be string", false)
		end
	end,
	
	findReadyToBoardGroupsForAll = function()
		local groupsReady = {}
		local nbGroups = 0
		for _, _group in pairs(ATMEDatas.readyToBoardGroups) do
			groupsReady[_group:getName()] = _group
			nbGroups = nbGroups + 1
		end
		
		return groupsReady, nbGroups
	end,
	
	findReadyToBoardGroupsForCoalition = function(_coalitionName)
		local fctDefinition = "Function ATME.C_Group.findReadyToBoardGroupsForCoalition(coalitionName)"
		
		if type(_coalitionName) == "string" then
			local coalitionName = string.upper(_coalitionName)
			if ATMEConst.coalitionIdsBlueRed[coalitionName] ~= nil then
				local groupsReady = {}
				local nbGroups = 0
				for _, _group in pairs(ATMEDatas.readyToBoardGroups) do
					local localMembersGroup = ATMEDatas.localMembers[_group:getIndex()]
					if localMembersGroup.coalition == coalitionName then
						groupsReady[localMembersGroup.name] = _group
						nbGroups = nbGroups + 1
					end
				end
				
				return groupsReady, nbGroups
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName must be 'BLUE' or 'RED'", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName type must be string", false)
		end
	end,
	foreach = function(_callback, _datas)
		local fctDefinition = "Function ATME.C_Group.foreach(callback, [datas])"
		
		if type(_callback) == "function" then
			ATMEFct.foreachObject(nil, "groups", _callback, _datas)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "callback type must be function", false)
		end
	end,
	foreachInCoalition = function(_coalitionName, _callback, _datas)
		local fctDefinition = "Function ATME.C_Group.foreachInCoalition(coalitionName, callback, [datas])"
		
		if type(_coalitionName) == "string" and type(_callback) == "function" then
			if ATMEConst.coalitionIdsBlueRed[_coalitionName] == nil then
				ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName must be 'BLUE' or 'RED'", false)
				return
			end
			
			ATMEFct.foreachObject(_coalitionName, "groups", _callback, _datas)
		else
			if type(_coalitionName) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName type must be string", false)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "callback type must be function", false)
			end
		end
	end,
}


ATMEFct.C_Group_addUnit = function(_group, _unit)
	local unitName = _unit:getName()

	local localMembers = ATMEDatas.localMembers[_group:getIndex()]
	local localMembersUnit = ATMEDatas.localMembers[_unit:getIndex()]
	localMembersUnit.num = localMembers.ATMEUnitNames:add(unitName)

	localMembers.ATMEUnits[unitName] = _unit

	if localMembers.onlyInfantry == true and localMembersUnit.objectDCS:hasAttribute("Infantry") == false then
		localMembers.onlyInfantry = false
	end
end


ATMEFct.C_Group_deleteUnit = function(_group, _unitName, _reserved)
	local localMembers = ATMEDatas.localMembers[_group:getIndex()]
	local localMembersUnit = ATMEDatas.localMembers[localMembers.ATMEUnits[_unitName]:getIndex()]
	
	localMembers.ATMEUnitNames:remove(localMembersUnit.num)
	localMembers.ATMEUnits[_unitName] = nil

	if localMembers.ATMEUnitNames:getCount() == 0 then
		ATMEFct.C_Group_delete(_group, _reserved)
	end
end

ATMEFct.C_Group_evalTypeDistance = {
	["MIN_H"] = function(_pA, _pB, _value)
		local d2 = (_pA.x - _pB.x) * (_pA.x - _pB.x) + (_pA.z - _pB.z) * (_pA.z - _pB.z)
		if _value == nil or _value > d2 then
			return d2
		else
			return _value
		end
	end,
	["MAX_H"] = function(_pA, _pB, _value)
		local d2 = (_pA.x - _pB.x) * (_pA.x - _pB.x) + (_pA.z - _pB.z) * (_pA.z - _pB.z)
		if _value == nil or _value < d2 then
			return d2
		else
			return _value
		end
	end,
	["MIN"] = function(_pA, _pB, _value)
		local d2 = (_pA.x - _pB.x) * (_pA.x - _pB.x) + (_pA.z - _pB.z) * (_pA.z - _pB.z) + (_pA.y - _pB.y) * (_pA.y - _pB.y)
		if _value == nil or _value > d2 then
			return d2
		else
			return _value
		end
	end,
	["MAX"] = function(_pA, _pB, _value)
		local d2 = (_pA.x - _pB.x) * (_pA.x - _pB.x) + (_pA.z - _pB.z) * (_pA.z - _pB.z) + (_pA.y - _pB.y) * (_pA.y - _pB.y)
		if _value == nil or _value < d2 then
			return d2
		else
			return _value
		end
	end,
}

ATMEFct.C_Group_evalDistances = function(_group, _point, _typeTest)
	local value
	local units = _group:getUnits()
	local localMembers = ATMEDatas.localMembers[_group:getIndex()]
	
	for _, _unit in pairs(units) do
		local localMembersUnit = ATMEDatas.localMembers[_unit:getIndex()]
		value = ATMEFct.C_Group_evalTypeDistance[_typeTest](_point, localMembersUnit.objectDCS:getPoint(), value)
	end
	
	if value == nil then
		value = ATMEFct.C_Group_evalTypeDistance[_typeTest](_point, localMembers.lastPosition, value)
	end
	
	return math.sqrt(value)
end


ATMEFct.C_Group_changeReadyToBoard = function(_group, _value, _nextWaypoint)
	local localMembers = ATMEDatas.localMembers[_group:getIndex()]

	if _value == true and ATMEDatas.readyToBoardGroups[localMembers.name] == nil then
		if _nextWaypoint == nil then
			localMembers.nextWaypoint = nil
		else
			local nextWaypoint = _nextWaypoint
			if nextWaypoint - localMembers.deltaWaypointId <= localMembers.lastPassedWaypoint then
				if localMembers.lastPassedWaypoint + localMembers.deltaWaypointId < localMembers.nbInitialWaypoints then
					localMembers.nextWaypoint = localMembers.lastPassedWaypoint + localMembers.deltaWaypointId + 1
					ATMEFct.outputCore(nil, "Group '" .. localMembers.name  .. "' next waypoint for unload : Already passed", 0)
				else
					localMembers.nextWaypoint = nil
				end
			elseif nextWaypoint <= localMembers.nbInitialWaypoints then
				localMembers.nextWaypoint = nextWaypoint
			else
				localMembers.nextWaypoint = nil
			end
		end
		
		ATMEDatas.readyToBoardGroups[localMembers.name] = _group
		trigger.action.groupStopMoving(localMembers.objectDCS)
		
		local temp = ATMEDatas.currentModuleName
		for _moduleName, _callback in pairs(localMembers.readyToBoardCallback) do
			ATMEDatas.currentModuleName = _moduleName
				
			ATMEFct.outputCore(nil,"Appel de readyToBoardCallback du groupe " .. localMembers.name .. " pour le module " .. _moduleName, 2)
			xpcall(function() _callback(_group) end, function() ATMEFct.errUserHandler('readyToBoardCallback du groupe ' .. localMembers.name) end)
		end
		
		ATMEDatas.currentModuleName = temp
	elseif _value == false and localMembers.boardingStarted == false and ATMEDatas.readyToBoardGroups[localMembers.name] ~= nil then
		trigger.action.groupContinueMoving(localMembers.objectDCS)
		ATMEDatas.readyToBoardGroups[localMembers.name] = nil
		localMembers.nextWaypoint = nil
		localMembers.pickupZone = nil
	end	
end

ATMEFct.C_Group_finalizeBoarding = function(_localMembersTransportUnit, _groupToBoard, _transportUnit)
	_localMembersTransportUnit.nbUnitsOnBoard = _localMembersTransportUnit.nbUnitsOnBoard + _groupToBoard.nbUnitsOnBoard
	
	if ATMEConst.canFly[ATMEDatas.localMembers[_localMembersTransportUnit.group:getIndex()].category] == true then
		trigger.action.setUnitInternalCargo(_localMembersTransportUnit.name, _localMembersTransportUnit.nbUnitsOnBoard * 100)
		
		ATMEFct.outputCore(nil, "Poids chargé = " .. _localMembersTransportUnit.nbUnitsOnBoard * 100 .. " pour unité transport " .. _localMembersTransportUnit.name, 2)
	end
	
	local index = 0
	for _i, _groupItem in ipairs(_localMembersTransportUnit.groupsBoarding) do
		if _groupToBoard.localMembers.name == _groupItem.groupName then
			index = _i
		end
	end
	
	ATMEDatas.reservedBoardingGroup[_groupToBoard.localMembers.name] = nil
	
	if index > 0 then
		table.remove(_localMembersTransportUnit.groupsBoarding, index)
	end
	
	
	-- TODO -> Ajuster les valeurs passées
	local idGroupOnBoard = ATMEFct.C_Unit_addOnBoardGroup(_transportUnit,  
																	{
																		localMembers = _groupToBoard.localMembers,
																		unitsOnBoard = _groupToBoard.unitsOnBoard,
																		trackingCallbacks = _groupToBoard.trackingCallbacks,
																		userPrivateDatas = _groupToBoard.userPrivateDatas,
																		nbDeadUnits = 0,
																		nbUnitsOnBoard = _groupToBoard.nbUnitsOnBoard,
																		newRouteCallback = _groupToBoard.newRouteCallback,
																		readyToBoardCallback = _groupToBoard.readyToBoardCallback,
																	})

	-- Evénement de fin d'embarquement
	-- On ne stocke que le nom. La vérification de l'existance de l'unité ne se fera que lors
	-- de l'activation de l'événement. 
	-- ATTENTION : Ici name est le nom du group qui va être disable... Donc pas de réactivation
	-- à l'activation de l'événement.

	local event = {
		type = "TRANSPORT",
		datas = {
			action = "LOADED",
			unitName = _localMembersTransportUnit.name,
			index = idGroupOnBoard,
			onBoardGroupName = _groupToBoard.localMembers.name,
			nbOnBoardUnits = _groupToBoard.nbUnitsOnBoard,
			nbDeadUnits = _groupToBoard.nbUnitsLost,
		},
	}
	
	ATMEFct.fireEvent(event)
	
	_localMembersTransportUnit.nbBoardings = _localMembersTransportUnit.nbBoardings - 1
	
	if _localMembersTransportUnit.nbBoardings == 0 then
		local event = {
			type = "TRANSPORT",
			datas = {
				action = "END_BOARDING",
				unitName = _localMembersTransportUnit.name,
			},
		}
		
		ATMEFct.fireEvent(event)
	end
end

ATMETimers.C_Group_startLoad_timer = function(_groupToBoard)
	if ATMEDatas.exitATME == false then
		local transportUnit = ATMEDatas.units[_groupToBoard.transportUnitName]
		local transportExist = false
		
		if transportUnit ~= nil and ATMEDatas.localMembers[transportUnit:getIndex()].objectDCS:isExist()then
			local localMembersTransportUnit = ATMEDatas.localMembers[transportUnit:getIndex()]
			local toBoardGroup = ATME.C_Group.getByName(_groupToBoard.localMembers.name)
			local vector = ATMEFct.vectors["3D"].createFromTable(nil, nil, localMembersTransportUnit.objectDCS:getVelocity())
			
			if toBoardGroup ~= nil then
				local localMembersGroup = ATMEDatas.localMembers[toBoardGroup:getIndex()]
				if _groupToBoard.boardingStopped == false then
					local localMembersTransportGroup = ATMEDatas.localMembers[localMembersTransportUnit.group:getIndex()]
					if (localMembersTransportUnit.objectDCS:hasAttribute("Ground vehicles") == true or (ATMEConst.canFly[localMembersTransportGroup.category] == true and localMembersTransportUnit.objectDCS:inAir() == true)) and
									vector:getModule() > 3 then
									
						_groupToBoard.delayBeforeStopped = _groupToBoard.delayBeforeStopped + 1
					
						if _groupToBoard.delayBeforeStopped == 40 then -- 40 cycles soit 2s
							_groupToBoard.boardingStopped = true
							_groupToBoard.currentUnitName = localMembersGroup.ATMEUnitNames:getFirst()
						end
					else
						_groupToBoard.delayBeforeStopped = 0
					end
				end
				
				-- Vitesse des infanteries à 5 m/s soit un temps de parcours = dist/5
				-- Marge de sécurité à 10s en plus. Au delà, tout le groupe est chargé directement
				-- Ceci permet de contourner le problème d'une unité capricieuse qui louperait le transport et le rayon de chargement.
					
				if _groupToBoard.boardingStopped == false then
					if ATMEDatas.units[_groupToBoard.currentUnitName] ~= nil then
						local unit = ATMEDatas.units[_groupToBoard.currentUnitName]
						local localMembersUnit = ATMEDatas.localMembers[unit:getIndex()]
						local disabled = false
						
						if _groupToBoard.boardingStart == false then
							local item = {
								groupName = _groupToBoard.localMembers.name,
								nbUnits = localMembersGroup.ATMEUnitNames:getCount(),
							}

							table.insert(localMembersTransportUnit.groupsBoarding, item)
							_groupToBoard.boardingStart = true
						end

						if timer.getTime() < _groupToBoard.maxDelay then
							local v = ATMEFct.vectors["3D"].createFromReference(nil, nil, localMembersUnit.objectDCS:getPoint(), localMembersTransportUnit.objectDCS:getPoint())
							local ok = ATMEFct.whichSide(_groupToBoard.A, _groupToBoard.B, localMembersUnit.objectDCS:getPoint()) == _groupToBoard.sideOfWP2

							if v:getHModule() < ATMEConst.personnelCarrierAttributes[localMembersTransportUnit.typeName].minDistance or ok == true then
								-- tant que l'unité de transport est au sol, les unités à transporter
								-- sont désactivées à l'approche de l'unité de transport. Distance paramétrée dans ATMEConst.personnelCarrierAttributes
								_groupToBoard.unitsOnBoard[localMembersUnit.name] = true
								_groupToBoard.nbUnitsOnBoard = _groupToBoard.nbUnitsOnBoard + 1
								ATMEFct.C_Unit_disable(unit, true)
								disabled = true
							end
						else
							_groupToBoard.unitsOnBoard[localMembersUnit.name] = true
							_groupToBoard.nbUnitsOnBoard = _groupToBoard.nbUnitsOnBoard + 1
							
							ATMEFct.C_Unit_disable(unit, true)
							disabled = true
						end

						if localMembersGroup.ATMEUnitNames == nil then
							localMembersGroup.onBoard = true
							ATMEFct.C_Group_finalizeBoarding(localMembersTransportUnit, _groupToBoard, transportUnit)
							
							_groupToBoard.unitsOnBoard = nil
						else
							if disabled == false then
								_groupToBoard.currentUnitName = localMembersGroup.ATMEUnitNames:getNext()
								
								if _groupToBoard.currentUnitName == nil then
									_groupToBoard.currentUnitName = localMembersGroup.ATMEUnitNames:getFirst()
								end
							else
								_groupToBoard.currentUnitName = localMembersGroup.ATMEUnitNames:getFirst()
							end
							
							return timer.getTime() + 0.5
						end
					else
						localMembersTransportUnit.nbBoardings = localMembersTransportUnit.nbBoardings - 1
					end
				else
					if _groupToBoard.nbUnitsOnBoard > 0 then
						local disabled = false
						
						-- Si l'unité de transport décolle ou se déplace plus vite que 3m/s, tout le groupe est disable
						-- Embarquement en urgence ! des pertes sont possibles
					
						if ATMEDatas.units[_groupToBoard.currentUnitName] ~= nil then
							local unit = ATMEDatas.units[_groupToBoard.currentUnitName]
							local localMembersUnit = ATMEDatas.localMembers[unit:getIndex()]
							
							if math.random(1,10) <= 7 - _groupToBoard.alea then
								_groupToBoard.unitsOnBoard[localMembersUnit.name] = true
								_groupToBoard.nbUnitsOnBoard = _groupToBoard.nbUnitsOnBoard + 1
								ATMEFct.C_Unit_disable(unit, true)
							else
								-- Unité détruite définitivement
								_groupToBoard.nbUnitsLost = _groupToBoard.nbUnitsLost + 1
								ATMEFct.C_Unit_disable(unit, false)
							end
							
							disabled = true
							
							if _groupToBoard.alea < 7 then
								_groupToBoard.alea = _groupToBoard.alea + 1
							end
						end
						
						if localMembersGroup.ATMEUnitNames == nil then 
							localMembersGroup.onBoard = true
							ATMEFct.C_Group_finalizeBoarding(localMembersTransportUnit, _groupToBoard, transportUnit)
							
							_groupToBoard.unitsOnBoard = nil
						else
							if disabled == false then
								_groupToBoard.currentUnitName = localMembersGroup.ATMEUnitNames:getNext()
								
								if _groupToBoard.currentUnitName == nil then
									_groupToBoard.currentUnitName = localMembersGroup.ATMEUnitNames:getFirst()
								end
							else
								_groupToBoard.currentUnitName = localMembersGroup.ATMEUnitNames:getFirst()
							end
							
							return timer.getTime() + 0.5
						end	
					else
						toBoardGroup:stop()
						localMembersGroup.boardingStarted = false
						localMembersTransportUnit.nbBoardings = localMembersTransportUnit.nbBoardings - 1
					end
				end
			else
				if _groupToBoard.nbUnitsOnBoard > 0 then
					ATMEFct.C_Group_finalizeBoarding(localMembersTransportUnit, _groupToBoard, transportUnit)
				else
					localMembersTransportUnit.nbBoardings = localMembersTransportUnit.nbBoardings - 1
				end
				
				_groupToBoard.unitsOnBoard = nil
			end
		else
			-- L'unité de transport est inexistante (détruit ou changement de slot)
			-- Tout est perdu. Seules restent les unités du groupe extérieures à l'unité de transport.
			
			_groupToBoard.unitsOnBoard = nil
			ATMEDatas.reservedBoardingGroup[_groupToBoard.localMembers.name] = nil
			local toBoardGroup = ATME.C_Group.getByName(_groupToBoard.localMembers.name)
			if toBoardGroup ~= nil then
				_groupToBoard.localMembers.boardingStarted = false
			end
		end
	end
end

ATMEFct.getFirstUnit = function(_group)
	local localMembers = ATMEDatas.localMembers[_group:getIndex()]

	return ATMEDatas.units[localMembers.objectDCS:getUnits()[1]:getName()]
end

ATMEFct.isGroupActivated = function(_group)
	local unit = ATMEFct.getFirstUnit(_group)

	if unit ~= nil then
		local localMembers = ATMEDatas.localMembers[unit:getIndex()]
		return localMembers.objectDCS:isActive()
	else
		return false
	end
end

ATMEFct.getGroupPosition = function(_group)
	local localMembers = ATMEDatas.localMembers[_group:getIndex()]
	
	if localMembers.objectDCS:isExist() and #localMembers.objectDCS:getUnits() ~= 0 then
		local unit = ATMEFct.getFirstUnit(_group)	
				
		if unit ~= nil then
			return unit:getPosition()
		else
			return localMembers.lastPosition
		end
	else
		return localMembers.lastPosition
	end
end

ATMEFct.C_Group_create = function(_ATMEUnit, _datas, _newObject)
	local newGroup = {}
	setmetatable(newGroup, ATMEDatas.C_Group_MT.mt)

	local index = ATMEDatas.C_Group_MT:add(_ATMEUnit:getDCSObject():getGroup():getName(), _ATMEUnit:getDCSObject():getGroup(), _datas, _ATMEUnit:getDCSObject():getCountry())
	local localMembers = ATMEDatas.localMembers[index]
	
	
	function newGroup.getIndex(self)
		return index
	end
		
	function newGroup.getName(self)
		return localMembers.name
	end
	
	function newGroup.getDCSID(self)
		return localMembers.objectDCSId
	end
	
	function newGroup.getNbPlayers(self)
		return localMembers.nbPlayers
	end
	
	function newGroup.getMenuRoot(self)
		if localMembers.destroyed == false then
			if localMembers.nbPlayers > 0 then
				return localMembers.itemsMenuCommandRoot[ATMEDatas.currentModuleName]
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newGroup) .. "getMenuRoot()", "already disabled", false)
		end
	end
	
	function newGroup.getDataSharing(self)
		if localMembers.destroyed == false then
			if localMembers.nbPlayers > 0 then
				return localMembers.dataSharings[ATMEDatas.currentModuleName]
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newGroup) .. "getMenuRoot()", "already disabled", false)
		end
	end
	
	function newGroup.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if localMembers.tagsName ~= nil and pos >= 1 and pos <= #localMembers.tagsName then
				return localMembers.tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newGroup.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newGroup.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if localMembers.tagsName ~= nil then
				local ok = ATMEFct.compareTags(localMembers.tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newGroup.isAlwaysVisible(self)
		if localMembers.destroyed == false then
			return localMembers.alwaysVisible
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newGroup) .. "isAlwaysVisible()", "already disabled", false)
		end
	end
	
	function newGroup.getCategoryName(self)
		if localMembers.destroyed == false then
			return localMembers.category
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newGroup) .. "getCategoryName()", "already disabled", false)
		end
	end
	
	function newGroup.getCoalitionName(self)
		if localMembers.destroyed == false then
			return localMembers.coalition
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newGroup) .. "getCoalitionName()", "already disabled", false)
		end
	end
	
	function newGroup.getCountryName(self)
		if localMembers.destroyed == false then
			return localMembers.country
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newGroup) .. "getCoalitionName()", "already disabled", false)
		end
	end
	
	function newGroup.getDCSObject(self)
		return localMembers.objectDCS
	end
	
	function newGroup.activate(self)
		if localMembers.destroyed == false then
			if ATMEFct.isGroupActivated(self) == false then
				localMembers.objectDCS:activate()
				localMembers.justActivated = 3
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newGroup) .. "activate()", "already disabled", false)
		end
	end
	
	function newGroup.isActivated(self)
		-- ATTENTION fonction DCS sur l'unité et non sur le groupe
		-- Risque de BUG DCS car pas homogène
		
		if localMembers.destroyed == false then
			return ATMEFct.isGroupActivated(self)
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newGroup) .. "isActivated()", "already disabled", false)
		end
	end
	
	function newGroup.getUserPrivateDatas(self, _moduleName)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "getUserPrivateDatas([moduleName])"
		local moduleName
		
		if _moduleName ~= nil then
			if type(_moduleName) == "string" then
				if localMembers.userPrivateDatas[_moduleName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
					return
				else
					moduleName = _moduleName
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
				return
			end
		else
			moduleName = ATMEDatas.currentModuleName
		end
		
		return localMembers.userPrivateDatas[moduleName]
	end
	
	function newGroup.getWaypoint(self, _waypointId)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "getWaypoint(waypointId)"
		
		if localMembers.destroyed == false then
			local waypointId
			if type(_waypointId) == "number" then
				waypointId = math.floor(_waypointId) + 1
			elseif type(_waypointId) == "string" then
				if localMembers.initialWaypointsName[_waypointId] ~= nil then
					waypointId = localMembers.initialWaypointsName[_waypointId] + 1
				else
					ATMEFct.setError("BAD_WAYPOINT")
					ATMEFct.warningOutputCore(nil, fctDefinition, "waypointId name '" .. _waypointId .. "' unknown in current route", false)
					return
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "waypointId type must be number or string", true)
				return
			end

			if waypointId > 0 and waypointId <= localMembers.nbInitialWaypoints then
				return {
					waypointID = waypointId - 1,
					name = localMembers.trackingDatas[waypointId - localMembers.deltaWaypointId].name,
					x = localMembers.trackingDatas[waypointId - localMembers.deltaWaypointId].pos.x,
					y = localMembers.trackingDatas[waypointId - localMembers.deltaWaypointId].pos.y,
					z = localMembers.trackingDatas[waypointId - localMembers.deltaWaypointId].pos.z,
				}
			else
				ATMEFct.setError("BAD_WAYPOINT")
				ATMEFct.warningOutputCore(nil, fctDefinition, "waypointId must be > last passed waypoint and <= last waypoint", false)
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end
	
	function newGroup.getNextWaypoint(self)
		if localMembers.destroyed == false then
			if localMembers.waypointId + localMembers.deltaWaypointId <= localMembers.nbInitialWaypoints then
				return {
					waypointID = localMembers.waypointId + localMembers.deltaWaypointId - 1,
					name = localMembers.trackingDatas[localMembers.waypointId].name,
					x = localMembers.trackingDatas[localMembers.waypointId].pos.x,
					y = localMembers.trackingDatas[localMembers.waypointId].pos.y,
					z = localMembers.trackingDatas[localMembers.waypointId].pos.z,
				}
			else
				return
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newGroup) .. "getNextWaypoint()", "already disabled", false)
		end
	end
	
	function newGroup.getLastWaypoint(self)
		if localMembers.destroyed == false then
			return {
				waypointID = localMembers.nbInitialWaypoints - 1,
				name = localMembers.trackingDatas[#localMembers.waypoints].name,
				x = localMembers.trackingDatas[#localMembers.waypoints].pos.x,
				y = localMembers.trackingDatas[#localMembers.waypoints].pos.y,
				z = localMembers.trackingDatas[#localMembers.waypoints].pos.z,
			}
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newGroup) .. "getLastWaypoint()", "already disabled", false)
		end
	end
	
	function newGroup.setInvisible(self, _invisible)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "setInvisible(invisible)"
		
		if localMembers.destroyed == false then
			if type(_invisible) == "boolean" then
				local controller = localMembers.objectDCS:getController()
				local command = {
					id = "SetInvisible", 
					params = { 
						value = _invisible, 
					} 
				}
				
				controller:setCommand(command)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "invisible type must be boolean", true)
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end
	
	function newGroup.setImmortal(self, _immortal)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "setImmortal(immortal)"
		
		if localMembers.destroyed == false then
			if type(_immortal) == "boolean" then
				local controller = localMembers.objectDCS:getController()
				local command = {
					id = "SetImmortal", 
					params = { 
						value = _immortal, 
					} 
				}
				
				controller:setCommand(command)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "immortal type must be boolean", true)
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end
	
	function newGroup.switchWaypoint(self, _fromWaypoint, _toWaypoint)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "switchWaypoint(fromWaypoint, toWaypoint)"
	
		if (type(_fromWaypoint) == "number" or type(_fromWaypoint) == "string") and (type(_toWaypoint) == "number" or type(_toWaypoint) == "string") then
		
			if  ATMEConst.canFly[localMembersGroup.category] == true then
				local fromWaypoint
				local toWaypoint
				if type(_fromWaypoint) == "number" then
					fromWaypoint = math.floor(_fromWaypoint) + 1
					
					if fromWaypoint <= 0 and fromWaypoint > localMembers.nbInitialWaypoints then
						ATMEFct.errorOutputCore(nil, fctDefinition, "fromWaypoint must be >0 and <= max waypoints" , false)
						return
					end
				else
					if localMembers.initialWaypointsName[_fromWaypoint] ~= nil then
						fromWaypoint = localMembers.waypointsName[_fromWaypoint]
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "fromWaypoint '" .. _fromWaypoint .. "' doesn't exist in current route" , false)
						return
					end
				end
				
				if type(_toWaypoint) == "number" then
					toWaypoint = math.floor(_toWaypoint) + 1
					
					if toWaypoint <= 0 and toWaypoint > localMembers.nbInitialWaypoints then
						ATMEFct.errorOutputCore(nil, fctDefinition, "toWaypoint must be >0 and <= max waypoints" , false)
						return
					end
				else
					if localMembers.initialWaypointsName[_toWaypoint] ~= nil then
						toWaypoint = localMembers.waypointsName[_toWaypoint]
					else
						
						ATMEFct.errorOutputCore(nil, fctDefinition, "toWaypoint '" .. _toWaypoint .. "' doesn't exist in current route" , false)
						return
					end
				end
				
				local controller = localMembers.objectDCS:getController()
			
				local command = {
					id = "SwitchWaypoint", 
					params = { 
						fromWaypointIndex = fromWaypoint,  
						goToWaypointIndex = toWaypoint, 
					}, 
				}
				
				controller:setCommand(command)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Only aircrafts can switch their waypoints, use setRoute for others", false)
			end
		else
			if type(_fromWaypoint) ~= "number" and type(_fromWaypoint) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "fromWaypoint type must be number or string", true)
			elseif type(_toWaypoint) ~= "number" and type(_toWaypoint) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "toWaypoint type must be number or string", false)
			end
		end
	end
	
	function newGroup.setRoute(self, _route)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "setRoute(route)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isObjectClass(_route, ATME.C_Route) == true then
				local localMembersRoute = ATMEDatas.localMembers[_route:getIndex()]

				if localMembers.spawnDatas.type == localMembersRoute.type then
					if localMembers.routeUncontrolled == false then
						local err, waypoints, waypointsName = ATMEFct.C_Route_createWaypoints(_route)

						if err == false then
							ATMEFct.C_Route_addCurrentPositionWaypoint(waypoints, self)

							localMembers.lastD2 = nil 
							localMembers.waypoints = waypoints
							localMembers.initialWaypointsName = waypointsName
							
							-- Cause désynchro, on remet ici à 0 si ajout de callback de tracking juste après dans le module utilisateur.
 							-- La bonne valeur sera mise dans C_Group_newRouteSet avec le redémarrage du tracking.
							
							localMembers.waypointId = 1
							localMembers.lastPassedWaypoint = 0
							
							ATMEFct.C_Group_CreateTrackingWaypoints(localMembers)

							for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
								localMembers.trackingCallbacks[_localModuleMembers.name] = {}
							end
		
							localMembers.spawnDatas.datas.route.points = waypoints
							localMembers.spawnDatas.waypointsName = waypointsName

							localMembers.offsetWaypoint = 0
							localMembers.deltaWaypointId = 0
							
							-- Remove timer de tracking jusqu'à prise en compte de la route
							if localMembers.idTimer ~= nil then
								timer.removeFunction(localMembers.idTimer)
								localMembers.idTimer = nil
							end
							
							ATMEFct.C_GroupCloningContext_setMission(self, localMembers.waypoints, true)
							
							-- Ajustement du nombre de waypoints si pour un aéronef le dernier waypoint
							-- n'est pas un waypoint d'atterrissage.
							
							localMembers.maxWaypoints = #localMembers.waypoints
							
							if ATMEConst.canFly[localMembers.category] == true then
								if 	localMembers.waypoints[#localMembers.waypoints].type ~= "Land" then
									localMembers.maxWaypoints = localMembers.maxWaypoints + 1
								end
							end
							
							localMembers.nbInitialWaypoints = localMembers.maxWaypoints
						else
							ATMEFct.setError("BAD_WAYPOINTS")
							ATMEFct.warningOutputCore(nil, fctDefinition, "Route with no or bad waypoints", false)
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, localMembers.name .. " is an uncontrolled group.", false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "route must have same type as " .. localMembers.name, false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "route type must be ATME.C_Route", true)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "Group " .. localMembers.name .. " can't have player units", false)
		end
	end
	
	function newGroup.setTask(self, _task)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "setTask(task)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isObjectClass(_task, ATME.C_Task) == true then
				
				local ok, taskDatas = ATMEFct.C_Task_getTaskDatas(_task, localMembers.spawnDatas.type)
				
				if taskDatas.params.startCondition ~= nil then
					ATMEFct.outputCore(nil, "Group '" .. localMembers.name .. "' : setTask - Start condition ignored", 0)
				end
				
				taskDatas.params.startCondition = nil
				if ok == true then
				
					local controller = localMembers.objectDCS:getController()

					controller:setTask(taskDatas)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, taskDatas .. " for group " .. localMembers.name, false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "task type must be ATME.C_Task", true)
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end
	
	function newGroup.resetTask(self)
		if localMembers.destroyed == false then
			local controller = localMembers.objectDCS:getController()

			controller:resetTask()
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newGroup) .. "resetTask()", "already disabled", false)
		end
	end
	
	function newGroup.pushTask(self, _task)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "pushTask(task)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isObjectClass(_task, ATME.C_Task) == true then
				local ok, taskDatas = ATMEFct.C_Task_getTaskDatas(_task, localMembers.spawnDatas.type)
				
				if taskDatas.params.startCondition ~= nil then
					ATMEFct.outputCore(nil, "Group '" .. localMembers.name .. "' : pushTask - Start condition ignored", 0)
				end
				
				taskDatas.params.startCondition = nil
				if ok == true then
					local controller = localMembers.objectDCS:getController()
					controller:pushTask(taskDatas)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, taskDatas .. " for group " .. localMembers.name, false)
				end	
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "task type must be ATME.C_Task", true)
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end
	
	function newGroup.popTask(self)
		if localMembers.destroyed == false then
			local controller = localMembers.objectDCS:getController()

			controller:popTask()
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newGroup) .. "popTask()", "already disabled", false)
		end
	end
	
	function newGroup.canFly(self)
		if localMembers.destroyed == false then
			return ATMEConst.canFly[localMembers.category]
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newGroup) .. "canFly()", "already disabled", false)
		end
	end
	
	function newGroup.getMinDistance(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "getMinDistance(reference)"
		
		if ATMEFct.isReference(_reference) == true then
			return ATMEFct.C_Group_evalDistances(self, ATMEFct.getPointFromReference(_reference), "MIN")
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
		end
	end
	
	function newGroup.getMaxDistance(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "getMaxDistance(reference)"
		
		if ATMEFct.isReference(_reference) == true then
			return ATMEFct.C_Group_evalDistances(self, ATMEFct.getPointFromReference(_reference), "MAX")
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
		end
	end
	
	function newGroup.getMinHDistance(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "getMinHDistance(reference)"
		
		if ATMEFct.isReference(_reference) == true then
			return ATMEFct.C_Group_evalDistances(self, ATMEFct.getPointFromReference(_reference), "MIN_H")
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
		end
	end
		
	function newGroup.getMaxHDistance(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "getMaxHDistance(reference)"
		
		if ATMEFct.isReference(_reference) == true then
			return ATMEFct.C_Group_evalDistances(self, ATMEFct.getPointFromReference(_reference), "MAX_H")
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
		end
	end
	
	function newGroup.getPosition(self)
		return ATMEFct.getGroupPosition(self)
	end
	
	function newGroup.disable(self)
		if localMembers.destroyed == false then
			ATMEDatas.groupsToDisable[localMembers.name] = self
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newGroup) .. "disable()", "already disabled", false)
		end
	end
	
	function newGroup.isDisabled(self)
		if localMembers.objectDCS:isExist() == true then
			return #localMembers.objectDCS:getUnits() == 0
		else
			return true
		end
	end
	
	function newGroup.addNewRouteCallback(self, _fct)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "addNewRouteCallback(fct)"
		
		if localMembers.destroyed == false then
			if type(_fct) == "function" then
				localMembers.newRouteCallback[ATMEDatas.currentModuleName] = _fct
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end
	
	function newGroup.removeNewRouteCallback(self)
		if localMembers.destroyed == false then
			localMembers.newRouteCallback[ATMEDatas.currentModuleName] = nil
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newGroup) .. "removeNewRouteCallback()", "already disabled", false)
		end
	end
	
	function newGroup.addReadyToBoardCallback(self, _fct)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "addReadyToBoardCallback(fct)"
		
		if localMembers.destroyed == false then
			if type(_fct) == "function" then
				localMembers.readyToBoardCallback[ATMEDatas.currentModuleName] = _fct
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end
	
	function newGroup.removeReadyToBoardCallback(self)
		if localMembers.destroyed == false then
			localMembers.readyToBoardCallback[ATMEDatas.currentModuleName] = nil
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newGroup) .. "removeReadyToBoardCallback()", "already disabled", false)
		end
	end

	
	function newGroup.addTrackingCallback(self, _fct, _waypointId)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "addTrackingCallback(fct, [waypointId])"
		
		if localMembers.destroyed == false then	
			if type(_fct) == "function" then
				if type(_waypointId) == "number" then
					local index = math.floor(_waypointId) + 1
					
					if index > localMembers.lastPassedWaypoint + localMembers.deltaWaypointId then
						if index <= localMembers.nbInitialWaypoints then
							localMembers.trackingCallbacks[ATMEDatas.currentModuleName][index - localMembers.deltaWaypointId] = _fct
							return true
						else
							ATMEFct.setError("BAD_WAYPOINT")
							ATMEFct.warningOutputCore(nil, fctDefinition, "Group '" .. localMembers.name .. "' no tracking added. waypointId must be <= last waypoint (" .. localMembers.nbInitialWaypoints - 1 ..")", false)
							return false
						end
					else
						ATMEFct.setError("BAD_WAYPOINT")
						ATMEFct.warningOutputCore(nil, fctDefinition, "Group '" .. localMembers.name .. "' no tracking added. waypointId must be > last passed waypoint (" .. localMembers.lastPassedWaypoint - 1 .. ")", false)
						return false
					end
				elseif type(_waypointId) == "string" then
					if localMembers.initialWaypointsName[_waypointId] ~= nil then
						localMembers.trackingCallbacks[ATMEDatas.currentModuleName][localMembers.initialWaypointsName[_waypointId] + 1 - localMembers.deltaWaypointId] = _fct
						return true
					else
						ATMEFct.setError("BAD_WAYPOINT")
						ATMEFct.warningOutputCore(nil, fctDefinition, "Group '" .. localMembers.name .. "' no tracking added. waypointId name '" .. _waypointId .. "' unknown in current route", false)
						return false
					end
				elseif _waypointId == nil then
					for _i = localMembers.offsetWaypoint + 1, #localMembers.waypoints do
						localMembers.trackingCallbacks[ATMEDatas.currentModuleName][_i] = _fct
					end
					
					return true
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "waypointId type must be nil, number or string", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
			return false
		end
	end
	
	function newGroup.removeTrackingCallback(self, _waypointId)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "removeTrackingCallback([waypointId])"
		
		if localMembers.destroyed == false then
			local index
			
			if type(_waypointId) == "number" then
				index = math.floor(_waypointId) + 1
				
				if index > 0 and index <= localMembers.nbInitialWaypoints then
					localMembers.trackingCallbacks[ATMEDatas.currentModuleName][index + localMembers.offsetWaypoint] = nil
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "waypointId must be >=0 and <= max waypoints of current route", false)
				end
			elseif type(_waypointId) == "string" then
				index = #localMembers.waypoints
				local ok = false
				while index > localMembers.offsetWaypoint and ok == false do
					if localMembers.waypoints[index].name ~= nil and localMembers.waypoints[index].name == _waypointId then
						ok = true
					else
						index = index - 1
					end
				end
				
				if ok == true then
					localMembers.trackingCallbacks[ATMEDatas.currentModuleName][index] = nil
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "waypointId name '" .. _waypointId .. "' unknown in current route", false)
				end
			elseif _waypointId == nil then
				for _i = localMembers.offsetWaypoint + 1, #localMembers.waypoints do
					localMembers.trackingCallbacks[ATMEDatas.currentModuleName][_i] = nil
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "waypointId type must be nil, number or string", true)
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end
	
	function newGroup.respawn(self, _rules)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "respawn([rules])"
		
		if localMembers.destroyed == false then
			if type(_rules) == "string" or type(_rules) == "number" then
				if type(_rules) == "number" then
					if _rules >= 0 then
						local value = math.floor(_rules)
						localMembers.autoRespawn = {
							activated = true,
							time = value,
							random = false,
							minTime = 0,
							maxTime = 0,
						}
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "rules must be >= 0 when number", false)
					end
				else
					local tokens = ATMEFct.analyzeString(_rules)
					
					if #tokens >= 1 and #tokens <= 3 then
						local ok = false
						local minRandom = 0
						local maxRandom = 0
						local value = 0
						local randomFlag = false
						
						if #tokens >= 2 and tokens[1].type == "TOKEN_TYPE_RANDOM" then
							randomFlag = true
							
							if tokens[2].type == "TOKEN_INTEGER" then
								if tokens[2].value >= 0 then
									maxRandom = tokens[2].value
									
									if #tokens == 3 then
										minRandom = tokens[2].value
										
										if tokens[3].type == "TOKEN_INTEGER" then
											if tokens[3].value >= 0 then
												maxRandom = tokens[3].value
												ok = true
											else
												ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #3 must be >= 0 in string format rules", false)
											end
										else
											ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #3 must be number or time value in string format rules", false)
										end
									else
										maxRandom = tokens[2].value
										ok = true
									end
								else
									ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be >= 0 in string format rules", false)
								end
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be number or time value in string format rules", false)
							end
						else
							if tokens[1].type == "TOKEN_INTEGER" then
								if tokens[1].value >= 0 then
									value = tokens[1].value
									ok = true
								else
									ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #1 must be >= 0 in string format rules", false)
								end
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #1 must be number or time value or 'RANDOM' in string format rules", false)
							end
						end
						
						if ok == true then
							if randomFlag == true then
								if maxRandom > minRandom then
									localMembers.autoRespawn = {
										activated = true,
										time = 0,
										random = true,
										minTime = minRandom,
										maxTime = maxRandom,
									}
								else
									ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be < parameter #3 in string format rules", false)
								end
							else
								localMembers.autoRespawn = {
									activated = true,
									time = value,
									random = false,
									minTime = 0,
									maxTime = 0,
								}
							end
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Bad string format rules", false)
					end
				end
			elseif _rules == nil then
				localMembers.autoRespawn = {
					activated = false,
					time = 0,
					random = false,
					minTime = 0,
					maxTime = 0,
				}
			else
				if type(_rules) ~= "string" and type(_rules) ~= "number" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "rules type must be nil, string or number", true)
				end
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end
	
	function newGroup.start(self)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "start()"
		
		if localMembers.destroyed == false then
			if ATMEConst.canFly[localMembers.category] == true then
				if localMembers.routeUncontrolled == true then
					localMembers.routeUncontrolled = false
					local controller = localMembers.objectDCS:getController()

					controller:setCommand({ 
											id = 'Start', 
											params = {},
										})
				end
			else
				localMembers.routeUncontrolled = false
				self:freeze(false)
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end
	
	function newGroup.isControlled(self)
		if localMembers.destroyed == false then
			return localMembers.routeUncontrolled == false
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newGroup) .. "isControlled()", "already disabled", false)
		end
	end
	
	function newGroup.getFirstUnit(self)
		if localMembers.destroyed == false then
			local unitName = localMembers.ATMEUnitNames:getFirst()
			if unitName ~= nil then
				return ATMEDatas.units[unitName]
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newGroup) .. "getFirstUnit()", "already disabled", false)
		end
	end

	function newGroup.getNextUnit(self)
		if localMembers.destroyed == false then
			local unitName = localMembers.ATMEUnitNames:getNext()
			
			if unitName ~= nil then
				return ATMEDatas.units[unitName]
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newGroup) .. "getNextUnit()", "already disabled", false)
		end
	end

	function newGroup.getPreviousUnit(self)
		if localMembers.destroyed == false then
			local unitName = localMembers.ATMEUnitNames:getPrevious()
			
			if unitName ~= nil then
				return ATMEDatas.units[unitName]
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newGroup) .. "getPreviousUnit()", "already disabled", false)
		end
	end

	function newGroup.getLastUnit(self)
		if localMembers.destroyed == false then
			local unitName = localMembers.ATMEUnitNames:getLast()
			
			if unitName ~= nil then
				return ATMEDatas.units[unitName]
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newGroup) .. "getLastUnit()", "already disabled", false)
		end
	end

	function newGroup.getUnitByName(self, _name)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "getUnitByName(name)"
		
		if localMembers.destroyed == false then
			if type(_name) == "string" then
				return localMembers.ATMEUnits[_name]
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", true)
			end	
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end

	function newGroup.getUnits(self)
		if localMembers.destroyed == false then
			return localMembers.ATMEUnits
		end
	end

	function newGroup.getNbUnits(self)
		return localMembers.ATMEUnitNames:getCount()
	end

	function newGroup.isOnlyInfantry(self)
		if localMembers.destroyed == false then
			return localMembers.onlyInfantry
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newGroup) .. "isOnlyInfantry()", "already disabled", false)
		end
	end
	
	function newGroup.radarOn(self)
		if localMembers.destroyed == false then
			localMembers.objectDCS:enableEmission(true)
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newGroup) .. "radarOn()", "already disabled", false)
		end
	end
	
	function newGroup.radarOff(self)
		if localMembers.destroyed == false then
			localMembers.objectDCS:enableEmission(false)
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newGroup) .. "radarOff()", "already disabled", false)
		end
	end

	function newGroup.isInZones(self, _zones, _allGroup)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "isInZones(zones, allGroup)"
	
		if ATMEFct.isObjectClass(_zones, ATME.C_Zones) == true and type(_allGroup) == "boolean" then
			if localMembers.destroyed == false then
				local allUnitInZones = true
				local found = false

				for _, _unit in pairs(localMembers.ATMEUnits) do
					local point = _unit:getPosition()
					local isInside, _ = ATMEFct.C_Zones_isInside(_zones, point)
					if  isInside == false then
						allUnitInZones = false
					else 
						found = true
					end
				end
				
				return(_allGroup == false and found == true) or (_allGroup == true and allUnitInZones == true)
			else
				local isInside, _ = ATMEFct.C_Zones_isInside(_zones, localMembers.lastPosition)
				return isInside
			end
		else
			if ATMEFct.isObjectClass(_zones, ATME.C_Zones) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "zones type must be ATME.C_Zones", true)
			elseif type(_allGroup) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "allGroup type must be boolean", false)
			end
		end
	end				
	
	function newGroup.isInDCSZone(self, _zoneName, _allGroup)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "isInDCSZone(zoneName, allGroup)"
		
		if type(_zoneName) == "string" and type(_allGroup) == "boolean" then
			local zone = ATMEFct.getDCSZone(_zoneName)
			if zone ~= nil then
				if _allGroup == false then
					return ATMEFct.C_Group_evalDistances(self, zone.point, "MIN_H") <= zone.radius
				else
					return ATMEFct.C_Group_evalDistances(self, zone.point, "MAX_H") <= zone.radius
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone " .. _zoneName .. " unknown", false)
			end
		else
			if type(_zoneName) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "zoneName type must be string", true)
			elseif type(_allGroup) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "allGroup type must be boolean", false)
			end
		end
	end

	function newGroup.isInZone(self, _zone, _allGroup)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "isInZone(zone, allGroup)"
		
		if ATMEFct.isObjectClass(_zone, ATME.C_Zone) == true and type(_allGroup) == "boolean" then
			local localMembersZone = ATMEDatas.localMembers[_zone:getIndex()]
			local datas = {
				zone = _zone,
				actualCenter = ATMEFct.C_Zone_getActualCenter(localMembersZone),
				azimuth = ATMEFct.C_Zone_getActualAzimuth(localMembersZone),
			}
					
			if localMembers.destroyed == false then
				local allGroupInZone = true
				local found = false
				
				for _, _unit in pairs(localMembers.ATMEUnits) do
					local point = _unit:getPosition()
					
					if ATMEFct[ATMEConst.indZoneConfig[_zone:getClassName()].fctIsInside](point, datas) == true then
						found = true
					end
				end
				
				return (_allGroup == true and allGroupInZone == true) or (_allGroup == false and found == true)
			else
				return ATMEFct[ATMEConst.indZoneConfig[_zone:getClassName()].fctIsInside](localMembers.lastPosition, datas)
			end
		else
			if ATMEFct.isObjectClass(_zone, ATME.C_Zone) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "zone type must be ATME.C_Zone", true)
			elseif type(_radius) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "radius type must be number", false)
			elseif type(_allGroup) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "allGroup type must be boolean", false)
			end
		end
	end	

	function newGroup.isReadyToBoard(self)
		if localMembers.destroyed == false then
			return ATMEDatas.readyToBoardGroups[localMembers.name] ~= nil
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newGroup) .. "isReadyToBoard()", "already disabled", false)
		end
	end

	function newGroup.setReadyToBoard(self, ...)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "setReadyToBoard([nextWaypoint])\nsetReadyToBoard(pickupZone, [nextWaypoint])"
		
		if localMembers.destroyed == false then
			local args = {...}
			local argId = 1
			local pickupZone 
			
			if #args > 0 then
				if ATMEFct.isObjectClass(args[1], ATME.C_PickupZone) == true then
					pickupZone = args[1]
					
					if args[2] == nil or type(args[2]) == "number" or type(args[2]) == "string" then
						argId = 2
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "nextWaypoint type must be nil, number or string", true)
						return
					end
				elseif args[1] ~=  nil and type(args[1]) ~= "number" and type(args[1]) ~= "string" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "nextWaypoint type must be nil, number or string", true)
					return
				end
			end
			
			if self:isOnlyInfantry() == true then
				if localMembers.readyToBoardSet == false then
					if localMembers.boardingStarted == false then
						local nextWaypoint
						if type(args[argId]) == "number" then
							nextWaypoint = math.floor(args[argId]) + 1
						elseif type(args[argId]) == "string" then
							if localMembers.initialWaypointsName[args[argId]] ~= nil then
								nextWaypoint = localMembers.initialWaypointsName[args[argId]] + 1
							else
								ATMEFct.setError("BAD_WAYPOINT_NAME")
								ATMEFct.warningOutputCore(nil, fctDefinition, "Group '" .. localMembers.name .. "' : nextWaypoint name '" .. args[argId] .. "' unknown in current route", false)
								return false
							end
						end
						
						if pickupZone == nil then
							ATMEFct.C_Group_changeReadyToBoard(self, true, nextWaypoint)
							localMembers.readyToBoardSet = true
						else
							local localMembersPickupZone = ATMEDatas.localMembers[pickupZone:getIndex()]
							if localMembersPickupZone.coalitionName == localMembers.coalition then
								local datas = {
									randomMax = 0.25 + math.random() / 2,
									name = localMembers.name,
									object = self,
									localMembersObject = localMembers,
									localMembersPickupZone = localMembersPickupZone,
									isInZone = false,
									nextWaypoint = nextWaypoint,
									nbTries = 0,
								}
								
								localMembers.pickupZone = pickupZone
								
								localMembers.idTimerPickupZone = timer.scheduleFunction(ATMETimers.C_PickupZone_readyToBoardTimer, datas, timer.getTime() + 1)
								localMembers.readyToBoardSet = true
							else
								ATMEFct.setError("BAD_COALITION")
								ATMEFct.warningOutputCore(nil, fctDefinition, "Group '" .. localMembers.name .. "' : pickup zone coalition is bad", false)
								return false
							end
						end
					else
						ATMEFct.setError("BOARDING_STARTED")
						ATMEFct.warningOutputCore(nil, fctDefinition, "Group '" .. localMembers.name .. "' bording already started", false)
						return false
					end
				else
					ATMEFct.setError("ALREADY_SET_READYTOBOARD")
					ATMEFct.warningOutputCore(nil, fctDefinition, "Group '" .. localMembers.name .. "' already ready to board", false)
					return false
				end
			else
				ATMEFct.setError("NOT_INFANTRY")
				ATMEFct.warningOutputCore(nil, fctDefinition, "Group '" .. localMembers.name .. "' must be infantry group", true)
				return false
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
			return false
		end
		
		return true
	end
	
	function newGroup.resetReadyToBoard(self)
		if localMembers.destroyed == false then
			if localMembers.pickupZone ~= nil then
				if localMembers.idTimerPickupZone ~= nil then
					timer.removeFunction(localMembers.idTimerPickupZone)
					localMembers.idTimerPickupZone = nil
				end
				
				local localMembersPickupZone = ATMEDatas.localMembers[localMembers.pickupZone:getIndex()]
				if ATMEDatas.readyToBoardGroups[self:getName()] ~= nil then
					localMembersPickupZone.nbReadyToBoard = localMembersPickupZone.nbReadyToBoard - 1
				end
				
				localMembers.pickupZone = nil
			end
			
			localMembers.readyToBoardSet = false
			ATMEFct.C_Group_changeReadyToBoard(self, false, nil)
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newGroup) .. "resetReadyToBoard()", "already disabled", false)
		end
	end

	function newGroup.isBoarding(self)
		if localMembers.destroyed == false then
			return localMembers.boardingStarted
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newGroup) .. "isBoarding()", "already disabled", false)
		end
	end
	
	function newGroup.move(self)
		if localMembers.destroyed == false then
			if ATMEConst.canFly[localMembers.category] == false then
				trigger.action.groupContinueMoving(localMembers.objectDCS)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "only ground groups can be moved", false)
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newGroup) .. "move()", "already disabled", false)
		end
	end

	function newGroup.stop(self)
		if localMembers.destroyed == false then
			if ATMEConst.canFly[localMembers.category] == false then
				trigger.action.groupStopMoving(localMembers.objectDCS)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "only ground groups can be stopped", false)
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newGroup) .. "stop()", "already disabled", false)
		end
	end

	function newGroup.freeze(self, _freezeOn)
		if localMembers.destroyed == false then
			if ATMEConst.canFly[localMembers.category] == false then
				local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "freeze(freezeOn)"
				
				if type(_freezeOn) == "boolean" then
					if _freezeOn == true then
						trigger.action.setGroupAIOff(localMembers.objectDCS)
					else
						trigger.action.setGroupAIOn(localMembers.objectDCS)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "freezeOn type must be boolean", true)
				end	
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "only ground groups can be freezed", false)
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end
	
	function newGroup.setOption(self, _option, _value)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "setOption(option, value)"

		if localMembers.destroyed == false then
			if type(_option) == "string" and _value ~= nil then
				local err, option, optionValue = ATMEFct.C_setOption(localMembers.spawnDatas.type, _option, _value) 

				if err == false then
					local controller = localMembers.objectDCS:getController()
					controller:setOption(ATMEConst.optionAI[localMembers.spawnDatas.type][option].id, optionValue)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, option, false)
				end
			else
				if type(_option) ~= "string" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "option type must be string", true)
				elseif _value == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "value type can't be nil", false)
				end
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end

	ATMEFct.C_Group_addUnit(newGroup, _ATMEUnit)

	if ATMEDatas.spawnedGroupsWaypoints[localMembers.name] == nil then  -- Groupe non spawné
		-- Les waypoints de la route initiale sont recopiés
		-- pour avoir la route courante.
		localMembers.waypoints = localMembers.spawnDatas.datas.route.points
		localMembers.initialWaypointsName = localMembers.spawnDatas.waypointsName
	else -- Groupe spawné
		ATMEDatas.reservedAsPlayerGroups[localMembers.name] = 0
		localMembers.waypoints = ATMEDatas.spawnedGroupsWaypoints[localMembers.name].waypoints
		localMembers.initialWaypointsName = ATMEDatas.spawnedGroupsWaypoints[localMembers.name].initialWaypointsName
		ATMEDatas.spawnedGroupsWaypoints[localMembers.name] = nil
	end
	
	-- Ajustement du nombre de waypoints si pour un aéronef le dernier waypoint
	-- n'est pas un waypoint d'atterrissage.
	localMembers.maxWaypoints = #localMembers.waypoints
	localMembers.nbInitialWaypoints = localMembers.maxWaypoints

	if ATMEConst.canFly[localMembers.category] == true then
		if localMembers.maxWaypoints > 0 then
			if 	localMembers.waypoints[#localMembers.waypoints].type ~= "Land" then
				localMembers.maxWaypoints = localMembers.maxWaypoints + 1
			end
		end
	end
	
	newGroup.modules = {}
	setmetatable(newGroup.modules, ATMEDatas.C_Group_MT.mtModules)
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
		localMembers.trackingCallbacks[_localModuleMembers.name] = {}
		if newGroup.modules[_localModuleMembers.name] == nil then
			newGroup.modules[_localModuleMembers.name] = {}
		end
	end
	
	ATMEDatas.newCreatedGroups[localMembers.name] = {
		timerOffset = 1,
		isSpawned = false,
		disembark = ATMEDatas.groupsOnBoard[localMembers.name] ~= nil,
		group = newGroup,
	}

	ATMEFct.outputCore(nil,"Prepare group = " .. localMembers.name .. " -> ok", 2)
end

ATMEFct.C_Group_delete = function(_group, _reserved)
	local localMembers = ATMEDatas.localMembers[_group:getIndex()]

	if localMembers.destroyed == false then
		local flag = "DESTROYED"
		if localMembers.isEnabled == false then
			flag = "DISABLED"
		end
		
		if localMembers.idTimer ~= nil then
			timer.removeFunction(localMembers.idTimer)
			localMembers.idTimer = nil
		end
		
		ATMEDatas.soundsOn.groups[localMembers.objectDCSId] = nil
		
		if localMembers.pickupZone ~= nil then
			if ATMEDatas.readyToBoardGroups[localMembers.name] ~= nil then
				local localMembersPickupZone = ATMEDatas.localMembers[localMembers.pickupZone:getIndex()]
				localMembersPickupZone.nbReadyToBoard = localMembersPickupZone.nbReadyToBoard - 1
			end
		end
		
		if localMembers.boardingStarted == false then
			for _i = #ATMEDatas.modulesNames, 1, -1 do
				local localModuleMembers = ATMEDatas.modulesLocalMembers[ATMEDatas.modulesNames[_i]]
				
				-- Les groupes chargés dans un autre véhicule ne génèrent pas d'appel 
				-- à onDeleteGroupHandler
				
				if localModuleMembers.handlers.onDeleteGroupHandler ~= nil then
					ATMEDatas.currentModuleName = localModuleMembers.name
					
					-- Appel du module
					ATMEFct.outputCore(nil,"Appel onDeleteGroupHandler du module " .. localModuleMembers.name, 2)
					xpcall(function() localModuleMembers.handlers.onDeleteGroupHandler(_group, flag) end, function() ATMEFct.errUserHandler('onDeleteGroupHandler') end)
				end
			end

			-- Si le groupe a un respawn automatique il est préparé
			if localMembers.autoRespawn.activated == true then
				local respawnTime = localMembers.autoRespawn.time
				if localMembers.autoRespawn.random == true then
					respawnTime = math.random(localMembers.autoRespawn.minTime, localMembers.autoRespawn.maxTime)
				end
				
				ATMEDatas.groupsToRespawn[localMembers.name] = true
				timer.scheduleFunction(ATMETimers.C_Group_autoRespawn, localMembers.initialSpawnDatas, timer.getTime() + respawnTime)
				ATMEDatas.respawnedGroups[localMembers.name] = true
			elseif localMembers.spawnDatas.idSpawnedList ~= nil then
				if localMembers.spawnDatas.idSpawnedList.groupSpawnedId ~= nil then
					table.insert(localMembers.spawnDatas.idSpawnedList.freeIds, localMembers.spawnDatas.idSpawnedList.groupSpawnedId)
				end
				
				for _name, _item in pairs(localMembers.spawnDatas.idSpawnedList.units) do
					table.insert(_item.freeIds, _item.id)
				end
			end
		
			_group.modules = nil
		end

		if _reserved == false and ATMEDatas.reservedAsPlayerGroups[localMembers.name] == 0 and localMembers.autoRespawn.activated == false then
			ATMEDatas.reservedGroupNames[localMembers.name] = nil
		end

		localMembers.destroyed = true
		ATMEDatas.C_Group_MT:remove(_group:getIndex())
		
		ATMEDatas.groups[localMembers.name] = nil
		ATMEDatas.readyToBoardGroups[localMembers.name] = nil
		ATMEDatas.newCreatedGroups[localMembers.name] = nil
		
		ATMEFct.outputCore(nil,"Delete group = " .. localMembers.name, 2)
	end
end

setmetatable(ATME.C_Group,ATMEDatas.C_Group_MT)

ATMEDatas.C_Group_MT.mt.__metatable = "Private"
ATMEDatas.C_Group_MT.mt.__index = ATME.C_Group
ATMEDatas.C_Group_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Group_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Group : Bad key " .. k, "table locked", false)
	end
end

ATMEDatas.C_Group_MT.mtModules.__metatable = "Private"
ATMEDatas.C_Group_MT.mtModules.__newindex = function(t, k, v)
	if ATMEDatas.modules[k] ~= nil then
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil,"ATME : ATME.C_Group.modules Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Unit : gère les unités AI et joueur au format ATME. Pas de contructeur accessible
-- pour les modules externes.
-- Les instances de cette classe sont intégralement gérées par le module ATME_Core. La création et la destruction
-- dépendent du contexte de début de mission et des événements qui se produisent dans DCS en cours de mission.
-- Les modules externes accèdent à cette gestion indirectement par les handlers définis dans ceux-ci et les méthodes
-- accessibles de cette classe. Ceci permet de garantir une cohérence sur les listes internes ATME.
----------------------------------------------------------------------------------------------------------

ATMEFct.C_Unit_unload_now = function(_transportUnit, _id, _currentOnBoardGroupDatas)
	local localMembers = ATMEDatas.localMembers[_transportUnit:getIndex()]
	local err, delta = ATMEFct.C_GroupCloningContext_spawnFromTransportUnit(_transportUnit, _currentOnBoardGroupDatas)

	localMembers.nbUnitsOnBoard = localMembers.nbUnitsOnBoard - _currentOnBoardGroupDatas.nbUnitsOnBoard
	localMembers.onBoardGroupDatas:remove(_id)
	
	if ATMEConst.canFly[ATMEDatas.localMembers[localMembers.group:getIndex()].category] == true then
		trigger.action.setUnitInternalCargo(localMembers.name, localMembers.nbUnitsOnBoard * 100)
		
		ATMEFct.outputCore(nil, "Poids restant à bord = " .. localMembers.nbUnitsOnBoard * 100 .. " pour unité transport " .. localMembers.name, 2)
	end
	
	ATMEDatas.groupsOnBoard[_currentOnBoardGroupDatas.localMembers.spawnDatas.datas.name] = nil
	
	if err == false then
		-- ATTENTION : Lors du spawn, les unités sont crées après la fonction addGroup car unités au sol.
		-- Donc, à ce stade, les infos ATME pour le groupe et les unités sont à jour.
		-- La mise à jour de ATMEDatas.groupsOnBoard se fait ici.
		
		local localSpawnGroupMembers = ATMEDatas.localMembers[ATMEDatas.newCreatedGroups[_currentOnBoardGroupDatas.localMembers.spawnDatas.datas.name].group:getIndex()]
		localSpawnGroupMembers.autoRespawn = _currentOnBoardGroupDatas.localMembers.autoRespawn
		localSpawnGroupMembers.disembarking = true
		localSpawnGroupMembers.offsetWaypoint = 1	-- 1 WP de débarquement en plus.
		localSpawnGroupMembers.nbInitialWaypoints = _currentOnBoardGroupDatas.localMembers.nbInitialWaypoints
		localSpawnGroupMembers.deltaWaypointId = _currentOnBoardGroupDatas.localMembers.deltaWaypointId + delta
		localSpawnGroupMembers.userPrivateDatas = _currentOnBoardGroupDatas.userPrivateDatas
		localSpawnGroupMembers.newRouteCallback = _currentOnBoardGroupDatas.newRouteCallback
		localSpawnGroupMembers.readyToBoardCallback = _currentOnBoardGroupDatas.readyToBoardCallback
		localSpawnGroupMembers.initialSpawnDatas = _currentOnBoardGroupDatas.localMembers.initialSpawnDatas
		-- Permet de savoir si le groupe a déjà débarqué, pour rappel 2 WP sont ajoutés pour sortir du véhicule
		-- (position initiale et position pour s'en aller)

		localSpawnGroupMembers.trackingCallbacks = {}
		
		for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
			localSpawnGroupMembers.trackingCallbacks[_localModuleMembers.name] = {}
			
			for _i = 2, #_currentOnBoardGroupDatas.localMembers.waypoints do	-- Le premier WP est celui de débarquement
				localSpawnGroupMembers.trackingCallbacks[_localModuleMembers.name][_i] = _currentOnBoardGroupDatas.trackingCallbacks[_localModuleMembers.name][_i + delta]
			end
		end
		
		_currentOnBoardGroupDatas.trackingCallbacks = nil
		_currentOnBoardGroupDatas.userPrivateDatas = nil
		_currentOnBoardGroupDatas.newRouteCallback = nil
		_currentOnBoardGroupDatas.readyToBoardCallback = nil
		
		ATMEDatas.newCreatedGroups[_currentOnBoardGroupDatas.localMembers.spawnDatas.datas.name].event = {
			type = "TRANSPORT",
			datas = {
				action = "UNLOADED",
				unitName = _transportUnit:getName(),
				index = _id,
				linkedGroupName = _currentOnBoardGroupDatas.localMembers.spawnDatas.datas.name,
				nbUnitsOnBoard = localMembers.nbUnitsOnBoard,
				nbDeadUnits = _currentOnBoardGroupDatas.nbDeadUnits,
			},
		}
	else
		-- Le groupe est mort au débarquement
		
		local event = {
			type = "TRANSPORT",
			datas = {
				action = "UNLOADED_GROUP_DEAD",
				unitName = _localMembersTransportUnit.name,
				index = _id,
			},
		}
		
		ATMEFct.fireEvent(event)
	end
end

ATMETimers.C_Unit_autoUnload = function(_localMembers)
	if ATMEDatas.exitATME == false then
		if _localMembers.objectDCS:isExist() == true then
			local unit = ATMEDatas.units[_localMembers.name]
			local vector = ATMEFct.vectors["3D"].createFromTable(nil, nil, _localMembers.objectDCS:getVelocity())
			if _localMembers.autoUnloadDatas.all == false then
				local ok = false
				for _id, _item in pairs(_localMembers.autoUnloadDatas.list) do
					ok = true
					if _item.armed == false then
						
						if (_localMembers.objectDCS:hasAttribute("Ground vehicles") == true and vector:getHModule() > 3) or 
								((_localMembers.objectDCS:hasAttribute("Helicopters") == true or _localMembers.objectDCS:hasAttribute("Planes") == true) and _localMembers.objectDCS:inAir() == true) then
							_item.armed = true
							_item.wait = 5
						end
					end
					
					if _item.armed == true then
						if vector:getHModule() < 1 and (_localMembers.objectDCS:hasAttribute("Ground vehicles") == true or 
								((_localMembers.objectDCS:hasAttribute("Helicopters") == true or _localMembers.objectDCS:hasAttribute("Planes") == true) and _localMembers.objectDCS:inAir() == false)) then
						
							_item.wait = _item.wait - 1
							
							if _item.wait == 0 then
								local currentOnBoardGroupDatas = _localMembers.onBoardGroupDatas:get(_id)
								if currentOnBoardGroupDatas ~= nil then
									ATMEFct.C_Unit_unload_now(unit, _id, currentOnBoardGroupDatas)
								end
								
								_localMembers.autoUnloadDatas.list[_id] = nil
							end
						else
							_item.wait = 5
						end
					end
				end

				if ok == false then
					_localMembers.autoUnloadDatas = nil
					ATMEFct.outputCore(nil, "C_Unit_autoUnload stoppé " .. _localMembers.name, 2)
				else
					return timer.getTime() + 3
				end
			else
				if _localMembers.autoUnloadDatas.armed == false then
					if (_localMembers.objectDCS:hasAttribute("Ground vehicles") == true and vector:getHModule() > 3) or 
							((_localMembers.objectDCS:hasAttribute("Helicopters") == true or _localMembers.objectDCS:hasAttribute("Planes") == true) and _localMembers.objectDCS:inAir() == true) then
						_localMembers.autoUnloadDatas.armed = true
						_localMembers.autoUnloadDatas.wait = 5
					else
						return timer.getTime() + 10
					end
				end
				
				if _localMembers.autoUnloadDatas.armed == true then
					if vector:getHModule() < 1 and (_localMembers.objectDCS:hasAttribute("Ground vehicles") == true or 
							((_localMembers.objectDCS:hasAttribute("Helicopters") or _localMembers.objectDCS:hasAttribute("Planes")) and _localMembers.objectDCS:inAir() == false)) then
					
						_localMembers.autoUnloadDatas.wait = _localMembers.autoUnloadDatas.wait - 1
						
						if _localMembers.autoUnloadDatas.wait == 0 then
							for _id, _current in _localMembers.onBoardGroupDatas:pairs() do
								ATMEFct.C_Unit_unload_now(unit, _id, _current)
							end
							
							_localMembers.autoUnloadDatas = nil
							ATMEFct.outputCore(nil, "C_Unit_autoUnload stoppé " .. _localMembers.name, 2)
						else
							return timer.getTime() + 3
						end
					else
						_localMembers.autoUnloadDatas.wait = 5
						return timer.getTime() + 3
					end
				end
			end
		end
	end
	
	_localMembers.idTimerUnload = nil
end

ATMEFct.C_Unit_unload = function(_localMembers, _id, _whenStopped)
	if _localMembers.autoUnloadDatas == nil then
		_localMembers.autoUnloadDatas = {}
	end
	
	if _id == nil then
		_localMembers.autoUnloadDatas.all = true
		_localMembers.autoUnloadDatas.armed = not _whenStopped
		_localMembers.autoUnloadDatas.wait = 1
		_localMembers.autoUnloadDatas.list = nil
	else
		_localMembers.autoUnloadDatas.all = false
		if _localMembers.autoUnloadDatas.list == nil then
			_localMembers.autoUnloadDatas.list = {}
		end
		if _localMembers.autoUnloadDatas.list[_id] == nil then
			_localMembers.autoUnloadDatas.list[_id] =  {
				armed = not _whenStopped,
				wait = 1,
			}
		end
	end
	
	if _localMembers.idTimerUnload == nil then
		ATMEFct.outputCore(nil, "C_Unit_unload - Start timer pour " .. _localMembers.name, 2)
		_localMembers.idTimerUnload = timer.scheduleFunction(ATMETimers.C_Unit_autoUnload, _localMembers, timer.getTime() + 1)
	end
end

ATMEFct.C_Unit_load_now = function(_datas)
	local localMembers = ATMEDatas.localMembers[_datas.transportUnit:getIndex()]
	local localMembersGroup = ATMEDatas.localMembers[_datas.groupToBoard:getIndex()]
	
	local maxUnits = ATMEConst.personnelCarrierAttributes[localMembers.typeName].maxUnits
	local nbUnits = localMembers.nbUnitsOnBoard
	local nbUnitsBoarding = 0
								
	for _, _groupItem in ipairs(localMembers.groupsBoarding) do
		nbUnitsBoarding = nbUnitsBoarding + _groupItem.nbUnits
	end
	
	if localMembersGroup.ATMEUnitNames:getCount() + nbUnits + nbUnitsBoarding <= maxUnits then
		local transportUnitPosition = localMembers.objectDCS:getPoint()
		if ATMEFct.C_Group_evalDistances(_datas.groupToBoard, transportUnitPosition, "MIN_H") <= _datas.radius then	
			-- On mémorise le nb waypoints avant le changement de route pour monter
			-- dans le transport.

			local pos = ATMEFct.getGroupPosition(_datas.groupToBoard)
			local v = ATMEFct.vectors["3D"].createFromReference(nil, nil, pos, transportUnitPosition)
			local v1 = v:toUnitVector()

			local waypoints = {}
			waypoints[1] = {
				action = "Off Road",
				x = transportUnitPosition.x,
				y = transportUnitPosition.z,
				type = "Turning Point",
				speed = 5,
			}
			
			waypoints[2] = {
				action = "Off Road",
				x = transportUnitPosition.x + v1:getX() * 500,
				y = transportUnitPosition.z + v1:getZ() * 500,
				type = "Turning Point",
				speed = 5,
			}

			local line = ATMEFct.lines["2D"].createFromReference(nil, nil, transportUnitPosition, pos)

			local axis = line:getPerpendicular(0)
			
			ATMEFct.C_GroupCloningContext_setWaypoints(_datas.groupToBoard, waypoints, false)
			
			localMembersGroup.boardingStarted = true
			
			if localMembersGroup.pickupZone ~= nil then
				local localMembersPickupZone = ATMEDatas.localMembers[localMembersGroup.pickupZone:getIndex()]
				localMembersPickupZone.nbReadyToBoard = localMembersPickupZone.nbReadyToBoard - 1
				localMembersGroup.pickupZone = nil
			end
				
			trigger.action.groupContinueMoving(localMembersGroup.objectDCS)
						
			local pA = axis:getOrigin()
			local pB = ATMEFct.C_Line2D_getPointFromOrigin(axis:getA(), axis:getB(), pA, 10)
			
			local itemGroupToBoard = {
				localMembers = localMembersGroup,							-- On mémoérise les données la table avec les données pour éviter les désynchro du waypointId
				trackingCallbacks = localMembersGroup.trackingCallbacks, 	-- Table conservée après la destruction du groupe, donc on récup le pointeur
				userPrivateDatas = localMembersGroup.userPrivateDatas, 		-- Table conservée après la destruction du groupe, donc on récup le pointeur
				transportUnitName = localMembers.name,
				unitsOnBoard = {},
				nbUnitsOnBoard = 0,
				nbUnitsLost = 0,
				maxDelay = timer.getTime() + ATMEFct.C_Group_evalDistances(_datas.groupToBoard, transportUnitPosition, "MAX_H") / 5 + 10,
				A = pA,
				B = pB,
				sideOfWP2 = ATMEFct.C_Unit_whichSide(_datas.transportUnit, waypoints[2]),
				newRouteCallback = localMembersGroup.newRouteCallback,			-- Table conservée après la destruction du groupe, donc on récup le pointeur
				readyToBoardCallback = localMembersGroup.readyToBoardCallback, 	-- Table conservée après la destruction du groupe, donc on récup le pointeur
				currentUnitName = localMembersGroup.ATMEUnitNames:getFirst(),	-- Init sur la première unité
				boardingStart = false,
				boardingStopped = false,
				delayBeforeStopped = 0,
				alea = 0,
			}
			
			localMembers.nbBoardings = localMembers.nbBoardings + 1
			timer.scheduleFunction(ATMETimers.C_Group_startLoad_timer, itemGroupToBoard, timer.getTime() + 1)
			
			return
		else
			ATMEFct.outputCore(nil, "Transport unit '" .. localMembers.name .. "' -> Loading group '" .. localMembersGroup.name .. "' to far", 0)
			return
		end
	else
		ATMEFct.outputCore(nil, "Transport unit '" .. localMembers.name .. "' -> Loading group '" .. localMembersGroup.name .. "' to much units", 0)
		return
	end
end

ATMETimers.C_Unit_autoLoad = function(_datas)
	if ATMEDatas.exitATME == false then
		local localMembers = ATMEDatas.localMembers[_datas.transportUnit:getIndex()]
		
		if localMembers.objectDCS:isExist() == true then
			local localMembersGroup
			if _datas.groupToBoard == nil then
				_datas.groupToBoard = ATMEFct.C_Unit_getNearestReadyToBoardGroup(_datas.transportUnit, _datas.radius)
				
				if _datas.groupToBoard ~= nil then
					localMembersGroup = ATMEDatas.localMembers[_datas.groupToBoard:getIndex()]
						
					ATMEFct.outputCore(nil, "Transport unit '" .. localMembers.name .. "' -> Group '" .. localMembersGroup.name .. "' designed - Loading", 2)
					localMembers.idTimerLoad[localMembersGroup.name] = localMembers.idTimerLoadUndesigned[_datas.idGroup]
					localMembers.idTimerLoadUndesigned[_datas.idGroup] = nil
				end
			else
				localMembersGroup = ATMEDatas.localMembers[_datas.groupToBoard:getIndex()]
			end
			
			if _datas.groupToBoard ~= nil then
				if ATMEDatas.reservedBoardingGroup[localMembersGroup.name] == nil then
					ATMEDatas.reservedBoardingGroup[localMembersGroup.name] = localMembers
					ATMEFct.outputCore(nil, "Transport unit '" .. localMembers.name .. "' -> Group '" .. localMembersGroup.name .. "' reserved ok", 2)
				elseif ATMEDatas.reservedBoardingGroup[localMembersGroup.name] ~= localMembers then
					ATMEFct.outputCore(nil, "Transport unit '" .. localMembers.name .. "' -> Group '" .. localMembersGroup.name .. "' already reserved", 0)
					localMembers.idTimerLoad[localMembersGroup.name] = nil
					return
				end
			end
			
			if localMembersGroup ~= nil then
				if #localMembersGroup.objectDCS:getUnits() > 0 then
					local vector = ATMEFct.vectors["3D"].createFromTable(nil, nil, localMembers.objectDCS:getVelocity())
					if _datas.armed == false then
						if vector:getHModule() < 1 and (localMembers.objectDCS:hasAttribute("Ground vehicles") == true or 
								((localMembers.objectDCS:hasAttribute("Helicopters") == true or localMembers.objectDCS:hasAttribute("Planes") == true) and localMembers.objectDCS:inAir() == false)) then
							_datas.armed = true
							_datas.wait = 5
						end
					end
					
					if _datas.armed == true then
						if vector:getHModule() < 1 and (localMembers.objectDCS:hasAttribute("Ground vehicles") == true or 
								((localMembers.objectDCS:hasAttribute("Helicopters") == true or localMembers.objectDCS:hasAttribute("Planes") == true) and localMembers.objectDCS:inAir() == false)) then
							_datas.wait = _datas.wait - 1
							
							if _datas.wait == 0 then
								if ATMEDatas.readyToBoardGroups[localMembersGroup.name] ~= nil then
									ATMEFct.C_Unit_load_now(_datas)
									ATMEFct.outputCore(nil, "Transport unit '" .. localMembers.name .. "' boarding group '" .. localMembersGroup.name .. "'", 2)
									
									localMembers.idTimerLoad[localMembersGroup.name] = nil
									return
								end
							end
						else
							ATMEFct.outputCore(nil, "Transport unit '" .. localMembers.name .. "' moves -> No load for group '" .. localMembersGroup.name .. "'", 0)
							localMembers.idTimerLoad[localMembersGroup.name] = nil
							ATMEDatas.reservedBoardingGroup[localMembersGroup.name] = nil
							return
						end
					end
				else
					ATMEFct.outputCore(nil, "Transport unit '" .. localMembers.name .. "' -> Group '" .. localMembersGroup.name .. "' to load destroyed", 0)
					localMembers.idTimerLoad[localMembersGroup.name] = nil
					ATMEDatas.reservedBoardingGroup[localMembersGroup.name] = nil
					return
				end
			elseif _datas.armed == true then
				ATMEFct.outputCore(nil, "Transport unit '" .. localMembers.name .. "' -> No group found", 0)
				return
			end
		else
			ATMEFct.outputCore(nil, "Transport unit '" .. localMembers.name .. "' destroyed", 0)
			return
		end
	end
	
	return timer.getTime() + 3
end

ATMEFct.C_Unit_load = function(_transportUnit, _nbGroupsToBoard, _groupToBoard, _radius)
	local localMembers = ATMEDatas.localMembers[_transportUnit:getIndex()]
	local localMembersGroup
	
	if _groupToBoard ~= nil then
		localMembersGroup = ATMEDatas.localMembers[_groupToBoard:getIndex()]
	end
	
	if localMembersGroup == nil or localMembers.idTimerLoad[localMembersGroup.name] == nil then
		if localMembersGroup ~= nil then
			ATMEFct.outputCore(nil, "C_Unit_load - Start timer pour transport " .. localMembers.name .. " et chargement group " .. localMembersGroup.name, 2)
			local datas = {
				transportUnit = _transportUnit,
				radius = _radius,
				groupToBoard = _groupToBoard,
				wait = 1,
				armed = false,
			}
			
			localMembers.idTimerLoad[localMembersGroup.name] = timer.scheduleFunction(ATMETimers.C_Unit_autoLoad, datas, timer.getTime() + 1)
		else
			ATMEFct.outputCore(nil, "C_Unit_load - Start timer pour transport " .. localMembers.name .. " et chargement group non désigné", 2)
			
			for _i = 1, _nbGroupsToBoard do
				local datas = {
					transportUnit = _transportUnit,
					radius = _radius,
					idGroup = _i,
					wait = 1,
					armed = false,
				}
				
				localMembers.idTimerLoadUndesigned[_i] = timer.scheduleFunction(ATMETimers.C_Unit_autoLoad, datas, timer.getTime() + 1)
			end
		end
	end
end

ATMEFct.C_Unit_whichSide = function(_unit, _point)
	local vector = _unit:getRollAxisVector()
	local rollAxis = ATMEFct.lines["2D"].createFromVector(nil, nil, vector, _unit:getPosition())

	return ATMEFct.whichSideLine2D(rollAxis, _point)
end

ATMEFct.C_Unit_searchNearestReadyToBoardGroup = function(_unit, _datas)
	local localMembers = ATMEDatas.localMembers[ATMEDatas.units[_unit:getName()]:getIndex()]
	local localMembersGroup = ATMEDatas.localMembers[localMembers.group:getIndex()]
	
	if ATMEDatas.reservedBoardingGroup[localMembersGroup.name] == nil then
		local localMembersTransport = ATMEDatas.localMembers[_datas.transportUnit:getIndex()]
		if ATMEDatas.readyToBoardGroups[localMembersGroup.name] ~= nil and localMembersGroup.coalition == ATMEConst.coalitionNames[localMembersTransport.objectDCS:getCoalition()] and 
					localMembersGroup.boardingStarted == false and ATMEFct.isGroupActivated(localMembers.group) == true and 
					ATMEConst.personnelCarrierAttributes[localMembersTransport.typeName].maxUnits >= localMembersGroup.ATMEUnitNames:getCount() then

			local posUnit = localMembers.objectDCS:getPoint()
			local pos = localMembersTransport.objectDCS:getPoint()
			local d2 = (pos.x - posUnit.x) * (pos.x - posUnit.x) + (pos.z - posUnit.z) * (pos.z - posUnit.z)
			
			if _datas.minD2 == nil or _datas.minD2 > d2 then
				_datas.group = localMembers.group
				_datas.minD2 = d2
			end
		end
	end
end

ATMEFct.C_Unit_getNearestReadyToBoardGroup = function(_transportUnit, _radius)
	local datas = {
		transportUnit = _transportUnit,
	}

	local volume = {
		id = world.VolumeType.SPHERE,
		params = {
			point = ATMEDatas.localMembers[_transportUnit:getIndex()].objectDCS:getPoint(),
			radius = _radius,
		},
	}

	world.searchObjects(Object.Category.UNIT, volume, ATMEFct.C_Unit_searchNearestReadyToBoardGroup, datas)
	return datas.group
end

ATMEFct.C_Unit_addOnBoardGroup = function(_transportUnit, _datas)
	local localMembers = ATMEDatas.localMembers[_transportUnit:getIndex()]
	local id = localMembers.onBoardGroupDatas:add(_datas)

	ATMEDatas.groupsOnBoard[_datas.localMembers.name] = {
		transportUnit = _transportUnit,
		spawnDatas = _datas.localMembers.spawnDatas,
	}
	
	return id
end

ATMETimers.C_Unit_spot_razTimer = function(_localMembers)
	_localMembers.spotDatas.updateTimer = nil
end

ATMEFct.C_Unit_spot_update = function(_localMembers)
	local event = {
		type = "SPOT",
		datas = {
			action = _localMembers.spotDatas.action,
			unitName = _localMembers.name,
			targetName = _localMembers.spotDatas.targetName,
			targetClassName = _localMembers.spotDatas.targetClassName,
			code = _localMembers.spotDatas.code,
			vector = _localMembers.spotDatas.vector,
		},
	}
	
	ATMEFct.fireEvent(event)
	
	_localMembers.spotDatas.updateTimer = timer.scheduleFunction(ATMETimers.C_Unit_spot_razTimer, _localMembers, timer.getTime() + 5)
end

ATMEFct.C_Unit_spotAssign = function(_localMembers)
	if ATMEDatas.exitATME == false then
		local targetName
		local targetClassName
		
		if _localMembers.spotDatas.groupName ~= nil and ATMEDatas.groups[_localMembers.spotDatas.groupName] ~= nil then
			_localMembers.spotDatas.targetName, _localMembers.spotDatas.targetClassName = ATMEFct.C_Unit_findTargetInGroup(ATMEDatas.groups[_localMembers.spotDatas.groupName], _localMembers.spotDatas.ignoredTargets)
		elseif _localMembers.spotDatas.zone ~= nil then
			if _localMembers.spotDatas.targetName ~= nil and ATMEFct.C_Event_target_class[_localMembers.spotDatas.targetClassName](_localMembers.spotDatas.targetName) ~= nil then
				if _localMembers.spotDatas.outOfZone == true then
					ATMEFct.outputCore(nil, "Spot par unité " .. _localMembers.name .. " : unité " .. _localMembers.spotDatas.targetName .. " -> Hors zone", 2)
					local event = {
						type = "SPOT",
						datas = {
							action = "TARGET_OUT_OF_ZONE",
							unitName = _localMembers.name,
							targetName = _localMembers.spotDatas.targetName,
							targetClassName = _localMembers.spotDatas.targetClassName,
						},
					}
					
					ATMEFct.fireEvent(event)
				elseif _localMembers.spotDatas.noLOS == 50 and _localMembers.spotDatas.neverHaveLOS == false then
					ATMEFct.outputCore(nil, "Spot par unité " .. _localMembers.name .. " : unité " .. _localMembers.spotDatas.targetName .. " -> Plus de visée", 2)
					local event = {
						type = "SPOT",
						datas = {
							action = "TARGET_NO_LOS",
							unitName = _localMembers.name,
							targetName = _localMembers.spotDatas.targetName,
							targetClassName = _localMembers.spotDatas.targetClassName,
						},
					}
					
					ATMEFct.fireEvent(event)
				end
			else
				_localMembers.spotDatas.targetClassName = "C_Unit"
			end

			_localMembers.spotDatas.targetName, _localMembers.spotDatas.targetClassName = ATMEFct.C_Unit_findTargetInZone2D(_localMembers.spotDatas.zone, ATMEConst.coalitionEnemy[ATMEConst.coalitionNames[_localMembers.objectDCS:getCoalition()]], _localMembers.spotDatas.ignoredTargets)
		end

		if _localMembers.spotDatas.targetName ~= nil then
			local point = ATMEDatas.localMembers[ATMEFct.C_Event_target_class[_localMembers.spotDatas.targetClassName](_localMembers.spotDatas.targetName):getIndex()].objectDCS:getPoint()
			local posUnit = _localMembers.objectDCS:getPoint()
			
			
			_localMembers.spotDatas.vector =  ATMEFct.vectors["3D"].createFromReference(nil, nil, posUnit, point)
			_localMembers.spotDatas.noLOS = 0
			_localMembers.spotDatas.ignoredTargetOn = false
			
			local offset = {x = 0, y = -2, z = 0}
			if _localMembers.spotDatas.fromAir == false then
				offset = {x = 0, y = 2, z = 0}
			end
					
			if _localMembers.spotDatas.code ~= nil then
				_localMembers.spotDatas.spot = Spot.createLaser(_localMembers.objectDCS,
																offset, 
																point, 
																_localMembers.spotDatas.code)
				ATMEFct.outputCore(nil, "Spot LASER par unité " .. _localMembers.name .. " : ciblage unité " .. _localMembers.spotDatas.targetName, 2)
			else
				_localMembers.spotDatas.spot = Spot.createInfraRed(_localMembers.objectDCS,
															offset, 
															point)
				ATMEFct.outputCore(nil, "Spot IR par unité " .. _localMembers.name .. " : ciblage unité " .. _localMembers.spotDatas.targetName, 2)
			end
			
			_localMembers.spotDatas.action = "TARGET_NEW"
			ATMEFct.C_Unit_spot_update(_localMembers)
		else
			ATMEFct.outputCore(nil, "Spot par unité " .. _localMembers.name .. " : plus d'unités à cibler. Arrêt", 2)
		
			local event = {
				type = "SPOT",
				datas = {
					action = "NO_TARGET",
					unitName = _localMembers.name,
				},
			}

			ATMEFct.fireEvent(event)		
		end
	end
end

ATMETimers.C_Unit_spot_timer = function(_localMembers)	
	if ATMEDatas.exitATME == false then
		if _localMembers.objectDCS:isExist() == true then	
			-- Suppression des spots actifs pour les unités AI au sol qui se mettent à bouger
			-- Tracking si pointeur sur objet dynamique
			
			local vector = ATMEFct.vectors["3D"].createFromTable(nil, nil, _localMembers.objectDCS:getVelocity())

			if _localMembers.spotDatas.fromAir == false and vector:getModule() > 0.01 and _localMembers.spotDatas.spot ~= nil then
				-- L'unité au sol (JTAC) se déplace, arrêt du spot
				
				_localMembers.spotDatas.spot:destroy()
				_localMembers.spotDatas.spot = nil
				
				local event = {
					type = "SPOT",
					datas = {
						action = "UNIT_MOVING",
						unitName = _localMembers.name,
					},
				}
				ATMEFct.outputCore(nil, "Spot par unité " .. _localMembers.name .. " : Arrêt cause déplacement", 2)		
				ATMEFct.fireEvent(event)
			else
				if _localMembers.spotDatas.spot == nil then -- Pas de spot actif
					ATMEFct.outputCore(nil, "Cible pas de spot actif, recherche suivante", 2)
					ATMEFct.C_Unit_spotAssign(_localMembers)
					
					if _localMembers.spotDatas.spot ~= nil then
						return timer:getTime() + 3 + math.random(1,5)
					end
				elseif _localMembers.spotDatas.spot ~= nil then -- Un spot est actif
					if ATMEFct.C_Event_target_class[_localMembers.spotDatas.targetClassName](_localMembers.spotDatas.targetName) ~= nil and _localMembers.spotDatas.ignoredTargetOn == false then
						local posUnit = _localMembers.objectDCS:getPoint()
						
						local localMembersTarget = ATMEDatas.localMembers[ATMEFct.C_Event_target_class[_localMembers.spotDatas.targetClassName](_localMembers.spotDatas.targetName):getIndex()]
						local point = localMembersTarget.objectDCS:getPoint()
						local vectorTarget = ATMEFct.vectors["3D"].createFromTable(nil, nil, localMembersTarget.objectDCS:getVelocity())
						
						if (_localMembers.spotDatas.fromAir == false and land.isVisible({x = posUnit.x, y = posUnit.y + 2, z = posUnit.z} , {x = point.x, y = point.y + 2, z = point.z}) == false) or 
						   (_localMembers.spotDatas.fromAir == true and land.isVisible({x = posUnit.x, y = posUnit.y - 2, z = posUnit.z}  , {x = point.x, y = point.y + 2, z = point.z}) == false) then
							_localMembers.spotDatas.noLOS = _localMembers.spotDatas.noLOS + 1
						else
							_localMembers.spotDatas.neverHaveLOS = false
							_localMembers.spotDatas.noLOS = 0
							_localMembers.spotDatas.ignoredTargets[_localMembers.spotDatas.targetName] = nil
						end
						
						if _localMembers.spotDatas.noLOS == 50 then -- Pas de ligne de visée pendant 5 secondes. Abandon
							_localMembers.spotDatas.spot:destroy()
							_localMembers.spotDatas.spot = nil
							_localMembers.spotDatas.targetName = nil
							_localMembers.spotDatas.targetClassName = nil
							_localMembers.spotDatas.ignoredTargets[_localMembers.spotDatas.targetName] = true
							ATMEFct.outputCore(nil, "Cible sans visée depuis 5s, recherche suivante", 2)
							ATMEFct.C_Unit_spotAssign(_localMembers)
							
							if _localMembers.spotDatas.spot ~= nil then
								return timer:getTime() + 3 + math.random(1,5)
							end
						elseif _localMembers.spotDatas.dynamicSpot == true and (vectorTarget:getModule() > 0.01 or _localMembers.spotDatas.fromAir == true) then
							local datas
							
							if _localMembers.spotDatas.zone ~= nil then
								local localMembers = ATMEDatas.localMembers[_localMembers.spotDatas.zone:getIndex()]
								datas = {
									zone = _localMembers.spotDatas.zone,
									actualCenter = localMembers.center,
									actualAzimuth = localMembers.azimuth,
								}
							end
							
							if _localMembers.spotDatas.zone == nil or 
								(datas ~= nil and ATMEFct.C_FindContext_Object_isInZone(ATMEDatas.units[_localMembers.spotDatas.targetName], datas) == true) then
								-- La cible existe et elle est dynamique... Le spot suit la cible.
								ATMEFct.outputCore(nil, "Cible dynamique", 2)
								_localMembers.spotDatas.vector =  ATMEFct.vectors["3D"].createFromReference(nil, nil, posUnit, point)
								_localMembers.spotDatas.spot:setPoint(point)
								
								if _localMembers.spotDatas.updateTimer == nil then
									_localMembers.spotDatas.action = "TARGET_UPDATE"
									ATMEFct.C_Unit_spot_update(_localMembers)
								end
								
								return timer:getTime() + 0.1
							else -- La cible est sortie de la zone de surveillance
								_localMembers.spotDatas.spot:destroy()
								_localMembers.spotDatas.spot = nil
								_localMembers.spotDatas.targetName = nil
								_localMembers.spotDatas.targetClassName = nil
								_localMembers.spotDatas.outOfZone = true
								ATMEFct.outputCore(nil, "Cible sortie surveillance, recherche suivante", 2)
								ATMEFct.C_Unit_spotAssign(_localMembers)
								
								if _localMembers.spotDatas.spot ~= nil then
									return timer:getTime() + 3 + math.random(1,5)
								end
							end
						else
							if _localMembers.spotDatas.spot ~= nil then
								return timer:getTime() + 3
							end
						end
					else -- La cible n'existe plus
						_localMembers.spotDatas.spot:destroy()
						_localMembers.spotDatas.spot = nil
						_localMembers.spotDatas.targetName = nil
						_localMembers.spotDatas.targetClassName = nil
						ATMEFct.outputCore(nil, "Cible détruite, recherche suivante",2)
						ATMEFct.C_Unit_spotAssign(_localMembers)

						if _localMembers.spotDatas.spot ~= nil then
							return timer:getTime() + 3 + math.random(1,5)
						end
					end
				end
			end
		end
	end
end

ATMETimers.C_Unit_targetTracking = function(_localMembers)
	if ATMEDatas.exitATME == false then
		if _localMembers.objectDCS:isExist() == true then
			local targets = ATMEFct.C_Unit_getDetectedTargets(_localMembers, _localMembers.targetsTracking.typeOfSensorName)
			for _targetName, _ in pairs(targets) do
				if targets[_targetName].stillDetected == false then
					if targets[_targetName].trackingOn == true then
						local event = {
							type = "SENSOR",
							datas = {
								action = "TRACK_END",
								unitName = _localMembers.name,
								targetName = _targetName,
								targetClassName = targets[_targetName].className,
							},
						}

						targets[_targetName] = nil
						ATMEFct.fireEvent(event)
					end
				else
					if targets[_targetName].trackingOn == false and (_localMembers.targetsTracking.flag == nil or ATME.getFlag(_localMembers.targetsTracking.flag) ~= 0) then	
						-- Event nouvelle target dans le tracking confirmée
						
						targets[_targetName].trackingOn = true
						local event = {
							type = "SENSOR",
							datas = {
								action = "TRACK_NEW",
								unitName = _localMembers.name,
								targetName = _targetName,
								targetClassName = targets[_targetName].className,
							},
						}

						ATMEFct.fireEvent(event)
					elseif targets[_targetName].trackingOn == true and _localMembers.targetsTracking.flag ~= nil and ATME.getFlag(_localMembers.targetsTracking.flag) == 0 then
						targets[_targetName].trackingOn = false
						local event = {
							type = "SENSOR",
							datas = {
								action = "TRACK_END",
								unitName = _localMembers.name,
								targetName = _targetName,
								targetClassName = targets[_targetName].className,
							},
						}
						
						ATMEFct.fireEvent(event)
					end
				end
			end
			
			return timer.getTime() + 1
		else		
			_localMembers.targetsTracking.idTimer = nil 
		end
	end
end

ATMEFct.C_Unit_startTransmission = function(_radioTransmission)
	if _radioTransmission.started == false and _radioTransmission.playlist[_radioTransmission.currentReadIndex].pause == false then
		local controller = ATMEDatas.units[_radioTransmission.unitName]:getDCSObject():getController()

		local commandFreq = { 
			id = "SetFrequency", 
			params = {
				frequency = _radioTransmission.localMembers.frequency, 
				modulation = ATMEConst.modulationValues[_radioTransmission.localMembers.modulation],
			},
		}
		
		controller:setCommand(commandFreq)
		
		local command = { 
			id = "TransmitMessage", 
			params = {
				file = "l10n/DEFAULT/" .. _radioTransmission.localMembers.currentVoice .. _radioTransmission.playlist[_radioTransmission.currentReadIndex].fileName,
				duration = _radioTransmission.playlist[_radioTransmission.currentReadIndex].subtitle.duration,
				subtitle = _radioTransmission.playlist[_radioTransmission.currentReadIndex].subtitle.text,
				loop = false,
			} 
		}
		controller:setCommand(command)
		_radioTransmission.started = true
	end
end

ATMEFct.C_Unit_stopTransmission = function(_radioTransmission)
	if _radioTransmission.started == true then
		local controller = ATMEDatas.units[_radioTransmission.unitName]:getDCSObject():getController()
		
		local command = { 
			id = "StopTransmission", 
			params = {},
				
		}
		
		controller:setCommand(command)
		_radioTransmission.started = false
	end
end

ATMEFct.C_Unit_loopAttributes = function(_unit, _datas)
	local datas = Unit.getDesc(_unit:getDCSObject())

	for _id, _ in pairs(datas.attributes) do
		if ATMEConst.backTargetsAttributes[_id] ~= nil then
			if _datas.currentPriority < ATMEConst.backTargetsAttributes[_id].priority then
				_datas.currentPriority = ATMEConst.backTargetsAttributes[_id].priority
				_datas.targetName = _unit:getName()
				_datas.targetClassName = _unit:getClassName()
			end
		end
	end
end

ATMEFct.C_Unit_isTargetOk = function(_unit, _ignoredTargets)
	local localMembers = ATMEDatas.localMembers[_unit:getIndex()]
	local localMembersGroup = ATMEDatas.localMembers[localMembers.group:getIndex()]
	local active = localMembers.objectDCS:isActive()
	local ok = false
	
	if _ignoredTargets[localMembers.name] ~= nil then
		ok = true
	end
	
	return (localMembers.category == "GROUND" or 
			(ATMEConst.canFly[localMembers.category] == true and localMembers.objectDCS:inAir() == false)) and 
			(active == true or (active == false and localMembersGroup.alwaysVisible == true)) and ok == false
end

ATMEFct.isTargetOk = {
	["C_Unit"] = function(_unit, _ignoredTargets)
		return ATMEFct.C_Unit_isTargetOk(_unit, _ignoredTargets)
	end,
	["C_Static"] = function(_object, _ignoredTargets)
		local ok = false
		
		if _ignoredTargets[_object:getName()] ~= nil then
			ok = true
		end
		
		return ok == false
	end,
	["C_Cargo"] = function(_object, _ignoredTargets)
		local ok = false
		
		if _ignoredTargets[_object:getName()] ~= nil then
			ok = true
		end
		
		return ok == false
	end,
}

ATMEFct.C_Unit_searchTargetsInBox = function(_unit, _datas)
	local unit = ATMEDatas.units[_unit:getName()]
	
	if ATMEFct.isTargetOk[unit:getClassName()](unit, _datas.ignoredTargets) == true and ATMEConst.coalitionNames[_unit:getCoalition()] == _datas.targetsCoalition then
		if ATMEFct.C_FindContext_Object_isInZone(unit, _datas) == true then
			ATMEFct.C_Unit_loopAttributes(unit, _datas)
		end
	end
end

ATMEFct.C_Unit_findTargetInGroup = function(_group , _ignoredTargets)
	local localMembers = ATMEDatas.localMembers[_group:getIndex()]
	local units = localMembers.objectDCS:getUnits()
	
	local datas = {
		currentPriority = -1,
	}
	
	for _, _unit in ipairs(units) do
		local unit = ATMEDatas.units[_unit:getName()]
		if ATMEFct.isTargetOk[unit:getClassName()](unit, _ignoredTargets) == true then
			local found = false
			local unitName = unit:getName()
			
			ATMEFct.C_Unit_loopAttributes(unit, datas)
			if datas.targetName == nil then
				datas.targetName = unitName
				datas.targetClassName = unit:getClassName()
			end
		end
	end
		
	return datas.targetName, datas.targetClassName
end

ATMEFct.C_Unit_findTargetInZone2D = function(_zone, _targetsCoalition, _ignoredTargets)
	local localMembers = ATMEDatas.localMembers[_zone:getIndex()]
	local volume = {
		id = world.VolumeType.BOX,
		params = localMembers.box,
	}
	
	local datas = {
		currentPriority = -1,
		zone = _zone,
		targetsCoalition = _targetsCoalition,
		ignoredTargets = _ignoredTargets,
		actualCenter = localMembers.center,
		actualAzimuth = localMembers.azimuth,
	}

	world.searchObjects(Object.Category.UNIT, volume, ATMEFct.C_Unit_searchTargetsInBox, datas)		
	
	return datas.targetName, datas.targetClassName
end

ATMEFct.C_Unit_getDetectedTargets = function(_localMembers, _typeOfSensorName)
	local controller = _localMembers.objectDCS:getController()
	local detectedTargetsList
	
	if _typeOfSensorName ~= nil then
		detectedTargetsList = controller:getDetectedTargets(ATMEConst.typesOfSensor[_typeOfSensorName])
	else
		detectedTargetsList = controller:getDetectedTargets()
	end

	for _, _target in pairs(_localMembers.targets) do
		_target.stillDetected = false
	end

	for _, _target in ipairs(detectedTargetsList) do
		if _target.object ~= nil then
			local targetName = _target.object:getName()
			if ATMEDatas.units[targetName] ~= nil then
				-- Test si la target est bien une unité ou un joueur
				
				if _localMembers.targets[targetName] ~= nil then
					_localMembers.targets[targetName].stillDetected = true
				else
					_localMembers.targets[targetName] = {
						className = ATMEDatas.units[targetName]:getClassName(),
						stillDetected = true,
						trackingOn = false,
					}
				end
			end
		end
	end
	
	return _localMembers.targets
end

ATMEFct.C_Unit_playerEnters = function(_unit)
	local localMembers = ATMEDatas.localMembers[_unit:getIndex()]
	local playerName = localMembers.objectDCS:getPlayerName()
	
	if playerName ~= nil then
		localMembers.player = playerName
		ATMEDatas.players[localMembers.name] = _unit
		
		ATMEFct.C_Group_addPlayer(localMembers.group)
	end
end

ATMEFct.C_Unit_playerLeaves = function(_unit, _localMembersGroup)
	local localMembers = ATMEDatas.localMembers[_unit:getIndex()]
	
	if localMembers.player ~= nil then
		localMembers.player = nil
		ATMEDatas.players[localMembers.name] = nil
		ATMEFct.C_Group_deletePlayer(_localMembersGroup)
	end
end

ATMEDatas.C_Unit_MT = {
	className = "C_Unit",

	mtModules = {},
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["getDCSObject"] = true,
		["getDCSID"] = true,
		["getName"] = true,
		["getAirbase"] = true,
		["getAttributes"] = true,
		["hasAttributes"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["getTypeName"] = true,
		["getGroup"] = true,
		["getPseudo"] = true,
		["isInDirection"] = true,
		["disable"] = true,
		["isDisabled"] = true,
		["setVoice"] = true,
		["getVoice"] = true,
		["getPosition"] = true,
		["getVelocityVector"] = true,
		["getRollAxisVector"] = true,
		["getFuelRatio"] = true,
		["getLife"] = true,
		["getLifeRatio"] = true,
		["getCallsign"] = true,
		["getSpeed"] = true,
		["getHSpeed"] = true,
		["getVSpeed"] = true,
		["getSpeedAzimuth"] = true,
		["getAzimuth"] = true,
		["getCountryName"] = true,
		["getCoalitionName"] = true,
		["getCategoryName"] = true,
		["isActivated"] = true,
		["isHelicopter"] = true,
		["isPlane"] = true,
		["isPersonnelCarrier"] = true,
		["isInfantry"] = true,
		["isGroundVehicle"] = true,
		["isAirDefence"] = true, 
		["isManPad"] = true,
		["isSAMVehicle"] = true,
		["isSAMSiteLauncher"] = true,
		["isSAMSiteRadar"] = true,
		["isSAMSiteCommandCenter"] = true,
		["isAAA"] = true,
		["inAir"] = true,
		["getMSLAltitude"] = true,  
		["getAGLAltitude"] = true,
		["explode"] = true,
		["isInZones"] = true,
		["isInDCSZone"] = true,
		["isInZone"] = true,
		["isPilotDead"] = true,
		["isPilotEjected"] = true,
		["isEngineStarted"] = true,
		["isPlayer"] = true,
		["whichSide"] = true,
		["getNbUnitsOnBoard"] = true,
		["getOnBoardGroupNames"] = true,
		["load"] = true,
		["unload"] = true,
		["getNearestReadyToBoardGroup"] = true,
		["fireFlare"] = true,
		["fireSmoke"] = true,
		["fireIlluminationBomb"] = true,
		["modules"] = true,
		["canFly"] = true,
		["getUserPrivateDatas"] = true,
		["spotActivate"] = true,
		["spotDesactivate"] = true,
		["spotIgnoreTarget"] = true,
		["assignTarget"] = true,
		["sensorStart"] = true,
		["sensorStop"] = true,
		["beaconOn"] = true,
		["beaconOff"] = true,
		["setFrequency"] = true,
		["addEventCallback"] = true,
		["removeEventCallback"] = true,
		["getModelDatas"] = true,
		["createStaticDatas"] = true,
		["radarOn"] = true,
		["radarOff"] = true,
		["pushTask"] = true,
		["popTask"] = true,
		["setTask"] = true,
		["resetTask"] = true,
	},

	add = function(self, _unit, _spawnDatas)
		local name = _unit:getName()
		local item = {
			name = name,
			objectDCS = _unit,
			objectDCSId = _unit:getID(),
			typeName = _unit:getTypeName(),
			engineOn = false,
			nbUnitsOnBoard = 0,
			nbBoardings = 0,
			groupsBoarding = {},
			spotDatas = {},
			onBoardGroupDatas = ATMEFct.C_List_create(),
			idTimerLoad = {},
			idTimerLoadUndesigned = {},
			isEnabled = true,
			currentVoice = "",
			userPrivateDatas = {},						-- Point d'entrée des données privées : un par module
			eventCallbacks = {},						-- Callbacks d'événement
			targets = {},								-- Table des cibles trackées à un instant T
			tagsName = ATMEFct.decompString(name),
			pilotDead = false,
			pilotEjected = false,
			hasTakenOff = false,
			destroyed = false;
			spawnDatas = _spawnDatas[name],
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].eventCallbacks = nil
		ATMEDatas.localMembers[_id].targets = nil
		ATMEDatas.localMembers[_id].idTimerLoad = nil
		ATMEDatas.localMembers[_id].idTimerLoadUndesigned = nil
	end,
}

ATME.C_Unit = {
	getClassName = function(self)
		return ATMEDatas.C_Unit_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Unit_MT.parentClass
	end,
	
	getByName = function(_unitName)
		if type(_unitName) == "string" then
			return ATMEDatas.units[_unitName]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Unit.getByName(name)", "name type must be string", false)
		end	
	end,
	
	foreach = function(_callback, _datas)
		local fctDefinition = "Function ATME.C_Unit.foreach(callback, [datas])"
		
		if type(_callback) == "function" then
			ATMEFct.foreachObject(nil, "units", _callback, _datas)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "callback type must be function", false)
		end
	end,
	
	foreachInCoalition = function(_coalitionName, _callback, _datas)
		local fctDefinition = "Function ATME.C_Unit.foreachInCoalition(coalitionName, callback, [datas])"
		
		if type(_coalitionName) == "string" and type(_callback) == "function" then
			if ATMEConst.coalitionIdsBlueRed[_coalitionName] == nil then
				ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName must be 'BLUE' or 'RED'", false)
				return
			end
			
			ATMEFct.foreachObject(_coalitionName, "units", _callback, _datas)
		else
			if type(_coalitionName) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName type must be string", false)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "callback type must be function", false)
			end
		end
	end,
	
	foreachPlayer = function(_callback, _datas)
		local fctDefinition = "Function ATME.C_Unit.foreachPlayer(callback, [datas])"
		
		if type(_callback) == "function" then
			ATMEFct.foreachObject(nil, "players", _callback, _datas)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "callback type must be function", false)
		end
	end,
	
	foreachPlayerInCoalition = function(_coalitionName, _callback, _datas)
		local fctDefinition = "Function ATME.C_Unit.foreachPlayerInCoalition(coalitionName, callback, [datas])"
		
		if type(_coalitionName) == "string" and type(_callback) == "function" then
			if ATMEConst.coalitionIdsBlueRed[_coalitionName] == nil then
				ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName must be 'BLUE' or 'RED'", false)
				return
			end
			
			ATMEFct.foreachObject(_coalitionName, "players", _callback, _datas)
		else
			if type(_coalitionName) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName type must be string", false)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "callback type must be function", false)
			end
		end
	end,
	
	getNbPlayers = function()
		return ATMEDatas.nbPlayers
	end,
	
	getRandomPlayer = function()
		return ATMEFct.getRandomPlayer()
	end,
	
	getRandomPlayerInCoalition = function(_coalitionName)
		if type(_coalitionName) == "string" then
			if ATMEConst.coalitionIdsBlueRed[_coalitionName] == nil then
				ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName must be 'BLUE' or 'RED'", false)
				return
			end
			
			return ATMEFct.getRandomPlayer(ATMEConst.coalitionIdsBlueRed[_coalitionName])
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName type must be string", false)
		end
	end,
}

ATMEFct.C_Unit_create = function(_unit, _datas, _newObject, _groupExist)
	local newUnit = {}

	setmetatable(newUnit, ATMEDatas.C_Unit_MT.mt)
	local index = ATMEDatas.C_Unit_MT:add(_unit, _datas.spawnDatas.unitsByName)
	local localMembers = ATMEDatas.localMembers[index]

	function newUnit.getIndex(self)
		return index
	end
	
	function newUnit.getAirbase(self)
		return linkedAirbase
	end
	
	function newUnit.getDCSObject(self)
		return localMembers.objectDCS
	end
	
	function newUnit.isPlayer(self)
		return localMembers.player ~= nil
	end
	
	function newUnit.getDCSID(self)
		return localMembers.objectDCSId
	end
	
	function newUnit.getModelDatas(self)
		return {
			type = localMembers.spawnDatas.type,
			livery_id = localMembers.spawnDatas.livery_id,
		}
	end
	
	function newUnit.getPseudo(self)
		return localMembers.player
	end
	
	function newUnit.isInDirection(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "isInDirection(reference)"
		
		if ATMEFct.isReference(_reference) == true then
			local vector = ATMEFct.vectors["3D"].createFromTable(nil, nil, self:getVelocityVector())
			local azimuth = 0
			
			if vector:getModule() > 1 then 
				azimuth =  vector:getAzimuth()
			end
			
			return ATMEFct.isAzimuthInLimits(localMembers.objectDCS:getPoint(), azimuth, ATMEFct.getPointFromReference(_reference))
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
		end
	end
	
	function newUnit.createStaticDatas(self, _name)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "createStaticDatas([_name])"
		
		if _name == nil or type(_name) == "string" then
			if localMembers.category ~= "SHIP" then
				return {
					country = country.name[localMembers.objectDCS:getCountry()],
					datas = {
						name = _name,
						type = localMembers.spawnDatas.type,
						livery_id = localMembers.spawnDatas.livery_id,
						category = localMembers.category,
					},
				}
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "staticDatas can't be created from ships", true)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be nil or string", true)
		end
	end
	
	function newUnit.getName(self)
		return localMembers.name
	end
	
	function newUnit.isActivated(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:isActive()
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "isActivated()", "already disabled", false)
		end
	end
	
	function newUnit.getCategoryName(self)
		if localMembers.destroyed == false then
			return localMembers.category
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "getCategoryName()", "already disabled", false)
		end
	end
	
	function newUnit.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if localMembers.tagsName ~= nil and pos >= 1 and pos <= #localMembers.tagsName then
				return localMembers.tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newUnit.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newUnit.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if localMembers.tagsName ~= nil then
				local ok = ATMEFct.compareTags(localMembers.tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newUnit.getTypeName(self)
		return localMembers.typeName
	end
	
	function newUnit.getGroup(self)
		return localMembers.group
	end
	
	function newUnit.getUserPrivateDatas(self, _moduleName)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "getUserPrivateDatas([moduleName])"
		local moduleName
		
		if _moduleName ~= nil then
			if type(_moduleName) == "string" then
				if localMembers.userPrivateDatas[_moduleName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
					return
				else
					moduleName = _moduleName
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
				return
			end
		else
			moduleName = ATMEDatas.currentModuleName
		end
		
		return localMembers.userPrivateDatas[moduleName]
	end
	
	function newUnit.getAttributes(self)
		if localMembers.destroyed == false then
			local datas = Unit.getDesc(localMembers.objectDCS)
			local attributes = {}
			for _id, _ in pairs(datas.attributes) do
				if ATMEConst.backTargetsAttributes[_id] ~= nil then
					table.insert(attributes, ATMEConst.backTargetsAttributes[_id].value)
				end
			end
			
			return attributes
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "getAttributes()", "already disabled", false)
		end
	end
	
	function newUnit.isEngineStarted(self)
		if localMembers.destroyed == false then
			local localMembersGroup = ATMEDatas.localMembers[localMembers.group:getIndex()]
			if ATMEConst.canFly[localMembersGroup.category] == true then
				return localMembers.engineOn
			else
				return true
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "isEngineStarted()", "already disabled", false)
		end
	end
	
	function newUnit.hasAttributes(self, _attributes)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "hasAttributes(attributes)"
		
		if localMembers.destroyed == false then
			if type(_attributes) == "string" or type(_attributes) == "table" then
				local err, datas = ATMEFct.extractDatasFromStringOrTable(_attributes)
		
				if err == false then
					return ATMEFct.hasAttributes(localMembers.objectDCS, datas)
				end
				
				ATMEFct.errorOutputCore(nil, fctDefinition, datas, false)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "attributes type must be string or table", true)
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end
	
	function newUnit.addEventCallback(self, _fct, _eventType)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "addEventCallback(fct, [eventType])"
		
		if localMembers.destroyed == false then
			if type(_fct) == "function" and (_eventType == nil or type(_eventType) == "string") then
				if _eventType == nil then
					for _typeName, _ in pairs(ATMEConst.coreEventTypes) do
						localMembers.eventCallbacks[ATMEDatas.currentModuleName][_typeName] = _fct
					end
				else
					local eventType = string.upper(_eventType)
					if ATMEConst.coreEventTypes[eventType] ~= nil then
						localMembers.eventCallbacks[ATMEDatas.currentModuleName][eventType] = _fct
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "eventType must be an existing event type (see manual)", false)
					end
				end
			else
				if type(_fct) ~= "function" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
				elseif type(_eventType) ~= "string" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "eventType type must be nil or string", false)
				end
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end
	
	function newUnit.removeEventCallback(self, _eventType)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "removeEventCallback([eventType])"
		
		if localMembers.destroyed == false then
			if _eventType == nil or type(_eventType) == "string" then
				if _eventType == nil then
					for _typeName, _ in pairs(ATMEConst.coreEventTypes) do
						localMembers.eventCallbacks[ATMEDatas.currentModuleName][_typeName] = nil
					end
				else
					local eventType = string.upper(_eventType)
					if ATMEConst.coreEventTypes[eventType] ~= nil then
						localMembers.eventCallbacks[ATMEDatas.currentModuleName][eventType] = nil
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "eventType must be an existing event type (see manual)", false)
					end
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "eventType type must be nil or string", true)
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end
	
	function newUnit.assignTarget(self, _target)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "assignTarget(target)"
		
		if localMembers.destroyed == false then
			if type(_target) == "string" or ATMEFct.isObjectClass(_target, ATME.C_Unit) == true then
				local unit = _target
				if type(_target) == "string" then
					unit = ATMEDatas.units[_target]
				end
				
				if unit ~= nil then
					local controller = ATMEDatas.localMembers[_ATMEUnit:getIndex()].objectDCS:getController()
					local localMembersTarget = ATMEDatas.localMembers[unit:getIndex()]
					
					controller:knowTarget(localMembersTarget.objectDCS, true, true)
				else
					ATMEFct.setError("NOT_TARGET")
					ATMEFct.warningOutputCore(nil, fctDefinition, "not target", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "target type must be string or ATME.C_Unit", true)
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end
	
	function newUnit.sensorStart(self, _typeOfSensorName, _flag)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "sensorStart(typeOfSensorName, [flag])"

		if localMembers.destroyed == false then
			if type(_typeOfSensorName) == "string" and (type(_flag) == "number" or type(_flag) == "string" or _flag == nil) then
				if ATMEConst.typesOfSensor[_typeOfSensorName] ~= nil then
					if localMembers.targetsTracking == nil then
						localMembers.targetsTracking = {
							flag = _flag,
							unit = self,
						}
						if _typeOfSensorName ~= "ALL" then
							localMembers.targetsTracking.typeOfSensorName = _typeOfSensorName
						end

						localMembers.targetsTracking.idTimer = timer.scheduleFunction(ATMETimers.C_Unit_targetTracking, localMembers, timer.getTime() + 1)
					else
						setError("ALREADY_TRACKING")
						ATMEFct.warningOutputCore(nil, fctDefinition, "sensor already activated", false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "typeOfSensorName must be 'ALL', 'RADAR', 'VISUAL', 'OPTIC', 'IRST', 'RWR', 'DLINK'", false)
				end
			else
				if type(_typeOfSensorName) ~= "string" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "typeOfSensorName type must be string", true)
				elseif type(_flag) ~= "number" and type(_flag) ~= "string" and _flag ~= nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "flag type must be nil or string or number", false)
				end
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end
	
	function newUnit.sensorStop(self)
		if localMembers.destroyed == false then
			if localMembers.targetsTracking ~= nil then
				if localMembers.targetsTracking.idTimer ~= nil then
					timer.removeFunction(localMembers.targetsTracking.idTimer)
				end
			
				localMembers.targetsTracking = nil
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "sensorStop()", "already disabled", false)
		end
	end
	
	function newUnit.setTask(self, _task)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "setTask(task)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isObjectClass(_task, ATME.C_Task) == true then
				local localMembersGroup = ATMEDatas.localMembers[localMembers.group:getIndex()]
				if ATMEConst.canFly[localMembersGroup.category] == true then 
					local ok, taskDatas = ATMEFct.C_Task_getTaskDatas(_task, localMembersGroup.spawnDatas.type)
					if taskDatas.params.startCondition ~= nil then
						ATMEFct.outputCore(nil, "Group '" .. localMembers.name .. "' : setTask - Start condition ignored", 0)
					end
					
					taskDatas.params.startCondition = nil
					if ok == true then
						local controller = localMembers.objectDCS:getController()
						controller:setTask(taskDatas)
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, taskDatas .. " for unit " .. localMembers.name, false)
					end
				else
					ATMEFct.setError("NOT AIRCRAFT")
					ATMEFct.warningOutputCore(nil, fctDefinition, "not aircraft group", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "task type must be ATME.C_Task", true)
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end
	
	function newUnit.resetTask(self)
		if localMembers.destroyed == false then
			if ATMEConst.canFly[ATMEDatas.localMembers[localMembers.group:getIndex()].category] == true then 
				local controller = localMembers.objectDCS:getController()

				controller:resetTask()
			else
				ATMEFct.setError("NOT AIRCRAFT")
				ATMEFct.warningOutputCore(nil, fctDefinition, "not aircraft group", false)
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "resetTask()", "already disabled", false)
		end
	end
	
	function newUnit.pushTask(self, _task)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "pushTask(task)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isObjectClass(_task, ATME.C_Task) == true then
				local localMembersGroup = ATMEDatas.localMembers[localMembers.group:getIndex()]
				if ATMEConst.canFly[localMembersGroup.category] == true then 
					local ok, taskDatas = ATMEFct.C_Task_getTaskDatas(_task, localMembersGroup.spawnDatas.type)
					if taskDatas.params.startCondition ~= nil then
						ATMEFct.outputCore(nil, "Group '" .. localMembers.name .. "' : pushTask - Start condition ignored", 0)
					end
					
					taskDatas.params.startCondition = nil
					if ok == true then
						local controller = localMembers.objectDCS:getController()
						controller:pushTask(taskDatas)
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, taskDatas .. " for unit " .. localMembers.name, false)
					end
				else
					ATMEFct.setError("NOT AIRCRAFT")
					ATMEFct.warningOutputCore(nil, fctDefinition, "not aircraft group", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "task type must be ATME.C_Task", true)
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end
	
	function newUnit.popTask(self)
		if localMembers.destroyed == false then
			if ATMEConst.canFly[ATMEDatas.localMembers[localMembers.group:getIndex()].category] == true then 
				local controller = localMembers.objectDCS:getController()

				controller:popTask()
			else
				ATMEFct.setError("NOT AIRCRAFT")
				ATMEFct.warningOutputCore(nil, fctDefinition, "not aircraft group", false)
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "popTask()", "already disabled", false)
		end
	end
	
	function newUnit.spotActivate(self, _target, _code)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "spotActivate(target, [code])"
		
		if localMembers.destroyed == false then
			if (ATMEFct.isObjectClass(_target, ATME.C_Unit) == true or ATMEFct.isObjectClass(_target, ATME.C_Static) == true or
					ATMEFct.isObjectClass(_target, ATME.C_Cargo) == true or ATMEFct.isObjectClass(_target, ATME.C_Group) == true or  
					ATMEFct.isObjectClass(_target, ATME.C_Zone2D) == true or type(_target) == "string") and (_code == nil or type(_code) == "number") then
					 
				
				local zone2D
				local ignoredTargets = {}
				
				if ATMEFct.isObjectClass(_target, ATME.C_Static) == true or ATMEFct.isObjectClass(_target, ATME.C_Cargo) == true or 
						ATMEFct.isObjectClass(_target, ATME.C_Unit) == true then
					if ATMEFct.isTargetOk[_target:getClassName()](_target, ignoredTargets) == true then
						localMembers.spotDatas.targetName = _target:getName()
						localMembers.spotDatas.targetClassName = _target:getClassName()
						localMembers.spotDatas.dynamicSpot = ATMEFct.isObjectClass(_target, ATME.C_Unit) == true or ATMEFct.isObjectClass(_target, ATME.C_Group) == true
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "bad target designation", false)
						return
					end
				elseif ATMEFct.isObjectClass(_target, ATME.C_Group) == true then
					local localMembersGroupTarget = ATMEDatas.localMembers[_target:getIndex()]
					if localMembersGroupTarget.category ~= "SHIP" then
						if ATMEConst.coalitionNames[localMembers.objectDCS:getCoalition()] ~= localMembersGroupTarget.coalition then
							localMembers.spotDatas.groupName = localMembersGroupTarget.name
							localMembers.spotDatas.dynamicSpot = true
						else
							ATMEFct.setError("BAD_COALITION")
							ATMEFct.warningOutputCore(nil, fctDefinition, "group coalition is bad", false)
							return
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "bad target designation", false)
						return
					end
				elseif type(_target) == "string" then
					local zone = ATMEFct.getDCSZone(_target)
					if zone ~= nil then
						localMembers.spotDatas.zone = ATMEFct.C_Circle_create(nil, nil, zone.point, zone.radius)
						localMembers.spotDatas.dynamicSpot = true
						localMembers.spotDatas.targetClassName = "C_Unit"
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone " .. _target .. " unknown", false)
						return
					end
				elseif ATMEFct.isObjectClass(_target, ATME.C_Zone2D) == true then
					localMembers.spotDatas.zone = _target
					localMembers.spotDatas.dynamicSpot = true
					localMembers.spotDatas.targetClassName = "C_Unit"
				end
				
				-- Lasing à partir d'un drone en vol
				local lasingInAir = false
				local localMembersGroup = ATMEDatas.localMembers[localMembers.group:getIndex()]
				if ATMEConst.canFly[localMembersGroup.category] == true and localMembers.objectDCS:inAir() == true and ATMEConst.aircraftTypeJTAC[typeName] == true then
					lasingInAir = true
				end
				
				if localMembers.category == "GROUND" or lasingInAir == true then
					local code

					if _code ~= nil then
						code = math.floor(_code)
						if code < 1111 or code > 1788 then
							ATMEFct.errorOutputCore(nil, fctDefinition, "code must be >= 1111 and <= 1788", false)
							return
						end
					end
					
					local vector = ATMEFct.vectors["3D"].createFromTable(nil, nil, localMembers.objectDCS:getVelocity())
					
					if (vector:getModule() < 0.01 and lasingInAir == false) or lasingInAir == true then
						if localMembers.spotDatas.spot ~= nil then
							localMembers.spotDatas.spot:destroy()
							localMembers.spotDatas.spot = nil
						end

						localMembers.spotDatas.code = code
						localMembers.spotDatas.noLOS = 0
						localMembers.spotDatas.neverHaveLOS = true
						localMembers.spotDatas.ignoredTargets = ignoredTargets
						localMembers.spotDatas.ignoredTargetOn = false
						localMembers.spotDatas.fromAir = lasingInAir
						
						timer.scheduleFunction(ATMETimers.C_Unit_spot_timer, localMembers, timer.getTime() + 0.2)
						
						return
					else
						ATMEFct.setError("UNIT_MOVING")
						ATMEFct.warningOutputCore(nil, fctDefinition, "not stopped for spotting", false)
						return
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "current unit must be ground unit or drone unit in air", false)
				end
			else
				if ATMEFct.isObjectClass(_target, ATME.C_Unit) == false and ATMEFct.isObjectClass(_target, ATME.C_Static) == false and
						ATMEFct.isObjectClass(_target, ATME.C_Group) == false and ATMEFct.isObjectClass(_target, ATME.C_Zone2D) == false and 
						type(_target) ~= "string" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "target type must be string, ATME.C_Zone2D, ATME.C_Group, ATME.C_Unit or ATME.C_Static", true)
				elseif _code ~= nil and type(_code) ~= "number" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "code type must be nil or number", false)
				end
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end
	
	function newUnit.spotDesactivate(self)
		if localMembers.destroyed == false then
			if localMembers.spotDatas.spot ~= nil then
				localMembers.spotDatas.spot.spot:destroy()
				localMembers.spotDatas.spot = nil
				localMembers.spotDatas.targetName = nil
				localMembers.spotDatas.targetClassName = nil
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "spotDesactivate()", "already disabled", false)
		end
	end
	
	function newUnit.spotIgnoreTarget(self)
		if localMembers.destroyed == false then
			if localMembers.spotDatas.spot ~= nil then
				if localMembers.spotDatas.targetName ~= nil then
					localMembers.spotDatas.ignoredTargets[localMembers.spotDatas.targetName] = true
					localMembers.spotDatas.ignoredTargetOn = true
				end
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "spotIgnoreTarget()", "already disabled", false)
		end
	end
	
	function newUnit.disable(self)
		if localMembers.destroyed == false then
			ATMEDatas.unitsToDisable[localMembers.name] = self
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "disable()", "already disabled", false)
		end
	end
	
	function newUnit.isDisabled(self)
		return localMembers.objectDCS:isExist() == false
	end
	
	function newUnit.getFuelRatio(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:getFuel()
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "getFuelRatio()", "already disabled", false)
		end
	end
	
	function newUnit.getLife(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:getLife()
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "getLife()", "already disabled", false)
		end
	end
	
	function newUnit.getLifeRatio(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:getLife()/localMembers.objectDCS:getLife0()
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "getLifeRatio()", "already disabled", false)
		end
	end
	
	function newUnit.radarOn(self)
		if localMembers.destroyed == false then
			localMembers.objectDCS:enableEmission(true)
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "radarOn()", "already disabled", false)
		end
	end
	
	function newUnit.radarOff(self)
		if localMembers.destroyed == false then
			localMembers.objectDCS:enableEmission(false)
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "radarOff()", "already disabled", false)
		end
	end
	
	function newUnit.getCallsign(self)
		if localMembers.destroyed == false then
			local identification
			local id1
			local id2
			
			if localMembers.objectDCS:hasAttribute("Helicopters") == true or localMembers.objectDCS:hasAttribute("Planes") == true then
				local callsign = localMembers.objectDCS:getCallsign()
				if localMembers.objectDCS:getCoalition() == 2 then
					identification = string.match(callsign, "^%a*")
					id1 = string.match(callsign, "%d")
					id2 = string.match(callsign, "%d$")
				else
					identification = callsign
				end
			end
			
			return identification, id1, id2
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "getCallsign()", "already disabled", false)
		end
	end
	
	function newUnit.getPosition(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:getPoint()
		else
			return localMembers.lastPosition.p
		end
	end

	function newUnit.getVelocityVector(self)
		local velocity
		if localMembers.destroyed == false then
			velocity = localMembers.objectDCS:getVelocity()
		else
			velocity = localMembers.lastVelocity
		end

		return ATMEFct.vectors["3D"].createFromTable(nil, nil, velocity)
	end

	function newUnit.getSpeed(self)
		local velocity
		if localMembers.destroyed == false then
			velocity = localMembers.objectDCS:getVelocity()
		else
			velocity = localMembers.lastVelocity
		end
			
		local vector = ATMEFct.vectors["3D"].createFromTable(nil, nil, velocity)
		return vector:getModule()
	end

	function newUnit.getHSpeed(self)
		local velocity
		if localMembers.destroyed == false then
			velocity = localMembers.objectDCS:getVelocity()
		else
			velocity = localMembers.lastVelocity
		end
		
		local vector = ATMEFct.vectors["3D"].createFromTable(nil, nil, velocity)
		return vector:getHModule()
	end

	function newUnit.getVSpeed(self)
		local velocity
		if localMembers.destroyed == false then
			velocity = localMembers.objectDCS:getVelocity()
		else
			velocity = localMembers.lastVelocity
		end
			
		return velocity.y
	end

	function newUnit.getSpeedAzimuth(self)
		local velocity
		if localMembers.destroyed == false then
			velocity = localMembers.objectDCS:getVelocity()
		else
			velocity = localMembers.lastVelocity
		end
		
		local vector = ATMEFct.vectors["3D"].createFromTable(nil, nil, velocity)
				
		if vector:getModule() > 1 then 
			return vector:getAzimuth()
		else
			return 0
		end
	end

	function newUnit.getRollAxisVector(self)
		local position
		if localMembers.destroyed == false then
			position = localMembers.objectDCS:getPosition().x
		else
			position = localMembers.lastPosition.x
		end

		return ATMEFct.vectors["3D"].createFromTable(nil, nil, position)
	end

	function newUnit.getAzimuth(self)
		local position
		if localMembers.destroyed == false then
			position = localMembers.objectDCS:getPosition().x
		else
			position = localMembers.lastPosition.x
		end
		
		local vector = ATMEFct.vectors["3D"].createFromTable(nil, nil, position)
		return vector:getAzimuth()
	end

	function newUnit.getCountryName(self)
		if localMembers.destroyed == false then
			local id = localMembers.objectDCS:getCountry()
			return country.name[id]
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "getCountryName()", "already disabled", false)
		end
	end

	function newUnit.getCoalitionName(self)
		if localMembers.destroyed == false then
			return ATMEConst.coalitionNames[localMembers.objectDCS:getCoalition()]
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "getCoalitionName()", "already disabled", false)
		end
	end

	function newUnit.isHelicopter(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:hasAttribute("Helicopters")
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "isHelicopter()", "already disabled", false)
		end
	end
	
	function newUnit.isPersonnelCarrier(self)
		if localMembers.destroyed == false then
			return ATMEConst.personnelCarrierAttributes[localMembers.typeName] ~= nil
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "isPersonnelCarrier()", "already disabled", false)
		end
	end


	function newUnit.isPlane(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:hasAttribute("Planes")
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "isPlane()", "already disabled", false)
		end
	end
	
	function newUnit.canFly(self)
		if localMembers.destroyed == false then
			local localMembersGroup = ATMEDatas.localMembers[localMembers.group:getIndex()]
			return ATMEConst.canFly[localMembersGroup.category]
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "canFly()", "already disabled", false)
		end
	end
	
	function newUnit.isInfantry(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:hasAttribute("Infantry")
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "isInfantry()", "already disabled", false)
		end
	end
	
	function newUnit.isGroundVehicle(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:hasAttribute("Ground vehicles")
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "isGroundVehicle()", "already disabled", false)
		end
	end

	function newUnit.isAirDefence(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:hasAttribute("Air Defence")
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "isAirDefence()", "already disabled", false)
		end
	end

	function newUnit.isManPad(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:hasAttribute("MANPADS")
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "isManPad()", "already disabled", false)
		end
	end

	function newUnit.isSAMVehicle(self, _range)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "isSAMVehicle([range])"
		
		if localMembers.destroyed == false then
			if _range == nil or type(_range) == "string" then
				if _range == nil then
					return localMembers.objectDCS:hasAttribute("SR SAM") or self.localMembers.objectDCS:hasAttribute("MR SAM") or self.localMembers.objectDCS:hasAttribute("LR SAM")
				elseif _range == "SHORT" then
					return localMembers.objectDCS:hasAttribute("SR SAM")
				elseif _range == "MEDIUM" then
					return self.localMembers.objectDCS:hasAttribute("MR SAM")
				elseif _range == "LONG" then
					return self.localMembers.objectDCS:hasAttribute("LR SAM")
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "range bad value (SHORT, MEDIUM, LONG)", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "range type must be nil or string", true)
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end

	function newUnit.isSAMSiteLauncher(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:hasAttribute("SAM LL")
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "isSAMSiteLauncher()", "already disabled", false)
		end
	end

	function newUnit.isSAMSiteRadar(self, _category)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "isSAMSiteRadar([category])"
		
		if localMembers.destroyed == false then
			if _category == nil or type(_category) == "string" then
				if _category== nil then
					return localMembers.objectDCS:hasAttribute("SAM SR") or localMembers.objectDCS:hasAttribute("SAM TR")
				elseif _category == "SURFACE" then
					return localMembers.objectDCS:hasAttribute("SAM SR")
				elseif _category == "TRACKING" then
					return self.localMembers.objectDCS:hasAttribute("SAM TR")
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "category bad value (SURFACE, TRACKING)", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "category type must be nil or string", true)
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end

	function newUnit.isSAMSiteCommandCenter(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:hasAttribute("SAM CC")
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "isSAMSiteCommandCenter()", "already disabled", false)
		end
	end

	function newUnit.isAAA(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:hasAttribute("AAA")
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "isAAA()", "already disabled", false)
		end
	end	

	function newUnit.inAir(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:inAir()
		else
			return false
		end
	end

	function newUnit.getMSLAltitude(self)
		return ATME.getPointAltitude(self:getPosition())
	end

	function newUnit.getAGLAltitude(self)
		local point = self:getPosition()
		
		local landHeight = ATMEFct.getSurfaceAltitude(point)
		return ATME.getPointAltitude(point) - landHeight
	end

	function newUnit.explode(self, _power)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "explode(power)"
		
		if type(_power) == "number" then
			if _power > 0 then
				trigger.action.explosion(self:getPosition(), _power)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "power must be >= 1", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "power type must be number", true)
		end
	end
	
	function newUnit.isInZones(self, _zones)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "isInZones(zones)"
		
		if ATMEFct.isObjectClass(_zones, ATME.C_Zones) == true then
			local isInside, _ = ATMEFct.C_Zones_isInside(_zones, self:getPosition())
			return isInside
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "zones type must be ATME.C_Zones", true)
		end
	end				

	function newUnit.isInDCSZone(self, _name)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "isInDCSZone(name)"
		
		if type(_name) == "string" then
			local zone = ATMEFct.getDCSZone(_name)
	
			if zone ~= nil then
				local vector = ATMEFct.vectors["3D"].createFromReference(nil, nil, self:getPosition(), zone.point)
				if vector:getHModule() <= zone.radius then
					return true
				else
					return false
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone " .. _name .. " unknown", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", true)
		end
	end
	
	function newUnit.isInZone(self, _zone)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "isInZone(zone)"
		
		if ATMEFct.isObjectClass(_zone, ATME.C_Zone) == true then
			local localMembersZone = ATMEDatas.localMembers[_zone:getIndex()]
			
			local datas = {
				zone = _zone,
				actualCenter = ATMEFct.C_Zone_getActualCenter(localMembersZone),
				azimuth = ATMEFct.C_Zone_getActualAzimuth(localMembersZone),
			}
					
			return ATMEFct[ATMEConst.indZoneConfig[_zone:getClassName()].fctIsInside](self:getPosition(), datas)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "zone type must be ATME.C_Zone", true)
		end
	end
	
	function newUnit.isPilotDead(self)
		if localMembers.destroyed == false then
			return localMembers.pilotDead
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "isPilotDead()", "already disabled", false)
		end
	end
	
	function newUnit.isPilotEjected(self)
		if localMembers.destroyed == false then
			return localMembers.pilotEjected
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "isPilotEjected()", "already disabled", false)
		end
	end
	
	function newUnit.whichSide(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "whichSide(reference)"
		
		if ATMEFct.isReference(_reference) == true  then
			local point = ATMEFct.getPointFromReference(_reference)

			return ATMEFct.C_Unit_whichSide(self, point)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
		end
	end
	
	function newUnit.getNbUnitsOnBoard(self)
		if localMembers.destroyed == false then
			return localMembers.nbUnitsOnBoard
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "getNbUnitsOnBoard()", "already disabled", false)
		end
	end

	function newUnit.getOnBoardGroupNames(self)
		if localMembers.destroyed == false then
			local temp = {}
			
			for _id, _item in localMembers.onBoardGroupDatas:pairs() do
				temp[_item.name] = _id
			end
			
			return temp
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "getOnBoardGroupNames()", "already disabled", false)
		end
	end

	function newUnit.unload(self, _param1, _param2)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "unload([whenStopped])\nunload(id, [whenStopped])"

		if localMembers.destroyed == false then
			if (_param1 == nil and _param2 == nil) or type(_param1) == "boolean" then
				local whenStopped = false
				 if type(_param1) == "boolean" then
					whenStopped = _param1
				end
				
				ATMEFct.C_Unit_unload(localMembers, nil, whenStopped)
			elseif type(_param1) == "number" and (_param2 == nil or type(_param2) == "boolean") then
				local id = math.floor(_param1)
				
				local whenStopped = _param2
				
				if _param2 == nil then
					whenStopped = false
				end
				
				ATMEFct.C_Unit_unload(localMembers, id, whenStopped)
			else
				if type(_param1) ~= "boolean" and type(_param1) ~= "number" and _param2 == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "whenStopped must be nil or boolean\nid type must be number", true)
				else
					if type(_param1) ~= "number" then
						ATMEFct.errorOutputCore(nil, fctDefinition, "id type must be number", true)
					elseif _param2 ~= nil and type(_param2) ~= "boolean" then
						ATMEFct.errorOutputCore(nil, fctDefinition, "whenStopped type must be nil or boolean", false)
					end
				end
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end

	function newUnit.load(self, _param1, _param2)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "load(radius)\nload(radius, nbGroups)\nload(radius, groupToBoard)"
		
		if localMembers.destroyed == false then
			if localMembers.objectDCS:isActive() == true then
				if type(_param1) == "number" then
					local groupToBoard
					local nbGroupsToBoard = 1
					
					if _param2 ~= nil then
						if  ATMEFct.isObjectClass(_param2, ATME.C_Group) == true then
							groupToBoard = _param2
						elseif type(_param2) == "number" then
							nbGroupsToBoard = math.floor(_param2)
							
							if nbGroupsToBoard <= 0 then
								ATMEFct.errorOutputCore(nil, fctDefinition, "unit " .. localMembers.name .. " - nbGroups must be > 0", false)
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "nbGroups type must be nil or number\ngroupToBoard type must be nil or ATME.C_Group", false)
							return
						end
					end

					if groupToBoard ~= nil then
						local localMembersGroup = ATMEDatas.localMembers[groupToBoard:getIndex()]
						if localMembersGroup.onlyInfantry == true then
							if _param1 > 0 then
								if ATMEConst.personnelCarrierAttributes[localMembers.typeName] ~= nil then
									ATMEFct.C_Unit_load(self, nil, groupToBoard, _param1)
								else
									ATMEFct.errorOutputCore(nil, fctDefinition, "unit " .. localMembers.name .. " not a transport unit", false)
								end
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "radius must be > 0", false)
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "group  " .. localMembersGroup.name .. " not infantry group", false)
						end
					else
						ATMEFct.C_Unit_load(self, nbGroupsToBoard, nil, _param1)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "radius type must be number", true)
				end
			else
				ATMEFct.setError("UNACTIVATED_UNIT")
				ATMEFct.warningOutputCore(nil, fctDefinition, "unit " .. localMembers.name .. " not activated", false)
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end

	function newUnit.getNearestReadyToBoardGroup(self, _radius)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "getNearestReadyToBoardGroup(radius)"
		
		if localMembers.destroyed == false then
			if type(_radius) == "number" then
				if _radius >= 0 then
					return ATMEFct.C_Unit_getNearestReadyToBoardGroup(self, _radius)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "radius must be > 0", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "radius must be number", true)
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end
	
	function newUnit.fireFlare(self, _color)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "fireFlare(color)"
		
		if type(_color) == "string" then
			if ATMEConst.flareColors[string.upper(_color)]~= nil then
				if localMembers.category == "GROUND" then
					if localMembers.flare == nil then
						localMembers.flare = ATMEFct.C_Flare_create(_color, self:getPosition())
					else
						ATMEDatas.localMembers[localMembers.flare:getIndex()].point = self:getPosition()
					end
					
					localMembers.flare:activate()
				else
					ATMEFct.setError("NOT_GROUND_UNIT")
					ATMEFct.warningOutputCore(nil, fctDefinition, "not ground unit, can't fire flare", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "color must be a flare color or RANDOM", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "color type must be string", true)
		end
	end

	function newUnit.fireSmoke(self, _color)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "fireSmoke(color)"
		
		if type(_color) == "string" then
			if ATMEConst.smokeColors[string.upper(_color)] ~= nil then 
				if localMembers.category == "GROUND" then
					local point = self:getPosition()
					local newPoint = {
						x = point.x + math.random(-20,20),
						y = point.y,
						z = point.z + math.random(-20,20),
					}
					
					local smoke = ATMEFct.C_Smoke_create(_color, newPoint)
					smoke:activate(false)
				else
					ATMEFct.setError("NOT_GROUND_UNIT")
					ATMEFct.warningOutputCore(nil, fctDefinition, "not ground unit, can't fire smoke", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "color must be a smoke color", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "color type must be string", true)
		end
	end

	function newUnit.fireIlluminationBomb(self, _power)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "fireIlluminationBomb(power)"
		
		if type(_power) == "number" then
			if _power >= 1 then
				if localMembers.category == "GROUND" then
					local point = self:getPosition()
					point.y = point.y + 200 + math.random(0, 100)

					trigger.action.illuminationBomb(point, _power)
				else
					ATMEFct.setError("NOT_GROUND_UNIT")
					ATMEFct.warningOutputCore(nil, fctDefinition, "not ground unit, can't fire illumination bomb", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "power must be >= 1", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "power type must be number", true)
		end
	end
	
	function newUnit.setVoice(self, _name)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "setVoice([name])"
		
		if localMembers.destroyed == false then
			if _name == nil or type(_name) == "string" then
				if _name ~= nil then
					local d, f, temp = _name:find("^(%a[%a%d_]*)")
					
					if temp == _name then
						localMembers.currentVoice = _name
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "name bad format", false)
					end
				else
					localMembers.currentVoice = ""
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be nil or string", true)
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end
	
	function newUnit.getVoice(self)
		if localMembers.destroyed == false then
			if localMembers.currentVoice ~= "" then
				return localMembers.currentVoice
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "getVoice()", "already disabled", false)
		end
	end

	function newUnit.beaconOn(self, _settings)
		-- BeaconSites.lua dans DCS World OpenBeta\Scripts\World\Radio : définition des types
		
		if localMembers.destroyed == false then
			local fcts = {
				["TACAN"] = function(_tokens)
					if _tokens[2].type == "TOKEN_TACAN_CHANNEL" then
						local mode = string.sub(string.upper(_tokens[2].value), -1)
						local channel = tonumber(string.sub(_tokens[2].value, 1, _tokens[2].value:len() - 1))
		
						local typeValue = 4
						local systemValue = 3
						local frequency
						
						if _tokens[1].type == "TOKEN_TYPE_TACAN_AA" then
							if localMembers.group:canFly() == true then
								channel = channel + 63

								if ATMEConst.specialAircraftTypes[localMembers.typeName] ~= nil then
									--Tanker
									
									typeValue = 4
									systemValue = 4 + ATMEConst.tacanMode[mode]
								else
									typeValue = 4
									systemValue = 13 + ATMEConst.tacanMode[mode]
								end
							else
								return true, "AA Tacan must be set on HELICOPTER or AIRPLAINE groups only"
							end
							
							frequency = (ATMEConst.tacanFrequencyOffset["Xmit"][mode] + channel) * 1000000
						else
							if mode == 'Y' then -- TACAN non AA limitation à mode X. Mode Y fonctionne mais pas de bearing pour l'instant.
								return true, "Tacan channel mode must be X only if not AA TACAN"
							end
							
							frequency = (ATMEConst.tacanFrequencyOffset["Recv"][mode] + channel) * 1000000
						end
					
						local beaconCallsign
						if #_tokens == 3 and _tokens[3].type == "TOKEN_STRING" then
							beaconCallsign = string.upper(_tokens[3].value)
						end
						
						local controller = ATMEDatas.localMembers[localMembers.group:getIndex()].objectDCS:getController()
						local command = { 
							id = 'ActivateBeacon', 
							params = { 
								type = typeValue, 
								system = systemValue, 
								callsign = beaconCallsign, 
								frequency = frequency,
							}
						}
						controller:setCommand(command)
						
						return false
					else
						return true, "Tacan channel must be >=1 and <=63, X or Y if AA TACAN, X only if not AA TACAN"
					end
				end,
				["ICLS"] = function(_tokens)
					if _tokens[2].type == "TOKEN_INTEGER" then
						local controller = ATMEDatas.localMembers[localMembers.group:getIndex()].objectDCS:getController()
						local frequency = 15415000000 + 30000000 * ((_tokens[2].value - 1) % 10)
						
						local command = { 
							id = 'ActivateBeacon', 
							params = {
								type = 131328,
								system = 16,
								frequency = frequency,
							}
						}
						controller:setCommand(command)
						
						command = { 
							id = 'ActivateBeacon', 
							params = {
								type = 131584,
								system = 17,
								frequency = frequency,
							}
						}
						controller:setCommand(command)
						
						return false
					else
						return true, "ICLS channel type must be integer"
					end
				end,
			}
			
			local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "beaconOn(settings)"
			
			if type(_settings) == "string" then
				local tokens = ATMEFct.analyzeString(_settings)
				
				if #tokens == 2 or #tokens == 3 then
					local beaconType = ATMEConst.beaconTypes[tokens[1].type]
					
					
					if fcts[beaconType] ~= nil then
						local err, errorDesc = fcts[beaconType](tokens)
						
						if err == true then
							ATMEFct.errorOutputCore(nil, fctDefinition, errorDesc, false)
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "settings must be 'TACAN', 'TACAN_AA' or 'ICLS'", false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "settings must have 2 or 3 parameters", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "settings type must be string", true)
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end
	
	function newUnit.beaconOff(self)
		if localMembers.destroyed == false then
			local controller = ATMEDatas.localMembers[localMembers.group:getIndex()].objectDCS:getController()
			local command = { 
				id = "DeactivateBeacon", 
				params = {},
			}
			
			controller:setCommand(command)
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newUnit) .. "beaconOff()", "already disabled", false)
		end
	end
	
	function newUnit.setFrequency(self, _modulation, _frequency)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "setFrequency(modulation, frequency)"

		if localMembers.destroyed == false then
			if type(_modulation) == "string" and type(_frequency) == "number" then
				local modulationValue = ATMEConst.modulationValues[_modulation]
				if modulationValue ~= nil then
					local frequencyValue = math.floor(_frequency)
					if frequencyValue > 0 then
						local controller = ATMEDatas.localMembers[localMembers.group:getIndex()].objectDCS:getController()

						local command = {
							id = "SetFrequency", 
							params = { 
								frequency = frequencyValue,
								modulation = modulationValue,							
							} 
						}
						
						controller:setCommand(command)
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "frequency type must > 0", false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "modulation must be 'AM' or 'FM'", true)
				end
			else
				if type(_modulation) ~= "string" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "modulation type must be string", true)
				elseif type(_frequency) ~= "number" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "frequency type must be number", false)
				end
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end

	newUnit.modules = {}
	setmetatable(newUnit.modules, ATMEDatas.C_Unit_MT.mtModules)
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
		localMembers.eventCallbacks[_localModuleMembers.name] = {}
		newUnit.modules[_localModuleMembers.name] = {}
	end
	
	if ATMEConst.personnelCarrierAttributes[localMembers.typeName] ~= nil and localMembers.objectDCS:hasAttribute("Helicopters") == true then
		localMembers.indexPickupZoneSignal = ATMEDatas.signalCapabilityUnits:add({ 
																					object = newUnit,
																					localMembers = localMembers})
	end

	local groupName = _unit:getGroup():getName()

	if _groupExist == true then
		-- Cas d'un ajout d'unité sur une groupe existant (cas changement d'unité par joueur)
		
		ATMEFct.C_Group_addUnit(ATMEDatas.groups[groupName], newUnit)

		ATMEDatas.units[localMembers.name] = newUnit
		localMembers.group = ATMEDatas.groups[groupName]

		ATMEFct.C_Unit_playerEnters(newUnit)

		for _, _nameModule in ipairs(ATMEDatas.modulesNames) do
			local localModuleMembers = ATMEDatas.modulesLocalMembers[_nameModule]	
			ATMEDatas.currentModuleName = _nameModule
								
			if localModuleMembers.handlers.onCreateUnitHandler ~= nil then
				ATMEFct.outputCore(nil,"Appel onCreateUnitHandler du module " .. localModuleMembers.name, 2)
				xpcall(function() localModuleMembers.handlers.onCreateUnitHandler(newUnit) end, function() ATMEFct.errUserHandler('onCreateUnitHandler') end)
			end
		end
	else
		if ATMEDatas.newCreatedGroups[groupName] == nil then
			ATMEFct.C_Group_create(newUnit, _datas, _newObject)
		else
			ATMEFct.C_Group_addUnit(ATMEDatas.newCreatedGroups[groupName].group, newUnit)
		end
		
		localMembers.group = ATMEDatas.newCreatedGroups[groupName].group
		ATMEFct.C_Unit_playerEnters(newUnit)
	end
	
	local localMembersGroup = ATMEDatas.localMembers[localMembers.group:getIndex()]
	if localMembersGroup.waypoints[1].airdromeId ~= nil or localMembersGroup.waypoints[1].helipadId ~= nil then
		if localMembersGroup.waypoints[1].airdromeId ~= nil then
			airbase = ATMEDatas.airbasesById[localMembersGroup.waypoints[1].airdromeId]:getName()
		elseif localMembersGroup.waypoints[1].helipadId ~= nil then
			airbase = ATMEDatas.airbasesById[localMembersGroup.waypoints[1].helipadId]:getName()
		end
		
		if localMembersGroup.waypoints[1].type == "TakeOffParkingHot" or localMembersGroup.waypoints[1].type == "TakeOffGroundHot" or
				localMembersGroup.waypoints[1].type == "TakeOff" then
			-- Départ moteurs On
			localMembers.engineOn = true
		end
	else
		-- En vol....
		
		localMembers.engineOn = true
		localMembers.hasTakenOff = true
	end
	
	localMembers.category = localMembersGroup.category

	if _groupExist == true and ATMEConst.canFly[localMembersGroup.category] == true and localMembers.engineOn == true then
		-- Emission de l'événement démarrage moteur
		local event = {
			type = "ENGINE",
			datas = {
				action = "START",
				unitName = localMembers.name,
			},
		}
		
		ATMEFct.outputCore(nil, "Initial Start engine pour " .. localMembers.name, 2)
		ATMEFct.C_Event_add(event)
	end
	
	ATMEFct.outputCore(nil,"Prepare unit = " .. localMembers.name .. " -> ok", 2)
end

ATMEFct.C_Unit_delete = function(_unit, _reserved)
	local localMembers = ATMEDatas.localMembers[_unit:getIndex()]
	
	if localMembers.destroyed == false then
		local localMembersGroup = ATMEDatas.localMembers[localMembers.group:getIndex()]
		local flag = ATMEDatas.flagDeleteUnit[localMembers.name]

		ATMEDatas.flagDeleteUnit[localMembers.name] = nil
		
		if localMembers.objectDCS:isExist() == true then
			-- N'existe plus lors de certains changement de slot avec un delete déphasé
			-- fait dans S_EVENT_BIRTH. Voir aussi S_EVENT_PLAYER_LEAVE_UNIT
			
			localMembers.lastPosition = localMembers.objectDCS:getPosition()
			localMembers.lastVelocity = localMembers.objectDCS:getVelocity()
		end
		
		if localMembersGroup.boardingStarted == false then
			-- Les unités d'un groupe chargé dans un véhicule ne génèrent pas 
			-- d'appel à onDeleteUnitHandler
			
			for _i = #ATMEDatas.modulesNames, 1, -1 do
				local localModuleMembers = ATMEDatas.modulesLocalMembers[ATMEDatas.modulesNames[_i]]
				
				if localModuleMembers.handlers.onDeleteUnitHandler ~= nil then
					ATMEDatas.currentModuleName = localModuleMembers.name
					
					-- Appel du module
					ATMEFct.outputCore(nil,"Appel onDeleteUnitHandler du module " .. localModuleMembers.name, 2)
					xpcall(function() localModuleMembers.handlers.onDeleteUnitHandler(_unit, flag) end, function() ATMEFct.errUserHandler('onDeleteUnitHandler') end)
				end
			end
		end
		
		ATMEFct.C_Airbase_delete(localMembers.objectDCS)
		
		if localMembers.indexPickupZoneSignal ~= nil then
			ATMEDatas.signalCapabilityUnits:remove(localMembers.indexPickupZoneSignal)
			localMembers.indexPickupZoneSignal = nil
		end
		
		if _reserved == false and localMembersGroup.autoRespawn.activated == false then
			ATMEDatas.reservedUnitNames[localMembers.name] = nil
		end
		ATMEDatas.reservedBoardingGroup[localMembers.name] = nil

		_unit.modules = nil

		for _, _item in localMembers.onBoardGroupDatas:pairs() do
			-- Si un des groupes embarqués a un respawn automatique il est préparé
			
			if _item.localMembers.autoRespawn.activated == true then
				local respawnTime = _item.localMembers.autoRespawn.time
				if _item.localMembers.autoRespawn.random == true then
					respawnTime = math.random(_item.localMembers.autoRespawn.minTime, _item.localMembers.autoRespawn.maxTime)
				end
				
				ATMEDatas.groupsToRespawn[_item.localMembers.name] = true
				timer.scheduleFunction(ATMETimers.C_Group_autoRespawn, _item.localMembers.initialSpawnDatas, timer.getTime() + respawnTime)
			end
			
			ATMEDatas.reservedGroupNames[_item.localMembers.name] = nil
			
			for _, _unitItem in pairs(ATMEDatas.groupsOnBoard[_item.localMembers.name].spawnDatas.datas.units) do
				ATMEDatas.reservedUnitNames[_unitItem.name] = nil
			end
			
			ATMEDatas.groupsOnBoard[_item.localMembers.name] = nil
		end

		localMembers.onBoardGroupDatas:removeAll()
		localMembers.nbBoardings = 0
		
		if localMembers.spotDatas.spot ~= nil then
			localMembers.spotDatas.spot:destroy()
			localMembers.spotDatas.spot = nil
		end

		ATMEFct.outputCore(nil,"Delete unit = " .. localMembers.name, 2)
		
		ATMEFct.C_Group_deleteUnit(localMembers.group, localMembers.name, _reserved)
		ATMEDatas.C_Unit_MT:remove(_unit:getIndex())
		
		localMembers.destroyed = true
		ATMEDatas.units[localMembers.name] = nil
	end
end

setmetatable(ATME.C_Unit,ATMEDatas.C_Unit_MT)


ATMEDatas.C_Unit_MT.mt.__metatable = "Private"
ATMEDatas.C_Unit_MT.mt.__index = ATME.C_Unit
ATMEDatas.C_Unit_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Unit_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Unit : Bad key " .. k, "table locked", false)
	end
end

ATMEDatas.C_Unit_MT.mtModules.__metatable = "Private"
ATMEDatas.C_Unit_MT.mtModules.__newindex = function(t, k, v)
	if ATMEDatas.modules[k] ~= nil then
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil,"ATME : ATME.C_Unit.modules Bad key " .. k, "table locked", false)
	end
end


----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Static : gère les objets statiques au format ATME. Pas de contructeur accessible
-- pour les modules externes. 
-- Sont exclus les objets statiques FARP et Airfieds qui sont gérés par ATME.C_Airbase.
-- Les instances de cette classe sont intégralement gérées par le module ATME_Core. La création et la destruction
-- dépendent du contexte de début de mission et des événements qui se produisent dans DCS en cours de mission.
-- Les modules externes accèdent à cette gestion indirectement par les handlers définis dans ceux-ci et les méthodes
-- accessibles de cette classe. Ceci permet de garantir une cohérence sur les listes internes ATME.
----------------------------------------------------------------------------------------------------------

ATMEDatas.C_Static_MT = {
	className = "C_Static",
	mtModules = {},
	mt = {},
	
	keys = {
		["getCategoryName"] = true,
		["getLife"] = true,
		["getDCSObject"] = true,
		["getDCSID"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["getIndex"] = true,
		["getTypeName"] = true,
		["getPosition"] = true,
		["getModelDatas"] = true,
		["getAzimuth"] = true,
		["getCountryName"] = true,
		["getCoalitionName"] = true,
		["getAttributes"] = true,
		["hasAttributes"] = true,
		["isInZones"] = true,
		["isInDCSZone"] = true,
		["isInZone"] = true,
		["explode"] = true,
		["fireFlare"] = true,
		["fireSmoke"] = true,
		["fireIlluminationBomb"] = true, 
		["disable"] = true,
		["isDisabled"] = true,
		["getUserPrivateDatas"] = true,
		["modules"] = true,
	},

	add = function(self, _object, _datas, _state)
		local category
		local name = _object:getName()

		if _state == false then
			category = ATMEConst.backStaticObjectCategories[ATMEDatas.allStaticsInMission[name].category]
		else
			category = ATMEConst.backStaticObjectCategories[ATMEDatas.staticsToBirth[name].category]
		end
		
		local item = {
			name = name,
			objectDCS = _object,
			objectDCSId = _object:getID(),
			typeName = _object:getTypeName(),
			isEnabled = true,
			destroyed = false,
			userPrivateDatas = {},						-- Point d'entrée des données privées : un par module
			tagsName = ATMEFct.decompString(name),
			category = category,
			spawnDatas = _datas,
		}
		
		item.spawnDatas.name = name
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
	end,
}

ATME.C_Static = {
	getClassName = function(self)
		return ATMEDatas.C_Static_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Static_MT.parentClass
	end,
	
	getByName = function(_objectName)
		if type(_objectName) == "string" then
			return ATMEDatas.statics[_objectName]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Static.getByName(name)", "name must be string", false)
		end	
	end,
	foreach = function(_callback, _datas)
		local fctDefinition = "Function ATME.C_Static.foreach(callback, [datas])"
		
		if type(_callback) == "function" then
			ATMEFct.foreachObject(nil, "statics", _callback, _datas)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "callback type must be function", false)
		end
	end,
	foreachInCoalition = function(_coalitionName, _callback, _datas)
		local fctDefinition = "Function ATME.C_Static.foreachInCoalition(coalitionName, callback, [datas])"
		
		if type(_coalitionName) == "string" and type(_callback) == "function" then
			if ATMEConst.coalitionIdsBlueRed[_coalitionName] == nil then
				ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName must be 'BLUE' or 'RED'", false)
				return
			end
			
			ATMEFct.foreachObject(_coalitionName, "statics", _callback, _datas)
		else
			if type(_coalitionName) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName type must be string", false)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "callback type must be function", false)
			end
		end
	end,
}


ATMEFct.C_Static_create = function(_object, _datas)
	local newObject = {}
	local datas
	if _datas == nil then
		local cloningContextLocalMembers = ATMEDatas.localMembers[ATMEDatas.staticsToBirth[_object:getName()].staticCloningContext:getIndex()]
		
		datas = {
			country = cloningContextLocalMembers.country,
			datas = {
				type = cloningContextLocalMembers.staticDatas.type,
				shape_name = cloningContextLocalMembers.staticDatas.shape_name,
				livery_id = cloningContextLocalMembers.staticDatas.livery_id,
				category = cloningContextLocalMembers.staticDatas.category,
			},
		}
	else
		datas = {
			country = _datas.country,
			datas = {
				type = _datas.type,
				shape_name = _datas.shape_name,
				livery_id = _datas.livery_id,
				category = _datas.category,
			},
		}
	end
	
	local index = ATMEDatas.C_Static_MT:add(_object, datas, _datas == nil)
	local localMembers = ATMEDatas.localMembers[index]
	
	setmetatable(newObject, ATMEDatas.C_Static_MT.mt)
	
	function newObject.getIndex(self)
		return index
	end
	
	function newObject.getDCSObject(self)
		return localMembers.objectDCS
	end
	
	function newObject.getDCSID(self)
		return localMembers.objectDCSId
	
	end
	
	function newObject.getName(self)
		return localMembers.name
	end
	
	function newObject.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if localMembers.tagsName ~= nil and pos >= 1 and pos <= #localMembers.tagsName then
				return localMembers.tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newObject.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newObject.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if localMembers.tagsName ~= nil then
				local ok = ATMEFct.compareTags(localMembers.tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newObject.getCategoryName(self)
		-- Catégories objets statiques - A ne pas confondre avec catégorie DE BASE dcs des statics/cargo
		-- 		Ships
		-- 		Planes
		-- 		Helicopters
		-- 		Warehouses
		-- 		Cargos	-> Traité par la classe ATME.C_Cargo
		-- 		Unarmed (tous ce qui est véhicules terrestre)
		-- 		Personnel
		--		Animal
		--		Heliports -> Traités par la classe ATME.C_Airbase
		-- 		Fortifications
		--		ADEquipment
		
		if localMembers.destroyed == false then
			return localMembers.category
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newObject) .. "getCategoryName()", "already disabled", false)
		end
	end
	
	function newObject.getTypeName(self)
		return localMembers.typeName
	end
	
	function newObject.getModelDatas(self)
		return {
			type = localMembers.spawnDatas.datas.type,
			shape_name = localMembers.spawnDatas.datas.shape_name,
			livery_id = localMembers.spawnDatas.livery_id,
		}
	end
	
	function newObject.getLife(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:getLife()
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newObject) .. "getLife()", "already disabled", false)
		end
	end
	
	function newObject.getPosition(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:getPoint()
		else
			return localMembers.lastPosition.p
		end
	end

	function newObject.getAzimuth(self)
		local position
		if localMembers.destroyed == false then
			position = localMembers.objectDCS:getPosition().x
		else
			position = localMembers.lastPosition.x
		end
		
		local vector = ATMEFct.vectors["3D"].createFromTable(nil, nil, position)
		return vector:getAzimuth()
	end

	function newObject.getCountryName(self)
		if localMembers.destroyed == false then
			local id = localMembers.objectDCS:getCountry()
			return country.name[id]
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newObject) .. "getCountryName()", "already disabled", false)
		end
	end

	function newObject.getCoalitionName(self)
		if localMembers.destroyed == false then
			local id = localMembers.objectDCS:getCoalition()
			return ATMEConst.coalitionNames[id]
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newObject) .. "getCoalitionName()", "already disabled", false)
		end
	end
	
	function newObject.disable(self)
		if localMembers.destroyed == false then
			ATMEDatas.staticsToDisable[localMembers.name] = self
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newObject) .. "disable()", "already disabled", false)
		end
	end
	
	function newObject.isDisabled(self)
		return localMembers.objectDCS:isExist() == false
	end
	
	function newObject.getUserPrivateDatas(self, _moduleName)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "getUserPrivateDatas([moduleName])"
		local moduleName
		
		if _moduleName ~= nil then
			if type(_moduleName) == "string" then
				if localMembers.userPrivateDatas[_moduleName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
					return
				else
					moduleName = _moduleName
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
				return
			end
		else
			moduleName = ATMEDatas.currentModuleName
		end
		
		return localMembers.userPrivateDatas[moduleName]
	end
	
	function newObject.getAttributes(self)
		if localMembers.destroyed == false then
			local datas = StaticObject.getDesc(localMembers.objectDCS)
			local attributes = {}
			for _id, _ in pairs(datas.attributes) do
				if ATMEConst.backTargetsAttributes[_id] ~= nil then
					table.insert(attributes, ATMEConst.backTargetsAttributes[_id].value)
				end
			end
			
			return attributes
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newObject) .. "getAttributes()", "already disabled", false)
		end
	end
	
	function newObject.hasAttributes(self, _attributes)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "hasAttributes(attributes)"
		
		if localMembers.destroyed == false then
			if type(_attributes) == "string" or type(_attributes) == "table" then
				local err, datas = ATMEFct.extractDatasFromStringOrTable(_attributes)
		
				if err == false then
					return ATMEFct.hasAttributes(localMembers.objectDCS, datas)
				end
				
				ATMEFct.errorOutputCore(nil, fctDefinition, datas, false)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "attributes type must be string or table", true)
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, fctDefinition, "already disabled", false)
		end
	end

	function newObject.explode(self, _power)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "explode(power)"
		
		if type(_power) == "number" then
			if _power > 0 then
				trigger.action.explosion(self:getPosition(), _power)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "power must be >= 1", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "power type must be number", true)
		end
	end
	
	function newObject.fireFlare(self, _color)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "fireFlare(color)"
		
		if type(_color) == "string" then
			if ATMEConst.flareColors[string.upper(_color)]~= nil then
				if localMembers.flare == nil then
					localMembers.flare = ATMEFct.C_Flare_create(_color, self:getPosition())
				end
				
				localMembers.flare:activate()
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "color must be a flare color or RANDOM", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "color type must be string", true)
		end
	end

	function newObject.fireSmoke(self, _color)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "fireSmoke(color)"
		
		if type(_color) == "string" then
			if ATMEConst.smokeColors[string.upper(_color)]~= nil then 
				local point = self:getPosition()
				
				local newPoint = {
					x = point.x + math.random(-20,20),
					y = point.y,
					z = point.z + math.random(-20,20),
				}
					
				local smoke = ATMEFct.C_Smoke_create(_color, newPoint)
				smoke:activate(false)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "color must be a smoke color", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "color type must be string", true)
		end
	end

	function newObject.fireIlluminationBomb(self, _power)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "fireIlluminationBomb(power)"
		
		if type(_power) == "number" then
			if _power >= 1 then
				local point = self:getPosition()
				point.y = point.y + 200 + math.random(0, 100)

				trigger.action.illuminationBomb(point, _power)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "power must be >= 1", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "power type must be number", true)
		end
	end
	
	function newObject.isInZones(self, _zones)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "isInZones(zones)"
		
		if ATMEFct.isObjectClass(_zones, ATME.C_Zones) == true then
			local isInside, _ = ATMEFct.C_Zones_isInside(_zones, self:getPosition())
			return isInside
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "zones type must be ATME.C_Zones", true)
		end
	end				
	
	function newObject.isInDCSZone(self, _name)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "isInDCSZone(name)"
		
		if type(_name) == "string" then
			local zone = ATMEFct.getDCSZone(_name)
			
			if zone ~= nil then
				local vector = ATMEFct.vectors["3D"].createFromReference(nil, nil, self:getPosition(), zone.point)
				if vector:getHModule() <= zone.radius then
					return true
				else
					return false
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone " .. _name .. " unknown", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", true)
		end
	end
	
	function newObject.isInZone(self, _zone)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "isInZone(zone)"
		
		if ATMEFct.isObjectClass(_zone, ATME.C_Zone) == true then
			local localMembersZone = ATMEDatas.localMembers[_zone:getIndex()]
			
			local datas = {
				zone = _zone,
				actualCenter = ATMEFct.C_Zone_getActualCenter(localMembersZone),
				azimuth = ATMEFct.C_Zone_getActualAzimuth(localMembersZone),
			}
		
			return ATMEFct[ATMEConst.indZoneConfig[_zone:getClassName()].fctIsInside](self:getPosition(), datas)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "zone type must be ATME.C_Zone", true)
		end
	end

	ATMEDatas.newCreatedStatics[localMembers.name] = {
		object = newObject,
	}
	
	if ATMEDatas.staticsToBirth[localMembers.name] ~= nil then
		ATMEDatas.newCreatedStatics[localMembers.name].isSpawned = ATMEDatas.staticsToBirth[localMembers.name].isSpawned
		ATMEDatas.newCreatedStatics[localMembers.name].spawnCallback = ATMEDatas.staticsToBirth[localMembers.name].spawnCallback
		ATMEDatas.newCreatedStatics[localMembers.name].staticCloningContext =  ATMEDatas.staticsToBirth[localMembers.name].staticCloningContext
		
		if ATMEDatas.staticsToBirth[localMembers.name].idSpawnedList ~= nil then
			localMembers.idSpawnedList = {
				datas = ATMEDatas.staticsToBirth[localMembers.name].idSpawnedList,
				id = ATMEDatas.staticsToBirth[localMembers.name].id,
			}
		end
		
		ATMEDatas.staticsToBirth[localMembers.name] = nil
	end
	
	newObject.modules = {}
	setmetatable(newObject.modules, ATMEDatas.C_Static_MT.mtModules)
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
		newObject.modules[_localModuleMembers.name] = {}
	end
	
	ATMEFct.outputCore(nil,"Prepare static unit " .. localMembers.name .. " -> ok", 2)
end

ATMEFct.C_Static_delete = function(_staticUnit)
	local localMembers = ATMEDatas.localMembers[_staticUnit:getIndex()]
	
	if localMembers.destroyed == false then
		local flag = "DESTROYED"
		if localMembers.isEnabled == false then
			flag = "DISABLED"
		end
		
		localMembers.lastPosition = localMembers.objectDCS:getPosition()
		
		for _i = #ATMEDatas.modulesNames, 1, -1 do
			local localModuleMembers = ATMEDatas.modulesLocalMembers[ATMEDatas.modulesNames[_i]]
			if localModuleMembers.handlers.onDeleteStaticHandler ~= nil then
				ATMEDatas.currentModuleName = localModuleMembers.name
				
				-- Appel du module
				ATMEFct.outputCore(nil,"Appel onDeleteStaticHandler du module " .. localModuleMembers.name, 2)
				xpcall(function() localModuleMembers.handlers.onDeleteStaticHandler(_staticUnit, flag) end, function() ATMEFct.errUserHandler('onDeleteStaticHandler') end)
			end
		end
		
		if localMembers.idSpawnedList ~= nil then
			if localMembers.idSpawnedList.id ~= nil then
				table.insert(localMembers.idSpawnedList.datas.freeIds, localMembers.idSpawnedList.id)
			end
		end
		
		localMembers.destroyed = true
		_staticUnit.modules = nil
		
		ATMEDatas.C_Static_MT:remove(_staticUnit:getIndex())
		
		ATMEDatas.statics[localMembers.name] = nil
		
		ATMEFct.outputCore(nil,"Delete static unit = " .. localMembers.name, 2)
	end
end

ATMEFct.C_Static_disable = function(_staticUnit)
	local staticDCS = _staticUnit:getDCSObject()
	local localMembers = ATMEDatas.localMembers[_staticUnit:getIndex()]
	localMembers.isEnabled = false
		
	staticDCS:destroy()
	ATMEFct.C_Static_delete(_staticUnit)
end

setmetatable(ATME.C_Static,ATMEDatas.C_Static_MT)

ATMEDatas.C_Static_MT.mt.__metatable = "Private"
ATMEDatas.C_Static_MT.mt.__index = ATME.C_Static
ATMEDatas.C_Static_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Static_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Static : Bad key " .. k, "table locked", false)
	end
end

ATMEDatas.C_Static_MT.mtModules.__metatable = "Private"
ATMEDatas.C_Static_MT.mtModules.__newindex = function(t, k, v)
	if ATMEDatas.modules[k] ~= nil then
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil,"ATME : ATME.C_Static.modules Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Cargo : gère les objets cargos au format ATME. Pas de contructeur accessible
-- pour les modules externes. 
-- Les instances de cette classe sont intégralement gérées par le module ATME_Core. La création et la destruction
-- dépendent du contexte de début de mission et des événements qui se produisent dans DCS en cours de mission.
-- Les modules externes accèdent à cette gestion indirectement par les handlers définis dans ceux-ci et les méthodes
-- accessibles de cette classe. Ceci permet de garantir une cohérence sur les listes internes ATME.
----------------------------------------------------------------------------------------------------------

ATMEDatas.C_Cargo_MT = {
	className = "C_Cargo",
	mtModules = {},
	mt = {},
	
	keys = {
		["getCategoryName"] = true,
		["getLife"] = true,
		["getDCSObject"] = true,
		["getDCSID"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["getIndex"] = true,
		["getTypeName"] = true,
		["getPosition"] = true,
		["inAir"] = true,
		["getWeight"] = true,
		["getAzimuth"] = true,
		["getCountryName"] = true,
		["getCoalitionName"] = true,
		["isInZones"] = true,
		["isInDCSZone"] = true,
		["isInZone"] = true,
		["explode"] = true,
		["fireFlare"] = true,
		["fireSmoke"] = true,
		["fireIlluminationBomb"] = true, 
		["disable"] = true,
		["isDisabled"] = true,
		["getUserPrivateDatas"] = true,
		["getModelDatas"] = true,
		["modules"] = true,
	},

	add = function(self, _name, _cargo, _datas)
		local item = {
			name = _name,
			objectDCS = _cargo,
			objectDCSId = _cargo:getID(),
			typeName = _cargo:getTypeName(),
			isEnabled = true,
			destroyed = false,
			userPrivateDatas = {},						-- Point d'entrée des données privées : un par module
			tagsName = ATMEFct.decompString(_name),
			category = "CARGO",
			spawnDatas = _datas,
		}
		
		item.spawnDatas.name = _name
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
	end,
}

ATME.C_Cargo = {
	getClassName = function(self)
		return ATMEDatas.C_Cargo_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Cargo_MT.parentClass
	end,
	
	getByName = function(_cargoName)
		if type(_cargoName) == "string" then
			return ATMEDatas.cargos[_cargoName]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Cargo.getByName(name)", "name must be string", false)
		end	
	end,
	foreach = function(_callback, _datas)
		local fctDefinition = "Function ATME.C_Cargo.foreach(callback, [datas])"
		
		if type(_callback) == "function" then
			ATMEFct.foreachObject(nil, "cargos", _callback, _datas)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "callback type must be function", false)
		end
	end,
	foreachInCoalition = function(_coalitionName, _callback, _datas)
		local fctDefinition = "Function ATME.C_Cargo.foreachInCoalition(coalitionName, callback, [datas])"
		
		if type(_coalitionName) == "string" and type(_callback) == "function" then
			if ATMEConst.coalitionIdsBlueRed[_coalitionName] == nil then
				ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName must be 'BLUE' or 'RED'", false)
				return
			end
			
			ATMEFct.foreachObject(_coalitionName, "cargos", _callback, _datas)
		else
			if type(_coalitionName) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName type must be string", false)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "callback type must be function", false)
			end
		end
	end,
}


ATMEFct.C_Cargo_create = function(_cargo, _datas)
	local newCargo = {}

	local datas
	if _datas == nil then
		local cloningContextLocalMembers = ATMEDatas.localMembers[ATMEDatas.cargosToBirth[_cargo:getName()].cargoCloningContext:getIndex()]
		
		datas = {
			country = cloningContextLocalMembers.country,
			datas = {
				type = cloningContextLocalMembers.cargoDatas.type,
				shape_name = cloningContextLocalMembers.cargoDatas.shape_name,
				rate = cloningContextLocalMembers.cargoDatas.rate,
			},
		}
	else
		datas = {
			country = _datas.country,
			datas = {
				type = _datas.type,
				shape_name = _datas.shape_name,
				rate = _datas.rate,
			},
		}
	end
	
	local index = ATMEDatas.C_Cargo_MT:add(_cargo:getName(), _cargo, datas)
	local localMembers = ATMEDatas.localMembers[index]

	setmetatable(newCargo, ATMEDatas.C_Cargo_MT.mt)

	function newCargo.getIndex(self)
		return index
	end
	
	function newCargo.getDCSObject(self)
		return localMembers.objectDCS
	end
	
	function newCargo.getDCSID(self)
		return localMembers.objectDCSId
	end
	
	function newCargo.getName(self)
		return localMembers.name
	end
	
	function newCargo.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargo) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if localMembers.tagsName ~= nil and pos >= 1 and pos <= #localMembers.tagsName then
				return localMembers.tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
		
	function newCargo.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newCargo.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargo) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if localMembers.tagsName ~= nil then
				local ok = ATMEFct.compareTags(localMembers.tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newCargo.getCategoryName(self)
		-- Catégories objets cargos (sous ensemble static) - A ne pas confondre avec catégorie des statics/cargo
		-- 		Cargos
		
		if localMembers.destroyed == false then
			return localMembers.category
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newCargo) .. "getCategoryName()", "already disabled", false)
		end
	end
	
	function newCargo.getTypeName(self)
		return localMembers.typeName
	end
	
	function newCargo.getModelDatas(self)
		return {
			type = localMembers.spawnDatas.datas.type,
			shape_name = localMembers.spawnDatas.datas.shape_name,
		}
	end

	function newCargo.getLife(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:getLife()
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newCargo) .. "getLife()", "already disabled", false)
		end
	end
	
	function newCargo.getPosition(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:getPoint()
		else
			return localMembers.lastPosition.p
		end
	end

	function newCargo.getWeight(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:getCargoWeight()
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newCargo) .. "getWeight()", "already disabled", false)
		end
	end

	function newCargo.inAir(self)
		if localMembers.destroyed == false then
			local pos = localMembers.objectDCS:getPoint()
			
			local delta = pos.y - ATMEFct.getSurfaceAltitude(pos)
			
			if delta < 2 then
				return false
			else
				return true
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newCargo) .. "inAir()", "already disabled", false)
		end
	end

	function newCargo.getAzimuth(self)
		local position
		if localMembers.destroyed == false then
			position = localMembers.objectDCS:getPosition().x
		else
			position = localMembers.lastPosition.x
		end
		
		local vector = ATMEFct.vectors["3D"].createFromTable(nil, nil, position)
		return vector:getAzimuth()
	end

	function newCargo.getCountryName(self)
		if localMembers.destroyed == false then
			local id = localMembers.objectDCS:getCountry()
			return country.name[id]
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newCargo) .. "getCountryName()", "already disabled", false)
		end
	end

	function newCargo.getCoalitionName(self)
		if localMembers.destroyed == false then
			local id = localMembers.objectDCS:getCoalition()
			return ATMEConst.coalitionNames[id]
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newCargo) .. "getCoalitionName()", "already disabled", false)
		end
	end
	
	function newCargo.disable(self)
		if localMembers.destroyed == false then
			ATMEDatas.cargosToDisable[localMembers.name] = self
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newCargo) .. "disable()", "already disabled", false)
		end
	end
	
	function newCargo.isDisabled(self)
		return localMembers.objectDCS:isExist() == false
	end
	
	function newCargo.getUserPrivateDatas(self, _moduleName)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargo) .. "getUserPrivateDatas([moduleName])"
		local moduleName
		
		if _moduleName ~= nil then
			if type(_moduleName) == "string" then
				if localMembers.userPrivateDatas[_moduleName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
					return
				else
					moduleName = _moduleName
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
				return
			end
		else
			moduleName = ATMEDatas.currentModuleName
		end
		
		return localMembers.userPrivateDatas[moduleName]
	end

	function newCargo.explode(self, _power)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargo) .. "explode(power)"
		
		if type(_power) == "number" then
			if _power > 0 then
				local point = self:getPosition()
				if localMembers.destroyed == true then
					local alt =  ATMEFct.getSurfaceAltitude(point)
					point.y = alt
				end
				
				trigger.action.explosion(point, _power)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "power must be >= 1", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "power type must be number", true)
		end
	end
	
	function newCargo.fireFlare(self, _color)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargo) .. "fireFlare(color)"
		
		if type(_color) == "string" then
			if ATMEConst.flareColors[string.upper(_color)]~= nil then
				local point = self:getPosition()
				if localMembers.destroyed == false then
					if localMembers.objectDCS:inAir() == true then
						ATMEFct.setError("IN_AIR")
						ATMEFct.warningOutputCore(nil, fctDefinition, "in air, can't fire flare", false)
						return
					end
				else
					local alt =  ATMEFct.getSurfaceAltitude(point)
					point.y = alt
				end
				
				if localMembers.flare == nil then
					localMembers.flare = ATMEFct.C_Flare_create(_color, point)
				else
					ATMEDatas.localMembers[localMembers.flare:getIndex()].point = point
				end
				
				localMembers.flare:activate()
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "color must be a flare color or RANDOM", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "color type must be string", true)
		end
	end

	function newCargo.fireSmoke(self, _color)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargo) .. "fireSmoke(color)"
		
		if type(_color) == "string" then
			if ATMEConst.smokeColors[string.upper(_color)]~= nil then
				local point = self:getPosition()
				if localMembers.destroyed == false then
					if localMembers.objectDCS:inAir() == true then
						ATMEFct.setError("IN_AIR")
						ATMEFct.warningOutputCore(nil, fctDefinition, "in air, can't fire smoke", false)
						return
					end
				else
					local alt =  ATMEFct.getSurfaceAltitude(point)
					point.y = alt
				end
				
				local newPoint = {
					x = point.x + math.random(-20,20),
					y = point.y,
					z = point.z + math.random(-20,20),
				}	
				
				local smoke = ATMEFct.C_Smoke_create(_color, newPoint)
				smoke:activate(false)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "color must be a smoke color", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "color type must be string", true)
		end
	end

	function newCargo.fireIlluminationBomb(self, _power)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargo) .. "fireIlluminationBomb(power)"
		
		if type(_power) == "number" then
			if _power >= 1 then
				local point = self:getPosition()
				if localMembers.destroyed == false then
					if localMembers.objectDCS:inAir() == true then
						ATMEFct.setError("IN_AIR")
						ATMEFct.warningOutputCore(nil, fctDefinition, "in air, can't fire illumination bomb", false)
						return
					end
				else
					local alt =  ATMEFct.getSurfaceAltitude(point)
					point.y = alt + 200 + math.random(0, 100)
				end
				
				trigger.action.illuminationBomb(point, _power)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "power must be >= 1", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "power type must be number", true)
		end
	end
	
	function newCargo.isInZones(self, _zones)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargo) .. "isInZones(zones)"
		
		if ATMEFct.isObjectClass(_zones, ATME.C_Zones) == true then
			local isInside, _ = ATMEFct.C_Zones_isInside(_zones, self:getPosition())
			return isInside
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "zones type must be ATME.C_Zones", true)
		end
	end				
	
	function newCargo.isInDCSZone(self, _name)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargo) .. "isInDCSZone(name)"
		
		if type(_name) == "string" then
			local zone = ATMEFct.getDCSZone(_name)
			
			if zone ~= nil then
				local vector = ATMEFct.vectors["3D"].createFromReference(nil, nil, self:getPosition(), zone.point)
				if vector:getHModule() <= zone.radius then
					return true
				else
					return false
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone " .. _name .. " unknown", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", true)
		end
	end
	
	function newCargo.isInZone(self, _zone)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargo) .. "isInZone(zone)"
		
		if ATMEFct.isObjectClass(_zone, ATME.C_Zone) == true then
			local localMembersZone = ATMEDatas.localMembers[_zone:getIndex()]
			
			local datas = {
				zone = _zone,
				actualCenter = ATMEFct.C_Zone_getActualCenter(localMembersZone),
				azimuth = ATMEFct.C_Zone_getActualAzimuth(localMembersZone),
			}
		
			return ATMEFct[ATMEConst.indZoneConfig[_zone:getClassName()].fctIsInside](self:getPosition(), datas)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "zone type must be ATME.C_Zone", true)
		end
	end
	
	ATMEDatas.newCreatedCargos[localMembers.name] = {
		cargo = newCargo,
	}
	
	if ATMEDatas.cargosToBirth[localMembers.name] ~= nil then
		ATMEDatas.newCreatedCargos[localMembers.name].isSpawned = ATMEDatas.cargosToBirth[localMembers.name].isSpawned
		ATMEDatas.newCreatedCargos[localMembers.name].spawnCallback = ATMEDatas.cargosToBirth[localMembers.name].spawnCallback
		ATMEDatas.newCreatedCargos[localMembers.name].cargoCloningContext =  ATMEDatas.cargosToBirth[localMembers.name].cargoCloningContext
		
		if ATMEDatas.cargosToBirth[localMembers.name].idSpawnedList ~= nil then
			localMembers.idSpawnedList = {
				datas = ATMEDatas.cargosToBirth[localMembers.name].idSpawnedList,
				id = ATMEDatas.cargosToBirth[localMembers.name].id,
			}
		end
		
		ATMEDatas.cargosToBirth[localMembers.name] = nil
	end
	
	newCargo.modules = {}
	setmetatable(newCargo.modules, ATMEDatas.C_Cargo_MT.mtModules)
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
		newCargo.modules[_localModuleMembers.name] = {}
	end
	
	ATMEFct.outputCore(nil,"Prepare Cargo " .. localMembers.name .. " -> ok", 2)
end

ATMEFct.C_Cargo_delete = function(_cargoUnit)
	local localMembers = ATMEDatas.localMembers[_cargoUnit:getIndex()]
	
	if localMembers.destroyed == false then
		local flag = "DESTROYED"
		if localMembers.isEnabled == false then
			flag = "DISABLED"
		end
		
		localMembers.lastPosition = localMembers.objectDCS:getPosition()
		
		for _i = #ATMEDatas.modulesNames, 1, -1 do
			local localModuleMembers = ATMEDatas.modulesLocalMembers[ATMEDatas.modulesNames[_i]]
			if localModuleMembers.handlers.onDeleteCargoHandler ~= nil then
				ATMEDatas.currentModuleName = localModuleMembers.name
				
				-- Appel du module
				ATMEFct.outputCore(nil,"Appel onDeleteCargoHandler du module " .. localModuleMembers.name.. " pour le cargo " .. _cargoUnit:getName(), 2)
				xpcall(function() localModuleMembers.handlers.onDeleteCargoHandler(_cargoUnit, flag) end, function() ATMEFct.errUserHandler('onDeleteCargoHandler') end)
			end
		end
		
		if localMembers.idSpawnedList ~= nil then
			if localMembers.idSpawnedList.id ~= nil then
				table.insert(localMembers.idSpawnedList.datas.freeIds, localMembers.idSpawnedList.id)
			end
		end
		
		localMembers.destroyed = true
		_cargoUnit.modules = nil
		
		ATMEDatas.C_Cargo_MT:remove(_cargoUnit:getIndex())
		
		ATMEDatas.cargos[localMembers.name] = nil
		
		ATMEFct.outputCore(nil,"Delete cargo = " .. localMembers.name, 2)
	end
end

ATMEFct.C_Cargo_disable = function(_cargoUnit)
	local cargoDCS = _cargoUnit:getDCSObject()
	local localMembers = ATMEDatas.localMembers[_cargoUnit:getIndex()]
	localMembers.isEnabled = false
		
	cargoDCS:destroy()
	ATMEFct.C_Cargo_delete(_cargoUnit)
end

setmetatable(ATME.C_Cargo,ATMEDatas.C_Cargo_MT)

ATMEDatas.C_Cargo_MT.mt.__metatable = "Private"
ATMEDatas.C_Cargo_MT.mt.__index = ATME.C_Cargo
ATMEDatas.C_Cargo_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Cargo_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Cargo : Bad key " .. k, "table locked", false)
	end
end

ATMEDatas.C_Cargo_MT.mtModules.__metatable = "Private"
ATMEDatas.C_Cargo_MT.mtModules.__newindex = function(t, k, v)
	if ATMEDatas.modules[k] ~= nil then
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil,"ATME : ATME.C_Cargo.modules Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Weapon : gère les armes au format ATME. Pas de contructeur accessible
-- pour les modules externes.
-- Les instances de cette classe sont intégralement gérées par le module ATME_Core. La création et la destruction
-- dépendent des événements qui se produisent dans DCS en cours de mission.
-- Seules les armes activent sont référencées.
-- Les noms d'objets sont des nombres
-- Ci dessous dans les fonctions de la classe, isExist permet de s'assurer de l'existance de l'arme car
-- pas de destruction détectée à l'impact si l'arme ne touche rien (bâtiment, unité....) - Bug DCS? 
----------------------------------------------------------------------------------------------------------

ATMEDatas.C_Weapon_MT = {
	className = "C_Weapon",
	
	mtModules = {},
	mt = {},
	
	keys = {
		["getDCSObject"] = true,
		["getIndex"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["getTypeName"] = true,
		["getCoalitionName"] = true,
		["getCountryName"] = true,
		["getPosition"] = true,
		["getMSLAltitude"] = true,
		["getAGLAltitude"] = true,
		["getVelocityVector"] = true,
		["getSpeed"] = true,
		["getHSpeed"] = true,
		["getVSpeed"] = true,
		["getSpeedAzimuth"] = true,
		["getRollAxisVector"] = true,
		["getAzimuth"] = true,
		["getLauncher"] = true,
		["getTarget"] = true,
		["explode"] = true,
		["isInZones"] = true,
		["isInDCSZone"] = true,
		["isInZone"] = true,
		["disable"] = true,
		["isDisabled"] = true,
		["getUserPrivateDatas"] = true,
		["modules"] = true,
	},
	
	add = function(self, _weapon)
		local item = {
			name = _weapon:getName(),
			objectDCS = _weapon,
			typeName = _weapon:getTypeName(),
			tagsName = ATMEFct.decompString(tostring(_weapon:getName())),
			userPrivateDatas = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
	end,
}

ATME.C_Weapon = {
	getClassName = function(self)
		return ATMEDatas.C_Weapon_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Weapon_MT.parentClass
	end,
}


ATMEFct.C_Weapon_create = function(_weapon)
	local newWeapon = {}
	
	setmetatable(newWeapon, ATMEDatas.C_Weapon_MT.mt)
	local index = ATMEDatas.C_Weapon_MT:add(_weapon)
	local localMembers = ATMEDatas.localMembers[index]
	
	function newWeapon.getIndex(self)
		return index
	end
	
	function newWeapon.getDCSObject(self)
		return localMembers.objectDCS
	end
	
	function newWeapon.getName(self)
		return localMembers.name
	end
	
	function newWeapon.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargo) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if localMembers.tagsName ~= nil and pos >= 1 and pos <= #localMembers.tagsName then
				return localMembers.tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
		
	function newWeapon.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newWeapon.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargo) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if localMembers.tagsName ~= nil then
				local ok = ATMEFct.compareTags(localMembers.tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newWeapon.getCountryName(self)
		if localMembers.objectDCS:isExist() == true then
			local id = localMembers.objectDCS:getCountry()
			return country.name[id]
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newWeapon) .. "getCountryName()", "already disabled", false)
		end
	end

	function newWeapon.getCoalitionName(self)
		if localMembers.objectDCS:isExist() == true then
			local id = localMembers.objectDCS:getCoalition()
			return ATMEConst.coalitionNames[id]
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newWeapon) .. "getCoalitionName()", "already disabled", false)
		end
	end
	
	function newWeapon.getTypeName(self)
		if localMembers.objectDCS:isExist() == true then
			return localMembers.typeName
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newWeapon) .. "getTypeName()", "already disabled", false)
		end
	end
	
	function newWeapon.getLauncher(self)
		if localMembers.objectDCS:isExist() == true then
			local launcher = localMembers.objectDCS:getLauncher()
			
			if launcher ~= nil then
				return ATMEDatas.units[launcher:getName()]
			end
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newWeapon) .. "getLauncher()", "already disabled", false)
		end
	end
	
	function newWeapon.getTarget(self)
		if localMembers.objectDCS:isExist() == true then
			local target = localMembers.objectDCS:getTarget()
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newWeapon) .. "getTarget()", "already disabled", false)
		end
	end
	
	function newWeapon.getPosition(self)
		if localMembers.objectDCS:isExist() == true then
			return localMembers.objectDCS:getPoint()
		else
			return localMembers.lastPosition.p
		end
	end
	
	function newWeapon.getVelocityVector(self)
		local velocity
		if localMembers.objectDCS:isExist() == true then
			velocity = localMembers.objectDCS:getVelocity()
		else
			velocity = localMembers.lastVelocity
		end

		return ATMEFct.vectors["3D"].createFromTable(nil, nil, velocity)
	end

	function newWeapon.getSpeed(self)
		local velocity
		if localMembers.objectDCS:isExist() == true then
			velocity = localMembers.objectDCS:getVelocity()
		else
			velocity = localMembers.lastVelocity
		end
			
		local vector = ATMEFct.vectors["3D"].createFromTable(nil, nil, velocity)
		return vector:getModule()
	end

	function newWeapon.getHSpeed(self)
		local velocity
		if localMembers.objectDCS:isExist() == true then
			velocity = localMembers.objectDCS:getVelocity()
		else
			velocity = localMembers.lastVelocity
		end
		
		local vector = ATMEFct.vectors["3D"].createFromTable(nil, nil, velocity)
		return vector:getHModule()
	end

	function newWeapon.getVSpeed(self)
		local velocity
		if localMembers.objectDCS:isExist() == true then
			velocity = localMembers.objectDCS:getVelocity()
		else
			velocity = localMembers.lastVelocity
		end
			
		return velocity.y
	end
	
	function newWeapon.getMSLAltitude(self)
		return ATME.getPointAltitude(self:getPosition())
	end

	function newWeapon.getAGLAltitude(self)
		local point = self:getPosition()
		
		local landHeight = ATMEFct.getSurfaceAltitude(point)
		return ATME.getPointAltitude(point) - landHeight
	end

	function newWeapon.getSpeedAzimuth(self)
		local velocity
		if localMembers.objectDCS:isExist() == true then
			velocity = localMembers.objectDCS:getVelocity()
		else
			velocity = localMembers.lastVelocity
		end
		
		local vector = ATMEFct.vectors["3D"].createFromTable(nil, nil, velocity)
				
		if vector:getModule() > 1 then 
			return vector:getAzimuth()
		else
			return 0
		end
	end

	function newWeapon.getRollAxisVector(self)
		local position
		if localMembers.objectDCS:isExist() == true then
			position = localMembers.objectDCS:getPosition().x
		else
			position = localMembers.lastPosition.x
		end

		return ATMEFct.vectors["3D"].createFromTable(nil, nil, position)
	end

	function newWeapon.getAzimuth(self)
		local position
		if localMembers.destroyed == false then
			position = localMembers.objectDCS:getPosition().x
		else
			position = localMembers.lastPosition.x
		end
		
		local vector = ATMEFct.vectors["3D"].createFromTable(nil, nil, position)
		return vector:getAzimuth()
	end
	
	function newWeapon.disable(self)
		if localMembers.objectDCS:isExist() == true then
			localMembers.objectDCS:destroy()
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newWeapon) .. "disable()", "already disabled", false)
		end
		
		ATMEFct.C_Weapon_delete(localMembers.name)
	end
	
	function newWeapon.isDisabled(self)
		return localMembers.objectDCS:isExist()
	end

	function newWeapon.explode(self, _power)
		local fctDefinition = ATMEFct.getErrorPrefix(newWeapon) .. "explode(power)"
		
		if type(_power) == "number" then
			if _power > 0 then
				trigger.action.explosion(self:getPosition(), _power)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "power must be >= 1", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "power type must be number", true)
		end
	end
	
	function newWeapon.getUserPrivateDatas(self, _moduleName)
		local fctDefinition = ATMEFct.getErrorPrefix(newWeapon) .. "getUserPrivateDatas([moduleName])"
		local moduleName
		
		if _moduleName ~= nil then
			if type(_moduleName) == "string" then
				if localMembers.userPrivateDatas[_moduleName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
					return
				else
					moduleName = _moduleName
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
				return
			end
		else
			moduleName = ATMEDatas.currentModuleName
		end
		
		return localMembers.userPrivateDatas[moduleName]
	end
	
	function newWeapon.isInZones(self, _zones)
		local fctDefinition = ATMEFct.getErrorPrefix(newWeapon) .. "isInZones(zones)"
		
		if ATMEFct.isObjectClass(_zones, ATME.C_Zones) == true then
			local isInside, _ = ATMEFct.C_Zones_isInside(_zones, self:getPosition())
			return isInside
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "zones type must be ATME.C_Zones", true)
		end
	end				

	function newWeapon.isInDCSZone(self, _name)
		local fctDefinition = ATMEFct.getErrorPrefix(newWeapon) .. "isInDCSZone(name)"
		
		if type(_name) == "string" then
			local zone = ATMEFct.getDCSZone(_name)
	
			if zone ~= nil then
				local vector = ATMEFct.vectors["3D"].createFromReference(nil, nil, self:getPosition(), zone.point)
				if vector:getHModule() <= zone.radius then
					return true
				else
					return false
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone " .. _name .. " unknown", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", true)
		end
	end
	
	function newWeapon.isInZone(self, _zone)
		local fctDefinition = ATMEFct.getErrorPrefix(newWeapon) .. "isInZone(zone)"
		
		if ATMEFct.isObjectClass(_zone, ATME.C_Zone) == true then
			local localMembersZone = ATMEDatas.localMembers[_zone:getIndex()]
			
			local datas = {
				zone = _zone,
				actualCenter = ATMEFct.C_Zone_getActualCenter(localMembersZone),
				azimuth = ATMEFct.C_Zone_getActualAzimuth(localMembersZone),
			}
					
			return ATMEFct[ATMEConst.indZoneConfig[_zone:getClassName()].fctIsInside](self:getPosition(), datas)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "zone type must be ATME.C_Zone", true)
		end
	end
	
	newWeapon.modules = {}
	setmetatable(newWeapon.modules, ATMEDatas.C_Weapon_MT.mtModules)
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
		newWeapon.modules[_localModuleMembers.name] = {}
	end
	
	ATMEDatas.weapons[localMembers.name] = newWeapon
			
	ATMEFct.outputCore(nil,"Création weapon " .. localMembers.name .. " -> ok", 2)
	
	return newWeapon
end

ATMEFct.C_Weapon_delete = function(_name)
	if ATMEDatas.weapons[_name] ~= nil then
		local localMembers = ATMEDatas.localMembers[ATMEDatas.weapons[_name]:getIndex()]
		if localMembers.objectDCS:isExist() == true then
			localMembers.lastPosition = localMembers.objectDCS:getPosition()
			localMembers.lastVelocity = localMembers.objectDCS:getVelocity()
		end
	
		ATMEDatas.C_Weapon_MT:remove(ATMEDatas.weapons[_name]:getIndex())
		ATMEDatas.weapons[_name].modules = nil
		ATMEDatas.weapons[_name] = nil
		ATMEDatas.weaponsToDelete[_name] = nil
		ATMEFct.outputCore(nil,"Delete weapon = " .. _name, 2)
	end
end

setmetatable(ATME.C_Weapon,ATMEDatas.C_Weapon_MT)

ATMEDatas.C_Weapon_MT.mt.__metatable = "Private"
ATMEDatas.C_Weapon_MT.mt.__index = ATME.C_Weapon
ATMEDatas.C_Weapon_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Weapon_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Weapon : Bad key " .. k, "table locked", false)
	end
end

ATMEDatas.C_Weapon_MT.mtModules.__metatable = "Private"
ATMEDatas.C_Weapon_MT.mtModules.__newindex = function(t, k, v)
	if ATMEDatas.modules[k] ~= nil then
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil,"ATME : ATME.C_Weapon_MT.modules Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_SceneryObject : gère les objects statiques positionnées sur les cartes de DCS.
-- Les instances de cette classe sont définies par un nom qui est un identifiant numérique.
-- Les instances référencées ne sont que les objets ayant été touchés par une arme mais non détruits
-- ou les objects retrouvés par ATME.C_FindContext.
-- Les noms d'objets sont des nombres
--
-- A noter : getDesc et getLife non fonctionnels arrêt immédiat du script - Bug DCS
--           destroy non opérationnel, l'objet reste visible - Bug DCS
--			 Paradoxalement getDescByName fonctionne. En fait, tous les objets retournés par C_FindContext
--			 sont des "Buildings".
----------------------------------------------------------------------------------------------------------

ATMEDatas.C_SceneryObject_MT = {
	className = "C_SceneryObject",
	
	mtModules = {},
	mt = {},
	
	keys = {
		["getDCSObject"] = true,
		["getName"] = true,
		["getTypeName"] = true,
		["getPosition"] = true,
		["explode"] = true,
		["getIndex"] = true,
		["getUserPrivateDatas"] = true,
		["disable"] = true,
		["isDisabled"] = true,
		["fireFlare"] = true,
		["fireSmoke"] = true,
		["fireIlluminationBomb"] = true, 
		["isInZones"] = true,
		["isInDCSZone"] = true,
		["isInZone"] = true,
		["modules"] = true,
	},
	
	add = function(self, _object)
		local item = {
			name = _object:getName(),
			objectDCS = _object,
			typeName = _object:getTypeName(),
			userPrivateDatas = {},
			destroyed = false,
			lastPosition = _object:getPosition()
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
	end,
}

ATME.C_SceneryObject = {
	getClassName = function(self)
		return ATMEDatas.C_SceneryObject_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_SceneryObject_MT.parentClass
	end,
}


ATMEFct.C_SceneryObject_create = function(_object)
	local newObject = {}
	
	setmetatable(newObject, ATMEDatas.C_SceneryObject_MT.mt)
	local index = ATMEDatas.C_SceneryObject_MT:add(_object)
	local localMembers = ATMEDatas.localMembers[index]

	function newObject.getIndex(self)
		return index
	end
	
	function newObject.getDCSObject(self)
		return localMembers.objectDCS
	end
	
	function newObject.getName(self)
		return localMembers.name
	end
	
	function newObject.getTypeName(self)
		return localMembers.typeName
	end
	
		
	function newObject.getPosition(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:getPoint()
		else
			return localMembers.lastPosition.p
		end
	end
	
	function newObject.explode(self, _power)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "explode(power)"
		
		if type(_power) == "number" then
			if _power > 0 then
				trigger.action.explosion(self:getPosition(), _power)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "power must be >= 1", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "power type must be number", true)
		end
	end
	
	function newObject.disable(self)
		if localMembers.destroyed == false then
			localMembers.objectDCS:destroy()
		else
			ATMEFct.setError("ALREADY_DISABLED")
			ATMEFct.warningOutputCore(nil, ATMEFct.getErrorPrefix(newObject) .. "disable()", "already disabled", false)
		end
		
		ATMEFct.C_SceneryObject_delete(localMembers.name)
	end
	
	function newObject.isDisabled(self)
		return localMembers.destroyed == false
	end
	
	function newObject.fireFlare(self, _color)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "fireFlare(color)"
		
		if type(_color) == "string" then
			if ATMEConst.flareColors[string.upper(_color)]~= nil then
				if localMembers.flare == nil then
					localMembers.flare = ATMEFct.C_Flare_create(_color, self:getPosition())
				end
				localMembers.flare:activate()
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "color must be a flare color or RANDOM", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "color type must be string", true)
		end
	end

	function newObject.fireSmoke(self, _color)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "fireSmoke(color)"
		
		if type(_color) == "string" then
			if ATMEConst.smokeColors[string.upper(_color)]~= nil then 
				local point = self:getPosition()
				local newPoint = {
					x = point.x + math.random(-20,20),
					y = point.y,
					z = point.z + math.random(-20,20),
				}
					
				local smoke = ATMEFct.C_Smoke_create(_color, newPoint)
				smoke:activate(false)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "color must be a smoke color", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "color type must be string", true)
		end
	end

	function newObject.fireIlluminationBomb(self, _power)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "fireIlluminationBomb(power)"
		
		if type(_power) == "number" then
			if _power >= 1 then
				local point = self:getPosition()
				
				point.y = point.y + 200 + math.random(0, 100)

				trigger.action.illuminationBomb(point, _power)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "power must be >= 1", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "power type must be number", true)
		end
	end
	
	function newObject.getUserPrivateDatas(self, _moduleName)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "getUserPrivateDatas([moduleName])"
		local moduleName
		
		if _moduleName ~= nil then
			if type(_moduleName) == "string" then
				if localMembers.userPrivateDatas[_moduleName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Module '" .. _moduleName .. "' doesn't exist in mission", true)
					return
				else
					moduleName = _moduleName
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "moduleName type must be nil or string", true)
				return
			end
		else
			moduleName = ATMEDatas.currentModuleName
		end
		
		return localMembers.userPrivateDatas[moduleName]
	end
	
	function newObject.isInZones(self, _zones)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "isInZones(zones)"
		
		if ATMEFct.isObjectClass(_zones, ATME.C_Zones) == true then
			return _zones:isInside(self) 
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "zones type must be ATME.C_Zones", true)
		end
	end				
	
	function newObject.isInDCSZone(self, _name)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "isInDCSZone(name)"
		
		if type(_name) == "string" then
			local zone = ATMEFct.getDCSZone(_name)
			
			if zone ~= nil then
				local vector = ATMEFct.vectors["3D"].createFromReference(nil, nil, self:getPosition(), zone.point)
				if vector:getHModule() <= zone.radius then
					return true
				else
					return false
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone " .. _name .. " unknown", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", true)
		end
	end
	
	function newObject.isInZone(self, _zone)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "isInZone(zone)"
		
		if ATMEFct.isObjectClass(_zone, ATME.C_Zone) == true then
			local point = self:getPosition()
			local localMembersZone = ATMEDatas.localMembers[_zone:getIndex()]
			
			local datas = {
				zone = _zone,
				actualCenter = ATMEFct.C_Zone_getActualCenter(localMembersZone),
				azimuth = ATMEFct.C_Zone_getActualAzimuth(localMembersZone),
			}
		
			return ATMEFct[ATMEConst.indZoneConfig[_zone:getClassName()].fctIsInside](point, datas)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "zone type must be ATME.C_Zone", true)
		end
	end
	
	newObject.modules = {}
	setmetatable(newObject.modules, ATMEDatas.C_SceneryObject_MT.mtModules)
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
		newObject.modules[_localModuleMembers.name] = {}
	end
	
	ATMEDatas.sceneryObjects[localMembers.name] = newObject
	
	ATMEFct.outputCore(nil,"Création objet de carte (sceneryObject) " .. localMembers.name .. " -> ok", 2)
end

ATMEFct.C_SceneryObject_delete = function(_name)
	if ATMEDatas.sceneryObjects[_name] ~= nil then
		local localMembers = ATMEDatas.localMembers[ATMEDatas.sceneryObjects[_name]:getIndex()]
		
		if localMembers.destroyed == false then
			localMembers.destroyed = true
			ATMEDatas.C_SceneryObject_MT:remove(ATMEDatas.sceneryObjects[_name]:getIndex())
			ATMEDatas.sceneryObjects[_name].modules = nil
			ATMEDatas.sceneryObjects[_name] = nil
			
			ATMEFct.outputCore(nil,"Delete sceneryObject = " .. _name, 2)
		end
	end
end

setmetatable(ATME.C_SceneryObject,ATMEDatas.C_SceneryObject_MT)

ATMEDatas.C_SceneryObject_MT.mt.__metatable = "Private"
ATMEDatas.C_SceneryObject_MT.mt.__index = ATME.C_SceneryObject
ATMEDatas.C_SceneryObject_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_SceneryObject_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_SceneryObject : Bad key " .. k, "table locked", false)
	end
end

ATMEDatas.C_SceneryObject_MT.mtModules.__metatable = "Private"
ATMEDatas.C_SceneryObject_MT.mtModules.__newindex = function(t, k, v)
	if ATMEDatas.modules[k] ~= nil then
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil,"ATME : ATME.C_SceneryObject_MT.modules Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- FIN DE DEFINITION DES CLASSES D'OBJETS DEFINIS DANS ATME
--
-- FONCTIONS DE GESTION ATME_CORE
----------------------------------------------------------------------------------------------------------

do
	local startMissionFcts = {
		[0] = function()
			-- Délai d'attente pour déclarer les initialisations OK : Actuellement 3 secondes pour considérer les groupes créés
			ATMEDatas.startOffset = ATMEDatas.startOffset + 1
			if ATMEDatas.startOffset == 3 then
				ATMEDatas.startMissionOk = 1
			end
		end,
		[1] = function()
			ATMEDatas.startMissionOk = 2
			
			-- Appel du module onStartMission après initialisation
			for _, _name in ipairs(ATMEDatas.modulesNames) do
				local localModuleMembers = ATMEDatas.modulesLocalMembers[_name]
				if localModuleMembers.handlers.onStartMissionHandler ~= nil then
					ATMEDatas.currentModuleName = localModuleMembers.name
					
					-- Appel du module
					ATMEFct.outputCore(nil,"Appel onStartMissionHandler après initialisation des unités & joueurs du module " .. localModuleMembers.name, 2)
					xpcall(localModuleMembers.handlers.onStartMissionHandler, function() ATMEFct.errUserHandler('onStartMissionHandler') end)
				end
			end
		end,
		[2] = function()
		end,
	}
	
	----------------------------------------------------------------------------------------------------------
	-- Fonction _timerMainFunction est appelée à l'initialisation voir plus bas, puis réarmée toutes les secondes.
	-- Cette fonction permet :
	--		- la mise à jour permanente des listes ATME des joueurs et unités AI.
	--      - la gestion des spawns multiples
	-- 		- la gestion des smokes (classe ATME.C_Smoke) qui doivent rester actifs et les tirs répétés de 
	-- 		   flares (classe ATME.C_Flare)
	--      - la gestion d'entrée/sortie des zones managées (classe ATME.C_Zones)
	--      - la gestion des pickupZones et des signaux éventuels
	--		- l'appel pour chaque module du handler onTimerHandler
	--		- l'appel des onCreateHandlers suite à la création des unités et groupes, suivi éventuellement d'un onSpawnHandler
	-- 		- l'appel des fonctions de mise des joueurs
	--		- après ces appels, la liste des events actifs est remise à 0.
	--		- Traitement des event core retardés. Certains evénéments sont retardés pour garantir leur bonne prise en 
	--        par DCS.
	--      - Gestion des routes pour les patrouilles
	--
	----------------------------------------------------------------------------------------------------------
	
	local _timerMainFunction
    _timerMainFunction = function()
		if ATMEDatas.exitATME == false then
			-- Fonction des mises à jour de la liste des joueurs ... 
			-- Important pour éviter les incohérences, les timers étant asynchrones
			
			ATMEFct.outputCore(nil,"Start main timer -> " .. ATMEDatas.startMissionOk, 2)
			
			-- Pour les groupes créés ou spawnés en attente, on enlève un cycle et on purge les groupes détruits entre temps
			
			for _name, _item in pairs(ATMEDatas.newCreatedGroups) do
				if _item.timerOffset > 0 then
					_item.timerOffset = _item.timerOffset - 1
				else
					local localMembersGroup = ATMEDatas.localMembers[_item.group:getIndex()]

					if localMembersGroup.objectDCS:isExist() == true and #localMembersGroup.objectDCS:getUnits() ~= 0 then
						local ok = false
						
						localMembersGroup.lastPosition = localMembersGroup.objectDCS:getUnits()[1]:getPoint()
						for _unitName, _unit in pairs(localMembersGroup.ATMEUnits) do
							local localMembers = ATMEDatas.localMembers[_unit:getIndex()]
							if localMembers.objectDCS:isExist() == true then
								ATMEDatas.units[_unitName] = _unit
								ok = true
							else
								ATMEFct.outputCore(nil, "newCreatedGroups -> Unité déjà détruite " .. _unitName, 2)
								ATMEDatas.flagDeleteUnit[_unitName] = "DESTROYED"
								ATMEFct.C_Unit_delete(_unit, ATMEDatas.reservedAsPlayerGroups[_name] > 0)
							end
						end
						
						if ok == true then
							ATMEDatas.groups[_name] = _item.group
							ATMEDatas.playerNamesByGroupId[localMembersGroup.objectDCSId] = _name
							
							ATMEDatas.newPreparedGroups[_name] = _item
							localMembersGroup.initOk = true
						else
							ATMEFct.outputCore(nil, "newCreatedGroups -> Groupe déjà détruit " .. _name, 2)
						end

						ATMEDatas.newCreatedGroups[_name] = nil
					end
				end
			end
			
			for _name, _item in pairs(ATMEDatas.newCreatedCargos) do
				local localMembers = ATMEDatas.localMembers[_item.cargo:getIndex()]
				
				if localMembers.objectDCS:isExist() == true then
					ATMEDatas.cargos[_name] = _item.cargo
					ATMEDatas.newPreparedCargos[_name] = _item
				else
					ATMEFct.outputCore(nil, "newCreatedCargos -> Cargo déjà détruit " .. _name, 2)
				end
				
				ATMEDatas.newCreatedCargos[_name] = nil
			end
			
			for _name, _item in pairs(ATMEDatas.newCreatedStatics) do
				local localMembers = ATMEDatas.localMembers[_item.object:getIndex()]
				
				if localMembers.objectDCS:isExist() == true then
					ATMEDatas.statics[_name] = _item.object
					ATMEDatas.newPreparedStatics[_name] = _item
				else
					ATMEFct.outputCore(nil, "newCreatedStatics -> Static déjà détruit " .. _name, 2)
				end
				
				ATMEDatas.newCreatedStatics[_name] = nil
			end
			
			startMissionFcts[ATMEDatas.startMissionOk]()
			
			if ATMEDatas.startMissionOk == 2 then
				local delay = 0.5
				for _name, _item in pairs(ATMEDatas.newPreparedGroups) do
					local localMembersGroup = ATMEDatas.localMembers[_item.group:getIndex()]

					if localMembersGroup.objectDCS:isExist() == true and #localMembersGroup.objectDCS:getUnits() ~= 0 then
						if ATMEFct.isGroupActivated(_item.group) == true then
							if localMembersGroup.justActivated > 0 then
								localMembersGroup.justActivated = localMembersGroup.justActivated - 1
							else

								ATMEFct.C_Group_CreateTrackingWaypoints(localMembersGroup)
								if localMembersGroup.idTimer == nil then
									localMembersGroup.idTimer = timer.scheduleFunction(ATMETimers.C_Group_timerTrackingWaypoint, _item.group, timer.getTime() + delay)
								end
								
								for _, _nameModule in ipairs(ATMEDatas.modulesNames) do
									local localModuleMembers = ATMEDatas.modulesLocalMembers[_nameModule]	
									ATMEDatas.currentModuleName = _nameModule
									
									if _item.disembark == false then
										
										for _, _unit in pairs(localMembersGroup.ATMEUnits) do
											if localModuleMembers.handlers.onCreateUnitHandler ~= nil then
												ATMEFct.outputCore(nil,"Appel onCreateUnitHandler du module " .. localModuleMembers.name, 2)
												xpcall(function() localModuleMembers.handlers.onCreateUnitHandler(_unit) end, function() ATMEFct.errUserHandler('onCreateUnitHandler') end)
											end
										end

										if localModuleMembers.handlers.onCreateGroupHandler ~= nil then
											ATMEFct.outputCore(nil,"Appel onCreateGroupHandler du module " .. localModuleMembers.name, 2)
											xpcall(function() localModuleMembers.handlers.onCreateGroupHandler(_item.group, ATMEDatas.respawnedGroups[_name] ~= nil) end, function() ATMEFct.errUserHandler('onCreateGroupHandler') end)
										end
									end
								end

								for _unitName, _unit in pairs(localMembersGroup.ATMEUnits) do
									if ATMEConst.canFly[localMembersGroup.category] == true and _unit:isEngineStarted() == true then
										-- Emission de l'événement démarrage moteur
										local event = {
											type = "ENGINE",
											datas = {
												action = "START",
												unitName = _unitName,
											},
										}
										
										ATMEFct.outputCore(nil, "Initial Start engine pour " .. _unitName, 2)
										ATMEFct.C_Event_add(event)
									end
								end

								if _item.disembark == false then
									if _item.isSpawned == true then
										for _moduleName, _callback in pairs(_item.spawnCallback) do
											ATMEDatas.currentModuleName = _moduleName
												
											ATMEFct.outputCore(nil,"Appel spawnCallback cloning context du module " .. _moduleName .. " pour le groupe " .. _name, 2)
											xpcall(function() _callback(_item.cloningContext, _item.group) end, function() ATMEFct.errUserHandler('spawnCallback cloning context pour le groupe ' .. _name) end)
										end
										
										for _, _nameModule in ipairs(ATMEDatas.modulesNames) do
											local localModuleMembers = ATMEDatas.modulesLocalMembers[_nameModule]
											ATMEDatas.currentModuleName = _nameModule
											
											if localModuleMembers.handlers.onSpawnGroupHandler ~= nil then
												ATMEFct.outputCore(nil,"Appel de onSpawnGroupHandler du module " .. localModuleMembers.name, 2)
												xpcall(function() localModuleMembers.handlers.onSpawnGroupHandler(_item.group) end, function() ATMEFct.errUserHandler('onSpawnGroupHandler') end)
											end
										end
									end
								else 	-- Groupe débarqué, on prépare l'event de fin de débarquement
									ATMEFct.fireEvent(_item.event)
								end

								ATMEDatas.newPreparedGroups[_name] = nil
								
								delay = delay + 0.1
							end
						else
							ATMEFct.outputCore(nil, "Groupe " .. _name .. " en activation retardée -> Attente activation", 2)
						end
					else
						ATMEDatas.newPreparedGroups[_name] = nil
						ATMEFct.outputCore(nil, "newPreparedGroups -> Groupe déjà détruit " .. _name, 2)
					end
				end
				
				for _name, _item in pairs(ATMEDatas.newPreparedCargos) do
					for _, _nameModule in ipairs(ATMEDatas.modulesNames) do
						local localModuleMembers = ATMEDatas.modulesLocalMembers[_nameModule]
						ATMEDatas.currentModuleName = _nameModule
						
						if localModuleMembers.handlers.onCreateCargoHandler ~= nil then
							ATMEFct.outputCore(nil,"Appel onCreateCargoHandler du module " .. localModuleMembers.name .. " pour le cargo " .. _name, 2)
							xpcall(function() localModuleMembers.handlers.onCreateCargoHandler(ATMEDatas.cargos[_name]) end, function() ATMEFct.errUserHandler('onCreateCargoHandler') end)
						end
					end
					
					if _item.isSpawned == true then
						for _moduleName, _callback in pairs(_item.spawnCallback) do
							ATMEDatas.currentModuleName = _moduleName
								
							ATMEFct.outputCore(nil,"Appel spawnCallback cargo cloning context du module " .. _moduleName .. " pour le cargo " .. _name, 2)
							xpcall(function() _callback(_item.cargoCloningContext, ATMEDatas.cargos[_name]) end, function() ATMEFct.errUserHandler('spawnCallback cargo context pour le cargo ' .. _name) end)
						end
						
						for _, _nameModule in ipairs(ATMEDatas.modulesNames) do
							local localModuleMembers = ATMEDatas.modulesLocalMembers[_nameModule]
							ATMEDatas.currentModuleName = _nameModule
							
							if localModuleMembers.handlers.onSpawnCargoHandler ~= nil then
								ATMEFct.outputCore(nil,"Appel de onSpawnCargoHandler du module " .. localModuleMembers.name, 2)
								xpcall(function() localModuleMembers.handlers.onSpawnCargoHandler(ATMEDatas.cargos[_name]) end, function() ATMEFct.errUserHandler('onSpawnCargoHandler') end)
							end
						end
					end
	
					ATMEDatas.newPreparedCargos[_name] = nil
				end
				
				for _name, _item in pairs(ATMEDatas.newPreparedStatics) do
					for _, _nameModule in ipairs(ATMEDatas.modulesNames) do
						local localModuleMembers = ATMEDatas.modulesLocalMembers[_nameModule]
						ATMEDatas.currentModuleName = _nameModule
						
						if localModuleMembers.handlers.onCreateStaticHandler ~= nil then
							xpcall(function() localModuleMembers.handlers.onCreateStaticHandler(ATMEDatas.statics[_name]) end, function() ATMEFct.errUserHandler('onCreateStaticHandler') end)
						end
					end
					
					if _item.isSpawned == true then
						for _moduleName, _callback in pairs(_item.spawnCallback) do
							ATMEDatas.currentModuleName = _moduleName
								
							ATMEFct.outputCore(nil,"Appel spawnCallback static cloning context du module " .. _moduleName .. " pour le static " .. _name, 2)
							xpcall(function() _callback(_item.staticCloningContext, ATMEDatas.statics[_name]) end, function() ATMEFct.errUserHandler('spawnCallback static cloning context pour le static ' .. _name) end)
						end
						
						for _, _nameModule in ipairs(ATMEDatas.modulesNames) do
							local localModuleMembers = ATMEDatas.modulesLocalMembers[_nameModule]
							ATMEDatas.currentModuleName = _nameModule
							
							if localModuleMembers.handlers.onSpawnStaticHandler ~= nil then
								ATMEFct.outputCore(nil,"Appel de onSpawnStaticHandler du module " .. localModuleMembers.name, 2)
								xpcall(function() localModuleMembers.handlers.onSpawnStaticHandler(ATMEDatas.statics[_name]) end, function() ATMEFct.errUserHandler('onSpawnStaticHandler') end)
							end
						end
					end
					
					ATMEDatas.newPreparedStatics[_name] = nil
				end

				-- Gestion disable des groupes et unités
				ATMEFct.outputCore(nil, "Gestion disable des groupes et unités", 2)
				for _name, _unit in pairs(ATMEDatas.unitsToDisable) do
					ATMEDatas.flagDeleteUnit[_name] = "DISABLED"
					ATMEFct.C_Unit_disable(_unit, false)
					ATMEDatas.unitsToDisable[_name] = nil
				end
				
				for _name, _group in pairs(ATMEDatas.groupsToDisable) do
					ATMEFct.C_Group_disable(_group, ATMEDatas.reservedAsPlayerGroups[_name] > 0)
					ATMEDatas.groupsToDisable[_name] = nil
				end
				
				for _name, _object in pairs(ATMEDatas.staticsToDisable) do
					ATMEFct.C_Static_disable(_object)
					ATMEDatas.staticsToDisable[_name] = nil
				end
				
				for _name, _object in pairs(ATMEDatas.cargosToDisable) do
					ATMEFct.C_Cargo_disable(_object)
					ATMEDatas.cargosToDisable[_name] = nil
				end
				
				-- Vérification de l'existance des unités & groupes liés aux events et
				-- récupération de l'unité/groupe à partir de son nom sinon suppression
				for _id, _event in events[activeEvents]:pairs() do
					ATMEFct.C_Event_update(_event, true)
				end
				
				-- On fige les événements qui vont être générés et émis aux handlers.
				-- On bascule activeEvents sur le nouveau pour prendre en compte au prochain coup les événements
				-- éventuellement générés par ces mêmes handlers (OnTimer)
				local fixedEvents = activeEvents
				activeEvents = 1 - activeEvents
				
				-- Gestion des handlers des events
				ATMEFct.outputCore(nil, "Gestion des événements", 2)
				for _id, _event in events[fixedEvents]:pairs() do
					ATMEFct.C_Event_fire(_event)
				end
					
				-- Déclenchement des handlers
				ATMEFct.outputCore(nil, "Gestion des Handlers par module", 2)
				for _, _nameModule in ipairs(ATMEDatas.modulesNames) do
					local localModuleMembers = ATMEDatas.modulesLocalMembers[_nameModule]
					ATMEDatas.currentModuleName = _nameModule
					
					if localModuleMembers.handlers.onTimerHandler ~= nil then
						ATMEFct.outputCore(nil,"Appel onTimerHandler du module " .. localModuleMembers.name , 2)
						xpcall(localModuleMembers.handlers.onTimerHandler, function() ATMEFct.errUserHandler('onTimerHandler') end)
					end
				end
				
				ATMEFct.C_Event_removeAll(fixedEvents)
			end
			
			ATMEFct.outputCore(nil,"End main timer", 2)
			timer.scheduleFunction(_timerMainFunction, {}, timer.getTime() + 1)
		end
	end
	
	local _randomTimer
	_randomTimer = function()
		return timer.getTime() + 0.2 + math.random() / 3
	end
	
	
	----------------------------------------------------------------------------------------------------------
	-- Fonction gestion événements DCS ... essentielle ... appelée si l'événement se produit
	-- En l'état ne gère pas tous les événements mais uniquement ceux nécessaires
	----------------------------------------------------------------------------------------------------------	

	ATMEFct.fireEvent = function(_event)
		local newEvent = ATMEFct.C_Event_create(_event)
		
		ATMEFct.C_Event_update(newEvent, false)
		ATMEFct.C_Event_fire(newEvent)
	end
	
	ATMEFct.hitTargetFct = {
		[Object.Category.VOID] = function(_event, _weaponName)
			local event = {
				type = "WEAPON",
				datas = {
					action = "HIT_TARGET",
					unitName = _event.initiator:getName(),
					targetName =  _event.target:getName(),
					targetClassName = "C_Airbase",
					weaponName = _weaponName,
				},
			}		
			
			ATMEFct.fireEvent(event)
			ATMEFct.outputCore(nil, "Event WEAPON HIT_TARGET généré pour objet " .. _event.target:getName() .. " par " .. _event.initiator:getName(), 2)
		end,
		[Object.Category.UNIT] = function(_event, _weaponName)
			if ATMEDatas.units[_event.target:getName()] ~= nil then
				local event = {
					type = "WEAPON",
					datas = {
						action = "HIT_TARGET",
						unitName = _event.initiator:getName(),
						targetName =  _event.target:getName(),
						targetClassName = ATMEDatas.units[_event.target:getName()]:getClassName(),
						weaponName = _weaponName,
					},
				}		
				
				ATMEFct.fireEvent(event)
				ATMEFct.outputCore(nil, "Event WEAPON HIT_TARGET généré pour unit " .. _event.target:getName() .. " par " .. _event.initiator:getName(), 2)
			end
		end,
		[Object.Category.WEAPON] = function(_event, _weaponName)
			if ATMEDatas.weapons[_event.target:getName()] ~= nil then
				local event = {
					type = "WEAPON",
					datas = {
						action = "HIT_TARGET",
						unitName = _event.initiator:getName(),
						targetName =  _event.target:getName(),
						targetClassName = ATMEDatas.weapons[_event.target:getName()]:getClassName(),
						weaponName = _weaponName,
					},
				}
				
				ATMEFct.fireEvent(event)
				ATMEFct.outputCore(nil, "Event WEAPON HIT_TARGET généré pour arme " .. _event.target:getName() .. " par " .. _event.initiator:getName(), 2)
			end
		end,
		[Object.Category.STATIC] = function(_event,_weaponName)
			if ATMEDatas.statics[_event.target:getName()] ~= nil then
				local event = {
					type = "WEAPON",
					datas = {
						action = "HIT_TARGET",
						unitName = _event.initiator:getName(),
						targetName =  _event.target:getName(),
						targetClassName = ATMEDatas.statics[_event.target:getName()]:getClassName(),
						weaponName = _weaponName,
					},
				}

				ATMEFct.fireEvent(event)
				ATMEFct.outputCore(nil, "Event WEAPON HIT_TARGET généré pour static object " .. _event.target:getName() .. " par " .. _event.initiator:getName(), 2)
			end
		end,
		[Object.Category.CARGO] = function(_event, _weaponName)
			if ATMEDatas.cargos[_event.target:getName()] ~= nil then
				local event = {
					type = "WEAPON",
					datas = {
						action = "HIT_TARGET",
						unitName = _event.initiator:getName(),
						targetName =  _event.target:getName(),
						targetClassName = ATMEDatas.cargos[_event.target:getName()]:getClassName(),
						weaponName = _weaponName,
					},
				}

				ATMEFct.fireEvent(event)
				ATMEFct.outputCore(nil, "Event WEAPON HIT_TARGET généré pour cargo " .. _event.target:getName() .. " par " .. _event.initiator:getName(), 2)
			end
		end,
		[Object.Category.SCENERY] = function(_event, _weaponName)
			if ATMEDatas.sceneryObjects[_event.target:getName()] == nil then
				ATMEFct.C_SceneryObject_create(_event.target)
			end
			
			local event = {
				type = "WEAPON",
				datas = {
					action = "HIT_TARGET",
					unitName = _event.initiator:getName(),
					targetName =  _event.target:getName(),
					targetClassName = ATMEDatas.sceneryObjects[_event.target:getName()]:getClassName(),
					weaponName = _weaponName,
				},
			}

			ATMEFct.fireEvent(event)
			ATMEFct.outputCore(nil, "Event WEAPON HIT_TARGET généré pour scenery object" .. _event.target:getName() .. " par " .. _event.initiator:getName(), 2)
		end,
	}
	
	ATMEFct.shootTargetFct = {
		[Object.Category.UNIT] = function(_event)
			if ATMEDatas.units[_event.target:getName()] ~= nil then
				local event = {
					type = "WEAPON",
					datas = {
						action = "SHOOTING_START",
						unitName = _event.initiator:getName(),
						targetName =  _event.target:getName(),
						targetClassName = ATMEDatas.units[_event.target:getName()]:getClassName(),
					},
				}

				ATMEFct.fireEvent(event)
				ATMEFct.outputCore(nil, "Event WEAPON SHOOTING_START généré pour unit " .. _event.target:getName() .. " par " .. _event.initiator:getName(), 2)
			else
				local event = {
						type = "WEAPON",
						datas = {
							action = "SHOOTING_START",
							unitName = _event.initiator:getName(),
						},
					}
				
				ATMEFct.fireEvent(event)
				ATMEFct.outputCore(nil, "Event WEAPON SHOOTING START généré pour " .. _event.initiator:getName(), 2)
			end
		end,
		[Object.Category.WEAPON] = function(_event)
			if ATMEDatas.weapons[_event.target:getName()] ~= nil then
				local event = {
					type = "WEAPON",
					datas = {
						action = "SHOOTING_START",
						unitName = _event.initiator:getName(),
						targetName =  _event.target:getName(),
						targetClassName = ATMEDatas.weapons[_event.target:getName()]:getClassName(),
					},
				}
				
				ATMEFct.fireEvent(event)
				ATMEFct.outputCore(nil, "Event WEAPON SHOOTING_START généré pour arme " .. _event.target:getName() .. " par " .. _event.initiator:getName(), 2)
			else
				local event = {
						type = "WEAPON",
						datas = {
							action = "SHOOTING_START",
							unitName = _event.initiator:getName(),
						},
					}
				
				ATMEFct.fireEvent(event)
				ATMEFct.outputCore(nil, "Event WEAPON SHOOTING START généré pour " .. _event.initiator:getName(), 2)
			end
		end,
		[Object.Category.STATIC] = function(_event)
			if ATMEDatas.statics[_event.target:getName()] ~= nil then
				local event = {
					type = "WEAPON",
					datas = {
						action = "SHOOTING_START",
						unitName = _event.initiator:getName(),
						targetName =  _event.target:getName(),
						targetClassName = ATMEDatas.statics[_event.target:getName()]:getClassName(),
					},
				}
				
				ATMEFct.fireEvent(event)
				ATMEFct.outputCore(nil, "Event WEAPON SHOOTING_START généré pour static object " .. _event.target:getName() .. " par " .. _event.initiator:getName(), 2)
			else
				local event = {
						type = "WEAPON",
						datas = {
							action = "SHOOTING_START",
							unitName = _event.initiator:getName(),
						},
					}
				
				ATMEFct.fireEvent(event)
				ATMEFct.outputCore(nil, "Event WEAPON SHOOTING START généré pour " .. _event.initiator:getName(), 2)
			end
		end,
		[Object.Category.CARGO] = function(_event)
			if ATMEDatas.cargos[_event.target:getName()] ~= nil then
				local event = {
					type = "WEAPON",
					datas = {
						action = "SHOOTING_START",
						unitName = _event.initiator:getName(),
						targetName =  _event.target:getName(),
						targetClassName = ATMEDatas.cargos[_event.target:getName()]:getClassName(),
					},
				}
				
				ATMEFct.fireEvent(event)
				ATMEFct.outputCore(nil, "Event WEAPON SHOOTING_START généré pour cargo " .. _event.target:getName() .. " par " .. _event.initiator:getName(), 2)
			else
				local event = {
						type = "WEAPON",
						datas = {
							action = "SHOOTING_START",
							unitName = _event.initiator:getName(),
						},
					}
				
				ATMEFct.fireEvent(event)
				ATMEFct.outputCore(nil, "Event WEAPON SHOOTING START généré pour " .. _event.initiator:getName(), 2)
			end
		end,
		[Object.Category.SCENERY] = function(_event)
			if ATMEDatas.sceneryObjects[_event.target:getName()] == nil then
				ATMEFct.C_SceneryObject_create(_event.target)
			end
			
			local event = {
				type = "WEAPON",
				datas = {
					action = "SHOOTING_START",
					unitName = _event.initiator:getName(),
					targetName =  _event.target:getName(),
					targetClassName = ATMEDatas.sceneryObjects[_event.target:getName()]:getClassName(),
				},
			}
			
			ATMEFct.fireEvent(event)
			ATMEFct.outputCore(nil, "Event WEAPON SHOOTING_START généré pour scenery object" .. _event.target:getName() .. " par " .. _event.initiator:getName(), 2)
		end,
	}
	
	---------------------------------------------------------------------------------------------------------------------------------------
	-- S_EVENT_PLAYER_ENTER_UNIT ne fonctionne pour le joueur hôte en SP et MP
	--
	-- S_EVENT_BIRTH et S_EVENT_PLAYER_LEAVE_UNIT - Fonctionnement DCS Mono joueur
	-- 	  Si changement de slot 2 événements S_EVENT_PLAYER_LEAVE_UNIT émis, sinon 1 seul puis un événement S_EVENT_BIRTH
	--	  S_EVENT_PLAYER_LEAVE_UNIT ne dispose plus du pseudo du joueur lorsqu'il quitte une unité AI sol (bug DCS ?)
	--    Joueur seul dans un groupe avec des AI : Initialement invisible, l'unité du joueur disparaît au changement de slot 
	--                                             mais pas les unités AI du groupe.
	--    Joueur passe d'un aéronef compétence Client/Player à aéronef AI : l'unité du joueur disparaît.
	--    Joueur passe d'un aéronef AI à un autre autre aéronef : l'unité du joueur ne disparaît pas
	--    Joueur passe d'un aéronef à une unité sol AI : l'unité du joueur ne disparait pas
	--	  Joueur passe d'une unité sol à un aéronef AI ou compétence Client/Player : l'unité sol ne disparait pas
	--	  Joueur passe d'une unité sol à une autre : l'unité sol ne disparaît pas
	--	  En mono ATME a une gestion basée sur l'échange d'unités avec une destruction en fonction du contexte de l'unité joueur.
	--
	-- S_EVENT_BIRTH et S_EVENT_PLAYER_LEAVE_UNIT - Fonctionnement DCS Multi joueur :
	--    Pas de possibilité de passer d'un aéronef joueur à un aéronef AI, ni d'un aéronef joueur à une unité AI sol
	--	  Possibilité de contrôle les unités AI sol avec une profil particulier de slot (GAME MASTER etc...). 
	--    Si changement de slot 1 évenement S_EVENT_PLAYER_LEAVE_UNIT émis puis un événement S_EVENT_BIRTH uniquement pour le joueur.
	--	  Si plusieurs joueurs et unités AI, comportement DCS compliqué, les unités apparaissent et disparaissent sur un joueur particulier
	--                        ATME interdit d'avoir plus d'un aéronef compétence Client/Player si présence unités AI dans un même groupe et
	--						  détruit toutes les unités du groupe au changement de slot (sur événement S_EVENT_PLAYER_LEAVE_UNIT)
	--						  Si le groupe n'a que des aéronefs compétence Client/Player, seule l'unité joueur concernée sera détruite.
	--
	-- Multi, Aéronef à 2 joueurs ou plus (Test F14), mais semble non définitif (A SUIVRE) : Pour le F14, le pilote doit entrer en premier.  
	-- Le copilote génère un event S_EVENT_BIRTH mais unitName vaut alors "" et le nom du joueur est le pilote (voir cas CAS ENTREE ci après) ! 
	-- Ensuite au décollage et atterrissage deux events S_EVENT_TAKEOFF and S_EVENT_LAND générés (1 par joueur), 
	-- ce n'est pas le cas pour l'arrêt moteur !!! Enfin quand le copilote quitte, pas d'event S_EVENT_PLAYER_LEAVE_UNIT et nouveau event  
	-- S_EVENT_BIRTH spécifique avec juste le pilote restant (voir CAS SORTIE ci après). 
	---------------------------------------------------------------------------------------------------------------------------------------
	
	local eventFunctions = {
		[world.event.S_EVENT_BIRTH] = {	-- Si un objet est spawné en cours de mission, on passe ici
			[Object.Category.UNIT] = function(_event)
				local unitName = _event.initiator:getName()

				if _event.initiator:getPlayerName() ~= nil then
					-- Bug DCS pas d'event S_EVENT_PLAYER_ENTER_UNIT en multi pour un joueur client (non hote)
					
					if ATMEDatas.multiplayers == nil then
						if _event.initiator:getPlayerName() == "New callsign" then
							ATMEDatas.multiplayers = false
						else
							ATMEDatas.multiplayers = true
						end
					end
					
					if ATMEDatas.playersControl[_event.initiator:getPlayerName()] ~= nil then
						if ATMEDatas.multiplayers ==  false then
							local localMembers = ATMEDatas.localMembers[ATMEDatas.units[ATMEDatas.playersControl[_event.initiator:getPlayerName()]]:getIndex()]
							
							if localMembers.spawnDatas.skill == "Client" or localMembers.spawnDatas.skill == "Player" then
								ATMEFct.outputCore(nil, "S_EVENT_BIRTH Unité joueur initiale " .. ATMEDatas.playersControl[_event.initiator:getPlayerName()] .. " -> supprimée", 2)
								ATMEDatas.flagDeleteUnit[ATMEDatas.playersControl[_event.initiator:getPlayerName()]] = "CHANGE_SLOT"
								ATMEFct.C_Unit_delete(ATMEDatas.units[ATMEDatas.playersControl[_event.initiator:getPlayerName()]], true)
							end
						end
						
						ATMEDatas.unitsOutOfControl[ATMEDatas.playersControl[_event.initiator:getPlayerName()]] = nil
						ATMEDatas.playersControl[_event.initiator:getPlayerName()] = nil
					end
					
					if ATMEDatas.units[unitName] ~= nil then
						-- CAS SORTIE, 2 joueurs dans une même unité (tests sur F14)
						-- En multi et si le copilote quitte l'aeronef, alors on passe ici
						-- On ne fait rien si multijoueur dans ce cas car 2ème joueur non géré.
						
						if ATMEDatas.multiplayers == false then
							ATMEFct.C_Unit_playerEnters(ATMEDatas.units[unitName])

							for _, _nameModule in ipairs(ATMEDatas.modulesNames) do
								local localModuleMembers = ATMEDatas.modulesLocalMembers[_nameModule]	
								ATMEDatas.currentModuleName = _nameModule

								if localModuleMembers.handlers.onCreateUnitHandler ~= nil then
									ATMEFct.outputCore(nil,"Appel onCreateUnitHandler du module " .. localModuleMembers.name, 2)
									xpcall(function() localModuleMembers.handlers.onCreateUnitHandler(ATMEDatas.units[unitName]) end, function() ATMEFct.errUserHandler('onCreateUnitHandler') end)
								end		
							end
							
							local localMembers = ATMEDatas.localMembers[ATMEDatas.units[unitName]:getIndex()]
							local localMembersGroup = ATMEDatas.localMembers[localMembers.group:getIndex()]
							
							if ATMEConst.canFly[localMembersGroup.category] == true and localMembers.engineOn == true then
								-- Emission de l'événement démarrage moteur
								local event = {
									type = "ENGINE",
									datas = {
										action = "START",
										unitName = unitName,
									},
								}
								
								ATMEFct.outputCore(nil, "Initial Start engine pour " .. unitName, 2)
								ATMEFct.C_Event_add(event)
							end
						end
						
						ATMEFct.outputCore(nil,"S_EVENT_BIRTH Unité " .. unitName .. " contrôlée par joueur " .. _event.initiator:getPlayerName(), 2)
					else
						if ATMEDatas.groups[_event.initiator:getGroup():getName()] ~= nil then
							ATMEFct.outputCore(nil,"S_EVENT_BIRTH '" .. unitName .. "' -> unité de groupe existant (" .. _event.initiator:getGroup():getName() .. ")", 2)
						end
						
						if ATMEDatas.allAIUnitsInMission[unitName] ~= nil then
							-- CAS ENTREE, 2 joueurs dans une même unité (tests sur F14)
							-- unitName vaut "", dans ce cas on ne fait rien. Le joueur est le pilote
							-- Du coup, le 2ème joueur n'est plus géré -> 2 places mais 1 unité et un pilote uniquement pour DCS et donc ATME.
							
							ATMEFct.C_Unit_create(_event.initiator, ATMEDatas.allAIUnitsInMission[unitName].datas, true, ATMEDatas.groups[_event.initiator:getGroup():getName()] ~= nil)
						end
						
						if _event.initiator:getPlayerName() ~= nil then
							ATMEFct.outputCore(nil,"S_EVENT_BIRTH " .. unitName .. " en création  pour joueur " .. _event.initiator:getPlayerName(), 2)
						else
							ATMEFct.outputCore(nil,"S_EVENT_BIRTH " .. unitName .. " en création  pour joueur", 2)
						end
					end
				elseif ATMEDatas.allAIUnitsInMission[unitName] ~= nil then
					if ATMEDatas.units[unitName] == nil then
						-- Unités liées à un groupe avec joueurs ou à création différée start_time > 0.
			
						ATMEFct.C_Unit_create(_event.initiator, ATMEDatas.allAIUnitsInMission[unitName].datas, true, false)
						ATMEFct.outputCore(nil,"Unit " .. unitName .. " en création", 2)
					else
						local localMembers = ATMEDatas.localMembers[ATMEDatas.units[unitName]:getIndex()]
						local localMembersGroup = ATMEDatas.localMembers[localMembers.group:getIndex()]

						if ATMEDatas.unitsOutOfControl[unitName] ~= nil then
							ATMEFct.outputCore(nil,"S_EVENT_BIRTH Unité " .. unitName .. " fin de controle joueur ", 2)
							
							for _, _nameModule in ipairs(ATMEDatas.modulesNames) do
								local localModuleMembers = ATMEDatas.modulesLocalMembers[_nameModule]	
								ATMEDatas.currentModuleName = _nameModule

								if localModuleMembers.handlers.onCreateUnitHandler ~= nil then
									ATMEFct.outputCore(nil,"Appel onCreateUnitHandler du module " .. localModuleMembers.name, 2)
									xpcall(function() localModuleMembers.handlers.onCreateUnitHandler(ATMEDatas.units[unitName]) end, function() ATMEFct.errUserHandler('onCreateUnitHandler') end)
								end		
							end
							
							if ATMEConst.canFly[localMembersGroup.category] == true and localMembers.engineOn == true then
								-- Emission de l'événement démarrage moteur
								local event = {
									type = "ENGINE",
									datas = {
										action = "START",
										unitName = unitName,
									},
								}
								
								ATMEFct.outputCore(nil, "Initial Start engine pour " .. unitName, 2)
								ATMEFct.C_Event_add(event)
							end
							
							ATMEDatas.playersControl[ATMEDatas.unitsOutOfControl[unitName]] = nil
							ATMEDatas.unitsOutOfControl[unitName] = nil
						elseif localMembersGroup.isEnabled == false then
							ATMEFct.outputCore(nil,"LATE ACTIVATION -> Activation de l'unité " .. localMembers.name, 2)
							
							localMembersGroup.nbUnitsToEnable = localMembersGroup.nbUnitsToEnable - 1
							
							localMembers.objectDCS = _event.initiator
							localMembers.objectDCSId = _event.initiator:getID()
							
							if localMembersGroup.nbUnitsToEnable == 0 then
								localMembersGroup.isEnabled = true
								local group = _event.initiator:getGroup()	
								localMembersGroup.objectDCS = group
								localMembersGroup.objectDCSId = group:getID()
								
								ATMEDatas.newCreatedGroups[localMembersGroup.name] = {
									timerOffset = 1,
									isSpawned = false,
									disembark = false,
									group = localMembers.group,
								}
								
								ATMEFct.outputCore(nil,"LATE ACTIVATION -> Activation du groupe " .. localMembersGroup.name, 2)
							end
						end
					end
				elseif ATMEDatas.units[unitName] == nil then
					-- ATTENTION deux cas :
					-- 	1*	L'événement S_EVENT_BIRTH survient pendant l'appel de la fonction addgroup
					--		pour les unités au sol et en mer ou lors de l'activation retardée d'une unité.
					-- 		Les groupes à activation retardées n'ont pas à être traités car initialisés
					--		lors du démarrage d'ATME. Parfois le groupe n'est pas encore initialisé 
					-- 		pendant l'appel de la fonction addgroup. Aussi, dans ce cas, le traitement
					-- 		se fait juste après la fonction addgroup et dans ce cas, l'unité créée est
					--		mémorisée dans ATMEDatas.unitsReadyToCreate.
					--	2*	L'événement survient après le spawn pour les groupes d'aéronefs. Dans ce cas,
					--		l'unité est mémorisée dans ATMEDatas.groupsReadyToBirth après la fonction
					--		addgroup. Le traitement se fait alors ici, lors de la création physique desactivateSpot
					--		unités du groupe.
				
					local found = false
					
					for _name, _group in pairs(ATMEDatas.groupsReadyToBirth) do
						-- L'unité à créer est un bateau avec FARP ou porte-avion

						if _event.initiator:hasAttribute("AircraftCarrier") == true then
							ATMEFct.C_Airbase_create_basic("SHIP", false, _event.initiator)
						elseif _event.initiator:hasAttribute("HelicopterCarrier") == true then
							ATMEFct.C_Airbase_create_basic("SHIP", true, _event.initiator)
						end
							
						-- Traitement du cas 2
						
						if _group.item.units[unitName] ~= nil then
							local datas = {
								uncontrolled = _group.item.uncontrolled,
								visible = _group.item.visible,
								spawnDatas = _group.item.spawnDatas,
							}
							
							ATMEFct.C_Unit_create(_event.initiator, datas, true, false)
							ATMEDatas.unitsReadyToBirth[unitName] = nil
													
							found = true
							
							_group.item.units[unitName] = nil
							_group.item.nbUnits = _group.item.nbUnits - 1
							
							if _group.item.nbUnits == 0 then
								ATMEDatas.newCreatedGroups[_name].isSpawned = true
								ATMEDatas.newCreatedGroups[_name].spawnCallback = _group.spawnCallback
								ATMEDatas.newCreatedGroups[_name].cloningContext = _group.cloningContext
								ATMEDatas.groupsReadyToBirth[_name] = nil
							end
						end
					end

					if found == false then
						-- Si found vaut false, alors on est dans le cas 1

						table.insert(ATMEDatas.unitsReadyToCreate, _event.initiator)
					end
				end
			end,
			[Object.Category.CARGO] = function(_event)
				local unitName = _event.initiator:getName()
	
				if StaticObject.getByName(unitName) ~= nil then -- Cargo est un StaticObject pour DCS
					if ATMEDatas.cargos[unitName] ~= nil then
						ATMEFct.C_Cargo_delete(ATMEDatas.cargo[unitName])
					end
						
					ATMEFct.C_Cargo_create(_event.initiator)
					ATMEFct.outputCore(nil,"Cargo " .. unitName .. " créé", 2)
				end
			end,
			[Object.Category.STATIC] = function(_event)
				local unitName = _event.initiator:getName()
	
				if StaticObject.getByName(unitName) ~= nil then
					if unitName ~= "static" then -- Bug DCS certains objets spawnés ont toujours leur nom à 'static'.
						if ATMEDatas.statics[unitName] ~= nil then
							ATMEFct.C_Static_delete(ATMEDatas.statics[unitName])
						end
						
						ATMEFct.C_Static_create(_event.initiator)
						ATMEFct.outputCore(nil,"StaticObject " .. unitName .. " créé", 2)
					else
						ATMEFct.outputCore(nil,"StaticObject spawné avec un nom valant 'static' - Ignoré par ATME", 2)
					end
				end
			end,
		},
		[world.event.S_EVENT_PLAYER_ENTER_UNIT] = {
			[Object.Category.UNIT] = function(_event)
				-- 	BUG DCS -> Aucun traitement car non pris en compte en multi pour les clients non hôtes
				ATMEFct.outputCore(nil,"S_EVENT_PLAYER_ENTER_UNIT Changement de slot - Player " .. _event.initiator:getPlayerName() .. " entre dans l'unité " .. _event.initiator:getName() .. " - Pas de traitement", 2)
				
				if ATMEDatas.multiplayers == nil then
					if _event.initiator:getPlayerName() == "New callsign" then
						ATMEDatas.multiplayers = false
					else
						ATMEDatas.multiplayers = true
					end
				end
			end,
		},
		[world.event.S_EVENT_PLAYER_LEAVE_UNIT] = {
			[Object.Category.UNIT] = function(_event)
				local unitName = _event.initiator:getName()
				local category = _event.initiator:getGroup():getCategory()

				if _event.initiator:getPlayerName() ~= nil then -- unitName vaut "" si départ sur binome (A TESTER)
					if category == Group.Category.AIRPLANE or category == Group.Category.HELICOPTER then
						if ATMEDatas.playersControl[_event.initiator:getPlayerName()] == nil then
							-- Si un joueur quitte l'unité on mémorise pour ne pas recréer tout le groupe à son retour
							-- Valable en multi et en mono
							
							local localMembersUnit = ATMEDatas.localMembers[ATMEDatas.units[unitName]:getIndex()]
							local localMembersGroup = ATMEDatas.localMembers[localMembersUnit.group:getIndex()]
							
							-- La destruction ne se fait que lors du BIRTH donc on mémorise les dernières positions ici.
							localMembersUnit.lastPosition = localMembersUnit.objectDCS:getPosition()
							localMembersUnit.lastVelocity = localMembersUnit.objectDCS:getVelocity()
							
							ATMEDatas.playersControl[_event.initiator:getPlayerName()] = unitName
							ATMEDatas.unitsOutOfControl[unitName] = _event.initiator:getPlayerName()
							
							if ATMEDatas.multiplayers == false then
								ATMEFct.C_Unit_playerLeaves(ATMEDatas.units[unitName], localMembersGroup)
								ATMEFct.outputCore(nil,"S_EVENT_PLAYER_LEAVE_UNIT (SP) Player " .. _event.initiator:getPlayerName() .. " quitte l'unité " .. unitName, 2)
							else
								if ATMEDatas.reservedAsPlayerGroups[localMembersGroup.name] == 1 then  
									-- Groupe avec un seul joueur potentiel (eventuellement avec des AI) -> Destruction du groupe
									if ATMEDatas.groups[_event.initiator:getGroup():getName()] ~= nil then
										ATMEFct.C_Group_disable(ATMEDatas.groups[_event.initiator:getGroup():getName()], true, "CHANGE_SLOT")
									else -- Non encore créé par ATME
										local units = _event.initiator:getGroup():getUnits()
										
										for _i, _unit in ipairs(units) do
											_unit:destroy()
										end
									end
									
									ATMEFct.outputCore(nil,"S_EVENT_PLAYER_LEAVE_UNIT (MP) Changement de slot - Player " .. _event.initiator:getPlayerName() .. " quitte l'unité " .. unitName .. " - Destruction du groupe", 2)
								else
									-- Groupe avec plusieurs joueurs potentiels
									ATMEDatas.flagDeleteUnit[unitName] = "CHANGE_SLOT"
									ATMEFct.C_Unit_playerLeaves(ATMEDatas.units[unitName], localMembersGroup)
									ATMEFct.C_Unit_delete(ATMEDatas.units[unitName], true)
									ATMEFct.outputCore(nil,"S_EVENT_PLAYER_LEAVE_UNIT (MP) Changement de slot - Player " .. _event.initiator:getPlayerName() .. " quitte l'unité " .. unitName, 2)
								end
							end
						end
					else
						ATMEFct.outputCore(nil,"S_EVENT_PLAYER_LEAVE_UNIT Player quitte l'unité " .. unitName, 2)
					end
				else
					ATMEFct.outputCore(nil,"S_EVENT_PLAYER_LEAVE_UNIT Aucun player identifié pour l'unité " .. unitName, 2)
				end
			end,
		},
		[world.event.S_EVENT_DEAD] = {
			[Object.Category.UNIT] = function(_event)
				local unitName = _event.initiator:getName()
				
				if ATMEDatas.units[unitName] ~= nil then
					local localMembers = ATMEDatas.localMembers[ATMEDatas.units[unitName]:getIndex()]
					local localMembersGroup = ATMEDatas.localMembers[localMembers.group:getIndex()]
					
					ATMEFct.C_Unit_playerLeaves(ATMEDatas.units[unitName], localMembersGroup)
					
					ATMEDatas.flagDeleteUnit[unitName] = "DESTROYED"
					ATMEFct.C_Unit_delete(ATMEDatas.units[unitName], ATMEDatas.reservedAsPlayerGroups[localMembersGroup.name] > 0)
					ATMEFct.outputCore(nil,"Unit " .. unitName .. " détruite", 2)
				end
			end,
			[Object.Category.CARGO] = function(_event)
				local unitName = _event.initiator:getName()
	
				if ATMEDatas.cargos[unitName] ~= nil then
					ATMEFct.C_Cargo_delete(ATMEDatas.cargos[unitName])
					ATMEFct.outputCore(nil,"Cargo " .. unitName .. " détruit", 2)
				end
			end,
			[Object.Category.STATIC] = function(_event)
				local unitName = _event.initiator:getName()
	
				if ATMEDatas.statics[unitName] ~= nil then
					ATMEFct.C_Static_delete(ATMEDatas.statics[unitName])
					ATMEFct.outputCore(nil,"StaticUnit " .. unitName .. " détruite", 2)
				else
					ATMEFct.outputCore(nil,"Unité statique inconnue " .. unitName .. " détruite", 2)
				end
			end,
			[Object.Category.SCENERY] = function(_event)
				if ATMEDatas.sceneryObjects[_name] ~= nil then
					ATMEFct.C_SceneryObject_delete(_event.initiator:getName())
					ATMEFct.outputCore(nil,"Scenery object " .. _event.initiator:getName() .. " détruit", 2)
				end
			end,
		},
		[world.event.S_EVENT_CRASH] = {
			[Object.Category.UNIT] = function(_event)
				local unitName = _event.initiator:getName()
				
				if ATMEDatas.units[unitName] ~= nil then
					local localMembers = ATMEDatas.localMembers[ATMEDatas.units[unitName]:getIndex()]
					local localMembersGroup = ATMEDatas.localMembers[localMembers.group:getIndex()]
					
					ATMEFct.C_Unit_playerLeaves(ATMEDatas.units[unitName], localMembersGroup)
					
					ATMEDatas.flagDeleteUnit[unitName] = "DESTROYED"
					local groupName = _event.initiator:getGroup():getName()
					ATMEFct.C_Unit_delete(ATMEDatas.units[unitName], ATMEDatas.reservedAsPlayerGroups[groupName] > 0)
					ATMEFct.outputCore(nil,"Unit " .. unitName .. " crash", 2)
				end
			end,
		},
		[world.event.S_EVENT_PILOT_DEAD] = {
			[Object.Category.UNIT] = function(_event)
				local unitName = _event.initiator:getName()
				
				if ATMEDatas.units[unitName] ~= nil then
					local localMembers = ATMEDatas.localMembers[ATMEDatas.units[unitName]:getIndex()]
					local localMembersGroup = ATMEDatas.localMembers[localMembers.group:getIndex()]
					
					ATMEFct.C_Unit_playerLeaves(ATMEDatas.units[unitName], localMembersGroup)
					
					localMembers.pilotDead = true
					ATMEDatas.flagDeleteUnit[unitName] = "DEAD"
					local groupName = _event.initiator:getGroup():getName()
					ATMEFct.C_Unit_delete(ATMEDatas.units[unitName], ATMEDatas.reservedAsPlayerGroups[groupName] > 0)
					ATMEFct.outputCore(nil,"Unit " .. unitName .. " pilote mort", 2)
				end
			end,
		},
		[world.event.S_EVENT_EJECTION] = {
			[Object.Category.UNIT] = function(_event)
				local unitName = _event.initiator:getName()
				if ATMEDatas.units[unitName] ~= nil then
					local localMembers = ATMEDatas.localMembers[ATMEDatas.units[unitName]:getIndex()]
					localMembers.pilotEjected = true
					ATMEFct.outputCore(nil,"Unit " .. unitName .. " pilote éjecté", 2)
				end
			end,
		},
		[world.event.S_EVENT_HUMAN_FAILURE] = {
			[Object.Category.UNIT] = function(_event)
				ATMEFct.outputCore(nil,"Human failure " .. _event.initiator:getName(), 2)
				
				-- Non implémenté
				ATMEFct.outputCore(nil, "Event S_EVENT_HUMAN_FAILURE généré pour " .. _event.initiator:getName() .. " non implémenté", 2)
			end,
		},
		[world.event.S_EVENT_TAKEOFF] = {
			[Object.Category.UNIT] = function(_event)
				local unitName = _event.initiator:getName()
				if ATMEDatas.units[unitName] ~= nil then
					local localMembers = ATMEDatas.localMembers[ATMEDatas.units[unitName]:getIndex()]
					
					if localMembers.hasTakenOff == false then
						localMembers.hasTakenOff = true
						
						local event = {
							type = "TAKE_OFF",
							datas = {
								unitName = _event.initiator:getName(),
							},
						}
						
						ATMEFct.fireEvent(event)
						ATMEFct.outputCore(nil, "Event S_EVENT_TAKEOFF généré pour " .. _event.initiator:getName(), 2)
					end
				end
			end,
		},
		[world.event.S_EVENT_LAND] = {
			[Object.Category.UNIT] = function(_event)
				local unitName = _event.initiator:getName()
				if ATMEDatas.units[unitName] ~= nil then
					local localMembers = ATMEDatas.localMembers[ATMEDatas.units[unitName]:getIndex()]
					
					if localMembers.hasTakenOff == true then
						localMembers.hasTakenOff = false
						
						local event = {
							type = "LAND",
							datas = {
								unitName = _event.initiator:getName(),
							},
						}
						
						ATMEFct.fireEvent(event)
						ATMEFct.outputCore(nil, "Event S_EVENT_LAND généré pour " .. _event.initiator:getName(), 2)
					end
				end
			end,
		},
		[world.event.S_EVENT_ENGINE_STARTUP] = {
			[Object.Category.UNIT] = function(_event)
				-- Bug DCS pour les avions créés en début de mission
				-- Cet événement est généré manuellement lors de la création de l'aéronef.
				-- A ce moment là ATMEDatas.units[unitName] est nil
				-- Par contre ok pour démarrage moteur postérieur.

				local unitName = _event.initiator:getName()
				
				if ATMEDatas.units[unitName] ~= nil then
					local event = {
						type = "ENGINE",
						datas = {
							action = "START",
							unitName = unitName,
						},
					}
					
					ATMEFct.outputCore(nil, "Event S_EVENT_ENGINE_STARTUP généré pour " .. unitName, 2)
					ATMEFct.fireEvent(event)
				end
			end,
		},
		[world.event.S_EVENT_ENGINE_SHUTDOWN] = {
			[Object.Category.UNIT] = function(_event)
				local unitName = _event.initiator:getName()
				if ATMEDatas.units[unitName] ~= nil then
					local localMembers = ATMEDatas.localMembers[ATMEDatas.units[unitName]:getIndex()]
					localMembers.engineOn = false

					local event = {
						type = "ENGINE",
						datas = {
							action = "SHUTDOWN",
							unitName = unitName,
						},
					}

					ATMEFct.fireEvent(event)
					ATMEFct.outputCore(nil, "Event S_EVENT_ENGINE_SHUTDOWN généré pour " .. unitName, 2)
				end
			end,
		},
		[world.event.S_EVENT_BASE_CAPTURED] = {
			[Object.Category.UNIT] = function(_event)
				local unitName = _event.initiator:getName()
				
				if ATMEDatas.units[unitName] ~= nil then
					for _, _nameModule in ipairs(ATMEDatas.modulesNames) do
						local localModuleMembers = ATMEDatas.modulesLocalMembers[_nameModule]
						ATMEDatas.currentModuleName = _nameModule
						if localModuleMembers.handlers.onBaseCapturedHandler ~= nil then
							local datas = {
								airbase = _event.place,
								subPlace = _event.subPlace,
							}
							
							ATMEFct.outputCore(nil,"Appel onBaseCapturedUnitHandler du module " .. localModuleMembers.name, 2)
							
							xpcall(function() localModuleMembers.handlers.onBaseCapturedUnitHandler(ATMEDatas.units[unitName], datas) end, function() ATMEFct.errUserHandler('onBaseCapturedUnitHandler') end)
						end
					end
					
					ATMEFct.outputCore(nil,"S_EVENT_BASE_CAPTURED Unit " .. unitName .. " capture de la base : " .. _event.place:getName(), 2)
				end
			end,
		},
		[world.event.S_EVENT_REFUELING] = {
			[Object.Category.UNIT] = function(_event)
				local event = {
					type = "REFUELING",
					datas = {
						action = "START",
						unitName = _event.initiator:getName(),
					},
				}
				
				ATMEFct.fireEvent(event)
				ATMEFct.outputCore(nil, "Event S_EVENT_REFUELING généré pour " .. _event.initiator:getName(), 2)
			end,
		},
		[world.event.S_EVENT_REFUELING_STOP] = {
			[Object.Category.UNIT] = function(_event)
				local event = {
					type = "REFUELING",
					datas = {
						action = "STOP",
						unitName = _event.initiator:getName(),
					},
				}
				
				ATMEFct.fireEvent(event)
				ATMEFct.outputCore(nil, "Event S_EVENT_REFUELING_STOP généré pour " .. _event.initiator:getName(), 2)
			end,
		},
		[world.event.S_EVENT_SHOT] = {
			[Object.Category.UNIT] = function(_event)
				if _event.weapon ~= nil then
					local weapon = ATMEFct.C_Weapon_create(_event.weapon)
					local event = {
						type = "WEAPON",
						datas = {
							action = "SHOT",
							unitName = _event.initiator:getName(),
							weaponName = _event.weapon:getName(),
						},
					}
					
					ATMEFct.fireEvent(event)
					ATMEFct.outputCore(nil, "Event S_EVENT_SHOT généré pour " .. _event.initiator:getName(), 2)
				end
			end,
		},
		[world.event.S_EVENT_SHOOTING_START] = {
			[Object.Category.UNIT] = function(_event)
				if _event.target ~= nil then
					ATMEFct.shootTargetFct[_event.target:getCategory()](_event)
				else
					local event = {
						type = "WEAPON",
						datas = {
							action = "SHOOTING_START",
							unitName = _event.initiator:getName(),
						},
					}
				
					ATMEFct.fireEvent(event)
					
					ATMEFct.outputCore(nil, "Event S_EVENT_SHOOTING_START généré pour " .. _event.initiator:getName(), 2)
				end
			end,
		},
		[world.event.S_EVENT_SHOOTING_END] = {
			[Object.Category.UNIT] = function(_event)
				local event = {
					type = "WEAPON",
					datas = {
						action = "SHOOTING_END",
						unitName = _event.initiator:getName(),
					},
				}
				
				ATMEFct.fireEvent(event)
				ATMEFct.outputCore(nil, "Event S_EVENT_SHOOTING_END généré pour " .. _event.initiator:getName(), 2)
			end,
		},
		[world.event.S_EVENT_HIT] = {
			[Object.Category.UNIT] = function(_event)
				if _event.target ~= nil then
					if _event.weapon ~= nil then
						if ATMEDatas.weapons[_event.weapon:getName()] ~= nil and ATMEDatas.weaponsToDelete[_event.weapon:getName()] == nil then
							ATMEFct.hitTargetFct[_event.target:getCategory()](_event, _event.weapon:getName())
							
							-- Destruction de l'arme après 2 secondes au cas où S_EVENT_BDA à suivre
							timer.scheduleFunction(ATMEFct.C_Weapon_delete, _event.weapon:getName(), timer.getTime() + 2)
							ATMEDatas.weaponsToDelete[_event.weapon:getName()] = true
							
							ATMEFct.outputCore(nil, "Event S_EVENT_HIT généré pour " .. _event.initiator:getName(), 2)
						end
					else
						ATMEFct.hitTargetFct[_event.target:getCategory()](_event, nil)
						ATMEFct.outputCore(nil, "Event S_EVENT_HIT généré pour " .. _event.initiator:getName() .. " - sans arme associée", 2)
					end
				end
			end,
		},
		[world.event.S_EVENT_BDA] = {
			[Object.Category.UNIT] = function(_event)
				if _event.target ~= nil and _event.target:getCategory() == Object.Category.UNIT then
					if _event.weapon ~= nil then
						if ATMEDatas.weapons[_event.weapon:getName()] ~= nil then
							local event = {
								type = "WEAPON",
								datas = {
									action = "BDA_TARGET",
									unitName = _event.initiator:getName(),
									targetName =  _event.target:getName(),
									targetClassName = ATMEDatas.units[_event.target:getName()]:getClassName(),
									weaponName = _event.weapon:getName(),
								},
							}		
							
							ATMEFct.fireEvent(event)
							ATMEFct.C_Weapon_delete(_event.weapon:getName())
							ATMEFct.outputCore(nil, "Event S_EVENT_BDA généré pour " .. _event.initiator:getName(), 2)
						end
					else
						local event = {
								type = "WEAPON",
								datas = {
									action = "BDA_TARGET",
									unitName = _event.initiator:getName(),
									targetName =  _event.target:getName(),
									targetClassName = ATMEDatas.units[_event.target:getName()]:getClassName(),
								},
						}	
						
						ATMEFct.fireEvent(event)
						ATMEFct.outputCore(nil, "Event S_EVENT_BDA généré pour " .. _event.initiator:getName() .. " - sans arme associée", 2)
					end
				end
			end,
		},
		[world.event.S_EVENT_MISSION_START] = function(_event)
			-- Non pris en compte
		end,
		[world.event.S_EVENT_MISSION_END] = function(_event)
			for _, _nameModule in ipairs(ATMEDatas.modulesNames) do
				local localModuleMembers = ATMEDatas.modulesLocalMembers[_nameModule]
				ATMEDatas.currentModuleName = _nameModule
				if localModuleMembers.handlers.onEndMissionHandler ~= nil then
					ATMEFct.outputCore(nil,"Appel onEndMissionHandler du module " .. localModuleMembers.name, 2)
					xpcall(localModuleMembers.handlers.onEndMissionHandler, function() ATMEFct.errUserHandler('onEndMissionHandler') end)
				end
			end
			
			ATMEDatas.exitATME = true
			ATMEFct.outputCore(nil, "S_EVENT_MISSION_END Fin de Mission", 2)
		end,
		[world.event.S_EVENT_MARK_ADDED] = {
			[Object.Category.UNIT] = function(_event)
				local event = {
					type = "MARK",
					datas = {
						action = "ADD",
					},
				}		

				event.datas.mark = ATMEFct.C_Mark_create(_event.idx, _event.text, _event.pos)
				
				if _event.initiator ~= nil then
					event.datas.unitName = _event.initiator:getName()
				end
				
				if _event.coalition ~= 255 then
					event.datas.coalitionName = ATMEConst.coalitionNames[_event.coalition]
				end

				if _event.groupID ~= -1  then
					event.datas.linkedGroupName = ATMEDatas.playerNamesByGroupId[_event.groupID]
				end
				
				ATMEFct.fireEvent(event)
			end,
		},
		[world.event.S_EVENT_MARK_CHANGE] = {
			[Object.Category.UNIT] = function(_event)
				local event = {
					type = "MARK",
					datas = {
						action = "CHANGE",
					},
				}
				
				event.datas.mark = ATMEDatas.marks[_event.idx]
				local localMembers = ATMEDatas.localMembers[event.datas.mark:getIndex()]
				localMembers.text = _event.text
				localMembers.point.x = _event.pos.x
				localMembers.point.y = _event.pos.y
				localMembers.point.z = _event.pos.z
				
				if _event.initiator ~= nil then
					event.datas.unitName = _event.initiator:getName()
				end
				
				if _event.coalition ~= 255 then
					event.datas.coalitionName = ATMEConst.coalitionNames[_event.coalition]
				end
				
				if _event.groupID ~= -1 then
					event.datas.linkedGroupName = ATMEDatas.playerNamesByGroupId[_event.groupID]
				end
				
				ATMEFct.fireEvent(event)
			end,
		},
		[world.event.S_EVENT_MARK_REMOVED] = {
			[Object.Category.UNIT] = function(_event)
				local event = {
					type = "MARK",
					datas = {
						action = "REMOVE",
					},
				}
				
				event.datas.mark = ATMEDatas.marks[_event.idx]
				
				if _event.initiator ~= nil then
					event.datas.unitName = _event.initiator:getName()
				end
				
				if _event.coalition ~= 255 then
					event.datas.coalitionName = ATMEConst.coalitionNames[_event.coalition]
				end
				
				if _event.groupID ~= -1 then
					event.datas.linkedGroupName = ATMEDatas.playerNamesByGroupId[_event.groupID]
				end
				
				ATMEFct.fireEvent(event)
				
				ATMEFct.C_Mark_delete(_event.idx)
			end,
		},
	}
	
	
	function customEventHandler:onEvent(_event)
		if ATMEDatas.exitATME == true then
			world.removeEventHandler(customEventHandler)
		else
			if type(eventFunctions[_event.id]) == "table" then
				if _event.initiator ~= nil then
					if eventFunctions[_event.id][_event.initiator:getCategory()] ~= nil then
						eventFunctions[_event.id][_event.initiator:getCategory()](_event)
					else
						ATMEFct.outputCore(nil, "Evénement non traité catégorie " .. _event.initiator:getCategory() .. " - Type " .. _event.id, 2)
					end
				else
					ATMEFct.outputCore(nil, "Evénement non traité Initiator à nil - Type " .. _event.id, 2)
				end
			elseif type(eventFunctions[_event.id]) == "function" then
				eventFunctions[_event.id](_event)
			else
				ATMEFct.outputCore(nil, "Evénement inconnu ou non traité - Type " .. _event.id, 2)
			end
		end
	end
	
	----------------------------------------------------------------------------------------------------------
	-- Fonction d'initialisation des objets définis dans la mission
	-- Adapté de MIST - MIssion Scripting Tools
	-- Complétée par la recherche des FARP et AirFields nécessaires à la gestion des airbases.
	----------------------------------------------------------------------------------------------------------
	local function initObjectsInMission()
		for _coalitionName, _coalitionDatas in pairs(env.mission.coalition) do
			if (_coalitionName == 'red' or _coalitionName == 'blue') and type(_coalitionDatas) == 'table' then
				if _coalitionDatas.country then
					for _country_id, _countryDatas in pairs(_coalitionDatas.country) do
						for _typeName, _datas in pairs(_countryDatas) do
							if _typeName == "helicopter" or _typeName == "ship" or _typeName == "plane" or _typeName == "vehicle" or _typeName == "static" then	
								local isStatic = _typeName == "static"
								for _, _groupDatas in pairs(_datas.group) do
									if _groupDatas ~= nil and _groupDatas.units ~= nil then
										local currentGroupName = env.getValueDictByKey(_groupDatas.name)
										local d,f = string.find(currentGroupName, "|", 1)
										
										if d == nil then
											if isStatic == false then
												local uncontrolled = _groupDatas.uncontrolled
												local visible
												
												if _typeName == "ship" or _typeName == "vehicle" then
													uncontrolled = false
													if _groupDatas.lateActivation == nil or _groupDatas.lateActivation == false then
														visible = true
													else
														visible = _groupDatas.visible
													end
												else
													visible = _groupDatas.lateActivation == false
												end
												
												ATMEDatas.allGroupsInMission[currentGroupName] = {
													name = currentGroupName,
													country = _countryDatas.id,
													coalition = string.upper(_coalitionName),
													groupId = _groupDatas.groupId,								-- Uniquement défini pour les groupes initiaux.
													category = ATMEConst.groupCategories[_typeName],
													type = _typeName,
													uncontrolled = uncontrolled,
													units = {},
													visible = visible,
													taskSelected = true,
													lateActivation = _groupDatas.lateActivation,
													hidden = _groupDatas.hidden,
													start_time = _groupDatas.start_time,
													task = _groupDatas.task,
													tasks = _groupDatas.tasks,
													radioSet = _groupDatas.radioSet,
													modulation = _groupDatas.modulation,
													frequency = _groupDatas.frequency,
													communication = _groupDatas.communication,
													waypoints = {},
													waypointsName = {},
												}
												
												local nbPlayersFound = 0
												local nbAIFound = 0
												
												for _key, _unitDatas in ipairs(_groupDatas.units) do
													local unitName = env.getValueDictByKey(_unitDatas.name)
													local d,f = string.find(unitName, "|", 1)
										
													if d == nil then
														ATMEDatas.allGroupsInMission[currentGroupName].units[unitName] = {
															name = unitName,
															row = _key,
															type = _unitDatas.type,
															id = _unitDatas.unitId,
															livery_id = _unitDatas.livery_id,
															speed = _unitDatas.speed,
															psi = _unitDatas.psi,
															payload = _unitDatas.payload,
															skill = _unitDatas.skill,
															x = _unitDatas.x,
															y = _unitDatas.y,
															playerCanDrive = _unitDatas.playerCanDrive,
															heading = _unitDatas.heading,
															alt = _unitDatas.alt,
															alt_type = _unitDatas.alt_type,
															callsign = _unitDatas.callsign,
															onboard_num = _unitDatas.onboard_num,
															parking = _unitDatas.parking,
															parking_id = _unitDatas.parking_id,
														}
														
														ATMEDatas.allAIUnitsInMission[unitName] = {
															group = ATMEDatas.allGroupsInMission[currentGroupName],
															unit = ATMEDatas.allGroupsInMission[currentGroupName].units[unitName],
														}
														
														if _unitDatas.skill == "Player" or _unitDatas.skill == "Client" then
															ATMEDatas.reservedAsPlayers[unitName] = true
															nbPlayersFound = nbPlayersFound + 1
														else
															nbAIFound = nbAIFound + 1
														end
																					
														ATMEDatas.reservedUnitNames[unitName] = true
													else
														env.error("[ATME] Unit name '" .. unitName .. "' can't contain '|' character in Mission Editor\nATME stopped", true)
														return true
													end
												end
												
												ATMEDatas.reservedGroupNames[currentGroupName] = true
												
												if nbPlayersFound > 1 and nbAIFound > 0 then
													env.error("[ATME] Group name '" .. currentGroupName .. "' can't have more than one player with AI units\nATME stopped", true)
													return true
												end
												
												ATMEDatas.reservedAsPlayerGroups[currentGroupName] = nbPlayersFound
												
												if _groupDatas.route ~= nil and _groupDatas.route.points and #_groupDatas.route.points > 0 then
													for _pointNum, _point in ipairs(_groupDatas.route.points) do
														local helipadId
														
														if _point.helipadId ~= nil then
															-- Pb DCS : helipad est un string pour les heliports et les bateaux. Mais dans les WP c'est un entier
															-- Afin de retrouver l'airbase ad'hoc, conversion en string nécessaire pour éviter d'écraser un base de carte (airdrome est entier)
															helipadId = tostring(_point.helipadId)
														end
														
														local pointName =  env.getValueDictByKey(_point.name)
														
														-- Changement DCS 2.7, pointName peut être nil
														if pointName == nil then 
															pointName = ""
														end

														if pointName == "" or ATMEDatas.allGroupsInMission[currentGroupName].waypointsName[pointName] == nil then
															ATMEDatas.allGroupsInMission[currentGroupName].waypointsName[pointName] = #ATMEDatas.allGroupsInMission[currentGroupName].waypoints
															ATMEDatas.allGroupsInMission[currentGroupName].waypoints[#ATMEDatas.allGroupsInMission[currentGroupName].waypoints + 1] = {
																name = pointName,
																x = _point.x,
																y = _point.y,
																speed = _point.speed,
																speed_locked = _point.speed_locked,
																alt = _point.alt,
																alt_type = _point.alt_type,
																ETA = _point.ETA,
																ETA_locked = _point.ETA_locked,
																airdromeId = _point.airdromeId,
																helipadId = helipadId,
																type = _point.type,
																action = _point.action,
																task = _point.task,
																properties = _point.properties,
																formation_template = _point.formation_template,
															}
														else
															env.error("[ATME] Group name '" .. currentGroupName .. "' has more than one waypoint named '" .. pointName .. "' in Mission Editor\nATME stopped", true)
															return true
														end
													end
												end
											else 
												for _, _unitDatas in pairs(_groupDatas.units) do
													local unitName = env.getValueDictByKey(_unitDatas.name)
													local d,f = string.find(unitName, "|", 1)
										
													if d == nil then
														if _unitDatas.category == "GrassAirfields" or _unitDatas.category == "Heliports" then
															ATMEDatas.allStaticAirbasesInMission[unitName] = {
																name = unitName,
																type = _unitDatas.type,
																id = _unitDatas.unitId,
																x = _unitDatas.x,
																y = _unitDatas.y,
																canCargo = _unitDatas.canCargo,
																heading = _unitDatas.heading,
																category = _unitDatas.category,
																coalition = string.upper(_coalitionName),
																dead = _groupDatas.dead,
															}
														elseif _unitDatas.category == "Cargos" then
															ATMEDatas.allCargosInMission[unitName] = {
																name = unitName,
																country = country.name[_countryDatas.id],
																type = _unitDatas.type,
																x = _unitDatas.x,
																y = _unitDatas.y,
																canCargo = _unitDatas.canCargo,
																heading = _unitDatas.heading,
																category = _unitDatas.category,
																dead = _groupDatas.dead,
																shape_name = _unitDatas.shape_name,
																mass = _unitDatas.mass,
																rate =_unitDatas.rate,
																livery_id = _unitDatas.livery_id,
															}
														else
															-- Bug DCS : Attention les catégories de statics évoluent pour les objets terrestres pour ceux créés dans le ME.
															-- Pour le spawn, la catégorie reste "Unarmed"
															
															local category = _unitDatas.category
															if ATMEConst.backStaticObjectCategories[category] == nil then
																category = "Unarmed"
															end
		
															ATMEDatas.allStaticsInMission[unitName] = {
																name = unitName,
																country = country.name[_countryDatas.id],
																type = _unitDatas.type,
																x = _unitDatas.x,
																y = _unitDatas.y,
																canCargo = _unitDatas.canCargo,
																heading = _unitDatas.heading,
																category = category,
																dead = _groupDatas.dead,
																shape_name = _unitDatas.shape_name,
																mass = _unitDatas.mass,
																rate =_unitDatas.rate,
																livery_id = _unitDatas.livery_id,
															}
														end
													else
														env.error("[ATME] Static or Cargo unit name '" .. unitName .. "' can't contain '|' character in Mission Editor\nATME stopped", true)
														return true
													end
												end
											end
										else
											env.error("[ATME] Group name '" .. currentGroupName .. "' can't contain '|' character in Mission Editor\nATME stopped", true)
											return true
										end
									end
								end 
							end
						end
					end
				end
			end
		end
		
		return false
	end
	
	local function createSpawnedFromRadical(_table, _name)
		-- Recherche du radical pour suivi des incréments (copier/coller éditeur et futurs spawns)
		local d = string.find(_name, "%-[1-9]%d*$")
		local radical, startAt
		
		if d ~= nil then
			radical = _name:sub(1,d - 1)
			startAt = tonumber(_name:sub(d+1))
		else
		  radical = _name
		  startAt = 0
		end
		
		if _table[radical] == nil then
			_table[radical] = {
					max = startAt,
					freeIds = {},
			}
		else
			if startAt > _table[radical].max then
				_table[radical].max = startAt
			end
		end
	end
	
	local _timerStartMission
	_timerStartMission = function()
		for _id, _current in ipairs(world.getAirbases()) do
			if _current:hasAttribute("Airfields") == true then
				local id = _current:getID()
				ATMEFct.C_Airbase_create_basic("AIRBASE", false, _current, id)
			end
		end
	
		world.addEventHandler(customEventHandler)

		for _, _object in pairs(ATMEDatas.allStaticsInMission) do
			local objectName = _object.name
			local object = StaticObject.getByName(objectName)
			if object ~= nil then 
				-- Bug DCS : Certains objets comme M1 Barrage Balloon sont encore avec object = nil.
				-- Non pris en compte par ATME
				
				createSpawnedFromRadical(ATMEDatas.spawnedStatics, objectName)
				ATMEFct.C_Static_create(object, ATMEDatas.allStaticsInMission[objectName])
			else
				ATMEFct.outputCore(nil,"Static object " .. objectName .. " ignored by ATME : nil in run function", 2)
			end
		end
		
		for _, _object in pairs(ATMEDatas.allStaticAirbasesInMission) do
			local objectName = _object.name
			local object = Airbase.getByName(objectName)

			if object ~= nil then 
				-- Bug DCS : Certains objets comme M1 Barrage Balloon sont encore avec object = nil.
				-- Non pris en compte par ATME

				ATMEFct.C_Airbase_create_basic("HELIPAD", true, object)
			else
				ATMEFct.outputCore(nil,"Static object " .. objectName .. " ignored by ATME : nil in run function", 2)
			end
		end
	
		for _, _object in pairs(ATMEDatas.allCargosInMission) do
			local objectName = _object.name
			local object = StaticObject.getByName(objectName)	-- Les cargos sont les staticObject pour DCS
			
			if object ~= nil then 
				-- Bug DCS : Certains objets comme M1 Barrage Balloon sont encore avec object = nil.
				-- Non pris en compte par ATME
				
				createSpawnedFromRadical(ATMEDatas.spawnedCargos, objectName)
				ATMEFct.C_Cargo_create(object, ATMEDatas.allCargosInMission[objectName])
			else
				ATMEFct.outputCore(nil,"Cargo " .. objectName .. " ignored by ATME : nil in run function", 2)
			end
		end
	
		
		-- ATTENTION ... On fait une première passe pour créer les airbases liées aux unités bateaux portant des aéronefs
		
		for _name, _group in pairs(ATMEDatas.allGroupsInMission) do
			for _, _unit in pairs(_group.units) do
				local unit = Unit.getByName(_unit.name)
				if unit ~= nil then
					if unit:hasAttribute("AircraftCarrier") == true then
						ATMEFct.C_Airbase_create_basic("SHIP", false, unit)
					elseif unit:hasAttribute("HelicopterCarrier") == true then
						ATMEFct.C_Airbase_create_basic("SHIP", true, unit)
					end
				end
			end
		end
		
		-- Puis on crée toutes les unités
		-- On notera que les unités avec une activation retardées sont bien créées ! 
		
		for _name, _group in pairs(ATMEDatas.allGroupsInMission) do
			local nbUnits = 0
			
			_group.spawnDatas = {
				country = _group.country,
				category = _group.category,
				type = _group.type,
				unitsByName = {},
				waypointsName = _group.waypointsName,
				datas = {
					name = _name,
					groupId = _group.groupId,
					hidden = _group.hidden,
					task = _group.task,
					visible = false,
					--taskSelected = true,
					lateActivation = false,
					tasks = {},
					units = {},
					route = {
						points = _group.waypoints,
					},
				},
			}
			
			for _unitName, _unit in pairs(_group.units) do
				_group.spawnDatas.unitsByName[_unitName] = {
					type = _unit.type,
					name = _unitName,
					skill = _unit.skill,
					x = _unit.x,
					y = _unit.y,
					alt = _unit.alt,
					alt_type = _unit.alt_type,
					playerCanDrive = _unit.playerCanDrive,
					heading = _unit.heading,
					payload = _unit.payload,
					speed = _unit.speed,
					callsign = _unit.callsign,
					onboard_num = _unit.onboard_num,
					psi = _unit.psi,
					livery_id = _unit.livery_id,
				}
				
				_group.spawnDatas.datas.units[_unit.row] = _group.spawnDatas.unitsByName[_unitName]
				
				createSpawnedFromRadical(ATMEDatas.spawnedUnits, _unitName)
			end
			
			local datas = {
				uncontrolled = _group.uncontrolled,
				visible = _group.visible,
				spawnDatas = _group.spawnDatas,
			}

			for _, _unit in pairs(_group.units) do
				if _unit.skill ~= "Player" and _unit.skill ~= "Client" then
					local unit = Unit.getByName(_unit.name)
					if unit ~= nil and (_group.start_time == 0 or _group.lateActivation == true) then
						-- Bug DCS sur start_time qui peut rester > 0 (à 0 dans l'éditeur cependant) si late activation.
						-- Du coup on test le lateActivation
						
						ATMEFct.C_Unit_create(unit, datas, true, false)
					end
					
					ATMEDatas.allAIUnitsInMission[_unit.name].datas = datas
				else
					ATMEDatas.allAIUnitsInMission[_unit.name].datas = datas
				end
				
				nbUnits = nbUnits + 1
			end
			
			createSpawnedFromRadical(ATMEDatas.spawnedGroups, _name)
			
			if _group.lateActivation == true then
				local group = ATMEDatas.newCreatedGroups[_name].group
				local currentLocalMembers = ATMEDatas.localMembers[group:getIndex()]
				currentLocalMembers.isEnabled = false
				currentLocalMembers.nbUnitsToEnable = nbUnits
				ATMEFct.outputCore(nil,"Activation retardee pour " .. _name, 2)
			end
			
			_group.inMissionSpawnDatas = ATMEFct.copyTable(_group.spawnDatas)
		end
	
		world.addEventHandler(customEventHandler)

		ATMEDatas.isRunning = true

		_timerMainFunction()
		timer.scheduleFunction(_randomTimer, {}, timer.getTime() + 1)
	end
	
	----------------------------------------------------------------------------------------------------------
	-- Fonction de chargement de DLL
	-- La DLL doit être dans ..\Parties enregistrées\DCS.openbeta\_path (_path paramètre)
	-- et le fichier _DLLName.dll doit exister dans ce répertoire
	-- l'éditeur de mission par un script lua.
	----------------------------------------------------------------------------------------------------------
	ATME.loadDLL = function(_path, _DLLName)
		if lfs ~= nil and require ~= nil then
			local dllATMEModPath = lfs.writedir() .. _path
			package.cpath = package.cpath..";" .. dllATMEModPath .. "?.dll"
			ATMEDatas.ATMEDll = require(_DLLName)   -- loads the DLL
		else
			env.info("[ATME] - lfs or require : a nil value - Can't load DLL " .. _DLLName .. " -> Check MissionScripting.lua in DCS program scripts folder", true)
		end
	end

----------------------------------------------------------------------------------------------------------
	-- Fonction de démmarrage d'ATME à faire après le chargement de tous les modules. Cette fonction se met dans
	-- l'éditeur de mission par un script lua.
	----------------------------------------------------------------------------------------------------------
	ATME.run = function(_config)
		local fctDefinition = ATMEFct.getErrorPrefix() .. "run([language])"
		
		if ATMEDatas.exitATME == false then
			if _config == nil or type(_config) == "string" then
				if type(_config) == "string" then
					if ATMEConst.supportedLanguages[_config] ~= nil then
						ATMEDatas.language = _config
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "config formatted string first parameter must be FR (default), GER, RUS, EN", false)
						return
					end
				end
			elseif _config == nil then
				ATMEDatas.language = "FR"
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "config type must be nil or string", false)
				return
			end

			_timerStartMission()
			
			ATMEFct.outputCore(nil, "ATME is running -> " .. ATMEDatas.language, 2)
		end
	end
	
	----------------------------------------------------------------------------------------------------------
	-- Fonction MAIN d'initialisation générale appelée une seule fois au lancement du fichier script
	----------------------------------------------------------------------------------------------------------
	do
		env.info("[ATME] - Init starts - version " .. versionATME, false)
		
		ATMEDatas.exitATME = initObjectsInMission()
		ATMEDatas.theatreName = string.upper(env.mission.theatre)
		
		
		
		env.info("[ATME] - Init ok - Theatre : " .. ATMEDatas.theatreName, false)
		
		-- Fichier script dans Racine DCS\Scripts\MissionScripting.lua, mettre sanitizeModule('lfs') en commentaire
		-- pour utilisation du gestionnaire de fichiers lua. lfs est alors non nil
		
		if lfs ~= nil then
			ATMEDatas.files = {}
			ATMEDatas.ATMEDir = lfs.writedir() .."ATME"
			
			env.info("[ATME] - lfs ok - " .. ATMEDatas.ATMEDir, false)
		end
	end
end